Package: bioLeak
Type: Package
Title: Leakage-Safe Modeling and Auditing for Genomic and Clinical Data
Version: 0.3.8
Authors@R: c(
    person("Selcuk", "Korkmaz", role = c("aut","cre"), email = "selcukorkmaz@gmail.com",
           comment = c(ORCID = "0000-0003-4632-6850"))
    )
Description: Prevents and detects information leakage in biomedical machine learning.
  Provides leakage-resistant split policies (subject-grouped, batch-blocked, study leave-out, time-ordered),
  guarded preprocessing (train-only imputation, normalization, filtering, feature selection),
  cross-validated fitting with common learners, permutation-gap auditing, batch and fold association tests,
  and duplicate detection.
License: MIT + file LICENSE
URL: https://github.com/selcukorkmaz/bioLeak
BugReports: https://github.com/selcukorkmaz/bioLeak/issues
Encoding: UTF-8
Depends: R (>= 4.3)
Imports: digest, generics, methods, stats, utils, SummarizedExperiment,
        graphics, hardhat, parsnip
Suggests: BiocParallel, splitGraph, cli, dials, FNN, future,
        future.apply, ggplot2, glmnet, mice, missForest, pkgload,
        ranger, randomForest, recipes, RANN, rsample, tune, VIM, withr,
        workflows, xgboost, yardstick, pROC, PRROC, survival, knitr,
        rmarkdown, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
RoxygenNote: 7.3.3
Collate: 'classes.R' 'accessors.R' 'audit.R' 'audit_report.R'
        'benchmark_suite.R' 'ci_delta.R' 'conditions.R'
        'cv_uncertainty.R' 'cvauc_if.R' 'delta_lsi.R'
        'diagnostics_extra.R' 'fit_resample.R' 'globals.R'
        'guard_to_recipe.R' 'guards.R' 'imports.R' 'impute_guarded.R'
        'make_split_plan.R' 'permute_labels.R' 'plotting_audit.R'
        'provenance.R' 'simulate_suite.R' 'splitgraph_adapter.R'
        'summary.R' 'tidymodels.R' 'time_series_blocks.R'
        'tune_resample.R' 'utils_metrics.R' 'utils_print.R'
        'utils_se.R'
NeedsCompilation: no
Packaged: 2026-05-21 18:14:56 UTC; selcuk
Author: Selcuk Korkmaz [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-4632-6850>)
Maintainer: Selcuk Korkmaz <selcukorkmaz@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-21 18:40:02 UTC
Built: R 4.5.2; ; 2026-05-25 02:05:10 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as_rsample,LeakSplits)
S3method(predict,GuardFit)
S3method(print,GuardFit)
S3method(print,LeakAuditList)
S3method(print,LeakTune)
S3method(print,summary.GuardFit)
S3method(summary,GuardFit)
S3method(summary,LeakAudit)
S3method(summary,LeakDeltaLSI)
S3method(summary,LeakFit)
S3method(summary,LeakTune)
export(as_leaksplits)
export(as_rsample)
export(audit_batch_assoc)
export(audit_duplicates)
export(audit_info)
export(audit_leakage)
export(audit_leakage_by_learner)
export(audit_perm_gap)
export(audit_report)
export(audit_target_assoc)
export(benchmark_leakage_suite)
export(calibration_summary)
export(check_split_overlap)
export(confounder_sensitivity)
export(cv_ci)
export(delta_lsi)
export(dlsi_R_eff)
export(dlsi_ci)
export(dlsi_metric)
export(dlsi_p_value)
export(dlsi_repeats)
export(dlsi_robust)
export(dlsi_tier)
export(fit_metrics)
export(fit_resample)
export(guard_ensure_levels)
export(guard_fit)
export(guard_to_recipe)
export(impute_guarded)
export(make_split_plan)
export(plot_calibration)
export(plot_confounder_sensitivity)
export(plot_dlsi_repeats)
export(plot_fold_balance)
export(plot_overlap_checks)
export(plot_perm_distribution)
export(plot_time_acf)
export(predict_guard)
export(simulate_leakage_suite)
export(tune_resample)
exportClasses(LeakAudit)
exportClasses(LeakDeltaLSI)
exportClasses(LeakFit)
exportClasses(LeakSplits)
exportMethods(audit_batch_assoc)
exportMethods(audit_duplicates)
exportMethods(audit_info)
exportMethods(audit_perm_gap)
exportMethods(audit_target_assoc)
exportMethods(dlsi_R_eff)
exportMethods(dlsi_ci)
exportMethods(dlsi_metric)
exportMethods(dlsi_p_value)
exportMethods(dlsi_repeats)
exportMethods(dlsi_robust)
exportMethods(dlsi_tier)
exportMethods(fit_metrics)
exportMethods(plot)
exportMethods(show)
importFrom(methods,new)
importFrom(stats,aggregate)
importFrom(stats,ave)
importFrom(stats,predict)
importFrom(stats,rbinom)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(utils,head)
importFrom(utils,modifyList)
importMethodsFrom(methods,show)

<img src="man/figures/bioLeak_hex.png" align="right" width="95"/>

# bioLeak: Leakage-Safe Modeling and Auditing for Genomic and Clinical Data

<!-- badges: start -->
[![CRAN status](https://www.r-pkg.org/badges/version/bioLeak)](https://CRAN.R-project.org/package=bioLeak)
[![R-CMD-check](https://github.com/selcukorkmaz/bioLeak/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/selcukorkmaz/bioLeak/actions/workflows/R-CMD-check.yaml)
[![CRAN downloads](https://cranlogs.r-pkg.org/badges/grand-total/bioLeak)](https://CRAN.R-project.org/package=bioLeak)
[![License: MIT](https://img.shields.io/badge/License-MIT-blue.svg)](https://opensource.org/licenses/MIT)
[![Lifecycle: stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
<!-- badges: end -->

`bioLeak` is an R package for detecting, quantifying, and diagnosing data leakage in biomedical machine-learning workflows. It provides leakage-resistant resampling, guarded preprocessing, post-hoc auditing, and inference tools for cross-validation and related evaluation settings.

## Purpose and scope

**In scope:**
- **Preprocessing leakage:** global imputation, scaling, filtering, or feature selection applied before resampling.
- **Dependence leakage:** repeated measures, subject-level grouping, batch/site/study effects, and near-duplicate samples.
- **Resampling violations:** group overlap, study holdout, time-ordered evaluation, and multi-axis split constraints.
- **Diagnostic evidence:** permutation-based performance gaps, batch/fold association tests, target leakage scans, duplicate detection, and mechanism-level risk summaries.

**Out of scope:**
- Proving the absence of leakage or guaranteeing unbiased performance.
- Production deployment tooling.
- Unsupervised learning.
- Broad, non-leakage-oriented data-quality diagnostics.

## Why bioLeak is needed

Standard cross-validation assumes independent samples and exchangeable labels. Biomedical datasets often violate these assumptions due to repeated measures, site effects, batch structure, and temporal dependence. These violations can inflate performance metrics even when a model does not generalize. `bioLeak` enforces leakage-aware resampling and provides post-hoc diagnostics that estimate how much apparent performance could be driven by leakage or confounding.

## Core functionality

### Splitting and validation

| Function | Description |
|---|---|
| `make_split_plan()` | Leakage-aware splits: subject-grouped, batch-blocked, study leave-out, time-series, and N-axis combined modes with compact storage option and constraint-aware train/test exclusion |
| `check_split_overlap()` | Explicit overlap-invariant validation across declared grouping axes |
| `as_rsample()` | Convert `LeakSplits` to an `rsample` rset for tidymodels interoperability |

### Guarded preprocessing and modeling

| Function | Description |
|---|---|
| `fit_resample()` | Cross-validated fitting with train-only imputation, normalization, filtering, and feature selection; supports binomial, multiclass, regression, and survival tasks; parallel execution via `future.apply` |
| `tune_resample()` | Nested hyperparameter tuning via tidymodels `tune`/`dials` with leakage-aware outer splits, hyperparameter aggregation across folds, and optional threshold tuning; survival tasks are not yet supported |
| `impute_guarded()` | Standalone train-only imputation (median, knn, missForest, none) |
| `guard_to_recipe()` | Convert guarded preprocessing specifications to `recipes` pipelines |

### Auditing and diagnostics

| Function | Description |
|---|---|
| `audit_leakage()` | Permutation gap test, batch/study association tests, univariate and multivariate target leakage scans, near-duplicate detection, and mechanism-level risk summaries |
| `audit_leakage_by_learner()` | Multi-learner auditing for multi-model fits |
| `audit_report()` | Self-contained HTML summary of audit results |
| `calibration_summary()` | Probability calibration checks |
| `confounder_sensitivity()` | Sensitivity analysis for confounding effects |

### Inference

| Function | Description |
|---|---|
| `delta_lsi()` | Leakage sensitivity index (ΔLSI) with Huber M-estimation, BCa confidence intervals, sign-flip inference, and blocked exchangeability for time-series designs |
| `cv_ci()` | Cross-validation confidence intervals with Nadeau-Bengio correction |

### Simulation and benchmarking

| Function | Description |
|---|---|
| `simulate_leakage_suite()` | Generate controlled leakage scenarios for benchmarking audit sensitivity |
| `benchmark_leakage_suite()` | Reproducible modality-by-mechanism benchmark grids with detection-rate summaries |

### Visualization

`plot_calibration()`, `plot_confounder_sensitivity()`, `plot_fold_balance()`, `plot_overlap_checks()`, `plot_perm_distribution()`, `plot_time_acf()`

## Installation

Requires R >= 4.3.

From CRAN:

```r
install.packages("bioLeak")
```

Development version from GitHub:

```r
install.packages("remotes")
remotes::install_github("selcukorkmaz/bioLeak")
```

Non-obvious dependencies:
- `SummarizedExperiment` and `BiocGenerics` are Bioconductor packages (installed automatically by `remotes`, but can be installed manually with `BiocManager::install()` if needed).
- Optional packages enable specific features: `glmnet`, `ranger`, `pROC`, `PRROC`, `survival`, `future.apply`, `RANN`, `rmarkdown`, `tune`, `dials`.

## Minimal working example

```r
library(bioLeak)

set.seed(1)
n_subject <- 40
rep_per_subject <- 3
n <- n_subject * rep_per_subject

subject <- rep(seq_len(n_subject), each = rep_per_subject)
batch <- rep(seq_len(6), length.out = n)

# Subject-level latent risk creates dependence across repeated measures
subj_risk <- rnorm(n_subject, sd = 1)
x1 <- subj_risk[subject] + rnorm(n, sd = 0.5)
x2 <- rnorm(n)
x3 <- rnorm(n)
p_subj <- stats::plogis(1.5 * subj_risk)
outcome <- factor(ifelse(runif(n) < p_subj[subject], "case", "control"),
                  levels = c("control", "case"))

df <- data.frame(subject, batch, outcome, x1, x2, x3)

# Leakage-aware splits (subjects do not cross folds)
splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "subject",
  v = 5,
  stratify = TRUE,
  seed = 1
)

# Guarded pipeline (train-only preprocessing)
spec <- parsnip::logistic_reg(mode = "classification") |>
  parsnip::set_engine("glm")

fit_guarded <- fit_resample(
  df,
  outcome = "outcome",
  splits = splits,
  learner = spec,
  metrics = "auc",
  preprocess = list(
    impute = list(method = "median"),
    normalize = list(method = "zscore"),
    filter = list(var_thresh = 0),
    fs = list(method = "none")
  ),
  refit = FALSE,
  seed = 1
)

# Leaky comparator: add a leakage feature computed on the full dataset
df_leaky <- within(df, {
  leak_subject <- ave(as.numeric(outcome == "case"), subject, FUN = mean)
})

fit_leaky <- fit_resample(
  df_leaky,
  outcome = "outcome",
  splits = splits,
  learner = spec,
  metrics = "auc",
  preprocess = list(
    impute = list(method = "none"),
    normalize = list(method = "none"),
    filter = list(var_thresh = 0),
    fs = list(method = "none")
  ),
  refit = FALSE,
  seed = 1
)

# Use unstratified permutations here to avoid warnings in small grouped data
audit_guarded <- audit_leakage(
  fit_guarded,
  metric = "auc",
  B = 30,
  perm_stratify = FALSE,
  X_ref = df[, c("x1", "x2", "x3")]
)

audit_leaky <- audit_leakage(
  fit_leaky,
  metric = "auc",
  B = 30,
  perm_stratify = FALSE,
  X_ref = df_leaky[, c("x1", "x2", "x3", "leak_subject")]
)

summary(fit_guarded)
summary(fit_leaky)
summary(audit_guarded)
summary(audit_leaky)
```

Interpretation notes:
- If the leaky comparator shows higher AUC and `leak_subject` ranks near the top of the target leakage scan, the performance gap is likely inflated by leakage.
- Similar guarded and leaky results do not prove the absence of leakage; they only reduce specific risks tested by the audit.

## Tidymodels interoperability

`bioLeak` integrates with the tidymodels ecosystem at multiple levels:

- **Splits:** `fit_resample()` and `tune_resample()` accept `rsample` rset/rsplit objects as `splits`. `as_rsample()` converts `LeakSplits` to an `rsample` rset.
- **Preprocessing:** `preprocess` can be a `recipes::recipe` (prepped on training folds, baked on test folds). `guard_to_recipe()` converts guarded preprocessing specifications into a `recipes` pipeline.
- **Learners:** `learner` accepts `parsnip::model_spec` or `workflows::workflow` objects.
- **Metrics:** `metrics` accepts `yardstick::metric_set` objects.

Note: When using recipes/workflows, the built-in guarded preprocessing list is not applied; ensure your recipe is leakage-safe.

Example (rsample + recipes + yardstick):
```r
if (requireNamespace("rsample", quietly = TRUE) &&
    requireNamespace("recipes", quietly = TRUE) &&
    requireNamespace("yardstick", quietly = TRUE)) {
  rs <- rsample::vfold_cv(df, v = 5)
  rec <- recipes::recipe(outcome ~ ., data = df) |>
    recipes::step_normalize(recipes::all_numeric_predictors())
  ys <- yardstick::metric_set(yardstick::roc_auc, yardstick::accuracy)

  fit_rs <- fit_resample(
    df,
    outcome = "outcome",
    splits = rs,
    learner = spec,
    preprocess = rec,
    metrics = ys,
    refit = FALSE
  )
}
```

## Supported tasks and learners

`bioLeak` supports four task types:
- **Binomial classification**: built-in character metrics are `auc`, `pr_auc`, and `accuracy`; additional metrics can be supplied via `yardstick::metric_set` when supported by the learner outputs.
- **Multiclass classification**: built-in character metrics are `accuracy`, `macro_f1`, and `log_loss`.
- **Regression**: built-in character metrics are `rmse` and `cindex`.
- **Survival analysis**: built-in character metric is `cindex` in `fit_resample()`.

The `learner` argument accepts parsnip model specs, workflows, or built-in learner strings (`"glmnet"`, `"ranger"`). Models such as base R `glm` or `xgboost` can still be used through parsnip/workflows or custom learners, but they are not built-in character learner names.

Survival outcomes are supported in `fit_resample()`, but support is less complete across the package. In particular, `tune_resample()` does not yet support survival tasks.

## Methodological assumptions

- The split mode matches the true dependence structure (subject, batch, study, or time).
- Leakage is inferred from performance gaps and diagnostic signals, not proven or ruled out.
- Permutation gaps assume the chosen resampling scheme reflects the intended evaluation setting.
- By default, permutation gaps use refit-based permutations when refit data are available (auto mode), falling back to fixed-prediction shuffles when they are not.
- Target leakage scans include univariate associations plus a multivariate/interaction check by default for supported tasks; proxies outside `X_ref` can still pass undetected.

## Interpretation guidance

- **Permutation gap:** large positive gaps indicate non-random signal; they do not by themselves indicate or refute leakage.
- **Batch/study association warnings** indicate that folds align with metadata; this can reflect leakage or study design constraints.
- **Target leakage flags** identify features overly aligned with the outcome; inspect data provenance before removing them. When enabled, FDR-adjusted columns (`p_value_adj`, `flag_fdr`) provide a more conservative screen.
- **Duplicate detection** flags near-identical samples across train/test by default; use `duplicate_scope = "all"` to include within-fold duplicates and review for data-quality issues.
- **Mechanism risk summary** (`audit@info$mechanism_summary`) provides a compact mechanism-level risk view across permutation, confounding, target-proxy, duplicate, and temporal signals.

Common misinterpretations:
- "Non-significant permutation test means no leakage": false.
- "High AUC implies good generalization": false if resampling is violated.
- "No flagged features means no leakage": false; audits are limited to available metadata and `X_ref`.

## Intended audience

Biomedical ML researchers, biostatisticians, and methodologists reviewing cross-validation and leakage risk.

## Citation

```r
citation("bioLeak")
```

When reporting results, include: split mode, grouping columns, random seeds, preprocessing steps, learner specification, metrics, and audit settings (B, target threshold, similarity method). Include both guarded and leaky comparator results when used.

## License

MIT

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE,
  eval = TRUE
)

## ----example-data-------------------------------------------------------------
library(bioLeak)

set.seed(123)
n <- 160
subject <- rep(seq_len(40), each = 4)
batch <- sample(paste0("B", 1:6), n, replace = TRUE)
study <- sample(paste0("S", 1:4), n, replace = TRUE)
time <- seq_len(n)

x1 <- rnorm(n)
x2 <- rnorm(n)
x3 <- rnorm(n)
linpred <- 0.7 * x1 - 0.4 * x2 + 0.2 * x3 + rnorm(n, sd = 0.5)
p <- stats::plogis(linpred)
outcome <- factor(ifelse(runif(n) < p, "case", "control"),
                  levels = c("control", "case"))

df <- data.frame(
  subject = subject,
  batch = batch,
  study = study,
  time = time,
  outcome = outcome,
  x1 = x1,
  x2 = x2,
  x3 = x3
)

df_leaky <- within(df, {
  leak_subject <- ave(as.numeric(outcome == "case"), subject, FUN = mean)
  leak_batch <- ave(as.numeric(outcome == "case"), batch, FUN = mean)
  leak_global <- mean(as.numeric(outcome == "case"))
})

df_time <- df
df_time$leak_future <- c(as.numeric(df_time$outcome == "case")[-1], 0)
predictors <- c("x1", "x2", "x3")


# Example data (first 6 rows)
head(df)

# Outcome class counts
as.data.frame(table(df$outcome))

## ----leaky-splits-------------------------------------------------------------
leaky_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "row_id",
  v = 5,
  repeats = 1,
  stratify = TRUE
)

cat("Leaky splits summary (sample-wise CV):\n")
leaky_splits

## ----safe-splits--------------------------------------------------------------
safe_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "subject",
  v = 5,
  repeats = 1,
  stratify = TRUE,
  seed = 10
)

cat("Leakage-safe splits summary (subject-grouped CV):\n")
safe_splits

## ----splits-other-modes-------------------------------------------------------
batch_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "batch_blocked",
  batch = "batch",
  v = 4,
  stratify = TRUE
)

cat("Batch-blocked splits summary:\n")
batch_splits

study_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "study_loocv",
  study = "study"
)

cat("Study leave-one-out splits summary:\n")
study_splits

time_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "time_series",
  time = "time",
  v = 4,
  horizon = 2
)

cat("Time-series splits summary:\n")
time_splits

nested_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "subject",
  v = 3,
  nested = TRUE,
  stratify = TRUE
)

cat("Nested CV splits summary:\n")
nested_splits

## ----combined-splits----------------------------------------------------------
# For combined mode, constraint-axis levels should not span all folds.
# Here each subject belongs to exactly one site, so site exclusion only
# removes training samples from the same site as the test subjects.
df_comb <- df
df_comb$site <- paste0("site", rep(1:8, each = 5)[df_comb$subject])

combined_splits <- make_split_plan(
  df_comb,
  outcome = "outcome",
  mode = "combined",
  constraints = list(
    list(type = "subject", col = "subject"),
    list(type = "batch",   col = "site")
  ),
  v = 4,
  stratify = TRUE,
  seed = 42
)

cat("Combined (N-axis) splits summary:\n")
combined_splits

## ----combined-overlap-check---------------------------------------------------
# Verify zero overlap on all constraint axes
overlap_result <- check_split_overlap(combined_splits)
overlap_result

## ----check-split-overlap------------------------------------------------------
# Run on the subject-grouped splits from earlier
overlap_safe <- check_split_overlap(safe_splits)
overlap_safe

## ----compact-splits-----------------------------------------------------------
# Efficient storage for large N
big_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "subject",
  v = 5,
  compact = TRUE  # <--- Saves memory
)

cat("Compact-mode splits summary:\n")
big_splits
cat(sprintf("Compact storage enabled: %s\n", big_splits@info$compact))

## ----strict-mode-demo, error=TRUE---------------------------------------------
try({
# Enable strict mode temporarily
withr::with_options(list(bioLeak.strict = TRUE), {
  strict_splits <- make_split_plan(
    df,
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 5,
    stratify = TRUE,
    seed = 42
  )
  cat("Strict mode splits completed without error.\n")
})
})

## ----strict-mode-trained-recipe, error=TRUE-----------------------------------
try({
# Strict mode catches a pre-trained recipe
if (requireNamespace("recipes", quietly = TRUE)) {
  rec <- recipes::recipe(outcome ~ ., data = df[, c("outcome", predictors)]) |>
    recipes::step_normalize(recipes::all_numeric_predictors()) |>
    recipes::prep(training = df[, c("outcome", predictors)])

  withr::with_options(list(bioLeak.strict = TRUE), {
    tryCatch(
      fit_resample(
        df, outcome = "outcome",
        splits = safe_splits,
        learner = "glmnet",
        preprocess = rec
      ),
      error = function(e) cat("Strict mode error:", conditionMessage(e), "\n")
    )
  })
}
})

## ----leaky-scaling------------------------------------------------------------
df_leaky_scaled <- df
df_leaky_scaled[predictors] <- scale(df_leaky_scaled[predictors])
scaled_summary <- data.frame(
  feature = predictors,
  mean = colMeans(df_leaky_scaled[predictors]),
  sd = apply(df_leaky_scaled[predictors], 2, stats::sd)
)
scaled_summary$mean <- round(scaled_summary$mean, 3)
scaled_summary$sd <- round(scaled_summary$sd, 3)

# Leaky global scaling: means ~0 and SDs ~1 computed on all samples
scaled_summary

## ----guarded-preprocess-------------------------------------------------------
fold1 <- safe_splits@indices[[1]]
train_x <- df[fold1$train, predictors]
test_x <- df[fold1$test, predictors]

guard <- guard_fit(
  X = train_x,
  y = df$outcome[fold1$train],
  steps = list(
    impute = list(method = "median", winsor = TRUE),
    normalize = list(method = "zscore"),
    filter = list(var_thresh = 0, iqr_thresh = 0),
    fs = list(method = "none")
  ),
  task = "binomial"
)

train_x_guarded <- predict(guard, train_x)
test_x_guarded <- predict(guard, test_x)

cat("GuardFit object:\n")
guard
cat("\nGuardFit summary:\n")
summary(guard)

# Guarded training data (first 6 rows)
head(train_x_guarded)

# Guarded test data (first 6 rows)
head(test_x_guarded)

## ----guard-ensure-levels------------------------------------------------------
raw_levels <- data.frame(
  site = c("A", "B", "B"),
  status = c("yes", "no", "yes"),
  stringsAsFactors = FALSE
)

level_state <- guard_ensure_levels(raw_levels)

# Aligned factor data with consistent levels
level_state$data

# Levels map
level_state$levels

## ----leaky-impute-------------------------------------------------------------
train <- data.frame(a = c(1, 2, NA, 4), b = c(NA, 1, 1, 0))
test <- data.frame(a = c(NA, 5), b = c(1, NA))

all_median <- vapply(rbind(train, test),
                     function(col) median(col, na.rm = TRUE),
                     numeric(1))
train_leaky <- as.data.frame(Map(function(col, m) { col[is.na(col)] <- m; col },
                                 train, all_median))
test_leaky <- as.data.frame(Map(function(col, m) { col[is.na(col)] <- m; col },
                                test, all_median))

# Leaky medians computed on train + test
data.frame(feature = names(all_median), median = all_median)

# Leaky-imputed training data
train_leaky

# Leaky-imputed test data
test_leaky

## ----safe-impute--------------------------------------------------------------
imp <- impute_guarded(
  train = train,
  test = test,
  method = "median",
  winsor = FALSE
)

# Guarded-imputed training data
imp$train

# Guarded-imputed test data
imp$test

## ----guard-to-recipe----------------------------------------------------------
if (requireNamespace("recipes", quietly = TRUE)) {
  guard_steps <- list(
    impute    = list(method = "median"),
    normalize = list(method = "zscore")
  )

  rec <- guard_to_recipe(
    steps = guard_steps,
    formula = outcome ~ .,
    training_data = df[, c("outcome", predictors)]
  )

  cat("Converted recipe:\n")
  rec
}

## ----parsnip-spec-------------------------------------------------------------
spec <- parsnip::logistic_reg(mode = "classification") |>
  parsnip::set_engine("glm")

## ----fit-leaky----------------------------------------------------------------
fit_leaky <- fit_resample(
  df_leaky,
  outcome = "outcome",
  splits = leaky_splits,
  learner = spec,
  metrics = c("auc", "accuracy"),
  preprocess = list(
    impute = list(method = "median"),
    normalize = list(method = "zscore"),
    filter = list(var_thresh = 0),
    fs = list(method = "none")
  )
)

cat("Leaky fit summary:\n")
summary(fit_leaky)
metrics_leaky <- as.data.frame(fit_leaky@metric_summary)
num_cols <- vapply(metrics_leaky, is.numeric, logical(1))
metrics_leaky[num_cols] <- lapply(metrics_leaky[num_cols], round, digits = 3)

# Leaky fit: mean and SD of metrics across folds
metrics_leaky

## ----fit-safe-----------------------------------------------------------------
fit_safe <- fit_resample(
  df,
  outcome = "outcome",
  splits = safe_splits,
  learner = spec,
  metrics = c("auc", "accuracy"),
  preprocess = list(
    impute = list(method = "median"),
    normalize = list(method = "zscore"),
    filter = list(var_thresh = 0),
    fs = list(method = "none")
  ),
  positive_class = "case",
  class_weights = c(control = 1, case = 1),
  refit = TRUE
)

cat("Leakage-safe fit summary:\n")
summary(fit_safe)
metrics_safe <- as.data.frame(fit_safe@metric_summary)
num_cols <- vapply(metrics_safe, is.numeric, logical(1))
metrics_safe[num_cols] <- lapply(metrics_safe[num_cols], round, digits = 3)

# Leakage-safe fit: mean and SD of metrics across folds
metrics_safe

# Per-fold metrics (first 6 rows)
head(fit_metrics(fit_safe))

## ----fit-fold-status----------------------------------------------------------
# Fold-level diagnostics for reproducible troubleshooting
head(fit_safe@info$fold_status, 5)

## ----multiclass-example-------------------------------------------------------
if (requireNamespace("ranger", quietly = TRUE)) {
  set.seed(11)
  df_multi <- df
  df_multi$outcome3 <- factor(sample(c("A", "B", "C"),
                                     nrow(df_multi), replace = TRUE))

  multi_splits <- make_split_plan(
    df_multi,
    outcome = "outcome3",
    mode = "subject_grouped",
    group = "subject",
    v = 4,
    stratify = TRUE,
    seed = 11
  )

  fit_multi <- fit_resample(
    df_multi,
    outcome = "outcome3",
    splits = multi_splits,
    learner = "ranger",
    metrics = c("accuracy", "macro_f1", "log_loss"),
    refit = FALSE
  )

  cat("Multiclass fit summary:\n")
  summary(fit_multi)
} else {
  cat("ranger not installed; skipping multiclass example.\n")
}

## ----survival-example---------------------------------------------------------
if (requireNamespace("survival", quietly = TRUE) &&
    requireNamespace("glmnet", quietly = TRUE)) {
  set.seed(12)
  df_surv <- df
  df_surv$time_to_event <- rexp(nrow(df_surv), rate = 0.1)
  df_surv$event <- rbinom(nrow(df_surv), 1, 0.7)

  surv_splits <- make_split_plan(
    df_surv,
    outcome = c("time_to_event", "event"),
    mode = "subject_grouped",
    group = "subject",
    v = 4,
    stratify = FALSE,
    seed = 12
  )

  fit_surv <- fit_resample(
    df_surv,
    outcome = c("time_to_event", "event"),
    splits = surv_splits,
    learner = "glmnet",
    metrics = "cindex",
    refit = FALSE
  )

  cat("Survival fit summary:\n")
  summary(fit_surv)
} else {
  cat("survival or glmnet not installed; skipping survival example.\n")
}

## ----learner-args-------------------------------------------------------------
if (requireNamespace("glmnet", quietly = TRUE)) {
  fit_glmnet <- fit_resample(
    df,
    outcome = "outcome",
    splits = safe_splits,
    learner = "glmnet",
    metrics = "auc",
    learner_args = list(glmnet = list(alpha = 0.5)),
    preprocess = list(
      impute = list(method = "median"),
      normalize = list(method = "zscore"),
      filter = list(var_thresh = 0),
      fs = list(method = "none")
    )
  )
  cat("GLMNET summary with learner-specific arguments:\n")
  summary(fit_glmnet)
} else {
  cat("glmnet not installed; skipping learner_args example.\n")
}

## ----se-example---------------------------------------------------------------
if (requireNamespace("SummarizedExperiment", quietly = TRUE)) {
  se <- SummarizedExperiment::SummarizedExperiment(
    assays = list(counts = t(as.matrix(df[, predictors]))),
    colData = df[, c("subject", "batch", "study", "time", "outcome"), drop = FALSE]
  )

  se_splits <- make_split_plan(
    se,
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 3
  )

  se_fit <- fit_resample(
    se,
    outcome = "outcome",
    splits = se_splits,
    learner = spec,
    metrics = "auc"
  )
  cat("SummarizedExperiment fit summary:\n")
  summary(se_fit)
} else {
  cat("SummarizedExperiment not installed; skipping SE example.\n")
}

## ----tidymodels-interop, eval = requireNamespace("recipes", quietly = TRUE) && requireNamespace("yardstick", quietly = TRUE)----
## This chunk requires the recipes and yardstick packages, which are
## listed in DESCRIPTION's Suggests rather than Imports. The eval=
## option above gates the chunk on their availability, so the vignette
## still builds when those packages are absent. Users copying this
## code into their own R session must have both packages installed
## (e.g., install.packages(c("recipes", "yardstick"))).
library(bioLeak)
library(parsnip)
library(recipes)
library(yardstick)

set.seed(123)
N <- 60
df <- data.frame(
  subject = factor(rep(paste0("S", 1:20), length.out = N)), 
  outcome = factor(rep(c("ClassA", "ClassB"), length.out = N)),
  x1 = rnorm(N),
  x2 = rnorm(N),
  x3 = rnorm(N)
)

spec <- logistic_reg() |> set_engine("glm")

# Use bioLeak's native split planner to avoid conversion errors
set.seed(13)

# Use make_split_plan instead of rsample::group_vfold_cv
# This creates a subject-grouped CV directly compatible with fit_resample
splits <- make_split_plan(
  df, 
  outcome = "outcome", 
  mode = "subject_grouped", 
  group = "subject", 
  v = 3
)

rec <- recipes::recipe(outcome ~ x1 + x2 + x3, data = df) |>
  recipes::step_impute_median(recipes::all_numeric_predictors()) |>
  recipes::step_normalize(recipes::all_numeric_predictors())

metrics_set <- yardstick::metric_set(yardstick::roc_auc, yardstick::accuracy)

fit_rs <- fit_resample(
  df,
  outcome = "outcome",
  splits = splits,
  learner = spec,
  preprocess = rec,
  metrics = metrics_set,
  refit = FALSE
)

if (exists("as_rsample", where = asNamespace("bioLeak"), mode = "function")) {
    rs_export <- as_rsample(fit_rs@splits, data = df)
    print(rs_export)
}

## ----workflow-example---------------------------------------------------------
if (requireNamespace("workflows", quietly = TRUE)) {
  wf <- workflows::workflow() |>
    workflows::add_model(spec) |>
    workflows::add_formula(outcome ~ x1 + x2 + x3)

  fit_wf <- fit_resample(
    df,
    outcome = "outcome",
    splits = splits,
    learner = wf,
    metrics = "auc",
    refit = FALSE
  )

  cat("Workflow fit summary:\n")
  summary(fit_wf)
} else {
  cat("workflows not installed; skipping workflow example.\n")
}

## ----tune-example-------------------------------------------------------------
if (requireNamespace("parsnip", quietly = TRUE) &&
    requireNamespace("recipes", quietly = TRUE) &&
    requireNamespace("tune", quietly = TRUE) &&
    requireNamespace("glmnet", quietly = TRUE)) {
  # --- 1. Create Data ---
  set.seed(123)
  N <- 60
  df <- data.frame(
    subject = factor(rep(paste0("S", 1:20), length.out = N)),
    outcome = factor(sample(c("ClassA", "ClassB"), N, replace = TRUE)),
    x1 = rnorm(N),
    x2 = rnorm(N),
    x3 = rnorm(N)
  )

  # --- 2. Generate Nested Splits ---
  set.seed(1)
  nested_splits <- make_split_plan(
    df,
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 3,
    nested = TRUE
  )

  # --- 3. Define Recipe & Model ---
  rec <- recipes::recipe(outcome ~ x1 + x2 + x3, data = df) |>
    recipes::step_impute_median(recipes::all_numeric_predictors()) |>
    recipes::step_normalize(recipes::all_numeric_predictors())

  spec_tune <- parsnip::logistic_reg(
    penalty = tune::tune(),
    mixture = 1,
    mode = "classification"
  ) |>
    parsnip::set_engine("glmnet")

  # --- 4. Run Tuning ---
  tuned <- tune_resample(
    df,
    outcome = "outcome",
    splits = nested_splits,
    learner = spec_tune,
    preprocess = rec,
    inner_v = 2,
    grid = 3,
    metrics = c("auc", "accuracy"),
    selection = "one_std_err",
    refit = TRUE,
    seed = 14
  )

  summary(tuned)
} else {
  cat("parsnip/recipes/tune/glmnet not installed; skipping nested tuning example.\n")
}

## ----tune-diagnostics---------------------------------------------------------
if (exists("tuned")) {
  # Fold-level status from the outer loop
  head(tuned$fold_status, 5)

  # Final tuned model is available when refit = TRUE
  is.null(tuned$final_model)
} else {
  cat("Nested tuning object not available (dependencies missing).\n")
}

## ----fit-xgboost--------------------------------------------------------------
if (requireNamespace("parsnip", quietly = TRUE) &&
    requireNamespace("xgboost", quietly = TRUE) &&
    requireNamespace("recipes", quietly = TRUE)) {
  
  # 1. Define the model spec
  xgb_spec <- parsnip::boost_tree(
    mode = "classification",
    trees = 100,
    tree_depth = 6,
    learn_rate = 0.01
  ) |>
    parsnip::set_engine("xgboost")
  
  # 2. Define a recipe on data without 'subject' because split metadata
  # columns are excluded from predictors by fit_resample().
  df_for_rec <- df[, !names(df) %in% "subject"]
  
  rec_xgb <- recipes::recipe(outcome ~ ., data = df_for_rec) |>
    recipes::step_dummy(recipes::all_nominal_predictors()) |>
    recipes::step_impute_median(recipes::all_numeric_predictors())
  # Note: No need for step_rm(subject) because it's already gone!
  
  # 3. Fit
  fit_xgb <- fit_resample(
    df,
    outcome = "outcome",
    splits = splits,
    learner = xgb_spec,
    metrics = "auc",
    preprocess = rec_xgb 
  )
  
  cat("XGBoost parsnip fit summary:\n")
  print(summary(fit_xgb))
} else {
  cat("parsnip/xgboost/recipes not installed.\n")
}

## ----custom-learners----------------------------------------------------------
custom_learners <- list(
  glm = list(
    fit = function(x, y, task, weights, ...) {
      df_fit <- data.frame(y = y, x, check.names = FALSE)
      stats::glm(y ~ ., data = df_fit,
                 family = stats::binomial(), weights = weights)
    },
    predict = function(object, newdata, task, ...) {
      as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
                                type = "response"))
    }
  )
)

cat("Custom learner names:\n")
names(custom_learners)
cat("Custom learner components (fit/predict):\n")
lapply(custom_learners, names)

## ----plot-fold-balance--------------------------------------------------------
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_fold_balance(fit_safe)
} else {
  cat("ggplot2 not installed; skipping fold balance plot.\n")
}

## ----plot-calibration---------------------------------------------------------
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_calibration(fit_safe, bins = 10)
} else {
  cat("ggplot2 not installed; skipping calibration plot.\n")
}

## ----plot-confounder-sensitivity----------------------------------------------
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_confounder_sensitivity(fit_safe, confounders = c("batch", "study"),
                              metric = "auc", min_n = 3)
} else {
  cat("ggplot2 not installed; skipping confounder sensitivity plot.\n")
}

## ----diagnostics-tables-------------------------------------------------------
cal <- calibration_summary(fit_safe, bins = 10, min_bin_n = 5)
conf_tbl <- confounder_sensitivity(fit_safe,
                                   confounders = c("batch", "study"),
                                   metric = "auc",
                                   min_n = 3)

# Calibration metrics
cal$metrics

# Confounder sensitivity table (first 6 rows)
head(conf_tbl)

## ----plot-overlap-------------------------------------------------------------
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_overlap_checks(fit_leaky, column = "subject")
  plot_overlap_checks(fit_safe, column = "subject")
} else {
  cat("ggplot2 not installed; skipping overlap plots.\n")
}

## ----audit-leakage------------------------------------------------------------
# Use df_leaky (the original 160-row data) so X_ref aligns with fit_safe's predictions.
# df may have been redefined to a smaller dataset in the tidymodels section above.
X_ref <- df_leaky[, predictors]
X_ref[c(1, 5), ] <- X_ref[1, ]

audit <- audit_leakage(
  fit_safe,
  metric = "auc",
  B = 20,
  perm_stratify = TRUE,
  batch_cols = c("batch", "study"),
  X_ref = X_ref,
  sim_method = "cosine",
  sim_threshold = 0.995,
  return_perm = TRUE
)

cat("Leakage audit summary:\n")
summary(audit)
pg <- audit_perm_gap(audit)
if (!is.null(pg) && nrow(pg) > 0) {
  # Permutation significance results
  pg
}
ba <- audit_batch_assoc(audit)
if (!is.null(ba) && nrow(ba) > 0) {
  # Batch/study association with folds (Cramer's V)
  ba
} else {
  cat("No batch or study associations detected.\n")
}
ta <- audit_target_assoc(audit)
if (!is.null(ta) && nrow(ta) > 0) {
  # Top features by target association score
  head(ta)
} else {
  cat("No target leakage scan results available.\n")
}
du <- audit_duplicates(audit)
if (!is.null(du) && nrow(du) > 0) {
  # Top duplicate/near-duplicate pairs by similarity
  head(du)
} else {
  cat("No near-duplicates detected.\n")
}

## ----mechanism-summary--------------------------------------------------------
mech <- audit@info$mechanism_summary
if (is.data.frame(mech) && nrow(mech) > 0) {
  mech
} else {
  cat("No mechanism summary available.\n")
}

## ----plot-perm----------------------------------------------------------------
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_perm_distribution(audit)
} else {
  cat("ggplot2 not installed; skipping permutation plot.\n")
}

## ----audit-by-learner---------------------------------------------------------
if (requireNamespace("ranger", quietly = TRUE) && 
    requireNamespace("parsnip", quietly = TRUE)) {
    
    # 1. Define specs
    # Standard GLM (no tuning)
    spec_glm <- parsnip::logistic_reg() |> 
        parsnip::set_engine("glm")
    
    # Random Forest
    spec_rf <- parsnip::rand_forest(
        mode = "classification",
        trees = 100
    ) |>
        parsnip::set_engine("ranger")
    
    # 2. Fit using the current split object
    fit_multi <- fit_resample(
        df,
        outcome = "outcome",
        splits = splits,
        learner = list(glm = spec_glm, rf = spec_rf),
        metrics = "auc"
    )
    
    # 3. Run the audit
    audits <- audit_leakage_by_learner(fit_multi, metric = "auc", B = 20)
    cat("Per-learner audit summary:\n")
    print(audits)
    
} else {
    cat("ranger/parsnip not installed.\n")
}

## ----audit-report, eval = FALSE-----------------------------------------------
# if (requireNamespace("rmarkdown", quietly = TRUE) && rmarkdown::pandoc_available()) {
#   report_path <- audit_report(audit, output_dir = ".")
#   cat("HTML report written to:\n", report_path, "\n")
# } else {
#   cat("rmarkdown or pandoc not available; skipping audit report rendering.\n")
# }

## ----time-series-fit----------------------------------------------------------
time_splits <- make_split_plan(
  df_time,
  outcome = "outcome",
  mode = "time_series",
  time = "time",
  v = 4,
  horizon = 1
)

cat("Time-series splits summary:\n")
time_splits

fit_time_leaky <- fit_resample(
  df_time,
  outcome = "outcome",
  splits = time_splits,
  learner = spec,
  metrics = "auc"
)

cat("Time-series leaky fit summary:\n")
summary(fit_time_leaky)

## ----time-series-safe---------------------------------------------------------
df_time_safe <- df_time
df_time_safe$leak_future <- NULL

fit_time_safe <- fit_resample(
  df_time_safe,
  outcome = "outcome",
  splits = time_splits,
  learner = spec,
  metrics = "auc"
)

cat("Time-series safe fit summary:\n")
summary(fit_time_safe)

audit_time <- audit_leakage(
  fit_time_safe,
  metric = "auc",
  B = 20,
  time_block = "stationary",
  block_len = 5
)

cat("Time-series leakage audit summary:\n")
summary(audit_time)
if (!is.null(audit_time@permutation_gap) && nrow(audit_time@permutation_gap) > 0) {
  # Time-series permutation significance results
  audit_time@permutation_gap
}

if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_time_acf(fit_time_safe, lag.max = 20)
} else {
  cat("ggplot2 not installed; skipping ACF plot.\n")
}

## ----cv-ci--------------------------------------------------------------------
# Standard 95% CI from the leakage-safe fit
ci_std <- cv_ci(fit_safe@metrics, level = 0.95, method = "normal")
cat("Standard 95% CI:\n")
print(ci_std)

# Nadeau-Bengio corrected CI
# For K-fold CV: n_train ≈ (K-1)/K × n, n_test ≈ n/K
K_folds  <- length(unique(fit_safe@metrics$fold))
n_total  <- nrow(fit_safe@splits@info$coldata)

ci_nb <- cv_ci(
  fit_safe@metrics,
  level   = 0.95,
  method  = "nadeau_bengio",
  n_train = round(n_total * (K_folds - 1L) / K_folds),
  n_test  = round(n_total / K_folds)
)
cat("Nadeau-Bengio corrected 95% CI:\n")
print(ci_nb)

## ----delta-lsi-data-----------------------------------------------------------
set.seed(100)
n_d    <- 120L
subj_d <- rep(paste0("P", seq_len(30L)), each = 4L)   # 30 subjects, 4 obs each

df_dlsi_guarded <- data.frame(
  subject = subj_d,
  outcome = factor(sample(c("case", "control"), n_d, replace = TRUE)),
  x1      = rnorm(n_d),
  x2      = rnorm(n_d)
)

# Leaky feature: subject-level mean of outcome
# With subject-grouped splits, test subjects' leak value encodes their outcome
df_dlsi_naive <- df_dlsi_guarded
df_dlsi_naive$leak <- ave(
  as.numeric(df_dlsi_guarded$outcome == "case"),
  subj_d, FUN = mean
)

# Shared splits: 5-fold × 5 repeats → R_eff = 5 (tier C: point + p-value)
splits_dlsi <- make_split_plan(
  df_dlsi_guarded,
  outcome = "outcome",
  mode    = "subject_grouped",
  group   = "subject",
  v       = 5L,
  repeats = 5L
)

## ----delta-lsi-fits-----------------------------------------------------------
# Reuse the parsnip spec (base R glm) defined earlier
# Naive fit: uses leaky feature (leak encodes test-subject outcome label)
fit_dlsi_naive <- fit_resample(
  df_dlsi_naive,
  outcome = "outcome",
  splits  = splits_dlsi,
  learner = spec,
  metrics = "auc",
  seed    = 1L
)

# Guarded fit: same splits, clean features only
fit_dlsi_guarded <- fit_resample(
  df_dlsi_guarded,
  outcome = "outcome",
  splits  = splits_dlsi,
  learner = spec,
  metrics = "auc",
  seed    = 1L
)

cat("Naive   AUC:", round(mean(fit_dlsi_naive@metrics$auc,   na.rm = TRUE), 3), "\n")
cat("Guarded AUC:", round(mean(fit_dlsi_guarded@metrics$auc, na.rm = TRUE), 3), "\n")

## ----delta-lsi-run------------------------------------------------------------
result_dlsi <- delta_lsi(
  fit_leaky   = fit_dlsi_naive,
  fit_guarded = fit_dlsi_guarded,
  metric      = "auc",
  M_boot      = 300L,
  M_flip      = 300L,
  seed        = 1L
)

summary(result_dlsi)

## ----delta-lsi-tier-----------------------------------------------------------
cat("Tier:          ", result_dlsi@tier, "\n")
cat("R_eff:         ", result_dlsi@R_eff, "\n")
cat("inference_ok:  ", result_dlsi@inference_ok, "\n")

## ----delta-lsi-components-----------------------------------------------------
# delta_metric: arithmetic mean of {Δ_r}
cat("delta_metric:  ", round(result_dlsi@delta_metric, 4), "\n")
# delta_lsi: Huber M-estimate of {Δ_r} (k=1.345, fixed MAD scale)
cat("delta_lsi:     ", round(result_dlsi@delta_lsi,    4), "\n")
# delta_lsi_ci: 95% BCa interval for delta_lsi (NA below tier B)
cat("95% BCa CI:    [",
    paste(round(result_dlsi@delta_lsi_ci, 4), collapse = ", "), "]\n")
# p_value: sign-flip randomization p-value (NA below tier C or if unpaired)
cat("p-value:       ",
    if (is.na(result_dlsi@p_value)) "NA (tier D or unpaired)"
    else format(result_dlsi@p_value, digits = 3), "\n")

## ----delta-lsi-unpaired, warning=TRUE-----------------------------------------
# Unpaired: naive uses sample-wise CV, guarded uses subject-grouped splits
splits_rand <- make_split_plan(
  df_dlsi_naive,
  outcome = "outcome",
  mode    = "subject_grouped",
  group   = "row_id",
  v       = 5L,
  repeats = 5L
)

fit_naive_rand <- fit_resample(
  df_dlsi_naive,
  outcome = "outcome",
  splits  = splits_rand,
  learner = spec,
  metrics = "auc",
  seed    = 1L
)

r_unpaired <- suppressWarnings(
  delta_lsi(fit_naive_rand, fit_dlsi_guarded, metric = "auc", seed = 1L)
)

cat(sprintf("Paired:        %s\n", r_unpaired@info$paired))
cat(sprintf("R_eff:         %d  (0 = unpaired, inference disabled)\n", r_unpaired@R_eff))
cat(sprintf("delta_metric:  %.4f  (naive raw mean - guarded raw mean)\n",
            r_unpaired@delta_metric))
cat(sprintf("p_value:       %s\n",
            if (is.na(r_unpaired@p_value)) "NA" else r_unpaired@p_value))

## ----delta-lsi-direction------------------------------------------------------
r_hib_true  <- suppressWarnings(delta_lsi(
  fit_dlsi_naive, fit_dlsi_guarded,
  metric           = "auc",
  higher_is_better = TRUE,
  seed             = 1L
))
r_hib_false <- suppressWarnings(delta_lsi(
  fit_dlsi_naive, fit_dlsi_guarded,
  metric           = "auc",
  higher_is_better = FALSE,
  seed             = 1L
))

cat(sprintf("hib = TRUE:   delta_metric = %+.4f  (positive = naive inflated)\n",
            r_hib_true@delta_metric))
cat(sprintf("hib = FALSE:  delta_metric = %+.4f  (sign flipped)\n",
            r_hib_false@delta_metric))

## ----delta-lsi-hib-slot-------------------------------------------------------
cat("higher_is_better used:", result_dlsi@info$higher_is_better, "\n")

## ----delta-lsi-slots----------------------------------------------------------
# Per-repeat summaries for custom plotting
head(result_dlsi@repeats_naive,   3)
head(result_dlsi@repeats_guarded, 3)

# Per-fold diagnostics
head(result_dlsi@folds_naive,   3)
head(result_dlsi@folds_guarded, 3)

# Metadata
str(result_dlsi@info)

## ----delta-lsi-audit-connect--------------------------------------------------
# Step 1: detect and characterise leakage
audit_naive_dlsi <- audit_leakage(fit_dlsi_naive, metric = "auc", B = 20L)
cat("Naive audit summary:\n")
summary(audit_naive_dlsi)

# Step 2: confirmed guards reduced inflation; check residual DLSI
cat("\ndelta_metric (naive vs guarded):", round(result_dlsi@delta_metric, 4), "\n")
if (!is.na(result_dlsi@p_value)) {
  cat("Sign-flip p-value:              ", format(result_dlsi@p_value, digits = 3), "\n")
}

## ----parallel-setup, eval=FALSE-----------------------------------------------
# ## future is a Suggests dependency. Install with install.packages("future")
# ## before running this code; otherwise library(future) errors.
# library(future)
# 
# # Use multiple cores (works on all OS)
# plan(multisession, workers = 4)
# 
# # Run a heavy simulation
# sim <- simulate_leakage_suite(..., parallel = TRUE)
# 
# # Parallel folds or audits
# # fit_resample(..., parallel = TRUE)
# # audit_leakage(..., parallel = TRUE)
# # audit_leakage_by_learner(..., parallel_learners = TRUE)
# 
# # Return to sequential processing
# plan(sequential)

## ----simulate-suite-----------------------------------------------------------
if (requireNamespace("glmnet", quietly = TRUE)) {
  sim <- simulate_leakage_suite(
    n = 80,
    p = 6,
    mode = "subject_grouped",
    learner = "glmnet",
    leakage = "subject_overlap",
    seeds = 1:2,
    B = 20,
    parallel = FALSE,
    signal_strength = 1
  )
  
  # Simulation results (first 6 rows)
  head(sim)
} else {
  cat("glmnet not installed; skipping simulation suite example.\n")
}

# test_that("audit_report renders an HTML report", {
#   skip_if_not_installed("rmarkdown")
#   pandoc_ok <- tryCatch(rmarkdown::pandoc_available(), error = function(e) FALSE)
#   if (!isTRUE(pandoc_ok)) {
#     skip("pandoc not available")
#   }
#
#   set.seed(3)
#   df <- data.frame(
#     subject = rep(1:6, each = 2),
#     outcome = rep(c(0, 1), 6),
#     x1 = rnorm(12),
#     x2 = rnorm(12)
#   )
#   splits <- make_split_plan(df, outcome = "outcome",
#                         mode = "subject_grouped", group = "subject",
#                         v = 3, seed = 1, progress = FALSE)
#
#   custom <- list(
#     glm = list(
#       fit = function(x, y, task, weights, ...) {
#         suppressWarnings(stats::glm(y ~ ., data = as.data.frame(x),
#                                     family = stats::binomial(), weights = weights))
#       },
#       predict = function(object, newdata, task, ...) {
#         as.numeric(suppressWarnings(stats::predict(object,
#                                                    newdata = as.data.frame(newdata),
#                                                    type = "response")))
#       }
#     )
#   )
#
#   fit <- fit_resample(
#     df,
#     outcome = "outcome",
#     splits = splits,
#     learner = "glm",
#     custom_learners = custom,
#     metrics = "auc",
#     refit = FALSE,
#     seed = 1
#   )
#   audit <- audit_leakage(fit, metric = "auc", B = 5, perm_stratify = FALSE)
#
#   out_file <- audit_report(audit, output_dir = tempdir(),
#                            output_file = "bioLeak_audit_report_test.html",
#                            quiet = TRUE, open = FALSE)
#   expect_true(file.exists(out_file))
# })

test_that("audit_report validates inputs and dependencies", {
  expect_error(audit_report(list()), "LeakAudit or LeakFit")

  if (!requireNamespace("rmarkdown", quietly = TRUE)) {
    df <- make_class_df(10)
    splits <- make_split_plan_quiet(df, outcome = "outcome",
                                mode = "subject_grouped", group = "subject",
                                v = 2, seed = 1)
    custom <- make_custom_learners()
    fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                              learner = "glm", custom_learners = custom,
                              metrics = "auc", refit = FALSE, seed = 1)
    expect_error(audit_report(fit), "Install 'rmarkdown'")
  }
})

test_that("audit_report template exists in source tree", {
  tpl <- system.file("reports", "audit_report.Rmd", package = "bioLeak")
  expect_true(nzchar(tpl))
  expect_true(file.exists(tpl))
})

make_class_df <- function(n = 20) {
  stopifnot(n >= 4)
  data.frame(
    subject = rep(seq_len(ceiling(n / 2)), each = 2)[seq_len(n)],
    batch = rep(letters[1:3], length.out = n),
    study = rep(LETTERS[1:4], length.out = n),
    time = seq_len(n),
    outcome = factor(rep(c(0, 1), length.out = n), levels = c(0, 1)),
    x1 = rnorm(n),
    x2 = rnorm(n),
    stringsAsFactors = FALSE
  )
}

make_regression_df <- function(n = 20) {
  stopifnot(n >= 4)
  data.frame(
    subject = rep(seq_len(ceiling(n / 2)), each = 2)[seq_len(n)],
    batch = rep(letters[1:3], length.out = n),
    study = rep(LETTERS[1:4], length.out = n),
    time = seq_len(n),
    y = rnorm(n),
    x1 = rnorm(n),
    x2 = rnorm(n),
    stringsAsFactors = FALSE
  )
}

make_multiclass_df <- function(n = 30, k = 3) {
  stopifnot(n >= 6, k >= 3)
  classes <- factor(rep(seq_len(k), length.out = n))
  data.frame(
    subject = rep(seq_len(ceiling(n / 2)), each = 2)[seq_len(n)],
    batch = rep(letters[1:3], length.out = n),
    study = rep(LETTERS[1:4], length.out = n),
    time = seq_len(n),
    outcome = classes,
    x1 = rnorm(n),
    x2 = rnorm(n),
    stringsAsFactors = FALSE
  )
}

make_custom_learners <- function() {
  list(
    glm = list(
      fit = function(x, y, task, weights, ...) {
        df <- data.frame(y = y, x, check.names = FALSE)
        if (identical(task, "binomial")) {
          suppressWarnings(stats::glm(y ~ ., data = df,
                                      family = stats::binomial(), weights = weights))
        } else {
          suppressWarnings(stats::lm(y ~ ., data = df, weights = weights))
        }
      },
      predict = function(object, newdata, task, ...) {
        df <- as.data.frame(newdata, check.names = FALSE)
        if (identical(task, "binomial")) {
          as.numeric(suppressWarnings(stats::predict(object,
                                                     newdata = df,
                                                     type = "response")))
        } else {
          as.numeric(suppressWarnings(stats::predict(object, newdata = df)))
        }
      }
    )
  )
}

make_split_plan_quiet <- function(...) {
  make_split_plan(..., progress = FALSE)
}

fit_resample_quiet <- function(...) {
  out <- NULL
  err <- NULL
  suppress_patterns <- c(
    "glm.fit",
    "algorithm did not converge",
    "fitted probabilities numerically 0 or 1 occurred",
    "rank-deficient",
    "non-estim",
    "one multinomial or binomial class has fewer than 8",
    "option grouped=false enforced in cv.glmnet"
  )
  seen_warnings <- character()
  capture.output({
    out <- withCallingHandlers(
      tryCatch(
        fit_resample(...),
        error = function(e) {
          err <<- e
          NULL
        }
      ),
      warning = function(w) {
        seen_warnings <<- c(seen_warnings, conditionMessage(w))
        invokeRestart("muffleWarning")
      }
    )
  })
  if (length(seen_warnings)) {
    keep <- !vapply(seen_warnings, function(msg) {
      msg <- tolower(msg)
      any(vapply(suppress_patterns, function(pat) {
        grepl(pat, msg, fixed = TRUE)
      }, logical(1)))
    }, logical(1))
    if (any(keep)) {
      for (msg in seen_warnings[keep]) {
        warning(msg, call. = FALSE)
      }
    }
  }
  if (!is.null(err)) {
    stop(err)
  }
  out
}

with_temp_plot_device <- function(expr) {
  tmp <- tempfile(fileext = ".pdf")
  grDevices::pdf(tmp)
  on.exit(grDevices::dev.off(), add = TRUE)
  eval.parent(substitute(expr))
}

expect_warning_match <- function(expr, pattern, all = FALSE) {
  warnings <- character()
  value <- withCallingHandlers(expr, warning = function(w) {
    warnings <<- c(warnings, conditionMessage(w))
    invokeRestart("muffleWarning")
  })
  if (!length(warnings)) {
    testthat::expect_true(FALSE, info = "Expected warning, but none were emitted.")
  } else if (isTRUE(all)) {
    testthat::expect_true(all(grepl(pattern, warnings)))
  } else {
    testthat::expect_true(any(grepl(pattern, warnings)))
  }
  value
}

test_that("fit_resample supports custom learners", {
  set.seed(1)
  df <- data.frame(
    subject = rep(1:10, each = 2),
    outcome = rep(c(0, 1), times = 10),
    x1 = rnorm(20),
    x2 = rnorm(20)
  )

  splits <- make_split_plan(
    df,
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 5,
    stratify = FALSE
  )

  custom <- list(
    glm = list(
      fit = function(x, y, task, weights, ...) {
        suppressWarnings(stats::glm(y ~ ., data = as.data.frame(x),
                                    family = stats::binomial(), weights = weights))
      },
      predict = function(object, newdata, task, ...) {
        as.numeric(suppressWarnings(stats::predict(object,
                                                   newdata = as.data.frame(newdata),
                                                   type = "response")))
      }
    )
  )

  fit <- fit_resample(
    df,
    outcome = "outcome",
    splits = splits,
    learner = "glm",
    custom_learners = custom,
    metrics = "accuracy",
    refit = FALSE,
    seed = 1
  )

  expect_s4_class(fit, "LeakFit")
  expect_true(nrow(fit@metrics) > 0)
})

test_that("audit_leakage reports batch association and duplicates", {
  set.seed(2)
  X <- matrix(rnorm(24), nrow = 12, ncol = 2)
  X[7, ] <- X[1, ]
  X[8, ] <- X[2, ]

  df <- data.frame(
    outcome = rep(c(0, 1), 6),
    batch = rep(c("A", "B"), each = 6),
    x1 = X[, 1],
    x2 = X[, 2]
  )

  fold1 <- c(1L, 2L, 3L, 7L, 8L, 9L)
  fold2 <- c(4L, 5L, 6L, 10L, 11L, 12L)
  indices <- list(
    list(train = setdiff(seq_len(12), fold1), test = fold1, fold = 1, repeat_id = 1),
    list(train = setdiff(seq_len(12), fold2), test = fold2, fold = 2, repeat_id = 1)
  )
  splits <- bioLeak:::LeakSplits(mode = "custom", indices = indices,
                                 info = list(outcome = "outcome", coldata = df))

  custom <- list(
    glm = list(
      fit = function(x, y, task, weights, ...) {
        suppressWarnings(stats::glm(y ~ ., data = as.data.frame(x),
                                    family = stats::binomial(), weights = weights))
      },
      predict = function(object, newdata, task, ...) {
        as.numeric(suppressWarnings(stats::predict(object,
                                                   newdata = as.data.frame(newdata),
                                                   type = "response")))
      }
    )
  )

  fit <- fit_resample(
    df,
    outcome = "outcome",
    splits = splits,
    learner = "glm",
    custom_learners = custom,
    metrics = "auc",
    refit = FALSE,
    seed = 1
  )

  audit <- audit_leakage(
    fit,
    metric = "auc",
    B = 10,
    perm_stratify = FALSE,
    batch_cols = "batch",
    X_ref = df[, c("x1", "x2")],
    sim_threshold = 0.999,
    duplicate_scope = "all",
    # ADD THIS LINE TO FIX WARNINGS:
    target_scan_multivariate = FALSE
  )

  expect_true(nrow(audit@batch_assoc) >= 1)
  expect_true(is.finite(audit@batch_assoc$pval[1]))
  expect_true(nrow(audit@duplicates) >= 1)
})

test_that("audit_leakage batch association respects repeated CV folds", {
  set.seed(3)
  df <- make_class_df(8)
  df$batch <- rep(c("A", "B"), each = 4)
  df <- df[, c("batch", "outcome", "x1", "x2")]

  fold1 <- c(1L, 2L, 5L, 6L)
  fold2 <- c(3L, 4L, 7L, 8L)
  fold3 <- c(1L, 2L, 7L, 8L)
  fold4 <- c(3L, 4L, 5L, 6L)
  indices <- list(
    list(train = setdiff(seq_len(8), fold1), test = fold1, fold = 1, repeat_id = 1),
    list(train = setdiff(seq_len(8), fold2), test = fold2, fold = 2, repeat_id = 1),
    list(train = setdiff(seq_len(8), fold3), test = fold3, fold = 1, repeat_id = 2),
    list(train = setdiff(seq_len(8), fold4), test = fold4, fold = 2, repeat_id = 2)
  )
  splits <- bioLeak:::LeakSplits(mode = "custom", indices = indices,
                                 info = list(outcome = "outcome", coldata = df, batch = "batch"))

  custom <- make_custom_learners()
  fit <- fit_resample_quiet(
    df,
    outcome = "outcome",
    splits = splits,
    learner = "glm",
    custom_learners = custom,
    metrics = "auc",
    refit = FALSE,
    seed = 1
  )

  audit <- audit_leakage(
    fit,
    metric = "auc",
    B = 2,
    perm_stratify = FALSE,
    batch_cols = "batch",
    return_perm = FALSE,
    target_scan = FALSE,
    target_scan_multivariate = FALSE
  )

  expect_equal(sort(unique(audit@batch_assoc$repeat_id)), c(1, 2))
  expect_true(all(audit@batch_assoc$df == 1))
})

test_that("audit_leakage supports refit-based permutations", {
  df <- make_class_df(20)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 3, seed = 1)
  custom <- make_custom_learners()

  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "glm", custom_learners = custom,
                            metrics = "auc", refit = FALSE, seed = 1)

  audit <- audit_leakage(
    fit,
    metric = "auc",
    B = 2,
    perm_refit = TRUE,
    perm_refit_spec = list(
      x = df,
      outcome = "outcome",
      learner = "glm",
      custom_learners = custom
    ),
    perm_stratify = FALSE
  )

  expect_true(identical(audit@info$perm_method, "refit"))
  expect_equal(audit@permutation_gap$n_perm[1], 2)
})

test_that("audit_leakage supports perm_refit auto mode", {
  df <- make_class_df(20)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 3, seed = 1)
  custom <- make_custom_learners()
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "glm", custom_learners = custom,
                            metrics = "auc", refit = FALSE, seed = 1)

  audit_fixed <- audit_leakage(
    fit,
    metric = "auc",
    B = 5,
    perm_refit = "auto",
    perm_refit_auto_max = 2,
    perm_stratify = FALSE
  )
  expect_true(identical(audit_fixed@info$perm_method, "fixed"))
  expect_true(grepl("^auto", audit_fixed@info$perm_refit_mode))

  audit_refit <- audit_leakage(
    fit,
    metric = "auc",
    B = 2,
    perm_refit = "auto",
    perm_refit_auto_max = 5,
    perm_refit_spec = list(
      x = df,
      outcome = "outcome",
      learner = "glm",
      custom_learners = custom
    ),
    perm_stratify = FALSE
  )
  expect_true(identical(audit_refit@info$perm_method, "refit"))
  expect_true(grepl("^auto", audit_refit@info$perm_refit_mode))
})

test_that("audit_leakage multivariate target scan runs", {
  df <- make_class_df(24)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 3, seed = 1)
  custom <- make_custom_learners()
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "glm", custom_learners = custom,
                            metrics = "auc", refit = FALSE, seed = 1)
  aud <- audit_leakage(
    fit,
    metric = "auc",
    B = 3,
    perm_stratify = FALSE,
    X_ref = df[, c("x1", "x2")],
    target_scan_multivariate = TRUE,
    target_scan_multivariate_B = 2,
    target_scan_multivariate_components = 2
  )
  mv <- aud@info$target_multivariate
  expect_true(is.data.frame(mv))
  expect_true(nrow(mv) > 0)
  expect_true(all(c("metric", "score", "p_value") %in% names(mv)))
})

test_that("audit_leakage adds mechanism taxonomy and FDR-adjusted target flags", {
  df <- make_class_df(24)
  splits <- make_split_plan_quiet(
    df,
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 3,
    seed = 7
  )
  custom <- make_custom_learners()
  fit <- fit_resample_quiet(
    df,
    outcome = "outcome",
    splits = splits,
    learner = "glm",
    custom_learners = custom,
    metrics = "auc",
    refit = FALSE,
    seed = 7
  )

  aud <- audit_leakage(
    fit,
    metric = "auc",
    B = 3,
    perm_stratify = FALSE,
    X_ref = df[, c("x1", "x2")],
    target_scan_multivariate = FALSE,
    target_p_adjust = "BH",
    target_alpha = 0.2
  )

  expect_true("mechanism_class" %in% names(aud@permutation_gap))
  if (nrow(aud@batch_assoc)) {
    expect_true("mechanism_class" %in% names(aud@batch_assoc))
  }
  if (nrow(aud@target_assoc)) {
    expect_true("mechanism_class" %in% names(aud@target_assoc))
    expect_true("p_value_adj" %in% names(aud@target_assoc))
    expect_true("flag_fdr" %in% names(aud@target_assoc))
  }
  if (nrow(aud@duplicates)) {
    expect_true("mechanism_class" %in% names(aud@duplicates))
  }

  expect_true(is.data.frame(aud@info$taxonomy))
  expect_true(all(c("mechanism_class", "evidence_slot") %in% names(aud@info$taxonomy)))
  expect_true(is.data.frame(aud@info$mechanism_summary))
  expect_true(all(c("mechanism_class", "flagged", "evidence") %in% names(aud@info$mechanism_summary)))
})

test_that("compact splits are resolved during fitting", {
  set.seed(1)
  df <- data.frame(
    outcome = rep(c(0, 1), times = 10),
    subject = rep(1:10, each = 2),
    x1 = rnorm(20),
    x2 = rnorm(20)
  )
  rownames(df) <- paste0("S", seq_len(nrow(df)))

  splits <- make_split_plan(df,
                        outcome = "outcome",
                        mode = "subject_grouped",
                        group = "subject",
                        v = 5,
                        compact = TRUE,
                        progress = FALSE)

  expect_true(isTRUE(splits@info$compact))
  expect_equal(length(splits@info$fold_assignments), 1)
  expect_true(all(!is.na(splits@info$fold_assignments[[1]])))

  custom <- list(
    glm = list(
      fit = function(x, y, task, weights, ...) {
        suppressWarnings(stats::glm(y ~ ., data = as.data.frame(x),
                                    family = stats::binomial(), weights = weights))
      },
      predict = function(object, newdata, task, ...) {
        as.numeric(suppressWarnings(stats::predict(object,
                                                   newdata = as.data.frame(newdata),
                                                   type = "response")))
      }
    )
  )

  fit <- fit_resample(
    df,
    outcome = "outcome",
    splits = splits,
    learner = "glm",
    custom_learners = custom,
    metrics = "accuracy",
    refit = FALSE,
    seed = 1
  )

  pred_ids <- unique(do.call(rbind, fit@predictions)$id)
  expect_true(all(pred_ids %in% rownames(df)))
})

test_that("compact time_series splits keep purge/embargo behavior in as_rsample", {
  skip_if_not_installed("rsample")
  df <- make_class_df(12)
  splits <- make_split_plan_quiet(
    df,
    outcome = "outcome",
    mode = "time_series",
    time = "time",
    v = 3,
    compact = TRUE,
    purge = 2,
    embargo = 5,
    seed = 1
  )
  rs <- as_rsample(splits, data = df)

  expect_equal(nrow(rs), 2)
  expect_equal(max(rsample::analysis(rs$splits[[1]])$time), 3)
  expect_equal(max(rsample::analysis(rs$splits[[2]])$time), 7)
})

test_that("fit_resample supports multiclass custom learners", {
  df <- make_multiclass_df(15, k = 3)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 3, seed = 1)
  custom <- list(
    freq = list(
      fit = function(x, y, task, weights, ...) {
        probs <- prop.table(table(y))
        list(probs = probs, levels = names(probs))
      },
      predict = function(object, newdata, task, ...) {
        n <- nrow(newdata)
        prob_mat <- matrix(rep(object$probs, each = n), nrow = n, byrow = FALSE)
        colnames(prob_mat) <- object$levels
        prob_mat
      }
    )
  )
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "freq", custom_learners = custom,
                            metrics = c("accuracy", "macro_f1", "log_loss"),
                            refit = FALSE)
  expect_true(all(c("accuracy", "macro_f1", "log_loss") %in% colnames(fit@metrics)))
})

# test_that("fit_resample supports survival tasks with custom learners", {
#   skip_if_not_installed("survival")
#   n <- 20
#   df <- data.frame(
#     subject = rep(seq_len(10), each = 2),
#     time = rexp(n, rate = 0.1),
#     status = rbinom(n, 1, 0.7),
#     x1 = rnorm(n),
#     x2 = rnorm(n),
#     stringsAsFactors = FALSE
#   )
#   df$surv <- survival::Surv(df$time, df$status)
#   splits <- make_split_plan_quiet(df, outcome = "surv",
#                               mode = "subject_grouped", group = "subject",
#                               v = 2, seed = 1, stratify = FALSE)
#   custom <- list(
#     cox = list(
#       fit = function(x, y, task, weights, ...) {
#         df_fit <- data.frame(y = y, x, check.names = FALSE)
#         survival::coxph(y ~ ., data = df_fit, weights = weights)
#       },
#       predict = function(object, newdata, task, ...) {
#         as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
#                                   type = "lp"))
#       }
#     )
#   )
#   fit <- fit_resample_quiet(df, outcome = "surv", splits = splits,
#                             learner = "cox", custom_learners = custom,
#                             metrics = "cindex", refit = FALSE)
#   expect_true(nrow(fit@metrics) > 0)
# })

## ---------------------------------------------------------------------
## test-predict-guardfit.R
## ---------------------------------------------------------------------
## Tests for the 0.3.7 conversion of `predict_guard()` to a proper S3
## `predict.GuardFit()` method. These tests assert that:
##   (1) `predict()` dispatches to `predict.GuardFit()` on a GuardFit
##       object (the editor's Comment 7 requirement);
##   (2) the new method registers correctly via the standard S3
##       generic so that `methods(class = "GuardFit")` includes
##       "predict";
##   (3) `predict_guard()` is preserved as a backward-compatible alias
##       that yields bit-identical output to the S3 method;
##   (4) error messages for non-GuardFit input are informative;
##   (5) the new method silently accepts `...` (matching the `predict`
##       generic's signature) so user code passing extra arguments does
##       not error.
## ---------------------------------------------------------------------

make_simple_guardfit <- function() {
  X <- data.frame(a = c(1, 2, NA, 4), b = c(10, 11, 12, 13))
  guard_fit(
    X, y = c(0.1, 0.2, 0.3, 0.4),
    steps = list(impute = list(method = "median")),
    task = "gaussian"
  )
}

test_that("predict() dispatches to predict.GuardFit on a GuardFit object", {
  fit <- make_simple_guardfit()
  X_new <- data.frame(a = c(NA, 5), b = c(9, 14))
  out <- predict(fit, X_new)

  ## We verify dispatch via structural equivalence: the output of
  ## predict() must be bit-identical to fit$transform(newdata), the
  ## raw transform that the S3 method wraps. This avoids hardcoding
  ## guard-fit-internal normalization details that may evolve.
  expect_s3_class(out, "data.frame")
  expect_equal(nrow(out), 2L)
  expect_identical(out, fit$transform(X_new))
  ## The output must NOT contain NA values (the guard fit was
  ## configured with median imputation; the NA in X_new$a should be
  ## filled before normalization).
  expect_false(any(is.na(out$a)))
})

test_that("methods(class = 'GuardFit') includes predict (Comment 7 check)", {
  cls_methods <- as.character(methods(class = "GuardFit"))
  ## Strip dispatch labels like "predict.GuardFit" to bare generics.
  ## utils::methods returns entries like "predict.GuardFit" so we
  ## check for that exact name.
  expect_true("predict.GuardFit" %in% cls_methods)
  expect_true("print.GuardFit"   %in% cls_methods)
  expect_true("summary.GuardFit" %in% cls_methods)
})

test_that("predict_guard() yields bit-identical output to predict.GuardFit()", {
  fit <- make_simple_guardfit()
  X_new <- data.frame(a = c(NA, 5), b = c(9, 14))
  expect_identical(predict_guard(fit, X_new), predict(fit, X_new))
  expect_identical(predict_guard(fit, X_new), predict.GuardFit(fit, X_new))
})

test_that("predict.GuardFit() rejects non-GuardFit input with a clear error", {
  X_new <- data.frame(a = 1:3, b = 4:6)
  expect_error(predict.GuardFit(NULL, X_new),
               "must be a GuardFit object")
  expect_error(predict.GuardFit(list(), X_new),
               "must be a GuardFit object")
  expect_error(predict.GuardFit(data.frame(x = 1:3), X_new),
               "must be a GuardFit object")
})

test_that("predict_guard() also rejects non-GuardFit input (via the S3 method)", {
  X_new <- data.frame(a = 1:3, b = 4:6)
  expect_error(predict_guard(NULL,    X_new), "must be a GuardFit object")
  expect_error(predict_guard(list(),  X_new), "must be a GuardFit object")
})

test_that("predict.GuardFit() accepts and silently drops `...` args", {
  fit <- make_simple_guardfit()
  X_new <- data.frame(a = c(NA, 5), b = c(9, 14))
  ## Extra arguments should not cause an error (S3 method signature).
  out_no_extra <- predict(fit, X_new)
  out_extra    <- predict(fit, X_new, type = "response", se.fit = FALSE)
  expect_identical(out_no_extra, out_extra)
})

test_that("predict.GuardFit() and predict_guard() are both exported", {
  exported <- getNamespaceExports("bioLeak")
  expect_true("predict_guard" %in% exported)
  ## predict.GuardFit is registered as an S3 method via S3method() and
  ## therefore appears in the bioLeak S3 method table. Check via
  ## utils::methods() filtered for our class.
  cls_methods <- as.character(methods(class = "GuardFit"))
  expect_true("predict.GuardFit" %in% cls_methods)
})

# Safe sample that avoids the R sample(x) gotcha when length(x) == 1 -----------
.safe_sample <- function(x) {
  if (length(x) <= 1L) return(x)
  x[sample.int(length(x))]
}

# Restricted permutation engines ------------------------------------------------

#' Quantile break cache for permutation stratification
#'
#' Internal environment used to cache quantile breakpoints for numeric
#' outcomes during restricted permutation testing. This avoids recomputing
#' quantiles across repeated calls in \code{audit_leakage()}.
#'
#' @format An environment used to cache quantile breakpoints.
#' @return An environment (internal data object, not a function).
#' @keywords internal
#' @docType data
# Cached quantile breaks for numeric stratification ------------------------
.quantile_break_cache <- new.env(parent = emptyenv())

.get_cached_quantile_breaks <- function(vals, probs) {
  vals_clean <- vals[!is.na(vals)]
  if (requireNamespace("digest", quietly = TRUE)) {
    key <- digest::digest(list(vals = vals_clean, probs = probs))
    if (exists(key, envir = .quantile_break_cache, inherits = FALSE)) {
      return(get(key, envir = .quantile_break_cache, inherits = FALSE))
    }
    breaks <- stats::quantile(vals_clean, probs = probs, na.rm = TRUE)
    breaks <- unique(breaks)
    assign(key, breaks, envir = .quantile_break_cache)
    return(breaks)
  } else {
    # Fallback: compute directly without caching
    return(unique(stats::quantile(vals_clean, probs = probs, na.rm = TRUE)))
  }
}

#.majority_level helper
.majority_level <- function(vals) {
  vals <- vals[!is.na(vals)]
  if (!length(vals)) return(NA_character_)
  tab <- table(vals)
  names(tab)[which.max(tab)]
}

.permute_subject_grouped <- function(y, subj, strata = NULL) {
  subj   <- factor(subj)
  blocks <- split(seq_along(y), subj)
  values <- lapply(blocks, function(ix) y[ix])
  ord    <- seq_along(blocks)
  sizes  <- vapply(blocks, length, integer(1L))

  # Two-stage permutation:
  # Stage 1 — block-swap subjects of equal size within each stratum.
  #   Only same-size blocks are swapped to avoid R recycling and label
  #   corruption.
  # Stage 2 — any subject whose block could not be swapped (unique size
  #   within its stratum) has its own observations element-wise shuffled
  #   within the stratum. Under H0 labels are exchangeable, so this is
  #   valid and avoids an over-conservative test at small n.
  perm_ord <- ord
  swapped  <- logical(length(ord))

  if (!is.null(strata)) {
    block_strata <- vapply(ord,
                           function(i) .majority_level(strata[blocks[[i]]]),
                           character(1L))
    for (nm in unique(block_strata)) {
      in_stratum <- ord[block_strata == nm]
      for (sz in unique(sizes[in_stratum])) {
        eligible <- in_stratum[sizes[in_stratum] == sz]
        if (length(eligible) > 1L) {
          perm_ord[eligible] <- sample(eligible)
          swapped[eligible]  <- TRUE
        }
      }
    }
  } else {
    for (sz in unique(sizes)) {
      eligible <- ord[sizes == sz]
      if (length(eligible) > 1L) {
        perm_ord[eligible] <- sample(eligible)
        swapped[eligible]  <- TRUE
      }
    }
  }

  # Apply block swaps for swapped blocks
  out <- y
  for (i in ord[swapped]) out[blocks[[i]]] <- values[[perm_ord[i]]]

  # Element-wise shuffle for un-swapped blocks (no stratum constraint).
  # Stratifying the fallback shuffle by the binary outcome would be identity
  # (all zeros stay zeros, all ones stay ones), so we shuffle all unswapped
  # positions together.  Under H0 labels are exchangeable, making this valid.
  unswapped_pos <- unlist(blocks[!swapped], use.names = FALSE)
  if (length(unswapped_pos) > 1L) {
    out[unswapped_pos] <- sample(out[unswapped_pos])
  }

  out
}

.permute_within_group <- function(y, group, strata = NULL) {
  group <- factor(group)
  out <- y
  if (!is.null(strata)) {
    # Permute within group x strata cells
    cells <- interaction(group, strata, drop = TRUE)
    for (cell in levels(cells)) {
      ix <- which(cells == cell)
      if (length(ix) > 1L) {
        out[ix] <- sample(out[ix])
      }
    }
  } else {
    for (lvl in levels(group)) {
      ix <- which(group == lvl)
      if (length(ix) > 1L) {
        out[ix] <- sample(out[ix])
      }
    }
  }
  out
}

.permute_within_batch <- function(y, batch, strata = NULL) {
  .permute_within_group(y, batch, strata = strata)
}

.permute_within_study <- function(y, study, strata = NULL) {
  .permute_within_group(y, study, strata = strata)
}

#' Restricted permutation label factory
#'
#' Builds a closure that generates permuted outcome vectors per fold while
#' respecting grouping/batch/study/time constraints used in
#' \code{audit_leakage()}. Numeric outcomes can be stratified by quantiles to
#' preserve outcome structure under permutation.
#'
#' @param cd data.frame of sample metadata.
#' @param outcome outcome column name.
#' @param mode resampling mode (subject_grouped, batch_blocked, study_loocv, time_series).
#' @param folds list of fold descriptors from \code{LeakSplits}. When compact
#'   splits are used, fold assignments are read from the
#'   \code{fold_assignments} attribute.
#' @param perm_stratify logical or "auto"; if TRUE, permute within strata.
#' @param time_block time-series block permutation method.
#' @param block_len block length for time-series permutations.
#' @param seed integer seed.
#' @param group_col,batch_col,study_col optional metadata columns.
#' @param time_col optional metadata column name for time-series ordering.
#' @param perm_refit logical; if TRUE model is retrained on permuted labels
#'   (block permutation preserves subject structure); if FALSE predictions are
#'   fixed and simple label shuffle is used for \code{subject_grouped} mode.
#' @param verbose logical; print progress messages.
#' @return A function that returns a list of permuted outcome vectors, one per fold.
#' @keywords internal
# Factory returning closure that produces permuted training labels per fold
.permute_labels_factory <- function(cd, outcome, mode, folds, perm_stratify,
                                    time_block, block_len, seed,
                                    group_col = NULL, batch_col = NULL,
                                    study_col = NULL, time_col = NULL,
                                    perm_refit = TRUE,
                                    verbose = FALSE) {
  if (is.null(cd) || !outcome %in% names(cd)) {
    stop("Metadata with outcome column required for restricted permutations.")
  }
  y_all <- cd[[outcome]]
  if (all(is.na(y_all))) {
    stop("Outcome column contains only NA values.")
  }

  fold_assignments <- attr(folds, "fold_assignments")
  resolve_test_idx <- function(fold) {
    if (!is.null(fold$test)) return(fold$test)
    if (is.null(fold_assignments) || !length(fold_assignments)) {
      stop("Fold assignments required to resolve test indices for compact splits.")
    }
    r <- fold$repeat_id
    if (is.null(r) || !is.finite(r)) r <- 1L
    assign_vec <- fold_assignments[[r]]
    if (is.null(assign_vec)) {
      stop(sprintf("Missing fold assignments for repeat %s.", r))
    }
    which(assign_vec == fold$fold)
  }

  MIN_SAMPLES_FOR_REGRESSION_STRATIFICATION <- 20L
  strata_vec <- NULL
  should_stratify <- FALSE
  if (isTRUE(perm_stratify)) {
    should_stratify <- TRUE
    if (is.numeric(y_all) &&
        length(stats::na.omit(y_all)) < MIN_SAMPLES_FOR_REGRESSION_STRATIFICATION) {
      warning("perm_stratify = TRUE requires at least 20 non-missing numeric outcomes; proceeding without stratification.")
      should_stratify <- FALSE
    }
  } else if (identical(perm_stratify, "auto")) {
    should_stratify <- is.factor(y_all) ||
      (is.numeric(y_all) &&
       length(stats::na.omit(y_all)) >= MIN_SAMPLES_FOR_REGRESSION_STRATIFICATION)
  }
  if (should_stratify) {
    if (is.factor(y_all)) {
      strata_vec <- y_all
      if (isTRUE(verbose)) {
        message("[permute_labels] Stratifying by factor outcome levels: ",
                paste(levels(strata_vec), collapse = ", "))
      }
    } else if (is.numeric(y_all)) {
      # for regression, bins by quantiles to maintain structure
      br <- .get_cached_quantile_breaks(y_all, probs = seq(0, 1, length.out = 5))
      strata_vec <- cut(y_all, breaks = br, include.lowest = TRUE, labels = FALSE)
      if (isTRUE(verbose)) {
        message("[permute_labels] Stratifying numeric outcome into ",
                length(unique(stats::na.omit(strata_vec))), " bins.")
      }
    }
  } else if (isTRUE(verbose)) {
    message("[permute_labels] Stratification disabled for outcome '", outcome, "'.")
  }
  if (identical(mode, "time_series")) {
    time_col_use <- time_col
    if (is.null(time_col_use) && "time" %in% names(cd)) {
      time_col_use <- "time"
    }
    if (is.null(time_col_use) || !time_col_use %in% names(cd)) {
      stop("time_series permutations require a time column in metadata.", call. = FALSE)
    }
    time_vec <- cd[[time_col_use]]
    if (!is.numeric(time_vec) && !inherits(time_vec, c("POSIXct", "Date"))) {
      stop("time_series time column must be numeric, Date, or POSIXct.", call. = FALSE)
    }
    if (!exists(".stationary_bootstrap", mode = "function")) {
      stop("Missing .stationary_bootstrap() implementation.")
    }
    if (!exists(".circular_block_permute", mode = "function")) {
      stop("Missing .circular_block_permute() implementation.")
    }
  }
  set.seed(seed)
  function(b) {
    set.seed(seed + b)
    if (isTRUE(verbose)) {
      message("[permute_labels] Generating permuted labels for replicate ", b, ".")
    }
    res <- vector("list", length(folds))
    for (i in seq_along(folds)) {
      if (isTRUE(verbose)) {
        message(sprintf("[permute_labels] Permuting fold %d/%d using mode '%s'.",
                        i, length(folds), mode))
      }
      te_idx <- resolve_test_idx(folds[[i]])
      if (isTRUE(verbose) && !is.null(strata_vec)) {
        strata_fold <- stats::na.omit(strata_vec[te_idx])
        if (length(unique(strata_fold)) < 2L) {
          message("[permute_labels] Warning: Fewer than two non-NA strata present in this test fold; stratification has limited effect.")
        }
      }
      permuted <- switch(mode,
        subject_grouped = {
          # For perm_refit=FALSE predictions are fixed; simple shuffle gives the
          # correct null and avoids degenerate/inflated-Type-I-error behaviour
          # from block permutation constrained by subject size and outcome strata.
          if (!isTRUE(perm_refit)) {
            .safe_sample(y_all[te_idx])
          } else {
            subj_col <- if (!is.null(group_col) && group_col %in% names(cd)) cd[[group_col]] else NULL
            if (is.null(subj_col) && "group" %in% names(cd)) subj_col <- cd[["group"]]
            if (is.null(subj_col)) subj_col <- seq_along(y_all)
            subj <- subj_col[te_idx]
            strata <- if (!is.null(strata_vec)) strata_vec[te_idx] else NULL
            if (isTRUE(verbose) && !is.null(strata)) {
              message("[permute_labels] Subject-grouped strata used: ",
                      paste(sort(unique(stats::na.omit(strata))), collapse = ", "))
            }
            .permute_subject_grouped(y_all[te_idx], subj, strata)
          }
        },
        batch_blocked = {
          # For perm_refit=FALSE predictions are fixed; simple shuffle gives the
          # correct null.  Restricted within-batch permutation with stratification
          # collapses to the identity when each test fold is a single batch
          # (because shuffling within batch × outcome cells changes nothing).
          if (!isTRUE(perm_refit)) {
            .safe_sample(y_all[te_idx])
          } else {
            batch_vals <- NULL
            if (!is.null(batch_col) && batch_col %in% names(cd)) batch_vals <- cd[[batch_col]]
            if (is.null(batch_vals) && "batch" %in% names(cd)) batch_vals <- cd[["batch"]]
            if (is.null(batch_vals)) stop("Batch column not found for batch_blocked mode.")
            if (isTRUE(verbose)) {
              ktab <- table(batch_vals[te_idx])
              if (any(ktab == 1L)) {
                message("[permute_labels] Note: ", sum(ktab == 1L),
                        " batch level(s) in this fold have only one sample; permutation within those is identity.")
              }
            }
            .permute_within_batch(y_all[te_idx], batch_vals[te_idx],
                                  strata = if (!is.null(strata_vec)) strata_vec[te_idx] else NULL)
          }
        },
        study_loocv = {
          # For perm_refit=FALSE predictions are fixed; simple shuffle gives the
          # correct null.  Restricted within-study permutation with stratification
          # collapses to the identity when each test fold is a single study
          # (because shuffling within study × outcome cells changes nothing).
          if (!isTRUE(perm_refit)) {
            .safe_sample(y_all[te_idx])
          } else {
            study_vals <- NULL
            if (!is.null(study_col) && study_col %in% names(cd)) study_vals <- cd[[study_col]]
            if (is.null(study_vals) && "study" %in% names(cd)) study_vals <- cd[["study"]]
            if (is.null(study_vals)) stop("Study column not found for study_loocv mode.")
            if (isTRUE(verbose)) {
              ktab <- table(study_vals[te_idx])
              if (any(ktab == 1L)) {
                message("[permute_labels] Note: ", sum(ktab == 1L),
                        " study level(s) in this fold have only one sample; permutation within those is identity.")
              }
            }
            .permute_within_study(y_all[te_idx], study_vals[te_idx],
                                  strata = if (!is.null(strata_vec)) strata_vec[te_idx] else NULL)
          }
        },
        time_series = {
          time_vals <- time_vec[te_idx]
          idx_order <- order(time_vals, te_idx, na.last = TRUE)
          te_idx_sorted <- te_idx[idx_order]
          L <- block_len
          if (is.null(L) || !is.finite(L) || L <= 0) {
            L <- max(5L, floor(length(te_idx_sorted) * 0.1))
          }
          perm_idx <- if (identical(time_block, "stationary")) {
            .stationary_bootstrap(te_idx_sorted, mean_block = L)
          } else {
            .circular_block_permute(te_idx_sorted, block_len = L)
          }
          stopifnot(length(perm_idx) == length(te_idx_sorted))
          if (any(!perm_idx %in% te_idx_sorted)) {
            stop(".stationary_bootstrap/.circular_block_permute must return a permutation of the provided indices.")
          }
          perm_time <- y_all[perm_idx]
          pos <- match(te_idx, te_idx_sorted)
          if (anyNA(pos)) {
            stop("Failed to align permuted time-series labels to fold order.", call. = FALSE)
          }
          perm_time[pos]
        },
        {
          .safe_sample(y_all[te_idx])
        }
      )
      res[[i]] <- permuted
    }
    res
  }
}

test_that("quantile break cache and majority level helpers behave", {
  vals <- c(1, 2, 3, 4, NA)
  br1 <- bioLeak:::.get_cached_quantile_breaks(vals, probs = c(0, 0.5, 1))
  br2 <- bioLeak:::.get_cached_quantile_breaks(vals, probs = c(0, 0.5, 1))
  expect_equal(br1, br2)

  expect_equal(bioLeak:::.majority_level(c("a", "b", "a", NA)), "a")
  expect_true(is.na(bioLeak:::.majority_level(c(NA, NA))))
})

test_that("grouped and within-group permutation functions preserve values", {
  set.seed(1)
  y <- c(1, 2, 3, 4, 5, 6)
  subj <- c("s1", "s1", "s2", "s2", "s3", "s3")
  perm <- bioLeak:::.permute_subject_grouped(y, subj)
  expect_equal(sort(perm), sort(y))

  group <- c("g1", "g1", "g2", "g2", "g2", "g1")
  perm2 <- bioLeak:::.permute_within_group(y, group)
  expect_equal(sort(perm2), sort(y))

  perm3 <- bioLeak:::.permute_within_batch(y, group)
  perm4 <- bioLeak:::.permute_within_study(y, group)
  expect_equal(sort(perm3), sort(y))
  expect_equal(sort(perm4), sort(y))
})

test_that("permute_labels_factory returns per-fold permutations", {
  set.seed(1)
  df <- make_class_df(20)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 4, stratify = FALSE, seed = 1)
  perm_fun <- bioLeak:::.permute_labels_factory(
    cd = df, outcome = "outcome", mode = "subject_grouped",
    folds = splits@indices, perm_stratify = TRUE,
    time_block = "circular", block_len = 2, seed = 1,
    group_col = "subject", batch_col = "batch", study_col = "study"
  )
  out <- perm_fun(1)
  expect_equal(length(out), length(splits@indices))
  expect_equal(length(out[[1]]), length(splits@indices[[1]]$test))
})

test_that("permute_labels_factory warns for small numeric stratification", {
  df <- make_class_df(12)
  df$outcome <- rnorm(12)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "batch_blocked", batch = "batch",
                              v = 3, stratify = FALSE, seed = 1)
  perm_fun <- expect_warning_match(
    bioLeak:::.permute_labels_factory(
      cd = df, outcome = "outcome", mode = "batch_blocked",
      folds = splits@indices, perm_stratify = TRUE,
      time_block = "circular", block_len = 2, seed = 1,
      batch_col = "batch"
    ),
    "requires at least 20"
  )
  out <- perm_fun(1)
  expect_equal(length(out), length(splits@indices))
})

test_that("permute_labels_factory handles time-series permutations", {
  df <- make_class_df(30)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "time_series", time = "time",
                              v = 4, seed = 1)
  perm_fun <- bioLeak:::.permute_labels_factory(
    cd = df, outcome = "outcome", mode = "time_series",
    folds = splits@indices, perm_stratify = FALSE,
    time_block = "stationary", block_len = 3, seed = 1,
    time_col = "time"
  )
  out <- perm_fun(1)
  expect_equal(length(out), length(splits@indices))
  expect_equal(length(out[[1]]), length(splits@indices[[1]]$test))
})

test_that("time-series permutations respect time order", {
  df <- make_class_df(20)
  df$time <- rev(seq_len(nrow(df)))
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "time_series", time = "time",
                              v = 3, seed = 1, compact = TRUE)
  folds <- splits@indices
  attr(folds, "fold_assignments") <- splits@info$fold_assignments
  fold <- folds[[1]]
  test_idx <- which(splits@info$fold_assignments[[1]] == fold$fold)
  block_len <- length(test_idx)

  perm_fun <- bioLeak:::.permute_labels_factory(
    cd = df, outcome = "outcome", mode = "time_series",
    folds = folds, perm_stratify = FALSE,
    time_block = "circular", block_len = block_len, seed = 1,
    time_col = "time"
  )
  out <- perm_fun(1)

  time_order <- test_idx[order(df$time[test_idx], test_idx)]
  set.seed(2)
  perm_idx <- bioLeak:::.circular_block_permute(time_order, block_len = block_len)
  perm_time <- df$outcome[perm_idx]
  expected <- perm_time[match(test_idx, time_order)]
  expect_equal(out[[1]], expected)
})

test_that("permute_labels_factory errors on invalid metadata", {
  df <- data.frame(outcome = c(NA, NA), x = 1:2)
  splits <- list(list(test = 1:2, train = integer(0)))
  expect_error(bioLeak:::.permute_labels_factory(
    cd = df, outcome = "outcome", mode = "subject_grouped",
    folds = splits, perm_stratify = FALSE,
    time_block = "circular", block_len = 2, seed = 1
  ), "only NA")
})

# Utilities ---------------------------------------------------------------

`%||%` <- function(a, b) if (!is.null(a)) a else b

.guard_is_binary <- function(y) {
  if (is.null(y)) return(FALSE)
  if (is.factor(y)) return(nlevels(y) == 2)
  ux <- unique(y[!is.na(y)])
  length(ux) == 2
}

.guard_majority <- function(x) {
  x <- x[!is.na(x)]
  if (length(x) == 0) return(NA)
  ux <- unique(x)
  ux[which.max(tabulate(match(x, ux)))]
}

.guard_mad_winsorize <- function(x, k = 5, center = NULL, scale = NULL) {
  if (!is.numeric(x)) return(x)
  if (all(!is.finite(x))) return(x)
  m <- if (is.null(center)) stats::median(x, na.rm = TRUE) else center
  s <- if (is.null(scale)) stats::mad(x, na.rm = TRUE, constant = 1) else scale
  if (!is.finite(m) || !is.finite(s) || s == 0) return(x)
  lo <- m - k * s; hi <- m + k * s
  x[x < lo] <- lo
  x[x > hi] <- hi
  x
}

.guard_align_cols <- function(X, ref_cols) {
  X <- as.data.frame(X, check.names = FALSE)
  missing_cols <- setdiff(ref_cols, names(X))
  if (length(missing_cols)) {
    for (nm in missing_cols) X[[nm]] <- NA_real_
  }
  extra_cols <- setdiff(names(X), ref_cols)
  if (length(extra_cols)) {
    X <- X[, setdiff(names(X), extra_cols), drop = FALSE]
  }
  X[, ref_cols, drop = FALSE]
}

.guard_fill_na <- function(X, fill) {
  X <- as.data.frame(X, check.names = FALSE)
  for (nm in names(fill)) {
    if (!nm %in% names(X)) next
    idx <- is.na(X[[nm]])
    if (any(idx)) X[[nm]][idx] <- fill[[nm]]
  }
  X
}

.guard_knn_search <- function(train, query, k) {
  train <- as.matrix(train)
  query <- as.matrix(query)
  if (!nrow(train) || !nrow(query) || k < 1L) {
    return(matrix(integer(0), nrow = nrow(query), ncol = 0))
  }
  k <- min(k, nrow(train))
  if (requireNamespace("RANN", quietly = TRUE)) {
    nn <- RANN::nn2(train, query, k = k)
    return(nn$nn.idx)
  }
  if (requireNamespace("FNN", quietly = TRUE)) {
    nn <- FNN::get.knnx(train, query, k = k)
    return(nn$nn.index)
  }
  idx <- matrix(NA_integer_, nrow(query), k)
  for (i in seq_len(nrow(query))) {
    diff <- sweep(train, 2, query[i, ], "-")
    d <- rowSums(diff * diff)
    ord <- order(d, na.last = NA)
    idx[i, ] <- ord[seq_len(k)]
  }
  idx
}

.guard_knn_impute_train <- function(X, k = 5) {
  X <- as.data.frame(X, check.names = FALSE)
  med <- vapply(X, function(col) {
    m <- stats::median(col, na.rm = TRUE)
    if (!is.finite(m)) 0 else m
  }, numeric(1))
  X_fill <- .guard_fill_na(X, med)
  scale <- vapply(X_fill, stats::sd, numeric(1))
  scale[!is.finite(scale) | scale == 0] <- 1
  X_scaled <- sweep(as.matrix(X_fill), 2, scale, "/")

  n <- nrow(X_fill)
  if (n < 2L) {
    return(list(imp = X_fill, med = med, scale = scale))
  }

  idx <- .guard_knn_search(X_scaled, X_scaled, k = min(k + 1L, n))
  X_imp <- X_fill
  for (i in seq_len(n)) {
    miss_cols <- which(is.na(X[i, ]))
    if (!length(miss_cols)) next
    neigh <- idx[i, ]
    neigh <- neigh[neigh != i]
    if (!length(neigh)) next
    if (length(neigh) > k) neigh <- neigh[seq_len(k)]
    for (j in miss_cols) {
      val <- mean(X_fill[neigh, j], na.rm = TRUE)
      if (!is.finite(val)) val <- med[j]
      X_imp[i, j] <- val
    }
  }
  list(imp = X_imp, med = med, scale = scale)
}

.guard_knn_impute_new <- function(Xnew, ref, k, med, scale) {
  Xnew <- .guard_align_cols(Xnew, names(med))
  Xnew_raw <- Xnew
  Xnew_fill <- .guard_fill_na(Xnew, med)
  if (!nrow(ref)) return(Xnew_fill)

  ref_scaled <- sweep(as.matrix(ref), 2, scale, "/")
  new_scaled <- sweep(as.matrix(Xnew_fill), 2, scale, "/")
  idx <- .guard_knn_search(ref_scaled, new_scaled, k = min(k, nrow(ref)))

  X_imp <- Xnew_fill
  for (i in seq_len(nrow(Xnew_fill))) {
    miss_cols <- which(is.na(Xnew_raw[i, ]))
    if (!length(miss_cols)) next
    neigh <- idx[i, ]
    if (!length(neigh)) next
    for (j in miss_cols) {
      val <- mean(ref[neigh, j], na.rm = TRUE)
      if (!is.finite(val)) val <- med[j]
      X_imp[i, j] <- val
    }
  }
  X_imp
}

.guard_rf_impute_train <- function(X, maxiter = 1, ntree = 100, seed = NULL) {
  X <- as.data.frame(X, check.names = FALSE)
  if (!requireNamespace("randomForest", quietly = TRUE)) {
    stop("Package 'randomForest' is required for impute$method='missForest'.", call. = FALSE)
  }
  med <- vapply(X, function(col) {
    m <- stats::median(col, na.rm = TRUE)
    if (!is.finite(m)) 0 else m
  }, numeric(1))
  X_fill <- .guard_fill_na(X, med)
  miss_cols <- names(which(vapply(X, function(col) any(is.na(col)), logical(1))))
  models <- list()

  if (!length(miss_cols)) {
    return(list(imp = X_fill, med = med, models = models))
  }

  for (iter in seq_len(maxiter)) {
    for (col in miss_cols) {
      obs <- !is.na(X[[col]])
      if (sum(obs) < 2L) next
      pred_cols <- setdiff(names(X_fill), col)
      if (!length(pred_cols)) next
      df_obs <- data.frame(y = X_fill[obs, col],
                           X_fill[obs, pred_cols, drop = FALSE],
                           check.names = FALSE)
      if (!is.null(seed)) set.seed(seed + iter)
      rf <- randomForest::randomForest(y ~ ., data = df_obs, ntree = ntree)
      if (iter == maxiter) {
        models[[col]] <- list(model = rf, predictors = pred_cols)
      }
      if (any(!obs)) {
        newdata <- X_fill[!obs, pred_cols, drop = FALSE]
        X_fill[!obs, col] <- stats::predict(rf, newdata = newdata)
      }
    }
  }

  med_out <- vapply(X_fill, function(col) {
    m <- stats::median(col, na.rm = TRUE)
    if (!is.finite(m)) 0 else m
  }, numeric(1))
  list(imp = X_fill, med = med_out, models = models)
}

.guard_rf_impute_new <- function(Xnew, models, med) {
  Xnew <- .guard_align_cols(Xnew, names(med))
  Xnew_raw <- Xnew
  Xnew_fill <- .guard_fill_na(Xnew, med)

  if (length(models)) {
    for (col in names(models)) {
      if (!col %in% names(Xnew_fill)) next
      idx <- is.na(Xnew_raw[[col]])
      if (!any(idx)) next
      pred_cols <- models[[col]]$predictors
      missing_pred <- setdiff(pred_cols, names(Xnew_fill))
      if (length(missing_pred)) {
        for (nm in missing_pred) Xnew_fill[[nm]] <- med[[nm]]
      }
      newdata <- Xnew_fill[idx, pred_cols, drop = FALSE]
      Xnew_fill[idx, col] <- stats::predict(models[[col]]$model, newdata = newdata)
    }
  }

  Xnew_fill <- .guard_fill_na(Xnew_fill, med)
  Xnew_fill
}

.guard_hash <- function(obj) {
  if (requireNamespace("digest", quietly = TRUE)) {
    return(digest::digest(obj))
  }
  raw <- serialize(obj, connection = NULL)
  val <- sum(as.numeric(raw) * seq_along(raw)) %% .Machine$integer.max
  paste0("h", sprintf("%08X", as.integer(val)))
}

#' Ensure consistent categorical levels for guarded preprocessing
#'
#' Converts character/logical columns to factors and aligns factor levels with
#' a training-time \code{levels_map}. Adds a dummy level when a column has only
#' one observed level so that downstream one-hot encoding retains a column.
#'
#' @param df data.frame to normalize factor levels.
#' @param levels_map optional named list of factor levels learned from training data.
#' @param dummy_prefix prefix used when adding a dummy level to single-level factors.
#' @return List with elements \code{data} (data.frame) and \code{levels} (named list of levels).
#' @keywords internal
#' @export
.guard_ensure_levels <- function(df, levels_map = NULL, dummy_prefix = "__dummy__") {
  stopifnot(is.data.frame(df))
  if (!is.null(levels_map)) {
    missing_cols <- setdiff(names(levels_map), names(df))
    if (length(missing_cols)) {
      new_cols_df <- as.data.frame(lapply(levels_map[missing_cols], function(lvls) {
        factor(rep(NA_character_, nrow(df)), levels = lvls)
      }))
      df <- cbind(df, new_cols_df)
    }
  }

  cat_cols <- names(df)[!vapply(df, is.numeric, logical(1))]
  if (!length(cat_cols) && is.null(levels_map)) {
    return(list(data = df, levels = list()))
  }

  if (is.null(levels_map)) levels_map <- list()

  for (nm in cat_cols) {
    col <- df[[nm]]
    if (is.logical(col)) {
      col <- factor(col, levels = c(FALSE, TRUE))
    } else if (!is.factor(col)) {
      col <- factor(col)
    }

    lvls <- levels_map[[nm]]
    if (is.null(lvls)) {
      lvls <- levels(col)
      if (length(lvls) <= 1) {
        dummy <- paste0(dummy_prefix, nm)
        while (dummy %in% lvls) dummy <- paste0(dummy, "_")
        lvls <- unique(c(lvls, dummy))
      }
    }

    col <- factor(col, levels = lvls)
    df[[nm]] <- col
    levels_map[[nm]] <- lvls
  }

  list(data = df, levels = levels_map)
}

#' @title Fit leakage-safe preprocessing pipeline
#' @description Builds and fits a guarded preprocessing pipeline on training data,
#' then constructs a transformer for consistent application to new data.
#' @details
#' The pipeline applies, in order:
#' \itemize{
#'   \item Winsorization (optional) to limit outliers.
#'   \item Imputation learned on training data only.
#'   \item Normalization (z-score or robust).
#'   \item Variance/IQR filtering.
#'   \item Feature selection (optional; t-test, lasso, PCA).
#' }
#' All statistics are estimated on the training data and re-used for new data.
#' @param X matrix/data.frame of predictors (training).
#' @param y Optional outcome for supervised feature selection.
#' @param steps List of configuration options (see Details).
#' @param task "binomial", "multiclass", "gaussian", or "survival".
#' @return An object of class "GuardFit" with elements `transform`, `state`, `p_out`, and `steps`.
#' @seealso [predict_guard()]
#' @examples
#' x <- data.frame(a = c(1, 2, NA), b = c(3, 4, 5))
#' fit <- .guard_fit(x, y = c(1, 2, 3),
#'                   steps = list(impute = list(method = "median")),
#'                   task = "gaussian")
#' fit$transform(x)
#' @export
.guard_fit <- function(X, y = NULL, steps = list(),
                       task = c("binomial", "multiclass", "gaussian", "survival")) {
  task <- match.arg(task)
  # Coerce to data.frame for flexible handling
  if (is.matrix(X)) X <- as.data.frame(X, check.names = FALSE)
  stopifnot(is.data.frame(X))
  # --- Coerce character columns to numeric when they look numeric ---
  for (nm in names(X)) {
    if (is.character(X[[nm]])) {
      nz <- X[[nm]][!is.na(X[[nm]])]
      if (length(nz) && all(grepl("^\\s*-?\\d+(\\.\\d+)?\\s*$", nz))) {
        X[[nm]] <- as.numeric(X[[nm]])
      } else {
        X[[nm]] <- as.factor(X[[nm]])
      }
    }
  }

  n0 <- nrow(X)

  # Keep original column names for diagnostics
  orig_colnames <- colnames(X)

  # 0) Mixed-type handling: one-hot encode non-numeric columns -----------------
  if (!ncol(X)) stop("Input X has no columns after preprocessing.")
  all_numeric <- all(vapply(X, is.numeric, logical(1)))
  if (all_numeric) {
    encoding_levels <- list()
    X <- as.data.frame(X, check.names = FALSE)
    p_after_encode <- ncol(X)
  } else {
    prep <- .guard_ensure_levels(X)
    X <- prep$data
    if (!ncol(X)) stop("Input X has no columns after preprocessing.")
    encoding_levels <- prep$levels
    mf <- stats::model.frame(~ . , data = as.data.frame(X), na.action = stats::na.pass)
    mm <- stats::model.matrix(~ . - 1, data = mf)  # preserve row count
    X <- as.data.frame(mm, check.names = FALSE)
    p_after_encode <- ncol(X)
  }

  # Replace non-finite with NA to unify missingness treatment
  for (j in seq_len(ncol(X))) {
    X[[j]][!is.finite(X[[j]])] <- NA
  }

  state <- list()
  state$encoding <- list(levels = encoding_levels, has_factors = !all_numeric)
  audit <- list()

  # 1) Robust Winsorization (train-only) --------------------------------------
  impute_cfg <- steps$impute %||% list()
  winsor_on <- impute_cfg$winsor %||% TRUE
  winsor_k  <- impute_cfg$winsor_k %||% 5

  if (isTRUE(winsor_on)) {
    winsor_med <- vapply(X, function(col) {
      if (is.numeric(col)) stats::median(col, na.rm = TRUE) else NA_real_
    }, numeric(1))
    winsor_mad <- vapply(X, function(col) {
      if (is.numeric(col)) stats::mad(col, na.rm = TRUE, constant = 1) else NA_real_
    }, numeric(1))
    X <- as.data.frame(Map(function(col, med, mad) {
      .guard_mad_winsorize(col, k = winsor_k, center = med, scale = mad)
    }, X, winsor_med, winsor_mad),
                       check.names = FALSE)
    state$winsor <- list(enabled = TRUE, k = winsor_k, med = winsor_med, mad = winsor_mad)
    audit <- c(audit, list(paste0("winsor: k=", winsor_k)))
  } else {
    state$winsor <- list(enabled = FALSE)
  }

  # 2) Imputation (train-only fitting) ----------------------------------------
  imp_method <- impute_cfg$method %||% "median"

  has_na <- any(vapply(X, function(col) any(is.na(col)), logical(1)))


  if (!has_na) {
    state$impute <- list(method = "none")
    state$impute$label <- "impute: none (no NA)"
  } else if (imp_method == "median") {
    meds <- vapply(X, function(col) stats::median(col, na.rm = TRUE), numeric(1))
    for (j in seq_len(ncol(X))) {
      idx <- is.na(X[[j]])
      if (any(idx)) X[[j]][idx] <- meds[j]
    }
    state$impute <- list(method = "median", med = meds)

  } else if (imp_method == "knn") {
    k <- impute_cfg$k %||% 5
    knn_fit <- .guard_knn_impute_train(X, k = k)
    X <- knn_fit$imp
    state$impute <- list(
      method = "knn",
      k = k,
      med = knn_fit$med,
      knn_ref = knn_fit$imp,
      knn_scale = knn_fit$scale
    )

  } else if (imp_method == "mice") {
    stop("impute$method='mice' is not supported for guarded prediction; use 'knn', 'missForest', or 'median'.",
         call. = FALSE)

  } else if (imp_method == "missForest") {
    rf_maxiter <- impute_cfg$maxiter %||% 10
    rf_ntree <- impute_cfg$ntree %||% 100
    rf_fit <- .guard_rf_impute_train(X, maxiter = rf_maxiter, ntree = rf_ntree)
    X <- rf_fit$imp
    state$impute <- list(
      method = "missForest",
      maxiter = rf_maxiter,
      ntree = rf_ntree,
      med = rf_fit$med,
      rf_models = rf_fit$models
    )
  } else if (imp_method == "none") {
    orig_cols <- colnames(X)
    med <- vapply(orig_cols, function(nm) {
      col <- X[[nm]]
      if (all(is.na(col))) return(0)
      stats::median(col, na.rm = TRUE)
    }, numeric(1))

    indicator_map <- list()
    fallback_used <- FALSE
    for (nm in orig_cols) {
      col <- X[[nm]]
      miss_idx <- is.na(col)
      if (!any(miss_idx)) next
      fallback_used <- TRUE
      ind_name <- paste0(nm, "__missing")
      indicator_map[[nm]] <- ind_name
      X[[ind_name]] <- as.numeric(miss_idx)
      fill_val <- med[[nm]]
      if (!is.finite(fill_val)) fill_val <- 0
      if (!any(!miss_idx)) {
        # Column is entirely NA; keep indicator but set baseline to 0
        X[[nm]] <- rep(fill_val, length(col))
      } else {
        X[[nm]][miss_idx] <- fill_val
      }
    }

    state$impute <- list(
      method = "none",
      fallback = if (fallback_used) "median" else NULL,
      med = med,
      indicators = indicator_map,
      base_cols = orig_cols
    )

    if (fallback_used) {
      warning("Missing values detected with impute$method='none'; applying median fallback with missingness indicators to ensure compatibility.",
              call. = FALSE)
      state$impute$label <- "impute: none (fallback median)"
    } else {
      state$impute$label <- "impute: none"
    }
  } else {
    stop("Unknown impute method. Use 'median', 'knn', 'missForest', or 'none'.")
  }

  state$impute$order <- colnames(X)
  if (is.null(state$impute$label))
    state$impute$label <- paste0("impute: ", state$impute$method)

  audit <- c(audit, list(state$impute$label))

  # 3) Normalization -----------------------------------------------------------
  norm_cfg <- steps$normalize %||% list()
  norm_method <- norm_cfg$method %||% "zscore"

  if (norm_method == "zscore") {
    mu <- vapply(X, mean, numeric(1))
    sdv <- vapply(X, stats::sd, numeric(1))
    sdv[sdv == 0 | !is.finite(sdv)] <- 1
    X <- sweep(X, 2, mu, "-")
    X <- sweep(X, 2, sdv, "/")
    state$normalize <- list(method = "zscore", mu = mu, sd = sdv)

  } else if (norm_method == "robust") {
    mu <- vapply(X, stats::median, numeric(1))
    sdv <- vapply(X, stats::mad, numeric(1))
    sdv[sdv < 1e-12 | !is.finite(sdv)] <- 1
    X <- sweep(X, 2, mu, "-")
    X <- sweep(X, 2, sdv, "/")
    state$normalize <- list(method = "robust", mu = mu, sd = sdv)

  } else if (norm_method == "none") {
    state$normalize <- list(method = "none")
  } else {
    stop("Unknown normalize method. Use 'zscore', 'robust', or 'none'.")
  }
  audit <- c(audit, list(paste0("normalize: ", state$normalize$method)))

  # 4) Filtering (variance / IQR) ---------------------------------------------
  filt_cfg <- steps$filter %||% list()
  var_th   <- filt_cfg$var_thresh %||% 0
  iqr_th   <- filt_cfg$iqr_thresh %||% 0
  min_keep <- filt_cfg$min_keep %||% NULL

  keep <- rep(TRUE, ncol(X))
  names(keep) <- colnames(X)
  if (var_th > 0) {
    v <- vapply(X, stats::var, numeric(1))
    keep <- keep & (v >= var_th)
  }
  if (iqr_th > 0) {
    i <- vapply(X, stats::IQR, numeric(1))
    keep <- keep & (i >= iqr_th)
  }

  # Adaptive: ensure at least min_keep variables remain
  if (!is.null(min_keep)) {
    stopifnot(is.numeric(min_keep), length(min_keep) == 1L, min_keep >= 1L)
    min_keep <- min(min_keep, ncol(X))
    if (sum(keep) < min_keep) {
      # Rank by variance then IQR to recover features until min_keep
      v <- vapply(X, stats::var, numeric(1))
      i <- vapply(X, stats::IQR, numeric(1))
      ord <- order(v + i, decreasing = TRUE)
      keep[ord[seq_len(min_keep)]] <- TRUE
    }
  }

  X <- X[, keep, drop = FALSE]
  state$filter <- list(var_thresh = var_th, iqr_thresh = iqr_th, keep = keep)
  audit <- c(audit, list(paste0("filter: var>=", var_th, ", iqr>=", iqr_th,
                                if (!is.null(min_keep)) paste0(", min_keep=", min_keep) else "")))

  # 5) Feature Selection -------------------------------------------------------
  fs_cfg <- steps$fs %||% list()
  fs_method <- fs_cfg$method %||% "none"
  selected <- seq_len(ncol(X))

  if (fs_method == "ttest") {
    if (!.guard_is_binary(y))
      stop("fs='ttest' requires a binary outcome 'y'.")
    yy <- as.factor(y)
    if (nlevels(yy) != 2) stop("fs='ttest' requires y with exactly 2 levels.")

    g1 <- X[which(yy == levels(yy)[2]), , drop = FALSE]
    g0 <- X[which(yy == levels(yy)[1]), , drop = FALSE]
    if (nrow(g1) < 2 || nrow(g0) < 2)
      stop("Not enough samples per class for t-test feature selection.")

    m1 <- vapply(g1, mean, numeric(1)); m0 <- vapply(g0, mean, numeric(1))
    s1 <- vapply(g1, stats::sd, numeric(1)); s0 <- vapply(g0, stats::sd, numeric(1))
    n1 <- nrow(g1); n0 <- nrow(g0)
    se <- sqrt(s1^2 / n1 + s0^2 / n0); se[se == 0] <- Inf
    tstat <- (m1 - m0) / se
    k <- if (!is.null(fs_cfg$top_k) && is.numeric(fs_cfg$top_k) && fs_cfg$top_k >= 1L)
      as.integer(fs_cfg$top_k)
    else
      max(50L, ceiling(0.10 * length(tstat)))
    ord <- order(abs(tstat), decreasing = TRUE)
    selected <- ord[seq_len(min(k, length(ord)))]
    X <- X[, selected, drop = FALSE]
    state$fs <- list(method = "ttest", sel = selected, sel_names = colnames(X))

  } else if (fs_method == "lasso") {
    if (!requireNamespace("glmnet", quietly = TRUE))
      stop("Package 'glmnet' is required for fs='lasso'.", call. = FALSE)
    fam <- if (task == "binomial") "binomial"
    else if (task == "multiclass") "multinomial"
    else if (task == "survival") "cox"
    else "gaussian"

    # guard against constant columns
    nonconst <- vapply(X, stats::sd, numeric(1)) > 1e-8
    if (!all(nonconst)) X <- X[, nonconst, drop = FALSE]

    fit0 <- glmnet::glmnet(as.matrix(X), y, family = fam, alpha = 1, standardize = FALSE)
    if (identical(fam, "multinomial") && is.list(fit0$beta)) {
      beta_list <- fit0$beta
      take <- ncol(beta_list[[1]])
      scores <- Reduce(`+`, lapply(beta_list, function(b) {
        abs(b[, take, drop = FALSE])
      }))
      scores <- as.numeric(scores)
      nz <- which(scores != 0)
      if (!length(nz)) {
        nz <- order(scores, decreasing = TRUE)[1:min(50, length(scores))]
      }
      selected <- nz
    } else {
      beta <- as.matrix(fit0$beta)
      take <- ncol(beta)
      nz <- which(beta[, take] != 0)
      if (length(nz) == 0) {
        nz <- order(rowSums(abs(beta)), decreasing = TRUE)[1:min(50, nrow(beta))]
      }
      selected <- nz
    }
    X <- X[, selected, drop = FALSE]
    state$fs <- list(method = "lasso", sel = selected, sel_names = colnames(X))

  } else if (fs_method == "pca") {
    p <- fs_cfg$pca_comp %||% min(50, ncol(X))
    pc <- stats::prcomp(X, center = FALSE, scale. = FALSE)
    X <- as.data.frame(pc$x[, seq_len(min(p, ncol(pc$x))), drop = FALSE])
    state$fs <- list(method = "pca",
                     p = p,
                     rotation = pc$rotation[, seq_len(min(p, ncol(pc$rotation))), drop = FALSE])

  } else if (fs_method == "none") {
    state$fs <- list(method = "none", sel = selected)
  } else {
    stop("Unknown fs method. Use 'none','ttest','lasso','pca'.")
  }
  audit <- c(audit, list(paste0("fs: ", state$fs$method)))

  # 6) Diagnostics, audit, hash -----------------------------------------------
  state$diagnostics <- list(
    n_train = n0,
    p_in    = length(orig_colnames),
    p_after_encode = p_after_encode,
    p_out   = ncol(X),
    removed_by_filter = sum(!state$filter$keep)
  )
  state$audit <- data.frame(
    step = seq_along(audit),
    action = as.character(unlist(audit)),
    stringsAsFactors = FALSE
  )
  impute_hash <- state$impute
  if (!is.null(impute_hash$knn_ref)) impute_hash$knn_ref <- NULL
  if (!is.null(impute_hash$rf_models)) impute_hash$rf_models <- NULL
  state$hash  <- .guard_hash(list(impute = impute_hash, normalize = state$normalize,
                                  filter = list(var_th = var_th, iqr_th = iqr_th),
                                  fs = state$fs))

  # 7) Build transformer (apply to new data) ----------------------------------
  transformer <- function(Xnew) {
    if (is.matrix(Xnew)) Xnew <- as.data.frame(Xnew, check.names = FALSE)
    stopifnot(is.data.frame(Xnew))

    # Handle mixed types with the same level structure learned during training
    has_factors <- isTRUE(state$encoding$has_factors) || length(state$encoding$levels) > 0
    if (has_factors) {
      prep_new <- .guard_ensure_levels(Xnew, state$encoding$levels)
      Xnew <- prep_new$data
      if (!ncol(Xnew)) stop("Input X has no columns after preprocessing.")
      mf <- stats::model.frame(~ ., data = as.data.frame(Xnew), na.action = stats::na.pass)
      Xnew <- stats::model.matrix(~ . - 1, data = mf)
      Xnew <- as.data.frame(Xnew, check.names = FALSE)
    } else {
      for (nm in names(Xnew)) {
        if (is.character(Xnew[[nm]])) {
          nz <- Xnew[[nm]][!is.na(Xnew[[nm]])]
          if (length(nz) && all(grepl("^\\s*-?\\d+(\\.\\d+)?\\s*$", nz))) {
            Xnew[[nm]] <- as.numeric(Xnew[[nm]])
          }
        }
      }
      if (!all(vapply(Xnew, is.numeric, logical(1)))) {
        stop("Non-numeric columns detected in new data; guarded fit was trained on numeric predictors only.")
      }
      if (!ncol(Xnew)) stop("Input X has no columns after preprocessing.")
      Xnew <- as.data.frame(Xnew, check.names = FALSE)
    }

    # Replace non-finite
    for (j in seq_len(ncol(Xnew))) {
      Xnew[[j]][!is.finite(Xnew[[j]])] <- NA
    }

    # Winsorization: apply train-fitted centers/scales to new data
    if (isTRUE(state$winsor$enabled)) {
      k <- state$winsor$k
      w_med <- state$winsor$med
      w_mad <- state$winsor$mad
      if (!is.null(w_med) && !is.null(w_mad)) {
        common <- intersect(names(w_med), colnames(Xnew))
        for (nm in common) {
          Xnew[[nm]] <- .guard_mad_winsorize(Xnew[[nm]], k = k,
                                             center = w_med[[nm]],
                                             scale = w_mad[[nm]])
        }
      } else {
        warning("Winsorization statistics missing; skipping winsorization to avoid leakage.",
                call. = FALSE)
      }
    }

    # Impute using TRAIN-FITTED parameters
    if (identical(state$impute$method, "none") && !is.null(state$impute$fallback)) {
      base_cols <- state$impute$base_cols %||% character(0)
      med <- state$impute$med %||% numeric(0)
      ind_map <- state$impute$indicators %||% list()

      # Ensure all training columns exist in new data
      missing_base <- setdiff(base_cols, colnames(Xnew))
      if (length(missing_base) > 0) {
        for (mc in missing_base) Xnew[[mc]] <- NA_real_
      }

      # Drop columns unseen during training imputation step
      expected_cols <- unique(c(base_cols, unlist(ind_map, use.names = FALSE)))
      extra_cols <- setdiff(colnames(Xnew), expected_cols)
      if (length(extra_cols) > 0)
        Xnew <- Xnew[, setdiff(colnames(Xnew), extra_cols), drop = FALSE]

      # Apply fallback imputation and create indicators
      for (nm in base_cols) {
        if (!nm %in% colnames(Xnew)) next
        idx <- is.na(Xnew[[nm]])
        ind_name <- ind_map[[nm]]
        if (!is.null(ind_name)) {
          Xnew[[ind_name]] <- as.numeric(idx)
        }
        fill_val <- med[[nm]]
        if (is.null(fill_val) || !is.finite(fill_val)) fill_val <- 0
        if (any(idx)) Xnew[[nm]][idx] <- fill_val
        if (!is.null(ind_name) && !ind_name %in% colnames(Xnew)) {
          Xnew[[ind_name]] <- 0
        }
      }

      # Ensure indicator columns exist even if no missing values in new data
      for (nm in base_cols) {
        ind_name <- ind_map[[nm]]
        if (!is.null(ind_name) && !ind_name %in% colnames(Xnew)) {
          Xnew[[ind_name]] <- 0
        }
      }

      # Reorder to match training order for downstream steps
      order_cols <- state$impute$order %||% colnames(Xnew)
      missing_ord <- setdiff(order_cols, colnames(Xnew))
      if (length(missing_ord) > 0) {
        for (mo in missing_ord) Xnew[[mo]] <- 0
      }
      Xnew <- Xnew[, order_cols, drop = FALSE]

    } else if (identical(state$impute$method, "knn")) {
      med <- state$impute$med
      ref <- state$impute$knn_ref
      scale <- state$impute$knn_scale
      k <- state$impute$k %||% 5
      if (is.null(ref) || is.null(scale) || is.null(med)) {
        warning("KNN imputation state missing; falling back to median imputation.",
                call. = FALSE)
        if (!is.null(med)) {
          Xnew <- .guard_align_cols(Xnew, names(med))
          Xnew <- .guard_fill_na(Xnew, med)
        }
      } else {
        Xnew <- .guard_knn_impute_new(Xnew, ref, k = k, med = med, scale = scale)
      }
    } else if (identical(state$impute$method, "missForest")) {
      med <- state$impute$med
      models <- state$impute$rf_models
      if (is.null(med)) {
        warning("missForest imputation state missing; skipping imputation.",
                call. = FALSE)
      } else if (is.null(models) || !length(models)) {
        warning("missForest models missing; falling back to median imputation.",
                call. = FALSE)
        Xnew <- .guard_align_cols(Xnew, names(med))
        Xnew <- .guard_fill_na(Xnew, med)
      } else {
        Xnew <- .guard_rf_impute_new(Xnew, models, med)
      }
    } else if (!identical(state$impute$method, "none")) {
      med <- state$impute$med
      # align columns if train/test encodings differ (missing columns -> fill with 0 then impute)
      missing_cols <- setdiff(names(med), colnames(Xnew))
      if (length(missing_cols) > 0) {
        for (mc in missing_cols) Xnew[[mc]] <- NA_real_
      }
      extra_cols <- setdiff(colnames(Xnew), names(med))
      if (length(extra_cols) > 0) {
        # drop unseen columns (keeps mapping to training space)
        Xnew <- Xnew[, setdiff(colnames(Xnew), extra_cols), drop = FALSE]
      }
      Xnew <- Xnew[, names(med), drop = FALSE]
      for (j in seq_along(med)) {
        idx <- is.na(Xnew[[j]])
        if (any(idx)) Xnew[[j]][idx] <- med[[j]]
      }
    }

    # Normalize with TRAIN-FITTED centers/scales
    if (identical(state$normalize$method, "zscore") || identical(state$normalize$method, "robust")) {
      mu <- state$normalize$mu; sdv <- state$normalize$sd
      # align columns
      missing_cols <- setdiff(names(mu), colnames(Xnew))
      if (length(missing_cols) > 0) for (mc in missing_cols) Xnew[[mc]] <- 0
      extra_cols <- setdiff(colnames(Xnew), names(mu))
      if (length(extra_cols) > 0) Xnew <- Xnew[, names(mu), drop = FALSE]
      Xnew <- sweep(Xnew, 2, mu, "-")
      Xnew <- sweep(Xnew, 2, sdv, "/")
    }

    # Filter: keep as per train
    keep <- state$filter$keep
    # keep may be indexed on pre-encoded space; rebuild logical selector by name
    keep_names <- names(keep)
    if (!is.null(keep_names)) {
      # ensure all expected columns exist and align order explicitly by name
      miss <- setdiff(keep_names, colnames(Xnew))
      if (length(miss) > 0) for (m in miss) Xnew[[m]] <- 0
      Xnew <- Xnew[, keep_names, drop = FALSE]
      keep <- keep[keep_names]
      Xnew <- Xnew[, keep, drop = FALSE]
    } else {
      # if names missing (rare), fall back to length match
      Xnew <- Xnew[, keep, drop = FALSE]
    }

    # Feature selection mapping
    if (state$fs$method %in% c("ttest", "lasso")) {
      sel_names <- state$fs$sel_names
      if (!is.null(sel_names)) {
        # Name-based selection: robust to upstream column removal (e.g. lasso
        # constant-column guard) that would invalidate positional indices.
        avail <- intersect(sel_names, colnames(Xnew))
        Xnew <- Xnew[, avail, drop = FALSE]
      } else {
        # Backward compatibility: positional fallback
        sel <- state$fs$sel
        sel <- sel[sel >= 1 & sel <= ncol(Xnew)]
        Xnew <- Xnew[, sel, drop = FALSE]
      }
    } else if (state$fs$method == "pca") {
      # Project using train PCA rotation
      R <- state$fs$rotation
      # align columns by names
      miss <- setdiff(rownames(R), colnames(Xnew))
      if (length(miss) > 0) for (m in miss) Xnew[[m]] <- 0
      Xnew <- Xnew[, rownames(R), drop = FALSE]
      Xnew <- as.matrix(Xnew) %*% as.matrix(R)
      Xnew <- as.data.frame(Xnew, check.names = FALSE)
    }

    Xnew
  }

  out <- structure(
    list(
      transform = transformer,
      state     = state,
      p_out     = ncol(X),
      steps     = list(
        impute = imp_method,
        normalize = norm_method,
        filter = list(var_thresh = var_th, iqr_thresh = iqr_th, min_keep = min_keep),
        fs = fs_method
      )
    ),
    class = "GuardFit"
  )
  out$hash <- state$hash
  out
}

# S3 helpers --------------------------------------------------------------

#' @export
print.GuardFit <- function(x, ...) {
  cat("Guarded preprocessing pipeline\n")
  cat(sprintf(" - Imputation: %s\n", x$steps$impute))
  cat(sprintf(" - Normalization: %s\n", x$steps$normalize))
  cat(sprintf(" - Filter: var>=%s, iqr>=%s\n",
              x$steps$filter$var_thresh, x$steps$filter$iqr_thresh))
  cat(sprintf(" - Feature selection: %s\n", x$steps$fs))
  cat(sprintf(" - Output features: %d\n", x$p_out))
  invisible(x)
}

#' @export
summary.GuardFit <- function(object, ...) {
  s <- object$state
  out <- list(
    diagnostics = s$diagnostics,
    audit = s$audit,
    hash  = s$hash
  )
  class(out) <- "summary.GuardFit"
  out
}

#' @export
print.summary.GuardFit <- function(x, ...) {
  cat("GuardFit summary\n")
  if (!is.null(x$diagnostics)) {
    d <- x$diagnostics
    cat(sprintf(" n_train=%s, p_out=%s, removed_by_filter=%s\n",
                d$n_train, d$p_out, d$removed_by_filter))
  }
  if (!is.null(x$audit)) {
    cat(" Steps:\n")
    print(x$audit, row.names = FALSE)
  }
  if (!is.null(x$hash))
    cat(sprintf(" hash: %s\n", x$hash))
  invisible(x)
}

# Convenience -------------------------------------------------------------

#' Apply a fitted GuardFit transformer to new data
#'
#' @description
#' Applies the preprocessing steps stored in a \code{GuardFit} object to new
#' data without refitting any statistics. This is designed to prevent
#' validation leakage that would occur if imputation, scaling, filtering, or
#' feature selection were recomputed on evaluation data. It enforces the
#' training schema by aligning columns and factor levels, and it errors when a
#' numeric-only training fit receives non-numeric predictors. It does not
#' detect label leakage, duplicate samples, or train/test contamination.
#'
#' @param fit A \code{GuardFit} object created by [.guard_fit()]. This required
#' argument (no default) contains the training-time preprocessing settings and
#' statistics. Changing \code{fit} (for example, a different imputation method
#' or feature selection step) changes the output columns and values.
#' @param newdata A matrix or data.frame of predictors with one row per sample.
#' This required argument (no default) is transformed using the training-time
#' parameters in \code{fit} only. Missing columns are added and filled, extra
#' columns are dropped, and factor levels are aligned to the training levels; if
#' the training fit was numeric-only, non-numeric columns in \code{newdata}
#' trigger an error.
#'
#' @return A data.frame of transformed predictors with the same number of rows
#' as \code{newdata}. Column order and content match the training pipeline and
#' may include derived features (one-hot encodings, missingness indicators, or
#' PCA components). This output is not a prediction; it is intended as input to
#' a downstream model and assumes the training-time preprocessing is valid for
#' the new data.
#'
#' @examples
#' x_train <- data.frame(a = c(1, 2, NA, 4), b = c(10, 11, 12, 13))
#' fit <- .guard_fit(
#'   x_train,
#'   y = c(0.1, 0.2, 0.3, 0.4),
#'   steps = list(impute = list(method = "median")),
#'   task = "gaussian"
#' )
#' x_new <- data.frame(a = c(NA, 5), b = c(9, 14))
#' out <- predict_guard(fit, x_new)
#' out
#' @export
predict_guard <- function(fit, newdata) {
  stopifnot(inherits(fit, "GuardFit"))
  fit$transform(newdata)
}

# tests/testthat/test-invariants.R
# Invariant tests: seed reproducibility, nested-CV independence,
# transform locality, split completeness, multi-axis edge cases

# --- Helper data ---
.make_invariant_df <- function(n_subjects = 10, per_subject = 2) {
  n <- n_subjects * per_subject
  set.seed(999)
  data.frame(
    subject = rep(seq_len(n_subjects), each = per_subject),
    batch   = rep(letters[seq_len(min(n_subjects, 5))], length.out = n),
    outcome = rbinom(n, 1, 0.5),
    x1      = rnorm(n),
    x2      = rnorm(n),
    stringsAsFactors = FALSE
  )
}

# ============================================================
# Seed reproducibility
# ============================================================

test_that("same seed -> identical splits (subject_grouped)", {
  df <- .make_invariant_df()
  s1 <- make_split_plan(df, outcome = "outcome",
                        mode = "subject_grouped", group = "subject",
                        v = 5, seed = 42, progress = FALSE)
  s2 <- make_split_plan(df, outcome = "outcome",
                        mode = "subject_grouped", group = "subject",
                        v = 5, seed = 42, progress = FALSE)
  expect_identical(s1@indices, s2@indices)
})

test_that("different seeds -> different splits", {
  df <- .make_invariant_df()
  s1 <- make_split_plan(df, outcome = "outcome",
                        mode = "subject_grouped", group = "subject",
                        v = 5, seed = 1, progress = FALSE)
  s2 <- make_split_plan(df, outcome = "outcome",
                        mode = "subject_grouped", group = "subject",
                        v = 5, seed = 99, progress = FALSE)
  # At least one fold should differ
  any_diff <- !identical(s1@indices, s2@indices)
  expect_true(any_diff)
})

test_that("same seed -> identical splits (batch_blocked)", {
  df <- .make_invariant_df()
  s1 <- make_split_plan(df, outcome = "outcome",
                        mode = "batch_blocked", batch = "batch",
                        v = 3, seed = 7, progress = FALSE)
  s2 <- make_split_plan(df, outcome = "outcome",
                        mode = "batch_blocked", batch = "batch",
                        v = 3, seed = 7, progress = FALSE)
  expect_identical(s1@indices, s2@indices)
})

test_that("same seed -> identical splits (combined mode)", {
  df <- .make_invariant_df()
  s1 <- suppressWarnings(make_split_plan(
    df, outcome = "outcome",
    mode = "combined",
    constraints = list(
      list(type = "subject", col = "subject"),
      list(type = "batch", col = "batch")
    ),
    v = 3, seed = 5, progress = FALSE
  ))
  s2 <- suppressWarnings(make_split_plan(
    df, outcome = "outcome",
    mode = "combined",
    constraints = list(
      list(type = "subject", col = "subject"),
      list(type = "batch", col = "batch")
    ),
    v = 3, seed = 5, progress = FALSE
  ))
  expect_identical(s1@indices, s2@indices)
})

test_that("same seed -> identical fit_resample metrics (custom GLM)", {
  df <- .make_invariant_df()
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, seed = 1, progress = FALSE)
  custom <- list(
    glm = list(
      fit = function(x, y, task, weights, ...) {
        suppressWarnings(
          stats::glm(y ~ ., data = as.data.frame(x),
                     family = stats::binomial(), weights = weights)
        )
      },
      predict = function(object, newdata, task, ...) {
        as.numeric(suppressWarnings(
          stats::predict(object, newdata = as.data.frame(newdata),
                         type = "response")
        ))
      }
    )
  )
  f1 <- fit_resample(df, outcome = "outcome", splits = splits,
                     learner = "glm", custom_learners = custom,
                     metrics = "auc", refit = FALSE, seed = 10)
  f2 <- fit_resample(df, outcome = "outcome", splits = splits,
                     learner = "glm", custom_learners = custom,
                     metrics = "auc", refit = FALSE, seed = 10)
  expect_identical(f1@metrics, f2@metrics)
})

# ============================================================
# Nested CV independence
# ============================================================

test_that("inner train indices do not intersect outer test indices", {
  df <- .make_invariant_df(n_subjects = 12, per_subject = 2)
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 4, nested = TRUE, seed = 1, progress = FALSE)
  inner <- splits@info$inner
  expect_true(!is.null(inner))

  for (i in seq_along(splits@indices)) {
    outer_test <- splits@indices[[i]]$test
    outer_train <- splits@indices[[i]]$train
    for (j in seq_along(inner[[i]])) {
      # Inner indices are local to outer_train; map to global
      inner_train_global <- outer_train[inner[[i]][[j]]$train]
      inner_test_global  <- outer_train[inner[[i]][[j]]$test]
      expect_length(intersect(inner_train_global, outer_test), 0)
    }
  }
})

test_that("inner test indices do not intersect outer test indices", {
  df <- .make_invariant_df(n_subjects = 12, per_subject = 2)
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 4, nested = TRUE, seed = 1, progress = FALSE)
  inner <- splits@info$inner

  for (i in seq_along(splits@indices)) {
    outer_test <- splits@indices[[i]]$test
    outer_train <- splits@indices[[i]]$train
    for (j in seq_along(inner[[i]])) {
      inner_test_global <- outer_train[inner[[i]][[j]]$test]
      expect_length(intersect(inner_test_global, outer_test), 0)
    }
  }
})

# ============================================================
# Transform locality
# ============================================================

test_that(".guard_fit zscore uses train stats, not test stats", {
  set.seed(1)
  n_train <- 50
  n_test <- 50
  # Use train mean=0, test mean=5 (within winsorization bounds)
  train_x <- matrix(rnorm(n_train * 2, mean = 0, sd = 1), ncol = 2,
                     dimnames = list(NULL, c("x1", "x2")))
  test_x  <- matrix(rnorm(n_test * 2, mean = 5, sd = 1), ncol = 2,
                     dimnames = list(NULL, c("x1", "x2")))
  train_y <- rbinom(n_train, 1, 0.5)

  guard_obj <- bioLeak:::.guard_fit(
    X = train_x, y = train_y,
    steps = list(
      impute = list(winsor = FALSE),
      normalize = list(method = "zscore")
    )
  )

  pred <- predict_guard(guard_obj, newdata = test_x)
  # After zscore normalization using train stats (mean~0, sd~1),
  # test values (mean~5) should be transformed to ~(5-0)/1 = ~5
  # If test stats were used, test mean would be ~0 after transform
  expect_true(mean(pred[, 1]) > 2)
  expect_true(mean(pred[, 2]) > 2)
})

# ============================================================
# Split completeness
# ============================================================

test_that("every sample appears in at least one test fold (subject_grouped)", {
  df <- .make_invariant_df()
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 5, seed = 1, progress = FALSE)
  all_test <- sort(unique(unlist(lapply(splits@indices, `[[`, "test"))))
  expect_equal(all_test, seq_len(nrow(df)))
})

test_that("every sample appears in at least one test fold (batch_blocked)", {
  df <- .make_invariant_df()
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "batch_blocked", batch = "batch",
                            v = 5, seed = 1, progress = FALSE)
  all_test <- sort(unique(unlist(lapply(splits@indices, `[[`, "test"))))
  expect_equal(all_test, seq_len(nrow(df)))
})

# ============================================================
# Multi-axis edge cases
# ============================================================

test_that("combined mode with 3 constraint axes works and has zero overlap on all 3", {
  set.seed(1)
  # Need enough unique levels per axis so that removing test-axis levels

  # still leaves training samples
  n_subj <- 30
  df <- data.frame(
    subject = rep(seq_len(n_subj), each = 3),
    batch   = rep(paste0("B", seq_len(n_subj)), each = 3),
    site    = rep(paste0("S", seq_len(n_subj)), each = 3),
    outcome = rbinom(n_subj * 3, 1, 0.5),
    x1      = rnorm(n_subj * 3),
    stringsAsFactors = FALSE
  )
  constraints <- list(
    list(type = "subject", col = "subject"),
    list(type = "batch", col = "batch"),
    list(type = "batch", col = "site")
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "combined", constraints = constraints,
                            v = 3, seed = 1, progress = FALSE)
  expect_true(length(splits@indices) > 0)

  for (fold in splits@indices) {
    for (col in c("subject", "batch", "site")) {
      tr_vals <- unique(df[[col]][fold$train])
      te_vals <- unique(df[[col]][fold$test])
      overlap <- intersect(tr_vals, te_vals)
      expect_equal(length(overlap), 0,
                   info = sprintf("overlap on '%s'", col))
    }
  }
})

test_that("perfectly confounded axes (1:1 subject-batch) still produce valid folds", {
  set.seed(1)
  n_subj <- 10
  df <- data.frame(
    subject = rep(1:n_subj, each = 2),
    batch   = rep(paste0("B", 1:n_subj), each = 2),
    outcome = rbinom(n_subj * 2, 1, 0.5),
    x1      = rnorm(n_subj * 2),
    stringsAsFactors = FALSE
  )
  constraints <- list(
    list(type = "subject", col = "subject"),
    list(type = "batch", col = "batch")
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "combined", constraints = constraints,
                            v = 5, seed = 1, progress = FALSE)
  expect_true(length(splits@indices) > 0)
  for (fold in splits@indices) {
    expect_length(intersect(df$subject[fold$train], df$subject[fold$test]), 0)
    expect_length(intersect(df$batch[fold$train], df$batch[fold$test]), 0)
  }
})

# ============================================================
# Mechanism summary in audit output
# ============================================================

test_that("summary(audit) output contains Mechanism Risk Assessment", {
  set.seed(1)
  df <- data.frame(
    subject = rep(1:6, each = 2),
    outcome = rbinom(12, 1, 0.5),
    x1 = rnorm(12),
    x2 = rnorm(12)
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, seed = 1, progress = FALSE)
  custom <- list(
    glm = list(
      fit = function(x, y, task, weights, ...) {
        suppressWarnings(
          stats::glm(y ~ ., data = as.data.frame(x),
                     family = stats::binomial(), weights = weights)
        )
      },
      predict = function(object, newdata, task, ...) {
        as.numeric(suppressWarnings(
          stats::predict(object, newdata = as.data.frame(newdata),
                         type = "response")
        ))
      }
    )
  )
  fit <- fit_resample(df, outcome = "outcome", splits = splits,
                      learner = "glm", custom_learners = custom,
                      metrics = "auc", refit = FALSE, seed = 1)
  audit <- audit_leakage(fit, metric = "auc", B = 5,
                         X_ref = df[, c("x1", "x2")], seed = 1)

  out <- capture.output(summary(audit))
  expect_true(any(grepl("Mechanism Risk Assessment", out)))
  expect_true(any(grepl("non_random_signal", out)))
})

test_that("guard_fit produces a GuardFit pipeline with median imputation", {
  X <- data.frame(a = c(1, 2, NA, 4), b = c(NA, 1, 2, 3))
  fit <- bioLeak:::.guard_fit(
    X, y = c(0, 1, 0, 1),
    steps = list(impute = list(method = "median"),
                 normalize = list(method = "zscore"),
                 filter = list(var_thresh = 0, iqr_thresh = 0),
                 fs = list(method = "none")),
    task = "binomial"
  )

  expect_true(inherits(fit, "GuardFit"))
  Xtr <- fit$transform(X)
  expect_true(is.data.frame(Xtr))
  expect_false(anyNA(Xtr))
})

test_that("guard_fit warns on impute none with missing values", {
  X <- data.frame(a = c(1, NA, 3), b = c(4, 5, NA))
  fit <- expect_warning_match(
    bioLeak:::.guard_fit(
      X, y = c(1, 0, 1),
      steps = list(impute = list(method = "none"),
                   normalize = list(method = "none"),
                   filter = list(var_thresh = 0, iqr_thresh = 0),
                   fs = list(method = "none")),
      task = "binomial"
    ),
    "impute\\$method='none'"
  )
  out <- fit$transform(X)
  expect_true(any(grepl("__missing", names(out))))
})

test_that("guard_fit supports knn imputation and fallback warnings", {
  X <- data.frame(a = c(1, NA, 3), b = c(4, 5, NA))
  fit <- bioLeak:::.guard_fit(
    X, y = c(1, 0, 1),
    steps = list(impute = list(method = "knn", k = 1),
                 normalize = list(method = "none"),
                 filter = list(var_thresh = 0, iqr_thresh = 0),
                 fs = list(method = "none")),
    task = "binomial"
  )

  env <- environment(fit$transform)
  env$state$impute$knn_ref <- NULL
  out <- expect_warning_match(fit$transform(X), "KNN imputation state missing")
})

test_that("guard_fit validates imputation method values", {
  X <- data.frame(a = c(1, NA), b = c(2, 3))
  expect_error(bioLeak:::.guard_fit(X, y = c(1, 0),
                                    steps = list(impute = list(method = "mice")),
                                    task = "binomial"),
               "mice")
  expect_error(bioLeak:::.guard_fit(X, y = c(1, 0),
                                    steps = list(impute = list(method = "bad")),
                                    task = "binomial"),
               "Unknown impute method")
})

test_that("guard_fit validates normalization and feature selection modes", {
  X <- data.frame(a = 1:4, b = 2:5)
  expect_error(bioLeak:::.guard_fit(X, y = 1:4,
                                    steps = list(normalize = list(method = "bad")),
                                    task = "gaussian"),
               "Unknown normalize method")

  expect_error(bioLeak:::.guard_fit(X, y = c(1, 1, 1, 1),
                                    steps = list(fs = list(method = "ttest")),
                                    task = "binomial"),
               "requires a binary outcome")
})

test_that("guard_fit filter respects min_keep", {
  X <- data.frame(a = rep(1, 5), b = 1:5, c = 2:6)
  fit <- bioLeak:::.guard_fit(
    X, y = c(0, 1, 0, 1, 0),
    steps = list(impute = list(method = "median"),
                 normalize = list(method = "none"),
                 filter = list(var_thresh = 10, iqr_thresh = 10, min_keep = 1),
                 fs = list(method = "none")),
    task = "binomial"
  )
  expect_true(fit$state$diagnostics$p_out >= 1)
})

test_that("guard_fit supports PCA feature selection", {
  X <- data.frame(a = 1:10, b = 2:11, c = 3:12)
  fit <- bioLeak:::.guard_fit(
    X, y = c(rep(0, 5), rep(1, 5)),
    steps = list(impute = list(method = "median"),
                 normalize = list(method = "none"),
                 filter = list(var_thresh = 0, iqr_thresh = 0),
                 fs = list(method = "pca", pca_comp = 2)),
    task = "binomial"
  )
  out <- fit$transform(X)
  expect_equal(ncol(out), 2)
})

test_that("guard_fit handles missForest when available", {
  X <- data.frame(a = c(1, NA, 3), b = c(4, 5, NA))
  if (!requireNamespace("randomForest", quietly = TRUE)) {
    expect_error(bioLeak:::.guard_fit(
      X, y = c(1, 0, 1),
      steps = list(impute = list(method = "missForest")),
      task = "binomial"
    ), "randomForest")
  } else {
    fit <- suppressWarnings(bioLeak:::.guard_fit(
      X, y = c(1, 0, 1),
      steps = list(impute = list(method = "missForest", maxiter = 1, ntree = 5),
                   normalize = list(method = "none"),
                   filter = list(var_thresh = 0, iqr_thresh = 0),
                   fs = list(method = "none")),
      task = "binomial"
    ))
    expect_true(inherits(fit, "GuardFit"))
  }
})

test_that("guard_fit handles lasso feature selection when available", {
  X <- data.frame(a = rnorm(20), b = rnorm(20), c = rnorm(20))
  y <- factor(rep(c(0, 1), each = 10))
  if (!requireNamespace("glmnet", quietly = TRUE)) {
    expect_error(bioLeak:::.guard_fit(
      X, y = y,
      steps = list(fs = list(method = "lasso")),
      task = "binomial"
    ), "glmnet")
  } else {
    fit <- suppressWarnings(bioLeak:::.guard_fit(
      X, y = y,
      steps = list(impute = list(method = "median"),
                   normalize = list(method = "none"),
                   filter = list(var_thresh = 0, iqr_thresh = 0),
                   fs = list(method = "lasso")),
      task = "binomial"
    ))
    expect_true(fit$state$fs$method == "lasso")
  }
})

test_that("guard_fit coerces character columns to numeric or factor", {
  X <- data.frame(a = c("1", "2", "3"), b = c("x", "y", "x"),
                  stringsAsFactors = FALSE)
  fit <- bioLeak:::.guard_fit(
    X, y = c(0, 1, 0),
    steps = list(impute = list(method = "median"),
                 normalize = list(method = "none"),
                 filter = list(var_thresh = 0, iqr_thresh = 0),
                 fs = list(method = "none")),
    task = "binomial"
  )
  out <- fit$transform(X)
  expect_true(ncol(out) >= 2)
})

test_that("predict_guard requires GuardFit objects", {
  X <- data.frame(a = 1:3, b = 2:4)
  fit <- bioLeak:::.guard_fit(X, y = c(1, 0, 1),
                              steps = list(impute = list(method = "median")),
                              task = "binomial")
  out <- predict_guard(fit, X)
  expect_true(is.data.frame(out))
  expect_error(predict_guard(list(), X))
})

test_that("basic guard utility helpers behave as expected", {
  expect_equal(bioLeak:::`%||%`(1, 2), 1)
  expect_equal(bioLeak:::`%||%`(NULL, 2), 2)

  expect_true(bioLeak:::.guard_is_binary(factor(c("a", "b"))))
  expect_true(bioLeak:::.guard_is_binary(c(0, 1, 1)))
  expect_false(bioLeak:::.guard_is_binary(c(0, 1, 2)))

  expect_equal(bioLeak:::.guard_majority(c(1, 1, 2, NA)), 1)
  expect_true(is.na(bioLeak:::.guard_majority(c(NA, NA))))
})

test_that("winsorization guards numeric inputs and ignores non-numeric", {
  x <- c(-100, 0, 1, 2, 100)
  out <- bioLeak:::.guard_mad_winsorize(x, k = 1)
  expect_true(min(out) >= 0)
  expect_true(max(out) <= 2)

  y <- c("a", "b")
  expect_equal(bioLeak:::.guard_mad_winsorize(y), y)
})

test_that("column alignment and NA filling behave consistently", {
  X <- data.frame(a = c(1, NA), b = c(2, 3))
  aligned <- bioLeak:::.guard_align_cols(X, c("b", "a", "c"))
  expect_equal(names(aligned), c("b", "a", "c"))
  expect_true(all(is.na(aligned$c)))

  fill <- list(a = 5, b = 6, c = 7)
  filled <- bioLeak:::.guard_fill_na(aligned, fill)
  expect_equal(filled$a[2], 5)
  expect_equal(filled$c[1], 7)
})

test_that("kNN search and imputation functions return valid shapes", {
  train <- data.frame(a = c(1, 2, 3), b = c(4, 5, 6))
  query <- data.frame(a = c(2, 3), b = c(5, 6))
  idx <- bioLeak:::.guard_knn_search(train, query, k = 2)
  expect_equal(dim(idx), c(2, 2))
  expect_true(all(idx >= 1 & idx <= nrow(train)))

  empty_idx <- bioLeak:::.guard_knn_search(train[0, ], query, k = 2)
  expect_equal(dim(empty_idx), c(2, 0))

  X <- data.frame(a = c(1, NA, 3), b = c(4, 5, NA))
  knn_fit <- bioLeak:::.guard_knn_impute_train(X, k = 1)
  expect_equal(dim(knn_fit$imp), dim(X))
  expect_true(all(is.finite(as.matrix(knn_fit$imp))))

  Xnew <- data.frame(a = c(NA, 2), b = c(5, NA))
  imputed <- bioLeak:::.guard_knn_impute_new(Xnew, knn_fit$imp, k = 1,
                                             med = knn_fit$med, scale = knn_fit$scale)
  expect_equal(dim(imputed), dim(Xnew))
  expect_true(all(is.finite(as.matrix(imputed))))
})

test_that("random-forest imputation guard handles missing dependency", {
  X <- data.frame(a = c(1, NA, 3), b = c(4, 5, NA))
  if (!requireNamespace("randomForest", quietly = TRUE)) {
    expect_error(bioLeak:::.guard_rf_impute_train(X, maxiter = 1, ntree = 10),
                 "randomForest")
  } else {
    rf_fit <- suppressWarnings(bioLeak:::.guard_rf_impute_train(X, maxiter = 1,
                                                                ntree = 10, seed = 1))
    expect_equal(dim(rf_fit$imp), dim(X))
    out <- suppressWarnings(bioLeak:::.guard_rf_impute_new(X, rf_fit$models, rf_fit$med))
    expect_equal(dim(out), dim(X))
  }
})

test_that("guard hashing and level normalization behave predictably", {
  h <- bioLeak:::.guard_hash(list(a = 1))
  expect_true(is.character(h))
  expect_true(nchar(h) > 0)

  df <- data.frame(flag = c(TRUE, FALSE), group = c("a", "a"))
  out <- bioLeak:::.guard_ensure_levels(df)
  expect_true(is.factor(out$data$flag))
  expect_true(is.factor(out$data$group))
  expect_true(length(out$levels$group) >= 2)

  out2 <- bioLeak:::.guard_ensure_levels(data.frame(flag = c(TRUE, FALSE)),
                                         levels_map = out$levels)
  expect_true("group" %in% names(out2$data))
  expect_true(is.factor(out2$data$group))
})

#' Summarize a leakage audit
#'
#' Prints a concise, human-readable report for a `LeakAudit` object produced by
#' [audit_leakage()]. The summary surfaces four diagnostics when available:
#' label-permutation gap (prediction-label association by default), batch/study
#' association tests (metadata aligned with fold splits), target leakage scan
#' (features strongly associated with the outcome), and near-duplicate detection
#' (high similarity in `X_ref`). The output reflects the stored audit results
#' only; it does not recompute any tests.
#'
#' @details
#' The permutation test quantifies prediction-label association when using fixed
#' predictions; refit-based permutations require `perm_refit = TRUE` (or `"auto"`
#' with refit data). It does not by itself prove or rule out leakage.
#' Batch association flags metadata that align with fold assignment; this may
#' reflect study design rather than leakage.
#' Target leakage scan uses univariate feature-outcome associations and can miss
#' multivariate proxies, interaction leakage, or features not included in `X_ref`.
#' The multivariate scan (enabled by default for supported tasks) reports an
#' additional model-based score.
#' Duplicate detection only considers the provided `X_ref` features and the
#' similarity threshold used during [audit_leakage()]. By default,
#' `duplicate_scope = "train_test"` filters to pairs that cross train/test;
#' set `duplicate_scope = "all"` to include within-fold duplicates.
#' Sections are reported as "not available" when the corresponding audit
#' component was not computed.
#'
#' @seealso [plot_perm_distribution()], [plot_fold_balance()], [plot_overlap_checks()]
#'
#' @param object A `LeakAudit` object from [audit_leakage()]. The summary reads
#'   stored results from `object` and prints them to the console.
#' @param digits Integer number of digits to show when formatting numeric
#'   statistics in the console output. Defaults to `3`. Increasing `digits`
#'   shows more precision; decreasing it shortens the printout without changing
#'   the underlying values.
#' @param ... Unused. Included for S3 method compatibility; additional
#'   arguments are ignored.
#' @return Invisibly returns `object` after printing the summary.
#' @examples
#' set.seed(1)
#' df <- data.frame(
#'   subject = rep(1:6, each = 2),
#'   outcome = rbinom(12, 1, 0.5),
#'   x1 = rnorm(12),
#'   x2 = rnorm(12)
#' )
#' splits <- make_split_plan(df, outcome = "outcome",
#'                       mode = "subject_grouped", group = "subject", v = 3)
#' custom <- list(
#'   glm = list(
#'     fit = function(x, y, task, weights, ...) {
#'       stats::glm(y ~ ., data = as.data.frame(x),
#'                  family = stats::binomial(), weights = weights)
#'     },
#'     predict = function(object, newdata, task, ...) {
#'       as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
#'                                 type = "response"))
#'     }
#'   )
#' )
#' fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                     learner = "glm", custom_learners = custom,
#'                     metrics = "auc", refit = FALSE, seed = 1)
#' audit <- audit_leakage(fit, metric = "auc", B = 5,
#'                        X_ref = df[, c("x1", "x2")], seed = 1)
#' summary(audit) # prints the audit report and returns `audit` invisibly
#'
#' @export
summary.LeakAudit <- function(object, digits = 3, ...) {
  if (!inherits(object, "LeakAudit"))
    stop("Object must be a 'LeakAudit'.", call. = FALSE)

  cat("\n==============================\n")
  cat(" bioLeak Leakage Audit Summary\n")
  cat("==============================\n\n")

  fit <- object@fit
  info <- fit@info
  warn_sym <- .bio_symbol("warn")
  ok_sym   <- .bio_symbol("check")
  sym_pm <- .bio_symbol("pm")
  sym_chi <- .bio_symbol("chi_sq")
  sym_ge <- .bio_symbol("ge")

  task <- fit@task %||% NA_character_
  outcome <- fit@outcome %||% NA_character_
  splits <- fit@splits
  mode <- if (!is.null(splits)) splits@mode %||% NA_character_ else NA_character_
  indices <- if (!is.null(splits)) splits@indices else NULL
  hash <- info$hash %||% "NA"

  pos_class <- fit@info$positive_class %||% NA_character_
  pos_label <- ""
  if (!is.null(pos_class) && isTRUE(task == "binomial") &&
      !is.na(pos_class) && nzchar(as.character(pos_class))) {
    pos_label <- paste0(" | Positive class: ", as.character(pos_class))
  }
  cat(sprintf("Task: %s | Outcome: %s | Splitting mode: %s%s\n",
              task, outcome, mode, pos_label))
  cat(sprintf("Hash: %s | Folds: %d | Repeats: %d\n\n",
              substr(hash, 1, 12),
              length(indices),
              info$repeats %||% 1))

  # --- Label-permutation association test ---
  if (!is.null(object@permutation_gap) && nrow(object@permutation_gap) > 0) {
    pg <- object@permutation_gap
    perm_method <- object@info$perm_method %||% "fixed"
    perm_mode <- object@info$perm_refit_mode %||% perm_method
    perm_label <- if (identical(perm_method, "refit")) {
      "refit per permutation"
    } else {
      "fixed predictions"
    }
    if (grepl("^auto", perm_mode)) {
      perm_label <- paste0(perm_label, " (auto)")
    }
    cat("Label-Permutation Association Test:\n")
    cat(sprintf("  Method: %s\n", perm_label))
    if (!is.null(object@info$perm_refit_reason) &&
        nzchar(object@info$perm_refit_reason)) {
      cat(sprintf("  Auto mode: %s\n", object@info$perm_refit_reason))
    }
    cat(sprintf("  Observed metric: %s\n",
                formatC(pg$metric_obs, digits = digits, format = "f")))
    cat(sprintf("  Permuted mean %s SD: %s %s %s\n",
                sym_pm,
                formatC(pg$perm_mean, digits = digits, format = "f"),
                sym_pm,
                formatC(pg$perm_sd, digits = digits, format = "f")))
    cat(sprintf("  Gap: %s (larger gap = stronger non-random signal)\n",
                formatC(pg$gap, digits = digits, format = "f")))
    if (!identical(perm_method, "refit")) {
      cat("  Note: Fixed-prediction label permutations quantify prediction-label association.\n")
      cat("  They do NOT refit models and are not a full null test of no signal.\n")
    }
    cat("  This test does NOT diagnose information leakage. Use the Batch Association,\n")
    cat("  Target Leakage Scan, and Duplicate Detection sections to check for leakage.\n\n")
  } else {
    cat("Label-Permutation Association Test: not available.\n\n")
  }

  # --- Batch association ---
  if (!is.null(object@batch_assoc) && nrow(object@batch_assoc) > 0) {
    ba <- object@batch_assoc
    cat("Batch / Study Association:\n")
    if ("batch_col" %in% names(ba)) {
      has_repeat <- "repeat_id" %in% names(ba)
      for (i in seq_len(nrow(ba))) {
        label <- if (has_repeat && is.finite(ba$repeat_id[i])) {
          sprintf("%s (repeat %s)", ba$batch_col[i], ba$repeat_id[i])
        } else {
          ba$batch_col[i]
        }
        cat(sprintf("  %s: %s = %s (df = %s), p = %s\n",
                    label,
                    sym_chi,
                    formatC(ba$stat[i], digits = digits, format = "f"),
                    formatC(ba$df[i], digits = digits, format = "f"),
                    formatC(ba$pval[i], digits = digits, format = "f")))
      }
      cat("\n")
    } else {
      cat(sprintf("  %s = %s (df = %s), p = %s\n\n",
                  sym_chi,
                  formatC(ba$stat, digits = digits, format = "f"),
                  formatC(ba$df, digits = digits, format = "f"),
                  formatC(ba$pval, digits = digits, format = "f")))
    }
  } else {
    cat("Batch / Study Association: none detected.\n\n")
  }

  # --- Target leakage scan ---
  if (!is.null(object@target_assoc) && nrow(object@target_assoc) > 0) {
    ta <- object@target_assoc
    threshold <- object@info$target_threshold %||% 0.9
    flagged <- ta[!is.na(ta$score) & ta$score >= threshold, , drop = FALSE]
    cat("Target Leakage Scan:\n")
    cat(sprintf("  Features checked: %d | Flagged (score %s %s): %d\n",
                nrow(ta),
                sym_ge,
                formatC(threshold, digits = digits, format = "f"),
                nrow(flagged)))
    if (nrow(flagged) > 0) {
      flagged <- flagged[order(flagged$score, decreasing = TRUE, na.last = TRUE), , drop = FALSE]
      top <- utils::head(flagged, 5)
      cols <- intersect(c("feature", "metric", "value", "score", "p_value"), names(top))
      print(top[, cols, drop = FALSE], row.names = FALSE)
    } else {
      cat("  No strong proxy features detected.\n")
    }
    cat("\n")
  } else {
    cat("Target Leakage Scan: not available.\n\n")
  }

  # --- Multivariate target scan ---
  mv <- object@info$target_multivariate %||% data.frame()
  if (is.data.frame(mv) && nrow(mv) > 0) {
    cat("Multivariate Target Scan:\n")
    mv_row <- mv[1, , drop = FALSE]
    cat(sprintf("  Metric: %s | Score: %s | p = %s\n",
                mv_row$metric,
                formatC(mv_row$score, digits = digits, format = "f"),
                formatC(mv_row$p_value, digits = digits, format = "f")))
    if (all(c("n_features", "n_components", "n_interactions", "n_perm") %in% names(mv_row))) {
      cat(sprintf("  Features: %s | Components: %s | Interactions: %s | Permutations: %s\n",
                  mv_row$n_features, mv_row$n_components, mv_row$n_interactions, mv_row$n_perm))
    }
    cat("\n")
  } else if (isTRUE(object@info$target_scan_multivariate)) {
    cat("Multivariate Target Scan: not available.\n\n")
  }

  # --- Duplicate detection ---
  if (!is.null(object@duplicates) && nrow(object@duplicates) > 0) {
    dd <- object@duplicates
    cat("Near-Duplicate Samples:\n")
    sim_label <- object@info$sim_method %||% "cosine"
    dup_scope <- object@info$duplicate_scope %||% "all"
    scope_label <- if (identical(dup_scope, "train_test")) "train/test only" else "all pairs"
    cat(sprintf("  Scope: %s\n", scope_label))
    cat(sprintf("  %d pairs detected above %s %s %s\n",
                nrow(dd),
                sim_label,
                sym_ge,
                formatC(object@info$duplicate_threshold, digits = digits, format = "f")))
    head_pairs <- utils::head(dd, 5)
    cat("  Example pairs:\n")
    print(head_pairs, row.names = FALSE)
    cat("\n")
  } else {
    cat("No near-duplicates detected.\n\n")
  }

  # --- Mechanism risk assessment ---
  mech <- object@info$mechanism_summary %||% data.frame()
  if (is.data.frame(mech) && nrow(mech) > 0) {
    cat("Mechanism Risk Assessment:\n")
    cols <- intersect(c("mechanism_class", "flagged", "evidence", "statistic", "p_value"),
                      names(mech))
    print(mech[, cols, drop = FALSE], row.names = FALSE)
    cat("\n")
  } else {
    cat("Mechanism Risk Assessment: not available.\n\n")
  }

  # --- Overall interpretation ---
  cat("Interpretation:\n")
  pg_available <- !is.null(object@permutation_gap) && nrow(object@permutation_gap) > 0

  if (!pg_available) {
    cat("  No permutation test results.\n")
  } else if (object@permutation_gap$gap < 0.01) {
    cat(sprintf("  %s Little non-random signal (gap near zero).\n", warn_sym))
  } else if (object@permutation_gap$gap < 0.05) {
    cat(sprintf("  %s Modest non-random signal.\n", ok_sym))
  } else {
    cat(sprintf("  %s Strong non-random signal.\n", ok_sym))
  }

  invisible(object)
}

#' Summarize a LeakFit object
#'
#' Prints a compact console report for a [LeakFit] object created by
#' [fit_resample()]. The report lists task/outcome metadata, learners,
#' total folds, and cross-validated metrics summarized as mean and standard
#' deviation across completed folds, plus a small audit table with per-fold
#' train/test sizes and retained feature counts.
#'
#' This summary is meant for quick sanity checks of the resampling setup and
#' performance. It does not run leakage diagnostics and will not detect target
#' leakage, duplicate samples, or batch/study confounding; use [audit_leakage()]
#' or `summary()` on a [LeakAudit] object for those checks.
#'
#' @param object A [LeakFit] object returned by [fit_resample()]. It should
#'   contain `metric_summary` and `audit` slots; missing entries result in empty
#'   sections in the printed report.
#' @param digits Integer scalar. Number of decimal places to print in numeric
#'   summary tables. Defaults to 3; affects printed output only, not the
#'   returned data.
#' @param ... Unused. Included for S3 method compatibility; changing these
#'   values has no effect.
#' @return Invisibly returns `object@metric_summary`, a data frame of per-learner
#'   metric means and standard deviations computed across folds. This function
#'   does not recompute metrics.
#' @examples
#' set.seed(1)
#' df <- data.frame(
#'   subject = rep(1:6, each = 2),
#'   outcome = factor(rep(c(0, 1), each = 6)),
#'   x1 = rnorm(12),
#'   x2 = rnorm(12)
#' )
#' splits <- make_split_plan(
#'   df,
#'   outcome = "outcome",
#'   mode = "subject_grouped",
#'   group = "subject",
#'   v = 3,
#'   stratify = TRUE,
#'   progress = FALSE
#' )
#' custom <- list(
#'   glm = list(
#'     fit = function(x, y, task, weights, ...) {
#'       stats::glm(y ~ ., data = data.frame(y = y, x),
#'                  family = stats::binomial(), weights = weights)
#'     },
#'     predict = function(object, newdata, task, ...) {
#'       as.numeric(stats::predict(object,
#'                                 newdata = as.data.frame(newdata),
#'                                 type = "response"))
#'     }
#'   )
#' )
#' fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                     learner = "glm", custom_learners = custom,
#'                     metrics = "auc", seed = 1)
#' summary_df <- summary(fit)
#' summary_df
#'
#' @export
summary.LeakFit <- function(object, digits = 3, ...) {
  if (!inherits(object, "LeakFit"))
    stop("Object must be of class 'LeakFit'.")

  cat("\n===========================\n")
  cat(" bioLeak Model Fit Summary\n")
  cat("===========================\n\n")

  sym_pm <- .bio_symbol("pm")

  # Basic info
  info <- object@info
  cat(sprintf("Task: %s\n", object@task))
  cat(sprintf("Outcome: %s\n", object@outcome))
  if (identical(object@task, "binomial") && !is.null(info$positive_class) &&
      !is.na(info$positive_class) && nzchar(as.character(info$positive_class))) {
    cat(sprintf("Positive class: %s\n", as.character(info$positive_class)))
  }
  cat(sprintf("Learners: %s\n", paste(unique(object@metrics$learner), collapse = ", ")))
  cat(sprintf("Total folds: %d\n", length(object@splits@indices)))
  if (is.data.frame(info$fold_status) && nrow(info$fold_status) > 0) {
    ok <- sum(info$fold_status$status == "success", na.rm = TRUE)
    sk <- sum(info$fold_status$status == "skipped", na.rm = TRUE)
    fl <- sum(info$fold_status$status == "failed", na.rm = TRUE)
    cat(sprintf("Fold status: %d success, %d skipped, %d failed\n", ok, sk, fl))
  }
  cat(sprintf("Refit performed: %s\n", if (isTRUE(info$refit)) "Yes" else "No"))
  cat(sprintf("Hash: %s\n\n", substr(info$hash, 1, 12)))

  # Metric summary
  if (nrow(object@metric_summary) > 0) {
    cat(sprintf("Cross-validated metrics (mean %s SD):\n", sym_pm))
    ms <- object@metric_summary
    metrics_fmt <- as.data.frame(ms)
    num_cols <- vapply(metrics_fmt, is.numeric, logical(1))
    if (any(num_cols)) {
      metrics_fmt[num_cols] <- lapply(metrics_fmt[num_cols], round, digits = digits)
    }
    print(metrics_fmt)
    cat("\n")
  } else {
    cat("No metric summary available.\n\n")
  }

  # Audit information
  if (nrow(object@audit) > 0) {
    cat("Audit overview:\n")
    audit_df <- head(object@audit, 5)
    print(audit_df, row.names = FALSE)
    cat("\n")
  } else {
    cat("No audit information stored.\n\n")
  }

  invisible(object@metric_summary)
}

#' Summarize a nested tuning result
#'
#' Prints a concise report for a `LeakTune` object produced by [tune_resample()].
#' The report highlights the tuning strategy, selection metric, and
#' cross-validated performance across outer folds, plus a glimpse of the selected
#' hyperparameters.
#'
#' @param object A [LeakTune] object returned by [tune_resample()].
#' @param digits Integer scalar. Number of decimal places to print in numeric
#'   summary tables. Defaults to 3.
#' @param ... Unused. Included for S3 method compatibility.
#' @return Invisibly returns `object$metric_summary`, the data frame of per-learner
#'   metric means and standard deviations computed across outer folds.
#' @export
summary.LeakTune <- function(object, digits = 3, ...) {
  if (!inherits(object, "LeakTune"))
    stop("Object must be of class 'LeakTune'.", call. = FALSE)

  cat("\n============================\n")
  cat(" bioLeak Tuning Summary\n")
  cat("============================\n\n")

  sym_pm <- .bio_symbol("pm")
  info <- object$info

  outer_fit_ok <- vapply(object$outer_fits, function(of) {
    !is.null(of) && methods::is(of, "LeakFit")
  }, logical(1))
  successful_outer <- sum(outer_fit_ok)
  total_outer <- length(object$outer_fits)

  # Retrieve outcome from the first successful outer fit if available
  outcome_label <- "Unknown"
  if (successful_outer > 0) {
    first_outer <- object$outer_fits[[which(outer_fit_ok)[1]]]
    outcome_label <- first_outer@outcome
  }

  cat(sprintf("Task: %s\n", info$task))
  cat(sprintf("Outcome: %s\n", outcome_label))

  if (identical(info$task, "binomial") && !is.null(info$positive_class)) {
    cat(sprintf("Positive class: %s\n", as.character(info$positive_class)))
  }

  # Tuning Info
  grid_info <- if (is.data.frame(info$grid)) paste(nrow(info$grid), "combinations") else info$grid
  cat(sprintf("Tuning Grid: %s\n", grid_info))
  cat(sprintf("Selection Rule: %s (Metric: %s)\n", info$selection, info$selection_metric))
  if (is.data.frame(info$fold_status) && nrow(info$fold_status) > 0) {
    ok <- sum(info$fold_status$status == "success", na.rm = TRUE)
    sk <- sum(info$fold_status$status == "skipped", na.rm = TRUE)
    fl <- sum(info$fold_status$status == "failed", na.rm = TRUE)
    cat(sprintf("Fold status: %d success, %d skipped, %d failed\n", ok, sk, fl))
  }
  cat(sprintf("Refit performed: %s\n", if (isTRUE(info$refit) && !is.null(object$final_model)) "Yes" else "No"))
  cat(sprintf("Outer Folds: %d successful / %d total\n\n", successful_outer, total_outer))

  # Metric Summary
  if (nrow(object$metric_summary) > 0) {
    cat(sprintf("Outer Loop Metrics (mean %s SD):\n", sym_pm))
    ms <- object$metric_summary
    metrics_fmt <- as.data.frame(ms)
    num_cols <- vapply(metrics_fmt, is.numeric, logical(1))
    if (any(num_cols)) {
      metrics_fmt[num_cols] <- lapply(metrics_fmt[num_cols], round, digits = digits)
    }
    print(metrics_fmt)
    cat("\n")
  } else {
    cat("No metric summary available.\n\n")
  }

  # Best Params Snapshot (Cleaned)
  if (nrow(object$best_params) > 0) {
    cat("Best Parameters (First 5 Folds):\n")
    bp <- head(object$best_params, 5)

    # Remove internal bookkeeping columns
    bp$id <- NULL
    bp$.config <- NULL

    # Check if any actual parameters remain
    param_cols <- setdiff(names(bp), c("fold", "learner"))

    if (length(param_cols) == 0) {
      cat("  (No tunable parameters detected in model spec)\n")
    } else {
      # rounding for display
      num_cols_bp <- vapply(bp, is.numeric, logical(1))
      if (any(num_cols_bp)) {
        bp[num_cols_bp] <- lapply(bp[num_cols_bp], round, digits = digits)
      }
      print(bp, row.names = FALSE)
    }
    cat("\n")
  }

  invisible(object$metric_summary)
}

# Metric utilities -----------------------------------------------------------

.cindex_pairwise <- function(pred, truth) {
  pred <- as.numeric(pred)
  truth <- as.numeric(truth)
  ok <- is.finite(pred) & is.finite(truth)
  pred <- pred[ok]
  truth <- truth[ok]
  n <- length(pred)
  if (n < 2L) return(NA_real_)
  if (length(unique(truth)) < 2L) return(NA_real_)

  conc <- 0L
  ties <- 0L
  total <- 0L

  for (i in seq_len(n - 1L)) {
    yi <- truth[i]
    pi <- pred[i]
    dy <- yi - truth[(i + 1L):n]
    valid <- dy != 0
    if (!any(valid)) next
    pj <- pred[(i + 1L):n][valid]
    dy <- dy[valid]
    dp <- pi - pj
    total <- total + length(dp)
    prod <- dp * dy
    conc <- conc + sum(prod > 0)
    ties <- ties + sum(dp == 0)
  }

  if (!total) return(NA_real_)
  (conc + 0.5 * ties) / total
}

.cindex_survival <- function(pred, truth) {
  if (!inherits(truth, "Surv")) return(NA_real_)
  if (!requireNamespace("survival", quietly = TRUE)) return(NA_real_)
  df <- data.frame(pred = as.numeric(pred))
  concord <- try(survival::concordance(truth ~ pred, data = df), silent = TRUE)
  if (inherits(concord, "try-error")) return(NA_real_)
  as.numeric(concord$concordance)
}

.multiclass_accuracy <- function(truth, pred_class) {
  if (is.null(pred_class)) return(NA_real_)
  mean(pred_class == truth, na.rm = TRUE)
}

.multiclass_macro_f1 <- function(truth, pred_class) {
  if (is.null(pred_class)) return(NA_real_)
  truth <- factor(truth)
  pred_class <- factor(pred_class, levels = levels(truth))
  lvls <- levels(truth)
  f1_vals <- vapply(lvls, function(lbl) {
    tp <- sum(pred_class == lbl & truth == lbl, na.rm = TRUE)
    fp <- sum(pred_class == lbl & truth != lbl, na.rm = TRUE)
    fn <- sum(pred_class != lbl & truth == lbl, na.rm = TRUE)
    prec <- if ((tp + fp) > 0) tp / (tp + fp) else NA_real_
    rec <- if ((tp + fn) > 0) tp / (tp + fn) else NA_real_
    if (is.na(prec) && is.na(rec)) return(NA_real_)
    if (is.na(prec) || is.na(rec)) return(NA_real_)
    if ((prec + rec) == 0) return(0)
    2 * prec * rec / (prec + rec)
  }, numeric(1))
  if (all(is.na(f1_vals))) return(NA_real_)
  f1_vals[is.na(f1_vals)] <- 0
  mean(f1_vals)
}

.multiclass_log_loss <- function(truth, prob, eps = 1e-15) {
  if (is.null(prob)) return(NA_real_)
  truth <- factor(truth)
  levels_truth <- levels(truth)
  if (is.data.frame(prob)) prob <- as.matrix(prob)
  if (ncol(prob) != length(levels_truth)) return(NA_real_)
  prob <- pmin(pmax(prob, eps), 1 - eps)
  idx <- cbind(seq_len(nrow(prob)), match(truth, levels_truth))
  vals <- prob[idx]
  -mean(log(vals), na.rm = TRUE)
}

test_that("cindex_pairwise computes concordance correctly", {
  pred <- c(0.1, 0.2, 0.3)
  truth <- c(1, 2, 3)
  expect_equal(bioLeak:::.cindex_pairwise(pred, truth), 1)

  pred2 <- c(0.3, 0.2, 0.1)
  expect_equal(bioLeak:::.cindex_pairwise(pred2, truth), 0)

  expect_true(is.na(bioLeak:::.cindex_pairwise(1, 1)))
  expect_true(is.na(bioLeak:::.cindex_pairwise(c(0.1, 0.2), c(1, 1))))
})

#' Simulation benchmark matrix for leakage diagnostics
#'
#' Runs a reproducible grid of simulation scenarios across modalities, leakage
#' mechanisms, and split modes using [simulate_leakage_suite()]. This function
#' is designed as a benchmarking harness to quantify detection rates and
#' performance inflation under controlled settings.
#'
#' @param modalities Character vector selecting predefined modality profiles.
#'   Supported values: `"omics"`, `"imaging_tabular"`, `"ehr_tabular"`.
#' @param leakages Character vector of leakage mechanisms passed to
#'   [simulate_leakage_suite()].
#' @param modes Character vector of split modes passed to [simulate_leakage_suite()].
#' @param learner Character scalar. `"glmnet"` (default) or `"ranger"`.
#' @param seeds Integer vector of Monte Carlo seeds.
#' @param B Integer scalar. Number of permutations per scenario.
#' @param alpha Numeric scalar in (0, 1). Detection threshold applied to
#'   permutation p-values.
#' @param parallel Logical scalar. If TRUE, evaluates scenarios in parallel
#'   when `future.apply` is available.
#' @return A data.frame with one row per simulation seed/scenario and columns:
#'   `modality`, `leakage`, `mode`, `seed`, observed metric, gap, p-value, and
#'   a logical `detected` flag. A scenario-level summary is attached as
#'   `attr(x, "summary")`.
#' @export
benchmark_leakage_suite <- function(modalities = c("omics", "imaging_tabular", "ehr_tabular"),
                                    leakages = c("none", "subject_overlap", "batch_confounded", "peek_norm", "lookahead"),
                                    modes = c("subject_grouped", "batch_blocked", "time_series"),
                                    learner = c("glmnet", "ranger"),
                                    seeds = 1:5,
                                    B = 200,
                                    alpha = 0.05,
                                    parallel = FALSE) {
  learner <- match.arg(learner)
  modalities <- unique(as.character(modalities))
  leakages <- unique(as.character(leakages))
  modes <- unique(as.character(modes))

  if (!length(modalities)) stop("modalities must be non-empty.", call. = FALSE)
  if (!length(leakages)) stop("leakages must be non-empty.", call. = FALSE)
  if (!length(modes)) stop("modes must be non-empty.", call. = FALSE)
  if (!is.numeric(alpha) || length(alpha) != 1L || !is.finite(alpha) || alpha <= 0 || alpha >= 1) {
    stop("alpha must be a single numeric value in (0,1).", call. = FALSE)
  }

  modality_profiles <- list(
    omics = list(n = 220L, p = 800L, prevalence = 0.5, rho = 0.35),
    imaging_tabular = list(n = 500L, p = 120L, prevalence = 0.4, rho = 0.15),
    ehr_tabular = list(n = 1400L, p = 60L, prevalence = 0.3, rho = 0.5)
  )
  bad_mod <- setdiff(modalities, names(modality_profiles))
  if (length(bad_mod)) {
    stop(sprintf("Unsupported modalities: %s", paste(bad_mod, collapse = ", ")), call. = FALSE)
  }

  scenario_grid <- expand.grid(
    modality = modalities,
    leakage = leakages,
    mode = modes,
    stringsAsFactors = FALSE
  )

  eval_one <- function(i) {
    row <- scenario_grid[i, , drop = FALSE]
    prof <- modality_profiles[[row$modality]]
    horizon_use <- if (identical(row$mode, "time_series")) 1 else 0

    res <- suppressWarnings(simulate_leakage_suite(
      n = prof$n,
      p = prof$p,
      prevalence = prof$prevalence,
      mode = row$mode,
      learner = learner,
      leakage = row$leakage,
      rho = prof$rho,
      K = 5,
      repeats = 1,
      horizon = horizon_use,
      B = B,
      seeds = seeds,
      parallel = FALSE
    ))
    if (!nrow(res)) return(NULL)

    res$modality <- row$modality
    res$detected <- is.finite(res$p_value) & res$p_value <= alpha & is.finite(res$gap) & res$gap > 0
    res
  }

  rows <- if (isTRUE(parallel) && requireNamespace("future.apply", quietly = TRUE)) {
    future.apply::future_lapply(seq_len(nrow(scenario_grid)), eval_one, future.seed = TRUE)
  } else {
    lapply(seq_len(nrow(scenario_grid)), eval_one)
  }
  rows <- rows[!vapply(rows, is.null, logical(1))]
  if (!length(rows)) {
    out <- data.frame()
    attr(out, "summary") <- data.frame()
    return(out)
  }

  out <- do.call(rbind, rows)
  out <- out[, c("modality", "leakage", "mode", "seed", "metric_obs", "gap", "p_value", "detected"), drop = FALSE]
  rownames(out) <- NULL

  out$detected_num <- as.numeric(out$detected)
  summary_df <- stats::aggregate(
    cbind(metric_obs, gap, detected_num) ~ modality + leakage + mode,
    data = out,
    FUN = function(x) mean(x, na.rm = TRUE)
  )
  out$detected_num <- NULL
  names(summary_df)[names(summary_df) == "detected_num"] <- "detection_rate"
  attr(out, "summary") <- summary_df
  out
}

test_that("benchmark_leakage_suite returns scenario matrix and summary", {
  learner <- NULL
  if (requireNamespace("glmnet", quietly = TRUE)) {
    learner <- "glmnet"
  } else if (requireNamespace("ranger", quietly = TRUE)) {
    learner <- "ranger"
  } else {
    skip("benchmark_leakage_suite requires glmnet or ranger.")
  }

  out <- benchmark_leakage_suite(
    modalities = "omics",
    leakages = c("none", "subject_overlap"),
    modes = "subject_grouped",
    learner = learner,
    seeds = 1,
    B = 2,
    alpha = 0.1,
    parallel = FALSE
  )

  expect_true(is.data.frame(out))
  expect_true(nrow(out) >= 1)
  expect_true(all(c("modality", "leakage", "mode", "seed", "metric_obs", "gap", "p_value", "detected") %in% names(out)))
  summ <- attr(out, "summary")
  expect_true(is.data.frame(summ))
  expect_true(all(c("modality", "leakage", "mode", "detection_rate") %in% names(summ)))
})

test_that("benchmark_leakage_suite validates modalities", {
  expect_error(
    benchmark_leakage_suite(modalities = "unknown_modality", seeds = 1, B = 1, parallel = FALSE),
    "Unsupported modalities"
  )
})

#' Simulate leakage scenarios and audit results
#'
#' @description
#' Simulates synthetic binary classification datasets with optional leakage
#' mechanisms, fits a model using a leakage-aware cross-validation scheme, and
#' summarizes the permutation-gap audit for each Monte Carlo seed. The suite is
#' designed to surface validation failures such as subject overlap across folds,
#' batch-confounded outcomes, global normalization/summary leakage, and
#' time-series look-ahead. The output is a per-seed summary of observed CV
#' performance and its gap versus a label-permutation null; it does not return
#' fitted models or the full audit object. Results are limited to the built-in
#' data generator and leakage types implemented here, and should be interpreted
#' as a simulation-based sanity check rather than a comprehensive leakage
#' detector for real data.
#'
#' @details
#' The generator draws \code{p} standard normal predictors, builds a linear
#' predictor from the first \code{min(5, p)} features, scales it by
#' \code{signal_strength}, and samples a binary outcome to achieve the requested
#' \code{prevalence}. Outcomes are returned as a two-level factor, so the audited
#' metric is AUC. Simulated metadata include subject, batch, study, and time
#' fields used by \code{mode} to create leakage-aware splits. Leakage mechanisms
#' are injected by adding a single extra predictor as described in
#' \code{leakage}. Parallel execution uses \code{future.apply} when installed and
#' does not change results.
#'
#' @param n Integer scalar. Number of samples to simulate (default 500). Larger
#'   values stabilize the Monte Carlo summary but increase runtime.
#' @param p Integer scalar. Number of baseline predictors before any leakage
#'   feature is added (default 20). Increasing \code{p} changes the signal-to-noise
#'   ratio and increases fitting time.
#' @param prevalence Numeric scalar in (0, 1). Target prevalence of class 1 in
#'   the simulated outcome (default 0.5). Changing this alters class imbalance
#'   and can affect AUC and the permutation gap.
#' @param mode Character scalar. Cross-validation scheme passed to
#'   \code{make_split_plan()}; one of \code{"subject_grouped"},
#'   \code{"batch_blocked"}, \code{"study_loocv"}, \code{"time_series"}.
#'   Defaults to \code{"subject_grouped"}. This controls how samples are grouped
#'   into folds (by subject, batch, study, or time) and therefore which leakage
#'   mechanisms are realistically challenged.
#' @param learner Character scalar. Base learner, \code{"glmnet"} (default) or
#'   \code{"ranger"}. Requires the corresponding package in \code{Suggests}.
#'   Switching learners changes the fitted model, runtime, and performance.
#' @param leakage Character scalar. Leakage mechanism to inject; one of
#'   \code{"none"}, \code{"subject_overlap"}, \code{"batch_confounded"},
#'   \code{"peek_norm"}, \code{"lookahead"}. Leakage is added as an extra
#'   predictor: \code{"subject_overlap"} adds per-subject mean outcome,
#'   \code{"batch_confounded"} adds per-batch mean outcome, \code{"peek_norm"}
#'   adds the globally normalized (z-scored) outcome, and \code{"lookahead"} adds the next-time
#'   outcome. Changing this controls whether and how leakage is present.
#' @param preprocess Optional preprocessing list or recipe passed to
#'   [fit_resample()]. When NULL (default), the simulator uses the
#'   fit_resample defaults; for \code{"peek_norm"} leakage, normalization is
#'   set to \code{"none"} to avoid attenuating the constant leakage feature.
#' @param rho Numeric scalar in [-1, 1]. AR(1)-style autocorrelation applied to
#'   each predictor across row order (default 0). Higher absolute values increase
#'   serial correlation and make time-ordered leakage more pronounced.
#' @param K Integer scalar. Number of folds/partitions (default 5). Used as the
#'   fold count for \code{"subject_grouped"} and \code{"batch_blocked"}, and as
#'   the number of rolling partitions for \code{"time_series"}. Ignored for
#'   \code{"study_loocv"} (folds equal the number of studies).
#' @param repeats Integer scalar >= 1. Number of repeated CV runs for
#'   \code{"subject_grouped"} and \code{"batch_blocked"} (default 1). Increasing
#'   \code{repeats} increases the number of folds and runtime. Ignored for
#'   \code{"study_loocv"} and \code{"time_series"}.
#' @param horizon Numeric scalar >= 0. Minimum time gap enforced between train
#'   and test for \code{"time_series"} splits (default 0). Larger values make the
#'   split more conservative and can reduce leakage from temporal proximity.
#' @param B Integer scalar >= 1. Number of permutations used by
#'   \code{audit_leakage()} to compute the permutation gap and p-value (default
#'   200). Larger values yield more stable p-values but increase runtime.
#' @param seeds Integer vector. Monte Carlo seeds (default \code{1:10}). One row
#'   of output is produced per seed; changing \code{seeds} changes the simulated
#'   datasets and splits.
#' @param parallel Logical scalar. If \code{TRUE}, evaluates seeds in parallel
#'   using \code{future.apply} (if installed). Results are identical to sequential
#'   execution; only runtime changes.
#' @param signal_strength Numeric scalar. Scales the linear predictor before
#'   sampling outcomes (default 1). Larger values increase class separation and
#'   tend to increase AUC; smaller values make the task harder.
#' @param verbose Logical scalar. If \code{TRUE}, prints progress messages for
#'   each seed. Does not affect results.
#'
#' @note
#' This function is a general-purpose utility and its data-generation logic
#' intentionally differs from the custom simulation used in the bioLeak
#' manuscript (\file{paper/run_simulation.R}). Specific differences:
#' \itemize{
#'   \item \strong{peek_norm leakage}: this function uses a z-scored binary
#'     outcome as the leak feature; the manuscript uses a noisy continuous
#'     version (\code{as.numeric(y) + rnorm(n, 0, 0.3)}).
#'   \item \strong{lookahead leakage}: this function shifts the binary outcome
#'     (\code{c(y[-1], y[n])}); the manuscript shifts a continuous biomarker
#'     (\code{linpred + noise}).
#'   \item \strong{signal generation}: this function applies AR correlation to
#'     predictors via \code{rho}; the manuscript adds AR(1) noise directly to
#'     the linear predictor.
#'   \item \strong{audit settings}: the manuscript uses
#'     \code{perm_refit = FALSE} and \code{perm_stratify = TRUE};
#'     this function uses \code{perm_refit = "auto"} and the
#'     \code{perm_stratify} default (\code{FALSE}).
#' }
#' Users wishing to reproduce manuscript figures should run
#' \file{paper/run_simulation.R} directly rather than calling this function.
#'
#' @return
#' A \code{LeakSimResults} data frame with one row per seed and columns:
#' \itemize{
#'   \item \code{seed}: seed used for data generation, splitting, and auditing.
#'   \item \code{metric_obs}: observed CV performance (AUC for this simulation).
#'   \item \code{gap}: permutation-gap statistic (observed minus permutation mean).
#'   \item \code{p_value}: permutation p-value for the gap.
#'   \item \code{leakage}: leakage scenario used.
#'   \item \code{mode}: CV mode used.
#' }
#' Only the permutation-gap summary is returned; fitted models, predictions, and
#' other audit components are not included.
#'
#' @examples
#' \donttest{
#'   if (requireNamespace("glmnet", quietly = TRUE)) {
#'     set.seed(1)
#'     res <- simulate_leakage_suite(
#'       n = 120, p = 6, prevalence = 0.4,
#'       mode = "subject_grouped",
#'       learner = "glmnet",
#'       leakage = "subject_overlap",
#'       K = 3, repeats = 1,
#'       B = 50, seeds = 1,
#'       parallel = FALSE
#'     )
#'     # One row per seed with observed AUC, permutation gap, and p-value
#'     res
#'   }
#' }
#' @export
simulate_leakage_suite <- function(
    n = 500, p = 20, prevalence = 0.5,
    mode = c("subject_grouped", "batch_blocked", "study_loocv", "time_series"),
    learner = c("glmnet", "ranger"),
    leakage = c("none", "subject_overlap", "batch_confounded", "peek_norm", "lookahead"),
    preprocess = NULL,
    rho = 0, K = 5, repeats = 1, horizon = 0,
    B = 200, seeds = 1:10, parallel = FALSE, signal_strength = 1,
    verbose = FALSE
) {
  mode <- match.arg(mode)
  learner <- match.arg(learner)
  leakage <- match.arg(leakage)

  FUN <- function(s) {
    if (verbose)
      message(sprintf("[Seed %d] Mode=%s, Leakage=%s", s, mode, leakage))

    set.seed(s)
    sim <- .simulate_dataset(n, p, prevalence, mode, leakage, rho, signal_strength)

    splits <- make_split_plan(
      sim$data, outcome = "y", mode = mode,
      group = sim$group_col, batch = sim$batch_col,
      study = sim$study_col, time = sim$time_col,
      v = K, repeats = repeats, stratify = TRUE,
      horizon = horizon, seed = s
    )

    # Automatically select metric based on outcome type
    metrics <- if (is.factor(sim$data$y)) {
      if (nlevels(sim$data$y) > 2) "accuracy" else "auc"
    } else {
      "rmse"
    }

    preprocess_use <- preprocess
    if (is.null(preprocess_use) && leakage == "peek_norm") {
      preprocess_use <- list(
        impute = list(method = "median"),
        normalize = list(method = "none"),
        filter = list(var_thresh = 0, iqr_thresh = 0),
        fs = list(method = "none")
      )
    }

    fit <- fit_resample(
      sim$data, outcome = "y", splits = splits,
      preprocess = preprocess_use,
      learner = learner, metrics = metrics, seed = s
    )

    aud <- audit_leakage(
      fit,
      metric = metrics,
      B = B,
      perm_refit = "auto",
      perm_refit_auto_max = min(B, 200),
      seed = s
    )

    data.frame(
      seed = s,
      metric_obs = aud@permutation_gap$metric_obs,
      gap = aud@permutation_gap$gap,
      p_value = aud@permutation_gap$p_value,
      leakage = leakage,
      mode = mode
    )
  }

  results <- if (parallel) {
    if (!requireNamespace("future.apply", quietly = TRUE)) {
      warning("parallel=TRUE requires the 'future.apply' package; running sequentially.")
      lapply(seeds, FUN)
    } else {
      future.apply::future_lapply(seeds, FUN, future.seed = TRUE)
    }
  } else {
    lapply(seeds, FUN)
  }

  out <- do.call(rbind, results)
  class(out) <- c("LeakSimResults", class(out))
  out
}


.simulate_dataset <- function(n, p, prevalence, mode, leakage, rho, signal_strength = 1) {
  if (abs(rho) > 1) stop("'rho' must be in [-1, 1]; got ", rho, call. = FALSE)
  X <- matrix(rnorm(n * p), n, p)
  colnames(X) <- sprintf("x%02d", seq_len(p))
  subject <- sample(seq_len(max(5, n %/% 5)), n, replace = TRUE)
  batch <- sample(letters[1:max(3, pmin(6, n %/% 50))], n, replace = TRUE)
  study <- sample(seq_len(max(3, n %/% 80)), n, replace = TRUE)
  time <- seq_len(n)
  if (rho != 0) {
    for (j in seq_len(p)) {
      for (i in 2:n) {
        X[i, j] <- rho * X[i - 1, j] + sqrt(1 - rho^2) * X[i, j]
      }
    }
  }
  linpred <- rowSums(X[, seq_len(min(5, p)), drop = FALSE])
  linpred <- scale(linpred) * signal_strength
  thr <- stats::qnorm(prevalence)
  y_prob <- stats::pnorm(linpred - thr)
  y <- rbinom(n, 1, y_prob)
  if (leakage == "subject_overlap") {
    X <- cbind(X, leak_subj = ave(y, subject, FUN = mean))
  } else if (leakage == "batch_confounded") {
    batch_mean <- ave(y, batch, FUN = mean)
    X <- cbind(X, leak_batch = batch_mean)
  } else if (leakage == "peek_norm") {
    global_mean <- mean(y)
    global_sd <- stats::sd(y)
    if (!is.finite(global_sd) || global_sd == 0) global_sd <- 1
    leak_global <- (y - global_mean) / global_sd
    X <- cbind(X, leak_global = leak_global)
  } else if (leakage == "lookahead") {
    lead <- c(y[-1], y[length(y)])
    X <- cbind(X, leak_future = lead)
  }
  df <- data.frame(X, y = factor(y))
  df$subject <- subject
  df$batch <- batch
  df$study <- study
  df$time <- time
  list(data = df, group_col = "subject", batch_col = "batch",
       study_col = "study", time_col = "time")
}

test_that("simulate_dataset adds leakage-specific columns", {
  sim_none <- bioLeak:::.simulate_dataset(
    n = 20, p = 5, prevalence = 0.5,
    mode = "subject_grouped", leakage = "none",
    rho = 0, signal_strength = 1
  )
  expect_false("leak_subj" %in% names(sim_none$data))

  sim_subj <- bioLeak:::.simulate_dataset(
    n = 20, p = 5, prevalence = 0.5,
    mode = "subject_grouped", leakage = "subject_overlap",
    rho = 0, signal_strength = 1
  )
  expect_true("leak_subj" %in% names(sim_subj$data))

  sim_batch <- bioLeak:::.simulate_dataset(
    n = 20, p = 5, prevalence = 0.5,
    mode = "batch_blocked", leakage = "batch_confounded",
    rho = 0, signal_strength = 1
  )
  expect_true("leak_batch" %in% names(sim_batch$data))

  sim_peek <- bioLeak:::.simulate_dataset(
    n = 20, p = 5, prevalence = 0.5,
    mode = "subject_grouped", leakage = "peek_norm",
    rho = 0, signal_strength = 1
  )
  expect_true("leak_global" %in% names(sim_peek$data))

  sim_future <- bioLeak:::.simulate_dataset(
    n = 20, p = 5, prevalence = 0.5,
    mode = "time_series", leakage = "lookahead",
    rho = 0, signal_strength = 1
  )
  expect_true("leak_future" %in% names(sim_future$data))
})

test_that("simulate_leakage_suite runs when required learners are available", {
  if (!requireNamespace("glmnet", quietly = TRUE)) {
    expect_error(simulate_leakage_suite(
      n = 80, p = 5, prevalence = 0.5,
      mode = "batch_blocked", learner = "glmnet",
      leakage = "none", rho = 0, K = 3, repeats = 1,
      B = 2, seeds = 1, parallel = FALSE, signal_strength = 1
    ), "glmnet")
  } else {
    res <- simulate_leakage_suite(
      n = 80, p = 5, prevalence = 0.5,
      mode = "batch_blocked", learner = "glmnet",
      leakage = "none", rho = 0, K = 3, repeats = 1,
      B = 2, seeds = 1, parallel = FALSE, signal_strength = 1
    )
    expect_true(inherits(res, "LeakSimResults"))
    expect_true(all(c("seed", "metric_obs", "gap", "p_value", "leakage", "mode") %in% names(res)))
  }
})

# test_that("simulate_leakage_suite accepts preprocess overrides", {
#   skip_if_not_installed("glmnet")
#   res <- simulate_leakage_suite(
#     n = 80, p = 4, prevalence = 0.5,
#     mode = "subject_grouped", learner = "glmnet",
#     leakage = "peek_norm",
#     preprocess = list(normalize = list(method = "none")),
#     rho = 0, K = 3, repeats = 1,
#     B = 2, seeds = 1, parallel = FALSE, signal_strength = 1
#   )
#   expect_true(inherits(res, "LeakSimResults"))
# })

test_that("signal_strength increases outcome separation", {
  set.seed(123)
  sim0 <- bioLeak:::.simulate_dataset(
    n = 500, p = 10, prevalence = 0.5,
    mode = "subject_grouped", leakage = "none",
    rho = 0, signal_strength = 0
  )
  set.seed(123)
  sim2 <- bioLeak:::.simulate_dataset(
    n = 500, p = 10, prevalence = 0.5,
    mode = "subject_grouped", leakage = "none",
    rho = 0, signal_strength = 2
  )

  feats <- paste0("x", sprintf("%02d", 1:5))
  score <- as.numeric(scale(rowSums(sim0$data[, feats, drop = FALSE])))
  y0 <- as.numeric(sim0$data$y) - 1
  y2 <- as.numeric(sim2$data$y) - 1

  cor0 <- cor(score, y0)
  cor2 <- cor(score, y2)

  expect_gt(cor2, cor0 + 0.2)
})

# Provenance capture for reproducibility tracking ---------------------

#' Capture session provenance information
#'
#' Internal helper that records R version, loaded packages, platform,
#' timestamp, and optionally git SHA and hardware details.
#'
#' @param capture_git Logical; if TRUE, attempts to capture the current
#'   git short SHA via \code{system2("git", ...)}. Default FALSE.
#' @param capture_hardware Logical; if TRUE, captures basic hardware info
#'   from \code{Sys.info()}. Default TRUE.
#' @return A list with elements \code{r_version}, \code{packages} (data.frame),
#'   \code{platform}, \code{timestamp}, \code{git_sha}, and \code{hardware}.
#' @keywords internal
.bio_capture_provenance <- function(capture_git = TRUE, capture_hardware = TRUE) {
  si <- utils::sessionInfo()

  # Extract package versions from sessionInfo
  base_pkgs <- data.frame(
    package = si$basePkgs,
    version = vapply(si$basePkgs, function(p) {
      tryCatch(as.character(utils::packageVersion(p)), error = function(e) NA_character_)
    }, character(1)),
    stringsAsFactors = FALSE,
    row.names = NULL
  )

  other_pkgs <- if (length(si$otherPkgs)) {
    data.frame(
      package = vapply(si$otherPkgs, function(p) p$Package, character(1)),
      version = vapply(si$otherPkgs, function(p) p$Version, character(1)),
      stringsAsFactors = FALSE,
      row.names = NULL
    )
  } else {
    data.frame(package = character(0), version = character(0),
               stringsAsFactors = FALSE)
  }

  packages <- rbind(base_pkgs, other_pkgs)

  # Git SHA
 git_sha <- NA_character_
  if (isTRUE(capture_git)) {
    git_sha <- tryCatch({
      res <- system2("git", c("rev-parse", "--short", "HEAD"),
                      stdout = TRUE, stderr = TRUE)
      if (length(res) && !inherits(res, "try-error") && nchar(res[1]) > 0) {
        res[1]
      } else {
        NA_character_
      }
    }, error = function(e) NA_character_,
       warning = function(w) NA_character_)
  }

  # Hardware info
  hardware <- NULL
  if (isTRUE(capture_hardware)) {
    sinfo <- Sys.info()
    hardware <- list(
      sysname = unname(sinfo["sysname"]),
      nodename = unname(sinfo["nodename"]),
      machine = unname(sinfo["machine"])
    )
  }

  list(
    r_version = paste0(R.version$major, ".", R.version$minor),
    packages = packages,
    platform = si$platform,
    timestamp = Sys.time(),
    git_sha = git_sha,
    hardware = hardware
  )
}

test_that(".bio_capture_provenance returns list with all required keys", {
  prov <- bioLeak:::.bio_capture_provenance()
  expect_true(is.list(prov))
  required_keys <- c("r_version", "packages", "platform", "timestamp",
                      "git_sha", "hardware")
  expect_true(all(required_keys %in% names(prov)))
})

test_that("packages is a data.frame with package and version columns", {
  prov <- bioLeak:::.bio_capture_provenance()
  expect_true(is.data.frame(prov$packages))
  expect_true("package" %in% names(prov$packages))
  expect_true("version" %in% names(prov$packages))
  expect_gt(nrow(prov$packages), 0)
})

test_that("git_sha is NA when capture_git = FALSE", {
  prov <- bioLeak:::.bio_capture_provenance(capture_git = FALSE)
  expect_true(is.na(prov$git_sha))
})

test_that("hardware is NULL when capture_hardware = FALSE", {
  prov <- bioLeak:::.bio_capture_provenance(capture_hardware = FALSE)
  expect_null(prov$hardware)
})

test_that("hardware is a list with expected keys when capture_hardware = TRUE", {
  prov <- bioLeak:::.bio_capture_provenance(capture_hardware = TRUE)
  expect_true(is.list(prov$hardware))
  expect_true(all(c("sysname", "nodename", "machine") %in% names(prov$hardware)))
})

test_that("fit_resample output has provenance in info slot", {
  set.seed(42)
  df <- data.frame(
    subject = rep(1:30, each = 3),
    outcome = rep(c(0, 1), length.out = 90),
    x1 = rnorm(90),
    x2 = rnorm(90)
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, progress = FALSE)
  fit <- fit_resample(df, outcome = "outcome", splits = splits,
                      learner = "glmnet", preprocess = list(
                        impute = list(method = "median"),
                        normalize = list(method = "zscore")
                      ))
  expect_true("provenance" %in% names(fit@info))
  prov <- fit@info$provenance
  expect_true(is.list(prov))
  expect_true(nchar(prov$r_version) > 0)
  expect_true(is.data.frame(prov$packages))
})

# audit_leakage(): permutation gap, batch association, target scan, duplicates, trail ----

.cosine_sim_block <- function(A, B = NULL) {
  if (is.null(B)) B <- A
  A <- as.matrix(A); B <- as.matrix(B)
  An <- sqrt(rowSums(A * A)); Bn <- sqrt(rowSums(B * B))
  S <- A %*% t(B)
  S / (An %o% Bn + 1e-12)
}

# Row-normalize (L2) to allow cosine distance via Euclidean NN
.row_l2_normalize <- function(M) {
  M <- as.matrix(M)
  nrm <- sqrt(rowSums(M * M))
  nrm[nrm == 0] <- 1
  M / nrm
}

.row_center <- function(M) {
  M <- as.matrix(M)
  mu <- rowMeans(M, na.rm = TRUE)
  sweep(M, 1, mu, "-")
}

# Safe chi^2 with Cramer's V
.chisq_assoc <- function(tab) {
  if (is.null(tab) || length(tab) == 0) {
    return(list(stat = NA_real_, df = NA_integer_, pval = NA_real_, cramer_v = NA_real_))
  }
  if (!is.null(dim(tab))) {
    row_sums <- rowSums(tab)
    col_sums <- colSums(tab)
    if (any(row_sums == 0) || any(col_sums == 0)) {
      tab <- tab[row_sums > 0, col_sums > 0, drop = FALSE]
    }
  }
  if (any(dim(tab) < 2) || sum(tab) == 0) {
    return(list(stat = NA_real_, df = NA_integer_, pval = NA_real_, cramer_v = NA_real_))
  }
  cs <- suppressWarnings(stats::chisq.test(tab))
  n <- sum(tab)
  v <- sqrt(unname(cs$statistic) / (n * (min(dim(tab)) - 1)))
  list(stat = unname(cs$statistic), df = unname(cs$parameter),
       pval = unname(cs$p.value), cramer_v = as.numeric(v))
}

.chisq_assoc_soft <- function(tab) {
  if (any(dim(tab) < 2) || sum(tab) == 0) {
    return(list(stat = NA_real_, df = NA_integer_, pval = NA_real_, cramer_v = NA_real_))
  }
  cs <- suppressWarnings(stats::chisq.test(tab))
  n <- sum(tab)
  v <- sqrt(unname(cs$statistic) / (n * (min(dim(tab)) - 1)))
  list(stat = unname(cs$statistic), df = unname(cs$parameter),
       pval = unname(cs$p.value), cramer_v = as.numeric(v))
}

.auc_rank <- function(x, y01) {
  x <- as.numeric(x)
  y01 <- as.numeric(y01)
  ok <- is.finite(x) & !is.na(y01)
  x <- x[ok]
  y01 <- y01[ok]
  if (!length(x)) return(NA_real_)
  n_pos <- sum(y01 == 1)
  n_neg <- sum(y01 == 0)
  if (n_pos == 0 || n_neg == 0) return(NA_real_)
  r <- rank(x, ties.method = "average")
  (sum(r[y01 == 1]) - n_pos * (n_pos + 1) / 2) / (n_pos * n_neg)
}

.cor_pval <- function(r, n) {
  if (!is.finite(r) || is.na(r) || n < 3L) return(NA_real_)
  if (abs(r) >= 1) return(0)
  tval <- r * sqrt((n - 2) / (1 - r^2))
  2 * stats::pt(-abs(tval), df = n - 2)
}

.bio_leakage_taxonomy <- function() {
  data.frame(
    mechanism_class = c(
      "non_random_signal",
      "confounding_alignment",
      "proxy_target_leakage",
      "duplicate_overlap",
      "temporal_lookahead"
    ),
    evidence_slot = c(
      "permutation_gap",
      "batch_assoc",
      "target_assoc",
      "duplicates",
      "duplicates"
    ),
    primary_statistics = c(
      "gap,p_value",
      "pval,cramer_v",
      "score,p_value,p_value_adj",
      "similarity,cross_fold",
      "cross_fold,time-aware split"
    ),
    description = c(
      "Model signal exceeds permutation null.",
      "Fold assignments align with batch/study metadata.",
      "Predictor-outcome association indicates potential target proxy leakage.",
      "Near-duplicate samples bridge train/test partitions.",
      "Similarity links violate temporal ordering constraints."
    ),
    stringsAsFactors = FALSE
  )
}

.bio_mechanism_summary <- function(perm_df, batch_df, target_df, dup_df, split_mode = NA_character_) {
  out <- list()

  perm_gap <- if (!is.null(perm_df) && nrow(perm_df)) perm_df$gap[[1]] else NA_real_
  perm_p <- if (!is.null(perm_df) && nrow(perm_df)) perm_df$p_value[[1]] else NA_real_
  out[[length(out) + 1L]] <- data.frame(
    mechanism_class = "non_random_signal",
    flagged = is.finite(perm_p) && perm_p <= 0.05 && is.finite(perm_gap) && perm_gap > 0,
    evidence = "permutation_gap",
    statistic = perm_gap,
    p_value = perm_p,
    stringsAsFactors = FALSE
  )

  batch_p <- if (!is.null(batch_df) && nrow(batch_df) && "pval" %in% names(batch_df)) {
    suppressWarnings(min(batch_df$pval, na.rm = TRUE))
  } else {
    NA_real_
  }
  if (!is.finite(batch_p)) batch_p <- NA_real_
  batch_v <- if (!is.null(batch_df) && nrow(batch_df) && "cramer_v" %in% names(batch_df)) {
    suppressWarnings(max(batch_df$cramer_v, na.rm = TRUE))
  } else {
    NA_real_
  }
  if (!is.finite(batch_v)) batch_v <- NA_real_
  out[[length(out) + 1L]] <- data.frame(
    mechanism_class = "confounding_alignment",
    flagged = is.finite(batch_p) && batch_p <= 0.05 && is.finite(batch_v) && batch_v >= 0.1,
    evidence = "batch_assoc",
    statistic = batch_v,
    p_value = batch_p,
    stringsAsFactors = FALSE
  )

  proxy_flag <- FALSE
  proxy_stat <- NA_real_
  proxy_p <- NA_real_
  if (!is.null(target_df) && nrow(target_df)) {
    if ("flag_fdr" %in% names(target_df) && any(target_df$flag_fdr %in% TRUE, na.rm = TRUE)) {
      proxy_flag <- TRUE
    } else if ("flag" %in% names(target_df) && any(target_df$flag %in% TRUE, na.rm = TRUE)) {
      proxy_flag <- TRUE
    }
    if ("score" %in% names(target_df)) {
      proxy_stat <- suppressWarnings(max(target_df$score, na.rm = TRUE))
      if (!is.finite(proxy_stat)) proxy_stat <- NA_real_
    }
    p_col <- if ("p_value_adj" %in% names(target_df)) "p_value_adj" else if ("p_value" %in% names(target_df)) "p_value" else NULL
    if (!is.null(p_col)) {
      proxy_p <- suppressWarnings(min(target_df[[p_col]], na.rm = TRUE))
      if (!is.finite(proxy_p)) proxy_p <- NA_real_
    }
  }
  out[[length(out) + 1L]] <- data.frame(
    mechanism_class = "proxy_target_leakage",
    flagged = proxy_flag,
    evidence = "target_assoc",
    statistic = proxy_stat,
    p_value = proxy_p,
    stringsAsFactors = FALSE
  )

  dup_sim <- NA_real_
  dup_flag <- FALSE
  dup_cols <- c("sim", "cos_sim", "pearson")
  if (!is.null(dup_df) && nrow(dup_df)) {
    sim_col <- dup_cols[dup_cols %in% names(dup_df)]
    if (length(sim_col)) {
      dup_sim <- suppressWarnings(max(dup_df[[sim_col[[1]]]], na.rm = TRUE))
      if (!is.finite(dup_sim)) dup_sim <- NA_real_
    }
    # Only flag as leakage when cross-fold duplicates exist
    if ("cross_fold" %in% names(dup_df)) {
      dup_flag <- any(dup_df$cross_fold %in% TRUE, na.rm = TRUE)
    } else {
      dup_flag <- TRUE
    }
  }
  out[[length(out) + 1L]] <- data.frame(
    mechanism_class = "duplicate_overlap",
    flagged = dup_flag,
    evidence = "duplicates",
    statistic = dup_sim,
    p_value = NA_real_,
    stringsAsFactors = FALSE
  )

  lookahead_flag <- FALSE
  if (identical(split_mode, "time_series") && !is.null(dup_df) && nrow(dup_df) &&
      "cross_fold" %in% names(dup_df)) {
    lookahead_flag <- any(dup_df$cross_fold %in% TRUE, na.rm = TRUE)
  }
  out[[length(out) + 1L]] <- data.frame(
    mechanism_class = "temporal_lookahead",
    flagged = lookahead_flag,
    evidence = "duplicates",
    statistic = NA_real_,
    p_value = NA_real_,
    stringsAsFactors = FALSE
  )

  do.call(rbind, out)
}

.align_by_ids <- function(X, ids_chr, sample_ids = NULL, warn = TRUE, strict_align = FALSE) {
  if (is.null(X) || (!is.data.frame(X) && !is.matrix(X))) return(NULL)
  rn <- rownames(X)
  if (!is.null(rn) && all(ids_chr %in% rn)) {
    return(X[ids_chr, , drop = FALSE])
  }
  if (!is.null(sample_ids) && length(sample_ids) == nrow(X)) {
    idx <- match(ids_chr, as.character(sample_ids))
    if (all(!is.na(idx))) return(X[idx, , drop = FALSE])
  }
  ids_int <- suppressWarnings(as.integer(ids_chr))
  if (all(!is.na(ids_int)) && max(ids_int, na.rm = TRUE) <= nrow(X)) {
    return(X[ids_int, , drop = FALSE])
  }
  if (nrow(X) == length(ids_chr)) {
    if (isTRUE(strict_align)) {
      stop("X_ref rownames do not match prediction ids and strict_align = TRUE; ",
           "provide matching row names or a 'row_id' column.", call. = FALSE)
    }
    if (warn) {
      warning("X_ref rownames do not match prediction ids; assuming row order aligns to predictions.",
              call. = FALSE)
    }
    return(X)
  }
  if (warn) warning("X_ref not aligned to predictions; target leakage scan skipped.", call. = FALSE)
  NULL
}

.target_assoc_scan <- function(X, y, task, positive_class = NULL, threshold = 0.9) {
  if (is.null(X) || is.null(y)) return(data.frame())
  if (!is.data.frame(X) && !is.matrix(X)) return(data.frame())
  is_df <- is.data.frame(X)
  if (is_df) {
    n_rows <- nrow(X)
    n_cols <- ncol(X)
    col_names <- colnames(X)
    get_col <- function(j) X[[j]]
  } else {
    X_mat <- as.matrix(X)
    n_rows <- nrow(X_mat)
    n_cols <- ncol(X_mat)
    col_names <- colnames(X_mat)
    get_col <- function(j) X_mat[, j]
  }
  if (!n_rows || !n_cols) return(data.frame())
  if (length(y) != n_rows) {
    warning("X_ref rows do not match outcome length; target leakage scan skipped.", call. = FALSE)
    return(data.frame())
  }

  if (is.null(col_names)) {
    col_names <- paste0("feature_", seq_len(n_cols))
  }

  y01 <- NULL
  y_num <- NULL
  y_fac <- NULL
  if (task == "survival") {
    warning("Target leakage scan is not available for survival outcomes.", call. = FALSE)
    return(data.frame())
  }
  if (task == "binomial") {
    y01 <- y
    if (is.factor(y01)) {
      y01 <- droplevels(y01)
      if (!is.null(positive_class)) {
        pos_chr <- as.character(positive_class)
        if (pos_chr %in% levels(y01) && !identical(levels(y01)[2], pos_chr)) {
          y01 <- factor(y01, levels = c(setdiff(levels(y01), pos_chr), pos_chr))
        }
      }
      y01 <- as.numeric(y01) - 1
    } else if (is.logical(y01)) {
      y01 <- as.numeric(y01)
    } else {
      y01 <- as.numeric(y01)
      uniq <- sort(unique(y01[is.finite(y01)]))
      if (length(uniq) == 2 && !all(uniq %in% c(0, 1))) {
        y01 <- as.numeric(factor(y01, levels = uniq)) - 1
      }
    }
  } else if (task == "multiclass") {
    y_fac <- as.factor(y)
  } else {
    y_num <- as.numeric(y)
  }

  results <- lapply(seq_len(n_cols), function(j) {
    x <- get_col(j)
    if (inherits(x, c("Date", "POSIXct", "POSIXt"))) {
      x <- as.numeric(x)
    }
    is_cat <- is.factor(x) || is.character(x)

    if (is_cat) {
      x_fac <- as.factor(x)
      if (task == "binomial") {
        ok <- !is.na(x_fac) & !is.na(y01)
        x_ok <- droplevels(x_fac[ok])
        y_ok <- y01[ok]
        if (length(x_ok) < 2L || length(unique(x_ok)) < 2L || length(unique(y_ok)) < 2L) {
          return(NULL)
        }
        tab <- table(x_ok, y_ok)
        assoc <- .chisq_assoc_soft(tab)
        data.frame(
          feature = col_names[j],
          type = "categorical",
          metric = "cramer_v",
          value = assoc$cramer_v,
          score = assoc$cramer_v,
          p_value = assoc$pval,
          n = length(y_ok),
          n_levels = nlevels(x_ok),
          stringsAsFactors = FALSE
        )
      } else if (task == "multiclass") {
        ok <- !is.na(x_fac) & !is.na(y_fac)
        x_ok <- droplevels(x_fac[ok])
        y_ok <- droplevels(y_fac[ok])
        if (length(x_ok) < 2L || length(unique(x_ok)) < 2L || length(unique(y_ok)) < 2L) {
          return(NULL)
        }
        tab <- table(x_ok, y_ok)
        assoc <- .chisq_assoc_soft(tab)
        data.frame(
          feature = col_names[j],
          type = "categorical",
          metric = "cramer_v",
          value = assoc$cramer_v,
          score = assoc$cramer_v,
          p_value = assoc$pval,
          n = length(y_ok),
          n_levels = nlevels(x_ok),
          stringsAsFactors = FALSE
        )
      } else {
        ok <- !is.na(x_fac) & is.finite(y_num)
        x_ok <- droplevels(x_fac[ok])
        y_ok <- y_num[ok]
        if (length(y_ok) < 3L || nlevels(x_ok) < 2L) return(NULL)
        fit <- try(stats::lm(y_ok ~ x_ok), silent = TRUE)
        if (inherits(fit, "try-error")) return(NULL)
        an <- stats::anova(fit)
        ss_total <- sum(an$`Sum Sq`, na.rm = TRUE)
        eta_sq <- if (is.finite(ss_total) && ss_total > 0) an$`Sum Sq`[1] / ss_total else NA_real_
        p_val <- an$`Pr(>F)`[1]
        data.frame(
          feature = col_names[j],
          type = "categorical",
          metric = "eta_sq",
          value = eta_sq,
          score = eta_sq,
          p_value = p_val,
          n = length(y_ok),
          n_levels = nlevels(x_ok),
          stringsAsFactors = FALSE
        )
      }
    } else {
      x_num <- as.numeric(x)
      if (task == "binomial") {
        ok <- is.finite(x_num) & !is.na(y01)
        x_ok <- x_num[ok]
        y_ok <- y01[ok]
        if (length(x_ok) < 3L || length(unique(y_ok)) < 2L) return(NULL)
        if (stats::sd(x_ok) == 0) return(NULL)
        auc <- .auc_rank(x_ok, y_ok)
        score <- if (is.na(auc)) NA_real_ else abs(auc - 0.5) * 2
        p_val <- tryCatch(
          stats::wilcox.test(x_ok[y_ok == 1], x_ok[y_ok == 0], exact = FALSE)$p.value,
          error = function(e) NA_real_
        )
        data.frame(
          feature = col_names[j],
          type = "numeric",
          metric = "auc",
          value = auc,
          score = score,
          p_value = p_val,
          n = length(y_ok),
          n_levels = NA_integer_,
          stringsAsFactors = FALSE
        )
      } else if (task == "multiclass") {
        ok <- is.finite(x_num) & !is.na(y_fac)
        x_ok <- x_num[ok]
        y_ok <- droplevels(y_fac[ok])
        if (length(x_ok) < 3L || length(unique(y_ok)) < 2L) return(NULL)
        if (stats::sd(x_ok) == 0) return(NULL)
        fit <- try(stats::lm(x_ok ~ y_ok), silent = TRUE)
        if (inherits(fit, "try-error")) return(NULL)
        an <- stats::anova(fit)
        ss_total <- sum(an$`Sum Sq`, na.rm = TRUE)
        eta_sq <- if (is.finite(ss_total) && ss_total > 0) an$`Sum Sq`[1] / ss_total else NA_real_
        p_val <- an$`Pr(>F)`[1]
        data.frame(
          feature = col_names[j],
          type = "numeric",
          metric = "eta_sq",
          value = eta_sq,
          score = eta_sq,
          p_value = p_val,
          n = length(y_ok),
          n_levels = nlevels(y_ok),
          stringsAsFactors = FALSE
        )
      } else {
        ok <- is.finite(x_num) & is.finite(y_num)
        x_ok <- x_num[ok]
        y_ok <- y_num[ok]
        if (length(x_ok) < 3L) return(NULL)
        if (stats::sd(x_ok) == 0 || stats::sd(y_ok) == 0) return(NULL)
        r <- stats::cor(x_ok, y_ok)
        p_val <- .cor_pval(r, length(x_ok))
        data.frame(
          feature = col_names[j],
          type = "numeric",
          metric = "cor",
          value = r,
          score = abs(r),
          p_value = p_val,
          n = length(y_ok),
          n_levels = NA_integer_,
          stringsAsFactors = FALSE
        )
      }
    }
  })

  results <- results[!vapply(results, is.null, logical(1))]
  if (!length(results)) return(data.frame())
  out <- do.call(rbind, results)
  out$flag <- !is.na(out$score) & out$score >= threshold
  out <- out[order(out$score, decreasing = TRUE, na.last = TRUE), , drop = FALSE]
  rownames(out) <- NULL
  out
}

.target_scan_prepare_matrix <- function(X) {
  if (is.null(X) || (!is.data.frame(X) && !is.matrix(X))) return(NULL)
  X_mat <- if (is.data.frame(X)) {
    mm <- try(stats::model.matrix(~ . - 1, data = X), silent = TRUE)
    if (inherits(mm, "try-error")) return(NULL)
    mm
  } else {
    as.matrix(X)
  }
  if (!nrow(X_mat) || !ncol(X_mat)) return(NULL)
  storage.mode(X_mat) <- "numeric"
  for (j in seq_len(ncol(X_mat))) {
    col <- X_mat[, j]
    med <- stats::median(col, na.rm = TRUE)
    if (!is.finite(med)) med <- 0
    col[is.na(col)] <- med
    X_mat[, j] <- col
  }
  sdv <- vapply(seq_len(ncol(X_mat)), function(j) {
    stats::sd(X_mat[, j], na.rm = TRUE)
  }, numeric(1))
  keep <- is.finite(sdv) & sdv > 0
  X_mat <- X_mat[, keep, drop = FALSE]
  if (!ncol(X_mat)) return(NULL)
  X_mat
}

.target_scan_multivariate <- function(X, y, task, splits, folds, coldata, seed, B,
                                      max_components = 10L, interactions = TRUE,
                                      positive_class = NULL, permute_outcome = NULL) {
  if (is.null(X) || is.null(y)) return(data.frame())
  if (!task %in% c("binomial", "gaussian")) return(data.frame())
  X_mat <- .target_scan_prepare_matrix(X)
  if (is.null(X_mat)) return(data.frame())
  if (length(y) != nrow(X_mat)) return(data.frame())

  if (isTRUE(splits@info$compact) && identical(splits@mode, "time_series") && is.null(coldata)) {
    warning("time_series compact splits require coldata for multivariate target scan; skipping.",
            call. = FALSE)
    return(data.frame())
  }

  max_components <- as.integer(max_components)
  if (!is.finite(max_components) || max_components < 1L) return(data.frame())
  n <- nrow(X_mat)
  k <- min(max_components, n - 1L, ncol(X_mat))
  if (!is.finite(k) || k < 1L) return(data.frame())

  X_scaled <- scale(X_mat, center = TRUE, scale = TRUE)
  pc <- try(stats::prcomp(X_scaled, center = FALSE, scale. = FALSE, rank. = k), silent = TRUE)
  if (inherits(pc, "try-error") || is.null(pc$x)) return(data.frame())
  k_eff <- ncol(pc$x)
  if (!is.finite(k_eff) || k_eff < 1L) return(data.frame())
  X_design <- pc$x[, seq_len(k_eff), drop = FALSE]
  colnames(X_design) <- paste0("PC", seq_len(ncol(X_design)))

  n_interactions <- 0L
  if (isTRUE(interactions) && ncol(X_design) >= 2L) {
    k_int <- min(5L, ncol(X_design))
    comb <- utils::combn(k_int, 2)
    int_mat <- matrix(NA_real_, nrow = n, ncol = ncol(comb))
    colnames(int_mat) <- apply(comb, 2, function(idx) {
      paste0("PC", idx[1], "_x_PC", idx[2])
    })
    for (j in seq_len(ncol(comb))) {
      int_mat[, j] <- X_design[, comb[1, j]] * X_design[, comb[2, j]]
    }
    X_design <- cbind(X_design, int_mat)
    n_interactions <- ncol(int_mat)
  }

  score_for_y <- function(y_vec) {
    if (identical(task, "binomial")) {
      y_fac <- if (is.factor(y_vec)) droplevels(y_vec) else factor(y_vec)
      if (nlevels(y_fac) < 2L) {
        return(list(score = NA_real_, value = NA_real_, n_obs = 0L, metric = "auc"))
      }
      pos <- positive_class
      if (is.null(pos) || !as.character(pos) %in% levels(y_fac)) {
        pos <- levels(y_fac)[2]
      }
      y01 <- as.numeric(y_fac == pos)
    } else {
      y_num <- as.numeric(y_vec)
    }

    preds <- rep(NA_real_, n)
    for (i in seq_along(folds)) {
      fold_full <- try(.bio_resolve_fold_indices(splits, folds[[i]], n = n, data = coldata),
                       silent = TRUE)
      if (inherits(fold_full, "try-error")) next
      tr <- fold_full$train
      te <- fold_full$test
      if (!length(tr) || !length(te)) next
      if (identical(task, "binomial")) {
        y_tr <- y01[tr]
        ok_tr <- which(!is.na(y_tr))
        if (length(ok_tr) < 2L) next
        y_tr <- y_tr[ok_tr]
        if (length(unique(y_tr)) < 2L) next
        df_tr <- data.frame(y = y_tr, X_design[tr[ok_tr], , drop = FALSE], check.names = FALSE)
        fit <- suppressWarnings(
          try(stats::glm(y ~ ., data = df_tr, family = stats::binomial()), silent = TRUE)
        )
        if (inherits(fit, "try-error")) next
        df_te <- as.data.frame(X_design[te, , drop = FALSE], check.names = FALSE)
        pr <- try(stats::predict(fit, newdata = df_te, type = "response"), silent = TRUE)
        if (inherits(pr, "try-error")) next
        preds[te] <- as.numeric(pr)
      } else {
        y_tr <- y_num[tr]
        ok_tr <- which(is.finite(y_tr))
        if (length(ok_tr) < 2L) next
        df_tr <- data.frame(y = y_tr[ok_tr], X_design[tr[ok_tr], , drop = FALSE], check.names = FALSE)
        fit <- try(stats::lm(y ~ ., data = df_tr), silent = TRUE)
        if (inherits(fit, "try-error")) next
        df_te <- as.data.frame(X_design[te, , drop = FALSE], check.names = FALSE)
        pr <- try(stats::predict(fit, newdata = df_te), silent = TRUE)
        if (inherits(pr, "try-error")) next
        preds[te] <- as.numeric(pr)
      }
    }

    if (identical(task, "binomial")) {
      ok <- is.finite(preds) & !is.na(y01)
      if (sum(ok) < 3L) {
        return(list(score = NA_real_, value = NA_real_, n_obs = sum(ok), metric = "auc"))
      }
      auc <- .auc_rank(preds[ok], y01[ok])
      score <- if (is.na(auc)) NA_real_ else abs(auc - 0.5) * 2
      return(list(score = score, value = auc, n_obs = sum(ok), metric = "auc"))
    }
    ok <- is.finite(preds) & is.finite(y_num)
    if (sum(ok) < 3L) {
      return(list(score = NA_real_, value = NA_real_, n_obs = sum(ok), metric = "cor"))
    }
    r <- stats::cor(preds[ok], y_num[ok])
    list(score = abs(r), value = r, n_obs = sum(ok), metric = "cor")
  }

  obs <- score_for_y(y)
  perm_scores <- numeric(0)
  if (is.finite(B) && B >= 1L) {
    perm_scores <- vapply(seq_len(B), function(b) {
      set.seed(seed + b)
      y_perm <- if (!is.null(permute_outcome)) permute_outcome(b) else sample(y)
      if (length(y_perm) != length(y)) return(NA_real_)
      score_for_y(y_perm)$score
    }, numeric(1))
  }
  pval <- NA_real_
  if (is.finite(obs$score) && length(perm_scores) && any(is.finite(perm_scores))) {
    finite_perm <- is.finite(perm_scores)
    pval <- (1 + sum(perm_scores[finite_perm] >= obs$score)) / (1 + sum(finite_perm))
  }

  data.frame(
    scan = "multivariate",
    metric = obs$metric,
    value = obs$value,
    score = obs$score,
    p_value = pval,
    n = obs$n_obs,
    n_features = ncol(X_mat),
    n_components = k_eff,
    n_interactions = n_interactions,
    n_perm = length(perm_scores),
    stringsAsFactors = FALSE
  )
}

# Compute metric from predictions
.metric_value <- function(metric, task, truth, pred, pred_df = NULL) {
  if (task == "survival") {
    if (!inherits(truth, "Surv") && !is.null(pred_df)) {
      if (!requireNamespace("survival", quietly = TRUE)) return(NA_real_)
      if (all(c("truth_time", "truth_event") %in% names(pred_df))) {
        truth <- survival::Surv(pred_df$truth_time, pred_df$truth_event)
      } else if (all(c("time", "status") %in% names(pred_df))) {
        truth <- survival::Surv(pred_df$time, pred_df$status)
      }
    }
  }
  if (metric == "rmse") {
    return(sqrt(mean((as.numeric(truth) - pred)^2)))
  }
  if (metric == "accuracy" && task %in% c("binomial", "multiclass")) {
    pred_class <- if (!is.null(pred_df) && "pred_class" %in% names(pred_df)) {
      pred_df$pred_class
    } else if (is.factor(pred) || is.character(pred)) {
      pred
    } else {
      if (task == "binomial") {
        if (is.factor(truth)) {
          factor(ifelse(pred >= 0.5, levels(truth)[2], levels(truth)[1]),
                 levels = levels(truth))
        } else {
          ifelse(pred >= 0.5, 1, 0)
        }
      } else {
        pred
      }
    }
    return(.multiclass_accuracy(truth, pred_class))
  }
  if (metric == "macro_f1" && task == "multiclass") {
    pred_class <- if (!is.null(pred_df) && "pred_class" %in% names(pred_df)) {
      pred_df$pred_class
    } else {
      pred
    }
    return(.multiclass_macro_f1(truth, pred_class))
  }
  if (metric == "log_loss" && task == "multiclass") {
    prob <- NULL
    if (!is.null(pred_df)) {
      prob_cols <- grep("^\\.pred_", names(pred_df), value = TRUE)
      if (length(prob_cols)) {
        prob <- as.matrix(pred_df[, prob_cols, drop = FALSE])
      }
    }
    return(.multiclass_log_loss(truth, prob))
  }
  if (metric == "log_loss" && task == "binomial") {
    yb <- if (is.factor(truth)) as.numeric(truth) - 1 else as.numeric(truth)
    p <- pmin(pmax(as.numeric(pred), 1e-15), 1 - 1e-15)
    return(-mean(yb * log(p) + (1 - yb) * log(1 - p), na.rm = TRUE))
  }
  if (metric == "auc") {
    if (requireNamespace("pROC", quietly = TRUE)) {
      roc <- pROC::roc(truth, pred, quiet = TRUE)
      return(as.numeric(pROC::auc(roc)))
    }
    yb <- if (is.factor(truth)) as.numeric(truth) - 1 else truth
    pos <- pred[yb == 1]; neg <- pred[yb == 0]
    if (length(pos) && length(neg)) {
      comp <- outer(pos, neg, function(a, b) (a > b) + 0.5 * (a == b))
      return(mean(comp))
    }
    return(NA_real_)
  }
  if (metric == "pr_auc") {
    if (requireNamespace("PRROC", quietly = TRUE)) {
      yb <- if (is.factor(truth)) as.numeric(truth) - 1 else truth
      pr <- PRROC::pr.curve(scores.class0 = pred[yb == 1],
                            scores.class1 = pred[yb == 0], curve = FALSE)
      return(pr$auc.integral)
    }
    return(NA_real_)
  }
  if (metric == "cindex") {
    if (task == "survival") {
      return(.cindex_survival(pred, truth))
    }
    return(.cindex_pairwise(pred, truth))
  }
  NA_real_
}

#' Audit leakage and confounding
#'
#' @description
#' Computes a post-hoc leakage audit for a resampled model fit. The audit
#' (1) runs a permutation-gap test comparing observed cross-validated
#' performance to a label-permutation null (by default refitting when data
#' are available; otherwise using fixed predictions),
#' (2) tests whether fold assignments are associated with batch or study
#' metadata (confounding by design),
#' (3) scans features for unusually strong outcome proxies, and (4) flags
#' duplicate or near-duplicate samples in a reference feature matrix.
#'
#' The returned [LeakAudit] summarizes these diagnostics. It relies on the
#' stored predictions, splits, and optional metadata; it does not refit models
#' unless `perm_refit = TRUE` (or `perm_refit = "auto"` with a valid
#' `perm_refit_spec`). Results are conditional on the chosen metric
#' and supplied metadata/features and should be interpreted as diagnostics,
#' not proof of leakage or its absence.
#'
#' @param fit A [LeakFit] object from [fit_resample()] containing cross-validated
#'   predictions and split metadata. If predictions include learner IDs for
#'   multiple models, you must supply `learner` to select one; if learner IDs are
#'   absent, the audit uses all predictions and may mix learners.
#' @param metric Character scalar. One of `"auc"`, `"pr_auc"`, `"accuracy"`,
#'   `"macro_f1"`, `"log_loss"`, `"rmse"`, or `"cindex"`. Defaults to `"auc"`.
#'   This controls the observed performance
#'   statistic, the permutation null, and the sign of the reported gap.
#' @param B Integer scalar. Number of permutations used to build the null
#'   distribution (default 200). Larger values reduce Monte Carlo error but
#'   increase runtime.
#' @param perm_stratify Logical scalar or `"auto"`. If TRUE, permutations
#'   are stratified within each fold (factor levels; numeric outcomes are binned
#'   into quantiles when enough non-missing values are available). If FALSE, no
#'   stratification is used. Defaults to FALSE. Stratification only applies when
#'   `coldata` supplies the outcome; otherwise labels are shuffled within each fold.
#' @param perm_refit Logical scalar or `"auto"`. If FALSE, permutations keep
#'   predictions fixed and shuffle labels (association test). If TRUE, each
#'   permutation refits the model on permuted outcomes using `perm_refit_spec`.
#'   Refit-based permutations are slower but better approximate a full null
#'   distribution. The default is `"auto"`, which refits only when
#'   `perm_refit_spec` is provided and `B` is less than or equal to
#'   `perm_refit_auto_max`; otherwise it falls back to fixed-prediction
#'   permutations.
#' @param perm_refit_auto_max Integer scalar. Maximum `B` allowed for
#'   `perm_refit = "auto"` to trigger refitting. Defaults to 200.
#' @param perm_refit_spec List of inputs used when `perm_refit = TRUE`.
#'   Required elements: `x` (data used for fitting) and `learner` (parsnip
#'   model_spec, workflow, or legacy learner). Optional elements: `outcome`
#'   (defaults to `fit@outcome`), `preprocess`, `learner_args`,
#'   `custom_learners`, `class_weights`, `positive_class`, and `parallel`.
#'   Survival outcomes are not supported for refit-based permutations.
#' @param perm_mode Optional character scalar to override the permutation mode
#'   used for restricted shuffles. One of `"subject_grouped"`, `"batch_blocked"`,
#'   `"study_loocv"`, or `"time_series"`. Defaults to the split metadata when
#'   available (including rsample-derived modes).
#' @param time_block Character scalar, `"circular"` or `"stationary"`. Controls
#'   block permutation for `time_series` splits; ignored for other split modes.
#'   Default is `"circular"`.
#' @param block_len Integer scalar or NULL. Block length for time-series
#'   permutations. NULL selects `max(5, floor(0.1 * fold_size))`. Larger values
#'   preserve more temporal structure and yield a more conservative null.
#' @param include_z Logical scalar. If TRUE (default), include the z-score for the
#'   permutation gap when a standard error is available; if FALSE, `z` is NA.
#' @param ci_method Character scalar, `"if"` or `"bootstrap"`. Controls how the
#'   standard error and confidence interval for the permutation gap are estimated.
#'   Default is `"if"`. `"if"` uses an influence-function estimate when available; `"bootstrap"`
#'   resamples permutation values `boot_B` times. Failed estimates yield NA.
#' @param boot_B Integer scalar. Number of bootstrap resamples when
#'   `ci_method = "bootstrap"` (default 400). Larger values are more stable but slower.
#' @param parallel Logical scalar. If TRUE and `future.apply` is available,
#'   permutations run in parallel. Results should match sequential execution.
#'   Default is FALSE.
#' @param seed Integer scalar. Random seed used for permutations and bootstrap
#'   resampling; changing it changes the randomization but not the observed metric.
#'   Default is 1.
#' @param return_perm Logical scalar. If TRUE (default), stores the permutation
#'   distribution in `audit@perm_values`. Set FALSE to reduce memory use.
#' @param batch_cols Character vector. Names of `coldata` columns to test for
#'   association with fold assignment. If NULL, defaults to any of
#'   `"batch"`, `"plate"`, `"center"`, `"site"`, `"study"` found in `coldata`.
#'   Changing this controls which batch tests appear in `batch_assoc`.
#' @param coldata Optional data.frame of sample-level metadata. Rows must align
#'   to prediction ids via row names, a `row_id` column, or row order. Used to
#'   build restricted permutations (when the outcome column is present), compute
#'   batch associations, and supply outcomes for target scans. If NULL, uses
#'   `fit@splits@info$coldata` when available. If alignment fails, restricted
#'   permutations are disabled with a warning.
#' @param X_ref Optional numeric matrix/data.frame (samples x features). Used for
#'   duplicate detection and the target leakage scan. If NULL, uses
#'   `fit@info$X_ref` when available. Rows must align to sample ids (split order)
#'   via row names, a `row_id` column, or row order; misalignment disables these
#'   checks.
#' @param target_scan Logical scalar. If TRUE (default), computes per-feature
#'   outcome associations on `X_ref` and flags proxy features; if FALSE, or if
#'   `X_ref`/outcomes are unavailable, `target_assoc` is empty. Not available
#'   for survival outcomes.
#' @param target_scan_multivariate Logical scalar. If TRUE (default), fits a simple
#'   multivariate/interaction model on `X_ref` using the stored splits and
#'   reports a permutation-based score/p-value. This is slower and only
#'   implemented for binomial and gaussian tasks.
#' @param target_scan_multivariate_B Integer scalar. Number of permutations for
#'   the multivariate scan (default 100). Larger values stabilize the p-value.
#' @param target_scan_multivariate_components Integer scalar. Maximum number of
#'   principal components used in the multivariate scan (default 10).
#' @param target_scan_multivariate_interactions Logical scalar. If TRUE (default),
#'   adds pairwise interactions among the top components in the multivariate scan.
#' @param target_threshold Numeric scalar in (0,1). Threshold applied to the
#'   association score used to flag proxy features. Higher values are stricter.
#'   Default is 0.9.
#' @param target_p_adjust Character scalar. Multiple-testing correction method
#'   applied to finite `target_assoc$p_value` values. One of `"none"` (default),
#'   `"BH"`, `"BY"`, `"holm"`, or `"bonferroni"`. Adds columns
#'   `p_value_adj` and `flag_fdr` to `target_assoc`.
#' @param target_alpha Numeric scalar in (0,1). Significance level used for
#'   `flag_fdr` when `target_p_adjust != "none"`. Default is 0.05.
#' @param feature_space Character scalar, `"raw"` or `"rank"`. If `"rank"`,
#'   each row of `X_ref` is rank-transformed before similarity calculations.
#'   This affects duplicate detection only. Default is `"raw"`.
#' @param sim_method Character scalar, `"cosine"` or `"pearson"`. Similarity
#'   metric for duplicate detection. `"pearson"` row-centers before cosine.
#'   Default is `"cosine"`.
#' @param sim_threshold Numeric scalar in (0,1). Similarity cutoff for reporting
#'   duplicate pairs (default 0.995). Higher values yield fewer pairs.
#' @param nn_k Integer scalar. For large datasets (`n > 3000`) with `RANN`
#'   installed, checks only the nearest `nn_k` neighbors per row. Larger values
#'   increase sensitivity but slow the search. Ignored when full comparisons are used.
#'   Default is 50.
#' @param max_pairs Integer scalar. Maximum number of duplicate pairs returned.
#'   If more pairs are found, only the most similar are kept. This does not
#'   affect permutation results. Default is 5000.
#' @param duplicate_scope Character scalar. One of `"train_test"` (default) or
#'   `"all"`. `"train_test"` retains only near-duplicate pairs that appear in
#'   train vs test in at least one repeat; `"all"` reports all near-duplicate
#'   pairs in `X_ref` regardless of fold assignment.
#' @param learner Optional character scalar. When predictions include multiple
#'   learner IDs, selects the learner to audit. If NULL and multiple learners
#'   are present, the function errors; if predictions lack learner IDs, this
#'   argument is ignored with a warning. Default is NULL.
#' @param strict_align Logical scalar. If TRUE, errors instead of warning when
#'   \code{X_ref} or \code{coldata} cannot be aligned to predictions by row
#'   names or IDs and would otherwise fall back to row-order matching. Default
#'   is FALSE for backward compatibility. Set to TRUE in production pipelines to
#'   catch silent misalignment.
#' @return A \code{\linkS4class{LeakAudit}} S4 object containing:
#'   \describe{
#'     \item{\code{fit}}{The \code{LeakFit} object that was audited.}
#'     \item{\code{permutation_gap}}{One-row data.frame from the permutation-gap
#'       test with columns: \code{metric_obs} (observed cross-validated metric),
#'       \code{perm_mean} (mean of permuted metrics), \code{perm_sd} (standard
#'       deviation), \code{gap} (observed minus permuted mean, or vice versa for
#'       loss metrics), \code{z} (standardized gap), \code{p_value}
#'       (permutation p-value), and \code{n_perm} (number of permutations). A
#'       large positive gap and small p-value suggest the model captures signal
#'       beyond random label assignment.}
#'     \item{\code{perm_values}}{Numeric vector of length \code{B} containing
#'       the metric value from each permutation. Useful for plotting the null
#'       distribution. Empty if \code{return_perm = FALSE}.}
#'     \item{\code{batch_assoc}}{Data.frame of chi-square association tests
#'       between fold assignment and batch/study metadata, with columns:
#'       \code{variable}, \code{stat} (chi-square statistic), \code{df}
#'       (degrees of freedom), \code{pval}, and \code{cramer_v} (effect size).
#'       Small p-values indicate potential confounding by design.}
#'     \item{\code{target_assoc}}{Data.frame of per-feature outcome associations
#'       with columns: \code{feature}, \code{type} (\code{"numeric"} or
#'       \code{"categorical"}), \code{metric} (AUC, correlation, eta_sq, or
#'       Cramer's V depending on task), \code{value}, \code{score} (scaled
#'       effect size), \code{p_value}, \code{n}, and \code{flag} (TRUE if
#'       \code{score >= target_threshold}). Flagged features may indicate
#'       target leakage.}
#'     \item{\code{duplicates}}{Data.frame of near-duplicate sample pairs with
#'       columns: \code{i}, \code{j} (row indices in \code{X_ref}), \code{sim}
#'       (similarity value), and \code{cross_fold} (whether the pair spans
#'       train vs test). Duplicates across folds can inflate performance.}
#'     \item{\code{trail}}{List capturing audit parameters and intermediate
#'       results for reproducibility, including \code{metric}, \code{B},
#'       \code{seed}, \code{perm_stratify}, \code{perm_refit}, and timing info.}
#'     \item{\code{info}}{List with additional metadata including multivariate
#'       scan results when \code{target_scan_multivariate = TRUE}.}
#'   }
#'   Use \code{summary()} to print a human-readable report, or access slots
#'   directly with \code{@}.
#' @details
#' The `permutation_gap` slot reports `metric_obs`, `perm_mean`, `perm_sd`,
#' `gap`, `z`, `p_value`, and `n_perm`. The gap is defined as
#' `metric_obs - perm_mean` for metrics where higher is better (AUC, PR-AUC,
#' accuracy, macro-F1, C-index) and `perm_mean - metric_obs` for RMSE/log-loss.
#
#' By default, `perm_refit = "auto"` refits models when refit data are available
#' and `B` is not too large; otherwise it keeps predictions fixed and shuffles
#' labels. Fixed-prediction permutations quantify prediction-label association
#' rather than a full refit null. Set `perm_refit = FALSE` to force fixed
#' predictions, or `perm_refit = TRUE` (with `perm_refit_spec`) to always refit.
#'
#' `batch_assoc` contains chi-square tests between fold assignment and each
#' `batch_cols` variable (`stat`, `df`, `pval`, `cramer_v`). `target_assoc`
#' reports feature-wise outcome associations on `X_ref`; numeric features use
#' AUC (binomial), `eta_sq` (multiclass), or correlation (gaussian), while
#' categorical features use Cramer's V (binomial/multiclass) or `eta_sq` from a
#' one-way ANOVA (gaussian). The
#' `score` column is the scaled effect size used for heuristic flagging
#' (`flag = score >= target_threshold`).
#' When `target_p_adjust != "none"`, finite `p_value` entries also receive
#' multiplicity-adjusted `p_value_adj` and `flag_fdr = (p_value_adj <= target_alpha)`.
#' The univariate target leakage scan can miss multivariate proxies, interaction
#' leakage, or features not included in `X_ref`. The multivariate scan (enabled
#' by default for supported tasks) adds a model-based proxy check but still only
#' covers features present in `X_ref`.
#'
#' Duplicate detection compares rows of `X_ref` using the chosen `sim_method`
#' (cosine on L2-normalized rows, or Pearson via row-centering), optionally after
#' rank transformation (`feature_space = "rank"`). By default,
#' `duplicate_scope = "train_test"` filters to pairs that appear in train vs test
#' in at least one repeat; set `duplicate_scope = "all"` to include within-fold
#' duplicates. The `duplicates` slot returns index pairs and similarity values
#' for near-duplicate samples. Only duplicates present in `X_ref` can be
#' detected, and checks are skipped if inputs cannot be aligned to splits.
#' @examples
#' set.seed(1)
#' df <- data.frame(
#'   subject = rep(1:6, each = 2),
#'   outcome = rbinom(12, 1, 0.5),
#'   x1 = rnorm(12),
#'   x2 = rnorm(12)
#' )
#'
#' splits <- make_split_plan(df, outcome = "outcome",
#'                       mode = "subject_grouped", group = "subject", v = 3,
#'                       progress = FALSE)
#'
#' custom <- list(
#'   glm = list(
#'     fit = function(x, y, task, weights, ...) {
#'       stats::glm(y ~ ., data = as.data.frame(x),
#'                  family = stats::binomial(), weights = weights)
#'     },
#'     predict = function(object, newdata, task, ...) {
#'       as.numeric(stats::predict(object,
#'                                 newdata = as.data.frame(newdata),
#'                                 type = "response"))
#'     }
#'   )
#' )
#'
#' fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                     learner = "glm", custom_learners = custom,
#'                     metrics = "auc", refit = FALSE, seed = 1)
#'
#' audit <- audit_leakage(fit, metric = "auc", B = 10,
#'                        X_ref = df[, c("x1", "x2")])
#'
#' @export
audit_leakage <- function(fit,
                          metric = c("auc", "pr_auc", "accuracy", "macro_f1", "log_loss", "rmse", "cindex"),
                          B = 200,
                          perm_stratify = FALSE,
                          perm_refit = "auto",
                          perm_refit_auto_max = 200,
                          perm_refit_spec = NULL,
                          perm_mode = NULL,
                          time_block = c("circular", "stationary"),
                          block_len = NULL,
                          include_z = TRUE,
                          ci_method = c("if", "bootstrap"),
                          boot_B = 400,
                          parallel = FALSE,
                          seed = 1,
                          return_perm = TRUE,
                          batch_cols = NULL,
                          coldata = NULL,
                          X_ref = NULL,
                          target_scan = TRUE,
                          target_scan_multivariate = TRUE,
                          target_scan_multivariate_B = 100,
                          target_scan_multivariate_components = 10,
                          target_scan_multivariate_interactions = TRUE,
                          target_threshold = 0.9,
                          target_p_adjust = c("none", "BH", "BY", "holm", "bonferroni"),
                          target_alpha = 0.05,
                          feature_space = c("raw", "rank"),
                          sim_method = c("cosine", "pearson"),
                          sim_threshold = 0.995,
                          nn_k = 50,
                          max_pairs = 5000,
                          duplicate_scope = c("train_test", "all"),
                          learner = NULL,
                          strict_align = FALSE) {

  # --- CRITICAL PATCH: Support LeakTune objects (via tune_resample) ---
  if (inherits(fit, "LeakTune")) {
    # If using 'tune_resample', we construct a proxy LeakFit object
    # by aggregating predictions from the outer resampling folds.

    # 1. Collect all predictions
    all_preds_list <- list()
    first_fit <- NULL
    for (i in seq_along(fit$outer_fits)) {
      of <- fit$outer_fits[[i]]
      if (is.null(of) || !methods::is(of, "LeakFit")) next
      if (is.null(first_fit)) first_fit <- of
      # LeakFit predictions are lists of dataframes
      if (length(of@predictions) > 0) {
        # Combine inner predictions (usually just one DF for outer fit)
        p_df <- do.call(rbind, lapply(of@predictions, as.data.frame))

        # CRITICAL FIX 1: Overwrite 'fold' index to match the outer loop index 'i'.
        # Individual outer_fits are trained as single-fold objects (fold=1),
        # so we must re-index them to match the original splits (1..5).
        if (nrow(p_df) > 0) {
          p_df$fold <- i
        }

        all_preds_list[[i]] <- p_df
      }
    }

    if (length(all_preds_list) == 0) {
      stop("LeakTune object contains no outer loop predictions to audit.")
    }

    # 2. Reconstruct a minimal LeakFit for auditing
    if (is.null(first_fit)) {
      stop("LeakTune object contains no valid outer fit objects to audit.")
    }

    # CRITICAL FIX 2: Safely extract feature names to satisfy S4 validation
    # (must be character, not NULL)
    fnames <- tryCatch(first_fit@feature_names, error = function(e) character(0))
    if (is.null(fnames)) fnames <- character(0)

    # Create the proxy S4 object using new() to bypass validation if needed
    fit <- new("LeakFit",
               splits = fit$splits,
               metrics = fit$metrics,
               metric_summary = fit$metric_summary,
               audit = data.frame(),
               predictions = all_preds_list,
               preprocess = list(),
               learners = list(),
               outcome = fit$info$outcome %||% first_fit@outcome,
               task = fit$info$task %||% first_fit@task,
               feature_names = fnames,
               info = list(
                 hash = fit$splits@info$hash,
                 metrics_requested = fit$info$metrics_requested,
                 metrics_used = fit$info$metrics_used,
                 class_weights = NULL,
                 positive_class = fit$info$positive_class,
                 # CRITICAL FIX 3: Use sample_ids from the first outer fit to enable valid permutation
                 sample_ids = first_fit@info$sample_ids,
                 fold_seeds = NULL,
                 refit = FALSE,
                 final_model = NULL,
                 final_preprocess = NULL,
                 learner_names = unique(fit$metrics$learner),
                 perm_refit_spec = NULL
               )
    )
  }
  # --- END PATCH ---

  metric <- match.arg(metric)
  feature_space <- match.arg(feature_space)
  sim_method    <- match.arg(sim_method)
  duplicate_scope <- match.arg(duplicate_scope)
  time_block <- match.arg(time_block)
  ci_method <- match.arg(ci_method)
  target_p_adjust <- match.arg(target_p_adjust)

  if (!is.numeric(target_threshold) || length(target_threshold) != 1L ||
      !is.finite(target_threshold) || target_threshold <= 0 || target_threshold >= 1) {
    stop("target_threshold must be a single numeric value in (0,1).")
  }
  if (!is.numeric(target_alpha) || length(target_alpha) != 1L ||
      !is.finite(target_alpha) || target_alpha <= 0 || target_alpha >= 1) {
    stop("target_alpha must be a single numeric value in (0,1).")
  }
  if (!is.logical(target_scan_multivariate) || length(target_scan_multivariate) != 1L) {
    stop("target_scan_multivariate must be TRUE or FALSE.")
  }
  target_scan_multivariate <- isTRUE(target_scan_multivariate)
  target_scan_multivariate_B <- as.integer(target_scan_multivariate_B)
  if (isTRUE(target_scan_multivariate)) {
    if (!is.finite(target_scan_multivariate_B) || target_scan_multivariate_B < 1L) {
      stop("target_scan_multivariate_B must be an integer >= 1.")
    }
    target_scan_multivariate_components <- as.integer(target_scan_multivariate_components)
    if (!is.finite(target_scan_multivariate_components) ||
        target_scan_multivariate_components < 1L) {
      stop("target_scan_multivariate_components must be an integer >= 1.")
    }
    if (!is.logical(target_scan_multivariate_interactions) ||
        length(target_scan_multivariate_interactions) != 1L) {
      stop("target_scan_multivariate_interactions must be TRUE or FALSE.")
    }
  }

  set.seed(seed)
  if (is.null(perm_refit_spec) && !is.null(fit@info$perm_refit_spec)) {
    perm_refit_spec <- fit@info$perm_refit_spec
  }


  select_refit_learner <- function(spec_learner, learner, learner_names = NULL) {
    if (is.null(spec_learner) || is.null(learner)) return(spec_learner)
    if (inherits(spec_learner, "workflow") || inherits(spec_learner, "model_spec")) {
      return(spec_learner)
    }
    if (is.list(spec_learner)) {
      nm <- names(spec_learner)
      if (!is.null(nm) && learner %in% nm) {
        return(spec_learner[[learner]])
      }
      if (!is.null(learner_names)) {
        idx <- match(learner, learner_names)
        if (is.finite(idx) && idx >= 1L && idx <= length(spec_learner)) {
          return(spec_learner[[idx]])
        }
      }
    }
    if (is.character(spec_learner) && length(spec_learner) > 1L) {
      nm <- names(spec_learner)
      if (!is.null(nm) && learner %in% nm) {
        return(spec_learner[[learner]])
      }
      if (!is.null(learner_names)) {
        idx <- match(learner, learner_names)
        if (is.finite(idx) && idx >= 1L && idx <= length(spec_learner)) {
          return(spec_learner[[idx]])
        }
      }
    }
    spec_learner
  }

  if (!is.null(perm_refit_spec) && is.list(perm_refit_spec) && !is.null(learner)) {
    perm_refit_spec$learner <- select_refit_learner(
      perm_refit_spec$learner %||% NULL,
      learner,
      fit@info$learner_names %||% NULL
    )
  }
  perm_refit_mode <- "fixed"
  perm_refit_reason <- NULL
  perm_refit_raw <- perm_refit
  if (is.character(perm_refit_raw)) {
    if (!identical(perm_refit_raw, "auto")) {
      stop("perm_refit must be TRUE, FALSE, or \"auto\".", call. = FALSE)
    }
    perm_refit_mode <- "auto-fixed"
    if (!is.numeric(perm_refit_auto_max) || length(perm_refit_auto_max) != 1L ||
        !is.finite(perm_refit_auto_max) || perm_refit_auto_max < 1L) {
      stop("perm_refit_auto_max must be a single integer >= 1.", call. = FALSE)
    }
    if (is.null(perm_refit_spec) || !is.list(perm_refit_spec)) {
      perm_refit_reason <- "perm_refit_spec missing; using fixed predictions."
      message("audit_leakage(): perm_refit_spec not supplied; using fixed-prediction ",
              "permutations (perm_refit = FALSE). This test can be slightly liberal ",
              "(type I error above nominal). For tighter calibration, supply ",
              "perm_refit_spec to enable refit-based permutations.")
      perm_refit <- FALSE
    } else if (!is.finite(B) || B > perm_refit_auto_max) {
      perm_refit_reason <- sprintf("B=%s exceeds perm_refit_auto_max=%s; using fixed predictions.",
                                   as.character(B), as.character(perm_refit_auto_max))
      perm_refit <- FALSE
    } else {
      perm_refit <- TRUE
      perm_refit_mode <- "auto-refit"
    }
  } else if (isTRUE(perm_refit_raw)) {
    perm_refit <- TRUE
    perm_refit_mode <- "refit"
  } else if (isFALSE(perm_refit_raw)) {
    perm_refit <- FALSE
    perm_refit_mode <- "fixed"
  } else {
    stop("perm_refit must be TRUE, FALSE, or \"auto\".", call. = FALSE)
  }
  perm_method <- if (isTRUE(perm_refit)) "refit" else "fixed"

  refit_x <- NULL
  refit_outcome <- NULL
  refit_learner <- NULL
  refit_preprocess <- NULL
  refit_learner_args <- list()
  refit_custom_learners <- list()
  refit_class_weights <- NULL
  refit_positive_class <- NULL
  refit_parallel <- FALSE
  refit_coldata <- NULL
  refit_coldata_supplied <- FALSE
  default_preprocess <- list(
    impute = list(method = "median"),
    normalize = list(method = "zscore"),
    filter = list(var_thresh = 0, iqr_thresh = 0),
    fs = list(method = "none")
  )
  if (perm_refit) {
    if (is.null(perm_refit_spec) || !is.list(perm_refit_spec)) {
      stop("perm_refit=TRUE requires perm_refit_spec as a list with x and learner.")
    }
    refit_x <- perm_refit_spec$x %||% NULL
    if (is.null(refit_x)) {
      stop("perm_refit_spec$x is required when perm_refit=TRUE.")
    }
    refit_outcome <- perm_refit_spec$outcome %||% fit@outcome
    if (is.null(refit_outcome) || !length(refit_outcome)) {
      stop("perm_refit_spec$outcome (or fit@outcome) is required when perm_refit=TRUE.")
    }
    refit_learner <- perm_refit_spec$learner %||% NULL
    if (is.null(refit_learner)) {
      stop("perm_refit_spec$learner is required when perm_refit=TRUE.")
    }
    if (!inherits(refit_learner, c("workflow", "model_spec")) &&
        ((is.list(refit_learner) && length(refit_learner) != 1L) ||
         (is.character(refit_learner) && length(refit_learner) != 1L))) {
      stop("perm_refit requires a single learner; supply learner= or a single perm_refit_spec$learner.",
           call. = FALSE)
    }
    refit_preprocess <- perm_refit_spec$preprocess %||% default_preprocess
    refit_learner_args <- perm_refit_spec$learner_args %||% list()
    refit_custom_learners <- perm_refit_spec$custom_learners %||% list()
    refit_class_weights <- perm_refit_spec$class_weights %||% fit@info$class_weights %||% NULL
    refit_positive_class <- perm_refit_spec$positive_class %||% fit@info$positive_class %||% NULL
    refit_parallel <- isTRUE(perm_refit_spec$parallel)
    refit_coldata_supplied <- !is.null(perm_refit_spec$coldata)
    refit_coldata <- perm_refit_spec$coldata %||% NULL
  }

  # Trail / provenance
  trail <- list(
    indices_hash = .bio_hash_indices(fit@splits@indices),
    mode = fit@splits@mode,
    info = fit@splits@info,
    seed = seed
  )
  trail$metric <- metric
  trail$perm_method <- perm_method
  if (!is.null(fit@info$learner)) trail$learner <- fit@info$learner

  # --- Reconstruct main metric from CV predictions --------------------------
  task <- fit@task
  valid_metrics <- switch(task,
                          binomial = c("auc", "pr_auc", "accuracy", "log_loss"),
                          multiclass = c("accuracy", "macro_f1", "log_loss"),
                          gaussian = c("rmse", "cindex"),
                          survival = c("cindex"),
                          c("auc", "pr_auc", "rmse", "cindex"))
  if (!metric %in% valid_metrics) {
    stop(sprintf("Metric '%s' is not supported for %s tasks.", metric, task))
  }
  folds <- fit@splits@indices
  compact <- isTRUE(fit@splits@info$compact)
  fold_assignments <- fit@splits@info$fold_assignments
  perm_mode_use <- .bio_perm_mode(fit@splits)
  if (!is.null(perm_mode)) {
    valid_modes <- c("subject_grouped", "batch_blocked", "study_loocv", "time_series")
    perm_mode <- as.character(perm_mode)
    perm_mode <- perm_mode[!is.na(perm_mode) & nzchar(perm_mode)]
    if (length(perm_mode) != 1L || !perm_mode %in% valid_modes) {
      stop("perm_mode must be one of: subject_grouped, batch_blocked, study_loocv, time_series.",
           call. = FALSE)
    }
    perm_mode_use <- perm_mode
  }
  if (identical(fit@splits@mode, "rsample") && identical(perm_mode_use, "rsample")) {
    stop(paste0("rsample splits require an explicit perm_mode; pass perm_mode= to ",
                "audit_leakage() or set split_cols/attr(splits, 'bioLeak_perm_mode') when fitting."),
         call. = FALSE)
  }
  trail$perm_mode <- perm_mode_use
  resolve_test_idx <- function(fold) {
    if (!isTRUE(compact) && !is.null(fold$test)) return(fold$test)
    if (is.null(fold_assignments) || !length(fold_assignments)) {
      stop("Compact splits require fold assignments to resolve test indices.")
    }
    r <- fold$repeat_id
    if (is.null(r) || !is.finite(r)) r <- 1L
    assign_vec <- fold_assignments[[r]]
    if (is.null(assign_vec)) {
      stop(sprintf("Missing fold assignments for repeat %s.", r))
    }
    which(assign_vec == fold$fold)
  }
  if (is.null(coldata) && !is.null(fit@splits@info$coldata)) {
    coldata <- fit@splits@info$coldata
  }
  outcome_col <- fit@splits@info$outcome
  if (is.null(outcome_col)) outcome_col <- fit@outcome
  if (length(outcome_col) != 1L || identical(task, "survival")) outcome_col <- NULL
  if (isTRUE(target_scan) && identical(task, "survival")) {
    warning("Target leakage scan is not available for survival outcomes.", call. = FALSE)
    target_scan <- FALSE
  }
  if (isTRUE(target_scan_multivariate) && !task %in% c("binomial", "gaussian")) {
    warning("Multivariate target scan is only available for binomial or gaussian tasks.",
            call. = FALSE)
    target_scan_multivariate <- FALSE
  }

  pred_list_raw <- lapply(fit@predictions, function(df) data.frame(df, stringsAsFactors = FALSE))
  pred_df <- if (length(pred_list_raw)) do.call(rbind, pred_list_raw) else NULL
  if (is.null(pred_df) || !nrow(pred_df)) {
    stop("No predictions available in LeakFit object.")
  }

  has_learner <- "learner" %in% names(pred_df)
  if (has_learner) {
    pred_df$learner <- as.character(pred_df$learner)
    learner_vals <- unique(pred_df$learner)
    if (is.null(learner)) {
      if (length(learner_vals) == 1L) {
        learner <- learner_vals[[1]]
      } else {
        stop("Multiple learners found in predictions; specify `learner` to audit a single model.")
      }
    } else {
      if (length(learner) != 1L) stop("learner must be a single value.")
      if (!learner %in% learner_vals) {
        stop(sprintf("Learner '%s' not found in predictions. Available: %s",
                     learner, paste(learner_vals, collapse = ", ")))
      }
    }
    pred_df <- pred_df[pred_df$learner == learner, , drop = FALSE]
    if (!nrow(pred_df)) {
      stop(sprintf("No predictions available for learner '%s'.", learner))
    }
  } else {
    if (!is.null(learner)) {
      warning("`learner` ignored: predictions do not include learner IDs.")
    } else if (!is.null(fit@metrics) && length(unique(fit@metrics$learner)) > 1L) {
      warning("Multiple learners were fit but predictions lack learner IDs; audit may mix learners. Refit with updated bioLeak.")
    }
  }
  if (!is.null(learner)) {
    trail$learner <- learner
  }

  resolve_sample_ids <- function(fit_obj, fallback_n = NULL) {
    ids <- fit_obj@info$sample_ids %||% NULL
    if (!is.null(ids) && length(ids)) return(as.character(ids))
    cd_lookup <- fit_obj@splits@info$coldata %||% NULL
    if (!is.null(cd_lookup)) {
      rn_cd <- rownames(cd_lookup)
      if (!is.null(rn_cd) && !anyNA(rn_cd) && all(nzchar(rn_cd)) && !anyDuplicated(rn_cd)) {
        return(as.character(rn_cd))
      }
      if ("row_id" %in% names(cd_lookup)) {
        rid <- as.character(cd_lookup[["row_id"]])
        if (length(rid) == nrow(cd_lookup) && !anyNA(rid) &&
            !anyDuplicated(rid) && all(nzchar(rid))) {
          return(rid)
        }
      }
    }
    if (!is.null(fallback_n) && is.finite(fallback_n)) {
      return(as.character(seq_len(fallback_n)))
    }
    NULL
  }

  align_coldata_for_perm <- function(cd, sample_ids, context, strict_align = FALSE) {
    if (is.null(cd)) return(NULL)
    if (is.null(sample_ids) || !length(sample_ids)) {
      warning(sprintf("%s coldata alignment failed (missing sample ids); restricted permutations disabled.", context),
              call. = FALSE)
      return(NULL)
    }
    cd <- as.data.frame(cd, check.names = FALSE)
    sample_ids <- as.character(sample_ids)
    rn <- rownames(cd)
    if (!is.null(rn) && length(rn)) {
      if (anyDuplicated(rn)) {
        warning(sprintf("%s coldata rownames are duplicated; restricted permutations disabled.", context),
                call. = FALSE)
        return(NULL)
      }
      if (all(sample_ids %in% rn)) {
        return(cd[match(sample_ids, rn), , drop = FALSE])
      }
      warning(sprintf("%s coldata rownames do not match sample ids; restricted permutations disabled.",
                      context), call. = FALSE)
      return(NULL)
    }
    if ("row_id" %in% names(cd)) {
      rid <- as.character(cd[["row_id"]])
      if (anyDuplicated(rid)) {
        warning(sprintf("%s coldata row_id values are duplicated; restricted permutations disabled.",
                        context), call. = FALSE)
        return(NULL)
      }
      if (all(sample_ids %in% rid)) {
        return(cd[match(sample_ids, rid), , drop = FALSE])
      }
      warning(sprintf("%s coldata row_id values do not match sample ids; restricted permutations disabled.",
                      context), call. = FALSE)
      return(NULL)
    }
    if (nrow(cd) == length(sample_ids)) {
      if (isTRUE(strict_align)) {
        stop(sprintf("%s coldata has no ids and strict_align = TRUE; provide matching row names or a 'row_id' column.",
                     context), call. = FALSE)
      }
      warning(sprintf("%s coldata has no ids; assuming row order aligns to splits for permutations.",
                      context), call. = FALSE)
      return(cd)
    }
    warning(sprintf("%s coldata not aligned to splits; restricted permutations disabled.", context),
            call. = FALSE)
    NULL
  }

  if ("fold" %in% names(pred_df)) {
    pred_list <- lapply(seq_along(folds), function(i) {
      pred_df[pred_df$fold == i, , drop = FALSE]
    })
  } else if (has_learner) {
    pred_list <- lapply(pred_list_raw, function(df) {
      if ("learner" %in% names(df)) df[df$learner == learner, , drop = FALSE] else df
    })
  } else {
    pred_list <- pred_list_raw
  }

  # Filter out empty (zero-row) folds to keep pred_list aligned with folds
  non_empty <- vapply(pred_list, function(df) nrow(df) > 0L, logical(1))
  if (!all(non_empty)) {
    pred_list <- pred_list[non_empty]
    folds <- folds[non_empty]
  }

  all_pred <- do.call(rbind, pred_list)

  metric_obs <- .metric_value(metric, task, all_pred$truth, all_pred$pred, pred_df = all_pred)

  delta <- NA_real_
  perm_mean <- NA_real_
  perm_sd <- NA_real_
  pval <- NA_real_
  p_se <- NA_real_

  if (is.na(metric_obs) || !is.finite(metric_obs)) {
    warning("Observed metric is NA or non-finite; skipping permutation gap calculation.")
    metric_obs <- NA_real_
  }

  # Weights per fold for IF aggregation
  weights <- vapply(folds, function(f) length(resolve_test_idx(f)), integer(1))
  weights <- weights / sum(weights)

  se_obs <- NA_real_
  if (metric == "auc" && exists(".cvauc_if", mode = "function")) {
    obs_truth <- lapply(pred_list, `[[`, "truth")
    obs_pred <- lapply(pred_list, `[[`, "pred")
    if_stat <- try(.cvauc_if(pred = obs_pred, truth = obs_truth, weights = weights), silent = TRUE)
    if (!inherits(if_stat, "try-error") && !is.null(if_stat$se_auc)) {
      se_obs <- if_stat$se_auc
    }
  }

  higher_better <- metric %in% c("auc", "pr_auc", "cindex", "accuracy", "macro_f1")

  # --- Permutations ----------------------------------------------------------
  perm_vals <- numeric(0)
  if (perm_refit) {
    if (identical(task, "survival")) {
      stop("perm_refit=TRUE is not supported for survival tasks.", call. = FALSE)
    }
    if (length(refit_outcome) != 1L) {
      stop("perm_refit=TRUE requires a single outcome column name.", call. = FALSE)
    }

    refit_x_mat <- .bio_get_x(refit_x)
    n_refit <- nrow(refit_x_mat)
    if (!is.finite(n_refit) || n_refit < 1L) {
      stop("perm_refit_spec$x has no rows.", call. = FALSE)
    }
    if (isTRUE(compact) && length(fold_assignments)) {
      expected_n <- max(vapply(fold_assignments, length, integer(1)), na.rm = TRUE)
      if (is.finite(expected_n) && n_refit != expected_n) {
        stop("perm_refit_spec$x row count does not match compact split assignments.", call. = FALSE)
      }
    } else {
      max_idx <- suppressWarnings(max(vapply(folds, function(z) {
        idx <- c(z$train, z$test)
        if (length(idx)) max(idx) else 0L
      }, integer(1)), na.rm = TRUE))
      if (is.finite(max_idx) && n_refit < max_idx) {
        stop("perm_refit_spec$x has fewer rows than required by splits.", call. = FALSE)
      }
    }

    if (is.null(refit_coldata)) {
      if (.bio_is_se(refit_x)) {
        refit_coldata <- .bio_coldata_df(refit_x)
      } else if (is.data.frame(refit_x)) {
        refit_coldata <- refit_x
      }
    } else if (isTRUE(refit_coldata_supplied)) {
      refit_ids <- NULL
      if (.bio_is_se(refit_x)) {
        refit_ids <- rownames(SummarizedExperiment::colData(refit_x))
      } else if (is.data.frame(refit_x)) {
        refit_ids <- rownames(refit_x)
        if ((is.null(refit_ids) || !all(nzchar(refit_ids))) &&
            "row_id" %in% names(refit_x)) {
          refit_ids <- as.character(refit_x[["row_id"]])
        }
      } else if (is.matrix(refit_x)) {
        refit_ids <- rownames(refit_x)
      }
      if (!is.null(refit_ids) && all(nzchar(refit_ids))) {
        refit_coldata <- align_coldata_for_perm(refit_coldata, refit_ids,
                                                context = "perm_refit",
                                                strict_align = strict_align)
      } else if (is.null(refit_ids) && nrow(refit_coldata) == n_refit) {
        if (isTRUE(strict_align)) {
          stop("perm_refit coldata has no ids and strict_align = TRUE; provide matching row names or a 'row_id' column.",
               call. = FALSE)
        }
        warning("perm_refit coldata has no ids; assuming row order aligns to refit data.",
                call. = FALSE)
      } else if (is.null(refit_ids)) {
        warning("perm_refit coldata not aligned to refit data; restricted permutations disabled.",
                call. = FALSE)
        refit_coldata <- NULL
      }
    }

    y_base <- NULL
    if (!is.null(refit_coldata) && refit_outcome %in% names(refit_coldata)) {
      y_base <- refit_coldata[[refit_outcome]]
    } else {
      y_base <- .bio_get_y(refit_x, refit_outcome)
    }
    if (length(y_base) != n_refit) {
      stop("Outcome length does not match rows in perm_refit_spec$x.", call. = FALSE)
    }

    perm_full_source <- NULL
    if (!is.null(refit_coldata) &&
        refit_outcome %in% names(refit_coldata) &&
        perm_mode_use %in% c("subject_grouped", "batch_blocked", "study_loocv", "time_series")) {
      folds_perm <- list(list(test = seq_len(n_refit), fold = 1L, repeat_id = 1L))
      perm_full_source <- .permute_labels_factory(
        cd = refit_coldata, outcome = refit_outcome, mode = perm_mode_use,
        folds = folds_perm, perm_stratify = perm_stratify, time_block = time_block,
        block_len = block_len, seed = seed,
        group_col = fit@splits@info$group, batch_col = fit@splits@info$batch,
        study_col = fit@splits@info$study, time_col = fit@splits@info$time,
        perm_refit = TRUE
      )
    }

    permute_outcome <- function(b) {
      if (!is.null(perm_full_source)) {
        perm_full_source(b)[[1]]
      } else {
        sample(y_base)
      }
    }

    apply_outcome <- function(x, outcome, y_perm) {
      y_perm <- .coerce_truth_like(y_base, y_perm)
      if (.bio_is_se(x)) {
        cd <- SummarizedExperiment::colData(x)
        if (!outcome %in% colnames(cd)) {
          stop("Outcome column not found in SummarizedExperiment colData.", call. = FALSE)
        }
        cd[[outcome]] <- y_perm
        SummarizedExperiment::colData(x) <- cd
        return(x)
      }
      if (is.data.frame(x)) {
        if (!outcome %in% names(x)) {
          stop("Outcome column not found in data.frame.", call. = FALSE)
        }
        x[[outcome]] <- y_perm
        return(x)
      }
      if (is.matrix(x)) {
        if (is.character(outcome) && outcome %in% colnames(x)) {
          x[, outcome] <- y_perm
          return(x)
        }
        stop("Outcome column not found in matrix input.", call. = FALSE)
      }
      stop("perm_refit_spec$x must be a SummarizedExperiment, data.frame, or matrix.", call. = FALSE)
    }

    perm_eval_refit <- function(b) {
      set.seed(seed + b)
      y_perm <- permute_outcome(b)
      x_perm <- apply_outcome(refit_x, refit_outcome, y_perm)
      fit_perm <- try(fit_resample(
        x_perm,
        outcome = refit_outcome,
        splits = fit@splits,
        preprocess = refit_preprocess,
        learner = refit_learner,
        learner_args = refit_learner_args,
        custom_learners = refit_custom_learners,
        metrics = metric,
        class_weights = refit_class_weights,
        positive_class = refit_positive_class,
        parallel = refit_parallel,
        refit = FALSE,
        seed = seed + b
      ), silent = TRUE)
      if (inherits(fit_perm, "try-error")) {
        warning(sprintf("Permutation refit %d failed: %s", b, attr(fit_perm, "condition")$message),
                call. = FALSE)
        return(NA_real_)
      }
      perm_pred <- if (length(fit_perm@predictions)) {
        do.call(rbind, lapply(fit_perm@predictions, function(df) data.frame(df, stringsAsFactors = FALSE)))
      } else {
        NULL
      }
      if (is.null(perm_pred) || !nrow(perm_pred)) return(NA_real_)
      .metric_value(metric, task, perm_pred$truth, perm_pred$pred, pred_df = perm_pred)
    }

    if (parallel && requireNamespace("future.apply", quietly = TRUE)) {
      perm_vals <- future.apply::future_sapply(seq_len(B), perm_eval_refit, future.seed = TRUE)
    } else {
      perm_vals <- sapply(seq_len(B), perm_eval_refit)
    }
  } else {
    perm_source <- NULL
    perm_coldata <- NULL
    if (!is.null(coldata)) {
      sample_ids <- resolve_sample_ids(fit, fallback_n = nrow(fit@splits@info$coldata %||% data.frame()))
      perm_coldata <- align_coldata_for_perm(coldata, sample_ids, context = "Permutation",
                                              strict_align = strict_align)
    }
    if (!is.null(perm_coldata) && !is.null(outcome_col) && outcome_col %in% names(perm_coldata)) {
      folds_perm <- folds
      if (isTRUE(compact)) {
        attr(folds_perm, "fold_assignments") <- fold_assignments
      }
      perm_source <- .permute_labels_factory(
        cd = perm_coldata, outcome = outcome_col, mode = perm_mode_use,
        folds = folds_perm, perm_stratify = perm_stratify, time_block = time_block,
        block_len = block_len, seed = seed,
        group_col = fit@splits@info$group, batch_col = fit@splits@info$batch,
        study_col = fit@splits@info$study, time_col = fit@splits@info$time,
        perm_refit = FALSE
      )
    }
    if (is.null(perm_source)) {
      perm_source <- function(b) {
        lapply(pred_list, function(df) {
          if (identical(task, "survival")) {
            if (!requireNamespace("survival", quietly = TRUE)) {
              stop("Package 'survival' is required for survival permutations.")
            }
            if (all(c("truth_time", "truth_event") %in% names(df))) {
              idx <- sample(seq_len(nrow(df)))
              return(survival::Surv(df$truth_time[idx], df$truth_event[idx]))
            }
          }
          if (length(df$truth) <= 1L) df$truth else df$truth[sample.int(length(df$truth))]
        })
      }
    }

    # Global label shuffle: concatenate all fold predictions and permute
    # labels across the full dataset.  Per-fold shuffling would preserve
    # within-fold class balance, inflating the null when fold structure
    # correlates with the outcome (e.g., study_loocv with unequal class
    # proportions across studies).  A global shuffle gives the correct
    # null: E[AUC] = 0.5, testing whether predictions are associated with
    # labels without conditioning on fold membership.
    agg_base <- do.call(rbind, pred_list)
    perm_eval <- function(b) {
      n <- nrow(agg_base)
      if (n <= 1L) return(NA_real_)
      agg <- agg_base
      perm_idx <- sample.int(n)
      if (identical(task, "survival") &&
          all(c("truth_time", "truth_event") %in% names(agg))) {
        agg$truth_time <- agg$truth_time[perm_idx]
        agg$truth_event <- agg$truth_event[perm_idx]
      } else {
        agg$truth <- agg$truth[perm_idx]
      }
      .metric_value(metric, task, agg$truth, agg$pred, pred_df = agg)
    }

    if (parallel && requireNamespace("future.apply", quietly = TRUE)) {
      perm_vals <- future.apply::future_sapply(seq_len(B), function(i) {
        set.seed(seed + i); perm_eval(i)
      }, future.seed = TRUE)
    } else {
      perm_vals <- sapply(seq_len(B), function(i) { set.seed(seed + i); perm_eval(i) })
    }
  }

  if (!is.na(metric_obs) && is.finite(metric_obs)) {
    if (!any(is.finite(perm_vals))) {
      perm_mean <- NA_real_
      perm_sd <- NA_real_
      pval <- NA_real_
      p_se <- NA_real_
      delta <- NA_real_
    } else {
      perm_mean <- mean(perm_vals, na.rm = TRUE)
      perm_sd <- stats::sd(perm_vals, na.rm = TRUE)
      finite_perm <- is.finite(perm_vals)
      pval <- if (higher_better) {
        (1 + sum(perm_vals[finite_perm] >= metric_obs, na.rm = TRUE)) / (1 + sum(finite_perm))
      } else {
        (1 + sum(perm_vals[finite_perm] <= metric_obs, na.rm = TRUE)) / (1 + sum(finite_perm))
      }
      p_se <- sqrt(pval * (1 - pval) / (sum(finite_perm) + 1))
      delta <- if (higher_better) metric_obs - perm_mean else perm_mean - metric_obs
    }
  }

  seci <- list(se = NA_real_, ci = c(NA_real_, NA_real_), z = NA_real_)
  if (ci_method == "if" && exists(".se_ci_delta", mode = "function") && is.finite(se_obs) && is.finite(delta)) {
    seci_try <- try(.se_ci_delta(delta, se_obs, perm_vals), silent = TRUE)
    if (!inherits(seci_try, "try-error")) {
      seci <- seci_try
    } else {
      seci <- list(se = NA_real_, ci = c(NA_real_, NA_real_), z = NA_real_)
    }
  } else if (ci_method == "bootstrap" && any(is.finite(perm_vals)) && is.finite(metric_obs)) {
    perm_vals_finite <- perm_vals[is.finite(perm_vals)]
    boot_vals <- replicate(boot_B, {
      sample_vals <- sample(perm_vals_finite, replace = TRUE)
      if (higher_better) metric_obs - mean(sample_vals) else mean(sample_vals) - metric_obs
    })
    alpha <- 0.025
    se_boot <- stats::sd(boot_vals)
    seci <- list(
      se = se_boot,
      ci = as.numeric(stats::quantile(boot_vals, probs = c(alpha, 1 - alpha))),
      z = ifelse(se_boot > 0, delta / se_boot, NA_real_)
    )
  }

  z_val <- if (isTRUE(include_z)) seci$z else NA_real_

  perm_df <- data.frame(
    metric_obs = metric_obs,
    perm_mean  = perm_mean,
    perm_sd    = perm_sd,
    gap        = delta,
    z          = z_val,
    p_value    = pval,
    n_perm     = length(perm_vals)
  )
  perm_df$mechanism_class <- "non_random_signal"
  perm_df <- perm_df[, c("mechanism_class", setdiff(names(perm_df), "mechanism_class")), drop = FALSE]

  perm_df[] <- lapply(perm_df, function(x) {
    if (!is.numeric(x)) return(x)
    signif(x, 6)
  })

  # --- Batch / study association with folds ---------------------------------
  ids_pred <- all_pred$id
  ids_pred_chr <- as.character(ids_pred)
  ids_all <- unique(ids_pred)
  ids_all_chr <- as.character(ids_all)
  fold_id <- if ("fold" %in% names(all_pred)) all_pred$fold else NULL
  repeat_vals <- if (length(fit@splits@indices)) {
    vapply(fit@splits@indices, function(z) as.integer(z$repeat_id %||% 1L), integer(1))
  } else {
    1L
  }
  repeat_vals <- repeat_vals[is.finite(repeat_vals)]
  has_repeats <- length(unique(repeat_vals)) > 1L
  skip_batch_assoc <- FALSE
  if ((is.null(fold_id) || !length(fold_id) || all(is.na(fold_id))) && isTRUE(has_repeats)) {
    warning("Predictions lack fold identifiers for repeated CV; batch association skipped to avoid pooling repeats.",
            call. = FALSE)
    skip_batch_assoc <- TRUE
  }

  sample_ids <- fit@info$sample_ids %||% NULL
  if (is.null(sample_ids) || !length(sample_ids)) {
    if (!is.null(fit@splits@info$coldata)) {
      cd_lookup <- fit@splits@info$coldata
      rn_cd <- rownames(cd_lookup)
      if (!is.null(rn_cd) && !anyNA(rn_cd) && all(nzchar(rn_cd)) && !anyDuplicated(rn_cd)) {
        sample_ids <- rn_cd
      } else if ("row_id" %in% names(cd_lookup)) {
        rid <- as.character(cd_lookup[["row_id"]])
        if (length(rid) == nrow(cd_lookup) && !anyNA(rid) &&
            !anyDuplicated(rid) && all(nzchar(rid))) {
          sample_ids <- rid
        }
      }
    }
  }
  if (!skip_batch_assoc && (is.null(fold_id) || !length(fold_id) || all(is.na(fold_id)))) {
    fold_id_map <- rep(NA_integer_, length(ids_all_chr))
    names(fold_id_map) <- ids_all_chr
    if (!is.null(sample_ids) && length(sample_ids)) {
      sample_ids <- as.character(sample_ids)
      max_idx <- max(vapply(fit@splits@indices, function(z) {
        te_idx <- resolve_test_idx(z)
        if (length(te_idx)) max(te_idx) else 0L
      }, integer(1)), na.rm = TRUE)
      if (length(sample_ids) < max_idx) {
        warning("Sample ID mapping is shorter than split indices; fold mapping may be incomplete.",
                call. = FALSE)
      } else {
        for (i in seq_along(fit@splits@indices)) {
          te_idx <- resolve_test_idx(fit@splits@indices[[i]])
          te_ids <- sample_ids[te_idx]
          fold_id_map[as.character(te_ids)] <- i
        }
      }
    } else if (is.numeric(ids_all)) {
      for (i in seq_along(fit@splits@indices)) {
        te_idx <- resolve_test_idx(fit@splits@indices[[i]])
        te_ids <- intersect(te_idx, ids_all)
        fold_id_map[as.character(te_ids)] <- i
      }
    }
    fold_id <- fold_id_map[ids_pred_chr]
  }

  fold_label <- fold_id
  repeat_id <- NULL
  if (!skip_batch_assoc && !is.null(fold_id) && length(fold_id) && length(fit@splits@indices)) {
    fold_meta <- data.frame(
      fold_seq = seq_along(fit@splits@indices),
      fold = vapply(fit@splits@indices, function(z) {
        as.integer(z$fold %||% NA_integer_)
      }, integer(1)),
      repeat_id = vapply(fit@splits@indices, function(z) {
        as.integer(z$repeat_id %||% 1L)
      }, integer(1)),
      stringsAsFactors = FALSE
    )
    fold_meta$fold[!is.finite(fold_meta$fold)] <- fold_meta$fold_seq[!is.finite(fold_meta$fold)]
    fold_idx <- match(fold_id, fold_meta$fold_seq)
    repeat_id <- fold_meta$repeat_id[fold_idx]
    fold_label <- fold_meta$fold[fold_idx]
  }

  if (is.null(coldata) && !is.null(fit@splits@info$coldata)) {
    coldata <- fit@splits@info$coldata
  }
  if (!is.null(coldata)) {
    rn <- rownames(coldata)
    aligned <- FALSE

    if (!is.null(rn) && all(ids_all_chr %in% rn)) {
      coldata <- coldata[ids_all_chr, , drop = FALSE]
      aligned <- TRUE
    } else if ("row_id" %in% names(coldata)) {
      rid_chr <- as.character(coldata[["row_id"]])
      if (!anyDuplicated(rid_chr) && all(ids_all_chr %in% rid_chr)) {
        coldata <- coldata[match(ids_all_chr, rid_chr), , drop = FALSE]
        aligned <- TRUE
      }
    }

    if (!aligned && is.numeric(ids_all) && max(ids_all, na.rm = TRUE) <= nrow(coldata)) {
      if (!is.null(rn) && !all(ids_all_chr %in% rn)) {
        if (isTRUE(strict_align)) {
          stop("`coldata` row names do not match prediction ids and strict_align = TRUE; ",
               "provide matching row names or a 'row_id' column.", call. = FALSE)
        }
        warning("`coldata` row names do not match prediction ids; aligning by row order.")
      }
      coldata <- coldata[ids_all, , drop = FALSE]
      aligned <- TRUE
    }

    if (!aligned) {
      warning("`coldata` not aligned to predictions; batch association skipped.")
      coldata <- NULL
    }
  }

  coldata_pred <- NULL
  if (!is.null(coldata)) {
    idx_pred <- match(ids_pred_chr, ids_all_chr)
    if (anyNA(idx_pred)) {
      warning("`coldata` not aligned to predictions; batch association skipped.")
    } else {
      coldata_pred <- coldata[idx_pred, , drop = FALSE]
    }
  }

  batch_df <- data.frame()
  if (!skip_batch_assoc && !is.null(coldata_pred)) {
    if (is.null(batch_cols)) {
      batch_cols <- intersect(c("batch", "plate", "center", "site", "study"), colnames(coldata_pred))
    }
    if (length(batch_cols) > 0) {
      fold_valid <- !is.na(fold_label)
      repeat_valid <- !is.null(repeat_id) && !all(is.na(repeat_id))
      batch_results <- lapply(batch_cols, function(bc) {
        if (!bc %in% colnames(coldata_pred)) return(NULL)
        if (isTRUE(repeat_valid) && any(fold_valid, na.rm = TRUE)) {
          reps <- sort(unique(repeat_id[fold_valid]))
          reps <- reps[is.finite(reps)]
          rep_rows <- lapply(reps, function(r) {
            idx <- fold_valid & repeat_id == r
            if (!any(idx)) return(NULL)
            tab <- table(
              fold = fold_label[idx],
              batch = as.factor(coldata_pred[idx, bc])
            )
            df <- as.data.frame(.chisq_assoc(tab))
            df$repeat_id <- r
            df
          })
          rep_rows <- rep_rows[!vapply(rep_rows, is.null, logical(1))]
          if (!length(rep_rows)) return(NULL)
          do.call(rbind, rep_rows)
        } else {
          tab <- table(
            fold = fold_label[fold_valid],
            batch = as.factor(coldata_pred[fold_valid, bc])
          )
          df <- as.data.frame(.chisq_assoc(tab))
          df$repeat_id <- 1L
          df
        }
      })
      names(batch_results) <- batch_cols
      keep_idx <- !vapply(batch_results, is.null, logical(1))
      if (any(keep_idx)) {
        batch_results <- batch_results[keep_idx]
        batch_df <- do.call(rbind, Map(function(nm, df) { df$batch_col <- nm; df }, names(batch_results), batch_results))
        batch_df <- batch_df[, c("batch_col", "repeat_id", "stat", "df", "pval", "cramer_v")]
      }
    }
  }

  # --- Target leakage scan (feature-wise outcome association) ----------------
  target_df <- data.frame()
  if (isTRUE(target_scan)) {
    y_outcome <- NULL
    if (!is.null(coldata) && !is.null(outcome_col) && outcome_col %in% names(coldata)) {
      y_outcome <- coldata[[outcome_col]]
    }
    if (is.null(y_outcome) || !length(y_outcome)) {
      id_first <- !duplicated(as.character(all_pred$id))
      truth_map <- all_pred$truth[id_first]
      names(truth_map) <- as.character(all_pred$id[id_first])
      y_outcome <- truth_map[ids_all_chr]
    }
    if (is.null(X_ref) && !is.null(fit@info$X_ref)) X_ref <- fit@info$X_ref
    if (!is.null(X_ref) && !is.null(y_outcome)) {
      X_scan <- .align_by_ids(X_ref, ids_all_chr, sample_ids = sample_ids,
                              strict_align = strict_align)
      if (!is.null(X_scan)) {
        target_df <- .target_assoc_scan(
          X_scan, y_outcome, task,
          positive_class = fit@info$positive_class,
          threshold = target_threshold
        )
      }
    }
  }

  # --- Multivariate/interaction target scan ----------------------------------
  target_multivariate <- data.frame()
  if (isTRUE(target_scan_multivariate)) {
    if (is.null(X_ref) && !is.null(fit@info$X_ref)) X_ref <- fit@info$X_ref
    max_idx <- suppressWarnings(max(vapply(folds, function(z) {
      idx <- c(z$train, z$test)
      if (length(idx)) max(idx) else 0L
    }, integer(1)), na.rm = TRUE))
    if (!is.finite(max_idx) || max_idx < 1L) max_idx <- NA_integer_
    sample_ids_scan <- resolve_sample_ids(fit, fallback_n = max_idx)
    if (!is.null(sample_ids_scan) && length(sample_ids_scan)) {
      sample_ids_scan <- as.character(sample_ids_scan)
    }
    coldata_scan <- fit@splits@info$coldata %||% NULL
    if (!is.null(coldata_scan) && !is.null(sample_ids_scan)) {
      coldata_scan <- .align_by_ids(coldata_scan, sample_ids_scan,
                                    sample_ids = sample_ids_scan, warn = FALSE)
    }
    y_outcome_scan <- NULL
    if (!is.null(coldata_scan) && !is.null(outcome_col) &&
        outcome_col %in% names(coldata_scan)) {
      y_outcome_scan <- coldata_scan[[outcome_col]]
    }
    if (is.null(y_outcome_scan) || !length(y_outcome_scan)) {
      id_first <- !duplicated(as.character(all_pred$id))
      truth_map <- all_pred$truth[id_first]
      names(truth_map) <- as.character(all_pred$id[id_first])
      if (!is.null(sample_ids_scan)) {
        y_outcome_scan <- truth_map[sample_ids_scan]
      }
    }
    if (!is.null(X_ref) && !is.null(y_outcome_scan) && length(y_outcome_scan) &&
        !is.null(sample_ids_scan) && length(sample_ids_scan)) {
      X_scan_mv <- .align_by_ids(X_ref, sample_ids_scan, sample_ids = sample_ids_scan)
      if (!is.null(X_scan_mv)) {
        permute_outcome <- NULL
        if (!is.null(coldata_scan) && !is.null(outcome_col) &&
            outcome_col %in% names(coldata_scan) &&
            perm_mode_use %in% c("subject_grouped", "batch_blocked", "study_loocv", "time_series")) {
          folds_perm <- list(list(test = seq_len(length(y_outcome_scan)), fold = 1L, repeat_id = 1L))
          perm_source <- .permute_labels_factory(
            cd = coldata_scan, outcome = outcome_col, mode = perm_mode_use,
            folds = folds_perm, perm_stratify = perm_stratify, time_block = time_block,
            block_len = block_len, seed = seed,
            group_col = fit@splits@info$group, batch_col = fit@splits@info$batch,
            study_col = fit@splits@info$study, time_col = fit@splits@info$time,
            perm_refit = FALSE
          )
          permute_outcome <- function(b) perm_source(b)[[1]]
        }
        target_multivariate <- .target_scan_multivariate(
          X_scan_mv, y_outcome_scan, task, fit@splits, folds, coldata_scan, seed,
          B = target_scan_multivariate_B,
          max_components = target_scan_multivariate_components,
          interactions = target_scan_multivariate_interactions,
          positive_class = fit@info$positive_class,
          permute_outcome = permute_outcome
        )
      }
    }
  }

  if (nrow(batch_df)) {
    batch_df$mechanism_class <- "confounding_alignment"
    batch_df <- batch_df[, c("mechanism_class", setdiff(names(batch_df), "mechanism_class")), drop = FALSE]
  }
  if (nrow(target_df)) {
    target_df$mechanism_class <- "proxy_target_leakage"
    target_df$p_value_adj <- NA_real_
    target_df$flag_fdr <- NA
    if (identical(target_p_adjust, "none")) {
      target_df$flag_fdr <- as.logical(target_df$flag)
    } else if ("p_value" %in% names(target_df)) {
      finite_idx <- which(is.finite(target_df$p_value))
      if (length(finite_idx)) {
        target_df$p_value_adj[finite_idx] <- stats::p.adjust(
          target_df$p_value[finite_idx],
          method = target_p_adjust
        )
        target_df$flag_fdr <- !is.na(target_df$p_value_adj) &
          is.finite(target_df$p_value_adj) &
          target_df$p_value_adj <= target_alpha
      } else {
        target_df$flag_fdr <- rep(FALSE, nrow(target_df))
      }
    }
    target_df <- target_df[, c("mechanism_class", setdiff(names(target_df), "mechanism_class")), drop = FALSE]
  }
  if (nrow(target_multivariate)) {
    target_multivariate$mechanism_class <- "proxy_target_leakage"
    target_multivariate <- target_multivariate[, c("mechanism_class", setdiff(names(target_multivariate), "mechanism_class")), drop = FALSE]
  }

  # --- Duplicate / near-duplicate detection ---------------------------------
  dup_df <- data.frame()
  if (is.null(X_ref) && !is.null(fit@info$X_ref)) X_ref <- fit@info$X_ref
  if (!is.null(X_ref)) {
    n_samples <- NA_integer_
    if (!is.null(fit@info$sample_ids)) {
      n_samples <- length(fit@info$sample_ids)
    }
    if (!is.finite(n_samples) || n_samples < 1L) {
      if (!is.null(fit@splits@info$coldata)) {
        n_samples <- nrow(fit@splits@info$coldata)
      }
    }
    if (!is.finite(n_samples) || n_samples < 1L) {
      if (isTRUE(compact) && length(fold_assignments)) {
        n_samples <- max(vapply(fold_assignments, length, integer(1)), na.rm = TRUE)
      }
    }
    if (!is.finite(n_samples) || n_samples < 1L) {
      n_samples <- suppressWarnings(max(vapply(folds, function(z) {
        idx <- c(z$train, z$test)
        if (length(idx)) max(idx) else 0L
      }, integer(1)), na.rm = TRUE))
    }
    if (!is.finite(n_samples) || n_samples < 1L) n_samples <- NA_integer_

    sample_ids_all <- resolve_sample_ids(fit, fallback_n = n_samples)
    if (!is.null(sample_ids_all) && length(sample_ids_all)) {
      sample_ids_all <- as.character(sample_ids_all)
    }

    X_use <- NULL
    if (!is.null(sample_ids_all) && length(sample_ids_all)) {
      X_use <- .align_by_ids(
        X_ref, sample_ids_all, sample_ids = sample_ids_all,
        warn = identical(duplicate_scope, "train_test"),
        strict_align = strict_align
      )
    }
    if (is.null(X_use)) {
      if (identical(duplicate_scope, "train_test")) {
        warning("X_ref not aligned to splits; duplicate_scope='train_test' skipped.", call. = FALSE)
      } else {
        warning("X_ref not aligned to splits; proceeding with duplicate_scope='all'.", call. = FALSE)
        X_use <- X_ref
      }
    }

    if (is.null(X_use)) {
      # skip duplicate detection when train/test alignment is required
    } else {
      X <- as.matrix(X_use)
    # Replace NA/NaN/Inf with column medians so RANN::nn2 doesn't fail
    if (anyNA(X) || any(!is.finite(X))) {
      col_med <- apply(X, 2, function(v) { v <- v[is.finite(v)]; if (length(v)) stats::median(v) else 0 })
      for (j in seq_len(ncol(X))) { bad <- !is.finite(X[, j]); if (any(bad)) X[bad, j] <- col_med[j] }
    }
    # choose feature space
    if (feature_space == "rank") {
      X <- t(apply(X, 1, function(row) rank(row, ties.method = "average", na.last = "keep")))
      X[is.na(X)] <- 0
    }
    if (sim_method == "pearson") {
      X <- .row_center(X)
    }
    # standardize to enable similarity via Euclidean NN
    Xn <- .row_l2_normalize(X)

    n <- nrow(Xn)
    p <- ncol(Xn)
    candidate_pairs <- NULL

    use_ann <- (n > 500 || (n * p > 1.5e6)) &&
               requireNamespace("RANN", quietly = TRUE)

    if (use_ann) {
      # approximate k-NN search; cosine/pearson ~ 1 - 0.5*||u - v||^2 for unit vectors
      nn <- RANN::nn2(Xn, k = min(nn_k + 1, n))  # includes self
      idx <- rep(seq_len(n), times = ncol(nn$nn.idx))
      jdx <- as.vector(nn$nn.idx)
      mask <- (idx != jdx)
      idx <- idx[mask]
      jdx <- jdx[mask]
      # compute similarity for candidate pairs
      simv <- rowSums(Xn[idx, , drop = FALSE] * Xn[jdx, , drop = FALSE])
      keep <- which(simv >= sim_threshold)
      if (length(keep)) {
        candidate_pairs <- data.frame(i = idx[keep], j = jdx[keep], sim = simv[keep])
      }
    } else if (n > 500) {
      # chunked exact computation — avoids full n*n allocation
      block_size <- 200L
      pair_list <- list()
      for (start in seq(1L, n, by = block_size)) {
        end <- min(start + block_size - 1L, n)
        block_idx <- start:end
        S_block <- Xn[block_idx, , drop = FALSE] %*% t(Xn)
        # zero out self-matches and upper triangle within the block
        for (bi in seq_along(block_idx)) {
          ri <- block_idx[bi]
          S_block[bi, ri:n] <- 0
        }
        hits <- which(S_block >= sim_threshold, arr.ind = TRUE)
        if (nrow(hits) > 0) {
          pair_list[[length(pair_list) + 1L]] <- data.frame(
            i = block_idx[hits[, 1]],
            j = hits[, 2],
            sim = S_block[hits]
          )
        }
      }
      if (length(pair_list)) {
        candidate_pairs <- do.call(rbind, pair_list)
      }
    } else {
      # exact small-n path
      S <- .cosine_sim_block(Xn)
      S[upper.tri(S, TRUE)] <- 0
      which_dup <- which(S >= sim_threshold, arr.ind = TRUE)
      if (nrow(which_dup) > 0)
        candidate_pairs <- data.frame(i = which_dup[, 1], j = which_dup[, 2], sim = S[which_dup])
    }

    if (!is.null(candidate_pairs)) {
      # Canonicalize and deduplicate pairs (kNN branch may produce (i,j) and (j,i))
      swap <- candidate_pairs$i > candidate_pairs$j
      tmp <- candidate_pairs$i[swap]
      candidate_pairs$i[swap] <- candidate_pairs$j[swap]
      candidate_pairs$j[swap] <- tmp
      candidate_pairs <- candidate_pairs[!duplicated(candidate_pairs[, c("i", "j")]), ]
      compute_cross_fold <- function(pairs) {
        n_pairs <- nrow(pairs)
        if (!n_pairs) return(logical(0))
        if (!is.finite(n_samples) || n_samples < 1L) return(rep(NA, n_pairs))
        if (is.null(folds) || !length(folds)) return(rep(NA, n_pairs))
        split_mode <- fit@splits@mode %||% NA_character_

        if (!identical(split_mode, "time_series")) {
          fold_map_by_repeat <- NULL
          if (isTRUE(compact) && length(fold_assignments)) {
            ok_len <- all(vapply(fold_assignments, function(vec) {
              length(vec) == n_samples
            }, logical(1)))
            if (isTRUE(ok_len)) {
              fold_map_by_repeat <- lapply(fold_assignments, function(vec) as.integer(vec))
            }
          }
          if (is.null(fold_map_by_repeat)) {
            repeat_ids <- vapply(folds, function(f) {
              as.integer(f$repeat_id %||% 1L)
            }, integer(1))
            n_rep <- suppressWarnings(max(repeat_ids, na.rm = TRUE))
            if (!is.finite(n_rep) || n_rep < 1L) return(rep(NA, n_pairs))
            fold_map_by_repeat <- lapply(seq_len(n_rep), function(r) rep(NA_integer_, n_samples))
            for (i in seq_along(folds)) {
              r <- repeat_ids[[i]]
              test <- resolve_test_idx(folds[[i]])
              if (length(test)) {
                fold_id <- folds[[i]]$fold %||% i
                fold_map_by_repeat[[r]][test] <- fold_id
              }
            }
          }
          cross <- rep(FALSE, n_pairs)
          i_idx <- pairs$i
          j_idx <- pairs$j
          for (fm in fold_map_by_repeat) {
            if (length(fm) != n_samples) next
            fi <- fm[i_idx]
            fj <- fm[j_idx]
            valid <- !is.na(fi) & !is.na(fj)
            cross <- cross | (valid & fi != fj)
          }
          return(cross)
        }

        fold_map <- rep(NA_integer_, n_samples)
        time_vec <- NULL
        time_col <- fit@splits@info$time %||% NULL
        cd_time <- fit@splits@info$coldata %||% NULL
        if (!is.null(cd_time) && !is.null(time_col) && time_col %in% names(cd_time)) {
          time_vec <- cd_time[[time_col]]
        }
        fold_tmin <- NULL
        if (!is.null(time_vec) && length(time_vec) == n_samples) {
          fold_tmin <- rep(time_vec[1], length(folds))
          fold_tmin[] <- NA
        }
        for (i in seq_along(folds)) {
          test <- resolve_test_idx(folds[[i]])
          if (!length(test)) next
          fold_map[test] <- i
          if (!is.null(fold_tmin)) {
            fold_tmin[i] <- suppressWarnings(min(time_vec[test]))
          }
        }
        if (!is.null(time_vec) && length(time_vec) == n_samples && !is.null(fold_tmin)) {
          split_horizon <- fit@splits@info$horizon %||% 0
          cross <- rep(FALSE, n_pairs)
          i_idx <- pairs$i
          j_idx <- pairs$j
          fi <- fold_map[i_idx]
          fj <- fold_map[j_idx]
          idx_i <- which(!is.na(fi))
          if (length(idx_i)) {
            tmin_i <- fold_tmin[fi[idx_i]]
            valid_i <- which(!is.na(tmin_i))
            if (length(valid_i)) {
              idx_i <- idx_i[valid_i]
              tmin_i <- tmin_i[valid_i]
              if (split_horizon == 0) {
                cross[idx_i] <- time_vec[j_idx[idx_i]] < tmin_i
              } else {
                cutoff <- tmin_i - split_horizon
                cross[idx_i] <- time_vec[j_idx[idx_i]] <= cutoff
              }
            }
          }
          idx_j <- which(!is.na(fj))
          if (length(idx_j)) {
            tmin_j <- fold_tmin[fj[idx_j]]
            valid_j <- which(!is.na(tmin_j))
            if (length(valid_j)) {
              idx_j <- idx_j[valid_j]
              tmin_j <- tmin_j[valid_j]
              if (split_horizon == 0) {
                cross[idx_j] <- cross[idx_j] | (time_vec[i_idx[idx_j]] < tmin_j)
              } else {
                cutoff <- tmin_j - split_horizon
                cross[idx_j] <- cross[idx_j] | (time_vec[i_idx[idx_j]] <= cutoff)
              }
            }
          }
          return(cross)
        }

        has_train <- any(vapply(folds, function(f) !is.null(f$train), logical(1)))
        if (!has_train) return(rep(NA, n_pairs))
        cross <- rep(FALSE, n_pairs)
        for (i in seq_along(folds)) {
          fold <- folds[[i]]
          if (is.null(fold$train) || is.null(fold$test)) next
          train <- fold$train
          test <- fold$test
          if (!length(train) || !length(test)) next
          train_mask <- rep(FALSE, n_samples)
          test_mask <- rep(FALSE, n_samples)
          train_mask[train] <- TRUE
          test_mask[test] <- TRUE
          cross <- cross | (test_mask[pairs$i] & train_mask[pairs$j]) |
            (test_mask[pairs$j] & train_mask[pairs$i])
        }
        cross
      }

      cross_fold <- compute_cross_fold(candidate_pairs)
      if (length(cross_fold)) {
        candidate_pairs$cross_fold <- cross_fold
      }
      if (identical(duplicate_scope, "train_test")) {
        keep <- which(!is.na(cross_fold) & cross_fold)
        if (length(keep)) {
          candidate_pairs <- candidate_pairs[keep, , drop = FALSE]
        } else {
          candidate_pairs <- candidate_pairs[0, , drop = FALSE]
        }
      }

      if (nrow(candidate_pairs) > 0) {
        if (sim_method == "cosine") {
          candidate_pairs$cos_sim <- candidate_pairs$sim
        } else if (sim_method == "pearson") {
          candidate_pairs$pearson <- candidate_pairs$sim
        }
        # sort and cap
        ord <- order(candidate_pairs$sim, decreasing = TRUE)
        candidate_pairs <- candidate_pairs[ord, , drop = FALSE]
        if (nrow(candidate_pairs) > max_pairs)
          candidate_pairs <- candidate_pairs[seq_len(max_pairs), , drop = FALSE]
        dup_df <- candidate_pairs
      }
    }
    }
  }
  if (nrow(dup_df)) {
    dup_df$mechanism_class <- "duplicate_overlap"
    dup_df <- dup_df[, c("mechanism_class", setdiff(names(dup_df), "mechanism_class")), drop = FALSE]
  }

  mechanism_summary <- .bio_mechanism_summary(
    perm_df = perm_df,
    batch_df = batch_df,
    target_df = target_df,
    dup_df = dup_df,
    split_mode = fit@splits@mode %||% NA_character_
  )

  # --- Assemble S4 object ----------------------------------------------------
  perm_values <- if (isTRUE(return_perm)) perm_vals else numeric(0)

  new("LeakAudit",
      fit = fit,
      permutation_gap = perm_df,
      perm_values = perm_values,
      batch_assoc = batch_df,
      target_assoc = target_df,
      duplicates = dup_df,
      trail = trail,
      info = list(
        n_perm = B,
        perm_stratify = perm_stratify,
        perm_method = perm_method,
        perm_mode = perm_mode_use,
        perm_refit_mode = perm_refit_mode,
        perm_refit_reason = perm_refit_reason,
        perm_refit_auto_max = perm_refit_auto_max,
        parallel = parallel,
        target_scan = target_scan,
        target_scan_multivariate = target_scan_multivariate,
        target_scan_multivariate_B = target_scan_multivariate_B,
        target_scan_multivariate_components = target_scan_multivariate_components,
        target_scan_multivariate_interactions = target_scan_multivariate_interactions,
        target_multivariate = target_multivariate,
        target_threshold = target_threshold,
        target_p_adjust = target_p_adjust,
        target_alpha = target_alpha,
        sim_method = sim_method,
        feature_space = feature_space,
        duplicate_threshold = sim_threshold,
        duplicate_scope = duplicate_scope,
        taxonomy = .bio_leakage_taxonomy(),
        mechanism_summary = mechanism_summary,
        nn_k = nn_k,
        max_pairs = max_pairs,
        batch_cols = batch_cols,
        permutation_se = p_se,
        ci_delta = seci$ci,
        se_delta = seci$se,
        ci_method = ci_method,
        include_z = include_z,
        provenance = fit@info$provenance %||% NULL
      ))
}

#' Audit leakage per learner
#'
#' Runs [audit_leakage()] separately for each learner recorded in a [LeakFit]
#' and returns a named list of [LeakAudit] objects. Use this when a single fit
#' contains predictions for multiple models and you want model-specific audits.
#' If predictions do not include learner IDs, only a single audit can be run and
#' requesting multiple learners is an error.
#'
#' @param fit A [LeakFit] object produced by [fit_resample()]. It must contain
#'   predictions and split metadata. Learner IDs must be present in predictions
#'   to audit multiple models.
#' @param metric Character scalar. One of `"auc"`, `"pr_auc"`, `"accuracy"`,
#'   `"macro_f1"`, `"log_loss"`, `"rmse"`, or `"cindex"`. Controls which metric
#'   is audited for each learner.
#' @param learners Character vector or NULL. If NULL (default), audits all
#'   learners found in predictions. If provided, must match learner IDs stored in
#'   the predictions. Supplying more than one learner requires learner IDs.
#' @param parallel_learners Logical scalar. If TRUE, runs per-learner audits in
#'   parallel using `future.apply` (if installed). This changes runtime but not
#'   the audit results.
#' @param mc.cores Integer scalar or NULL. Number of workers used when
#'   `parallel_learners = TRUE`. Defaults to the minimum of available cores and
#'   the number of learners.
#' @param ... Additional named arguments forwarded to [audit_leakage()] for each
#'   learner. These control the audit itself. Common options include:
#'   `B` (integer permutations), `perm_stratify` (logical or `"auto"`),
#'   `perm_refit` (logical), `perm_refit_spec` (list),
#'   `time_block` (character), `block_len` (integer or NULL), `include_z`
#'   (logical), `ci_method` (character), `boot_B` (integer), `parallel`
#'   (logical), `seed` (integer), `return_perm` (logical), `batch_cols`
#'   (character vector), `coldata` (data.frame), `X_ref` (matrix/data.frame),
#'   `target_scan` (logical), `target_threshold` (numeric),
#'   `target_p_adjust` (character), `target_alpha` (numeric),
#'   `feature_space` (character), `sim_method` (character), `sim_threshold` (numeric),
#'   `nn_k` (integer), `max_pairs` (integer), and `duplicate_scope` (character).
#'   See [audit_leakage()] for full definitions; changing these values changes
#'   each learner's audit.
#' @return A named list of \code{\linkS4class{LeakAudit}} objects, where each
#'   element is keyed by the learner ID (character string). Each
#'   \code{LeakAudit} object contains the same slots as described in
#'   \code{\link{audit_leakage}}: \code{fit}, \code{permutation_gap},
#'   \code{perm_values}, \code{batch_assoc}, \code{target_assoc},
#'   \code{duplicates}, \code{trail}, and \code{info}. Use \code{names()} to
#'   retrieve learner IDs, and access individual audits with \code{[[learner_id]]}
#'   or \code{$learner_id}. Each audit reflects the performance and diagnostics
#'   for that specific learner's predictions.
#' @examples
#' set.seed(1)
#' df <- data.frame(
#'   subject = rep(1:6, each = 2),
#'   outcome = factor(rep(c(0, 1), 6)),
#'   x1 = rnorm(12),
#'   x2 = rnorm(12)
#' )
#' splits <- make_split_plan(df, outcome = "outcome",
#'                       mode = "subject_grouped", group = "subject",
#'                       v = 3, progress = FALSE)
#' custom <- list(
#'   glm = list(
#'     fit = function(x, y, task, weights, ...) {
#'       stats::glm(y ~ ., data = data.frame(y = y, x),
#'                  family = stats::binomial(), weights = weights)
#'     },
#'     predict = function(object, newdata, task, ...) {
#'       as.numeric(stats::predict(object,
#'                                 newdata = as.data.frame(newdata),
#'                                 type = "response"))
#'     }
#'   )
#' )
#' custom$glm2 <- custom$glm
#' fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                     learner = c("glm", "glm2"), custom_learners = custom,
#'                     metrics = "auc", refit = FALSE, seed = 1)
#' audits <- audit_leakage_by_learner(fit, metric = "auc", B = 10,
#'                                    perm_stratify = FALSE)
#' names(audits)
#'
#' @export
audit_leakage_by_learner <- function(fit,
                                     metric = c("auc", "pr_auc", "accuracy", "macro_f1", "log_loss", "rmse", "cindex"),
                                     learners = NULL,
                                     parallel_learners = FALSE,
                                     mc.cores = NULL,
                                     ...) {
  metric <- match.arg(metric)
  dots <- list(...)
  if ("learner" %in% names(dots)) {
    stop("Use `learners` to select models; do not pass `learner` to audit_leakage_by_learner().")
  }

  pred_list <- lapply(fit@predictions, function(df) data.frame(df, stringsAsFactors = FALSE))
  pred_df <- if (length(pred_list)) do.call(rbind, pred_list) else NULL
  if (is.null(pred_df) || !nrow(pred_df)) {
    stop("No predictions available in LeakFit object.")
  }

  has_learner <- "learner" %in% names(pred_df)
  if (is.null(learners)) {
    if (has_learner) {
      learners <- unique(as.character(pred_df$learner))
    } else if (!is.null(fit@metrics) && "learner" %in% names(fit@metrics)) {
      learners <- unique(as.character(fit@metrics$learner))
    } else {
      learners <- character(0)
    }
  }

  if (!length(learners)) {
    stop("No learner names found to audit.")
  }

  if (has_learner) {
    available <- unique(as.character(pred_df$learner))
    missing <- setdiff(learners, available)
    if (length(missing)) {
      stop(sprintf("Learner(s) not found in predictions: %s", paste(missing, collapse = ", ")))
    }
  } else if (length(learners) > 1L) {
    stop("Predictions do not include learner IDs; cannot audit per learner. Refit with updated bioLeak.")
  } else {
    warning("Predictions do not include learner IDs; returning a single audit without learner filtering.")
  }

  if (isTRUE(parallel_learners)) {
    if (!requireNamespace("future.apply", quietly = TRUE)) {
      warning("parallel_learners=TRUE requires the 'future.apply' package; using sequential execution.")
      parallel_learners <- FALSE
    }
  }
  if (isTRUE(parallel_learners)) {
    if (is.null(mc.cores)) {
      detected <- parallel::detectCores()
      if (is.na(detected) || detected < 1L) detected <- 1L
      mc.cores <- min(length(learners), detected)
    }
    mc.cores <- max(1L, as.integer(mc.cores))
    if (requireNamespace("future", quietly = TRUE)) {
      old_plan <- future::plan()
      on.exit(future::plan(old_plan), add = TRUE)
      future::plan(future::multisession, workers = mc.cores)
    }
  }

  runner <- function(ln) {
    audit_leakage(fit, metric = metric, learner = if (has_learner) ln else NULL, ...)
  }
  audits <- if (isTRUE(parallel_learners)) {
    future.apply::future_lapply(learners, runner, future.seed = TRUE)
  } else {
    lapply(learners, runner)
  }
  names(audits) <- learners
  class(audits) <- c("LeakAuditList", class(audits))
  audits
}

#' @export
print.LeakAuditList <- function(x, digits = 3, ...) {
  n <- length(x)
  cat(sprintf("LeakAuditList with %d learner%s\n", n, ifelse(n == 1L, "", "s")))
  if (!n) return(invisible(x))

  learners <- names(x)
  if (is.null(learners) || any(!nzchar(learners))) {
    learners <- paste0("learner_", seq_len(n))
  }

  summary_rows <- lapply(seq_along(x), function(i) {
    aud <- x[[i]]
    pg <- if (inherits(aud, "LeakAudit")) aud@permutation_gap else NULL
    metric_obs <- NA_real_
    perm_mean <- NA_real_
    gap <- NA_real_
    p_value <- NA_real_
    z_val <- NA_real_
    n_perm <- NA_real_
    if (!is.null(pg) && nrow(pg) > 0) {
      metric_obs <- pg$metric_obs[1]
      perm_mean <- pg$perm_mean[1]
      gap <- pg$gap[1]
      p_value <- pg$p_value[1]
      if ("z" %in% names(pg)) z_val <- pg$z[1]
      if ("n_perm" %in% names(pg)) n_perm <- pg$n_perm[1]
    }

    batch_df <- if (inherits(aud, "LeakAudit")) aud@batch_assoc else NULL
    batch_p_min <- NA_real_
    batch_col_min_p <- NA_character_
    batch_v_max <- NA_real_
    batch_col_max_v <- NA_character_
    if (!is.null(batch_df) && nrow(batch_df) > 0) {
      pvals <- batch_df$pval
      if (any(is.finite(pvals))) {
        idx <- which.min(pvals)
        batch_p_min <- pvals[idx]
        batch_col_min_p <- as.character(batch_df$batch_col[idx])
      }
      vvals <- batch_df$cramer_v
      if (any(is.finite(vvals))) {
        idx <- which.max(vvals)
        batch_v_max <- vvals[idx]
        batch_col_max_v <- as.character(batch_df$batch_col[idx])
      }
    }

    dup_df <- if (inherits(aud, "LeakAudit")) aud@duplicates else NULL
    dup_pairs <- NA_real_
    dup_max_sim <- NA_real_
    if (!is.null(dup_df)) {
      if (nrow(dup_df) > 0) {
        dup_pairs <- nrow(dup_df)
        if ("sim" %in% names(dup_df)) {
          dup_max_sim <- max(dup_df$sim, na.rm = TRUE)
        } else if ("cos_sim" %in% names(dup_df)) {
          dup_max_sim <- max(dup_df$cos_sim, na.rm = TRUE)
        } else if ("pearson" %in% names(dup_df)) {
          dup_max_sim <- max(dup_df$pearson, na.rm = TRUE)
        }
      } else if (ncol(dup_df) > 0) {
        dup_pairs <- 0
      }
    }
    dup_threshold <- if (inherits(aud, "LeakAudit") &&
                         !is.null(aud@info$duplicate_threshold)) {
      aud@info$duplicate_threshold
    } else {
      NA_real_
    }

    data.frame(
      learner = learners[[i]],
      metric_obs = metric_obs,
      perm_mean = perm_mean,
      gap = gap,
      p_value = p_value,
      z = z_val,
      n_perm = n_perm,
      batch_p_min = batch_p_min,
      batch_col_min_p = batch_col_min_p,
      batch_v_max = batch_v_max,
      batch_col_max_v = batch_col_max_v,
      dup_pairs = dup_pairs,
      dup_threshold = dup_threshold,
      dup_max_sim = dup_max_sim,
      stringsAsFactors = FALSE
    )
  })

  summary_df <- do.call(rbind, summary_rows)
  if (nrow(summary_df)) {
    numeric_cols <- vapply(summary_df, is.numeric, logical(1))
    summary_df[numeric_cols] <- lapply(summary_df[numeric_cols], function(v) round(v, digits))
    print(summary_df, row.names = FALSE)
  }
  invisible(x)
}

.coerce_truth_like <- function(template, values) {
  if (inherits(template, "Surv")) {
    values
  } else if (is.factor(template)) {
    factor(values, levels = levels(template))
  } else if (is.logical(template)) {
    as.logical(values)
  } else if (is.numeric(template)) {
    as.numeric(values)
  } else {
    values
  }
}

skip_if_tune_deps <- function() {
  skip_if_not_installed("tune")
  skip_if_not_installed("dials")
  skip_if_not_installed("glmnet")
  skip_if_not_installed("recipes")
  skip_if_not_installed("rsample")
  skip_if_not_installed("yardstick")
  skip_if_not_installed("workflows")
  skip_if_not_installed("parsnip")
}

tune_resample_quiet <- function(...) {
  out <- NULL
  capture.output({
    out <- suppressWarnings(tune_resample(...))
  })
  out
}

test_that("tune_resample selects a deterministic simple model with one_std_err", {
  skip_if_tune_deps()

  df <- make_class_df(24)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                                  mode = "subject_grouped", group = "subject",
                                  v = 2, nested = TRUE, stratify = FALSE, seed = 11)

  spec <- parsnip::logistic_reg(penalty = tune::tune(), mixture = 1) |>
    parsnip::set_engine("glmnet")
  rec <- recipes::recipe(outcome ~ x1 + x2, data = df)

  local_mocked_bindings(
    tune_grid = function(...) structure(list(), class = "mock_tune_results"),
    collect_metrics = function(...) {
      data.frame(
        .metric = rep("accuracy", 4),
        .estimator = rep("binary", 4),
        .estimate = c(0.80, 0.76, 0.77, 0.77),
        .config = c("cfg_a", "cfg_a", "cfg_b", "cfg_b"),
        penalty = c(0.01, 0.01, 1.0, 1.0),
        stringsAsFactors = FALSE
      )
    },
    finalize_workflow = function(x, parameters, ...) x,
    .package = "tune"
  )
  local_mocked_bindings(
    fit_resample = function(x, outcome, splits, ...) {
      truth <- factor(rep(c(0, 1), length.out = 4), levels = c(0, 1))
      methods::new(
        "LeakFit",
        splits = splits,
        metrics = data.frame(fold = 1, learner = "mock", accuracy = 0.75),
        metric_summary = data.frame(learner = "mock", accuracy_mean = 0.75, accuracy_sd = 0),
        audit = data.frame(),
        predictions = list(data.frame(
          id = as.character(seq_len(4)),
          truth = truth,
          pred = c(0.45, 0.55, 0.40, 0.60),
          fold = 1,
          learner = "mock",
          stringsAsFactors = FALSE
        )),
        preprocess = list(),
        learners = list(),
        outcome = outcome,
        task = "binomial",
        feature_names = c("x1", "x2"),
        info = list(sample_ids = as.character(seq_len(nrow(x))))
      )
    },
    .package = "bioLeak"
  )

  tuned_best <- tune_resample_quiet(df, outcome = "outcome", splits = splits,
                                    learner = spec, preprocess = rec, grid = 2,
                                    metrics = "accuracy", selection = "best", seed = 11)
  tuned_ose <- tune_resample_quiet(df, outcome = "outcome", splits = splits,
                                   learner = spec, preprocess = rec, grid = 2,
                                   metrics = "accuracy", selection = "one_std_err", seed = 11)

  expect_true(nrow(tuned_best$best_params) > 0)
  expect_true(nrow(tuned_ose$best_params) > 0)
  expect_true(all(tuned_best$best_params$penalty == 0.01))
  expect_true(all(tuned_ose$best_params$penalty == 1.0))
})

test_that("tune_resample supports final refit and stores fold status", {
  skip_if_tune_deps()

  df <- make_class_df(80)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                                  mode = "subject_grouped", group = "subject",
                                  v = 2, nested = TRUE, stratify = FALSE, seed = 2)

  spec <- parsnip::logistic_reg(penalty = tune::tune(), mixture = 1) |>
    parsnip::set_engine("glmnet")
  rec <- recipes::recipe(outcome ~ x1 + x2, data = df)

  tuned <- tune_resample_quiet(df, outcome = "outcome", splits = splits,
                               learner = spec, preprocess = rec, grid = 2,
                               metrics = "accuracy", selection = "best",
                               refit = TRUE, seed = 2)

  expect_true(isTRUE(tuned$info$refit))
  expect_true(!is.null(tuned$final_model))
  expect_s3_class(tuned$final_workflow, "workflow")
  expect_true(nrow(tuned$final_params) > 0)
  expect_true(is.data.frame(tuned$fold_status))
  expect_true(nrow(tuned$fold_status) == length(splits@indices))
  expect_true(all(tuned$fold_status$status %in% c("success", "skipped", "failed")))
})

test_that("final refit uses aggregated params, not single best outer fold", {
  skip_if_tune_deps()

  df <- make_class_df(80)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                                  mode = "subject_grouped", group = "subject",
                                  v = 2, nested = TRUE, stratify = FALSE, seed = 2)

  spec <- parsnip::logistic_reg(penalty = tune::tune(), mixture = 1) |>
    parsnip::set_engine("glmnet")
  rec <- recipes::recipe(outcome ~ x1 + x2, data = df)

  tuned <- tune_resample_quiet(df, outcome = "outcome", splits = splits,
                               learner = spec, preprocess = rec, grid = 3,
                               metrics = "accuracy", selection = "best",
                               refit = TRUE, seed = 2)

  # refit_method must be "aggregate", not a single-fold selection
  expect_identical(tuned$info$refit_method, "aggregate")

  # refit_fold must be NA — no single fold was selected based on outer metrics
  expect_true(is.na(tuned$info$refit_fold))

  # final_params must have exactly the hyperparameter columns, no fold column
  expect_false("fold" %in% names(tuned$final_params))

  # --- Behavioral proof that outer test metrics are not consulted ---
  # final_params must exactly equal the coordinate-wise aggregate of ALL folds'
  # best_params. This is the core independence guarantee: the old (leaky) code
  # did which.max(metrics_df[[metric]]) to pick a single fold, so final_params
  # would equal that fold's row. The new code aggregates all folds equally.
  bp <- tuned$best_params
  param_cols <- setdiff(names(bp), "fold")
  expect_true(nrow(bp) >= 2,
              info = "Need at least 2 outer folds to test aggregation")
  expect_true(length(param_cols) >= 1,
              info = "Need at least 1 hyperparameter column")

  for (col in param_cols) {
    if (is.numeric(bp[[col]])) {
      expected <- stats::median(bp[[col]], na.rm = TRUE)
      expect_equal(tuned$final_params[[col]], expected,
                   info = paste("Param", col,
                                "must equal median across ALL folds, not a single fold's value"))
    } else {
      tbl <- table(bp[[col]])
      expected <- names(tbl)[which.max(tbl)]
      expect_equal(as.character(tuned$final_params[[col]]), expected,
                   info = paste("Param", col,
                                "must equal majority vote across ALL folds"))
    }
  }

  # Verify final_params is not simply copied from the best-metric fold.
  # Identify the fold that had the best outer metric — if it differs from the
  # aggregate, the old (leaky) logic would have returned that fold's params.
  metric_col <- tuned$info$selection_metric
  if (!is.null(metric_col) && metric_col %in% names(tuned$metrics)) {
    metric_vals <- tuned$metrics[[metric_col]]
    minimize <- metric_col %in% c("rmse", "mae", "log_loss", "mn_log_loss")
    best_idx <- if (minimize) which.min(metric_vals) else which.max(metric_vals)
    best_fold <- tuned$metrics$fold[best_idx]
    best_fold_params <- bp[bp$fold == best_fold, param_cols, drop = FALSE]

    # Check if folds actually disagree on params
    folds_disagree <- FALSE
    for (col in param_cols) {
      if (length(unique(bp[[col]])) > 1L) {
        folds_disagree <- TRUE
        break
      }
    }

    if (folds_disagree && nrow(best_fold_params) == 1L) {
      # When folds disagree, the aggregate should NOT equal the best-metric
      # fold's params. This is the direct proof: the old code would have
      # returned best_fold_params, the new code returns the aggregate.
      params_match_best <- all(vapply(param_cols, function(col) {
        identical(tuned$final_params[[col]], best_fold_params[[col]])
      }, logical(1)))
      expect_false(params_match_best,
                   info = paste("final_params must not equal fold", best_fold,
                                "params (the best-metric fold) — that would be",
                                "nested-CV leakage"))
    }
  }
})

test_that("LeakTune summary and audit handle skipped outer folds", {
  skip_if_tune_deps()

  set.seed(1)
  df <- data.frame(
    subject = paste0("s", seq_len(60)),
    outcome = factor(c(rep(0, 30), rep(1, 30)), levels = c(0, 1)),
    x1 = rnorm(60),
    x2 = rnorm(60),
    stringsAsFactors = FALSE
  )

  indices <- list(
    list(train = 1:20, test = 21:60, fold = 1L, repeat_id = 1L),
    list(train = c(1:20, 31:60), test = 21:30, fold = 2L, repeat_id = 1L)
  )
  split_info <- list(
    outcome = "outcome",
    v = 2L,
    repeats = 1L,
    seed = 1L,
    mode = "subject_grouped",
    perm_mode = "subject_grouped",
    group = "subject",
    batch = NULL,
    study = NULL,
    time = NULL,
    stratify = FALSE,
    nested = FALSE,
    horizon = 0,
    summary = data.frame(
      fold = c(1L, 2L),
      repeat_id = c(1L, 1L),
      train_n = c(20L, 50L),
      test_n = c(40L, 10L)
    ),
    hash = "manual",
    inner = NULL,
    compact = FALSE,
    fold_assignments = NULL,
    coldata = df
  )
  splits <- bioLeak:::LeakSplits(mode = "subject_grouped", indices = indices, info = split_info)

  spec <- parsnip::logistic_reg(penalty = tune::tune(), mixture = 1) |>
    parsnip::set_engine("glmnet")
  rec <- recipes::recipe(outcome ~ x1 + x2, data = df)

  tuned <- tune_resample_quiet(df, outcome = "outcome", splits = splits,
                               learner = spec, preprocess = rec, grid = 2,
                               metrics = "accuracy", seed = 1)

  expect_true(is.null(tuned$outer_fits[[1]]))
  expect_true(!is.null(tuned$outer_fits[[2]]))
  expect_true(is.data.frame(tuned$fold_status))
  expect_equal(nrow(tuned$fold_status), 2)
  expect_true(any(tuned$fold_status$status == "skipped"))
  expect_true(any(tuned$fold_status$status == "success"))

  sum_out <- paste(capture.output(summary(tuned)), collapse = "\n")
  expect_match(sum_out, "Outer Folds: 1 successful / 2 total", fixed = TRUE)

  audit <- suppressWarnings(expect_no_error(
    audit_leakage(
      tuned,
      metric = "accuracy",
      B = 3,
      boot_B = 20,
      perm_stratify = FALSE,
      return_perm = FALSE,
      target_scan = FALSE,
      target_scan_multivariate = FALSE
    )
  ))
  expect_s4_class(audit, "LeakAudit")
})

test_that("tune_resample tunes binomial thresholds from inner predictions", {
  skip_if_tune_deps()

  df <- make_class_df(24)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                                  mode = "subject_grouped", group = "subject",
                                  v = 2, nested = TRUE, stratify = FALSE, seed = 21)

  spec <- parsnip::logistic_reg(penalty = tune::tune(), mixture = 1) |>
    parsnip::set_engine("glmnet")
  rec <- recipes::recipe(outcome ~ x1 + x2, data = df)

  seen_save_pred <- logical()
  seen_thresholds <- numeric()

  local_mocked_bindings(
    tune_grid = function(..., control) {
      seen_save_pred <<- c(seen_save_pred, isTRUE(control$save_pred))
      structure(list(), class = "mock_tune_results")
    },
    collect_metrics = function(...) {
      data.frame(
        .metric = rep("accuracy", 4),
        .estimator = rep("binary", 4),
        .estimate = c(0.8, 0.8, 0.7, 0.7),
        .config = c("cfg_a", "cfg_a", "cfg_b", "cfg_b"),
        penalty = c(0.01, 0.01, 1.0, 1.0),
        stringsAsFactors = FALSE
      )
    },
    collect_predictions = function(...) {
      data.frame(
        .config = rep("cfg_a", 4),
        outcome = factor(c(0, 1, 0, 1), levels = c(0, 1)),
        .pred_0 = c(0.9, 0.6, 0.55, 0.4),
        .pred_1 = c(0.1, 0.4, 0.45, 0.6),
        stringsAsFactors = FALSE
      )
    },
    finalize_workflow = function(x, parameters, ...) x,
    .package = "tune"
  )
  local_mocked_bindings(
    fit_resample = function(x, outcome, splits, classification_threshold = 0.5, ...) {
      seen_thresholds <<- c(seen_thresholds, classification_threshold)
      truth <- factor(rep(c(0, 1), length.out = 4), levels = c(0, 1))
      methods::new(
        "LeakFit",
        splits = splits,
        metrics = data.frame(fold = 1, learner = "mock", accuracy = 0.75),
        metric_summary = data.frame(learner = "mock", accuracy_mean = 0.75, accuracy_sd = 0),
        audit = data.frame(),
        predictions = list(data.frame(
          id = as.character(seq_len(4)),
          truth = truth,
          pred = c(0.1, 0.4, 0.45, 0.6),
          fold = 1,
          learner = "mock",
          stringsAsFactors = FALSE
        )),
        preprocess = list(),
        learners = list(),
        outcome = outcome,
        task = "binomial",
        feature_names = c("x1", "x2"),
        info = list(sample_ids = as.character(seq_len(nrow(x))))
      )
    },
    .package = "bioLeak"
  )

  tuned <- tune_resample_quiet(df, outcome = "outcome", splits = splits,
                               learner = spec, preprocess = rec, grid = 2,
                               metrics = "accuracy", selection = "best", seed = 21,
                               tune_threshold = TRUE, threshold_grid = c(0.2, 0.8))

  expect_true(all(seen_save_pred))
  expect_equal(length(seen_thresholds), length(splits@indices))
  expect_true(all(abs(seen_thresholds - 0.2) < 1e-12))
  expect_true(is.data.frame(tuned$thresholds))
  expect_equal(nrow(tuned$thresholds), length(splits@indices))
  expect_true(all(abs(tuned$thresholds$threshold - 0.2) < 1e-12))
  expect_true(isTRUE(tuned$info$threshold_tuned))
  expect_equal(tuned$info$threshold_metric, "accuracy")
})

test_that("duplicate detection completes for n=520 without error (chunked path)", {
  skip_on_cran()
  set.seed(42)
  n <- 520
  p <- 10
  X <- matrix(rnorm(n * p), nrow = n)
  # inject a near-duplicate
  X[2, ] <- X[1, ] + rnorm(p, sd = 0.001)
  df <- data.frame(
    subject = seq_len(n),
    outcome = rbinom(n, 1, 0.5),
    X
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, progress = FALSE)
  fit <- fit_resample(df, outcome = "outcome", splits = splits,
                      learner = "glmnet",
                      preprocess = list(
                        impute = list(method = "median"),
                        normalize = list(method = "zscore")
                      ))
  # Should complete without error — tests the chunked path (n > 500, no RANN)
  audit <- tryCatch(
    audit_leakage(fit, B = 5, X_ref = as.matrix(df[, paste0("X", 1:p)])),
    error = function(e) e
  )
  expect_false(inherits(audit, "error"))
})

test_that("feature-count trigger logic: n=300, p=6000 should trigger ANN path", {
  # This tests that the condition (n * p > 1.5e6) is correct
  n <- 300
  p <- 6000
  # n * p = 1.8e6 > 1.5e6 should trigger ANN when RANN available
  use_ann <- (n > 500 || (n * p > 1.5e6)) &&
             requireNamespace("RANN", quietly = TRUE)
  if (requireNamespace("RANN", quietly = TRUE)) {
    expect_true(use_ann)
  } else {
    expect_false(use_ann)
  }
})

test_that("injected near-duplicate is detected in the n>500 regime", {
  skip_on_cran()
  set.seed(42)
  n <- 520
  p <- 10
  X <- matrix(rnorm(n * p), nrow = n)
  # inject a near-duplicate
  X[2, ] <- X[1, ] + rnorm(p, sd = 1e-6)
  df <- data.frame(
    subject = seq_len(n),
    outcome = rbinom(n, 1, 0.5),
    X
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, progress = FALSE)
  fit <- fit_resample(df, outcome = "outcome", splits = splits,
                      learner = "glmnet",
                      preprocess = list(
                        impute = list(method = "median"),
                        normalize = list(method = "zscore")
                      ))
  audit <- audit_leakage(fit, B = 5,
                         X_ref = as.matrix(df[, paste0("X", 1:p)]),
                         sim_threshold = 0.99)
  expect_true(nrow(audit@duplicates) > 0)
})

test_that("audit_report renders without error with expanded template", {
  skip_if_not_installed("rmarkdown")
  skip_on_cran()
  skip_if_not(rmarkdown::pandoc_available(), "pandoc not available")
  set.seed(42)
  df <- data.frame(
    subject = rep(1:30, each = 3),
    outcome = rep(c(0, 1), length.out = 90),
    x1 = rnorm(90),
    x2 = rnorm(90)
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, progress = FALSE)
  fit <- fit_resample(df, outcome = "outcome", splits = splits,
                      learner = "glmnet",
                      preprocess = list(
                        impute = list(method = "median"),
                        normalize = list(method = "zscore")
                      ))
  audit <- audit_leakage(fit, B = 5)
  outfile <- tempfile(fileext = ".html")
  expect_no_error(audit_report(audit, output_file = outfile))
  expect_true(file.exists(outfile))
  unlink(outfile)
})

test_that("calibration_summary works for binomial tasks", {
  set.seed(42)
  df <- data.frame(
    subject = rep(1:30, each = 3),
    outcome = rep(c(0, 1), length.out = 90),
    x1 = rnorm(90),
    x2 = rnorm(90)
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, progress = FALSE)
  fit <- fit_resample(df, outcome = "outcome", splits = splits,
                      learner = "glmnet",
                      preprocess = list(
                        impute = list(method = "median"),
                        normalize = list(method = "zscore")
                      ))
  cal <- tryCatch(calibration_summary(fit), error = function(e) NULL)
  # calibration_summary may return NULL or a non-data.frame depending on
  # predictions structure — just verify it doesn't error
  if (!is.null(cal) && is.data.frame(cal)) {
    expect_true(nrow(cal) >= 0)
  } else {
    expect_true(TRUE)  # graceful fallback
  }
})

test_that("audit helper utilities behave as expected", {
  A <- matrix(c(1, 0, 0, 1), nrow = 2, byrow = TRUE)
  S <- bioLeak:::.cosine_sim_block(A)
  expect_equal(dim(S), c(2, 2))
  expect_equal(diag(S), c(1, 1))

  B <- matrix(c(3, 4, 0, 0), nrow = 2, byrow = TRUE)
  Bn <- bioLeak:::.row_l2_normalize(B)
  expect_equal(round(rowSums(Bn * Bn), 6), c(1, 0))

  C <- matrix(c(1, 2, 3, 4), nrow = 2, byrow = TRUE)
  Cc <- bioLeak:::.row_center(C)
  expect_equal(rowMeans(Cc), c(0, 0))
})

test_that("chi-squared association returns NA for small or sparse tables", {
  tab <- matrix(c(1, 0), nrow = 1)
  res <- bioLeak:::.chisq_assoc(tab)
  expect_true(all(is.na(unlist(res))))

  tab2 <- matrix(c(5, 2, 1, 3), nrow = 2)
  res2 <- bioLeak:::.chisq_assoc(tab2)
  expect_true(is.finite(res2$stat))
  expect_true(is.finite(res2$pval))
})

test_that("metric_value computes supported metrics", {
  truth <- factor(c(0, 1, 0, 1))
  pred <- c(0.1, 0.9, 0.2, 0.8)
  rmse <- bioLeak:::.metric_value("rmse", "gaussian", c(1, 2, 3), c(1, 2, 4))
  expect_equal(rmse, sqrt(1 / 3))

  auc <- bioLeak:::.metric_value("auc", "binomial", truth, pred)
  expect_true(is.finite(auc))

  pr <- bioLeak:::.metric_value("pr_auc", "binomial", truth, pred)
  if (requireNamespace("PRROC", quietly = TRUE)) {
    expect_true(is.finite(pr))
  } else {
    expect_true(is.na(pr))
  }

  cidx <- bioLeak:::.metric_value("cindex", "gaussian", c(1, 2, 3), c(0.2, 0.3, 0.1))
  expect_true(is.finite(cidx))
  expect_true(is.na(bioLeak:::.metric_value("unknown", "gaussian", 1:3, 1:3)))
})

test_that("coerce_truth_like preserves types", {
  expect_true(is.factor(bioLeak:::.coerce_truth_like(factor(c("a", "b")), c("b", "a"))))
  expect_true(is.logical(bioLeak:::.coerce_truth_like(c(TRUE, FALSE), c(1, 0))))
  expect_true(is.numeric(bioLeak:::.coerce_truth_like(c(1, 2), c("3", "4"))))
})

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE,
  eval = TRUE
)

## ----example-data-------------------------------------------------------------
library(bioLeak)

set.seed(123)
n <- 160
subject <- rep(seq_len(40), each = 4)
batch <- sample(paste0("B", 1:6), n, replace = TRUE)
study <- sample(paste0("S", 1:4), n, replace = TRUE)
time <- seq_len(n)

x1 <- rnorm(n)
x2 <- rnorm(n)
x3 <- rnorm(n)
linpred <- 0.7 * x1 - 0.4 * x2 + 0.2 * x3 + rnorm(n, sd = 0.5)
p <- stats::plogis(linpred)
outcome <- factor(ifelse(runif(n) < p, "case", "control"),
                  levels = c("control", "case"))

df <- data.frame(
  subject = subject,
  batch = batch,
  study = study,
  time = time,
  outcome = outcome,
  x1 = x1,
  x2 = x2,
  x3 = x3
)

df_leaky <- within(df, {
  leak_subject <- ave(as.numeric(outcome == "case"), subject, FUN = mean)
  leak_batch <- ave(as.numeric(outcome == "case"), batch, FUN = mean)
  leak_global <- mean(as.numeric(outcome == "case"))
})

df_time <- df
df_time$leak_future <- c(as.numeric(df_time$outcome == "case")[-1], 0)
predictors <- c("x1", "x2", "x3")


# Example data (first 6 rows)
head(df)

# Outcome class counts
as.data.frame(table(df$outcome))

## ----leaky-splits-------------------------------------------------------------
leaky_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "row_id",
  v = 5,
  repeats = 1,
  stratify = TRUE
)

cat("Leaky splits summary (sample-wise CV):\n")
leaky_splits

## ----safe-splits--------------------------------------------------------------
safe_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "subject",
  v = 5,
  repeats = 1,
  stratify = TRUE,
  seed = 10
)

cat("Leakage-safe splits summary (subject-grouped CV):\n")
safe_splits

## ----splits-other-modes-------------------------------------------------------
batch_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "batch_blocked",
  batch = "batch",
  v = 4,
  stratify = TRUE
)

cat("Batch-blocked splits summary:\n")
batch_splits

study_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "study_loocv",
  study = "study"
)

cat("Study leave-one-out splits summary:\n")
study_splits

time_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "time_series",
  time = "time",
  v = 4,
  horizon = 2
)

cat("Time-series splits summary:\n")
time_splits

nested_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "subject",
  v = 3,
  nested = TRUE,
  stratify = TRUE
)

cat("Nested CV splits summary:\n")
nested_splits

## ----combined-splits----------------------------------------------------------
# For combined mode, constraint-axis levels should not span all folds.
# Here each subject belongs to exactly one site, so site exclusion only
# removes training samples from the same site as the test subjects.
df_comb <- df
df_comb$site <- paste0("site", rep(1:8, each = 5)[df_comb$subject])

combined_splits <- make_split_plan(
  df_comb,
  outcome = "outcome",
  mode = "combined",
  constraints = list(
    list(type = "subject", col = "subject"),
    list(type = "batch",   col = "site")
  ),
  v = 4,
  stratify = TRUE,
  seed = 42
)

cat("Combined (N-axis) splits summary:\n")
combined_splits

## ----combined-overlap-check---------------------------------------------------
# Verify zero overlap on all constraint axes
overlap_result <- check_split_overlap(combined_splits)
overlap_result

## ----check-split-overlap------------------------------------------------------
# Run on the subject-grouped splits from earlier
overlap_safe <- check_split_overlap(safe_splits)
overlap_safe

## ----compact-splits-----------------------------------------------------------
# Efficient storage for large N
big_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "subject",
  v = 5,
  compact = TRUE  # <--- Saves memory
)

cat("Compact-mode splits summary:\n")
big_splits
cat(sprintf("Compact storage enabled: %s\n", big_splits@info$compact))

## ----strict-mode-demo, error=TRUE---------------------------------------------
try({
# Enable strict mode temporarily
withr::with_options(list(bioLeak.strict = TRUE), {
  strict_splits <- make_split_plan(
    df,
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 5,
    stratify = TRUE,
    seed = 42
  )
  cat("Strict mode splits completed without error.\n")
})
})

## ----strict-mode-trained-recipe, error=TRUE-----------------------------------
try({
# Strict mode catches a pre-trained recipe
if (requireNamespace("recipes", quietly = TRUE)) {
  rec <- recipes::recipe(outcome ~ ., data = df[, c("outcome", predictors)]) |>
    recipes::step_normalize(recipes::all_numeric_predictors()) |>
    recipes::prep(training = df[, c("outcome", predictors)])

  withr::with_options(list(bioLeak.strict = TRUE), {
    tryCatch(
      fit_resample(
        df, outcome = "outcome",
        splits = safe_splits,
        learner = "glmnet",
        preprocess = rec
      ),
      error = function(e) cat("Strict mode error:", conditionMessage(e), "\n")
    )
  })
}
})

## ----leaky-scaling------------------------------------------------------------
df_leaky_scaled <- df
df_leaky_scaled[predictors] <- scale(df_leaky_scaled[predictors])
scaled_summary <- data.frame(
  feature = predictors,
  mean = colMeans(df_leaky_scaled[predictors]),
  sd = apply(df_leaky_scaled[predictors], 2, stats::sd)
)
scaled_summary$mean <- round(scaled_summary$mean, 3)
scaled_summary$sd <- round(scaled_summary$sd, 3)

# Leaky global scaling: means ~0 and SDs ~1 computed on all samples
scaled_summary

## ----guarded-preprocess-------------------------------------------------------
fold1 <- safe_splits@indices[[1]]
train_x <- df[fold1$train, predictors]
test_x <- df[fold1$test, predictors]

guard <- .guard_fit(
  X = train_x,
  y = df$outcome[fold1$train],
  steps = list(
    impute = list(method = "median", winsor = TRUE),
    normalize = list(method = "zscore"),
    filter = list(var_thresh = 0, iqr_thresh = 0),
    fs = list(method = "none")
  ),
  task = "binomial"
)

train_x_guarded <- predict_guard(guard, train_x)
test_x_guarded <- predict_guard(guard, test_x)

cat("GuardFit object:\n")
guard
cat("\nGuardFit summary:\n")
summary(guard)

# Guarded training data (first 6 rows)
head(train_x_guarded)

# Guarded test data (first 6 rows)
head(test_x_guarded)

## ----guard-ensure-levels------------------------------------------------------
raw_levels <- data.frame(
  site = c("A", "B", "B"),
  status = c("yes", "no", "yes"),
  stringsAsFactors = FALSE
)

level_state <- .guard_ensure_levels(raw_levels)

# Aligned factor data with consistent levels
level_state$data

# Levels map
level_state$levels

## ----leaky-impute-------------------------------------------------------------
train <- data.frame(a = c(1, 2, NA, 4), b = c(NA, 1, 1, 0))
test <- data.frame(a = c(NA, 5), b = c(1, NA))

all_median <- vapply(rbind(train, test),
                     function(col) median(col, na.rm = TRUE),
                     numeric(1))
train_leaky <- as.data.frame(Map(function(col, m) { col[is.na(col)] <- m; col },
                                 train, all_median))
test_leaky <- as.data.frame(Map(function(col, m) { col[is.na(col)] <- m; col },
                                test, all_median))

# Leaky medians computed on train + test
data.frame(feature = names(all_median), median = all_median)

# Leaky-imputed training data
train_leaky

# Leaky-imputed test data
test_leaky

## ----safe-impute--------------------------------------------------------------
imp <- impute_guarded(
  train = train,
  test = test,
  method = "median",
  winsor = FALSE
)

# Guarded-imputed training data
imp$train

# Guarded-imputed test data
imp$test

## ----guard-to-recipe----------------------------------------------------------
if (requireNamespace("recipes", quietly = TRUE)) {
  guard_steps <- list(
    impute    = list(method = "median"),
    normalize = list(method = "zscore")
  )

  rec <- guard_to_recipe(
    steps = guard_steps,
    formula = outcome ~ .,
    training_data = df[, c("outcome", predictors)]
  )

  cat("Converted recipe:\n")
  rec
}

## ----parsnip-spec-------------------------------------------------------------
spec <- parsnip::logistic_reg(mode = "classification") |>
  parsnip::set_engine("glm")

## ----fit-leaky----------------------------------------------------------------
fit_leaky <- fit_resample(
  df_leaky,
  outcome = "outcome",
  splits = leaky_splits,
  learner = spec,
  metrics = c("auc", "accuracy"),
  preprocess = list(
    impute = list(method = "median"),
    normalize = list(method = "zscore"),
    filter = list(var_thresh = 0),
    fs = list(method = "none")
  )
)

cat("Leaky fit summary:\n")
summary(fit_leaky)
metrics_leaky <- as.data.frame(fit_leaky@metric_summary)
num_cols <- vapply(metrics_leaky, is.numeric, logical(1))
metrics_leaky[num_cols] <- lapply(metrics_leaky[num_cols], round, digits = 3)

# Leaky fit: mean and SD of metrics across folds
metrics_leaky

## ----fit-safe-----------------------------------------------------------------
fit_safe <- fit_resample(
  df,
  outcome = "outcome",
  splits = safe_splits,
  learner = spec,
  metrics = c("auc", "accuracy"),
  preprocess = list(
    impute = list(method = "median"),
    normalize = list(method = "zscore"),
    filter = list(var_thresh = 0),
    fs = list(method = "none")
  ),
  positive_class = "case",
  class_weights = c(control = 1, case = 1),
  refit = TRUE
)

cat("Leakage-safe fit summary:\n")
summary(fit_safe)
metrics_safe <- as.data.frame(fit_safe@metric_summary)
num_cols <- vapply(metrics_safe, is.numeric, logical(1))
metrics_safe[num_cols] <- lapply(metrics_safe[num_cols], round, digits = 3)

# Leakage-safe fit: mean and SD of metrics across folds
metrics_safe

# Per-fold metrics (first 6 rows)
head(fit_safe@metrics)

## ----fit-fold-status----------------------------------------------------------
# Fold-level diagnostics for reproducible troubleshooting
head(fit_safe@info$fold_status, 5)

## ----multiclass-example-------------------------------------------------------
if (requireNamespace("ranger", quietly = TRUE)) {
  set.seed(11)
  df_multi <- df
  df_multi$outcome3 <- factor(sample(c("A", "B", "C"),
                                     nrow(df_multi), replace = TRUE))

  multi_splits <- make_split_plan(
    df_multi,
    outcome = "outcome3",
    mode = "subject_grouped",
    group = "subject",
    v = 4,
    stratify = TRUE,
    seed = 11
  )

  fit_multi <- fit_resample(
    df_multi,
    outcome = "outcome3",
    splits = multi_splits,
    learner = "ranger",
    metrics = c("accuracy", "macro_f1", "log_loss"),
    refit = FALSE
  )

  cat("Multiclass fit summary:\n")
  summary(fit_multi)
} else {
  cat("ranger not installed; skipping multiclass example.\n")
}

## ----survival-example---------------------------------------------------------
if (requireNamespace("survival", quietly = TRUE) &&
    requireNamespace("glmnet", quietly = TRUE)) {
  set.seed(12)
  df_surv <- df
  df_surv$time_to_event <- rexp(nrow(df_surv), rate = 0.1)
  df_surv$event <- rbinom(nrow(df_surv), 1, 0.7)

  surv_splits <- make_split_plan(
    df_surv,
    outcome = c("time_to_event", "event"),
    mode = "subject_grouped",
    group = "subject",
    v = 4,
    stratify = FALSE,
    seed = 12
  )

  fit_surv <- fit_resample(
    df_surv,
    outcome = c("time_to_event", "event"),
    splits = surv_splits,
    learner = "glmnet",
    metrics = "cindex",
    refit = FALSE
  )

  cat("Survival fit summary:\n")
  summary(fit_surv)
} else {
  cat("survival or glmnet not installed; skipping survival example.\n")
}

## ----learner-args-------------------------------------------------------------
if (requireNamespace("glmnet", quietly = TRUE)) {
  fit_glmnet <- fit_resample(
    df,
    outcome = "outcome",
    splits = safe_splits,
    learner = "glmnet",
    metrics = "auc",
    learner_args = list(glmnet = list(alpha = 0.5)),
    preprocess = list(
      impute = list(method = "median"),
      normalize = list(method = "zscore"),
      filter = list(var_thresh = 0),
      fs = list(method = "none")
    )
  )
  cat("GLMNET summary with learner-specific arguments:\n")
  summary(fit_glmnet)
} else {
  cat("glmnet not installed; skipping learner_args example.\n")
}

## ----se-example---------------------------------------------------------------
if (requireNamespace("SummarizedExperiment", quietly = TRUE)) {
  se <- SummarizedExperiment::SummarizedExperiment(
    assays = list(counts = t(as.matrix(df[, predictors]))),
    colData = df[, c("subject", "batch", "study", "time", "outcome"), drop = FALSE]
  )

  se_splits <- make_split_plan(
    se,
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 3
  )

  se_fit <- fit_resample(
    se,
    outcome = "outcome",
    splits = se_splits,
    learner = spec,
    metrics = "auc"
  )
  cat("SummarizedExperiment fit summary:\n")
  summary(se_fit)
} else {
  cat("SummarizedExperiment not installed; skipping SE example.\n")
}

## ----tidymodels-interop-------------------------------------------------------
library(bioLeak)
library(parsnip)
library(recipes)
library(yardstick)

set.seed(123)
N <- 60
df <- data.frame(
  subject = factor(rep(paste0("S", 1:20), length.out = N)), 
  outcome = factor(rep(c("ClassA", "ClassB"), length.out = N)),
  x1 = rnorm(N),
  x2 = rnorm(N),
  x3 = rnorm(N)
)

spec <- logistic_reg() |> set_engine("glm")

# Use bioLeak's native split planner to avoid conversion errors
set.seed(13)

# Use make_split_plan instead of rsample::group_vfold_cv
# This creates a subject-grouped CV directly compatible with fit_resample
splits <- make_split_plan(
  df, 
  outcome = "outcome", 
  mode = "subject_grouped", 
  group = "subject", 
  v = 3
)

rec <- recipes::recipe(outcome ~ x1 + x2 + x3, data = df) |>
  recipes::step_impute_median(recipes::all_numeric_predictors()) |>
  recipes::step_normalize(recipes::all_numeric_predictors())

metrics_set <- yardstick::metric_set(yardstick::roc_auc, yardstick::accuracy)

fit_rs <- fit_resample(
  df,
  outcome = "outcome",
  splits = splits,
  learner = spec,
  preprocess = rec,
  metrics = metrics_set,
  refit = FALSE
)

if (exists("as_rsample", where = asNamespace("bioLeak"), mode = "function")) {
    rs_export <- as_rsample(fit_rs@splits, data = df)
    print(rs_export)
}

## ----workflow-example---------------------------------------------------------
if (requireNamespace("workflows", quietly = TRUE)) {
  wf <- workflows::workflow() |>
    workflows::add_model(spec) |>
    workflows::add_formula(outcome ~ x1 + x2 + x3)

  fit_wf <- fit_resample(
    df,
    outcome = "outcome",
    splits = splits,
    learner = wf,
    metrics = "auc",
    refit = FALSE
  )

  cat("Workflow fit summary:\n")
  summary(fit_wf)
} else {
  cat("workflows not installed; skipping workflow example.\n")
}

## ----tune-example-------------------------------------------------------------
if (requireNamespace("parsnip", quietly = TRUE) &&
    requireNamespace("recipes", quietly = TRUE) &&
    requireNamespace("tune", quietly = TRUE) &&
    requireNamespace("glmnet", quietly = TRUE)) {
  # --- 1. Create Data ---
  set.seed(123)
  N <- 60
  df <- data.frame(
    subject = factor(rep(paste0("S", 1:20), length.out = N)),
    outcome = factor(sample(c("ClassA", "ClassB"), N, replace = TRUE)),
    x1 = rnorm(N),
    x2 = rnorm(N),
    x3 = rnorm(N)
  )

  # --- 2. Generate Nested Splits ---
  set.seed(1)
  nested_splits <- make_split_plan(
    df,
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 3,
    nested = TRUE
  )

  # --- 3. Define Recipe & Model ---
  rec <- recipes::recipe(outcome ~ x1 + x2 + x3, data = df) |>
    recipes::step_impute_median(recipes::all_numeric_predictors()) |>
    recipes::step_normalize(recipes::all_numeric_predictors())

  spec_tune <- parsnip::logistic_reg(
    penalty = tune::tune(),
    mixture = 1,
    mode = "classification"
  ) |>
    parsnip::set_engine("glmnet")

  # --- 4. Run Tuning ---
  tuned <- tune_resample(
    df,
    outcome = "outcome",
    splits = nested_splits,
    learner = spec_tune,
    preprocess = rec,
    inner_v = 2,
    grid = 3,
    metrics = c("auc", "accuracy"),
    selection = "one_std_err",
    refit = TRUE,
    seed = 14
  )

  summary(tuned)
} else {
  cat("parsnip/recipes/tune/glmnet not installed; skipping nested tuning example.\n")
}

## ----tune-diagnostics---------------------------------------------------------
if (exists("tuned")) {
  # Fold-level status from the outer loop
  head(tuned$fold_status, 5)

  # Final tuned model is available when refit = TRUE
  is.null(tuned$final_model)
} else {
  cat("Nested tuning object not available (dependencies missing).\n")
}

## ----fit-xgboost--------------------------------------------------------------
if (requireNamespace("parsnip", quietly = TRUE) &&
    requireNamespace("xgboost", quietly = TRUE) &&
    requireNamespace("recipes", quietly = TRUE)) {
  
  # 1. Define the model spec
  xgb_spec <- parsnip::boost_tree(
    mode = "classification",
    trees = 100,
    tree_depth = 6,
    learn_rate = 0.01
  ) |>
    parsnip::set_engine("xgboost")
  
  # 2. Define a recipe on data without 'subject' because split metadata
  # columns are excluded from predictors by fit_resample().
  df_for_rec <- df[, !names(df) %in% "subject"]
  
  rec_xgb <- recipes::recipe(outcome ~ ., data = df_for_rec) |>
    recipes::step_dummy(recipes::all_nominal_predictors()) |>
    recipes::step_impute_median(recipes::all_numeric_predictors())
  # Note: No need for step_rm(subject) because it's already gone!
  
  # 3. Fit
  fit_xgb <- fit_resample(
    df,
    outcome = "outcome",
    splits = splits,
    learner = xgb_spec,
    metrics = "auc",
    preprocess = rec_xgb 
  )
  
  cat("XGBoost parsnip fit summary:\n")
  print(summary(fit_xgb))
} else {
  cat("parsnip/xgboost/recipes not installed.\n")
}

## ----custom-learners----------------------------------------------------------
custom_learners <- list(
  glm = list(
    fit = function(x, y, task, weights, ...) {
      df_fit <- data.frame(y = y, x, check.names = FALSE)
      stats::glm(y ~ ., data = df_fit,
                 family = stats::binomial(), weights = weights)
    },
    predict = function(object, newdata, task, ...) {
      as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
                                type = "response"))
    }
  )
)

cat("Custom learner names:\n")
names(custom_learners)
cat("Custom learner components (fit/predict):\n")
lapply(custom_learners, names)

## ----plot-fold-balance--------------------------------------------------------
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_fold_balance(fit_safe)
} else {
  cat("ggplot2 not installed; skipping fold balance plot.\n")
}

## ----plot-calibration---------------------------------------------------------
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_calibration(fit_safe, bins = 10)
} else {
  cat("ggplot2 not installed; skipping calibration plot.\n")
}

## ----plot-confounder-sensitivity----------------------------------------------
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_confounder_sensitivity(fit_safe, confounders = c("batch", "study"),
                              metric = "auc", min_n = 3)
} else {
  cat("ggplot2 not installed; skipping confounder sensitivity plot.\n")
}

## ----diagnostics-tables-------------------------------------------------------
cal <- calibration_summary(fit_safe, bins = 10, min_bin_n = 5)
conf_tbl <- confounder_sensitivity(fit_safe,
                                   confounders = c("batch", "study"),
                                   metric = "auc",
                                   min_n = 3)

# Calibration metrics
cal$metrics

# Confounder sensitivity table (first 6 rows)
head(conf_tbl)

## ----plot-overlap-------------------------------------------------------------
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_overlap_checks(fit_leaky, column = "subject")
  plot_overlap_checks(fit_safe, column = "subject")
} else {
  cat("ggplot2 not installed; skipping overlap plots.\n")
}

## ----audit-leakage------------------------------------------------------------
# Use df_leaky (the original 160-row data) so X_ref aligns with fit_safe's predictions.
# df may have been redefined to a smaller dataset in the tidymodels section above.
X_ref <- df_leaky[, predictors]
X_ref[c(1, 5), ] <- X_ref[1, ]

audit <- audit_leakage(
  fit_safe,
  metric = "auc",
  B = 20,
  perm_stratify = TRUE,
  batch_cols = c("batch", "study"),
  X_ref = X_ref,
  sim_method = "cosine",
  sim_threshold = 0.995,
  return_perm = TRUE
)

cat("Leakage audit summary:\n")
summary(audit)
if (!is.null(audit@permutation_gap) && nrow(audit@permutation_gap) > 0) {
  # Permutation significance results
  audit@permutation_gap
}
if (!is.null(audit@batch_assoc) && nrow(audit@batch_assoc) > 0) {
  # Batch/study association with folds (Cramer's V)
  audit@batch_assoc
} else {
  cat("No batch or study associations detected.\n")
}
if (!is.null(audit@target_assoc) && nrow(audit@target_assoc) > 0) {
  # Top features by target association score
  head(audit@target_assoc)
} else {
  cat("No target leakage scan results available.\n")
}
if (!is.null(audit@duplicates) && nrow(audit@duplicates) > 0) {
  # Top duplicate/near-duplicate pairs by similarity
  head(audit@duplicates)
} else {
  cat("No near-duplicates detected.\n")
}

## ----mechanism-summary--------------------------------------------------------
mech <- audit@info$mechanism_summary
if (is.data.frame(mech) && nrow(mech) > 0) {
  mech
} else {
  cat("No mechanism summary available.\n")
}

## ----plot-perm----------------------------------------------------------------
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_perm_distribution(audit)
} else {
  cat("ggplot2 not installed; skipping permutation plot.\n")
}

## ----audit-by-learner---------------------------------------------------------
if (requireNamespace("ranger", quietly = TRUE) && 
    requireNamespace("parsnip", quietly = TRUE)) {
    
    # 1. Define specs
    # Standard GLM (no tuning)
    spec_glm <- parsnip::logistic_reg() |> 
        parsnip::set_engine("glm")
    
    # Random Forest
    spec_rf <- parsnip::rand_forest(
        mode = "classification",
        trees = 100
    ) |>
        parsnip::set_engine("ranger")
    
    # 2. Fit using the current split object
    fit_multi <- fit_resample(
        df,
        outcome = "outcome",
        splits = splits,
        learner = list(glm = spec_glm, rf = spec_rf),
        metrics = "auc"
    )
    
    # 3. Run the audit
    audits <- audit_leakage_by_learner(fit_multi, metric = "auc", B = 20)
    cat("Per-learner audit summary:\n")
    print(audits)
    
} else {
    cat("ranger/parsnip not installed.\n")
}

## ----audit-report, eval = FALSE-----------------------------------------------
# if (requireNamespace("rmarkdown", quietly = TRUE) && rmarkdown::pandoc_available()) {
#   report_path <- audit_report(audit, output_dir = ".")
#   cat("HTML report written to:\n", report_path, "\n")
# } else {
#   cat("rmarkdown or pandoc not available; skipping audit report rendering.\n")
# }

## ----time-series-fit----------------------------------------------------------
time_splits <- make_split_plan(
  df_time,
  outcome = "outcome",
  mode = "time_series",
  time = "time",
  v = 4,
  horizon = 1
)

cat("Time-series splits summary:\n")
time_splits

fit_time_leaky <- fit_resample(
  df_time,
  outcome = "outcome",
  splits = time_splits,
  learner = spec,
  metrics = "auc"
)

cat("Time-series leaky fit summary:\n")
summary(fit_time_leaky)

## ----time-series-safe---------------------------------------------------------
df_time_safe <- df_time
df_time_safe$leak_future <- NULL

fit_time_safe <- fit_resample(
  df_time_safe,
  outcome = "outcome",
  splits = time_splits,
  learner = spec,
  metrics = "auc"
)

cat("Time-series safe fit summary:\n")
summary(fit_time_safe)

audit_time <- audit_leakage(
  fit_time_safe,
  metric = "auc",
  B = 20,
  time_block = "stationary",
  block_len = 5
)

cat("Time-series leakage audit summary:\n")
summary(audit_time)
if (!is.null(audit_time@permutation_gap) && nrow(audit_time@permutation_gap) > 0) {
  # Time-series permutation significance results
  audit_time@permutation_gap
}

if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_time_acf(fit_time_safe, lag.max = 20)
} else {
  cat("ggplot2 not installed; skipping ACF plot.\n")
}

## ----cv-ci--------------------------------------------------------------------
# Standard 95% CI from the leakage-safe fit
ci_std <- cv_ci(fit_safe@metrics, level = 0.95, method = "normal")
cat("Standard 95% CI:\n")
print(ci_std)

# Nadeau-Bengio corrected CI
# For K-fold CV: n_train ≈ (K-1)/K × n, n_test ≈ n/K
K_folds  <- length(unique(fit_safe@metrics$fold))
n_total  <- nrow(fit_safe@splits@info$coldata)

ci_nb <- cv_ci(
  fit_safe@metrics,
  level   = 0.95,
  method  = "nadeau_bengio",
  n_train = round(n_total * (K_folds - 1L) / K_folds),
  n_test  = round(n_total / K_folds)
)
cat("Nadeau-Bengio corrected 95% CI:\n")
print(ci_nb)

## ----delta-lsi-data-----------------------------------------------------------
set.seed(100)
n_d    <- 120L
subj_d <- rep(paste0("P", seq_len(30L)), each = 4L)   # 30 subjects, 4 obs each

df_dlsi_guarded <- data.frame(
  subject = subj_d,
  outcome = factor(sample(c("case", "control"), n_d, replace = TRUE)),
  x1      = rnorm(n_d),
  x2      = rnorm(n_d)
)

# Leaky feature: subject-level mean of outcome
# With subject-grouped splits, test subjects' leak value encodes their outcome
df_dlsi_naive <- df_dlsi_guarded
df_dlsi_naive$leak <- ave(
  as.numeric(df_dlsi_guarded$outcome == "case"),
  subj_d, FUN = mean
)

# Shared splits: 5-fold × 5 repeats → R_eff = 5 (tier C: point + p-value)
splits_dlsi <- make_split_plan(
  df_dlsi_guarded,
  outcome = "outcome",
  mode    = "subject_grouped",
  group   = "subject",
  v       = 5L,
  repeats = 5L
)

## ----delta-lsi-fits-----------------------------------------------------------
# Reuse the parsnip spec (base R glm) defined earlier
# Naive fit: uses leaky feature (leak encodes test-subject outcome label)
fit_dlsi_naive <- fit_resample(
  df_dlsi_naive,
  outcome = "outcome",
  splits  = splits_dlsi,
  learner = spec,
  metrics = "auc",
  seed    = 1L
)

# Guarded fit: same splits, clean features only
fit_dlsi_guarded <- fit_resample(
  df_dlsi_guarded,
  outcome = "outcome",
  splits  = splits_dlsi,
  learner = spec,
  metrics = "auc",
  seed    = 1L
)

cat("Naive   AUC:", round(mean(fit_dlsi_naive@metrics$auc,   na.rm = TRUE), 3), "\n")
cat("Guarded AUC:", round(mean(fit_dlsi_guarded@metrics$auc, na.rm = TRUE), 3), "\n")

## ----delta-lsi-run------------------------------------------------------------
result_dlsi <- delta_lsi(
  fit_leaky   = fit_dlsi_naive,
  fit_guarded = fit_dlsi_guarded,
  metric      = "auc",
  M_boot      = 300L,
  M_flip      = 300L,
  seed        = 1L
)

summary(result_dlsi)

## ----delta-lsi-tier-----------------------------------------------------------
cat("Tier:          ", result_dlsi@tier, "\n")
cat("R_eff:         ", result_dlsi@R_eff, "\n")
cat("inference_ok:  ", result_dlsi@inference_ok, "\n")

## ----delta-lsi-components-----------------------------------------------------
# delta_metric: arithmetic mean of {Δ_r}
cat("delta_metric:  ", round(result_dlsi@delta_metric, 4), "\n")
# delta_lsi: Huber M-estimate of {Δ_r} (k=1.345, fixed MAD scale)
cat("delta_lsi:     ", round(result_dlsi@delta_lsi,    4), "\n")
# delta_lsi_ci: 95% BCa interval for delta_lsi (NA below tier B)
cat("95% BCa CI:    [",
    paste(round(result_dlsi@delta_lsi_ci, 4), collapse = ", "), "]\n")
# p_value: sign-flip randomization p-value (NA below tier C or if unpaired)
cat("p-value:       ",
    if (is.na(result_dlsi@p_value)) "NA (tier D or unpaired)"
    else format(result_dlsi@p_value, digits = 3), "\n")

## ----delta-lsi-unpaired, warning=TRUE-----------------------------------------
# Unpaired: naive uses sample-wise CV, guarded uses subject-grouped splits
splits_rand <- make_split_plan(
  df_dlsi_naive,
  outcome = "outcome",
  mode    = "subject_grouped",
  group   = "row_id",
  v       = 5L,
  repeats = 5L
)

fit_naive_rand <- fit_resample(
  df_dlsi_naive,
  outcome = "outcome",
  splits  = splits_rand,
  learner = spec,
  metrics = "auc",
  seed    = 1L
)

r_unpaired <- suppressWarnings(
  delta_lsi(fit_naive_rand, fit_dlsi_guarded, metric = "auc", seed = 1L)
)

cat(sprintf("Paired:        %s\n", r_unpaired@info$paired))
cat(sprintf("R_eff:         %d  (0 = unpaired, inference disabled)\n", r_unpaired@R_eff))
cat(sprintf("delta_metric:  %.4f  (naive raw mean - guarded raw mean)\n",
            r_unpaired@delta_metric))
cat(sprintf("p_value:       %s\n",
            if (is.na(r_unpaired@p_value)) "NA" else r_unpaired@p_value))

## ----delta-lsi-direction------------------------------------------------------
r_hib_true  <- suppressWarnings(delta_lsi(
  fit_dlsi_naive, fit_dlsi_guarded,
  metric           = "auc",
  higher_is_better = TRUE,
  seed             = 1L
))
r_hib_false <- suppressWarnings(delta_lsi(
  fit_dlsi_naive, fit_dlsi_guarded,
  metric           = "auc",
  higher_is_better = FALSE,
  seed             = 1L
))

cat(sprintf("hib = TRUE:   delta_metric = %+.4f  (positive = naive inflated)\n",
            r_hib_true@delta_metric))
cat(sprintf("hib = FALSE:  delta_metric = %+.4f  (sign flipped)\n",
            r_hib_false@delta_metric))

## ----delta-lsi-hib-slot-------------------------------------------------------
cat("higher_is_better used:", result_dlsi@info$higher_is_better, "\n")

## ----delta-lsi-slots----------------------------------------------------------
# Per-repeat summaries for custom plotting
head(result_dlsi@repeats_naive,   3)
head(result_dlsi@repeats_guarded, 3)

# Per-fold diagnostics
head(result_dlsi@folds_naive,   3)
head(result_dlsi@folds_guarded, 3)

# Metadata
str(result_dlsi@info)

## ----delta-lsi-audit-connect--------------------------------------------------
# Step 1: detect and characterise leakage
audit_naive_dlsi <- audit_leakage(fit_dlsi_naive, metric = "auc", B = 20L)
cat("Naive audit summary:\n")
summary(audit_naive_dlsi)

# Step 2: confirmed guards reduced inflation; check residual DLSI
cat("\ndelta_metric (naive vs guarded):", round(result_dlsi@delta_metric, 4), "\n")
if (!is.na(result_dlsi@p_value)) {
  cat("Sign-flip p-value:              ", format(result_dlsi@p_value, digits = 3), "\n")
}

## ----parallel-setup, eval=FALSE-----------------------------------------------
# library(future)
# 
# # Use multiple cores (works on all OS)
# plan(multisession, workers = 4)
# 
# # Run a heavy simulation
# sim <- simulate_leakage_suite(..., parallel = TRUE)
# 
# # Parallel folds or audits
# # fit_resample(..., parallel = TRUE)
# # audit_leakage(..., parallel = TRUE)
# # audit_leakage_by_learner(..., parallel_learners = TRUE)
# 
# # Return to sequential processing
# plan(sequential)

## ----simulate-suite-----------------------------------------------------------
if (requireNamespace("glmnet", quietly = TRUE)) {
  sim <- simulate_leakage_suite(
    n = 80,
    p = 6,
    mode = "subject_grouped",
    learner = "glmnet",
    leakage = "subject_overlap",
    seeds = 1:2,
    B = 20,
    parallel = FALSE,
    signal_strength = 1
  )
  
  # Simulation results (first 6 rows)
  head(sim)
} else {
  cat("glmnet not installed; skipping simulation suite example.\n")
}

# Diagnostic plotting helpers --------------------------------------------------

#' Plot permutation distribution for a LeakAudit object
#'
#' Visualizes the label-permutation metric distribution and marks the observed
#' and permuted-mean values to help assess leakage signals. Requires ggplot2.
#'
#' @param audit LeakAudit.
#' @return A list containing the observed value, permuted mean, permutation values,
#'   and a ggplot object.
#' @examples
#' if (requireNamespace("ggplot2", quietly = TRUE)) {
#'   set.seed(42)
#'   df <- data.frame(
#'     subject = rep(1:15, each = 2),
#'     outcome = factor(rep(c(0, 1), 15)),
#'     x1 = rnorm(30),
#'     x2 = rnorm(30)
#'   )
#'   splits <- make_split_plan(df, outcome = "outcome",
#'                             mode = "subject_grouped", group = "subject",
#'                             v = 3, progress = FALSE)
#'   custom <- list(
#'     glm = list(
#'       fit = function(x, y, task, weights, ...) {
#'         stats::glm(y ~ ., data = as.data.frame(x),
#'                    family = stats::binomial(), weights = weights)
#'       },
#'       predict = function(object, newdata, task, ...) {
#'         as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
#'                                   type = "response"))
#'       }
#'     )
#'   )
#'   fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                       learner = "glm", custom_learners = custom,
#'                       metrics = "auc", refit = FALSE, seed = 1)
#'   audit <- audit_leakage(fit, metric = "auc", B = 20)
#'   plot_perm_distribution(audit)
#' }
#'
#' @export
plot_perm_distribution <- function(audit) {
  stopifnot(inherits(audit, "LeakAudit"))
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package 'ggplot2' is required for plotting. Install it to use plot_perm_distribution().",
         call. = FALSE)
  }
  perm <- audit@perm_values
  perm <- perm[is.finite(perm)]
  if (!length(perm)) {
    stop("No finite permutation values available for plotting.", call. = FALSE)
  }
  obs <- audit@permutation_gap$metric_obs
  perm_mean <- mean(perm, na.rm = TRUE)

  n_bins <- max(grDevices::nclass.FD(perm), 1L)
  brks <- seq(min(perm), max(perm), length.out = n_bins + 1L)
  hist_obj <- graphics::hist(perm, breaks = brks, plot = FALSE)
  df <- data.frame(mid = hist_obj$mids, count = hist_obj$counts)
  bin_width <- if (length(brks) > 1L) diff(brks)[1] else 1
  line_df <- data.frame(
    value = c(obs, perm_mean),
    type = factor(c("Observed", "Permuted mean"), levels = c("Observed", "Permuted mean"))
  )

  p <- ggplot2::ggplot(df, ggplot2::aes(x = mid, y = count)) +
    ggplot2::geom_col(width = bin_width, fill = "grey80", color = "white") +
    ggplot2::geom_vline(data = line_df,
                        ggplot2::aes(xintercept = value, color = type, linetype = type),
                        linewidth = 1) +
    ggplot2::labs(title = "Permutation distribution", x = "Metric", y = "Count",
                  color = NULL, linetype = NULL) +
    ggplot2::scale_color_manual(values = c("Observed" = "red", "Permuted mean" = "blue")) +
    ggplot2::scale_linetype_manual(values = c("Observed" = "solid",
                                              "Permuted mean" = "dashed")) +
    ggplot2::theme_minimal() +
    ggplot2::theme(legend.position = "top")

  if (interactive()) print(p)
  invisible(list(
    observed = obs,
    permuted_mean = perm_mean,
    perm_values = perm,
    plot = p
  ))
}

#' Plot fold balance of class counts per fold
#'
#' Displays a bar chart of class counts per fold. For binomial tasks, it also
#' overlays the positive proportion to diagnose stratification issues. The
#' positive class is taken from \code{fit@info$positive_class} when available;
#' otherwise the second factor level is used. For multiclass tasks, the plot
#' shows per-class counts without a proportion line. Only available for
#' classification tasks. Requires ggplot2.
#'
#' @param fit LeakFit.
#' @return A list containing the fold summary, positive class (if binomial),
#'   and a ggplot object.
#' @examples
#' if (requireNamespace("ggplot2", quietly = TRUE)) {
#'   set.seed(42)
#'   df <- data.frame(
#'     subject = rep(1:15, each = 2),
#'     outcome = factor(rep(c(0, 1), 15)),
#'     x1 = rnorm(30),
#'     x2 = rnorm(30)
#'   )
#'   splits <- make_split_plan(df, outcome = "outcome",
#'                             mode = "subject_grouped", group = "subject",
#'                             v = 3, progress = FALSE)
#'   custom <- list(
#'     glm = list(
#'       fit = function(x, y, task, weights, ...) {
#'         stats::glm(y ~ ., data = as.data.frame(x),
#'                    family = stats::binomial(), weights = weights)
#'       },
#'       predict = function(object, newdata, task, ...) {
#'         as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
#'                                   type = "response"))
#'       }
#'     )
#'   )
#'   fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                       learner = "glm", custom_learners = custom,
#'                       metrics = "auc", refit = FALSE, seed = 1)
#'   plot_fold_balance(fit)
#' }
#'
#' @export
plot_fold_balance <- function(fit) {
  stopifnot(inherits(fit, "LeakFit"))
  if (!fit@task %in% c("binomial", "multiclass")) {
    stop("plot_fold_balance is only available for classification tasks.", call. = FALSE)
  }
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package 'ggplot2' is required for plotting. Install it to use plot_fold_balance().",
         call. = FALSE)
  }
  if (fit@task == "multiclass") {
    class_levels <- NULL
    for (df in fit@predictions) {
      if (is.factor(df$truth)) {
        class_levels <- levels(df$truth)
        break
      }
    }
    if (is.null(class_levels)) {
      class_levels <- sort(unique(unlist(lapply(fit@predictions, function(df) {
        as.character(df$truth)
      }))))
    }
    if (!length(class_levels)) {
      stop("No class labels available for plotting.", call. = FALSE)
    }
    tab <- lapply(seq_along(fit@predictions), function(i) {
      df <- fit@predictions[[i]]
      y <- factor(as.character(df$truth), levels = class_levels)
      counts <- table(y)
      data.frame(
        fold = i,
        class = factor(names(counts), levels = class_levels),
        count = as.numeric(counts),
        stringsAsFactors = FALSE
      )
    })
    tab <- do.call(rbind, tab)
    p <- ggplot2::ggplot(tab, ggplot2::aes(x = fold, y = count, fill = class)) +
      ggplot2::geom_col(width = 0.7, color = "white") +
      ggplot2::scale_x_continuous(breaks = unique(tab$fold)) +
      ggplot2::labs(title = "Fold class balance", x = "Fold", y = "Count", fill = "Class") +
      ggplot2::theme_minimal() +
      ggplot2::theme(legend.position = "top")
    if (interactive()) print(p)
    return(invisible(list(
      fold_summary = tab,
      positive_class = NA_character_,
      plot = p
    )))
  }
  pos_class <- fit@info$positive_class
  if (length(pos_class) != 1L) pos_class <- NULL
  pos_class <- if (!is.null(pos_class)) as.character(pos_class) else NULL
  if (!is.null(pos_class) && (is.na(pos_class) || !nzchar(pos_class))) {
    pos_class <- NULL
  }
  if (is.null(pos_class)) {
    for (df in fit@predictions) {
      if (is.factor(df$truth) && nlevels(df$truth) >= 2) {
        pos_class <- levels(df$truth)[2]
        break
      }
    }
  }
  resolve_pos_label <- function(y, pos_label) {
    if (!is.null(pos_label)) return(pos_label)
    if (is.factor(y) && nlevels(y) >= 2) return(levels(y)[2])
    "1"
  }
  tab <- lapply(seq_along(fit@predictions), function(i) {
    df <- fit@predictions[[i]]
    y <- df$truth
    pos_label <- resolve_pos_label(y, pos_class)
    y_chr <- as.character(y)
    is_pos <- y_chr == pos_label
    valid <- !is.na(is_pos)
    data.frame(fold = i,
               positives = sum(is_pos[valid]),
               negatives = sum(!is_pos[valid]))
  })
  tab <- do.call(rbind, tab)
  totals <- tab$positives + tab$negatives
  tab$prop_pos <- ifelse(totals > 0, tab$positives / totals, NA_real_)
  pos_legend <- if (!is.null(pos_class)) {
    paste0("Positives (", pos_class, ")")
  } else {
    "Positives"
  }
  df_counts <- data.frame(
    fold = rep(tab$fold, times = 2),
    class = factor(
      rep(c(pos_legend, "Negatives"), each = nrow(tab)),
      levels = c(pos_legend, "Negatives")
    ),
    count = c(tab$positives, tab$negatives)
  )
  max_count <- max(df_counts$count, na.rm = TRUE)
  if (!is.finite(max_count) || max_count <= 0) max_count <- 1
  prop_df <- data.frame(
    fold = tab$fold,
    prop_pos = tab$prop_pos,
    prop_scaled = tab$prop_pos * max_count
  )
  prop_df <- prop_df[is.finite(prop_df$prop_scaled), , drop = FALSE]
  prop_df$series <- "Positive proportion"

  p <- ggplot2::ggplot(df_counts, ggplot2::aes(x = fold, y = count, fill = class)) +
    ggplot2::geom_col(position = ggplot2::position_dodge(width = 0.8),
                      width = 0.7, color = "white") +
    ggplot2::scale_fill_manual(values = setNames(c("steelblue", "tan"),
                                                 c(pos_legend, "Negatives"))) +
    ggplot2::scale_x_continuous(breaks = tab$fold) +
    ggplot2::labs(title = "Fold class balance", x = "Fold", y = "Count", fill = NULL)
  if (nrow(prop_df) > 0) {
    p <- p +
      ggplot2::geom_line(data = prop_df,
                         ggplot2::aes(x = fold, y = prop_scaled,
                                      color = series, linetype = series),
                         inherit.aes = FALSE) +
      ggplot2::geom_point(data = prop_df,
                          ggplot2::aes(x = fold, y = prop_scaled, color = series, shape = series),
                          inherit.aes = FALSE) +
      ggplot2::scale_color_manual(values = c("Positive proportion" = "blue")) +
      ggplot2::scale_linetype_manual(values = c("Positive proportion" = "dashed")) +
      ggplot2::scale_shape_manual(values = c("Positive proportion" = 17)) +
      ggplot2::scale_y_continuous(
        sec.axis = ggplot2::sec_axis(~ . / max_count,
                                     name = "Positive proportion",
                                     labels = function(x) sprintf("%s%%", round(x * 100)))
      )
  }
  p <- p +
    ggplot2::theme_minimal() +
    ggplot2::theme(legend.position = "top") +
    ggplot2::guides(fill = ggplot2::guide_legend(order = 1),
                    color = ggplot2::guide_legend(order = 2),
                    linetype = ggplot2::guide_legend(order = 2),
                    shape = ggplot2::guide_legend(order = 2))
  if (interactive()) print(p)
  invisible(list(
    fold_summary = tab,
    positive_class = pos_class,
    plot = p
  ))
}

#' Plot overlap diagnostics between train/test groups
#'
#' @description
#' Checks whether the same group identifiers appear in both the training and
#' test partitions within each resample. This is designed to detect leakage
#' from grouped or repeated-measures data (for example, the same subject,
#' batch, plate, or study appearing on both sides of a fold) when group-wise
#' splitting is expected.
#'
#' @details
#' For each resample in `fit@splits@indices`, the function counts the number of
#' unique values of `column` in the train and test sets and the size of their
#' intersection. Any non-zero overlap indicates that at least one group appears
#' in both train and test for that resample. The check is metadata-based only:
#' it relies on exact matches of the supplied column and does not inspect
#' features or outcomes. It only checks train vs test within each resample, so
#' it will not detect overlaps across different resamples or other leakage
#' mechanisms. Inconsistent IDs or missing values in the metadata can hide or
#' inflate overlaps. `NA` values are treated as regular identifiers and will
#' count toward overlap if they appear in both partitions. Requires ggplot2.
#'
#' @param fit A `LeakFit` object produced by [fit_resample()]. It must contain
#'   the split indices and the associated metadata in `fit@splits@info$coldata`.
#'   The metadata rows must align with the data used to create the splits.
#' @param column Character scalar naming the metadata column to check (for
#'   example `"subject"` or `"batch"`). The function compares unique values of
#'   this column between train and test within each resample. There is no
#'   default: `NULL` or an unknown column triggers an error. Changing `column`
#'   changes which kind of leakage (subject-level, batch-level, etc.) is tested
#'   and therefore the overlap counts.
#' @return A list returned invisibly with:
#'   \itemize{
#'     \item `overlap_counts`: data.frame with one row per resample and columns
#'       `fold` (resample index in `fit@splits@indices`), `overlap` (unique IDs
#'       shared by train and test), `train` (unique IDs in train), and `test`
#'       (unique IDs in test).
#'     \item `column`: the metadata column name used for the check.
#'     \item `plot`: the ggplot object showing the three count series across folds.
#'   }
#'   The plot is also printed. When any overlap is detected, the plot adds a
#'   warning annotation.
#' @examples
#' set.seed(1)
#' df <- data.frame(
#'   subject = rep(1:6, each = 2),
#'   outcome = rbinom(12, 1, 0.5),
#'   x1 = rnorm(12),
#'   x2 = rnorm(12)
#' )
#' splits <- make_split_plan(df, outcome = "outcome",
#'                       mode = "subject_grouped", group = "subject", v = 3)
#' custom <- list(
#'   glm = list(
#'     fit = function(x, y, task, weights, ...) {
#'       stats::glm(y ~ ., data = as.data.frame(x),
#'                  family = stats::binomial(), weights = weights)
#'     },
#'     predict = function(object, newdata, task, ...) {
#'       as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
#'                                 type = "response"))
#'     }
#'   )
#' )
#' fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                     learner = "glm", custom_learners = custom,
#'                     metrics = "accuracy", refit = FALSE)
#' if (requireNamespace("ggplot2", quietly = TRUE)) {
#'   out <- plot_overlap_checks(fit, column = "subject")
#'   out$overlap_counts
#' }
#'
#' @export
plot_overlap_checks <- function(fit, column = NULL) {
  stopifnot(inherits(fit, "LeakFit"))
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package 'ggplot2' is required for plotting. Install it to use plot_overlap_checks().",
         call. = FALSE)
  }
  cd <- fit@splits@info$coldata
  if (is.null(cd) || is.null(column) || !column %in% names(cd)) {
    stop("Column not available in metadata.")
  }
  n <- nrow(cd)
  counts <- lapply(seq_along(fit@splits@indices), function(i) {
    idx <- fit@splits@indices[[i]]
    idx <- .bio_resolve_fold_indices(fit@splits, idx, n = n, data = cd)
    tr <- unique(cd[[column]][idx$train])
    te <- unique(cd[[column]][idx$test])
    data.frame(fold = i, overlap = length(intersect(tr, te)),
               train = length(tr), test = length(te))
  })
  counts <- do.call(rbind, counts)
  plot_df <- data.frame(
    fold = rep(counts$fold, times = 3),
    metric = factor(rep(c("Overlap", "Train unique", "Test unique"),
                        each = nrow(counts)),
                    levels = c("Overlap", "Train unique", "Test unique")),
    count = c(counts$overlap, counts$train, counts$test)
  )
  p <- ggplot2::ggplot(plot_df,
                       ggplot2::aes(x = fold, y = count, color = metric, linetype = metric)) +
    ggplot2::geom_line() +
    ggplot2::geom_point() +
    ggplot2::scale_color_manual(values = c("Overlap" = "red",
                                           "Train unique" = "grey40",
                                           "Test unique" = "grey70")) +
    ggplot2::scale_linetype_manual(values = c("Overlap" = "solid",
                                              "Train unique" = "dashed",
                                              "Test unique" = "dotted")) +
    ggplot2::scale_x_continuous(breaks = counts$fold) +
    ggplot2::labs(title = sprintf("Overlap diagnostics: %s", column),
                  x = "Fold", y = "Count", color = NULL, linetype = NULL) +
    ggplot2::theme_minimal() +
    ggplot2::theme(legend.position = "top")
  if (any(counts$overlap > 0)) {
    y_max <- max(plot_df$count, na.rm = TRUE)
    if (!is.finite(y_max)) y_max <- 1
    p <- p +
      ggplot2::expand_limits(y = y_max * 1.1) +
      ggplot2::annotate("text",
                        x = max(counts$fold),
                        y = y_max * 1.05,
                        label = "WARNING: Overlaps detected!",
                        hjust = 1, vjust = 0,
                        color = "red", fontface = "bold")
  }
  if (interactive()) print(p)
  invisible(list(
    overlap_counts = counts,
    column = column,
    plot = p
  ))
}

#' Plot ACF of test predictions for time-series leakage checks
#'
#' Uses the autocorrelation function of out-of-fold predictions to detect
#' temporal dependence that may indicate leakage. Predictions are ordered by
#' the split time column before computing the ACF. Requires numeric predictions
#' (regression or survival). Requires ggplot2.
#'
#' @param fit LeakFit.
#' @param lag.max maximum lag to show.
#' @return A list with the autocorrelation results, \code{lag.max}, and a ggplot object.
#' @examples
#' if (requireNamespace("ggplot2", quietly = TRUE)) {
#'   set.seed(42)
#'   df <- data.frame(
#'     id = 1:30,
#'     time = seq.Date(as.Date("2020-01-01"), by = "day", length.out = 30),
#'     y = rnorm(30),
#'     x1 = rnorm(30),
#'     x2 = rnorm(30)
#'   )
#'   splits <- make_split_plan(df, outcome = "y", mode = "time_series",
#'                             time = "time", v = 3, progress = FALSE)
#'   custom <- list(
#'     lm = list(
#'       fit = function(x, y, task, weights, ...) {
#'         stats::lm(y ~ ., data = data.frame(y = y, x))
#'       },
#'       predict = function(object, newdata, task, ...) {
#'         as.numeric(stats::predict(object, newdata = as.data.frame(newdata)))
#'       }
#'     )
#'   )
#'   fit <- fit_resample(df, outcome = "y", splits = splits,
#'                       learner = "lm", custom_learners = custom,
#'                       metrics = "rmse", refit = FALSE, seed = 1)
#'   plot_time_acf(fit, lag.max = 10)
#' }
#'
#' @export
plot_time_acf <- function(fit, lag.max = 20) {
  stopifnot(inherits(fit, "LeakFit"))
  all_pred <- do.call(rbind, fit@predictions)
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package 'ggplot2' is required for plotting. Install it to use plot_time_acf().",
         call. = FALSE)
  }
  if (!"pred" %in% names(all_pred)) {
    stop("Predictions missing 'pred' column for ACF plotting.", call. = FALSE)
  }
  pred <- all_pred$pred
  if (!is.numeric(pred)) {
    stop("plot_time_acf requires numeric predictions (regression or survival).", call. = FALSE)
  }
  if (!"id" %in% names(all_pred)) {
    stop("Predictions are missing sample ids; time ordering unavailable.", call. = FALSE)
  }
  time_col <- fit@splits@info$time %||% NULL
  coldata <- fit@splits@info$coldata %||% NULL
  if (is.null(coldata)) {
    stop("plot_time_acf requires split metadata with time values.", call. = FALSE)
  }
  coldata <- as.data.frame(coldata, check.names = FALSE)
  if (is.null(time_col) || !time_col %in% names(coldata)) {
    stop("plot_time_acf requires a time column in split metadata.", call. = FALSE)
  }

  ids_chr <- as.character(all_pred$id)
  time_vals <- NULL
  rn <- rownames(coldata)
  if (!is.null(rn) && all(ids_chr %in% rn)) {
    time_vals <- coldata[match(ids_chr, rn), time_col]
  } else if ("row_id" %in% names(coldata)) {
    rid <- as.character(coldata[["row_id"]])
    if (!anyDuplicated(rid) && all(ids_chr %in% rid)) {
      time_vals <- coldata[match(ids_chr, rid), time_col]
    }
  } else if (!is.null(fit@info$sample_ids) &&
             length(fit@info$sample_ids) == nrow(coldata)) {
    sample_ids <- as.character(fit@info$sample_ids)
    idx <- match(ids_chr, sample_ids)
    if (all(!is.na(idx))) time_vals <- coldata[idx, time_col]
  } else {
    ids_int <- suppressWarnings(as.integer(ids_chr))
    if (all(!is.na(ids_int)) && max(ids_int, na.rm = TRUE) <= nrow(coldata)) {
      time_vals <- coldata[ids_int, time_col]
    }
  }
  if (is.null(time_vals) || length(time_vals) != length(pred)) {
    stop("plot_time_acf could not align time metadata to predictions.", call. = FALSE)
  }
  if (!is.numeric(time_vals) && !inherits(time_vals, c("POSIXct", "Date"))) {
    stop("plot_time_acf requires numeric, Date, or POSIXct time values.", call. = FALSE)
  }

  ok_time <- !is.na(time_vals)
  if (!all(ok_time)) {
    warning("plot_time_acf dropped predictions with missing time values.", call. = FALSE)
  }
  pred <- pred[ok_time]
  time_vals <- time_vals[ok_time]
  ok_pred <- is.finite(pred)
  pred <- pred[ok_pred]
  time_vals <- time_vals[ok_pred]
  if (length(pred) < 2) {
    stop("Not enough finite predictions for ACF plotting.", call. = FALSE)
  }
  ord <- order(time_vals, seq_along(time_vals), na.last = TRUE)
  pred <- pred[ord]
  acf_res <- stats::acf(pred, lag.max = lag.max, plot = FALSE)
  lag_vals <- as.numeric(acf_res$lag)
  acf_vals <- as.numeric(acf_res$acf)
  df <- data.frame(lag = lag_vals, acf = acf_vals)
  conf <- if (is.finite(acf_res$n.used) && acf_res$n.used > 0) {
    1.96 / sqrt(acf_res$n.used)
  } else {
    NA_real_
  }

  p <- ggplot2::ggplot(df, ggplot2::aes(x = lag, y = acf)) +
    ggplot2::geom_hline(yintercept = 0, color = "grey50") +
    ggplot2::geom_segment(ggplot2::aes(xend = lag, yend = 0), color = "steelblue") +
    ggplot2::geom_point(color = "steelblue") +
    ggplot2::labs(title = "Prediction autocorrelation", x = "Lag", y = "ACF") +
    ggplot2::theme_minimal()
  if (is.finite(conf)) {
    p <- p +
      ggplot2::geom_hline(yintercept = c(conf, -conf),
                          color = "red", linetype = "dashed")
  }
  if (interactive()) print(p)
  invisible(list(acf = acf_res, lag.max = lag.max, plot = p))
}

#' Plot calibration curve for binomial predictions
#'
#' Visualizes observed outcome rates versus predicted probabilities across
#' bins to diagnose calibration (binomial tasks only). Requires ggplot2.
#'
#' @param fit LeakFit.
#' @param bins Number of probability bins to use.
#' @param min_bin_n Minimum samples per bin shown in the plot.
#' @param learner Optional character scalar. When predictions include multiple
#'   learners, selects the learner to summarize.
#' @return A list containing the calibration curve, metrics, and a ggplot object.
#' @examples
#' if (requireNamespace("ggplot2", quietly = TRUE)) {
#'   set.seed(42)
#'   df <- data.frame(
#'     subject = rep(1:15, each = 2),
#'     outcome = factor(rep(c(0, 1), 15)),
#'     x1 = rnorm(30),
#'     x2 = rnorm(30)
#'   )
#'   splits <- make_split_plan(df, outcome = "outcome",
#'                             mode = "subject_grouped", group = "subject",
#'                             v = 3, progress = FALSE)
#'   custom <- list(
#'     glm = list(
#'       fit = function(x, y, task, weights, ...) {
#'         stats::glm(y ~ ., data = as.data.frame(x),
#'                    family = stats::binomial(), weights = weights)
#'       },
#'       predict = function(object, newdata, task, ...) {
#'         as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
#'                                   type = "response"))
#'       }
#'     )
#'   )
#'   fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                       learner = "glm", custom_learners = custom,
#'                       metrics = "auc", refit = FALSE, seed = 1)
#'   plot_calibration(fit, bins = 5)
#' }
#'
#' @export
plot_calibration <- function(fit, bins = 10, min_bin_n = 5, learner = NULL) {
  stopifnot(inherits(fit, "LeakFit"))
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package 'ggplot2' is required for plotting. Install it to use plot_calibration().",
         call. = FALSE)
  }
  cal <- calibration_summary(fit, bins = bins, min_bin_n = min_bin_n, learner = learner)
  df <- cal$curve
  df <- df[is.finite(df$pred_mean) & is.finite(df$obs_rate), , drop = FALSE]
  df_plot <- df[df$n >= min_bin_n, , drop = FALSE]
  if (!nrow(df_plot)) {
    stop("Not enough data for calibration plotting.", call. = FALSE)
  }
  p <- ggplot2::ggplot(df_plot, ggplot2::aes(x = pred_mean, y = obs_rate)) +
    ggplot2::geom_abline(slope = 1, intercept = 0, linetype = "dashed",
                         color = "grey50") +
    ggplot2::geom_line(color = "steelblue") +
    ggplot2::geom_point(ggplot2::aes(size = n), color = "steelblue") +
    ggplot2::scale_size_continuous(range = c(2, 6)) +
    ggplot2::labs(title = "Calibration curve",
                  x = "Mean predicted probability",
                  y = "Observed event rate",
                  size = "Bin n") +
    ggplot2::theme_minimal()
  if (interactive()) print(p)
  invisible(list(curve = df, metrics = cal$metrics, plot = p))
}

#' Plot confounder sensitivity
#'
#' Shows performance metrics across confounder strata to assess sensitivity to
#' batch/study effects. Requires ggplot2.
#'
#' @param fit LeakFit.
#' @param confounders Character vector of columns in `coldata` to evaluate.
#' @param metric Metric name to compute within each stratum.
#' @param min_n Minimum samples per stratum to display.
#' @param coldata Optional data.frame of sample metadata.
#' @param numeric_bins Number of quantile bins for numeric confounders.
#' @param learner Optional character scalar. When predictions include multiple
#'   learners, selects the learner to summarize.
#' @return A list containing the sensitivity table and a ggplot object.
#' @examples
#' if (requireNamespace("ggplot2", quietly = TRUE)) {
#'   set.seed(42)
#'   df <- data.frame(
#'     subject = rep(1:15, each = 2),
#'     outcome = factor(rep(c(0, 1), 15)),
#'     batch = factor(rep(c("A", "B", "C"), 10)),
#'     x1 = rnorm(30),
#'     x2 = rnorm(30)
#'   )
#'   splits <- make_split_plan(df, outcome = "outcome",
#'                             mode = "subject_grouped", group = "subject",
#'                             v = 3, progress = FALSE)
#'   custom <- list(
#'     glm = list(
#'       fit = function(x, y, task, weights, ...) {
#'         stats::glm(y ~ ., data = as.data.frame(x),
#'                    family = stats::binomial(), weights = weights)
#'       },
#'       predict = function(object, newdata, task, ...) {
#'         as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
#'                                   type = "response"))
#'       }
#'     )
#'   )
#'   fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                       learner = "glm", custom_learners = custom,
#'                       metrics = "auc", refit = FALSE, seed = 1)
#'   plot_confounder_sensitivity(fit, confounders = "batch", coldata = df)
#' }
#'
#' @export
plot_confounder_sensitivity <- function(fit, confounders = NULL, metric = NULL,
                                        min_n = 10, coldata = NULL, numeric_bins = 4,
                                        learner = NULL) {
  stopifnot(inherits(fit, "LeakFit"))
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package 'ggplot2' is required for plotting. Install it to use plot_confounder_sensitivity().",
         call. = FALSE)
  }
  df <- confounder_sensitivity(fit, confounders = confounders, metric = metric,
                               min_n = min_n, coldata = coldata,
                               numeric_bins = numeric_bins, learner = learner)
  if (is.null(df) || !nrow(df)) {
    stop("No confounder sensitivity results available for plotting.", call. = FALSE)
  }
  df_plot <- df[is.finite(df$value) & df$n >= min_n, , drop = FALSE]
  if (!nrow(df_plot)) {
    stop("No strata meet the minimum size for plotting.", call. = FALSE)
  }
  direction <- df_plot$direction[1] %||% "higher"
  p <- ggplot2::ggplot(df_plot, ggplot2::aes(x = level, y = value)) +
    ggplot2::geom_col(fill = "steelblue") +
    ggplot2::facet_wrap(~ confounder, scales = "free_x") +
    ggplot2::coord_flip() +
    ggplot2::labs(title = "Confounder sensitivity",
                  subtitle = sprintf("%s (better is %s)", df_plot$metric[1], direction),
                  x = NULL, y = "Metric value") +
    ggplot2::theme_minimal()
  if (interactive()) print(p)
  invisible(list(data = df, plot = p))
}

test_that("plotting helpers validate ggplot2 availability", {
  skip_if_not_installed("ggplot2")
  df <- make_class_df(10)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 2, seed = 1)
  custom <- make_custom_learners()
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "glm", custom_learners = custom,
                            metrics = "accuracy", refit = FALSE, seed = 1)
  audit <- audit_leakage(fit, metric = "auc", B = 3, perm_stratify = FALSE)

  if (requireNamespace("ggplot2", quietly = TRUE)) {
    expect_true(TRUE)
  } else {
    expect_error(plot_perm_distribution(audit), "ggplot2")
    expect_error(plot_fold_balance(fit), "ggplot2")
    expect_error(plot_overlap_checks(fit, column = "subject"), "ggplot2")
    expect_error(plot_time_acf(fit), "ggplot2")
  }
})

test_that("plotting helpers return objects when ggplot2 is installed", {
  skip_if_not_installed("ggplot2")
  df <- make_class_df(12)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 3, seed = 1)
  custom <- make_custom_learners()
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "glm", custom_learners = custom,
                            metrics = "auc", refit = FALSE, seed = 1)
  audit <- audit_leakage(fit, metric = "auc", B = 5, perm_stratify = FALSE)

  with_temp_plot_device({
    res1 <- plot_perm_distribution(audit)
    expect_true(all(c("observed", "permuted_mean", "plot") %in% names(res1)))

    res2 <- plot_fold_balance(fit)
    expect_true(all(c("fold_summary", "plot") %in% names(res2)))

    res3 <- plot_overlap_checks(fit, column = "subject")
    expect_true(all(c("overlap_counts", "plot") %in% names(res3)))

    # plot_time_acf requires time_series splits; skip here
    expect_error(plot_time_acf(fit, lag.max = 3), "time")
  })
})

test_that("plot_overlap_checks and plot_time_acf validate inputs", {
  skip_if_not_installed("ggplot2")
  df <- make_class_df(8)
  df <- df[, c("subject", "outcome", "x1", "x2")]
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 2, seed = 1)
  custom <- make_custom_learners()
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "glm", custom_learners = custom,
                            metrics = "accuracy", refit = FALSE, seed = 1)
  expect_error(plot_overlap_checks(fit, column = "missing"), "Column not available")

  fit@predictions <- list(data.frame(pred = 1, truth = factor(1), id = 1, fold = 1, learner = "glm"))
  expect_error(plot_time_acf(fit, lag.max = 2), "time")
})

test_that("plot_perm_distribution errors without permutation values", {
  skip_if_not_installed("ggplot2")
  df <- make_class_df(10)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 2, seed = 1)
  custom <- make_custom_learners()
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "glm", custom_learners = custom,
                            metrics = "auc", refit = FALSE, seed = 1)
  audit <- audit_leakage(fit, metric = "auc", B = 3, return_perm = FALSE,
                         perm_stratify = FALSE)
  expect_error(plot_perm_distribution(audit), "No finite permutation values")
})

# Tidymodels interoperability helpers ---------------------------------------

.bio_is_rset <- function(x) inherits(x, "rset")
.bio_is_rsplit <- function(x) inherits(x, "rsplit")
.bio_is_rsample <- function(x) .bio_is_rset(x) || .bio_is_rsplit(x)
.bio_is_recipe <- function(x) inherits(x, "recipe")
.bio_is_workflow <- function(x) inherits(x, "workflow")
.bio_is_yardstick_metric <- function(x) inherits(x, "metric") || inherits(x, "metric_function")

.bio_validation_mode <- function(mode = NULL) {
  if (.bio_is_strict()) return("error")
  mode <- mode %||% getOption("bioLeak.validation_mode", "warn")
  if (is.null(mode) || length(mode) != 1L || !is.character(mode)) {
    mode <- "warn"
  }
  mode <- tolower(mode)
  if (!mode %in% c("warn", "error", "off")) {
    warning("Unknown bioLeak validation mode; using 'warn'.", call. = FALSE)
    mode <- "warn"
  }
  mode
}

.bio_signal_policy <- function(message, mode = "warn") {
  mode <- .bio_validation_mode(mode)
  if (identical(mode, "off")) return(invisible(FALSE))
  if (identical(mode, "error")) {
    stop(message, call. = FALSE)
  }
  warning(message, call. = FALSE)
  invisible(TRUE)
}

.bio_recipe_step_label <- function(step) {
  cls <- class(step)
  cls <- cls[!cls %in% c("step", "list")]
  if (!length(cls)) return("step")
  cls[[1]]
}

.bio_recipe_is_trained <- function(recipe_obj) {
  if (!.bio_is_recipe(recipe_obj)) return(FALSE)
  has_tr_info <- !is.null(recipe_obj$tr_info)
  has_orig_levels <- !is.null(recipe_obj$orig_lvls)
  has_fit_times <- !is.null(recipe_obj$fit_times)
  isTRUE(has_tr_info || has_orig_levels || has_fit_times)
}

.bio_validate_recipe_graph <- function(recipe_obj, context = "fit_resample", mode = NULL) {
  if (!.bio_is_recipe(recipe_obj)) return(invisible(TRUE))
  mode <- .bio_validation_mode(mode)
  if (identical(mode, "off")) return(invisible(TRUE))

  if (.bio_recipe_is_trained(recipe_obj)) {
    .bio_signal_policy(
      sprintf(
        "%s: received a trained recipe. Pass an untrained recipe to avoid fold-global leakage.",
        context
      ),
      mode
    )
  }

  steps <- recipe_obj$steps %||% list()
  if (length(steps)) {
    for (i in seq_along(steps)) {
      st <- steps[[i]]
      if (inherits(st, "step_lag")) {
        lag_vals <- suppressWarnings(as.numeric(st$lag %||% numeric(0)))
        if (length(lag_vals) && any(is.finite(lag_vals) & lag_vals < 0)) {
          .bio_signal_policy(
            sprintf(
              "%s: recipe step %d (%s) uses negative lag values; this can introduce lookahead leakage.",
              context, i, .bio_recipe_step_label(st)
            ),
            mode
          )
        }
      }
    }
  }

  invisible(TRUE)
}

.bio_workflow_is_trained <- function(workflow_obj) {
  if (!.bio_is_workflow(workflow_obj) || !requireNamespace("workflows", quietly = TRUE)) {
    return(FALSE)
  }

  is_trained_fn <- get0("is_trained_workflow", envir = asNamespace("workflows"), inherits = FALSE)
  if (is.function(is_trained_fn)) {
    out <- tryCatch(is_trained_fn(workflow_obj), error = function(e) FALSE)
    return(isTRUE(out))
  }

  has_fit <- tryCatch({
    workflows::extract_fit_parsnip(workflow_obj)
    TRUE
  }, error = function(e) FALSE)
  isTRUE(has_fit)
}

.bio_validate_workflow_graph <- function(workflow_obj, context = "fit_resample", mode = NULL) {
  if (!.bio_is_workflow(workflow_obj)) return(invisible(TRUE))
  mode <- .bio_validation_mode(mode)
  if (identical(mode, "off")) return(invisible(TRUE))

  if (.bio_workflow_is_trained(workflow_obj)) {
    .bio_signal_policy(
      sprintf(
        "%s: received a trained workflow. Pass an unfitted workflow to avoid fold-global leakage.",
        context
      ),
      mode
    )
  }

  if (requireNamespace("workflows", quietly = TRUE)) {
    pre <- tryCatch(workflows::extract_preprocessor(workflow_obj), error = function(e) NULL)
    if (.bio_is_recipe(pre)) {
      .bio_validate_recipe_graph(
        pre,
        context = paste0(context, " (workflow preprocessor)"),
        mode = mode
      )
    }
  }

  invisible(TRUE)
}

.bio_perm_mode <- function(splits) {
  if (!inherits(splits, "LeakSplits")) return(NA_character_)
  mode <- splits@mode %||% NA_character_
  valid_modes <- c("subject_grouped", "batch_blocked", "study_loocv", "time_series")
  if (!identical(mode, "rsample")) {
    return(mode)
  }
  perm_mode <- splits@info$perm_mode %||% NULL
  if (is.null(perm_mode)) return(mode)
  perm_mode <- as.character(perm_mode)
  perm_mode <- perm_mode[!is.na(perm_mode) & nzchar(perm_mode)]
  if (!length(perm_mode)) return(mode)
  perm_mode <- perm_mode[[1]]
  if (perm_mode %in% valid_modes) perm_mode else mode
}

.bio_rsample_col_name <- function(val) {
  if (is.null(val) || is.logical(val)) return(NULL)
  if (is.character(val)) {
    val <- val[!is.na(val) & nzchar(val)]
    return(if (length(val)) val else NULL)
  }
  if (inherits(val, "formula")) {
    vars <- all.vars(val)
    return(if (length(vars)) vars else NULL)
  }
  if (is.name(val)) return(as.character(val))
  if (is.call(val)) {
    vars <- all.vars(val)
    return(if (length(vars)) vars else NULL)
  }
  NULL
}

.bio_rsample_split_cols <- function(splits) {
  grab <- function(nm) .bio_rsample_col_name(attr(splits, nm, exact = TRUE))
  list(
    group = grab("group") %||% grab("groups"),
    batch = grab("batch"),
    study = grab("study"),
    time = grab("time") %||% grab("index") %||% grab("date")
  )
}

.bio_rsplit_indices <- function(split, n = NULL) {
  if (!inherits(split, "rsplit")) {
    stop("Expected an rsample rsplit object.", call. = FALSE)
  }
  in_id <- split$in_id %||% split$analysis
  out_id <- split$out_id %||% split$assessment
  if (is.null(in_id)) {
    stop("rsplit object does not expose analysis indices.", call. = FALSE)
  }
  # Treat NULL, zero-length, or all-NA out_id as "assessment not provided"
  if (is.null(out_id) || length(out_id) == 0L || all(is.na(out_id))) {
    if (is.null(n)) {
      stop("rsplit object does not expose assessment indices; provide n.", call. = FALSE)
    }
    out_id <- setdiff(seq_len(n), in_id[!is.na(in_id)])
  }
  # Drop any NA values from indices
  in_id  <- in_id[!is.na(in_id)]
  out_id <- out_id[!is.na(out_id)]
  if (!is.null(n)) {
    if (length(in_id) && any(in_id < 1L | in_id > n)) {
      stop("rsplit analysis indices exceed data rows.", call. = FALSE)
    }
    if (length(out_id) && any(out_id < 1L | out_id > n)) {
      stop("rsplit assessment indices exceed data rows.", call. = FALSE)
    }
  }
  list(train = as.integer(in_id), test = as.integer(out_id))
}

.bio_as_leaksplits_from_rsample <- function(splits, n = NULL, coldata = NULL, split_cols = NULL) {
  if (!requireNamespace("rsample", quietly = TRUE)) {
    stop("Package 'rsample' is required for rsample split objects.", call. = FALSE)
  }

  if (.bio_is_rsplit(splits)) {
    split_list <- list(splits)
    split_ids <- "Fold1"
    split_ids2 <- NULL
  } else if (.bio_is_rset(splits)) {
    split_list <- splits$splits
    split_ids <- if ("id" %in% names(splits)) as.character(splits$id) else NULL
    split_ids2 <- if ("id2" %in% names(splits)) as.character(splits$id2) else NULL
  } else {
    stop("splits must be a LeakSplits or rsample rset/rsplit.", call. = FALSE)
  }

  if (!length(split_list)) {
    stop("rsample object contains no splits.", call. = FALSE)
  }

  if (is.null(split_ids)) {
    split_ids <- paste0("Fold", seq_along(split_list))
  }
  fold_map <- as.integer(factor(split_ids, levels = unique(split_ids)))

  if (is.null(split_ids2)) {
    repeat_map <- rep(1L, length(split_list))
  } else {
    repeat_map <- as.integer(factor(split_ids2, levels = unique(split_ids2)))
  }

  rsample_cols_attr <- .bio_rsample_split_cols(splits)
  rsample_cols <- rsample_cols_attr
  split_cols_norm <- NULL
  auto_cols <- FALSE
  if (!is.null(split_cols)) {
    if (is.character(split_cols) && length(split_cols) == 1L && identical(split_cols, "auto")) {
      auto_cols <- TRUE
      split_cols <- NULL
    } else if (is.character(split_cols)) {
      if (is.null(names(split_cols)) || !length(names(split_cols))) {
        stop("split_cols must be a named character vector or list.", call. = FALSE)
      }
      split_cols <- as.list(split_cols)
    } else if (!is.list(split_cols)) {
      stop("split_cols must be a named character vector or list.", call. = FALSE)
    }
    if (!is.null(split_cols)) {
      split_cols_norm <- list(
        group = .bio_rsample_col_name(split_cols$group %||% split_cols$groups),
        batch = .bio_rsample_col_name(split_cols$batch),
        study = .bio_rsample_col_name(split_cols$study),
        time = .bio_rsample_col_name(split_cols$time)
      )
      for (nm in names(split_cols_norm)) {
        if (!is.null(split_cols_norm[[nm]])) {
          rsample_cols[[nm]] <- split_cols_norm[[nm]]
        }
      }
    }
  }
  if (isTRUE(auto_cols) && !is.null(coldata)) {
    cd_names <- names(coldata)
    if (is.null(rsample_cols$group)) {
      if ("group" %in% cd_names) {
        rsample_cols$group <- "group"
      } else if ("subject" %in% cd_names) {
        rsample_cols$group <- "subject"
      }
    }
    if (is.null(rsample_cols$batch)) {
      if ("batch" %in% cd_names) rsample_cols$batch <- "batch"
    }
    if (is.null(rsample_cols$study)) {
      if ("study" %in% cd_names) rsample_cols$study <- "study"
    }
    if (is.null(rsample_cols$time)) {
      if ("time" %in% cd_names) {
        rsample_cols$time <- "time"
      } else if ("date" %in% cd_names) {
        rsample_cols$time <- "date"
      } else if ("index" %in% cd_names) {
        rsample_cols$time <- "index"
      }
    }
  }

  valid_modes <- c("subject_grouped", "batch_blocked", "study_loocv", "time_series")
  mode_attr <- attr(splits, "bioLeak_mode", exact = TRUE)
  if (!is.null(mode_attr)) {
    mode_attr <- as.character(mode_attr)
    mode_attr <- mode_attr[!is.na(mode_attr) & nzchar(mode_attr)]
    mode_attr <- if (length(mode_attr)) mode_attr[[1]] else NULL
  }
  perm_mode_attr <- attr(splits, "bioLeak_perm_mode", exact = TRUE)
  if (!is.null(perm_mode_attr)) {
    perm_mode_attr <- as.character(perm_mode_attr)
    perm_mode_attr <- perm_mode_attr[!is.na(perm_mode_attr) & nzchar(perm_mode_attr)]
    perm_mode_attr <- if (length(perm_mode_attr)) perm_mode_attr[[1]] else NULL
  }
  mode_attr_valid <- !is.null(mode_attr) && mode_attr %in% valid_modes
  perm_mode_attr_valid <- !is.null(perm_mode_attr) && perm_mode_attr %in% valid_modes
  if (isTRUE(mode_attr_valid) && isTRUE(perm_mode_attr_valid) &&
      !identical(mode_attr, perm_mode_attr)) {
    stop("bioLeak_mode and bioLeak_perm_mode conflict; supply only one or ensure they match.",
         call. = FALSE)
  }
  split_mode <- "rsample"
  perm_mode <- NULL
  if (isTRUE(mode_attr_valid)) {
    split_mode <- mode_attr
    perm_mode <- mode_attr
  } else if (isTRUE(perm_mode_attr_valid)) {
    perm_mode <- perm_mode_attr
  } else {
    time_ok <- function(col, cd) {
      if (is.null(cd)) return(FALSE)
      if (!is.null(col)) {
        col <- col[!is.na(col) & nzchar(col)]
        col <- if (length(col)) col[[1]] else NULL
      }
      if (is.null(col)) {
        if (!"time" %in% names(cd)) return(FALSE)
        col <- "time"
      }
      if (!col %in% names(cd)) return(FALSE)
      vals <- cd[[col]]
      is.numeric(vals) || inherits(vals, c("POSIXct", "Date"))
    }
    time_col <- rsample_cols$time
    time_series_classes <- c("rolling_origin", "sliding_index", "sliding_period", "sliding_window")
    if (inherits(splits, time_series_classes) && time_ok(time_col, coldata)) {
      perm_mode <- "time_series"
    } else if (inherits(splits, "group_rset")) {
      perm_mode <- "subject_grouped"
    } else {
      explicit_cols <- rsample_cols_attr
      if (!is.null(split_cols_norm)) {
        for (nm in names(split_cols_norm)) {
          explicit_cols[[nm]] <- split_cols_norm[[nm]]
        }
      }
      has_explicit <- any(vapply(explicit_cols, function(x) {
        !is.null(x) && length(x)
      }, logical(1)))
      perm_cols <- if (isTRUE(has_explicit)) explicit_cols else rsample_cols
      candidates <- character(0)
      if (!is.null(perm_cols$time) && time_ok(perm_cols$time, coldata)) {
        candidates <- c(candidates, "time_series")
      }
      if (!is.null(perm_cols$study)) {
        candidates <- c(candidates, "study_loocv")
      }
      if (!is.null(perm_cols$batch)) {
        candidates <- c(candidates, "batch_blocked")
      }
      if (!is.null(perm_cols$group)) {
        candidates <- c(candidates, "subject_grouped")
      }
      if (length(candidates) == 1L) perm_mode <- candidates[[1]]
    }
  }
  if (is.null(perm_mode)) {
    stop(paste0("rsample splits require an explicit perm_mode (set attr(splits, ",
                "'bioLeak_perm_mode') or 'bioLeak_mode') or a single inferable ",
                "group/batch/study/time column via split_cols."),
         call. = FALSE)
  }

  indices <- vector("list", length(split_list))
  summary_rows <- vector("list", length(split_list))
  for (i in seq_along(split_list)) {
    idx <- .bio_rsplit_indices(split_list[[i]], n = n)
    indices[[i]] <- list(
      train = idx$train,
      test = idx$test,
      fold = fold_map[[i]],
      repeat_id = repeat_map[[i]]
    )
    summary_rows[[i]] <- data.frame(
      fold = fold_map[[i]],
      repeat_id = repeat_map[[i]],
      train_n = length(idx$train),
      test_n = length(idx$test),
      stringsAsFactors = FALSE
    )
  }

  split_summary <- do.call(rbind, summary_rows)
  info <- list(
    outcome = NULL,
    v = length(unique(fold_map)),
    repeats = length(unique(repeat_map)),
    seed = NA_integer_,
    mode = split_mode,
    perm_mode = perm_mode,
    group = rsample_cols$group,
    batch = rsample_cols$batch,
    study = rsample_cols$study,
    time = rsample_cols$time,
    stratify = NA,
    nested = FALSE,
    horizon = 0,
    purge = 0,
    embargo = 0,
    summary = split_summary,
    hash = .bio_hash_indices(indices),
    inner = NULL,
    compact = FALSE,
    fold_assignments = NULL,
    coldata = coldata,
    source = "rsample",
    id = split_ids,
    id2 = split_ids2,
    split_cols_override = split_cols_norm
  )

  methods::new("LeakSplits", mode = split_mode, indices = indices, info = info)
}

.bio_resolve_fold_indices <- function(splits, fold, n, data = NULL) {
  if (!inherits(splits, "LeakSplits")) {
    stop("splits must be a LeakSplits object.", call. = FALSE)
  }
  if (!isTRUE(splits@info$compact) || !is.null(fold$train)) return(fold)

  fold_assignments <- splits@info$fold_assignments
  if (is.null(fold_assignments) || !length(fold_assignments)) {
    stop("Compact splits require fold assignments to compute indices.", call. = FALSE)
  }
  r <- fold$repeat_id %||% 1L
  assign_vec <- fold_assignments[[r]]
  if (is.null(assign_vec)) {
    stop(sprintf("Missing fold assignments for repeat %s.", r), call. = FALSE)
  }
  test <- which(assign_vec == fold$fold)
  if (identical(splits@mode, "time_series")) {
    time_col <- splits@info$time
    time_vec <- NULL
    if (!is.null(time_col) && !is.null(data) && time_col %in% names(data)) {
      time_vec <- data[[time_col]]
    } else if (!is.null(time_col) && !is.null(splits@info$coldata) &&
               time_col %in% names(splits@info$coldata)) {
      time_vec <- splits@info$coldata[[time_col]]
    }
    if (is.null(time_vec)) {
      stop("time_series compact splits require time column in data.", call. = FALSE)
    }
    split_horizon <- splits@info$horizon %||% 0
    split_purge <- splits@info$purge %||% 0
    split_embargo <- splits@info$embargo %||% 0
    train <- .bio_time_series_train_indices(
      time_vec = time_vec,
      test_idx = test,
      candidate_idx = seq_len(n),
      horizon = split_horizon,
      purge = split_purge,
      embargo = split_embargo
    )
  } else if (identical(splits@mode, "combined")) {
    # Combined mode: exclude samples sharing constraint-axis levels with test
    remaining <- setdiff(seq_len(n), test)
    cst <- splits@info$constraints
    cd <- data %||% splits@info$coldata
    if (!is.null(cst) && length(cst) > 1L && !is.null(cd)) {
      for (ax in cst[-1]) {
        ax_vec <- cd[[ax$col]]
        test_levels <- unique(ax_vec[test])
        remaining <- remaining[!ax_vec[remaining] %in% test_levels]
      }
    }
    train <- remaining
  } else {
    train <- setdiff(seq_len(n), test)
  }

  list(train = train, test = test, fold = fold$fold, repeat_id = fold$repeat_id)
}

.bio_make_rsplit <- function(train, test, data, id = NULL) {
  rsample_make_splits <- utils::getFromNamespace("make_splits", "rsample")
  args <- list(list(analysis = train, assessment = test), data = data)
  if (!is.null(id)) args$id <- id
  out <- try(do.call(rsample_make_splits, args), silent = TRUE)
  if (inherits(out, "try-error") && !is.null(id)) {
    args$id <- NULL
    out <- do.call(rsample_make_splits, args)
  }
  out
}

#' Convert LeakSplits to an rsample resample set
#'
#' @param x LeakSplits object created by [make_split_plan()].
#' @param data Optional data.frame used to populate rsample splits. When NULL,
#'   the stored `coldata` from `x` is used (if available).
#' @param ... Additional arguments passed to methods (unused).
#' @return An rsample \code{rset} object compatible with tidymodels workflows.
#'   The returned object is a tibble with class \code{rset} containing:
#'   \describe{
#'     \item{\code{splits}}{List-column of \code{rsplit} objects, each with
#'       \code{analysis} (training indices) and \code{assessment} (test indices).}
#'     \item{\code{id}}{Character column with fold identifiers (e.g., "Fold1").}
#'     \item{\code{id2}}{Character column with repeat identifiers (e.g., "Repeat1")
#'       when multiple repeats are present; otherwise absent.}
#'   }
#'   The object also carries attributes for \code{group}, \code{batch},
#'   \code{study}, \code{time} (when available from the original \code{LeakSplits}),
#'   and \code{bioLeak_mode} indicating the original splitting mode. This allows
#'   the splits to be used with \code{tune::tune_grid()}, \code{rsample::fit_resamples()},
#'   and other tidymodels functions.
#' @examples
#' if (requireNamespace("rsample", quietly = TRUE)) {
#'   df <- data.frame(
#'     subject = rep(1:10, each = 2),
#'     outcome = rbinom(20, 1, 0.5),
#'     x1 = rnorm(20),
#'     x2 = rnorm(20)
#'   )
#'   splits <- make_split_plan(df, outcome = "outcome",
#'                         mode = "subject_grouped", group = "subject", v = 5)
#'   rset <- as_rsample(splits, data = df)
#' }
#' @export
as_rsample <- function(x, data = NULL, ...) UseMethod("as_rsample")

#' @export
as_rsample.LeakSplits <- function(x, data = NULL, ...) {
  if (!requireNamespace("rsample", quietly = TRUE)) {
    stop("Package 'rsample' is required to export rsample splits.", call. = FALSE)
  }
  if (!inherits(x, "LeakSplits")) {
    stop("x must be a LeakSplits object.", call. = FALSE)
  }
  if (is.null(data)) data <- x@info$coldata
  if (is.null(data)) {
    stop("Provide 'data' to export rsample splits.", call. = FALSE)
  }
  data <- as.data.frame(data, check.names = FALSE)
  n <- nrow(data)
  folds <- x@indices
  if (!length(folds)) stop("LeakSplits object contains no folds.", call. = FALSE)

  resolved <- lapply(folds, function(fold) {
    .bio_resolve_fold_indices(x, fold, n = n, data = data)
  })

  ids <- vapply(resolved, function(fold) {
    paste0("Fold", fold$fold)
  }, character(1))
  repeats <- vapply(resolved, function(fold) fold$repeat_id %||% 1L, integer(1))
  ids2 <- if (length(unique(repeats)) > 1L) {
    paste0("Repeat", repeats)
  } else {
    NULL
  }

  split_objs <- lapply(seq_along(resolved), function(i) {
    fold <- resolved[[i]]
    .bio_make_rsplit(fold$train, fold$test, data = data, id = ids[[i]])
  })

  manual_rset <- utils::getFromNamespace("manual_rset", "rsample")
  args <- list(splits = split_objs, ids = ids)
  if (!is.null(ids2) && "ids2" %in% names(formals(manual_rset))) {
    args$ids2 <- ids2
  }
  rset <- do.call(manual_rset, args)
  info <- x@info
  if (!is.null(info$group)) attr(rset, "group") <- info$group
  if (!is.null(info$batch)) attr(rset, "batch") <- info$batch
  if (!is.null(info$study)) attr(rset, "study") <- info$study
  if (!is.null(info$time)) attr(rset, "time") <- info$time
  if (!is.null(info$mode)) attr(rset, "bioLeak_mode") <- info$mode
  rset
}

test_that("recipe leakage validator flags trained recipes", {
  skip_if_not_installed("recipes")

  df <- data.frame(
    outcome = factor(rep(c(0, 1), each = 5), levels = c(0, 1)),
    x1 = rnorm(10),
    x2 = rnorm(10)
  )

  rec <- recipes::recipe(outcome ~ x1 + x2, data = df)
  rec_trained <- recipes::prep(rec, training = df, retain = TRUE)

  expect_warning(
    bioLeak:::.bio_validate_recipe_graph(rec_trained, context = "test", mode = "warn"),
    "trained recipe"
  )
  expect_error(
    bioLeak:::.bio_validate_recipe_graph(rec_trained, context = "test", mode = "error"),
    "trained recipe"
  )
})

test_that("recipe leakage validator flags negative lag steps", {
  skip_if_not_installed("recipes")

  df <- data.frame(
    outcome = factor(rep(c(0, 1), each = 5), levels = c(0, 1)),
    x1 = rnorm(10),
    x2 = rnorm(10)
  )
  rec_lag <- recipes::recipe(outcome ~ x1 + x2, data = df) |>
    recipes::step_lag(x1, lag = -1)

  expect_warning(
    bioLeak:::.bio_validate_recipe_graph(rec_lag, context = "test", mode = "warn"),
    "negative lag"
  )
})

test_that("workflow leakage validator flags trained workflows", {
  skip_if_not_installed("workflows")
  skip_if_not_installed("parsnip")

  df <- data.frame(
    outcome = factor(rep(c(0, 1), each = 6), levels = c(0, 1)),
    x1 = rnorm(12),
    x2 = rnorm(12)
  )
  wf <- workflows::workflow() |>
    workflows::add_model(
      parsnip::logistic_reg(mode = "classification") |>
        parsnip::set_engine("glm")
    ) |>
    workflows::add_formula(outcome ~ x1 + x2)
  wf_trained <- generics::fit(wf, data = df)

  expect_warning(
    bioLeak:::.bio_validate_workflow_graph(wf_trained, context = "test", mode = "warn"),
    "trained workflow"
  )
})

test_that("fit_resample accepts rsample splits", {
  skip_if_not_installed("rsample")
  df <- make_class_df(12)
  rs <- rsample::vfold_cv(df, v = 2)
  attr(rs, "bioLeak_perm_mode") <- "subject_grouped"
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = rs,
                     learner = "glm", custom_learners = make_custom_learners(),
                     metrics = "accuracy", refit = FALSE)
  expect_true(nrow(fit@metrics) > 0)
})

test_that("rsample grouped splits drop grouping columns", {
  skip_if_not_installed("rsample")
  df <- make_class_df(12)
  rs <- rsample::group_vfold_cv(df, group = subject, v = 2)
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = rs,
                     learner = "glm", custom_learners = make_custom_learners(),
                     metrics = "accuracy", refit = FALSE)
  expect_true(length(fit@feature_names) > 0)
  expect_false("subject" %in% fit@feature_names)
})

test_that("rsample split_cols override drops identifiers", {
  skip_if_not_installed("rsample")
  df <- make_class_df(12)
  rs <- rsample::vfold_cv(df, v = 2)
  attr(rs, "bioLeak_perm_mode") <- "subject_grouped"
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = rs,
                     split_cols = list(group = "subject", batch = "batch"),
                     learner = "glm", custom_learners = make_custom_learners(),
                     metrics = "accuracy", refit = FALSE)
  expect_false("subject" %in% fit@feature_names)
  expect_false("batch" %in% fit@feature_names)
})

test_that("rsample auto split_cols drops common identifiers", {
  skip_if_not_installed("rsample")
  df <- make_class_df(12)
  rs <- rsample::vfold_cv(df, v = 2)
  attr(rs, "bioLeak_perm_mode") <- "subject_grouped"
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = rs,
                     learner = "glm", custom_learners = make_custom_learners(),
                     metrics = "accuracy", refit = FALSE)
  expect_false("subject" %in% fit@feature_names)
  expect_false("batch" %in% fit@feature_names)
  expect_false("time" %in% fit@feature_names)
})

test_that("as_rsample converts LeakSplits", {
  skip_if_not_installed("rsample")
  df <- make_class_df(12)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject", v = 3, seed = 1)
  rs <- as_rsample(splits, data = df)
  expect_true(inherits(rs, "rset"))
  expect_equal(nrow(rs), length(splits@indices))
  expect_equal(attr(rs, "group"), "subject")
  expect_equal(attr(rs, "bioLeak_mode"), "subject_grouped")
  back <- bioLeak:::.bio_as_leaksplits_from_rsample(rs, n = nrow(df), coldata = df, split_cols = "auto")
  expect_equal(back@mode, "subject_grouped")
  expect_equal(back@info$perm_mode, "subject_grouped")
})

test_that("rsample group splits enable restricted permutations", {
  skip_if_not_installed("rsample")
  df <- make_class_df(12)
  rs <- rsample::group_vfold_cv(df, group = subject, v = 2)
  splits <- bioLeak:::.bio_as_leaksplits_from_rsample(rs, n = nrow(df), coldata = df, split_cols = "auto")
  expect_equal(splits@info$perm_mode, "subject_grouped")
  expect_equal(bioLeak:::.bio_perm_mode(splits), "subject_grouped")
})

test_that("fit_resample accepts recipe preprocessing", {
  skip_if_not_installed("recipes")
  df <- make_class_df(12)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject", v = 3, seed = 1)
  rec <- recipes::recipe(outcome ~ ., data = df) |>
    recipes::step_normalize(recipes::all_numeric_predictors())
  fit <- tryCatch(
    fit_resample_quiet(df, outcome = "outcome", splits = splits,
                       preprocess = rec,
                       learner = "glm", custom_learners = make_custom_learners(),
                       metrics = "accuracy", refit = FALSE),
    error = function(e) {
      skip(paste("recipe compatibility issue:", conditionMessage(e)))
    }
  )
  expect_true(nrow(fit@metrics) > 0)
})

test_that("fit_resample accepts workflow learners", {
  skip_if_not_installed("workflows")
  skip_if_not_installed("parsnip")
  df <- make_class_df(12)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject", v = 3, seed = 1)
  spec <- parsnip::logistic_reg(mode = "classification") |>
    parsnip::set_engine("glm")
  wf <- workflows::workflow() |>
    workflows::add_model(spec) |>
    workflows::add_formula(outcome ~ .)
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = wf, metrics = "accuracy", refit = FALSE)
  expect_true(nrow(fit@metrics) > 0)
})

test_that("fit_resample accepts yardstick metric sets", {
  skip_if_not_installed("yardstick")
  df <- make_class_df(12)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject", v = 3, seed = 1)
  ys <- yardstick::metric_set(yardstick::accuracy, yardstick::roc_auc)
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "glm", custom_learners = make_custom_learners(),
                            metrics = ys, refit = FALSE)
  expect_true(all(c("accuracy", "roc_auc") %in% colnames(fit@metrics)))
})

# Adapter: consume splitGraph split_spec objects as bioLeak LeakSplits.
#
# splitGraph emits a tool-agnostic `split_spec` describing sample-level
# grouping, blocking, and ordering. This file converts that artifact into a
# bioLeak `LeakSplits` by joining the spec's sample_data to the user's
# feature/outcome frame on sample_id and delegating to `make_split_plan()`.

#' Convert a splitGraph split_spec into bioLeak splits
#'
#' Consume a `split_spec` produced by \pkg{splitGraph} and build a
#' corresponding \code{LeakSplits} object via \code{\link{make_split_plan}}.
#' The spec supplies the grouping/blocking/ordering assignments; the caller
#' supplies the observation frame (features + outcome), joined on sample id.
#'
#' The mapping from \code{spec$constraint_mode} to
#' \code{make_split_plan(mode=)} is:
#' \itemize{
#'   \item \code{"subject"}   -> \code{"subject_grouped"}
#'   \item \code{"batch"}     -> \code{"batch_blocked"}
#'   \item \code{"study"}     -> \code{"study_loocv"}
#'   \item \code{"time"}      -> \code{"time_series"}
#'   \item \code{"composite"} -> \code{"combined"}
#' }
#'
#' Blocking variables declared on the spec (\code{batch_group},
#' \code{study_group}) and ordering (\code{order_rank}) are forwarded
#' automatically when relevant.
#'
#' @param spec A \code{split_spec} object from \pkg{splitGraph}.
#' @param data A data.frame (or SummarizedExperiment) containing at least one
#'   identifier column matching \code{sample_id_col} and an \code{outcome}
#'   column.
#' @param outcome Name of the outcome column in \code{data}.
#' @param sample_id_col Name of the sample-id column in \code{data}
#'   (default \code{"sample_id"}).
#' @param v Number of CV folds to request from \code{make_split_plan()}.
#' @param ... Additional arguments forwarded to \code{\link{make_split_plan}}
#'   (e.g. \code{stratify}, \code{seed}, \code{horizon}, \code{purge}).
#'
#' @return A \code{LeakSplits} object.
#'
#' @seealso \code{\link{make_split_plan}}, \code{\link{as_rsample}}
#' @export
as_leaksplits <- function(spec, data, outcome,
                          sample_id_col = "sample_id",
                          v = 5, ...) {
  if (!inherits(spec, "split_spec")) {
    stop("`spec` must be a `split_spec` (produced by splitGraph::as_split_spec()).",
         call. = FALSE)
  }
  if (!is.data.frame(data)) {
    stop("`data` must be a data.frame.", call. = FALSE)
  }
  if (!sample_id_col %in% names(data)) {
    stop(sprintf("Column '%s' not found in `data`.", sample_id_col),
         call. = FALSE)
  }
  if (is.null(outcome) || !outcome %in% names(data)) {
    stop("`outcome` must name a column in `data`.", call. = FALSE)
  }

  sd <- spec$sample_data
  keep <- c("sample_id", spec$group_var,
            intersect(c("batch_group", "study_group",
                        "timepoint_id", "order_rank"), names(sd)))
  sd <- sd[, keep, drop = FALSE]

  merged <- merge(data, sd,
                  by.x = sample_id_col, by.y = "sample_id",
                  all.x = FALSE, all.y = FALSE, sort = FALSE)
  if (nrow(merged) != nrow(sd)) {
    stop(sprintf(
      "Join on '%s' produced %d rows but spec has %d samples; check ID alignment.",
      sample_id_col, nrow(merged), nrow(sd)
    ), call. = FALSE)
  }

  mode_map <- c(subject   = "subject_grouped",
                batch     = "batch_blocked",
                study     = "study_loocv",
                time      = "time_series",
                composite = "combined")
  src_mode <- spec$constraint_mode %||% "composite"
  bl_mode <- mode_map[[src_mode]]
  if (is.null(bl_mode)) bl_mode <- "subject_grouped"

  args <- list(
    x       = merged,
    outcome = outcome,
    mode    = bl_mode,
    v       = v
  )

  args$group <- spec$group_var
  if ("batch_group" %in% names(merged) &&
      !all(is.na(merged$batch_group))) {
    args$batch <- "batch_group"
  }
  if ("study_group" %in% names(merged) &&
      !all(is.na(merged$study_group))) {
    args$study <- "study_group"
  }
  if (!is.null(spec$time_var) && spec$time_var %in% names(merged)) {
    args$time <- spec$time_var
  }

  user <- list(...)
  args[names(user)] <- user

  do.call(make_split_plan, args)
}

#' Render an HTML audit report
#'
#' Creates an HTML report that summarizes a leakage audit for a resampled model.
#' The report is built from a [LeakAudit] (or created from a [LeakFit]) and
#' presents: cross-validated metric summaries, a label-permutation association
#' test of the chosen performance metric (auto-refit when refit data are
#' available; otherwise fixed predictions), batch or study association tests
#' between metadata and predictions, confounder sensitivity plots, calibration
#' checks for binomial tasks, a target leakage scan based on feature-outcome
#' similarity (with multivariate scan enabled by default for supported tasks),
#' and duplicate
#' detection across training and test folds. The
#' output is a self-contained HTML file with tables and plots for these checks
#' plus the audit parameters used.
#'
#' The report does not refit models or reprocess data unless `perm_refit` triggers
#' refitting (`TRUE` or `"auto"` with a valid `perm_refit_spec`); it otherwise
#' inspects the predictions and metadata stored in the input.
#' Results are conditional on the
#' provided splits, selected metric, and any feature matrix supplied to
#' [audit_leakage()]. The univariate target leakage scan can miss multivariate
#' proxies, interaction leakage, or features not included in `X_ref`; the
#' multivariate scan (enabled by default for supported tasks) adds a model-based
#' check but still only uses features in `X_ref`. A
#' non-significant result does not prove the absence of leakage, especially with
#' small `B` or incomplete metadata. Rendering requires the
#' `rmarkdown` package and `ggplot2` for plots.
#'
#' @param audit A [LeakAudit] object from [audit_leakage()] or a [LeakFit] object
#'   from [fit_resample()]. If a [LeakAudit] is supplied, the report uses its
#'   stored results verbatim. If a [LeakFit] is supplied, `audit_report()` first
#'   computes a new audit via [audit_leakage(...)]; the fit must contain
#'   predictions and split metadata. When multiple learners were fit, pass a
#'   `learner` argument via `...` to select a single model.
#' @param output_file Character scalar. File name for the HTML report. Defaults
#'   to `"bioLeak_audit_report.html"`. If a relative name is provided, it is
#'   created inside `output_dir`. Changing this value only changes the file name,
#'   not the audit content.
#' @param output_dir Character scalar. Directory path where the report is
#'   written. Defaults to [tempdir()]. The directory must exist or be creatable
#'   by `rmarkdown::render()`. Changing this value only changes the output
#'   location.
#' @param quiet Logical scalar passed to `rmarkdown::render()`. Defaults to
#'   `TRUE`. When `FALSE`, knitting output and warnings are printed to the
#'   console. This does not change audit results.
#' @param open Logical scalar. Defaults to `FALSE`. When `TRUE`, opens the
#'   generated report in a browser via [utils::browseURL()]. This does not change
#'   the report contents.
#' @param ... Additional named arguments forwarded to [audit_leakage()] only
#'   when `audit` is a [LeakFit]. These control how the audit is computed and
#'   therefore change the report. Typical examples include `metric` (character),
#'   `B` (integer), `perm_stratify` (logical), `batch_cols` (character vector),
#'   `X_ref` (matrix/data.frame), `sim_method` (character), and
#'   `duplicate_scope` (character). When omitted, [audit_leakage()] defaults are
#'   used. Ignored when `audit` is already a [LeakAudit].
#' @return Character string containing the absolute file path to the generated
#'   HTML report. The report is a self-contained HTML file that can be opened
#'   in any web browser. It includes sections for: cross-validated metric
#'   summaries, label-permutation test results (gap, p-value), batch/study
#'   association tests, confounder sensitivity analysis, calibration diagnostics
#'   (for binomial tasks), target leakage scan results, and duplicate detection
#'   findings. The path can be used with \code{\link[utils]{browseURL}} to open
#'   the report programmatically.
#' @examples
#' set.seed(1)
#' df <- data.frame(
#'   subject = rep(1:6, each = 2),
#'   outcome = factor(rep(c(0, 1), 6)),
#'   x1 = rnorm(12),
#'   x2 = rnorm(12)
#' )
#'
#' splits <- make_split_plan(df, outcome = "outcome",
#'                       mode = "subject_grouped", group = "subject",
#'                       v = 3, progress = FALSE)
#'
#' custom <- list(
#'   glm = list(
#'     fit = function(x, y, task, weights, ...) {
#'       stats::glm(y ~ ., data = data.frame(y = y, x),
#'                  family = stats::binomial(), weights = weights)
#'     },
#'     predict = function(object, newdata, task, ...) {
#'       as.numeric(stats::predict(object,
#'                                 newdata = as.data.frame(newdata),
#'                                 type = "response"))
#'     }
#'   )
#' )
#'
#' fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                     learner = "glm", custom_learners = custom,
#'                     metrics = "auc", refit = FALSE, seed = 1)
#'
#' audit <- audit_leakage(fit, metric = "auc", B = 5, perm_stratify = FALSE)
#'
#' if (requireNamespace("rmarkdown", quietly = TRUE) &&
#'     requireNamespace("ggplot2", quietly = TRUE) &&
#'     isTRUE(rmarkdown::pandoc_available("1.12.3"))) {
#'   out_file <- audit_report(audit, output_dir = tempdir(), quiet = TRUE)
#'   out_file
#' }
#' @export
audit_report <- function(audit,
                         output_file = "bioLeak_audit_report.html",
                         output_dir = tempdir(),
                         quiet = TRUE,
                         open = FALSE,
                         ...) {
  if (inherits(audit, "LeakFit")) {
    audit <- audit_leakage(audit, ...)
  }
  if (!inherits(audit, "LeakAudit")) {
    stop("audit must be a LeakAudit or LeakFit object.")
  }
  if (!requireNamespace("rmarkdown", quietly = TRUE)) {
    stop("Install 'rmarkdown' to render the HTML report.", call. = FALSE)
  }
  template <- system.file("reports", "audit_report.Rmd", package = "bioLeak")
  if (!nzchar(template)) {
    stop("Report template not found. Reinstall bioLeak.", call. = FALSE)
  }

  # Render from a temporary copy to avoid writing intermediates into the
  # installed package directory, which may be read-only during checks.
  render_dir <- tempfile("bioLeak-audit-report-")
  ok_dir <- dir.create(render_dir, recursive = TRUE, showWarnings = FALSE)
  if (!isTRUE(ok_dir)) {
    stop("Failed to create temporary report directory.", call. = FALSE)
  }
  on.exit(unlink(render_dir, recursive = TRUE, force = TRUE), add = TRUE)

  render_input <- file.path(render_dir, basename(template))
  ok_copy <- file.copy(template, render_input, overwrite = TRUE)
  if (!isTRUE(ok_copy)) {
    stop("Failed to prepare report template in temporary directory.", call. = FALSE)
  }

  out <- rmarkdown::render(
    input = render_input,
    output_file = output_file,
    output_dir = output_dir,
    intermediates_dir = render_dir,
    params = list(audit = audit),
    quiet = quiet,
    envir = new.env(parent = globalenv())
  )

  if (isTRUE(open)) {
    utils::browseURL(out)
  }
  out
}

# Structured condition constructors for bioLeak ---------------------------------
#
# These use base-R condition construction (no rlang dependency).
# Users can catch specific error categories via:
#   tryCatch(..., bioLeak_input_error = function(e) ...)

#' @keywords internal
.bio_stop <- function(message, class = NULL, ..., call = NULL) {
  cond <- structure(
    class = c(class, "bioLeak_error", "error", "condition"),
    list(message = message, call = call, ...)
  )
  stop(cond)
}

#' @keywords internal
.bio_warn <- function(message, class = NULL, ..., call = NULL) {
  cond <- structure(
    class = c(class, "bioLeak_warning", "warning", "condition"),
    list(message = message, call = call, ...)
  )
  warning(cond)
}

#' @keywords internal
.bio_is_strict <- function() {
  isTRUE(getOption("bioLeak.strict", FALSE))
}

#' @keywords internal
.bio_strict_checks <- function(context = "bioLeak", seed = NULL,
                               nested = FALSE, mode = NA_character_) {
  if (!.bio_is_strict()) return(invisible(NULL))

  if (is.null(seed) || is.na(seed)) {
    .bio_warn(
      sprintf("%s [strict]: No explicit seed provided. Set 'seed' for reproducibility.", context),
      "bioLeak_validation_warning"
    )
  }

  if (isTRUE(nested) && identical(mode, "combined")) {
    .bio_warn(
      sprintf("%s [strict]: nested=TRUE with mode='combined' may produce empty inner folds.", context),
      "bioLeak_validation_warning"
    )
  }

  invisible(NULL)
}

# Condition class taxonomy:
#
# Errors:
#   bioLeak_input_error      — Bad argument type/value/missing
#   bioLeak_column_error     — Missing or invalid metadata column
#   bioLeak_split_error      — Split generation failures
#   bioLeak_overlap_error    — Overlap invariant violations
#   bioLeak_fit_error        — Model fitting failures
#   bioLeak_audit_error      — Audit-specific failures
#   bioLeak_dependency_error — Missing optional package
#
# Warnings:
#   bioLeak_validation_warning — Recipe/workflow leakage guardrail
#   bioLeak_fold_warning       — Fold-level issues
#   bioLeak_fallback_warning   — Fallback behavior triggered

test_that(".bio_stop raises error inheriting from bioLeak_error", {
  expect_error(
    bioLeak:::.bio_stop("test error", "bioLeak_input_error"),
    "test error"
  )
  err <- tryCatch(
    bioLeak:::.bio_stop("test error", "bioLeak_input_error"),
    error = function(e) e
  )
  expect_s3_class(err, "bioLeak_error")
  expect_s3_class(err, "bioLeak_input_error")
  expect_s3_class(err, "error")
  expect_s3_class(err, "condition")
})

test_that(".bio_warn raises warning inheriting from bioLeak_warning", {
  expect_warning(
    bioLeak:::.bio_warn("test warning", "bioLeak_fold_warning"),
    "test warning"
  )
  w <- tryCatch(
    bioLeak:::.bio_warn("test warning", "bioLeak_fold_warning"),
    warning = function(w) w
  )
  expect_s3_class(w, "bioLeak_warning")
  expect_s3_class(w, "bioLeak_fold_warning")
  expect_s3_class(w, "warning")
  expect_s3_class(w, "condition")
})

test_that("bioLeak_column_error is catchable from make_split_plan", {
  df <- data.frame(x1 = rnorm(10), outcome = rbinom(10, 1, 0.5))
  caught <- tryCatch(
    make_split_plan(df, outcome = "outcome",
                    mode = "subject_grouped", group = "nonexistent",
                    v = 3, progress = FALSE),
    bioLeak_column_error = function(e) "caught_column"
  )
  expect_equal(caught, "caught_column")
})

test_that("bioLeak_input_error is catchable from make_split_plan", {
  df <- data.frame(x1 = rnorm(10), outcome = rbinom(10, 1, 0.5))
  caught <- tryCatch(
    make_split_plan(df, outcome = "outcome",
                    mode = "subject_grouped", group = NULL,
                    v = 3, progress = FALSE),
    bioLeak_input_error = function(e) "caught_input"
  )
  expect_equal(caught, "caught_input")
})

test_that("check_split_overlap raises bioLeak_overlap_error", {
  indices <- list(
    list(train = 1:8, test = 7:10, fold = 1L, repeat_id = 1L)
  )
  cd <- data.frame(subject = c(1,1,2,2,3,3,4,4,5,5))
  splits_bad <- bioLeak:::LeakSplits(
    mode = "subject_grouped",
    indices = indices,
    info = list(v = 1L, repeats = 1L, seed = 1L, mode = "subject_grouped",
                group = "subject", batch = NULL, study = NULL, time = NULL,
                primary_axis = NULL, secondary_axis = NULL,
                stratify = FALSE, nested = FALSE,
                horizon = 0, purge = 0, embargo = 0,
                summary = data.frame(fold=1L, repeat_id=1L, train_n=8L, test_n=4L),
                hash = NA_character_, inner = NULL,
                compact = FALSE, fold_assignments = NULL,
                coldata = cd)
  )
  caught <- tryCatch(
    check_split_overlap(splits_bad, coldata = cd, cols = "subject"),
    bioLeak_overlap_error = function(e) "caught_overlap"
  )
  expect_equal(caught, "caught_overlap")
})

test_that("fit_resample with bad input raises bioLeak_input_error", {
  caught <- tryCatch(
    fit_resample(data.frame(x = 1:5), outcome = "x",
                 splits = "not_a_split",
                 learner = "glmnet"),
    bioLeak_input_error = function(e) "caught_input"
  )
  expect_equal(caught, "caught_input")
})

test_that("missing dependency raises bioLeak_dependency_error", {
  caught <- tryCatch(
    bioLeak:::.bio_stop("Install 'foo' package", "bioLeak_dependency_error"),
    bioLeak_dependency_error = function(e) "caught_dep"
  )
  expect_equal(caught, "caught_dep")
})

test_that("guard_to_recipe fallback raises bioLeak_fallback_warning", {
  skip_if_not_installed("recipes")
  df <- data.frame(outcome = factor(c(0,1,0,1)), x1 = rnorm(4))
  w <- tryCatch(
    guard_to_recipe(
      steps = list(impute = list(method = "missForest")),
      formula = outcome ~ .,
      training_data = df
    ),
    bioLeak_fallback_warning = function(w) w
  )
  expect_s3_class(w, "bioLeak_fallback_warning")
})

# Internal helper shared by explicit and compact time-series split builders.
.bio_time_series_train_indices <- function(time_vec, test_idx, candidate_idx = seq_along(time_vec),
                                           horizon = 0, purge = 0, embargo = 0) {
  if (!length(test_idx)) return(integer(0))
  tmin <- min(time_vec[test_idx])
  train <- if (horizon == 0 && purge == 0) {
    candidate_idx[time_vec[candidate_idx] < tmin]
  } else {
    candidate_idx[time_vec[candidate_idx] <= (tmin - horizon - purge)]
  }
  if (embargo > 0 && length(train)) {
    tmax <- max(time_vec[test_idx])
    train <- train[time_vec[train] <= (tmax - embargo)]
  }
  train
}

# Internal helper for combined (N-axis) splits.
.bio_make_combined_splits <- function(cd, n, constraints,
                                      v, repeats, stratify, outcome, seed,
                                      compact) {
  p_col <- constraints[[1]]$col
  p_groups <- as.factor(cd[[p_col]])
  p_levels <- levels(p_groups)

  # Precompute constraint axis vectors (all axes beyond the primary)
  constraint_axes <- lapply(constraints[-1], function(ax) cd[[ax$col]])

  indices <- list()
  summary_rows <- list()
  fold_assignments <- if (isTRUE(compact)) vector("list", repeats) else NULL

  .majority_class <- function(y) {
    ux <- unique(y)
    if (length(ux) == 0) return(NA)
    ux[which.max(tabulate(match(y, ux)))]
  }

  # Optional stratification at primary group level
  g_lab <- NULL
  if (isTRUE(stratify) && !is.null(outcome) && outcome %in% names(cd)) {
    y <- cd[[outcome]]
    g_lab <- tapply(seq_len(n), p_groups, function(ix) .majority_class(y[ix]))
    g_lab <- factor(g_lab)
    if (length(unique(g_lab)) < 2) g_lab <- NULL
  }

  # Build constraint-aware components of primary groups.
  # Subjects sharing any constraint-axis level form a component and should
  # be assigned to the same fold to avoid excessive exclusion.
  comp_map <- setNames(seq_along(p_levels), p_levels)
  for (ax_vec in constraint_axes) {
    ax_to_p <- split(as.character(p_groups), ax_vec)
    for (members in ax_to_p) {
      members <- unique(members)
      if (length(members) > 1L) {
        target_comp <- comp_map[members[1L]]
        for (m in members[-1L]) {
          old_comp <- comp_map[m]
          if (old_comp != target_comp) comp_map[comp_map == old_comp] <- target_comp
        }
      }
    }
  }
  components <- split(p_levels, comp_map)
  n_comp <- length(components)
  use_components <- n_comp >= v
  if (!use_components && length(constraint_axes) > 0L) {
    warning(sprintf(
      paste0("Only %d constraint-linked component(s) of primary groups ",
             "but v=%d folds; falling back to random group assignment."),
      n_comp, v), call. = FALSE)
  }

  for (r in seq_len(repeats)) {
    set.seed(seed + 1000 * r)
    if (isTRUE(compact)) fold_assign <- rep(NA_integer_, n)

    # Assign primary groups to folds
    gfold <- integer(length(p_levels))
    names(gfold) <- p_levels
    if (use_components && length(constraint_axes) > 0L) {
      # Constraint-aware: assign components via greedy bin packing
      comp_order <- sample(n_comp)
      if (!is.null(g_lab)) {
        comp_labels <- vapply(components, function(comp) {
          as.character(.majority_class(g_lab[comp]))
        }, character(1))
        g_levels <- levels(factor(g_lab))
        fold_counts <- matrix(0L, nrow = v, ncol = length(g_levels),
                              dimnames = list(NULL, g_levels))
        for (ci in comp_order) {
          cl <- comp_labels[ci]
          candidates <- which(fold_counts[, cl] == min(fold_counts[, cl]))
          target_fold <- if (length(candidates) == 1L) candidates else sample(candidates, 1L)
          for (pg in components[[ci]]) gfold[pg] <- target_fold
          fold_counts[target_fold, cl] <- fold_counts[target_fold, cl] + length(components[[ci]])
        }
      } else {
        fold_counts <- rep(0L, v)
        for (ci in comp_order) {
          candidates <- which(fold_counts == min(fold_counts))
          target_fold <- if (length(candidates) == 1L) candidates else sample(candidates, 1L)
          for (pg in components[[ci]]) gfold[pg] <- target_fold
          fold_counts[target_fold] <- fold_counts[target_fold] + length(components[[ci]])
        }
      }
    } else {
      # Random assignment (original behavior / fallback for highly connected data)
      if (!is.null(g_lab)) {
        for (cl in levels(g_lab)) {
          g_in <- names(g_lab)[g_lab == cl]
          gfold[g_in] <- sample(rep(seq_len(v), length.out = length(g_in)))
        }
      } else {
        gfold <- sample(rep(seq_len(v), length.out = length(p_levels)))
        names(gfold) <- p_levels
      }
    }

    for (k in seq_len(v)) {
      # Test = all samples belonging to primary groups assigned to fold k
      test_p <- p_levels[gfold == k]
      test <- which(p_groups %in% test_p)
      if (length(test) == 0L) next

      # Step 2-3: for each constraint axis, remove from remaining any samples
      # whose constraint-axis level appears in test
      remaining <- setdiff(seq_len(n), test)
      for (ax_vec in constraint_axes) {
        test_levels <- unique(ax_vec[test])
        remaining <- remaining[!ax_vec[remaining] %in% test_levels]
      }
      train <- remaining

      if (length(train) == 0L) next

      if (isTRUE(compact)) {
        fold_assign[test] <- k
        indices[[length(indices) + 1L]] <- list(fold = k, repeat_id = r)
      } else {
        indices[[length(indices) + 1L]] <- list(
          train = train, test = test, fold = k, repeat_id = r
        )
      }
      summary_rows[[length(summary_rows) + 1L]] <- data.frame(
        fold = k, repeat_id = r,
        train_n = length(train), test_n = length(test),
        stringsAsFactors = FALSE
      )
    }
    if (isTRUE(compact)) fold_assignments[[r]] <- fold_assign
  }

  list(indices = indices, summary_rows = summary_rows,
       fold_assignments = fold_assignments)
}

#' Create leakage-resistant splits
#'
#' @description
#' Generates leakage-safe cross-validation splits for common biomedical setups:
#' subject-grouped, batch-blocked, study leave-one-out, and time-series
#' rolling-origin. Supports repeats, optional stratification, nested inner CV,
#' and optional prediction horizon/purge/embargo gaps for time series. Note that splits store
#' explicit indices, which can be memory-intensive for large \code{n} and many
#' repeats.
#'
#' @param x SummarizedExperiment or data.frame/matrix (samples x features).
#'   If SummarizedExperiment, metadata are taken from colData(x). If data.frame,
#'   metadata are taken from x (columns referenced by \code{group}, \code{batch}, \code{study}, \code{time}, \code{outcome}).
#' @param outcome character, outcome column name (used for stratification).
#' @param mode one of "subject_grouped","batch_blocked","study_loocv","time_series","combined".
#' @param group subject/group id column (for subject_grouped). Required when
#'   mode is `subject_grouped`; use `group = "row_id"` to explicitly request
#'   sample-wise CV.
#' @param batch batch/plate/center column (for batch_blocked).
#' @param study study id column (for study_loocv).
#' @param time time column (numeric or POSIXct) for time_series.
#' @param primary_axis List with elements \code{type} (one of \code{"subject"},
#'   \code{"batch"}, \code{"study"}) and \code{col} (column name). Used only
#'   when \code{mode = "combined"} to define the primary grouping axis.
#'   Deprecated in favor of \code{constraints}; still supported for backward
#'   compatibility.
#' @param secondary_axis List with elements \code{type} and \code{col}. Used
#'   only when \code{mode = "combined"} to define the secondary constraint axis.
#'   Training sets exclude samples whose secondary-axis levels appear in the
#'   test set. Deprecated in favor of \code{constraints}; still supported for
#'   backward compatibility.
#' @param constraints A list of constraint specifications for \code{mode = "combined"}.
#'   Each element is a list with \code{type} (one of \code{"subject"}, \code{"batch"},
#'   \code{"study"}) and \code{col} (column name). The first element defines the
#'   primary grouping axis (fold driver); subsequent elements define exclusion
#'   constraints (training samples sharing constraint-axis levels with the test
#'   set are removed). Requires at least 2 elements. Cannot be used together with
#'   \code{primary_axis}/\code{secondary_axis}.
#' @param v integer, number of folds (k) or rolling partitions.
#' @param repeats integer, number of repeats (>=1) for non-LOOCV modes.
#' @param stratify logical, keep outcome proportions similar across folds.
#'   For grouped modes, stratification is applied at the group level (by
#'   majority class per group) if \code{outcome} is provided; otherwise ignored.
#' @param nested logical, whether to attach inner CV splits (per outer fold)
#'   using the same \code{mode} on the outer training set (with \code{v} folds, 1 repeat).
#' @param seed integer seed.
#' @param horizon numeric (>=0), minimal time gap for time_series so that the
#'   training set only contains samples with time < min(test_time) when horizon = 0,
#'   and time <= min(test_time) - horizon otherwise.
#' @param purge numeric (>=0), additional gap removed immediately before each
#'   time-series test block.
#' @param embargo numeric (>=0), additional exclusion window anchored at the end
#'   of each time-series test block. Training rows with
#'   \code{time > max(test_time) - embargo} are removed.
#' @param progress logical, print progress for large jobs.
#' @param compact logical; store fold assignments instead of explicit train/test
#'   indices to reduce memory usage for large datasets. Not supported when
#'   \code{nested = TRUE}.
#' @param strict logical; deprecated and ignored. `subject_grouped` always
#'   requires a non-NULL `group`.
#'
#' @return A \code{\linkS4class{LeakSplits}} S4 object containing:
#'   \describe{
#'     \item{\code{mode}}{Character string indicating the splitting mode
#'       (\code{"subject_grouped"}, \code{"batch_blocked"}, \code{"study_loocv"},
#'       or \code{"time_series"}).}
#'     \item{\code{indices}}{List of fold descriptors, each containing
#'       \code{train} (integer vector of training indices), \code{test}
#'       (integer vector of test indices), \code{fold} (fold number), and
#'       \code{repeat_id} (repeat identifier). When \code{compact = TRUE},
#'       indices are stored as fold assignments instead.}
#'     \item{\code{info}}{List of metadata including \code{outcome}, \code{v},
#'       \code{repeats}, \code{seed}, grouping columns (\code{group},
#'       \code{batch}, \code{study}, \code{time}), \code{stratify},
#'       \code{nested}, \code{horizon}, \code{purge}, \code{embargo},
#'       \code{summary} (data.frame of fold
#'       sizes), \code{hash} (reproducibility checksum), \code{inner}
#'       (nested inner splits if \code{nested = TRUE}), and \code{coldata}
#'       (sample metadata).}
#'   }
#'   Use the \code{show} method to print a summary, or access slots directly
#'   with \code{@}.
#' @examples
#' set.seed(1)
#' df <- data.frame(
#'   subject = rep(1:10, each = 2),
#'   outcome = rbinom(20, 1, 0.5),
#'   x1 = rnorm(20),
#'   x2 = rnorm(20)
#' )
#' splits <- make_split_plan(df, outcome = "outcome",
#'                       mode = "subject_grouped", group = "subject", v = 5)
#' @export
make_split_plan <- function(x, outcome = NULL,
                        mode = c("subject_grouped", "batch_blocked", "study_loocv", "time_series", "combined"),
                        group = NULL, batch = NULL, study = NULL, time = NULL,
                        primary_axis = NULL, secondary_axis = NULL,
                        constraints = NULL,
                        v = 5, repeats = 1, stratify = FALSE, nested = FALSE,
                        seed = 1, horizon = 0, purge = 0, embargo = 0,
                        progress = TRUE, compact = FALSE,
                        strict = TRUE) {

  mode <- match.arg(mode)
  .bio_strict_checks(context = "make_split_plan", seed = seed,
                     nested = nested, mode = mode)
  stopifnot(is.numeric(v), v >= 2 || mode == "study_loocv")
  stopifnot(is.numeric(repeats), repeats >= 1)
  stopifnot(is.numeric(horizon), horizon >= 0)
  stopifnot(is.numeric(purge), purge >= 0)
  stopifnot(is.numeric(embargo), embargo >= 0)
  set.seed(seed)
  # Seed lineage:
  #   make_split_plan:  set.seed(seed), repeats: seed+1000*r, nested: seed+1
  #   fit_resample:     set.seed(seed), per-fold: seed+fold_id
  #   audit_leakage:    set.seed(seed), per-perm: seed+b
  # These are NOT isolated streams. For practical v/repeats/B values,
  # collisions do not occur.

  if (!identical(mode, "time_series") && (purge > 0 || embargo > 0)) {
    .bio_warn("purge and embargo are only used when mode = 'time_series'; ignoring both.",
              "bioLeak_fold_warning")
  }

  if (isTRUE(compact) && isTRUE(nested)) {
    .bio_warn("compact=TRUE is not supported with nested=TRUE; using compact=FALSE.",
              "bioLeak_fold_warning")
    compact <- FALSE
  }

  # ---- Extract metadata ----
  X <- .bio_get_x(x)
  n <- nrow(X)

  if (.bio_is_se(x)) {
    cd <- .bio_coldata_df(x)
  } else if (is.data.frame(x)) {
    cd <- x
  } else if (is.matrix(x)) {
    cd <- data.frame(row_id = seq_len(n))
  } else {
    .bio_stop("x must be SummarizedExperiment, data.frame, or matrix.",
              "bioLeak_input_error")
  }

  # Ensure unique row_id always exists
  if (!"row_id" %in% names(cd)) cd$row_id <- seq_len(nrow(cd))

  # Helper to ensure column existence
  .need_col <- function(col, label) {
    if (is.null(col)) .bio_stop(sprintf("Provide '%s' column name.", label),
                                "bioLeak_column_error")
    if (!col %in% names(cd)) .bio_stop(sprintf("Column '%s' not found in metadata.", col),
                                       "bioLeak_column_error")
    if (anyNA(cd[[col]])) .bio_stop(sprintf("Missing values found in column '%s'.", col),
                                    "bioLeak_column_error")
  }

  # Column checks
  if (mode == "subject_grouped") {
    if (is.null(group)) {
      .bio_stop("subject_grouped mode requires a non-NULL 'group' column.",
                "bioLeak_input_error")
    }
    .need_col(group, "group")
  }
  if (mode == "batch_blocked")   .need_col(batch, "batch")
  if (mode == "study_loocv")     .need_col(study, "study")
  if (mode == "time_series")     .need_col(time, "time")
  if (mode == "combined") {
    valid_axis_types <- c("subject", "batch", "study")
    .validate_axis <- function(ax, label) {
      if (is.null(ax) || !is.list(ax))
        .bio_stop(sprintf("'%s' must be a list with 'type' and 'col' elements.", label),
                  "bioLeak_input_error")
      if (!all(c("type", "col") %in% names(ax)))
        .bio_stop(sprintf("'%s' must have 'type' and 'col' elements.", label),
                  "bioLeak_input_error")
      if (!ax$type %in% valid_axis_types)
        .bio_stop(sprintf("'%s$type' must be one of: %s.", label, paste(valid_axis_types, collapse = ", ")),
                  "bioLeak_input_error")
      .need_col(ax$col, paste0(label, "$col"))
    }

    # Resolve constraints vs primary_axis/secondary_axis
    has_legacy <- !is.null(primary_axis) || !is.null(secondary_axis)
    has_constraints <- !is.null(constraints)

    if (has_constraints && has_legacy) {
      .bio_stop("Cannot specify both 'constraints' and 'primary_axis'/'secondary_axis'. Use one or the other.",
                "bioLeak_input_error")
    }

    if (has_constraints) {
      # Validate constraints list
      if (!is.list(constraints) || length(constraints) < 2L)
        .bio_stop("'constraints' must be a list with at least 2 elements (primary + constraint).",
                  "bioLeak_input_error")
      for (i in seq_along(constraints)) {
        .validate_axis(constraints[[i]], sprintf("constraints[[%d]]", i))
      }
      # Derive primary_axis/secondary_axis for backward compat in info slot
      primary_axis <- constraints[[1]]
      secondary_axis <- constraints[[2]]
    } else {
      # Legacy path: validate and convert to constraints
      .validate_axis(primary_axis, "primary_axis")
      .validate_axis(secondary_axis, "secondary_axis")
      constraints <- list(primary_axis, secondary_axis)
    }
  }

  if (isTRUE(stratify) && (is.null(outcome) || !outcome %in% names(cd)))
    .bio_warn("Stratification requested but 'outcome' not found; proceeding unstratified.",
              "bioLeak_fold_warning")
  if (isTRUE(stratify) && !is.null(outcome) && length(outcome) == 2L) {
    .bio_warn("Stratification ignored for time/event outcomes; proceeding unstratified.",
              "bioLeak_fold_warning")
    stratify <- FALSE
  }
  if (isTRUE(stratify) && !is.null(outcome) && outcome %in% names(cd)) {
    if (inherits(cd[[outcome]], "Surv")) {
      .bio_warn("Stratification ignored for survival outcomes; proceeding unstratified.",
                "bioLeak_fold_warning")
      stratify <- FALSE
    }
  }

  # Utilities
  .as_factor <- function(z) as.factor(z)
  .majority_class <- function(y) {
    ux <- unique(y)
    if (length(ux) == 0) return(NA)
    ux[which.max(tabulate(match(y, ux)))]
  }
  .msg <- function(...) if (isTRUE(progress)) message(sprintf(...))

  indices <- list()
  inner_indices <- NULL
  summary_rows <- list()
  repeats_eff <- if (mode %in% c("study_loocv", "time_series")) 1 else repeats
  fold_assignments <- if (isTRUE(compact)) vector("list", repeats_eff) else NULL

  # ---- Subject-grouped ----
  if (mode == "subject_grouped") {
    gid <- .as_factor(cd[[group]])
    lev <- levels(gid)

    # Stratification
    if (isTRUE(stratify) && !is.null(outcome) && outcome %in% names(cd)) {
      y <- cd[[outcome]]
      g_lab <- tapply(seq_len(n), gid, function(ix) .majority_class(y[ix]))
      g_lab <- factor(g_lab)
      if (length(unique(g_lab)) < 2) {
        warning("Stratification ignored: only one class present at group level.")
        stratify <- FALSE
      }
    }

    for (r in seq_len(repeats)) {
      set.seed(seed + 1000 * r)
      if (isTRUE(compact)) {
        fold_assign <- rep(NA_integer_, n)
      }
      if (isTRUE(stratify) && exists("g_lab")) {
        gfold <- integer(length(lev)); names(gfold) <- lev
        for (cl in levels(g_lab)) {
          g_in <- names(g_lab)[g_lab == cl]
          gfold[g_in] <- sample(rep(seq_len(v), length.out = length(g_in)))
        }
      } else {
        gfold <- sample(rep(seq_len(v), length.out = length(lev)))
        names(gfold) <- lev
      }

      for (k in seq_len(v)) {
        test_g <- lev[gfold == k]
        test <- which(gid %in% test_g)
        train <- setdiff(seq_len(n), test)
        if (length(test) == 0L || length(train) == 0L) next
        if (isTRUE(compact)) {
          fold_assign[test] <- k
          indices[[length(indices) + 1L]] <- list(fold = k, repeat_id = r)
        } else {
          indices[[length(indices) + 1L]] <- list(train = train, test = test, fold = k, repeat_id = r)
        }
        summary_rows[[length(summary_rows) + 1L]] <- data.frame(
          fold = k,
          repeat_id = r,
          train_n = length(train),
          test_n = length(test),
          stringsAsFactors = FALSE
        )
      }
      if (isTRUE(compact)) fold_assignments[[r]] <- fold_assign
      .msg("subject_grouped: repeat %d/%d done.", r, repeats)
    }
  }

  # ---- Batch-blocked ----
  if (mode == "batch_blocked") {
    bid <- factor(cd[[batch]], levels = unique(cd[[batch]]))  # preserve order
    blevels <- levels(bid)

    if (length(blevels) < v) v <- length(blevels)

    if (isTRUE(stratify) && !is.null(outcome) && outcome %in% names(cd)) {
      y <- cd[[outcome]]
      b_lab <- tapply(seq_len(n), bid, function(ix) .majority_class(y[ix]))
      b_lab <- factor(b_lab)
      if (length(unique(b_lab)) < 2) {
        warning("Stratification ignored: only one class present at group level.")
        stratify <- FALSE
      }
    }

    for (r in seq_len(repeats)) {
      set.seed(seed + 1000 * r)
      if (isTRUE(compact)) {
        fold_assign <- rep(NA_integer_, n)
      }
      # if (isTRUE(stratify) && exists("b_lab")) {
      #   bfold <- integer(length(blevels)); names(bfold) <- blevels
      #   for (cl in levels(b_lab)) {
      #     b_in <- names(b_lab)[b_lab == cl]
      #     bfold[b_in] <- sample(rep(seq_len(v), length.out = length(b_in)))
      #   }
      # } else {
      #   if (v >= length(blevels)) {
      #     bfold <- seq_len(length(blevels))
      #     names(bfold) <- blevels
      #     bfold <- sample(bfold)
      #   } else {
      #     bfold <- sample(rep(seq_len(v), length.out = length(blevels)))
      #     names(bfold) <- blevels
      #   }
      # }

      if (v >= length(blevels)) {
        # One batch per fold: force unique assignment, ignore stratify
        bfold <- seq_len(length(blevels))
        names(bfold) <- blevels
        bfold <- sample(bfold)  # randomize order
      } else if (isTRUE(stratify) && exists("b_lab")) {
        bfold <- integer(length(blevels)); names(bfold) <- blevels
        for (cl in levels(b_lab)) {
          b_in <- names(b_lab)[b_lab == cl]
          bfold[b_in] <- sample(rep(seq_len(v), length.out = length(b_in)))
        }
      } else {
        bfold <- sample(rep(seq_len(v), length.out = length(blevels)))
        names(bfold) <- blevels
      }


      for (k in seq_len(v)) {
        test_b <- blevels[bfold == k]
        test <- which(bid %in% test_b)
        train <- setdiff(seq_len(n), test)
        if (length(train) == 0L || length(test) == 0L) next
        if (isTRUE(compact)) {
          fold_assign[test] <- k
          indices[[length(indices) + 1L]] <- list(fold = k, repeat_id = r)
        } else {
          indices[[length(indices) + 1L]] <- list(train = train, test = test, fold = k, repeat_id = r)
        }
        summary_rows[[length(summary_rows) + 1L]] <- data.frame(
          fold = k,
          repeat_id = r,
          train_n = length(train),
          test_n = length(test),
          stringsAsFactors = FALSE
        )
      }

      if (isTRUE(compact)) fold_assignments[[r]] <- fold_assign
      .msg("batch_blocked: repeat %d/%d done.", r, repeats)
    }
  }


  # ---- Study LOOCV ----
  if (mode == "study_loocv") {
    sid <- .as_factor(cd[[study]])
    if (isTRUE(compact)) {
      fold_assign <- rep(NA_integer_, n)
    }
    for (s in levels(sid)) {
      test <- which(sid == s)
      train <- setdiff(seq_len(n), test)
      if (length(train) == 0L || length(test) == 0L) next
      fold_id <- as.integer(which(levels(sid) == s))
      if (isTRUE(compact)) {
        fold_assign[test] <- fold_id
        indices[[length(indices) + 1L]] <- list(
          fold = fold_id,
          repeat_id = as.integer(1)
        )
      } else {
        indices[[length(indices) + 1L]] <- list(
          train = train,
          test = test,
          fold = fold_id,
          repeat_id = as.integer(1)
        )
      }
      summary_rows[[length(summary_rows) + 1L]] <- data.frame(
        fold = fold_id,
        repeat_id = 1,
        train_n = length(train),
        test_n = length(test),
        stringsAsFactors = FALSE
      )
    }
    if (isTRUE(compact)) fold_assignments[[1]] <- fold_assign
  }

  # ---- Time-series ----
  if (mode == "time_series") {
    tt <- cd[[time]]
    if (!is.numeric(tt) && !inherits(tt, c("POSIXct", "Date")))
      .bio_stop("'time' column must be numeric, Date, or POSIXct.",
                "bioLeak_input_error")
    ord <- order(tt)
    Xidx <- seq_len(n)[ord]

    block_ids <- cut(seq_along(Xidx), breaks = v, include.lowest = TRUE, labels = FALSE)
    blocks <- split(Xidx, block_ids)
    if (isTRUE(compact)) {
      fold_assign <- rep(NA_integer_, n)
    }
    for (k in seq_len(v)) {
      test <- blocks[[k]]
      if (!length(test)) next
      train <- .bio_time_series_train_indices(
        time_vec = tt,
        test_idx = test,
        candidate_idx = Xidx,
        horizon = horizon,
        purge = purge,
        embargo = embargo
      )
      if (length(train) < 1L) next
      if (length(test) < 3L) {
        warning(sprintf("time_series fold %d skipped: only %d test sample(s) (minimum 3).", k, length(test)),
                call. = FALSE)
        next
      }
      if (isTRUE(compact)) {
        fold_assign[test] <- k
        indices[[length(indices) + 1L]] <- list(fold = as.integer(k), repeat_id = as.integer(1))
      } else {
        indices[[length(indices) + 1L]] <- list(train = train, test = test, fold = as.integer(k), repeat_id = as.integer(1))
      }
      summary_rows[[length(summary_rows) + 1L]] <- data.frame(
        fold = as.integer(k),
        repeat_id = 1,
        train_n = length(train),
        test_n = length(test),
        stringsAsFactors = FALSE
      )
    }
    if (isTRUE(compact)) fold_assignments[[1]] <- fold_assign
  }

  # ---- Combined ----
  if (mode == "combined") {
    combined <- .bio_make_combined_splits(
      cd = cd, n = n,
      constraints = constraints,
      v = v, repeats = repeats, stratify = stratify, outcome = outcome,
      seed = seed, compact = compact
    )
    indices <- combined$indices
    summary_rows <- combined$summary_rows
    fold_assignments <- combined$fold_assignments
  }

  # ---- Nested ----
  if (isTRUE(nested) && length(indices) > 0) {
    inner_indices <- vector("list", length(indices))
    for (i in seq_along(indices)) {
      tr_idx <- indices[[i]]$train
      x_inner <- if (.bio_is_se(x)) x[, tr_idx] else cd[tr_idx, , drop = FALSE]
      set.seed(seed + 1L)
      inner <- make_split_plan(
        x_inner, outcome = outcome, mode = mode, group = group,
        batch = batch, study = study, time = time, v = v,
        constraints = constraints,
        repeats = 1, stratify = stratify, nested = FALSE,
        seed = seed + 1L, horizon = horizon, purge = purge, embargo = embargo,
        progress = FALSE
      )
      inner_indices[[i]] <- inner@indices
    }
  }

  # ---- Summary ----
  if (length(indices) == 0L)
    .bio_stop("No valid folds generated. Check inputs (group/batch/study/time, v, repeats, horizon).",
              "bioLeak_split_error")

  fold_repeat_keys <- vapply(indices, function(z) paste(z$fold, z$repeat_id), character(1))
  if (anyDuplicated(fold_repeat_keys)) {
    warning("Duplicate fold/repeat combinations detected; check v and repeats settings.")
  }

  split_summary <- if (isTRUE(compact)) {
    do.call(rbind, summary_rows)
  } else {
    data.frame(
      fold      = vapply(indices, `[[`, integer(1), "fold"),
      repeat_id = vapply(indices, `[[`, integer(1), "repeat_id"),
      train_n   = vapply(indices, function(z) length(z$train), integer(1)),
      test_n    = vapply(indices, function(z) length(z$test), integer(1)),
      stringsAsFactors = FALSE
    )
  }

  indices_hash <- .bio_hash_indices(indices)
  hash_val <- tryCatch({
    if (requireNamespace("digest", quietly = TRUE)) {
      digest::digest(list(
        mode = mode, n = n, v = v, repeats = repeats, stratify = stratify,
        seed = seed, horizon = horizon, purge = purge, embargo = embargo,
        group = group, batch = batch,
        study = study, time = time, indices_hash = indices_hash
      ))
    } else {
      indices_hash
    }
  }, error = function(e) NA_character_)

  info <- list(
    outcome = outcome, v = v, repeats = repeats_eff, seed = seed, mode = mode,
    group = group, batch = batch, study = study, time = time,
    primary_axis = primary_axis, secondary_axis = secondary_axis,
    constraints = constraints,
    stratify = stratify, nested = nested, horizon = horizon,
    purge = purge, embargo = embargo,
    summary = split_summary, hash = hash_val, inner = inner_indices,
    compact = compact, fold_assignments = fold_assignments,
    coldata = cd
  )

  result <- new("LeakSplits", mode = mode, indices = indices, info = info)

  if (.bio_is_strict() && !isTRUE(compact)) {
    check_split_overlap(result)
  }

  result
}


#' @title Display summary for LeakSplits objects
#' @description Prints fold counts, sizes, and hash metadata for quick inspection.
#' @param object LeakSplits object.
#' @return No return value, called for side effects (prints a summary to the
#'   console showing mode, fold count, repeats, outcome, stratification status,
#'   nested status, per-fold train/test sizes, and the reproducibility hash).
#' @examples
#' df <- data.frame(
#'   subject = rep(1:10, each = 2),
#'   outcome = rbinom(20, 1, 0.5),
#'   x1 = rnorm(20),
#'   x2 = rnorm(20)
#' )
#' splits <- make_split_plan(df, outcome = "outcome",
#'                       mode = "subject_grouped", group = "subject", v = 5)
#' show(splits)
#' @importMethodsFrom methods show
#' @export
setMethod("show", "LeakSplits", function(object) {
  cat(sprintf("LeakSplits object (mode = %s, v = %d, repeats = %d)\n",
              object@mode, object@info$v, object@info$repeats))
  cat(sprintf("Outcome: %s | Stratified: %s | Nested: %s\n",
              ifelse(is.null(object@info$outcome), "None", object@info$outcome),
              object@info$stratify, object@info$nested))
  cat("------------------------------------------------------\n")
  print(object@info$summary)
  cat("------------------------------------------------------\n")
  cat(sprintf("Total folds: %d | Hash: %s\n",
              nrow(object@info$summary), object@info$hash))
})


#' Check split overlap invariants
#'
#' Verifies that a \code{\linkS4class{LeakSplits}} object satisfies the
#' expected no-overlap constraints for one or more grouping columns. For each
#' fold, the function checks that no group-level value appearing in the test
#' set is also present in the training set.
#'
#' @param splits A \code{LeakSplits} object from \code{\link{make_split_plan}}.
#' @param coldata A data.frame of sample metadata. When \code{NULL} (default),
#'   the function uses \code{splits@@info$coldata} if available.
#' @param cols Character vector of column names to check for overlap. When
#'   \code{NULL} (default), the function infers columns from the split mode
#'   (e.g., \code{group} for \code{subject_grouped}, \code{batch} for
#'   \code{batch_blocked}, both axes for \code{combined}).
#' @return A data.frame with one row per fold-by-column combination and
#'   columns \code{fold}, \code{repeat_id}, \code{col}, \code{n_overlap}
#'   (number of overlapping group values), and \code{pass} (logical).
#'   Invisible. Raises an error if any fold fails and \code{stop_on_fail = TRUE}.
#' @param stop_on_fail Logical; if \code{TRUE} (default), raises an error when
#'   any overlap is detected.
#' @export
check_split_overlap <- function(splits, coldata = NULL, cols = NULL,
                                stop_on_fail = TRUE) {
  stopifnot(inherits(splits, "LeakSplits"))

  cd <- coldata %||% splits@info$coldata
  if (is.null(cd) || !is.data.frame(cd)) {
    .bio_stop("coldata must be provided or present in splits@info$coldata.",
              "bioLeak_input_error")
  }

  # Infer columns to check from split mode when not supplied
  if (is.null(cols)) {
    mode <- splits@mode
    cols <- character(0)
    if (mode == "subject_grouped" && !is.null(splits@info$group))
      cols <- c(cols, splits@info$group)
    if (mode == "batch_blocked" && !is.null(splits@info$batch))
      cols <- c(cols, splits@info$batch)
    if (mode == "study_loocv" && !is.null(splits@info$study))
      cols <- c(cols, splits@info$study)
    if (mode == "combined") {
      if (!is.null(splits@info$constraints)) {
        for (ax in splits@info$constraints) {
          if (!is.null(ax$col)) cols <- c(cols, ax$col)
        }
      } else {
        if (!is.null(splits@info$primary_axis$col))
          cols <- c(cols, splits@info$primary_axis$col)
        if (!is.null(splits@info$secondary_axis$col))
          cols <- c(cols, splits@info$secondary_axis$col)
      }
    }
    if (!length(cols)) {
      message("No grouping columns detected; nothing to check.")
      return(invisible(data.frame()))
    }
  }

  missing_cols <- setdiff(cols, names(cd))
  if (length(missing_cols)) {
    .bio_stop(sprintf("Column(s) not found in coldata: %s",
                      paste(missing_cols, collapse = ", ")),
              "bioLeak_column_error")
  }

  # Resolve explicit indices only (compact mode not supported)
  folds <- splits@indices
  has_explicit <- all(vapply(folds, function(f) !is.null(f$train), logical(1)))
  if (!has_explicit) {
    message("compact splits (no explicit train/test indices) cannot be checked; skipping.")
    return(invisible(data.frame()))
  }

  result_rows <- list()
  for (fi in seq_along(folds)) {
    f <- folds[[fi]]
    tr <- f$train
    te <- f$test
    for (col in cols) {
      train_vals <- unique(cd[[col]][tr])
      test_vals  <- unique(cd[[col]][te])
      overlap    <- intersect(train_vals, test_vals)
      result_rows[[length(result_rows) + 1L]] <- data.frame(
        fold       = f$fold %||% fi,
        repeat_id  = f$repeat_id %||% 1L,
        col        = col,
        n_overlap  = length(overlap),
        pass       = length(overlap) == 0L,
        stringsAsFactors = FALSE
      )
    }
  }

  result <- do.call(rbind, result_rows)
  failures <- result[!result$pass, , drop = FALSE]

  if (nrow(failures) > 0L && isTRUE(stop_on_fail)) {
    .bio_stop(sprintf(
      "Overlap detected in %d fold(s). First failure: fold=%s, col=%s, n_overlap=%d.",
      nrow(failures),
      failures$fold[[1]], failures$col[[1]], failures$n_overlap[[1]]
    ), "bioLeak_overlap_error")
  } else if (nrow(failures) > 0L) {
    .bio_warn(sprintf("Overlap detected in %d fold(s).", nrow(failures)),
              "bioLeak_overlap_error")
  }

  invisible(result)
}

test_that("make_split_plan respects grouping constraints", {
  set.seed(1)
  df <- data.frame(
    outcome = rep(c(0, 1), each = 10),
    subject = rep(1:10, each = 2),
    batch = rep(letters[1:4], length.out = 20),
    study = rep(LETTERS[1:5], length.out = 20),
    time = seq_len(20),
    x1 = rnorm(20),
    x2 = rnorm(20)
  )

  splits_subject <- make_split_plan(df, outcome = "outcome",
                                mode = "subject_grouped", group = "subject",
                                v = 5, repeats = 1, stratify = TRUE,
                                seed = 1, progress = FALSE)
  for (fold in splits_subject@indices) {
    tr <- df$subject[fold$train]
    te <- df$subject[fold$test]
    expect_equal(length(intersect(unique(tr), unique(te))), 0)
  }

  splits_batch <- make_split_plan(df, outcome = "outcome",
                              mode = "batch_blocked", batch = "batch",
                              v = 4, repeats = 1, stratify = FALSE,
                              seed = 1, progress = FALSE)
  for (fold in splits_batch@indices) {
    tr <- df$batch[fold$train]
    te <- df$batch[fold$test]
    expect_equal(length(intersect(unique(tr), unique(te))), 0)
  }

  splits_study <- make_split_plan(df, outcome = "outcome",
                              mode = "study_loocv", study = "study",
                              seed = 1, progress = FALSE)
  for (fold in splits_study@indices) {
    te_study <- unique(df$study[fold$test])
    expect_equal(length(te_study), 1)
    expect_false(te_study %in% unique(df$study[fold$train]))
  }

  splits_time <- make_split_plan(df, outcome = "outcome",
                             mode = "time_series", time = "time",
                             v = 4, horizon = 1, seed = 1, progress = FALSE)
  expect_true(length(splits_time@indices) > 0)
  for (fold in splits_time@indices) {
    tmin <- min(df$time[fold$test])
    expect_true(all(df$time[fold$train] <= (tmin - 1)))
  }
})

test_that("combined mode produces valid splits with no subject overlap AND no batch overlap", {
  set.seed(42)
  df <- data.frame(
    subject = rep(1:20, each = 3),
    batch = rep(rep(c("A", "B", "C", "D"), each = 5), 3),
    outcome = rep(c(0, 1), length.out = 60),
    x1 = rnorm(60),
    x2 = rnorm(60)
  )
  splits <- suppressWarnings(make_split_plan(
    df, outcome = "outcome",
    mode = "combined",
    primary_axis = list(type = "subject", col = "subject"),
    secondary_axis = list(type = "batch", col = "batch"),
    v = 3, progress = FALSE
  ))
  expect_s4_class(splits, "LeakSplits")
  expect_equal(splits@mode, "combined")

  # Check no subject overlap between train and test
  for (fold in splits@indices) {
    if (is.null(fold$train) || is.null(fold$test)) next
    train_subjects <- unique(df$subject[fold$train])
    test_subjects <- unique(df$subject[fold$test])
    expect_length(intersect(train_subjects, test_subjects), 0)

    # Check no batch overlap between train and test
    train_batches <- unique(df$batch[fold$train])
    test_batches <- unique(df$batch[fold$test])
    expect_length(intersect(train_batches, test_batches), 0)
  }
})

test_that("error when primary_axis or secondary_axis missing", {
  df <- data.frame(subject = 1:10, batch = rep("A", 10),
                   outcome = rbinom(10, 1, 0.5), x1 = rnorm(10))
  expect_error(
    make_split_plan(df, outcome = "outcome", mode = "combined",
                    primary_axis = list(type = "subject", col = "subject"),
                    v = 3, progress = FALSE),
    "secondary_axis"
  )
  expect_error(
    make_split_plan(df, outcome = "outcome", mode = "combined",
                    secondary_axis = list(type = "batch", col = "batch"),
                    v = 3, progress = FALSE),
    "primary_axis"
  )
})

test_that("error when axis list is malformed", {
  df <- data.frame(subject = 1:10, batch = rep("A", 10),
                   outcome = rbinom(10, 1, 0.5), x1 = rnorm(10))
  # not a list
  expect_error(
    make_split_plan(df, outcome = "outcome", mode = "combined",
                    primary_axis = "subject",
                    secondary_axis = list(type = "batch", col = "batch"),
                    v = 3, progress = FALSE),
    "primary_axis"
  )
  # missing type
  expect_error(
    make_split_plan(df, outcome = "outcome", mode = "combined",
                    primary_axis = list(col = "subject"),
                    secondary_axis = list(type = "batch", col = "batch"),
                    v = 3, progress = FALSE),
    "primary_axis"
  )
})

test_that("compact = TRUE produces fold_assignments", {
  set.seed(42)
  df <- data.frame(
    subject = rep(1:20, each = 3),
    batch = rep(rep(c("A", "B", "C", "D"), each = 5), 3),
    outcome = rbinom(60, 1, 0.5),
    x1 = rnorm(60),
    x2 = rnorm(60)
  )
  splits <- suppressWarnings(make_split_plan(
    df, outcome = "outcome",
    mode = "combined",
    primary_axis = list(type = "subject", col = "subject"),
    secondary_axis = list(type = "batch", col = "batch"),
    v = 3, compact = TRUE, progress = FALSE
  ))
  expect_true(isTRUE(splits@info$compact))
  expect_true(length(splits@info$fold_assignments) > 0)
})

test_that("combined splits pass through to fit_resample successfully", {
  set.seed(42)
  df <- data.frame(
    subject = rep(1:20, each = 3),
    batch = rep(rep(c("A", "B", "C", "D"), each = 5), 3),
    outcome = rbinom(60, 1, 0.5),
    x1 = rnorm(60),
    x2 = rnorm(60)
  )
  splits <- suppressWarnings(make_split_plan(
    df, outcome = "outcome",
    mode = "combined",
    primary_axis = list(type = "subject", col = "subject"),
    secondary_axis = list(type = "batch", col = "batch"),
    v = 3, progress = FALSE
  ))
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "glmnet",
                            preprocess = list(
                              impute = list(method = "median"),
                              normalize = list(method = "zscore")
                            ))
  expect_s4_class(fit, "LeakFit")
  expect_gt(nrow(fit@metrics), 0)
})

test_that("stratification works with combined mode", {
  set.seed(42)
  df <- data.frame(
    subject = rep(1:20, each = 3),
    batch = rep(rep(c("A", "B", "C", "D"), each = 5), 3),
    outcome = rep(c(0, 1), 30),
    x1 = rnorm(60),
    x2 = rnorm(60)
  )
  splits <- suppressWarnings(make_split_plan(
    df, outcome = "outcome",
    mode = "combined",
    primary_axis = list(type = "subject", col = "subject"),
    secondary_axis = list(type = "batch", col = "batch"),
    v = 3, stratify = TRUE, progress = FALSE
  ))
  expect_s4_class(splits, "LeakSplits")
  expect_gt(length(splits@indices), 0)
})

test_that("check_split_overlap passes for correct subject_grouped splits", {
  set.seed(1)
  df <- data.frame(
    subject = rep(1:10, each = 2),
    outcome = rbinom(20, 1, 0.5),
    x1 = rnorm(20)
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, progress = FALSE)
  result <- check_split_overlap(splits)
  expect_true(all(result$pass))
})

test_that("check_split_overlap detects injected overlap", {
  # Manually construct a splits object with deliberate subject overlap
  indices <- list(
    list(train = 1:8, test = 7:10, fold = 1L, repeat_id = 1L)
  )
  cd <- data.frame(subject = c(1,1,2,2,3,3,4,4,5,5))
  splits_bad <- bioLeak:::LeakSplits(
    mode = "subject_grouped",
    indices = indices,
    info = list(v = 1L, repeats = 1L, seed = 1L, mode = "subject_grouped",
                group = "subject", batch = NULL, study = NULL, time = NULL,
                primary_axis = NULL, secondary_axis = NULL,
                stratify = FALSE, nested = FALSE,
                horizon = 0, purge = 0, embargo = 0,
                summary = data.frame(fold=1L, repeat_id=1L, train_n=8L, test_n=4L),
                hash = NA_character_, inner = NULL,
                compact = FALSE, fold_assignments = NULL,
                coldata = cd)
  )
  expect_error(
    check_split_overlap(splits_bad, coldata = cd, cols = "subject"),
    "Overlap detected"
  )
})

test_that("check_split_overlap stop_on_fail=FALSE returns warning not error", {
  indices <- list(
    list(train = 1:8, test = 7:10, fold = 1L, repeat_id = 1L)
  )
  cd <- data.frame(subject = c(1,1,2,2,3,3,4,4,5,5))
  splits_bad <- bioLeak:::LeakSplits(
    mode = "subject_grouped",
    indices = indices,
    info = list(v = 1L, repeats = 1L, seed = 1L, mode = "subject_grouped",
                group = "subject", batch = NULL, study = NULL, time = NULL,
                primary_axis = NULL, secondary_axis = NULL,
                stratify = FALSE, nested = FALSE,
                horizon = 0, purge = 0, embargo = 0,
                summary = data.frame(fold=1L, repeat_id=1L, train_n=8L, test_n=4L),
                hash = NA_character_, inner = NULL,
                compact = FALSE, fold_assignments = NULL,
                coldata = cd)
  )
  expect_warning(
    result <- check_split_overlap(splits_bad, coldata = cd, cols = "subject",
                                  stop_on_fail = FALSE),
    "Overlap detected"
  )
  expect_false(all(result$pass))
})

test_that("check_split_overlap works for combined mode and checks both axes", {
  set.seed(42)
  df <- data.frame(
    subject = rep(1:20, each = 3),
    batch = rep(rep(c("A","B","C","D"), each=5), 3),
    outcome = rbinom(60, 1, 0.5),
    x1 = rnorm(60)
  )
  splits <- suppressWarnings(make_split_plan(
    df, outcome = "outcome",
    mode = "combined",
    primary_axis = list(type = "subject", col = "subject"),
    secondary_axis = list(type = "batch", col = "batch"),
    v = 3, progress = FALSE
  ))
  result <- check_split_overlap(splits)
  # Both subject and batch columns should show no overlap
  expect_true(all(result$pass))
  expect_true(all(c("subject", "batch") %in% result$col))
})

test_that("check_split_overlap returns a data.frame with expected columns", {
  set.seed(1)
  df <- data.frame(
    subject = rep(1:10, each = 2),
    outcome = rbinom(20, 1, 0.5),
    x1 = rnorm(20)
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, progress = FALSE)
  result <- check_split_overlap(splits)
  expect_true(is.data.frame(result))
  expect_true(all(c("fold", "repeat_id", "col", "n_overlap", "pass") %in% names(result)))
})

test_that("LeakAudit info slot carries provenance", {
  set.seed(42)
  df <- data.frame(
    subject = rep(1:30, each = 3),
    outcome = rep(c(0, 1), length.out = 90),
    x1 = rnorm(90),
    x2 = rnorm(90)
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, progress = FALSE)
  fit <- fit_resample(df, outcome = "outcome", splits = splits,
                      learner = "glmnet",
                      preprocess = list(
                        impute = list(method = "median"),
                        normalize = list(method = "zscore")
                      ))
  audit <- audit_leakage(fit, B = 5)
  expect_true("provenance" %in% names(audit@info))
  expect_true(is.list(audit@info$provenance))
})

test_that("fold_status contains elapsed_sec column", {
  set.seed(1)
  df <- data.frame(
    subject = rep(1:30, each = 2),
    outcome = rep(c(0, 1), length.out = 60),
    x1 = rnorm(60),
    x2 = rnorm(60)
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, progress = FALSE)
  fit <- fit_resample(df, outcome = "outcome", splits = splits,
                      learner = "glmnet",
                      preprocess = list(
                        impute = list(method = "median"),
                        normalize = list(method = "zscore")
                      ))
  fs <- fit@info$fold_status
  expect_true("elapsed_sec" %in% names(fs))
})

test_that("successful folds have non-negative elapsed_sec", {
  set.seed(1)
  df <- data.frame(
    subject = rep(1:30, each = 2),
    outcome = rep(c(0, 1), length.out = 60),
    x1 = rnorm(60),
    x2 = rnorm(60)
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, progress = FALSE)
  fit <- fit_resample(df, outcome = "outcome", splits = splits,
                      learner = "glmnet",
                      preprocess = list(
                        impute = list(method = "median"),
                        normalize = list(method = "zscore")
                      ))
  fs <- fit@info$fold_status
  success_rows <- fs[fs$status == "success", ]
  expect_true(all(!is.na(success_rows$elapsed_sec)))
  expect_true(all(success_rows$elapsed_sec >= 0))
})

test_that("schema consistency between successful and failed folds", {
  set.seed(1)
  df <- data.frame(
    subject = rep(1:30, each = 2),
    outcome = rep(c(0, 1), length.out = 60),
    x1 = rnorm(60),
    x2 = rnorm(60)
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, progress = FALSE)
  fit <- fit_resample(df, outcome = "outcome", splits = splits,
                      learner = "glmnet",
                      preprocess = list(
                        impute = list(method = "median"),
                        normalize = list(method = "zscore")
                      ))
  fs <- fit@info$fold_status
  expected_cols <- c("fold", "stage", "status", "reason", "notes", "elapsed_sec")
  expect_true(all(expected_cols %in% names(fs)))
})

# tests/testthat/test-strict-mode.R
# Tests for options(bioLeak.strict = TRUE) strict leakage mode

test_that("strict mode forces bioLeak:::.bio_validation_mode() to return 'error'", {
  withr::local_options(list(bioLeak.strict = TRUE))
  expect_equal(bioLeak:::.bio_validation_mode(), "error")
  expect_equal(bioLeak:::.bio_validation_mode("warn"), "error")
  expect_equal(bioLeak:::.bio_validation_mode("off"), "error")
})

test_that("strict mode warns when seed is NULL", {
  withr::local_options(list(bioLeak.strict = TRUE))
  expect_warning(
    bioLeak:::.bio_strict_checks(context = "test", seed = NULL),
    "\\[strict\\]:.*seed"
  )
  expect_warning(
    bioLeak:::.bio_strict_checks(context = "test", seed = NA),
    "\\[strict\\]:.*seed"
  )
})

test_that("strict mode warns on nested + combined", {
  withr::local_options(list(bioLeak.strict = TRUE))
  expect_warning(
    bioLeak:::.bio_strict_checks(context = "test", seed = 1, nested = TRUE, mode = "combined"),
    "\\[strict\\]:.*nested.*combined"
  )
})

test_that("strict mode auto-runs check_split_overlap (correct splits succeed)", {
  withr::local_options(list(bioLeak.strict = TRUE))
  set.seed(1)
  df <- data.frame(
    subject = rep(1:10, each = 2),
    outcome = rbinom(20, 1, 0.5),
    x1 = rnorm(20)
  )
  # Should succeed without error — splits are correct
  expect_no_error(
    make_split_plan(df, outcome = "outcome",
                    mode = "subject_grouped", group = "subject",
                    v = 5, seed = 1, progress = FALSE)
  )
})

test_that("strict mode rejects trained recipes in fit_resample", {
  skip_if_not_installed("recipes")
  withr::local_options(list(bioLeak.strict = TRUE))

  set.seed(1)
  df <- data.frame(
    subject = rep(1:6, each = 2),
    outcome = factor(rep(c(0, 1), each = 6)),
    x1 = rnorm(12),
    x2 = rnorm(12)
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, seed = 1, progress = FALSE)

  rec <- recipes::recipe(outcome ~ x1 + x2, data = df)
  trained_rec <- recipes::prep(rec, training = df)

  custom <- list(
    glm = list(
      fit = function(x, y, task, weights, ...) {
        stats::glm(y ~ ., data = as.data.frame(x),
                   family = stats::binomial(), weights = weights)
      },
      predict = function(object, newdata, task, ...) {
        as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
                                  type = "response"))
      }
    )
  )

  # Strict mode forces validation_mode to "error", so trained recipe should error

  expect_error(
    fit_resample(df, outcome = "outcome", splits = splits,
                 learner = "glm", custom_learners = custom,
                 preprocess = trained_rec, metrics = "auc",
                 refit = FALSE, seed = 1),
    "trained recipe"
  )
})

test_that("constraints with 2 axes produces identical results to primary_axis/secondary_axis", {
  set.seed(42)
  df <- data.frame(
    subject = rep(1:20, each = 3),
    batch = rep(rep(c("A", "B", "C", "D"), each = 5), 3),
    outcome = rbinom(60, 1, 0.5),
    x1 = rnorm(60),
    x2 = rnorm(60)
  )

  splits_legacy <- suppressWarnings(make_split_plan(
    df, outcome = "outcome",
    mode = "combined",
    primary_axis = list(type = "subject", col = "subject"),
    secondary_axis = list(type = "batch", col = "batch"),
    v = 3, seed = 1, progress = FALSE
  ))

  splits_new <- suppressWarnings(make_split_plan(
    df, outcome = "outcome",
    mode = "combined",
    constraints = list(
      list(type = "subject", col = "subject"),
      list(type = "batch", col = "batch")
    ),
    v = 3, seed = 1, progress = FALSE
  ))

  expect_equal(length(splits_legacy@indices), length(splits_new@indices))
  for (i in seq_along(splits_legacy@indices)) {
    expect_equal(splits_legacy@indices[[i]]$train, splits_new@indices[[i]]$train)
    expect_equal(splits_legacy@indices[[i]]$test, splits_new@indices[[i]]$test)
  }
})

test_that("3-axis constraint (subject + batch + site) enforces all three", {
  set.seed(42)
  # Design: 9 groups of 5 subjects each. Each group has a unique batch-site pair.
  # 3 batches x 3 sites, nested so batch i maps to sites {i} only.
  # This makes the 3 axes perfectly aligned in 3 blocks, enabling v=3 splits.
  n_groups <- 9
  subjects_per_group <- 5
  group_batch <- rep(paste0("B", 1:3), each = 3)
  group_site <- rep(paste0("S", 1:3), times = 3)
  df <- data.frame(
    subject = rep(seq_len(n_groups), each = subjects_per_group),
    batch = rep(group_batch, each = subjects_per_group),
    site = rep(group_site, each = subjects_per_group),
    outcome = rbinom(n_groups * subjects_per_group, 1, 0.5),
    x1 = rnorm(n_groups * subjects_per_group),
    stringsAsFactors = FALSE
  )
  splits <- suppressWarnings(make_split_plan(
    df, outcome = "outcome",
    mode = "combined",
    constraints = list(
      list(type = "subject", col = "subject"),
      list(type = "batch", col = "batch"),
      list(type = "study", col = "site")
    ),
    v = 3, progress = FALSE
  ))
  expect_s4_class(splits, "LeakSplits")

  for (fold in splits@indices) {
    if (is.null(fold$train) || is.null(fold$test)) next
    # No subject overlap
    expect_length(intersect(unique(df$subject[fold$train]),
                            unique(df$subject[fold$test])), 0)
    # No batch overlap
    expect_length(intersect(unique(df$batch[fold$train]),
                            unique(df$batch[fold$test])), 0)
    # No site overlap
    expect_length(intersect(unique(df$site[fold$train]),
                            unique(df$site[fold$test])), 0)
  }
})

test_that("check_split_overlap validates all N axes", {
  set.seed(42)
  n_groups <- 9
  subjects_per_group <- 5
  group_batch <- rep(paste0("B", 1:3), each = 3)
  group_site <- rep(paste0("S", 1:3), times = 3)
  df <- data.frame(
    subject = rep(seq_len(n_groups), each = subjects_per_group),
    batch = rep(group_batch, each = subjects_per_group),
    site = rep(group_site, each = subjects_per_group),
    outcome = rbinom(n_groups * subjects_per_group, 1, 0.5),
    x1 = rnorm(n_groups * subjects_per_group),
    stringsAsFactors = FALSE
  )
  splits <- suppressWarnings(make_split_plan(
    df, outcome = "outcome",
    mode = "combined",
    constraints = list(
      list(type = "subject", col = "subject"),
      list(type = "batch", col = "batch"),
      list(type = "study", col = "site")
    ),
    v = 3, progress = FALSE
  ))
  result <- check_split_overlap(splits)
  expect_true(all(result$pass))
  expect_true(all(c("subject", "batch", "site") %in% result$col))
})

test_that("legacy primary_axis/secondary_axis still works", {
  set.seed(42)
  df <- data.frame(
    subject = rep(1:20, each = 3),
    batch = rep(rep(c("A", "B", "C", "D"), each = 5), 3),
    outcome = rbinom(60, 1, 0.5),
    x1 = rnorm(60)
  )
  splits <- suppressWarnings(make_split_plan(
    df, outcome = "outcome",
    mode = "combined",
    primary_axis = list(type = "subject", col = "subject"),
    secondary_axis = list(type = "batch", col = "batch"),
    v = 3, progress = FALSE
  ))
  expect_s4_class(splits, "LeakSplits")
  # constraints should be populated in info
  expect_equal(length(splits@info$constraints), 2)
})

test_that("error when both constraints and primary_axis provided", {
  df <- data.frame(subject = 1:10, batch = rep("A", 10),
                   outcome = rbinom(10, 1, 0.5), x1 = rnorm(10))
  expect_error(
    make_split_plan(df, outcome = "outcome", mode = "combined",
                    primary_axis = list(type = "subject", col = "subject"),
                    constraints = list(
                      list(type = "subject", col = "subject"),
                      list(type = "batch", col = "batch")
                    ),
                    v = 3, progress = FALSE),
    "Cannot specify both"
  )
})

test_that("error when constraints has fewer than 2 elements", {
  df <- data.frame(subject = 1:10, batch = rep("A", 10),
                   outcome = rbinom(10, 1, 0.5), x1 = rnorm(10))
  expect_error(
    make_split_plan(df, outcome = "outcome", mode = "combined",
                    constraints = list(
                      list(type = "subject", col = "subject")
                    ),
                    v = 3, progress = FALSE),
    "at least 2 elements"
  )
})

test_that("error when constraint col doesn't exist", {
  df <- data.frame(subject = 1:10, batch = rep("A", 10),
                   outcome = rbinom(10, 1, 0.5), x1 = rnorm(10))
  expect_error(
    make_split_plan(df, outcome = "outcome", mode = "combined",
                    constraints = list(
                      list(type = "subject", col = "subject"),
                      list(type = "batch", col = "nonexistent")
                    ),
                    v = 3, progress = FALSE),
    "not found"
  )
})

test_that("constraints info slot stores all axes", {
  set.seed(42)
  n_groups <- 9
  subjects_per_group <- 5
  group_batch <- rep(paste0("B", 1:3), each = 3)
  group_site <- rep(paste0("S", 1:3), times = 3)
  df <- data.frame(
    subject = rep(seq_len(n_groups), each = subjects_per_group),
    batch = rep(group_batch, each = subjects_per_group),
    site = rep(group_site, each = subjects_per_group),
    outcome = rbinom(n_groups * subjects_per_group, 1, 0.5),
    x1 = rnorm(n_groups * subjects_per_group),
    stringsAsFactors = FALSE
  )
  splits <- suppressWarnings(make_split_plan(
    df, outcome = "outcome",
    mode = "combined",
    constraints = list(
      list(type = "subject", col = "subject"),
      list(type = "batch", col = "batch"),
      list(type = "study", col = "site")
    ),
    v = 3, progress = FALSE
  ))
  expect_equal(length(splits@info$constraints), 3)
  # backward compat: primary_axis and secondary_axis still populated
  expect_equal(splits@info$primary_axis$col, "subject")
  expect_equal(splits@info$secondary_axis$col, "batch")
})

test_that("make_split_plan validates required columns and time types", {
  df <- make_class_df(6)
  expect_error(make_split_plan_quiet(df, outcome = "outcome",
                                 mode = "subject_grouped", group = NULL,
                                 v = 2, seed = 1),
               "requires a non-NULL 'group'")
  expect_error(make_split_plan_quiet(df, outcome = "outcome",
                                 mode = "batch_blocked", batch = "missing"),
               "Column 'missing' not found")
  expect_error(make_split_plan_quiet(df, outcome = "outcome",
                                 mode = "study_loocv", study = "missing"),
               "Column 'missing' not found")
  expect_error(make_split_plan_quiet(df, outcome = "outcome",
                                 mode = "time_series", time = "missing"),
               "Column 'missing' not found")

  df$time <- as.character(df$time)
  expect_error(make_split_plan_quiet(df, outcome = "outcome",
                                 mode = "time_series", time = "time"),
               "time' column must be numeric")
})

test_that("make_split_plan warns on stratification issues and compact/nested conflicts", {
  df <- make_class_df(10)
  expect_warning_match(
    make_split_plan_quiet(df, outcome = "missing",
                      mode = "subject_grouped", group = "subject",
                      v = 2, stratify = TRUE),
    "Stratification requested"
  )

  df$outcome <- 1
  expect_warning_match(
    make_split_plan_quiet(df, outcome = "outcome",
                      mode = "subject_grouped", group = "subject",
                      v = 2, stratify = TRUE),
    "Stratification ignored"
  )

  splits <- expect_warning_match(
    make_split_plan_quiet(df, outcome = "outcome",
                      mode = "subject_grouped", group = "subject",
                      v = 2, nested = TRUE, compact = TRUE),
    "compact=TRUE is not supported"
  )
  expect_false(splits@info$compact)
})

test_that("make_split_plan supports compact assignments and nested splits", {
  df <- make_class_df(12)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 3, compact = TRUE, seed = 1)
  expect_true(isTRUE(splits@info$compact))
  expect_true(length(splits@info$fold_assignments) > 0)
  expect_true(all(vapply(splits@indices, function(x) !is.null(x$fold), logical(1))))

  nested <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 3, nested = TRUE, seed = 1)
  expect_true(is.list(nested@info$inner))
  expect_true(length(nested@info$inner) == length(nested@indices))
})

test_that("make_split_plan errors when no valid folds are produced", {
  df <- make_class_df(4)
  expect_error(make_split_plan_quiet(df, outcome = "outcome",
                                 mode = "time_series", time = "time",
                                 v = 2, horizon = 100, seed = 1),
               "No valid folds generated")
})

test_that("make_split_plan time_series includes tail samples when n not divisible by v", {
  df <- make_class_df(10)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "time_series", time = "time",
                              v = 3, horizon = 0, seed = 1)
  test_idx <- sort(unique(unlist(lapply(splits@indices, `[[`, "test"))))
  expect_true(max(test_idx) == nrow(df))
})

test_that("make_split_plan time_series applies purge and embargo gaps", {
  df <- make_class_df(12)

  base <- make_split_plan_quiet(df, outcome = "outcome",
                           mode = "time_series", time = "time",
                           v = 3, horizon = 0, seed = 1)
  with_purge <- make_split_plan_quiet(df, outcome = "outcome",
                                 mode = "time_series", time = "time",
                                 v = 3, horizon = 0, purge = 2, seed = 1)
  with_embargo <- make_split_plan_quiet(df, outcome = "outcome",
                                   mode = "time_series", time = "time",
                                   v = 3, horizon = 0, embargo = 5, seed = 1)

  fold2_base <- base@indices[[which(vapply(base@indices, function(z) z$fold == 2L, logical(1)))]]
  fold2_purge <- with_purge@indices[[which(vapply(with_purge@indices, function(z) z$fold == 2L, logical(1)))]]
  fold2_embargo <- with_embargo@indices[[which(vapply(with_embargo@indices, function(z) z$fold == 2L, logical(1)))]]

  expect_equal(max(df$time[fold2_base$train]), 4)
  expect_equal(max(df$time[fold2_purge$train]), 3)
  expect_equal(max(df$time[fold2_embargo$train]), 3)
  expect_equal(with_purge@info$purge, 2)
  expect_equal(with_purge@info$embargo, 0)
  expect_equal(with_embargo@info$embargo, 5)
})

test_that("make_split_plan nested + combined propagates constraints", {
  set.seed(1)
  # 16 batches × 3 subjects/batch = 48 subjects, 2 obs/subject = 96 obs.
  # seed = 6 produces a fold assignment where outer and inner both succeed.
  df <- data.frame(
    subject = rep(1:48, each = 2),
    batch   = rep(paste0("B", sprintf("%02d", rep(1:16, each = 3))), each = 2),
    outcome = rep(c(0, 1), 48),
    x1      = rnorm(96)
  )
  constraints <- list(
    list(type = "subject", col = "subject"),
    list(type = "batch",   col = "batch")
  )
  # Before fix: recursive call dropped constraints → "'primary_axis' must be a list"
  splits <- make_split_plan_quiet(df, outcome = "outcome", mode = "combined",
    constraints = constraints, v = 2, nested = TRUE, seed = 6)
  expect_true(inherits(splits, "LeakSplits"))
  expect_true(is.list(splits@info$inner))
  expect_true(length(splits@info$inner) > 0)
})

test_that("make_split_plan supports matrix and SummarizedExperiment inputs", {
  mat <- matrix(rnorm(20), nrow = 10)
  splits <- make_split_plan(mat, outcome = NULL,
                        mode = "subject_grouped", group = "row_id",
                        v = 2, seed = 1)
  expect_true(inherits(splits, "LeakSplits"))
  expect_true("row_id" %in% names(splits@info$coldata))

  skip_if_not_installed("SummarizedExperiment")
  se <- SummarizedExperiment::SummarizedExperiment(
    assays = list(expr = matrix(rnorm(40), nrow = 4)),
    colData = data.frame(outcome = rep(0:1, each = 5),
                         subject = rep(1:5, 2))
  )
  splits_se <- make_split_plan(se, outcome = "outcome",
                           mode = "subject_grouped", group = "subject",
                           v = 2, seed = 1, progress = FALSE)
  expect_true(inherits(splits_se, "LeakSplits"))
})

# Cross-validation uncertainty estimation -----------------------------------

#' Nadeau-Bengio corrected variance
#'
#' Computes the corrected variance for repeated K-fold CV as described in
#' Nadeau & Bengio (2003). The correction factor accounts for the overlap
#' between training sets across folds.
#'
#' @param per_fold_vals Numeric vector of per-fold metric values.
#' @param n_train Average number of training samples per fold. NULL to skip
#'   correction.
#' @param n_test Average number of test samples per fold. NULL to skip
#'   correction.
#' @return Scalar corrected variance, or \code{NA_real_} for single-fold input.
#' @keywords internal
.nb_corrected_var <- function(per_fold_vals, n_train = NULL, n_test = NULL) {
  per_fold_vals <- per_fold_vals[is.finite(per_fold_vals)]
  K <- length(per_fold_vals)
  if (K < 2L) return(NA_real_)
  v <- stats::var(per_fold_vals)
  if (!is.null(n_train) && !is.null(n_test) &&
      is.finite(n_train) && is.finite(n_test) && n_train > 0) {
    # Nadeau-Bengio correction: var_corrected = (1/K + n_test/n_train) * var
    corrected <- (1 / K + n_test / n_train) * v
  } else {
    # Fallback: simple var/K
    corrected <- v / K
  }
  corrected
}

#' Confidence intervals for cross-validated metrics
#'
#' Computes per-learner confidence intervals for each metric column in a
#' per-fold metrics data.frame. Supports the standard normal/t approach and the
#' Nadeau-Bengio (2003) corrected variance for repeated K-fold CV.
#'
#' @param metrics_df Data.frame with columns \code{fold}, \code{learner}, and
#'   one or more numeric metric columns.
#' @param level Confidence level (default 0.95).
#' @param method One of \code{"normal"} or \code{"nadeau_bengio"}.
#' @param n_train Average number of training samples per fold. Used only when
#'   \code{method = "nadeau_bengio"}. NULL to use fallback variance.
#' @param n_test Average number of test samples per fold. Used only when
#'   \code{method = "nadeau_bengio"}. NULL to use fallback variance.
#' @return A data.frame with \code{learner} and, for each metric, columns
#'   \code{<metric>_mean}, \code{<metric>_sd}, \code{<metric>_ci_lo}, and
#'   \code{<metric>_ci_hi}.
#' @export
cv_ci <- function(metrics_df, level = 0.95,
                  method = c("normal", "nadeau_bengio"),
                  n_train = NULL, n_test = NULL) {
  method <- match.arg(method)
  stopifnot(is.data.frame(metrics_df))
  stopifnot("fold" %in% names(metrics_df), "learner" %in% names(metrics_df))
  stopifnot(level > 0 && level < 1)

  metric_cols <- setdiff(names(metrics_df), c("fold", "learner"))
  if (!length(metric_cols)) {
    .bio_stop("No numeric metric columns found in metrics_df.",
              "bioLeak_input_error")
  }

  learners <- unique(metrics_df$learner)
  result_rows <- vector("list", length(learners))

  for (li in seq_along(learners)) {
    ln <- learners[[li]]
    sub <- metrics_df[metrics_df$learner == ln, , drop = FALSE]
    K <- nrow(sub)
    row <- list(learner = ln)

    for (mc in metric_cols) {
      vals <- sub[[mc]]
      if (!is.numeric(vals)) next
      vals_finite <- vals[is.finite(vals)]
      K_eff <- length(vals_finite)
      m <- if (K_eff > 0L) mean(vals_finite) else NA_real_
      s <- if (K_eff > 1L) stats::sd(vals_finite) else NA_real_
      row[[paste0(mc, "_mean")]] <- m
      row[[paste0(mc, "_sd")]] <- s

      if (K_eff < 2L) {
        row[[paste0(mc, "_ci_lo")]] <- NA_real_
        row[[paste0(mc, "_ci_hi")]] <- NA_real_
        next
      }

      if (identical(method, "nadeau_bengio")) {
        corrected_var <- .nb_corrected_var(vals_finite, n_train = n_train, n_test = n_test)
        se <- sqrt(corrected_var)
      } else {
        se <- s / sqrt(K_eff)
      }

      # Use t-distribution for K_eff < 30, normal otherwise
      alpha <- 1 - level
      if (K_eff < 30) {
        q <- stats::qt(1 - alpha / 2, df = K_eff - 1)
      } else {
        q <- stats::qnorm(1 - alpha / 2)
      }

      row[[paste0(mc, "_ci_lo")]] <- m - q * se
      row[[paste0(mc, "_ci_hi")]] <- m + q * se
    }

    result_rows[[li]] <- as.data.frame(row, stringsAsFactors = FALSE)
  }

  do.call(rbind, result_rows)
}

test_that("cv_ci returns correct column names", {
  df <- data.frame(
    fold = 1:5,
    learner = rep("glmnet", 5),
    auc = runif(5, 0.6, 0.9),
    accuracy = runif(5, 0.7, 0.95)
  )
  res <- cv_ci(df)
  expect_true("learner" %in% names(res))
  expect_true(all(c("auc_mean", "auc_sd", "auc_ci_lo", "auc_ci_hi") %in% names(res)))
  expect_true(all(c("accuracy_mean", "accuracy_sd", "accuracy_ci_lo", "accuracy_ci_hi") %in% names(res)))
})

test_that("CI bounds satisfy ci_lo < mean < ci_hi", {
  set.seed(1)
  df <- data.frame(
    fold = 1:10,
    learner = rep("glmnet", 10),
    auc = runif(10, 0.6, 0.9)
  )
  res <- cv_ci(df)
  expect_true(res$auc_ci_lo < res$auc_mean)
  expect_true(res$auc_mean < res$auc_ci_hi)
})

test_that("Nadeau-Bengio CIs are wider than normal CIs", {
  set.seed(1)
  df <- data.frame(
    fold = 1:10,
    learner = rep("glmnet", 10),
    auc = runif(10, 0.6, 0.9)
  )
  res_nb <- cv_ci(df, method = "nadeau_bengio", n_train = 90, n_test = 10)
  res_normal <- cv_ci(df, method = "normal")
  nb_width <- res_nb$auc_ci_hi - res_nb$auc_ci_lo
  normal_width <- res_normal$auc_ci_hi - res_normal$auc_ci_lo
  expect_gt(nb_width, normal_width)
})

test_that("single-fold input returns NA for CIs", {
  df <- data.frame(fold = 1, learner = "glmnet", auc = 0.8)
  res <- cv_ci(df)
  expect_true(is.na(res$auc_ci_lo))
  expect_true(is.na(res$auc_ci_hi))
})

test_that("multiple learners produce separate rows", {
  df <- data.frame(
    fold = rep(1:5, 2),
    learner = rep(c("glmnet", "ranger"), each = 5),
    auc = runif(10, 0.6, 0.9)
  )
  res <- cv_ci(df)
  expect_equal(nrow(res), 2)
  expect_true(all(c("glmnet", "ranger") %in% res$learner))
})

test_that("fit_resample metric_summary gains _ci_lo/_ci_hi columns", {
  set.seed(1)
  df <- data.frame(
    subject = rep(1:30, each = 2),
    outcome = rep(c(0, 1), length.out = 60),
    x1 = rnorm(60),
    x2 = rnorm(60)
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, progress = FALSE)
  fit <- fit_resample(df, outcome = "outcome", splits = splits,
                      learner = "glmnet",
                      preprocess = list(
                        impute = list(method = "median"),
                        normalize = list(method = "zscore")
                      ))
  ms <- fit@metric_summary
  ci_cols <- grep("_ci_lo$|_ci_hi$", names(ms), value = TRUE)
  expect_gt(length(ci_cols), 0)
})

test_that(".nb_corrected_var returns NA for single fold", {
  expect_true(is.na(bioLeak:::.nb_corrected_var(0.8)))
})

test_that(".nb_corrected_var without n_train falls back to var/K", {
  vals <- c(0.7, 0.8, 0.9, 0.75, 0.85)
  K <- length(vals)
  res <- bioLeak:::.nb_corrected_var(vals, n_train = NULL, n_test = NULL)
  expected <- var(vals) / K
  expect_equal(res, expected)
})

test_that("cv_ci handles NA fold metrics correctly", {
  df <- data.frame(
    fold = 1:5,
    learner = rep("glm", 5),
    auc = c(0.9, NA, NA, NA, 0.7)
  )

  # Normal method: CI should be based on K_eff = 2, not K = 5
  res_n <- cv_ci(df, method = "normal")
  expect_equal(res_n$auc_mean, 0.8, tolerance = 1e-10)
  # With only 2 informative folds, df = 1, CI should be wide
  expect_true(all(is.finite(c(res_n$auc_ci_lo, res_n$auc_ci_hi))))

  # Nadeau-Bengio method: should also produce finite CIs
  res_nb <- cv_ci(df, method = "nadeau_bengio", n_train = 80, n_test = 20)
  expect_true(all(is.finite(c(res_nb$auc_ci_lo, res_nb$auc_ci_hi))))
  expect_equal(res_nb$auc_mean, 0.8, tolerance = 1e-10)
})

test_that("cv_ci returns NA CIs when only one finite fold", {
  df <- data.frame(
    fold = 1:3,
    learner = rep("glm", 3),
    auc = c(0.8, NA, NA)
  )
  res <- cv_ci(df, method = "normal")
  expect_true(is.na(res$auc_ci_lo))
  expect_true(is.na(res$auc_ci_hi))
})

# Time-series block permutation utilities --------------------------------------

#' Circular block permutation indices
#'
#' Generates a permutation of time indices by concatenating random-length blocks
#' sampled circularly from the ordered sequence. Used for creating
#' block-permuted surrogates that preserve short-range temporal structure.
#'
#' @param idx Integer vector of ordered indices.
#' @param block_len Positive integer block length (>= 1).
#'
#' @return Integer vector of permuted indices of the same length as `idx`.
#'
#' @keywords internal
.circular_block_permute <- function(idx, block_len) {
  stopifnot(length(idx) > 0L, is.numeric(block_len), block_len > 0L)
  n <- length(idx)
  block_len <- max(1L, as.integer(round(block_len)))

  out <- integer(0L)
  while (length(out) < n) {
    start <- sample.int(n, 1L)
    block_seq <- ((start - 1L) + seq_len(block_len)) %% n + 1L
    out <- c(out, idx[block_seq])
  }

  out <- out[seq_len(n)]
  stopifnot(length(out) == n)
  out
}


#' Stationary bootstrap indices
#'
#' Implements the stationary bootstrap of Politis & Romano (1994),
#' which resamples contiguous blocks of variable length to preserve
#' weak temporal dependence while maintaining ergodicity.
#'
#' @param idx Integer vector of ordered indices.
#' @param mean_block Positive numeric, expected block length.
#'
#' @return Integer vector of permuted indices of the same length as `idx`.
#'
#' @references
#' Politis, D. N., & Romano, J. P. (1994).
#' *The stationary bootstrap.* Journal of the American Statistical Association, 89(428), 1303-1313.
#'
#' @keywords internal
.stationary_bootstrap <- function(idx, mean_block) {
  stopifnot(length(idx) > 0L, is.numeric(mean_block), mean_block > 0)
  n <- length(idx)
  if (n == 1L) return(idx)
  mean_block <- max(1, mean_block)
  p <- 1 / mean_block

  out <- integer(n)
  pos <- sample.int(n, 1L)
  out[1L] <- idx[pos]

  for (i in 2:n) {
    if (runif(1L) < p) {
      pos <- sample.int(n, 1L)
    } else {
      pos <- pos + 1L
      if (pos > n) pos <- 1L
    }
    out[i] <- idx[pos]
  }

  stopifnot(length(out) == n)
  out
}

test_that("time-series block permutation utilities return valid indices", {
  set.seed(1)
  idx <- 1:10
  out1 <- bioLeak:::.circular_block_permute(idx, block_len = 3)
  expect_equal(length(out1), length(idx))
  expect_true(all(out1 %in% idx))

  set.seed(2)
  out2 <- bioLeak:::.stationary_bootstrap(idx, mean_block = 3)
  expect_equal(length(out2), length(idx))
  expect_true(all(out2 %in% idx))
})

test_that("time-series block permutation utilities validate inputs", {
  expect_error(bioLeak:::.circular_block_permute(integer(0), block_len = 1))
  expect_error(bioLeak:::.circular_block_permute(1:5, block_len = 0))
  expect_error(bioLeak:::.stationary_bootstrap(integer(0), mean_block = 1))
  expect_error(bioLeak:::.stationary_bootstrap(1:5, mean_block = 0))
})

# Gap statistic Delta = observed - mean(perm) standard errors ------------------
.se_ci_delta <- function(delta, se_obs, perm_values, level = 0.95) {
  perm_values <- perm_values[is.finite(perm_values)]
  if (length(perm_values) < 2L) {
    return(list(se = NA_real_, ci = c(NA_real_, NA_real_), z = NA_real_))
  }

  var_perm <- stats::var(perm_values)
  if (!is.finite(var_perm) || var_perm == 0) {
    se_obs <- ifelse(is.finite(se_obs), se_obs, 0)
    z <- if (se_obs > 0) delta / se_obs else NA_real_
    alpha <- (1 - level) / 2
    q0 <- stats::qnorm(c(alpha, 1 - alpha))
    ci <- delta + q0 * se_obs
    return(list(se = se_obs, ci = round(ci, 6), z = z))
  }

  se_perm <- sqrt(var_perm / length(perm_values))
  se_obs <- ifelse(is.finite(se_obs), se_obs, 0)
  se_delta <- sqrt(se_obs^2 + se_perm^2)
  z <- if (se_delta > 0) delta / se_delta else NA_real_

  alpha <- (1 - level) / 2
  q <- if (length(perm_values) < 30) {
    stats::qt(c(alpha, 1 - alpha), df = length(perm_values) - 1)
  } else {
    stats::qnorm(c(alpha, 1 - alpha))
  }

  ci <- round(delta + q * se_delta, 6)
  list(se = se_delta, ci = ci, z = z)
}

# Utilities to work with SummarizedExperiment (SE) or plain matrices

.bio_is_se <- function(x) inherits(x, "SummarizedExperiment")

.bio_coldata_df <- function(x) {
  cd <- if (.bio_is_se(x)) SummarizedExperiment::colData(x) else x
  out <- as.list(cd)
  out <- lapply(out, function(col) {
    if (is.factor(col) || is.character(col) || is.numeric(col) ||
        is.integer(col) || is.logical(col) ||
        inherits(col, "Date") || inherits(col, "POSIXt")) {
      return(col)
    }
    if (methods::is(col, "Rle") || methods::is(col, "Vector")) {
      return(as.vector(col))
    }
    if (methods::is(col, "List")) {
      return(as.list(col))
    }
    col
  })
  out <- as.data.frame(out, check.names = FALSE, stringsAsFactors = FALSE)
  rn <- tryCatch(rownames(cd), error = function(e) NULL)
  if (!is.null(rn) && length(rn) == nrow(out)) {
    rownames(out) <- rn
  }
  out
}

.bio_get_x <- function(x, assay_name = NULL) {
  if (.bio_is_se(x)) {
    if (is.null(assay_name)) {
      an <- SummarizedExperiment::assayNames(x)
      if (length(an) == 0) stop("No assays found in SummarizedExperiment.")
      assay_name <- an[1L]
    }
    out <- t(SummarizedExperiment::assay(x, assay_name))  # samples x features
    out <- as.data.frame(out, stringsAsFactors = FALSE)
  } else if (is.data.frame(x)) {
    out <- x  # keep original structure
  } else if (is.matrix(x)) {
    # convert to numeric data.frame if possible
    out <- as.data.frame(x, stringsAsFactors = FALSE)
    # attempt numeric conversion for columns that look numeric
    out[] <- lapply(out, function(col) {
      suppressWarnings(num <- as.numeric(col))
      if (is.numeric(col)) {
        col
      } else if (!anyNA(num)) {
        num
      } else {
        col
      }
    })
  } else {
    stop("Unsupported x type.")
  }
  return(out)
}


.bio_get_y <- function(x, outcome) {
  if (.bio_is_se(x)) {
    cd <- SummarizedExperiment::colData(x)
    if (is.character(outcome) && length(outcome) == 2L) {
      if (!all(outcome %in% colnames(cd))) stop("Outcome columns not in colData.")
      if (!requireNamespace("survival", quietly = TRUE)) {
        stop("Package 'survival' is required for time/event outcomes.")
      }
      time <- cd[[outcome[[1]]]]
      event <- cd[[outcome[[2]]]]
      return(survival::Surv(time = time, event = event))
    }
    if (!(outcome %in% colnames(cd))) stop("Outcome column not in colData.")
    cd[[outcome]]
  } else {
    if (is.null(outcome))
      stop("Provide outcome column name when x is not a SummarizedExperiment.")
    if (is.character(outcome) && length(outcome) == 2L) {
      if (!all(outcome %in% colnames(x))) stop("Outcome columns not found in data.")
      if (!requireNamespace("survival", quietly = TRUE)) {
        stop("Package 'survival' is required for time/event outcomes.")
      }
      if (is.matrix(x)) {
        time <- x[, outcome[[1]]]
        event <- x[, outcome[[2]]]
      } else {
        time <- x[[outcome[[1]]]]
        event <- x[[outcome[[2]]]]
      }
      return(survival::Surv(time = time, event = event))
    }
    if (is.character(outcome) && outcome %in% colnames(x)) {
      if (is.data.frame(x)) {
        x[[outcome]]
      } else if (is.matrix(x)) {
        x[, outcome]
      } else {
        x[[outcome]]
      }
    } else {
      stop("Outcome not found in data frame or matrix.")
    }
  }
}

.bio_get_meta <- function(x, cols) {
  if (.bio_is_se(x)) {
    cd <- SummarizedExperiment::colData(x)
    out <- list()
    for (nm in cols) {
      out[[nm]] <- if (nm %in% colnames(cd)) cd[[nm]] else NULL
    }
    out
  } else {
    as.list(rep(list(NULL), length(cols))) |> stats::setNames(cols)
  }
}

.bio_hash_indices <- function(idx) {
  if (requireNamespace("digest", quietly = TRUE))
    return(paste0("h", substr(digest::digest(idx), 1, 8)))
  paste0("h", sprintf("%08X", as.integer(sum(unlist(idx)) %% .Machine$integer.max)))
}

.bio_is_survival <- function(y) inherits(y, "Surv")
.bio_is_classification <- function(y) {
  if (.bio_is_survival(y)) return(FALSE)
  is.factor(y) || (is.numeric(y) && length(unique(y)) <= 10)
}
.bio_is_multiclass <- function(y) {
  is.factor(y) && nlevels(y) > 2
}
.bio_is_regression <- function(y) is.numeric(y) && !.bio_is_binomial(y)
.bio_is_binomial <- function(y) {
  if (.bio_is_survival(y)) return(FALSE)
  if (is.factor(y)) return(nlevels(y) == 2)
  if (is.numeric(y)) {
    u <- sort(unique(y[!is.na(y)]))
    return(length(u) == 2 && all(u %in% c(0,1)))
  }
  FALSE
}

test_that("SummarizedExperiment helpers behave for data.frame and matrix inputs", {
  df <- make_class_df(6)

  expect_false(bioLeak:::.bio_is_se(df))
  expect_equal(bioLeak:::.bio_get_x(df)$x1, df$x1)
  expect_equal(bioLeak:::.bio_get_y(df, "outcome"), df$outcome)

  expect_error(bioLeak:::.bio_get_y(df, "missing"), "Outcome not found")
  expect_error(bioLeak:::.bio_get_y(df, NULL), "Provide outcome")

  mat <- as.matrix(df[, c("x1", "x2")])
  mat_out <- bioLeak:::.bio_get_x(mat)
  expect_true(is.data.frame(mat_out))
  expect_true(all(vapply(mat_out, is.numeric, logical(1))))
})

test_that("SummarizedExperiment helpers behave for SE inputs", {
  skip_if_not_installed("SummarizedExperiment")
  se <- SummarizedExperiment::SummarizedExperiment(
    assays = list(counts = matrix(1:9, nrow = 3)),
    colData = data.frame(outcome = c(0, 1, 0))
  )
  expect_true(bioLeak:::.bio_is_se(se))
  x <- bioLeak:::.bio_get_x(se)
  expect_equal(dim(x), c(3, 3))
  expect_equal(bioLeak:::.bio_get_y(se, "outcome"), c(0, 1, 0))
  expect_error(bioLeak:::.bio_get_y(se, "missing"), "Outcome column not in colData")
})

test_that("metadata and hash helpers are stable", {
  df <- make_class_df(6)
  meta <- bioLeak:::.bio_get_meta(df, c("batch", "study"))
  expect_equal(names(meta), c("batch", "study"))
  expect_true(all(vapply(meta, is.null, logical(1))))

  idx <- list(list(train = 1:3, test = 4:6))
  h <- bioLeak:::.bio_hash_indices(idx)
  expect_true(is.character(h))
  expect_true(nchar(h) >= 2)
})

test_that("classification helper predicates behave as expected", {
  expect_true(bioLeak:::.bio_is_binomial(c(0, 1, 0)))
  expect_false(bioLeak:::.bio_is_binomial(c(0, 1, 2)))
  expect_true(bioLeak:::.bio_is_classification(factor(c("a", "b"))))
  expect_true(bioLeak:::.bio_is_regression(c(1.1, 2.2, 3.3)))
  expect_false(bioLeak:::.bio_is_regression(factor(c("a", "b"))))
})

## ---------------------------------------------------------------------
## R/accessors.R
## ---------------------------------------------------------------------
## Public accessor functions for the bioLeak S4 result classes.
##
## These wrappers expose slot data as named functions so that downstream
## code (replication scripts, vignettes, end-user analyses) can read the
## components of LeakFit, LeakAudit, and LeakDeltaLSI objects without
## reaching into S4 internals via `@`. Slot definitions are unchanged;
## these accessors are purely additive.
## ---------------------------------------------------------------------


# ---- LeakFit ---------------------------------------------------------

#' Per-fold metric data frame from a LeakFit
#'
#' Returns the per-fold metric data frame stored in a `LeakFit` object.
#' Each row is one (fold, repeat, learner) combination; columns include
#' the requested metric values such as `auc` and any task-specific
#' performance scores.
#'
#' @param fit A `LeakFit` object returned by [fit_resample()].
#' @return A `data.frame` with one row per (fold, repeat, learner)
#'   combination.
#' @seealso [fit_resample()], [audit_perm_gap()]
#' @examples
#' set.seed(1)
#' df <- data.frame(
#'   subject = rep(1:6, each = 2),
#'   outcome = factor(rep(c("a","b"), 6), levels = c("a","b")),
#'   x1 = rnorm(12), x2 = rnorm(12)
#' )
#' splits <- make_split_plan(df, outcome = "outcome",
#'                           mode = "subject_grouped", group = "subject", v = 3)
#' \dontrun{
#'   fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                       learner = parsnip::logistic_reg() |>
#'                                 parsnip::set_engine("glm"),
#'                       metrics = "auc")
#'   fit_metrics(fit)
#' }
#' @export
fit_metrics <- function(fit) {
  if (!methods::is(fit, "LeakFit")) {
    stop("`fit` must be a LeakFit object (got: ",
         paste(class(fit), collapse = "/"), ").",
         call. = FALSE)
  }
  fit@metrics
}


# ---- LeakAudit ------------------------------------------------------

#' Permutation-gap test results from a LeakAudit
#'
#' Returns the permutation-gap test data frame stored in a `LeakAudit`
#' object. Columns include the observed metric, permuted-null mean and
#' SD, gap, z-score, and permutation p-value.
#'
#' @param audit A `LeakAudit` object returned by [audit_leakage()].
#' @return A `data.frame` with one row per (mechanism class, repeat),
#'   summarising the permutation-gap test.
#' @seealso [audit_leakage()], [audit_target_assoc()]
#' @export
audit_perm_gap <- function(audit) {
  if (!methods::is(audit, "LeakAudit")) {
    stop("`audit` must be a LeakAudit object (got: ",
         paste(class(audit), collapse = "/"), ").",
         call. = FALSE)
  }
  audit@permutation_gap
}

#' Batch / study association results from a LeakAudit
#'
#' Returns the batch/study chi-squared association data frame stored in
#' a `LeakAudit` object. Columns include the metadata column, repeat,
#' chi-squared statistic, degrees of freedom, p-value, and Cramer's V.
#'
#' @param audit A `LeakAudit` object returned by [audit_leakage()].
#' @return A `data.frame` (possibly empty) with one row per
#'   (metadata column, repeat).
#' @seealso [audit_leakage()], [audit_perm_gap()]
#' @export
audit_batch_assoc <- function(audit) {
  if (!methods::is(audit, "LeakAudit")) {
    stop("`audit` must be a LeakAudit object (got: ",
         paste(class(audit), collapse = "/"), ").",
         call. = FALSE)
  }
  audit@batch_assoc
}

#' Target leakage scan results from a LeakAudit
#'
#' Returns the per-feature target-association data frame stored in a
#' `LeakAudit` object. Columns include the feature name, association
#' value, score, and a logical flag indicating whether the feature
#' exceeded the configured `target_threshold`.
#'
#' @param audit A `LeakAudit` object returned by [audit_leakage()].
#' @return A `data.frame` (possibly empty) with one row per scanned
#'   feature.
#' @seealso [audit_leakage()], [audit_perm_gap()]
#' @export
audit_target_assoc <- function(audit) {
  if (!methods::is(audit, "LeakAudit")) {
    stop("`audit` must be a LeakAudit object (got: ",
         paste(class(audit), collapse = "/"), ").",
         call. = FALSE)
  }
  audit@target_assoc
}

#' Near-duplicate sample pairs from a LeakAudit
#'
#' Returns the near-duplicate detection data frame stored in a
#' `LeakAudit` object. Each row is one (i, j) sample pair whose
#' similarity exceeds the configured threshold.
#'
#' @param audit A `LeakAudit` object returned by [audit_leakage()].
#' @return A `data.frame` (possibly empty) with one row per detected
#'   near-duplicate pair.
#' @seealso [audit_leakage()]
#' @export
audit_duplicates <- function(audit) {
  if (!methods::is(audit, "LeakAudit")) {
    stop("`audit` must be a LeakAudit object (got: ",
         paste(class(audit), collapse = "/"), ").",
         call. = FALSE)
  }
  audit@duplicates
}

#' Auxiliary information list from a LeakAudit
#'
#' Returns the auxiliary `info` list stored in a `LeakAudit` object.
#' This list carries non-tabular audit components such as
#' multivariate-target-scan results, configuration flags, and
#' permutation-test diagnostics.
#'
#' @param audit A `LeakAudit` object returned by [audit_leakage()].
#' @return A named list of supplementary audit components.
#' @seealso [audit_leakage()]
#' @export
audit_info <- function(audit) {
  if (!methods::is(audit, "LeakAudit")) {
    stop("`audit` must be a LeakAudit object (got: ",
         paste(class(audit), collapse = "/"), ").",
         call. = FALSE)
  }
  audit@info
}


# ---- LeakDeltaLSI ---------------------------------------------------

#' Raw delta-metric estimate from a LeakDeltaLSI
#'
#' Returns the arithmetic mean of the per-repeat raw metric differences
#' (leaky minus guarded) stored in a `LeakDeltaLSI` object.
#'
#' @param dlsi A `LeakDeltaLSI` object returned by [delta_lsi()].
#' @return A length-one numeric scalar.
#' @seealso [delta_lsi()], [dlsi_robust()], [dlsi_ci()]
#' @export
dlsi_metric <- function(dlsi) {
  if (!methods::is(dlsi, "LeakDeltaLSI")) {
    stop("`dlsi` must be a LeakDeltaLSI object (got: ",
         paste(class(dlsi), collapse = "/"), ").",
         call. = FALSE)
  }
  dlsi@delta_metric
}

#' Huber-robust delta_lsi point estimate from a LeakDeltaLSI
#'
#' Returns the Huber-robust point estimate of the per-repeat delta
#' values stored in a `LeakDeltaLSI` object.
#'
#' @param dlsi A `LeakDeltaLSI` object returned by [delta_lsi()].
#' @return A length-one numeric scalar.
#' @seealso [delta_lsi()], [dlsi_metric()], [dlsi_ci()]
#' @export
dlsi_robust <- function(dlsi) {
  if (!methods::is(dlsi, "LeakDeltaLSI")) {
    stop("`dlsi` must be a LeakDeltaLSI object (got: ",
         paste(class(dlsi), collapse = "/"), ").",
         call. = FALSE)
  }
  dlsi@delta_lsi
}

#' BCa confidence interval from a LeakDeltaLSI
#'
#' Returns the bias-corrected and accelerated (BCa) bootstrap confidence
#' interval stored in a `LeakDeltaLSI` object. By default returns the
#' interval for the Huber-robust delta estimate; set
#' `which = "metric"` to return the interval for the raw metric
#' difference instead.
#'
#' @param dlsi A `LeakDeltaLSI` object returned by [delta_lsi()].
#' @param which Either `"robust"` (default) for the Huber estimate's
#'   confidence interval, or `"metric"` for the raw arithmetic mean's
#'   confidence interval.
#' @return A length-two numeric vector `c(lower, upper)`. Returns
#'   `c(NA_real_, NA_real_)` when the interval is not computed (e.g.,
#'   the inference tier did not include CIs).
#' @seealso [delta_lsi()], [dlsi_robust()], [dlsi_metric()]
#' @export
dlsi_ci <- function(dlsi, which = c("robust", "metric")) {
  if (!methods::is(dlsi, "LeakDeltaLSI")) {
    stop("`dlsi` must be a LeakDeltaLSI object (got: ",
         paste(class(dlsi), collapse = "/"), ").",
         call. = FALSE)
  }
  which <- match.arg(which)
  if (identical(which, "robust"))   dlsi@delta_lsi_ci  else dlsi@delta_metric_ci
}

#' Sign-flip p-value from a LeakDeltaLSI
#'
#' Returns the paired sign-flip randomization-test p-value stored in a
#' `LeakDeltaLSI` object.
#'
#' @param dlsi A `LeakDeltaLSI` object returned by [delta_lsi()].
#' @return A length-one numeric scalar in `[0, 1]`, or `NA_real_` when
#'   the inference tier did not include hypothesis testing.
#' @seealso [delta_lsi()], [dlsi_tier()]
#' @export
dlsi_p_value <- function(dlsi) {
  if (!methods::is(dlsi, "LeakDeltaLSI")) {
    stop("`dlsi` must be a LeakDeltaLSI object (got: ",
         paste(class(dlsi), collapse = "/"), ").",
         call. = FALSE)
  }
  dlsi@p_value
}

#' Inference tier from a LeakDeltaLSI
#'
#' Returns the inference tier label stored in a `LeakDeltaLSI` object.
#' Possible values are `"A_full_inference"` (`R_eff >= 20`),
#' `"B_signflip_ci"` (`R_eff >= 10`), `"C_signflip"`
#' (`R_eff >= 5`), or `"D_insufficient"` (`R_eff < 5`).
#'
#' @param dlsi A `LeakDeltaLSI` object returned by [delta_lsi()].
#' @return A length-one character string giving the tier label.
#' @seealso [delta_lsi()], [dlsi_R_eff()]
#' @export
dlsi_tier <- function(dlsi) {
  if (!methods::is(dlsi, "LeakDeltaLSI")) {
    stop("`dlsi` must be a LeakDeltaLSI object (got: ",
         paste(class(dlsi), collapse = "/"), ").",
         call. = FALSE)
  }
  dlsi@tier
}

#' Effective number of paired repeats from a LeakDeltaLSI
#'
#' Returns the effective number of paired repeats `R_eff` stored in a
#' `LeakDeltaLSI` object. This is the count of repeats that contribute
#' to the inference; it equals the smaller of the leaky and guarded
#' fits' repeat counts when the comparison is paired.
#'
#' @param dlsi A `LeakDeltaLSI` object returned by [delta_lsi()].
#' @return A length-one integer.
#' @seealso [delta_lsi()], [dlsi_tier()]
#' @export
dlsi_R_eff <- function(dlsi) {
  if (!methods::is(dlsi, "LeakDeltaLSI")) {
    stop("`dlsi` must be a LeakDeltaLSI object (got: ",
         paste(class(dlsi), collapse = "/"), ").",
         call. = FALSE)
  }
  dlsi@R_eff
}

#' Per-repeat metric data frames from a LeakDeltaLSI
#'
#' Returns the per-repeat metric data frame for one of the two
#' pipelines stored in a `LeakDeltaLSI` object. The naive (or leaky)
#' pipeline's repeats are returned by default; set
#' `which = "guarded"` to return the guarded pipeline's repeats.
#'
#' @param dlsi A `LeakDeltaLSI` object returned by [delta_lsi()].
#' @param which Either `"naive"` (default) for the naive/leaky
#'   pipeline's per-repeat data frame, or `"guarded"` for the guarded
#'   pipeline's per-repeat data frame.
#' @return A `data.frame` with one row per repeat.
#' @seealso [delta_lsi()]
#' @export
dlsi_repeats <- function(dlsi, which = c("naive", "guarded")) {
  if (!methods::is(dlsi, "LeakDeltaLSI")) {
    stop("`dlsi` must be a LeakDeltaLSI object (got: ",
         paste(class(dlsi), collapse = "/"), ").",
         call. = FALSE)
  }
  which <- match.arg(which)
  if (identical(which, "naive"))   dlsi@repeats_naive  else dlsi@repeats_guarded
}

## ---------------------------------------------------------------------
## test-accessors.R
## ---------------------------------------------------------------------
## Tests for the public accessor functions added in 0.3.7. Each accessor
## should:
##   (a) return exactly the same object as the corresponding @ slot read,
##   (b) error informatively when called on the wrong class, and
##   (c) (where applicable) honor the `which` argument that selects
##       between two related slots.
## ---------------------------------------------------------------------


# ---- Build minimal fixtures by direct slot construction --------------
# These tests avoid running fit_resample / audit_leakage / delta_lsi to
# keep the suite fast and to isolate accessor behavior from upstream
# computational changes. The fixtures populate every slot the accessors
# read, and only those slots, so an accessor regression is unambiguous.

make_leakfit_fixture <- function() {
  splits <- make_split_plan_quiet(
    make_class_df(8),
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 2,
    seed = 1
  )
  metrics_df <- data.frame(
    learner = "stub",
    fold = 1:2,
    repeat_id = 1L,
    auc = c(0.71, 0.65),
    stringsAsFactors = FALSE
  )
  bioLeak:::LeakFit(
    splits         = splits,
    metrics        = metrics_df,
    metric_summary = data.frame(),
    audit          = data.frame(),
    predictions    = list(),
    preprocess     = list(),
    learners       = list(),
    outcome        = "outcome",
    task           = "binomial",
    feature_names  = character(),
    info           = list()
  )
}

make_leakaudit_fixture <- function() {
  fit <- make_leakfit_fixture()
  perm_gap <- data.frame(
    mechanism_class = "non_random_signal",
    repeat_id = 1L,
    metric_obs = 0.68,
    perm_mean = 0.50,
    perm_sd = 0.05,
    gap = 0.18,
    z = 3.6,
    p_value = 0.02,
    n_perm = 50L,
    stringsAsFactors = FALSE
  )
  batch_assoc <- data.frame(
    mechanism_class = "confounding_alignment",
    batch_col = "batch",
    repeat_id = 1L,
    stat = 4.0,
    df = 2L,
    pval = 0.135,
    cramer_v = 0.30,
    stringsAsFactors = FALSE
  )
  target_assoc <- data.frame(
    mechanism_class = "proxy_target_leakage",
    feature = c("x1", "x2"),
    type = "numeric",
    metric = "auc",
    value = c(0.55, 0.72),
    score = c(0.10, 0.45),
    p_value = c(0.40, 0.01),
    n = 8L,
    n_levels = NA_integer_,
    flag = c(FALSE, FALSE),
    p_value_adj = NA_real_,
    flag_fdr = FALSE,
    stringsAsFactors = FALSE
  )
  duplicates <- data.frame(
    mechanism_class = character(0),
    i = integer(0),
    j = integer(0),
    sim = numeric(0),
    cross_fold = logical(0),
    cos_sim = numeric(0),
    stringsAsFactors = FALSE
  )
  info <- list(
    target_multivariate = list(score = 0.42, p_value = 0.07),
    perm_method = "fixed_predictions"
  )
  bioLeak:::LeakAudit(
    fit             = fit,
    permutation_gap = perm_gap,
    perm_values     = c(0.48, 0.51, 0.55),
    batch_assoc     = batch_assoc,
    target_assoc    = target_assoc,
    duplicates      = duplicates,
    trail           = list(),
    info            = info
  )
}

make_leakdeltalsi_fixture <- function() {
  repeats_naive <- data.frame(
    repeat_id = 1:3,
    metric = c(0.91, 0.92, 0.90),
    stringsAsFactors = FALSE
  )
  repeats_guarded <- data.frame(
    repeat_id = 1:3,
    metric = c(0.62, 0.60, 0.58),
    stringsAsFactors = FALSE
  )
  bioLeak:::LeakDeltaLSI(
    metric          = "auc",
    exchangeability = "iid",
    tier            = "B_signflip_ci",
    strict          = FALSE,
    R_eff           = 3L,
    delta_lsi       = 0.310,
    delta_lsi_ci    = c(0.295, 0.325),
    delta_metric    = 0.305,
    delta_metric_ci = c(0.290, 0.320),
    p_value         = 0.0033,
    inference_ok    = FALSE,
    folds_naive     = data.frame(),
    folds_guarded   = data.frame(),
    repeats_naive   = repeats_naive,
    repeats_guarded = repeats_guarded,
    info            = list()
  )
}


# ---- LeakFit accessors ----------------------------------------------

test_that("fit_metrics() returns the @metrics slot for a LeakFit", {
  fit <- make_leakfit_fixture()
  expect_identical(fit_metrics(fit), fit@metrics)
  expect_s3_class(fit_metrics(fit), "data.frame")
  expect_named(fit_metrics(fit), c("learner", "fold", "repeat_id", "auc"))
})

test_that("fit_metrics() rejects non-LeakFit input", {
  expect_error(fit_metrics(NULL),
               "must be a LeakFit object")
  expect_error(fit_metrics(data.frame(x = 1)),
               "must be a LeakFit object")
  audit <- make_leakaudit_fixture()
  expect_error(fit_metrics(audit),
               "must be a LeakFit object")
})


# ---- LeakAudit accessors --------------------------------------------

test_that("audit_perm_gap() returns the @permutation_gap slot", {
  audit <- make_leakaudit_fixture()
  expect_identical(audit_perm_gap(audit), audit@permutation_gap)
  expect_true("metric_obs" %in% names(audit_perm_gap(audit)))
})

test_that("audit_perm_gap() rejects non-LeakAudit input", {
  expect_error(audit_perm_gap(NULL), "must be a LeakAudit object")
  fit <- make_leakfit_fixture()
  expect_error(audit_perm_gap(fit), "must be a LeakAudit object")
})

test_that("audit_batch_assoc() returns the @batch_assoc slot", {
  audit <- make_leakaudit_fixture()
  expect_identical(audit_batch_assoc(audit), audit@batch_assoc)
  expect_true("cramer_v" %in% names(audit_batch_assoc(audit)))
})

test_that("audit_batch_assoc() rejects non-LeakAudit input", {
  expect_error(audit_batch_assoc("not an audit"),
               "must be a LeakAudit object")
})

test_that("audit_target_assoc() returns the @target_assoc slot", {
  audit <- make_leakaudit_fixture()
  expect_identical(audit_target_assoc(audit), audit@target_assoc)
  expect_true("score" %in% names(audit_target_assoc(audit)))
})

test_that("audit_target_assoc() rejects non-LeakAudit input", {
  expect_error(audit_target_assoc(list(target_assoc = data.frame())),
               "must be a LeakAudit object")
})

test_that("audit_duplicates() returns the @duplicates slot", {
  audit <- make_leakaudit_fixture()
  expect_identical(audit_duplicates(audit), audit@duplicates)
  expect_s3_class(audit_duplicates(audit), "data.frame")
})

test_that("audit_duplicates() rejects non-LeakAudit input", {
  expect_error(audit_duplicates(1L),
               "must be a LeakAudit object")
})

test_that("audit_info() returns the @info slot", {
  audit <- make_leakaudit_fixture()
  expect_identical(audit_info(audit), audit@info)
  expect_named(audit_info(audit),
               c("target_multivariate", "perm_method"))
  expect_equal(audit_info(audit)$target_multivariate$p_value, 0.07)
})

test_that("audit_info() rejects non-LeakAudit input", {
  expect_error(audit_info(list()), "must be a LeakAudit object")
})


# ---- LeakDeltaLSI accessors -----------------------------------------

test_that("dlsi_metric() returns the @delta_metric slot", {
  dlsi <- make_leakdeltalsi_fixture()
  expect_identical(dlsi_metric(dlsi), dlsi@delta_metric)
  expect_equal(dlsi_metric(dlsi), 0.305)
})

test_that("dlsi_robust() returns the @delta_lsi slot", {
  dlsi <- make_leakdeltalsi_fixture()
  expect_identical(dlsi_robust(dlsi), dlsi@delta_lsi)
  expect_equal(dlsi_robust(dlsi), 0.310)
})

test_that("dlsi_ci(which='robust') returns the @delta_lsi_ci slot (default)", {
  dlsi <- make_leakdeltalsi_fixture()
  expect_identical(dlsi_ci(dlsi), dlsi@delta_lsi_ci)
  expect_identical(dlsi_ci(dlsi, which = "robust"), dlsi@delta_lsi_ci)
  expect_equal(dlsi_ci(dlsi), c(0.295, 0.325))
})

test_that("dlsi_ci(which='metric') returns the @delta_metric_ci slot", {
  dlsi <- make_leakdeltalsi_fixture()
  expect_identical(dlsi_ci(dlsi, which = "metric"), dlsi@delta_metric_ci)
  expect_equal(dlsi_ci(dlsi, which = "metric"), c(0.290, 0.320))
})

test_that("dlsi_ci() rejects unknown `which` values", {
  dlsi <- make_leakdeltalsi_fixture()
  expect_error(dlsi_ci(dlsi, which = "raw"))      # match.arg() error
  expect_error(dlsi_ci(dlsi, which = "huber"))
})

test_that("dlsi_p_value() returns the @p_value slot", {
  dlsi <- make_leakdeltalsi_fixture()
  expect_identical(dlsi_p_value(dlsi), dlsi@p_value)
  expect_equal(dlsi_p_value(dlsi), 0.0033)
})

test_that("dlsi_tier() returns the @tier slot", {
  dlsi <- make_leakdeltalsi_fixture()
  expect_identical(dlsi_tier(dlsi), dlsi@tier)
  expect_equal(dlsi_tier(dlsi), "B_signflip_ci")
})

test_that("dlsi_R_eff() returns the @R_eff slot", {
  dlsi <- make_leakdeltalsi_fixture()
  expect_identical(dlsi_R_eff(dlsi), dlsi@R_eff)
  expect_equal(dlsi_R_eff(dlsi), 3L)
})

test_that("dlsi_repeats(which='naive') returns the @repeats_naive slot (default)", {
  dlsi <- make_leakdeltalsi_fixture()
  expect_identical(dlsi_repeats(dlsi), dlsi@repeats_naive)
  expect_identical(dlsi_repeats(dlsi, which = "naive"), dlsi@repeats_naive)
  expect_equal(dlsi_repeats(dlsi)$metric, c(0.91, 0.92, 0.90))
})

test_that("dlsi_repeats(which='guarded') returns the @repeats_guarded slot", {
  dlsi <- make_leakdeltalsi_fixture()
  expect_identical(dlsi_repeats(dlsi, which = "guarded"), dlsi@repeats_guarded)
  expect_equal(dlsi_repeats(dlsi, which = "guarded")$metric,
               c(0.62, 0.60, 0.58))
})

test_that("dlsi_repeats() rejects unknown `which` values", {
  dlsi <- make_leakdeltalsi_fixture()
  expect_error(dlsi_repeats(dlsi, which = "leaky"))   # match.arg() error
})

test_that("LeakDeltaLSI accessors all reject non-LeakDeltaLSI input", {
  not_dlsi <- list(delta_lsi = 1, delta_metric = 1)   # mimic-y but wrong class
  expect_error(dlsi_metric(not_dlsi),   "must be a LeakDeltaLSI object")
  expect_error(dlsi_robust(not_dlsi),   "must be a LeakDeltaLSI object")
  expect_error(dlsi_ci(not_dlsi),       "must be a LeakDeltaLSI object")
  expect_error(dlsi_p_value(not_dlsi),  "must be a LeakDeltaLSI object")
  expect_error(dlsi_tier(not_dlsi),     "must be a LeakDeltaLSI object")
  expect_error(dlsi_R_eff(not_dlsi),    "must be a LeakDeltaLSI object")
  expect_error(dlsi_repeats(not_dlsi),  "must be a LeakDeltaLSI object")
})


# ---- Round-trip equivalence to direct @ access ----------------------
# These tests assert that, for every accessor, calling it produces the
# bit-identical R object as direct slot access. This guards against
# any future helper that silently copies, reformats, or transforms the
# underlying data.

test_that("Every accessor is bit-identical to its corresponding @ slot read", {
  fit <- make_leakfit_fixture()
  audit <- make_leakaudit_fixture()
  dlsi <- make_leakdeltalsi_fixture()

  expect_identical(fit_metrics(fit),               fit@metrics)
  expect_identical(audit_perm_gap(audit),          audit@permutation_gap)
  expect_identical(audit_batch_assoc(audit),       audit@batch_assoc)
  expect_identical(audit_target_assoc(audit),      audit@target_assoc)
  expect_identical(audit_duplicates(audit),        audit@duplicates)
  expect_identical(audit_info(audit),              audit@info)
  expect_identical(dlsi_metric(dlsi),              dlsi@delta_metric)
  expect_identical(dlsi_robust(dlsi),              dlsi@delta_lsi)
  expect_identical(dlsi_ci(dlsi, "robust"),        dlsi@delta_lsi_ci)
  expect_identical(dlsi_ci(dlsi, "metric"),        dlsi@delta_metric_ci)
  expect_identical(dlsi_p_value(dlsi),             dlsi@p_value)
  expect_identical(dlsi_tier(dlsi),                dlsi@tier)
  expect_identical(dlsi_R_eff(dlsi),               dlsi@R_eff)
  expect_identical(dlsi_repeats(dlsi, "naive"),    dlsi@repeats_naive)
  expect_identical(dlsi_repeats(dlsi, "guarded"),  dlsi@repeats_guarded)
})

# During R CMD check, bioLeak is loaded via tests/testthat.R (library(bioLeak)).
# During interactive development, load via pkgload if not already attached.
if (!"package:bioLeak" %in% search()) {
  pkg_root <- normalizePath("../..", winslash = "/", mustWork = FALSE)
  if (file.exists(file.path(pkg_root, "DESCRIPTION"))) {
    pkgload::load_all(pkg_root, export_all = FALSE)
  }
}

# Additional diagnostic helpers ----------------------------------------------

.select_predictions_for_diagnostics <- function(fit, context, learner = NULL) {
  pred_df <- if (length(fit@predictions)) {
    do.call(rbind, lapply(fit@predictions, function(df) data.frame(df, stringsAsFactors = FALSE)))
  } else {
    NULL
  }
  if (is.null(pred_df) || !nrow(pred_df)) {
    stop(sprintf("No predictions available for %s.", context), call. = FALSE)
  }

  has_learner <- "learner" %in% names(pred_df)
  if (has_learner) {
    pred_df$learner <- as.character(pred_df$learner)
    learner_vals <- unique(pred_df$learner)
    if (is.null(learner)) {
      if (length(learner_vals) == 1L) {
        learner <- learner_vals[[1]]
      } else {
        stop("Multiple learners found in predictions; specify `learner` to select a single model.",
             call. = FALSE)
      }
    } else {
      if (length(learner) != 1L) stop("learner must be a single value.", call. = FALSE)
      if (!learner %in% learner_vals) {
        stop(sprintf("Learner '%s' not found in predictions. Available: %s",
                     learner, paste(learner_vals, collapse = ", ")),
             call. = FALSE)
      }
    }
    pred_df <- pred_df[pred_df$learner == learner, , drop = FALSE]
    if (!nrow(pred_df)) {
      stop(sprintf("No predictions available for learner '%s'.", learner), call. = FALSE)
    }
  } else {
    if (!is.null(learner)) {
      warning("`learner` ignored: predictions do not include learner IDs.", call. = FALSE)
    } else if (!is.null(fit@metrics) && "learner" %in% names(fit@metrics) &&
               length(unique(fit@metrics$learner)) > 1L) {
      warning("Multiple learners were fit but predictions lack learner IDs; diagnostics may mix learners. Refit with updated bioLeak.",
              call. = FALSE)
    }
  }

  pred_df
}

#' Calibration diagnostics for binomial predictions
#'
#' Computes reliability curve summaries and calibration metrics for a
#' binomial [LeakFit] using out-of-fold predictions.
#'
#' @param fit A [LeakFit] object from [fit_resample()].
#' @param bins Integer number of probability bins for the calibration curve.
#' @param min_bin_n Minimum samples per bin used in plotting; bins smaller than
#'   this are retained in the output but can be filtered by the caller.
#' @param learner Optional character scalar. When predictions include multiple
#'   learners, selects the learner to summarize.
#' @return A list with a `curve` data.frame and a one-row `metrics` data.frame
#'   containing ECE, MCE, and Brier score.
#' @examples
#' set.seed(42)
#' df <- data.frame(
#'   subject = rep(1:15, each = 2),
#'   outcome = factor(rep(c(0, 1), 15)),
#'   x1 = rnorm(30),
#'   x2 = rnorm(30)
#' )
#' splits <- make_split_plan(df, outcome = "outcome",
#'                           mode = "subject_grouped", group = "subject",
#'                           v = 3, progress = FALSE)
#' custom <- list(
#'   glm = list(
#'     fit = function(x, y, task, weights, ...) {
#'       stats::glm(y ~ ., data = as.data.frame(x),
#'                  family = stats::binomial(), weights = weights)
#'     },
#'     predict = function(object, newdata, task, ...) {
#'       as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
#'                                 type = "response"))
#'     }
#'   )
#' )
#' fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                     learner = "glm", custom_learners = custom,
#'                     metrics = "auc", refit = FALSE, seed = 1)
#' cal <- calibration_summary(fit, bins = 5)
#' cal$metrics
#'
#' @export
calibration_summary <- function(fit, bins = 10, min_bin_n = 5, learner = NULL) {
  stopifnot(inherits(fit, "LeakFit"))
  if (!identical(fit@task, "binomial")) {
    stop("calibration_summary is only available for binomial tasks.", call. = FALSE)
  }
  bins <- as.integer(bins)
  if (!is.finite(bins) || bins < 2L) {
    stop("bins must be an integer >= 2.", call. = FALSE)
  }
  min_bin_n <- as.integer(min_bin_n)
  if (!is.finite(min_bin_n) || min_bin_n < 1L) {
    stop("min_bin_n must be an integer >= 1.", call. = FALSE)
  }

  pred_df <- .select_predictions_for_diagnostics(fit, "calibration", learner = learner)
  if (!"pred" %in% names(pred_df)) {
    stop("Predictions missing 'pred' column for calibration.", call. = FALSE)
  }
  if (!"truth" %in% names(pred_df)) {
    stop("Predictions missing 'truth' values for calibration.", call. = FALSE)
  }

  pred <- as.numeric(pred_df$pred)
  truth <- pred_df$truth
  if (!is.factor(truth)) truth <- factor(truth)
  pos_class <- fit@info$positive_class
  if (is.null(pos_class) || !as.character(pos_class) %in% levels(truth)) {
    if (nlevels(truth) < 2L) {
      warning("calibration_summary requires at least two outcome levels.", call. = FALSE)
      return(data.frame(bin = integer(0), mean_pred = numeric(0),
                        obs_rate = numeric(0), n = integer(0)))
    }
    pos_class <- levels(truth)[2]
  }
  y <- as.integer(truth == pos_class)
  ok <- is.finite(pred) & !is.na(y)
  pred <- pred[ok]
  y <- y[ok]
  if (!length(pred)) {
    stop("No finite predictions available for calibration.", call. = FALSE)
  }

  if (any(pred < 0 | pred > 1, na.rm = TRUE)) {
    warning("Predictions outside [0,1] were clipped for calibration.", call. = FALSE)
    pred <- pmin(pmax(pred, 0), 1)
  }

  breaks <- seq(0, 1, length.out = bins + 1L)
  bin_id <- cut(pred, breaks = breaks, include.lowest = TRUE, right = TRUE)
  n_bin <- tapply(pred, bin_id, length)
  pred_mean <- tapply(pred, bin_id, mean)
  obs_rate <- tapply(y, bin_id, mean)
  pred_min <- tapply(pred, bin_id, min)
  pred_max <- tapply(pred, bin_id, max)

  curve <- data.frame(
    bin = factor(names(n_bin), levels = levels(bin_id), ordered = TRUE),
    n = as.integer(n_bin),
    pred_mean = as.numeric(pred_mean),
    obs_rate = as.numeric(obs_rate),
    pred_min = as.numeric(pred_min),
    pred_max = as.numeric(pred_max),
    stringsAsFactors = FALSE
  )

  valid <- is.finite(curve$pred_mean) & is.finite(curve$obs_rate) & curve$n > 0
  n_total <- sum(curve$n[valid])
  if (!n_total) {
    stop("No valid bins available for calibration metrics.", call. = FALSE)
  }
  ece <- sum((curve$n[valid] / n_total) * abs(curve$obs_rate[valid] - curve$pred_mean[valid]))
  mce <- max(abs(curve$obs_rate[valid] - curve$pred_mean[valid]))
  brier <- mean((pred - y)^2, na.rm = TRUE)

  metrics <- data.frame(
    n = n_total,
    bins = bins,
    min_bin_n = min_bin_n,
    ece = ece,
    mce = mce,
    brier = brier,
    stringsAsFactors = FALSE
  )

  list(curve = curve, metrics = metrics, positive_class = pos_class)
}

#' Confounder sensitivity summaries
#'
#' Computes performance metrics within confounder strata to surface potential
#' confounding. Requires aligned metadata in `coldata`.
#'
#' @param fit A [LeakFit] object from [fit_resample()].
#' @param confounders Character vector of columns in `coldata` to evaluate.
#'   Defaults to common batch/study identifiers when available.
#' @param metric Metric name to compute within each stratum. Defaults to the
#'   first metric used in the fit (or task defaults if unavailable).
#' @param min_n Minimum samples per stratum; smaller strata return NA metrics.
#' @param coldata Optional data.frame of sample metadata. Defaults to
#'   `fit@splits@info$coldata` when available.
#' @param numeric_bins Integer number of quantile bins for numeric confounders
#'   with many unique values.
#' @param learner Optional character scalar. When predictions include multiple
#'   learners, selects the learner to summarize.
#' @param strict_align Logical scalar. If TRUE, errors when coldata cannot be
#'   aligned by row names or IDs and would fall back to row-order matching.
#'   Default is FALSE.
#' @return A data.frame with per-confounder, per-level metrics and counts.
#' @examples
#' set.seed(42)
#' df <- data.frame(
#'   subject = rep(1:15, each = 2),
#'   outcome = factor(rep(c(0, 1), 15)),
#'   batch = factor(rep(c("A", "B", "C"), 10)),
#'   x1 = rnorm(30),
#'   x2 = rnorm(30)
#' )
#' splits <- make_split_plan(df, outcome = "outcome",
#'                           mode = "subject_grouped", group = "subject",
#'                           v = 3, progress = FALSE)
#' custom <- list(
#'   glm = list(
#'     fit = function(x, y, task, weights, ...) {
#'       stats::glm(y ~ ., data = as.data.frame(x),
#'                  family = stats::binomial(), weights = weights)
#'     },
#'     predict = function(object, newdata, task, ...) {
#'       as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
#'                                 type = "response"))
#'     }
#'   )
#' )
#' fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                     learner = "glm", custom_learners = custom,
#'                     metrics = "auc", refit = FALSE, seed = 1)
#' confounder_sensitivity(fit, confounders = "batch", coldata = df)
#'
#' @export
confounder_sensitivity <- function(fit, confounders = NULL, metric = NULL,
                                   min_n = 10, coldata = NULL, numeric_bins = 4,
                                   learner = NULL, strict_align = FALSE) {
  stopifnot(inherits(fit, "LeakFit"))
  pred_df <- .select_predictions_for_diagnostics(fit, "confounder sensitivity", learner = learner)
  if (!"id" %in% names(pred_df)) {
    stop("Predictions are missing sample ids; confounder sensitivity unavailable.", call. = FALSE)
  }

  if (is.null(coldata)) coldata <- fit@splits@info$coldata
  if (is.null(coldata)) {
    stop("No coldata available for confounder sensitivity.", call. = FALSE)
  }
  coldata <- as.data.frame(coldata, check.names = FALSE)

  ids_chr <- as.character(pred_df$id)
  align_coldata <- function(cd, ids, sample_ids = NULL) {
    rn <- rownames(cd)
    if (!is.null(rn) && all(ids %in% rn)) {
      return(cd[ids, , drop = FALSE])
    }
    if (!is.null(sample_ids) && length(sample_ids) == nrow(cd)) {
      idx <- match(ids, as.character(sample_ids))
      if (all(!is.na(idx))) return(cd[idx, , drop = FALSE])
    }
    if ("row_id" %in% names(cd)) {
      rid <- as.character(cd[["row_id"]])
      if (!anyDuplicated(rid) && all(ids %in% rid)) {
        return(cd[match(ids, rid), , drop = FALSE])
      }
    }
    ids_int <- suppressWarnings(as.integer(ids))
    if (all(!is.na(ids_int)) && max(ids_int, na.rm = TRUE) <= nrow(cd)) {
      return(cd[ids_int, , drop = FALSE])
    }
    if (nrow(cd) == length(ids)) {
      if (isTRUE(strict_align)) {
        stop("coldata rownames do not match prediction ids and strict_align = TRUE; ",
             "provide matching row names or a 'row_id' column.", call. = FALSE)
      }
      warning("coldata rownames do not match prediction ids; assuming row order aligns to predictions.",
              call. = FALSE)
      return(cd)
    }
    warning("coldata not aligned to predictions; confounder sensitivity skipped.", call. = FALSE)
    NULL
  }

  aligned <- align_coldata(coldata, ids_chr, sample_ids = fit@info$sample_ids)
  if (is.null(aligned)) return(data.frame())

  if (is.null(confounders)) {
    confounders <- intersect(c("batch", "study", "plate", "center", "site", "group"),
                             names(aligned))
  }
  if (!length(confounders)) {
    stop("No confounder columns available in coldata.", call. = FALSE)
  }

  min_n <- as.integer(min_n)
  if (!is.finite(min_n) || min_n < 1L) {
    stop("min_n must be an integer >= 1.", call. = FALSE)
  }
  numeric_bins <- as.integer(numeric_bins)
  if (!is.finite(numeric_bins) || numeric_bins < 2L) {
    stop("numeric_bins must be an integer >= 2.", call. = FALSE)
  }

  default_metric <- fit@info$metrics_used %||% NULL
  if (is.null(metric)) {
    metric <- default_metric[1] %||% {
      if (identical(fit@task, "binomial")) "auc"
      else if (identical(fit@task, "multiclass")) "accuracy"
      else if (identical(fit@task, "survival")) "cindex"
      else "rmse"
    }
  }

  valid_metrics <- switch(fit@task,
                          binomial = c("auc", "roc_auc", "pr_auc", "accuracy", "log_loss"),
                          multiclass = c("accuracy", "macro_f1", "log_loss", "mn_log_loss"),
                          gaussian = c("rmse", "cindex"),
                          survival = c("cindex"),
                          c("auc", "rmse"))
  if (!metric %in% valid_metrics) {
    stop(sprintf("Metric '%s' is not supported for %s tasks.", metric, fit@task), call. = FALSE)
  }

  metric_internal <- metric
  if (identical(metric_internal, "roc_auc")) metric_internal <- "auc"
  if (identical(metric_internal, "mn_log_loss")) metric_internal <- "log_loss"

  direction <- if (metric %in% c("rmse", "log_loss", "mn_log_loss")) "lower" else "higher"

  pos_class <- fit@info$positive_class %||% NULL
  if (identical(fit@task, "binomial")) {
    if (!is.factor(pred_df$truth)) pred_df$truth <- factor(pred_df$truth)
    if (is.null(pos_class) || !as.character(pos_class) %in% levels(pred_df$truth)) {
      if (nlevels(pred_df$truth) < 2L) {
        warning("confounder_sensitivity requires at least two outcome levels.",
                call. = FALSE)
        return(data.frame())
      }
      pos_class <- levels(pred_df$truth)[2]
    }
  }

  out <- list()
  for (conf in confounders) {
    if (!conf %in% names(aligned)) next
    vec <- aligned[[conf]]
    if (all(is.na(vec))) next
    if (is.numeric(vec)) {
      uniq <- unique(vec[is.finite(vec)])
      if (length(uniq) > numeric_bins) {
        probs <- seq(0, 1, length.out = numeric_bins + 1L)
        brks <- unique(stats::quantile(vec, probs = probs, na.rm = TRUE, names = FALSE))
        if (length(brks) >= 2L) {
          vec <- cut(vec, breaks = brks, include.lowest = TRUE, dig.lab = 10)
        } else {
          vec <- factor(vec)
        }
      } else {
        vec <- factor(vec)
      }
    } else {
      vec <- factor(vec)
    }

    df <- pred_df
    df$confounder_level <- vec
    lvls <- levels(df$confounder_level)
    for (lvl in lvls) {
      sub <- df[df$confounder_level == lvl & !is.na(df$confounder_level), , drop = FALSE]
      n_sub <- nrow(sub)
      value <- NA_real_
      if (n_sub >= min_n) {
        truth_val <- if ("truth" %in% names(sub)) sub$truth else sub$pred
        if (identical(fit@task, "binomial") && metric_internal %in% c("auc", "pr_auc") &&
            length(unique(truth_val)) < 2L) {
          value <- NA_real_
          warning(sprintf("Confounder '%s' level '%s': only one class present; %s is undefined.",
                          conf, lvl, metric_internal), call. = FALSE)
        } else {
          value <- .metric_value(metric_internal, fit@task, truth_val, sub$pred, pred_df = sub)
        }
      }
      pos_rate <- NA_real_
      if (identical(fit@task, "binomial") && n_sub > 0L) {
        pos_rate <- mean(sub$truth == pos_class, na.rm = TRUE)
      }
      out[[length(out) + 1L]] <- data.frame(
        confounder = conf,
        level = as.character(lvl),
        metric = metric,
        direction = direction,
        value = value,
        n = n_sub,
        positive_rate = pos_rate,
        stringsAsFactors = FALSE
      )
    }
  }

  if (!length(out)) return(data.frame())
  do.call(rbind, out)
}

test_that("calibration_summary computes metrics for binomial fits", {
  df <- make_class_df(30)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 3, seed = 1)
  custom <- make_custom_learners()
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "glm", custom_learners = custom,
                            metrics = "accuracy", refit = FALSE, seed = 1)
  cal <- calibration_summary(fit, bins = 5)
  expect_true(is.data.frame(cal$curve))
  expect_true(is.data.frame(cal$metrics))
  expect_true(all(c("ece", "mce", "brier") %in% names(cal$metrics)))
})

test_that("confounder_sensitivity summarizes metrics by confounder", {
  df <- make_class_df(30)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 3, seed = 1)
  custom <- make_custom_learners()
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "glm", custom_learners = custom,
                            metrics = "accuracy", refit = FALSE, seed = 1)
  cs <- confounder_sensitivity(fit, confounders = "batch", metric = "accuracy", min_n = 2)
  expect_true(nrow(cs) > 0)
  expect_true(all(c("confounder", "level", "metric", "value", "n") %in% names(cs)))
})

test_that("diagnostics require learner selection when multiple models are present", {
  df <- make_class_df(40)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 4, seed = 2)
  custom <- make_custom_learners()
  custom$glm2 <- custom$glm
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = c("glm", "glm2"), custom_learners = custom,
                            metrics = "accuracy", refit = FALSE, seed = 2)

  expect_error(calibration_summary(fit, bins = 5), "Multiple learners found")
  expect_error(confounder_sensitivity(fit, confounders = "batch", metric = "accuracy", min_n = 2),
               "Multiple learners found")

  cal <- calibration_summary(fit, bins = 5, learner = "glm")
  expect_true(is.data.frame(cal$curve))
  cs <- confounder_sensitivity(fit, confounders = "batch", metric = "accuracy",
                               min_n = 2, learner = "glm")
  expect_true(nrow(cs) > 0)
})

#' Leakage-safe data imputation via guarded preprocessing
#'
#' @description
#' Fits imputation parameters on the training data only, then applies the same
#' guarded transformation to the test data. This function is a thin wrapper
#' around the guarded preprocessing used by \code{fit_resample()}.
#' Output is the transformed feature matrix used by the guarded pipeline
#' (categorical variables are one-hot encoded).
#'
#' @param train data frame (training set)
#' @param test data frame (test set)
#' @param method one of "median", "knn", "missForest", or "none"
#' @param constant_value unused; retained for backward compatibility
#' @param k number of neighbors for kNN imputation (if method = "knn")
#' @param seed unused; retained for backward compatibility. Set seed before
#'   calling this function if reproducibility is needed.
#' @param winsor logical; apply MAD-based winsorization before imputation
#' @param winsor_thresh numeric; MAD cutoff (default = 3)
#' @param parallel logical; unused (kept for compatibility)
#' @param return_outliers logical; unused (outlier flags not returned)
#' @param vars optional character vector; impute only selected variables
#' @return A list (S3 class \code{"LeakImpute"}) with elements \code{train},
#'   \code{test}, \code{model}, \code{method}, \code{summary}, and \code{outliers}.
#' @seealso [fit_resample()], [predict_guard()]
#' @examples
#' train <- data.frame(x = c(1, 2, NA, 4), y = c(NA, 1, 1, 0))
#' test <- data.frame(x = c(NA, 5), y = c(1, NA))
#' imp <- impute_guarded(train, test, method = "median", winsor = FALSE)
#' imp$train
#' imp$test
#' @export
impute_guarded <- function(train,
                           test,
                           method = c("median", "knn", "missForest", "none"),
                           constant_value = 0,
                           k = 5,
                           seed = 123,
                           winsor = TRUE,
                           winsor_thresh = 3,
                           parallel = FALSE,
                           return_outliers = FALSE,
                           vars = NULL) {
  method <- match.arg(method)
  if (!is.data.frame(train) || !is.data.frame(test)) {
    stop("train and test must be data frames.", call. = FALSE)
  }

  if (!is.null(vars)) {
    vars <- intersect(vars, names(train))
    train <- train[, vars, drop = FALSE]
    test  <- test[, vars, drop = FALSE]
  }

  if (isTRUE(parallel)) {
    warning("parallel is unused in impute_guarded; guarded preprocessing runs sequentially.",
            call. = FALSE)
  }
  if (isTRUE(return_outliers)) {
    warning("return_outliers is not supported in impute_guarded; outlier flags are not returned.",
            call. = FALSE)
  }
  if (!identical(constant_value, 0) && method != "none") {
    warning("constant_value is unused in impute_guarded and will be ignored.",
            call. = FALSE)
  }

  steps <- list(
    impute = list(method = method, k = k, winsor = winsor, winsor_k = winsor_thresh),
    normalize = list(method = "none"),
    filter = list(var_thresh = 0, iqr_thresh = 0),
    fs = list(method = "none")
  )

  guard <- .guard_fit(train, y = NULL, steps = steps, task = "gaussian")
  train_imp <- guard$transform(train)
  test_imp <- guard$transform(test)

  diagnostics <- data.frame(
    variable = names(train),
    missing_train = colSums(is.na(train)),
    missing_test  = colSums(is.na(test)),
    stringsAsFactors = FALSE
  )
  diagnostics$missing_train <- as.integer(diagnostics$missing_train)
  diagnostics$missing_test  <- as.integer(diagnostics$missing_test)

  structure(
    list(
      train = train_imp,
      test  = test_imp,
      model = guard$state,
      method = method,
      summary = list(
        nmiss_train = sum(is.na(train)),
        nmiss_test  = sum(is.na(test)),
        winsorized = winsor,
        winsor_thresh = winsor_thresh,
        diagnostics = diagnostics,
        guard_audit = guard$state$audit
      ),
      outliers = NULL
    ),
    class = "LeakImpute"
  )
}

test_that("impute_guarded matches guarded preprocessing output", {
  train <- data.frame(
    a = c(1, 2, NA, 4),
    b = c(NA, 1, 2, 3)
  )
  test <- data.frame(
    a = c(NA, 5),
    b = c(1, NA)
  )

  res <- impute_guarded(train, test, method = "median", winsor = FALSE, seed = 1)
  guard <- guard_fit(
    train,
    y = NULL,
    steps = list(
      impute = list(method = "median", winsor = FALSE, winsor_k = 3),
      normalize = list(method = "none"),
      filter = list(var_thresh = 0, iqr_thresh = 0),
      fs = list(method = "none")
    ),
    task = "gaussian"
  )

  expect_equal(res$train, guard$transform(train))
  expect_equal(res$test, guard$transform(test))
})

test_that("impute_guarded validates inputs and emits warnings", {
  train <- data.frame(a = c(1, NA), b = c(2, 3))
  test <- data.frame(a = c(NA, 4), b = c(1, NA))

  expect_error(impute_guarded(1, test), "data frames")
  expect_error(impute_guarded(train, 1), "data frames")

  res <- expect_warning_match(
    impute_guarded(train, test, method = "median",
                   parallel = TRUE, return_outliers = TRUE,
                   constant_value = 1),
    "unused|not supported",
    all = TRUE
  )
  expect_equal(class(res), "LeakImpute")
})

test_that("impute_guarded respects vars selection", {
  train <- data.frame(a = c(1, NA), b = c(2, 3), c = c(4, 5))
  test <- data.frame(a = c(NA, 4), b = c(1, NA), c = c(6, 7))
  res <- impute_guarded(train, test, method = "median", vars = c("a", "b"))
  expect_equal(names(res$train), c("a", "b"))
  expect_equal(names(res$test), c("a", "b"))
})

# Influence-function based cvAUC standard errors --------------------------------
.psi_auc <- function(pos, neg) {
  outer(pos, neg, function(a, b) (a > b) + 0.5 * (a == b))
}

.cvauc_if <- function(pred, truth, weights = NULL) {
  stopifnot(length(pred) == length(truth))
  m <- length(pred)

  fold_auc <- numeric(m)
  fold_var <- numeric(m)
  fold_n   <- integer(m)
  fold_details <- vector("list", m)

  for (i in seq_len(m)) {
    pr <- pred[[i]]
    tr <- truth[[i]]
    if (length(pr) != length(tr)) stop("pred and truth length mismatch per fold")

    y <- if (is.factor(tr)) as.numeric(tr) - 1 else as.numeric(tr)
    pos <- pr[y == 1]
    neg <- pr[y == 0]
    n1 <- length(pos); n0 <- length(neg)
    fold_n[i] <- n1 + n0

    if (n1 == 0L || n0 == 0L) {
      fold_auc[i] <- NA_real_
      fold_var[i] <- NA_real_
      fold_details[[i]] <- data.frame(
        fold = i, auc = NA_real_, se = NA_real_, n_pos = n1, n_neg = n0
      )
      next
    }

    psi <- .psi_auc(pos, neg)
    v_i <- rowMeans(psi)
    w_j <- colMeans(psi)
    auc <- mean(v_i)
    var_pos <- if (n1 > 1L) stats::var(v_i, na.rm = TRUE) else 0
    var_neg <- if (n0 > 1L) stats::var(w_j, na.rm = TRUE) else 0
    var_fold <- var_pos / n1 + var_neg / n0
    fold_auc[i] <- auc
    fold_var[i] <- var_fold
    fold_details[[i]] <- data.frame(
      fold = i, auc = auc, se = sqrt(max(var_fold, 0)),
      n_pos = n1, n_neg = n0
    )
  }

  if (sum(fold_n, na.rm = TRUE) == 0L) {
    return(list(mean_auc = NA_real_, var_auc = NA_real_, se_auc = NA_real_,
                fold_stats = data.frame()))
  }

  if (is.null(weights)) weights <- fold_n / sum(fold_n)
  valid <- is.finite(fold_auc) & is.finite(fold_var)
  if (!any(valid)) {
    return(list(mean_auc = NA_real_, var_auc = NA_real_, se_auc = NA_real_,
                fold_stats = data.frame()))
  }
  weights <- weights[valid] / sum(weights[valid])
  fold_auc <- fold_auc[valid]
  fold_var <- fold_var[valid]

  auc_mean <- sum(weights * fold_auc)
  var_mean <- sum((weights^2) * fold_var)
  fold_stats <- do.call(rbind, fold_details[valid])

  list(
    mean_auc = auc_mean,
    var_auc  = var_mean,
    se_auc   = sqrt(max(var_mean, 0)),
    fold_stats = fold_stats
  )
}

test_that("se_ci_delta handles empty and zero-variance permutations", {
  out_empty <- bioLeak:::.se_ci_delta(0.1, 0.2, numeric(0))
  expect_true(all(is.na(unlist(out_empty))))

  perm <- rep(0.2, 10)
  out_zero <- bioLeak:::.se_ci_delta(0.1, 0.05, perm)
  expect_equal(out_zero$se, 0.05)
  expect_equal(length(out_zero$ci), 2)
  expect_true(is.finite(out_zero$z))
})

test_that("se_ci_delta handles singleton permutation vector", {
  out <- bioLeak:::.se_ci_delta(0.1, 0.05, 0.2)
  # Should return NA gracefully instead of erroring
  expect_true(is.na(out$se))
  expect_true(all(is.na(out$ci)))
  expect_true(is.na(out$z))
})

test_that("se_ci_delta handles finite variance permutations", {
  perm <- seq(0.1, 0.5, length.out = 25)
  out <- bioLeak:::.se_ci_delta(0.2, 0.1, perm, level = 0.9)
  expect_true(is.finite(out$se))
  expect_equal(length(out$ci), 2)
})

test_that("psi_auc and cvauc_if compute fold statistics", {
  pos <- c(0.9, 0.8)
  neg <- c(0.1, 0.2)
  psi <- bioLeak:::.psi_auc(pos, neg)
  expect_equal(dim(psi), c(2, 2))
  expect_true(all(psi %in% c(0, 0.5, 1)))

  pred <- list(c(0.9, 0.1, 0.8, 0.2), c(0.7, 0.3))
  truth <- list(factor(c(1, 0, 1, 0)), factor(c(1, 0)))
  res <- bioLeak:::.cvauc_if(pred, truth, weights = c(0.7, 0.3))
  expect_true(is.finite(res$mean_auc))
  expect_true(is.finite(res$se_auc))
  expect_equal(nrow(res$fold_stats), 2)
})

test_that("cvauc_if returns NA when no valid folds", {
  pred <- list(c(0.1, 0.2), c(0.3, 0.4))
  truth <- list(factor(c(1, 1)), factor(c(0, 0)))
  res <- bioLeak:::.cvauc_if(pred, truth)
  expect_true(all(is.na(c(res$mean_auc, res$var_auc, res$se_auc))))
  expect_equal(nrow(res$fold_stats), 0)
})

# Suppress NSE notes for ggplot2 aesthetics in R CMD check.
if (getRversion() >= "2.15.1") {
  utils::globalVariables(c(
    "acf",
    "count",
    "fold",
    "lag",
    "metric",
    "mid",
    "n",
    "obs_rate",
    "pred_mean",
    ".pred",
    ".pred_class",
    "prop_scaled",
    "series",
    "truth",
    "type",
    "level",
    "value"
  ))
}

# Nested tuning ---------------------------------------------------------------

#' Leakage-aware nested tuning with tidymodels
#'
#' Runs nested cross-validation for hyperparameter tuning using leakage-aware
#' splits. Inner resamples are constructed from each outer training fold to
#' avoid information leakage during tuning. Requires tidymodels tuning
#' packages and a workflow or recipe-based preprocessing. Survival tasks are
#' not yet supported.
#'
#' @param x SummarizedExperiment or matrix/data.frame.
#' @param outcome Outcome column name (if x is SE or data.frame).
#' @param splits LeakSplits object defining the outer resamples. If the splits
#'   do not already include inner folds, they are created from each outer
#'   training fold using the same split metadata. rsample splits must already
#'   include inner folds.
#' @param split_cols Optional named list/character vector or `"auto"` (default)
#'   overriding group/batch/study/time column names when `splits` is an rsample
#'   object and its attributes are missing. `"auto"` falls back to common
#'   metadata column names (e.g., `group`, `subject`, `batch`, `study`, `time`).
#'   Supported names are `group`, `batch`, `study`, and `time`.
#' @param learner A parsnip model_spec with tunable parameters, or a workflows
#'   workflow. When a model_spec is provided, a workflow is built using
#'   `preprocess` or a formula.
#' @param preprocess Optional `recipes::recipe`. Required when you need
#'   preprocessing for tuning. Ignored when `learner` is already a workflow.
#'   Recipe/workflow leakage guardrails run before tuning; configure policy via
#'   \code{options(bioLeak.validation_mode = "warn" | "error" | "off")}.
#' @param grid Tuning grid passed to `tune::tune_grid()`. Can be a data.frame or
#'   an integer size.
#' @param metrics Character vector of metric names (`auc`, `pr_auc`, `accuracy`,
#'   `macro_f1`, `log_loss`, `rmse`) or a yardstick metric set/list. Metrics are
#'   computed with yardstick; unsupported metrics are dropped with a warning.
#'   For binomial tasks, if any inner assessment fold contains a single class,
#'   probability metrics (`auc`, `roc_auc`, `pr_auc`) are dropped for tuning with
#'   a warning.
#' @param positive_class Optional value indicating the positive class for
#'   binomial outcomes. When set, the outcome levels are reordered so the
#'   positive class is second.
#' @param selection Selection rule for tuning, either `"best"` or `"one_std_err"`.
#' @param selection_metric Metric name used for selecting hyperparameters.
#'   Defaults to the first metric in `metrics`. If the chosen metric yields
#'   no valid results, the first available metric is used with a warning.
#' @param inner_v Optional number of folds for inner CV when inner splits are
#'   not precomputed. Defaults to the outer `v`.
#' @param inner_repeats Optional number of repeats for inner CV when inner
#'   splits are not precomputed. Defaults to 1.
#' @param inner_seed Optional seed for inner split generation when inner splits
#'   are not precomputed. Defaults to the outer split seed.
#' @param control Optional `tune::control_grid()` settings for tuning.
#' @param tune_threshold Logical; when `TRUE` for binomial tasks, selects a
#'   probability threshold from inner-fold predictions and applies it only to the
#'   corresponding outer-fold evaluation.
#' @param threshold_grid Numeric vector of thresholds in `[0, 1]` considered when
#'   `tune_threshold = TRUE`.
#' @param threshold_metric Metric used to pick thresholds when
#'   `tune_threshold = TRUE`. Supported values are `"accuracy"`,
#'   `"balanced_accuracy"`, and `"f1"`, or a custom function with signature
#'   `function(truth, pred_class, prob, threshold)`.
#' @param parallel Logical; passed to [fit_resample()] when evaluating outer
#'   folds (single-fold, no refit).
#' @param refit Logical; if TRUE, refits a final tuned workflow on the full
#'   dataset using aggregated hyperparameters across all outer folds (median
#'   for numeric parameters, majority vote for categorical). This avoids
#'   nested-CV leakage that would occur from selecting a single fold's params.
#' @param seed Integer seed for reproducibility.
#' @return A list of class `"LeakTune"` with components:
#'   \item{metrics}{Outer-fold metrics.}
#'   \item{metric_summary}{Mean/SD metrics across outer folds with columns
#'     \code{learner}, and \code{<metric>_mean} and \code{<metric>_sd} for
#'     each metric.}
#'   \item{best_params}{Best hyperparameters per outer fold.}
#'   \item{inner_results}{List of inner tuning results.}
#'   \item{outer_fits}{List of outer LeakFit objects.}
#'   \item{thresholds}{Per-fold threshold choices when threshold tuning is enabled.}
#'   \item{fold_status}{Outer-fold status log with stage, status, reason, and notes.}
#'   \item{final_model}{Optional final workflow fit when `refit = TRUE`.}
#'   \item{info}{Metadata about the tuning run.}
#' @examples
#' \donttest{
#'   if (requireNamespace("tune", quietly = TRUE) &&
#'       requireNamespace("recipes", quietly = TRUE) &&
#'       requireNamespace("glmnet", quietly = TRUE) &&
#'       requireNamespace("rsample", quietly = TRUE) &&
#'       requireNamespace("workflows", quietly = TRUE) &&
#'       requireNamespace("yardstick", quietly = TRUE) &&
#'      requireNamespace("dials", quietly = TRUE)) {
#'     df <- data.frame(
#'       subject = rep(1:10, each = 2),
#'       outcome = factor(rep(c(0, 1), each = 10)),
#'       x1 = rnorm(20),
#'       x2 = rnorm(20)
#'     )
#'     splits <- make_split_plan(df, outcome = "outcome",
#'                          mode = "subject_grouped", group = "subject",
#'                          v = 3, nested = TRUE, stratify = TRUE)
#'     spec <- parsnip::logistic_reg(penalty = tune::tune(), mixture = 1) |>
#'       parsnip::set_engine("glmnet")
#'     rec <- recipes::recipe(outcome ~ x1 + x2, data = df)
#'     tuned <- tune_resample(df, outcome = "outcome", splits = splits,
#'                           learner = spec, preprocess = rec, grid = 5)
#'     tuned$metric_summary
#'   }
#'}
#'
#' @export
tune_resample <- function(x, outcome, splits,
                          learner,
                          preprocess = NULL,
                          grid = 10,
                          metrics = NULL,
                          positive_class = NULL,
                          selection = c("best", "one_std_err"),
                          selection_metric = NULL,
                          inner_v = NULL,
                          inner_repeats = 1,
                          inner_seed = NULL,
                          control = NULL,
                          parallel = FALSE,
                          refit = FALSE,
                          seed = 1,
                          split_cols = "auto",
                          tune_threshold = FALSE,
                          threshold_grid = seq(0.1, 0.9, by = 0.05),
                          threshold_metric = "accuracy") {

  selection <- match.arg(selection)
  .bio_strict_checks(context = "tune_resample", seed = seed)
  if (!is.logical(refit) || length(refit) != 1L || is.na(refit)) {
    .bio_stop("refit must be TRUE or FALSE.", "bioLeak_input_error")
  }
  if (!is.logical(tune_threshold) || length(tune_threshold) != 1L || is.na(tune_threshold)) {
    .bio_stop("tune_threshold must be TRUE or FALSE.", "bioLeak_input_error")
  }
  if (!is.numeric(threshold_grid) || !length(threshold_grid)) {
    .bio_stop("threshold_grid must be a non-empty numeric vector.", "bioLeak_input_error")
  }
  threshold_grid <- sort(unique(as.numeric(threshold_grid)))
  if (any(!is.finite(threshold_grid)) || any(threshold_grid < 0 | threshold_grid > 1)) {
    .bio_stop("threshold_grid must contain finite values in [0, 1].", "bioLeak_input_error")
  }
  threshold_metric_is_fn <- is.function(threshold_metric)
  if (!threshold_metric_is_fn) {
    if (!is.character(threshold_metric) || length(threshold_metric) != 1L) {
      .bio_stop("threshold_metric must be one of: accuracy, balanced_accuracy, f1, or a function.",
                "bioLeak_input_error")
    }
    threshold_metric <- match.arg(threshold_metric, c("accuracy", "balanced_accuracy", "f1"))
  }

  if (!requireNamespace("tune", quietly = TRUE)) {
    .bio_stop("Package 'tune' is required for tune_resample().", "bioLeak_dependency_error")
  }
  if (!requireNamespace("dials", quietly = TRUE)) {
    .bio_stop("Package 'dials' is required for tune_resample().", "bioLeak_dependency_error")
  }
  if (!requireNamespace("yardstick", quietly = TRUE)) {
    .bio_stop("Package 'yardstick' is required for tune_resample().", "bioLeak_dependency_error")
  }
  if (!requireNamespace("rsample", quietly = TRUE)) {
    .bio_stop("Package 'rsample' is required for tune_resample().", "bioLeak_dependency_error")
  }
  if (!requireNamespace("workflows", quietly = TRUE)) {
    .bio_stop("Package 'workflows' is required for tune_resample().", "bioLeak_dependency_error")
  }

  is_parsnip_spec <- function(obj) inherits(obj, "model_spec")
  is_workflow <- function(obj) inherits(obj, "workflow")

  if (!is_workflow(learner) && !is_parsnip_spec(learner)) {
    .bio_stop("learner must be a parsnip model_spec or a workflows::workflow.", "bioLeak_input_error")
  }

  if (is_workflow(learner) && !is.null(preprocess)) {
    warning("preprocess ignored when learner is a workflow.")
  }
  if (!is.null(preprocess) && !inherits(preprocess, "recipe") && !is_workflow(learner)) {
    .bio_stop("tune_resample requires a recipe (or a workflow) for preprocessing.",
              "bioLeak_input_error")
  }
  if (!is_workflow(learner) && inherits(preprocess, "recipe") &&
      !requireNamespace("recipes", quietly = TRUE)) {
    .bio_stop("Package 'recipes' is required when preprocess is a recipe.",
              "bioLeak_dependency_error")
  }
  validation_mode <- .bio_validation_mode()
  if (inherits(preprocess, "recipe") && !is_workflow(learner)) {
    .bio_validate_recipe_graph(
      preprocess,
      context = "tune_resample",
      mode = validation_mode
    )
  }
  if (is_workflow(learner)) {
    .bio_validate_workflow_graph(
      learner,
      context = "tune_resample",
      mode = validation_mode
    )
  }

  # --- Learner Name Extraction ---
  derive_learner_label <- function(obj) {
    spec <- obj
    if (inherits(obj, "workflow")) {
      if (requireNamespace("workflows", quietly = TRUE)) {
        spec <- tryCatch(workflows::extract_spec_parsnip(obj), error = function(e) NULL)
      }
    }
    if (inherits(spec, "model_spec")) {
      cls <- class(spec)
      cls <- cls[!cls %in% c("model_spec", "object")]
      model_type <- if (length(cls) > 0) cls[1] else "model"
      engine <- spec$engine
      if (!is.null(engine)) {
        return(paste0(model_type, "/", engine))
      } else {
        return(model_type)
      }
    }
    return("tuned_model")
  }
  learner_label <- derive_learner_label(learner)
  # -------------------------------

  if (!inherits(splits, "LeakSplits")) {
    if (.bio_is_rsample(splits)) {
      coldata <- if (.bio_is_se(x)) {
        .bio_coldata_df(x)
      } else if (is.data.frame(x)) {
        x
      } else if (is.matrix(x)) {
        data.frame(row_id = seq_len(nrow(x)))
      } else {
        NULL
      }
      splits <- .bio_as_leaksplits_from_rsample(splits, n = nrow(.bio_get_x(x)), coldata = coldata,
                                                split_cols = split_cols)
    } else {
      .bio_stop("splits must be a LeakSplits or rsample rset/rsplit.", "bioLeak_input_error")
    }
  }
  if (identical(splits@mode, "rsample") && is.null(splits@info$inner)) {
    .bio_stop("rsample splits require precomputed inner folds for tune_resample().", "bioLeak_input_error")
  }

  Xall <- .bio_get_x(x)
  y_orig <- .bio_get_y(x, outcome)
  yall <- y_orig
  y_data <- y_orig

  if (.bio_is_survival(yall)) {
    .bio_stop("tune_resample does not yet support survival tasks.", "bioLeak_input_error")
  }

  drop_cols <- outcome
  split_info <- splits@info
  drop_cols <- unique(c(drop_cols,
                        split_info$group,
                        split_info$batch,
                        split_info$study,
                        split_info$time))
  drop_cols <- drop_cols[!is.na(drop_cols) & nzchar(drop_cols)]
  if (length(drop_cols) && !is.null(colnames(Xall))) {
    Xall <- Xall[, setdiff(colnames(Xall), drop_cols), drop = FALSE]
  }

  task <- if (.bio_is_binomial(yall)) "binomial"
  else if (.bio_is_multiclass(yall)) "multiclass"
  else if (.bio_is_regression(yall)) "gaussian"
  else if (is.factor(yall) && nlevels(yall) == 2) "binomial"
  else if (is.factor(yall) && nlevels(yall) > 2) "multiclass"
  else .bio_stop("Unsupported outcome type for tuning.", "bioLeak_input_error")

  class_levels <- NULL
  if (task == "binomial") {
    if (!is.factor(yall)) {
      yall <- factor(yall)
      y_data <- factor(y_data)
    }
    yall <- droplevels(yall)
    if (nlevels(yall) != 2) {
      .bio_stop("Binomial task requires exactly two outcome levels.", "bioLeak_input_error")
    }
    if (!is.null(positive_class)) {
      pos_chr <- as.character(positive_class)
      if (length(pos_chr) != 1L) {
        .bio_stop("positive_class must be a single value.", "bioLeak_input_error")
      }
      levels_y <- levels(yall)
      if (!pos_chr %in% levels_y) {
        .bio_stop(sprintf("positive_class '%s' not found in outcome levels: %s",
                          pos_chr, paste(levels_y, collapse = ", ")),
                  "bioLeak_input_error")
      }
      if (!identical(pos_chr, levels_y[2])) {
        levels_y <- c(setdiff(levels_y, pos_chr), pos_chr)
        yall <- factor(yall, levels = levels_y)
        if (!inherits(preprocess, "recipe")) {
          y_data <- factor(y_data, levels = levels_y)
        }
      }
    }
    class_levels <- levels(yall)
  } else if (task == "multiclass") {
    if (!is.factor(yall)) {
      yall <- factor(yall)
      y_data <- factor(y_data)
    }
    yall <- droplevels(yall)
    if (nlevels(yall) < 3) {
      .bio_stop("Multiclass task requires 3 or more outcome levels.", "bioLeak_input_error")
    }
    if (!is.null(positive_class)) {
      warning("positive_class is ignored for multiclass tasks.")
    }
    class_levels <- levels(yall)
  } else if (!is.numeric(yall)) {
    yall <- as.numeric(yall)
    y_data <- as.numeric(y_data)
    if (anyNA(yall)) stop("Gaussian task requires numeric outcome values.", call. = FALSE)
  }

  if (isTRUE(tune_threshold) && !identical(task, "binomial")) {
    warning("tune_threshold is only supported for binomial tasks; disabling threshold tuning.",
            call. = FALSE)
    tune_threshold <- FALSE
  }

  if (is.null(control)) control <- tune::control_grid()
  if (isTRUE(tune_threshold)) {
    if (!"save_pred" %in% names(control)) {
      stop("control must be created by tune::control_grid() when tune_threshold = TRUE.",
           call. = FALSE)
    }
    control$save_pred <- TRUE
  }

  resolve_metrics <- function(metrics, task) {
    macro_f1 <- yardstick::metric_tweak("macro_f1", yardstick::f_meas, estimator = "macro")
    auc_metric <- yardstick::metric_tweak("auc", yardstick::roc_auc, event_level = "second")
    roc_metric <- yardstick::metric_tweak("roc_auc", yardstick::roc_auc, event_level = "second")
    pr_metric <- yardstick::metric_tweak("pr_auc", yardstick::pr_auc, event_level = "second")
    log_loss_metric <- yardstick::metric_tweak("log_loss", yardstick::mn_log_loss)
    mn_log_loss_metric <- yardstick::metric_tweak("mn_log_loss", yardstick::mn_log_loss)

    defaults <- if (task == "binomial") c("auc", "pr_auc", "accuracy")
    else if (task == "multiclass") c("accuracy", "macro_f1")
    else c("rmse")

    metric_map <- list(
      auc = auc_metric,
      roc_auc = roc_metric,
      pr_auc = pr_metric,
      accuracy = yardstick::accuracy,
      macro_f1 = macro_f1,
      log_loss = log_loss_metric,
      mn_log_loss = mn_log_loss_metric,
      rmse = yardstick::rmse
    )

    if (is.null(metrics)) metrics <- defaults

    if (is.character(metrics)) {
      metrics[metrics == "auc"] <- "roc_auc"
      valid_keys <- intersect(metrics, names(metric_map))
      invalid <- setdiff(metrics, names(metric_map))

      if (length(invalid)) {
        warning(sprintf("Dropping unsupported metrics for %s task: %s", task,
                        paste(invalid, collapse = ", ")))
      }

      if (length(valid_keys) == 0) valid_keys <- defaults[defaults %in% names(metric_map)]

      fns <- metric_map[valid_keys]
      metric_set <- do.call(yardstick::metric_set, fns)
      return(list(set = metric_set, names = valid_keys))
    }

    if (inherits(metrics, "metric_set")) {
      return(list(set = metrics, names = names(attr(metrics, "metrics"))))
    }

    if (is.list(metrics)) {
      metric_set <- do.call(yardstick::metric_set, metrics)
      return(list(set = metric_set, names = names(attr(metric_set, "metrics"))))
    }

    stop("metrics must be a character vector or yardstick metric set.", call. = FALSE)
  }

  subset_metric_set <- function(metric_set, keep) {
    metric_list <- attr(metric_set, "metrics")
    metric_list <- metric_list[names(metric_list) %in% keep]
    do.call(yardstick::metric_set, metric_list)
  }

  metrics_resolved <- resolve_metrics(metrics, task)
  tune_metrics <- metrics_resolved$set
  metric_names <- metrics_resolved$names

  if (is.null(selection_metric)) {
    if ("roc_auc" %in% metric_names) selection_metric <- "roc_auc"
    else if ("auc" %in% metric_names) selection_metric <- "auc"
    else selection_metric <- metric_names[[1]]
  }

  if (!selection_metric %in% metric_names) {
    if (selection_metric == "auc" && "roc_auc" %in% metric_names) selection_metric <- "roc_auc"
    else if (selection_metric == "roc_auc" && "auc" %in% metric_names) selection_metric <- "auc"
    else stop(sprintf("selection_metric '%s' not found in metrics.", selection_metric), call. = FALSE)
  }

  make_fold_df <- function(X, y) {
    df <- as.data.frame(X, check.names = FALSE)
    df[[outcome]] <- y
    df[, c(outcome, setdiff(names(df), outcome)), drop = FALSE]
  }

  has_two_classes <- function(y) {
    y <- y[!is.na(y)]
    length(unique(y)) >= 2L
  }

  evaluate_threshold_metric <- function(truth, prob, threshold, metric) {
    pred_class <- factor(
      ifelse(prob >= threshold, class_levels[2], class_levels[1]),
      levels = class_levels
    )
    if (is.function(metric)) {
      return(as.numeric(metric(truth, pred_class, prob, threshold)))
    }
    if (identical(metric, "accuracy")) {
      return(mean(pred_class == truth))
    }

    truth_pos <- truth == class_levels[2]
    pred_pos <- pred_class == class_levels[2]
    tp <- sum(truth_pos & pred_pos)
    tn <- sum(!truth_pos & !pred_pos)
    fp <- sum(!truth_pos & pred_pos)
    fn <- sum(truth_pos & !pred_pos)

    if (identical(metric, "balanced_accuracy")) {
      sens <- if ((tp + fn) > 0) tp / (tp + fn) else NA_real_
      spec <- if ((tn + fp) > 0) tn / (tn + fp) else NA_real_
      return(mean(c(sens, spec), na.rm = TRUE))
    }
    if (identical(metric, "f1")) {
      precision <- if ((tp + fp) > 0) tp / (tp + fp) else NA_real_
      recall <- if ((tp + fn) > 0) tp / (tp + fn) else NA_real_
      if (!is.finite(precision) || !is.finite(recall) || (precision + recall) == 0) {
        return(NA_real_)
      }
      return(2 * precision * recall / (precision + recall))
    }
    NA_real_
  }

  select_threshold <- function(truth, prob, grid, metric) {
    scores <- vapply(grid, function(thr) {
      evaluate_threshold_metric(truth = truth, prob = prob, threshold = thr, metric = metric)
    }, numeric(1))
    valid <- is.finite(scores)
    if (!any(valid)) {
      return(list(threshold = 0.5, metric_value = NA_real_))
    }
    grid_ok <- grid[valid]
    scores_ok <- scores[valid]
    best <- which(scores_ok == max(scores_ok))
    if (length(best) > 1L) {
      best <- best[order(abs(grid_ok[best] - 0.5), grid_ok[best])[1L]]
    } else {
      best <- best[1L]
    }
    list(threshold = grid_ok[[best]], metric_value = scores_ok[[best]])
  }

  available_metrics <- function(metrics_df) {
    if (is.null(metrics_df) || !nrow(metrics_df)) return(character())
    if (!".metric" %in% names(metrics_df)) return(character())
    value_col <- NULL
    if (".estimate" %in% names(metrics_df)) {
      value_col <- ".estimate"
    } else if ("estimate" %in% names(metrics_df)) {
      value_col <- "estimate"
    } else if ("mean" %in% names(metrics_df)) {
      value_col <- "mean"
    }
    if (is.null(value_col)) return(character())
    metric_ok <- tapply(!is.na(metrics_df[[value_col]]), metrics_df$.metric, any)
    names(metric_ok)[metric_ok]
  }

  infer_simplicity_direction <- function(name) {
    nm <- tolower(name)
    simpler_when_higher <- c(
      "penalty", "lambda", "cost_complexity", "cp", "gamma",
      "min_n", "min.node.size", "min_child_weight", "loss_reduction",
      "sample_size", "margin"
    )
    simpler_when_lower <- c(
      "mtry", "trees", "num_trees", "tree_depth", "max_depth", "depth",
      "degree", "neighbors", "k", "num_terms", "num_comp", "num_components",
      "hidden_units", "size", "epochs", "iter", "stop_iter", "learn_rate",
      "mixture"
    )
    if (nm %in% simpler_when_higher) return("higher")
    if (nm %in% simpler_when_lower) return("lower")
    "lower"
  }

  simplify_rank <- function(df, config_col = ".config") {
    if (!nrow(df)) return(numeric(0))
    param_cols <- setdiff(names(df), c(config_col, "mean", "sd", "n", "std_err"))
    if (!length(param_cols)) return(rep(0, nrow(df)))

    score <- rep(0, nrow(df))
    for (col in param_cols) {
      v <- df[[col]]
      dir <- infer_simplicity_direction(col)
      if (is.logical(v)) v <- as.numeric(v)
      if (is.factor(v)) v <- as.character(v)
      if (is.character(v)) {
        ord <- rank(v, ties.method = "average", na.last = "keep")
      } else if (is.numeric(v)) {
        ord <- rank(v, ties.method = "average", na.last = "keep")
      } else {
        ord <- rank(as.character(v), ties.method = "average", na.last = "keep")
      }
      if (dir == "higher") {
        finite_ord <- ord[is.finite(ord)]
        max_ord <- if (length(finite_ord)) max(finite_ord) else 0
        ord <- ifelse(is.finite(ord), (max_ord + 1) - ord, ord)
      }
      ord[!is.finite(ord)] <- max(ord[is.finite(ord)], 0) + 1
      score <- score + ord
    }
    score
  }

  select_config <- function(metrics_df, metric_name, selection_rule) {
    res_sub <- metrics_df[
      metrics_df$.metric == metric_name & is.finite(metrics_df$.estimate),
      c(".config", ".estimate"),
      drop = FALSE
    ]
    if (!nrow(res_sub)) return(NULL)

    agg_mean <- aggregate(.estimate ~ .config, data = res_sub, FUN = mean, na.rm = TRUE)
    names(agg_mean)[names(agg_mean) == ".estimate"] <- "mean"
    agg_sd <- aggregate(.estimate ~ .config, data = res_sub, FUN = stats::sd, na.rm = TRUE)
    names(agg_sd)[names(agg_sd) == ".estimate"] <- "sd"
    agg_n <- aggregate(.estimate ~ .config, data = res_sub, FUN = length)
    names(agg_n)[names(agg_n) == ".estimate"] <- "n"
    cfg <- Reduce(function(x, y) merge(x, y, by = ".config", all = TRUE),
                  list(agg_mean, agg_sd, agg_n))
    cfg$std_err <- ifelse(cfg$n > 1L & is.finite(cfg$sd),
                          cfg$sd / sqrt(cfg$n),
                          0)

    param_cols <- setdiff(names(metrics_df), c(".metric", ".estimator", ".estimate", "n", "std_err", ".config",
                                               "id", "id2", ".notes"))
    params_by_cfg <- unique(metrics_df[, c(".config", param_cols), drop = FALSE])
    cfg <- merge(cfg, params_by_cfg, by = ".config", all.x = TRUE, sort = FALSE)

    minimize <- metric_name %in% c("rmse", "mae", "log_loss", "mn_log_loss")
    best_idx <- if (minimize) which.min(cfg$mean) else which.max(cfg$mean)
    if (!length(best_idx) || !is.finite(cfg$mean[best_idx[1]])) return(NULL)
    best_idx <- best_idx[1]

    cand <- cfg
    if (identical(selection_rule, "one_std_err")) {
      best_mean <- cfg$mean[best_idx]
      best_se <- cfg$std_err[best_idx]
      if (!is.finite(best_se)) best_se <- 0
      keep <- if (minimize) {
        cfg$mean <= (best_mean + best_se)
      } else {
        cfg$mean >= (best_mean - best_se)
      }
      cand <- cfg[keep & is.finite(cfg$mean), , drop = FALSE]
      if (!nrow(cand)) cand <- cfg[best_idx, , drop = FALSE]
    } else {
      tol <- sqrt(.Machine$double.eps)
      tied <- which(abs(cfg$mean - cfg$mean[best_idx]) <= tol)
      cand <- cfg[tied, , drop = FALSE]
      if (!nrow(cand)) cand <- cfg[best_idx, , drop = FALSE]
    }

    cand$.simplicity <- simplify_rank(cand)
    ord <- order(cand$.simplicity, cand$.config, na.last = TRUE)
    chosen <- cand[ord[1], , drop = FALSE]

    params <- chosen[, setdiff(names(chosen), c(".config", "mean", "sd", "n", "std_err", ".simplicity")),
                     drop = FALSE]
    list(config = chosen$.config[[1]], params = params, metric = metric_name)
  }

  build_workflow <- function() {
    if (is_workflow(learner)) return(learner)
    if (inherits(preprocess, "recipe")) {
      return(workflows::workflow() |> workflows::add_model(learner) |>
               workflows::add_recipe(preprocess))
    }
    if (length(outcome) != 1L) {
      stop("Formula workflows require a single outcome column.", call. = FALSE)
    }
    form <- stats::as.formula(paste0("`", outcome, "` ~ ."))
    workflows::workflow() |>
      workflows::add_model(learner) |>
      workflows::add_formula(form)
  }

  base_workflow <- build_workflow()

  make_inner_rset <- function(inner_idx, data) {
    if (!length(inner_idx)) stop("Inner splits are empty.", call. = FALSE)
    ids <- vapply(seq_along(inner_idx), function(i) {
      fold <- inner_idx[[i]]
      if (!is.null(fold$fold)) {
        rep_id <- fold$repeat_id %||% 1L
        paste0("Fold", fold$fold, "_Repeat", rep_id)
      } else {
        paste0("Fold", i)
      }
    }, character(1))
    split_objs <- lapply(seq_along(inner_idx), function(i) {
      fold <- inner_idx[[i]]
      if (is.null(fold$train) || is.null(fold$test)) {
        stop("Inner split is missing train/test indices.", call. = FALSE)
      }
      .bio_make_rsplit(fold$train, fold$test, data = data, id = ids[[i]])
    })
    manual_rset <- utils::getFromNamespace("manual_rset", "rsample")
    do.call(manual_rset, list(splits = split_objs, ids = ids))
  }

  make_outer_splits <- function(train, test) {
    info <- splits@info
    info$v <- 1L
    info$repeats <- 1L
    info$nested <- FALSE
    info$inner <- NULL
    info$compact <- FALSE
    info$fold_assignments <- NULL
    info$summary <- data.frame(
      fold = 1L,
      repeat_id = 1L,
      train_n = length(train),
      test_n = length(test),
      stringsAsFactors = FALSE
    )
    indices <- list(list(train = train, test = test, fold = 1L, repeat_id = 1L))
    info$hash <- .bio_hash_indices(indices)
    new("LeakSplits", mode = splits@mode, indices = indices, info = info)
  }

  inner_seed <- inner_seed %||% (splits@info$seed %||% seed)
  inner_v <- inner_v %||% (splits@info$v %||% 5L)
  inner_repeats <- inner_repeats %||% 1L

  df_all <- if (is.data.frame(x)) {
    x
  } else {
    make_fold_df(Xall, y_data)
  }

  coldata <- if (.bio_is_se(x)) {
    .bio_coldata_df(x)
  } else if (is.data.frame(x)) {
    x
  } else if (is.matrix(x)) {
    data.frame(row_id = seq_len(nrow(Xall)))
  } else {
    NULL
  }

  folds <- splits@indices
  metrics_rows <- list()
  best_rows <- list()
  inner_results <- list()
  outer_fits <- list()
  threshold_rows <- list()
  threshold_metric_label <- if (is.function(threshold_metric)) "custom" else threshold_metric
  skipped_single_class_outer <- 0L
  skipped_single_class_inner <- 0L
  skipped_no_results <- 0L
  metrics_used <- character()
  selection_metric_requested <- selection_metric
  selection_metric_used <- selection_metric
  selection_metric_warned <- FALSE
  fold_status_rows <- vector("list", length(folds))
  mark_fold_status <- function(fold_id, stage, status, reason = NA_character_, notes = NA_character_) {
    fold_status_rows[[fold_id]] <<- data.frame(
      fold = fold_id,
      stage = as.character(stage),
      status = as.character(status),
      reason = as.character(reason),
      notes = as.character(notes),
      stringsAsFactors = FALSE
    )
  }

  for (i in seq_along(folds)) {
    fold <- folds[[i]]
    fold <- .bio_resolve_fold_indices(splits, fold, n = nrow(df_all), data = coldata)
    tr <- fold$train
    te <- fold$test
    if (!length(tr) || !length(te)) {
      warning(sprintf("Outer fold %d skipped: empty train/test.", i), call. = FALSE)
      mark_fold_status(i, stage = "outer_split", status = "skipped",
                       reason = "empty_train_or_test", notes = "Outer split has empty train or test indices.")
      next
    }

    df_train <- df_all[tr, , drop = FALSE]
    y_train_logic <- yall[tr]

    if (task %in% c("binomial", "multiclass") && !has_two_classes(y_train_logic)) {
      warning(sprintf("Outer fold %d skipped: training data has a single outcome class.", i),
              call. = FALSE)
      skipped_single_class_outer <- skipped_single_class_outer + 1L
      mark_fold_status(i, stage = "outer_train", status = "skipped",
                       reason = "single_class_training",
                       notes = "Outer training data contains fewer than two outcome classes.")
      next
    }

    inner_idx <- NULL
    inner_precomputed <- FALSE
    if (!is.null(splits@info$inner) && length(splits@info$inner) >= i) {
      inner_idx <- splits@info$inner[[i]]
      inner_precomputed <- TRUE
    }
    if (is.null(inner_idx)) {
      if (identical(splits@mode, "rsample")) {
        inner_mode <- "subject_grouped"
        if (!is.null(splits@info$perm_mode)) inner_mode <- splits@info$perm_mode
        grp <- splits@info$group
        if(is.null(grp) && "subject" %in% names(df_train)) grp <- "subject"
        if(is.null(grp) && "id" %in% names(df_train)) grp <- "id"
        x_inner <- if (.bio_is_se(x)) x[, tr] else x[tr, , drop = FALSE]
      } else {
        x_inner <- if (.bio_is_se(x)) x[, tr] else x[tr, , drop = FALSE]
      }

      inner <- make_split_plan(
        x_inner,
        outcome = outcome,
        mode = splits@mode,
        group = splits@info$group,
        batch = splits@info$batch,
        study = splits@info$study,
        time = splits@info$time,
        constraints = splits@info$constraints,
        v = inner_v,
        repeats = inner_repeats,
        stratify = isTRUE(splits@info$stratify),
        nested = FALSE,
        seed = inner_seed + i,
        horizon = splits@info$horizon %||% 0,
        purge = splits@info$purge %||% 0,
        embargo = splits@info$embargo %||% 0,
        progress = FALSE
      )
      inner_idx <- inner@indices
    }

    inner_global <- FALSE
    if (inner_precomputed) {
      inner_positions <- unlist(lapply(inner_idx, function(fold) {
        c(fold$train, fold$test)
      }), use.names = FALSE)
      inner_positions <- inner_positions[!is.na(inner_positions)]
      if (!length(inner_positions)) {
        stop(sprintf("Outer fold %d: inner splits are empty.", i), call. = FALSE)
      }
      if (any(inner_positions < 1L)) {
        stop(sprintf("Outer fold %d: inner split indices must be positive.", i), call. = FALSE)
      }
      if (any(inner_positions > nrow(df_all))) {
        stop(sprintf("Outer fold %d: inner split indices exceed available rows.", i),
             call. = FALSE)
      }
      inner_in_train <- all(inner_positions %in% seq_len(nrow(df_train)))
      inner_in_outer <- all(inner_positions %in% tr)
      if (!inner_in_train && inner_in_outer) {
        inner_global <- TRUE
      } else if (inner_in_outer && !identical(tr, seq_len(nrow(df_train)))) {
        inner_global <- TRUE
      } else if (!inner_in_train && !inner_in_outer) {
        stop(sprintf(
          "Outer fold %d: inner split indices do not match outer training data.",
          i
        ), call. = FALSE)
      }
    }

    # Precomputed inner indices may be global; choose the matching data source.
    y_inner <- if (inner_global) yall else y_train_logic
    if (task %in% c("binomial", "multiclass")) {
      inner_has_class <- vapply(inner_idx, function(fold) {
        idx <- fold$train
        if (is.null(idx)) return(FALSE)
        has_two_classes(y_inner[idx])
      }, logical(1))
      if (!any(inner_has_class)) {
        warning(sprintf(
          "Outer fold %d skipped: inner training folds have a single outcome class. Consider `stratify = TRUE` or reducing `inner_v`.",
          i
        ), call. = FALSE)
        skipped_single_class_inner <- skipped_single_class_inner + 1L
        mark_fold_status(i, stage = "inner_train", status = "skipped",
                         reason = "single_class_training",
                         notes = "All inner training folds have fewer than two outcome classes.")
        next
      }
      if (task == "binomial") {
        drop_count <- sum(!inner_has_class)
        if (drop_count > 0L) {
          inner_idx <- inner_idx[inner_has_class]
          warning(sprintf(
            "Outer fold %d: dropped %d inner fold(s) with single-class training data.",
            i, drop_count
          ), call. = FALSE)
        }
      }
    }

    tune_metrics_fold <- tune_metrics
    metric_names_fold <- metric_names
    if (task == "binomial") {
      inner_assess_two <- vapply(inner_idx, function(fold) {
        idx <- fold$test
        if (is.null(idx)) return(FALSE)
        has_two_classes(y_inner[idx])
      }, logical(1))
      if (!all(inner_assess_two)) {
        drop_metrics <- intersect(metric_names_fold, c("auc", "roc_auc", "pr_auc"))
        keep_metrics <- setdiff(metric_names_fold, drop_metrics)
        if (length(drop_metrics) && length(keep_metrics)) {
          tune_metrics_fold <- subset_metric_set(tune_metrics_fold, keep_metrics)
          metric_names_fold <- keep_metrics
          warning(sprintf(
            "Outer fold %d: inner assessment folds lack both classes; dropping metrics %s for tuning.",
            i, paste(drop_metrics, collapse = ", ")
          ), call. = FALSE)
        }
      }
    }

    metrics_used <- unique(c(metrics_used, metric_names_fold))

    inner_data <- if (inner_global) df_all else df_train
    inner_rset <- make_inner_rset(inner_idx, inner_data)
    set.seed(seed + i)
    tune_res <- tune::tune_grid(
      base_workflow,
      resamples = inner_rset,
      grid = grid,
      metrics = tune_metrics_fold,
      control = control
    )
    inner_results[[i]] <- tune_res
    no_results <- FALSE
    metrics_raw <- tryCatch(
      tune::collect_metrics(tune_res, summarize = FALSE),
      error = function(e) {
        if (grepl("No results are available", conditionMessage(e), fixed = TRUE)) {
          no_results <<- TRUE
          return(NULL)
        }
        stop(e)
      }
    )
    if (isTRUE(no_results) || is.null(metrics_raw) || !nrow(metrics_raw)) {
      warning(sprintf(
        "Outer fold %d skipped: inner tuning produced no results. Check `tune::collect_notes()` for failures.",
        i
      ), call. = FALSE)
      skipped_no_results <- skipped_no_results + 1L
      mark_fold_status(i, stage = "inner_tuning", status = "skipped",
                       reason = "no_results",
                       notes = "tune::collect_metrics() returned no rows.")
      next
    }
    avail_metrics <- available_metrics(metrics_raw)
    metric_used <- selection_metric_used

    if (!metric_used %in% avail_metrics) {
      if ("<prb_mtrc>" %in% avail_metrics && metric_used %in% c("roc_auc", "auc", "pr_auc")) {
        metric_used <- "<prb_mtrc>"
      } else if ("<clss_mtr>" %in% avail_metrics && metric_used %in% c("accuracy", "kap")) {
        metric_used <- "<clss_mtr>"
      } else {
        if (length(avail_metrics)) {
          metric_used <- avail_metrics[1]
          if (!selection_metric_warned) {
            warning(sprintf("Selected metric not found; falling back to '%s'.", metric_used), call. = FALSE)
            selection_metric_warned <- TRUE
          }
        } else {
          warning(sprintf("Outer fold %d skipped: no usable metrics found.", i), call. = FALSE)
          skipped_no_results <- skipped_no_results + 1L
          mark_fold_status(i, stage = "metric_resolution", status = "skipped",
                           reason = "no_usable_metrics",
                           notes = "No metric with finite estimates was available for model selection.")
          next
        }
      }
    }

    selection_metric_used <- metric_used
    selected_cfg <- select_config(metrics_raw, metric_name = metric_used, selection_rule = selection)
    if (is.null(selected_cfg) || is.null(selected_cfg$params) || !nrow(selected_cfg$params)) {
      warning(sprintf("Outer fold %d skipped: could not determine best parameters.", i), call. = FALSE)
      mark_fold_status(i, stage = "model_selection", status = "skipped",
                       reason = "selection_failed",
                       notes = sprintf("Failed to select hyperparameters using metric '%s'.", metric_used))
      next
    }
    best_config <- selected_cfg$config
    best_params <- selected_cfg$params

    fold_threshold <- 0.5
    fold_threshold_metric <- NA_real_
    fold_threshold_n <- 0L
    if (isTRUE(tune_threshold) && identical(task, "binomial")) {
      inner_preds <- tryCatch(
        tune::collect_predictions(tune_res),
        error = function(e) NULL
      )
      if (is.null(inner_preds) || !nrow(inner_preds)) {
        warning(sprintf("Outer fold %d: threshold tuning skipped (no inner predictions).", i),
                call. = FALSE)
      } else {
        pred_cfg <- inner_preds[inner_preds$.config == best_config, , drop = FALSE]
        if (!nrow(pred_cfg)) {
          warning(sprintf("Outer fold %d: threshold tuning skipped (selected config has no predictions).", i),
                  call. = FALSE)
        } else {
          pos_col <- paste0(".pred_", make.names(class_levels[2]))
          prob_cols <- setdiff(grep("^\\.pred_", names(pred_cfg), value = TRUE), ".pred_class")
          if (!pos_col %in% prob_cols && length(prob_cols) >= 2L) {
            pos_col <- prob_cols[[2]]
          }
          if (!pos_col %in% names(pred_cfg)) {
            warning(sprintf("Outer fold %d: threshold tuning skipped (positive probability column not found).", i),
                    call. = FALSE)
          } else {
            prob_vals <- as.numeric(pred_cfg[[pos_col]])
            if (".row" %in% names(pred_cfg)) {
              row_idx <- as.integer(pred_cfg$.row)
              valid_rows <- is.finite(row_idx) & row_idx >= 1L & row_idx <= length(y_inner)
              row_idx <- row_idx[valid_rows]
              prob_vals <- prob_vals[valid_rows]
              truth_vals <- factor(y_inner[row_idx], levels = class_levels)
            } else if (length(outcome) == 1L && outcome %in% names(pred_cfg)) {
              truth_vals <- factor(pred_cfg[[outcome]], levels = class_levels)
            } else if ("truth" %in% names(pred_cfg)) {
              truth_vals <- factor(pred_cfg$truth, levels = class_levels)
            } else {
              truth_vals <- factor(character(0), levels = class_levels)
            }
            keep <- is.finite(prob_vals) & !is.na(truth_vals)
            prob_vals <- prob_vals[keep]
            truth_vals <- truth_vals[keep]

            if (length(prob_vals) && has_two_classes(truth_vals)) {
              threshold_pick <- select_threshold(
                truth = truth_vals,
                prob = prob_vals,
                grid = threshold_grid,
                metric = threshold_metric
              )
              fold_threshold <- threshold_pick$threshold
              fold_threshold_metric <- threshold_pick$metric_value
              fold_threshold_n <- length(prob_vals)
            } else {
              warning(sprintf(
                "Outer fold %d: threshold tuning skipped (insufficient inner predictions with both classes).",
                i
              ), call. = FALSE)
            }
          }
        }
      }
      threshold_rows[[length(threshold_rows) + 1L]] <- data.frame(
        fold = i,
        threshold = fold_threshold,
        metric_value = fold_threshold_metric,
        n_inner = fold_threshold_n,
        stringsAsFactors = FALSE
      )
    }

    final_workflow <- tune::finalize_workflow(base_workflow, best_params)
    outer_splits <- make_outer_splits(tr, te)

    # CRITICAL: Force the correct name in fit_resample by passing a named list
    learner_for_fit <- list()
    learner_for_fit[[learner_label]] <- final_workflow

    outer_fit_args <- list(
      x = x,
      outcome = outcome,
      splits = outer_splits,
      preprocess = list(),
      learner = learner_for_fit,
      metrics = tune_metrics_fold,
      positive_class = if (task == "binomial") class_levels[2] else NULL,
      parallel = parallel,
      refit = FALSE,
      seed = seed + i
    )
    if (identical(task, "binomial")) {
      outer_fit_args$classification_threshold <- fold_threshold
    }
    outer_fit <- do.call(fit_resample, outer_fit_args)

    outer_fits[[i]] <- outer_fit
    fold_metrics <- outer_fit@metrics
    fold_metrics$fold <- i
    metrics_rows[[length(metrics_rows) + 1L]] <- fold_metrics
    mark_fold_status(i, stage = "outer_evaluation", status = "success",
                     reason = NA_character_,
                     notes = sprintf("Selected config '%s' using metric '%s'.",
                                     as.character(best_config), as.character(metric_used)))

    best_df <- as.data.frame(best_params)
    best_df$fold <- i
    best_rows[[length(best_rows) + 1L]] <- best_df
  }

  if (!length(metrics_rows)) {
    msg <- "No successful outer folds were completed."
    reasons <- character()
    if (skipped_single_class_outer) {
      reasons <- c(reasons, "Some outer training folds had a single outcome class.")
    }
    if (skipped_single_class_inner) {
      reasons <- c(reasons, "Some inner training folds had a single outcome class.")
    }
    if (skipped_no_results) {
      reasons <- c(reasons, "Inner tuning produced no results for some folds.")
    }
    if (length(reasons)) {
      msg <- paste(msg, paste(reasons, collapse = " "),
                   "Consider `stratify = TRUE` in make_split_plan() or reducing `inner_v`.")
    }
    stop(msg, call. = FALSE)
  }

  # Unify metric row schemas before binding (folds may differ in computed metrics)
  all_cols <- unique(unlist(lapply(metrics_rows, names)))
  metrics_rows <- lapply(metrics_rows, function(row) {
    for (col in setdiff(all_cols, names(row))) row[[col]] <- NA_real_
    row[, all_cols, drop = FALSE]
  })
  metrics_df <- do.call(rbind, metrics_rows)
  best_params_df <- if (length(best_rows)) do.call(rbind, best_rows) else data.frame()
  fold_status_df <- if (length(fold_status_rows)) {
    do.call(rbind, lapply(seq_along(fold_status_rows), function(i) {
      row <- fold_status_rows[[i]]
      if (is.null(row)) {
        data.frame(
          fold = i,
          stage = "outer_loop",
          status = "failed",
          reason = "unknown",
          notes = "Fold ended without a recorded status.",
          stringsAsFactors = FALSE
        )
      } else {
        row
      }
    }))
  } else {
    data.frame(fold = integer(0), stage = character(0), status = character(0),
               reason = character(0), notes = character(0), stringsAsFactors = FALSE)
  }
  thresholds_df <- if (length(threshold_rows)) {
    do.call(rbind, threshold_rows)
  } else {
    data.frame(fold = integer(0), threshold = numeric(0),
               metric_value = numeric(0), n_inner = integer(0), stringsAsFactors = FALSE)
  }
  threshold_stability <- if (nrow(thresholds_df)) {
    list(
      mean = mean(thresholds_df$threshold, na.rm = TRUE),
      sd = stats::sd(thresholds_df$threshold, na.rm = TRUE),
      min = min(thresholds_df$threshold, na.rm = TRUE),
      max = max(thresholds_df$threshold, na.rm = TRUE)
    )
  } else {
    list(mean = NA_real_, sd = NA_real_, min = NA_real_, max = NA_real_)
  }

  metric_cols <- setdiff(names(metrics_df), "fold")
  metric_summary_raw <- aggregate(. ~ learner, data = metrics_df[, metric_cols, drop = FALSE],
                                  FUN = function(x) c(mean = mean(x, na.rm = TRUE),
                                                      sd = sd(x, na.rm = TRUE)),
                                  na.action = stats::na.pass)
  # flatten embedded matrices from aggregate() into separate columns
  metric_summary <- data.frame(learner = metric_summary_raw$learner,
                               stringsAsFactors = FALSE)
  for (col in setdiff(colnames(metric_summary_raw), "learner")) {
    mat <- metric_summary_raw[[col]]
    if (is.matrix(mat)) {
      metric_summary[[paste0(col, "_mean")]] <- mat[, "mean"]
      metric_summary[[paste0(col, "_sd")]] <- mat[, "sd"]
    } else {
      metric_summary[[col]] <- mat
    }
  }

  # confidence intervals -------------------------------------------------------
  tryCatch({
    ci_df <- cv_ci(metrics_df, method = "nadeau_bengio")
    ci_cols <- grep("_ci_lo$|_ci_hi$", names(ci_df), value = TRUE)
    if (length(ci_cols) && nrow(ci_df) == nrow(metric_summary)) {
      for (cc in ci_cols) {
        metric_summary[[cc]] <- ci_df[[cc]]
      }
    }
  }, error = function(e) NULL)

  final_model <- NULL
  final_workflow <- NULL
  final_params <- data.frame()
  refit_fold <- NA_integer_
  refit_metric <- NA_character_
  if (isTRUE(refit) && nrow(best_params_df) > 0) {
    refit_metric <- selection_metric_used

    # Aggregate hyperparameters across all outer folds to avoid nested-CV
    # leakage. The old approach selected the fold with the best outer-fold
    # test metric and used that fold's inner-CV params — a violation because
    # final params become informed by outer test performance.
    #
    # This coordinate-wise aggregation (median for numeric, majority vote
    # for categorical) is leakage-safe but may produce parameter combinations
    # never jointly evaluated in any single inner CV. This is an accepted
    # tradeoff: the aggregated values lack direct validation grounding but
    # preserve the independence of the outer evaluation loop.
    param_cols <- setdiff(names(best_params_df), "fold")
    if (length(param_cols)) {
      agg_list <- lapply(param_cols, function(col) {
        vals <- best_params_df[[col]]
        if (is.numeric(vals)) {
          stats::median(vals, na.rm = TRUE)
        } else {
          # majority vote
          tbl <- table(vals)
          names(tbl)[which.max(tbl)]
        }
      })
      names(agg_list) <- param_cols
      final_params <- as.data.frame(agg_list, stringsAsFactors = FALSE)
      # Restore numeric types where needed
      for (col in param_cols) {
        if (is.numeric(best_params_df[[col]])) {
          final_params[[col]] <- as.numeric(final_params[[col]])
        }
      }
      refit_fold <- NA_integer_
      final_workflow <- tune::finalize_workflow(base_workflow, final_params)
      final_model <- generics::fit(final_workflow, data = df_all)
    } else {
      warning("Refit requested but skipped: no hyperparameter columns found.", call. = FALSE)
    }
  }

  structure(
    list(
      splits = splits,
      metrics = metrics_df,
      metric_summary = metric_summary,
      best_params = best_params_df,
      inner_results = inner_results,
      outer_fits = outer_fits,
      thresholds = thresholds_df,
      fold_status = fold_status_df,
      final_model = final_model,
      final_workflow = final_workflow,
      final_params = final_params,
      info = list(
        task = task,
        metrics_requested = metric_names,
        metrics_used = if (length(metrics_used)) metrics_used else metric_names,
        positive_class = if (task == "binomial") class_levels[2] else NULL,
        selection = selection,
        selection_metric = selection_metric_used,
        selection_metric_requested = selection_metric_requested,
        threshold_tuned = isTRUE(tune_threshold) && identical(task, "binomial"),
        threshold_metric = threshold_metric_label,
        threshold_grid = threshold_grid,
        thresholds = thresholds_df,
        threshold_stability = threshold_stability,
        refit = isTRUE(refit),
        refit_method = if (isTRUE(refit)) "aggregate" else NA_character_,
        refit_metric = refit_metric,
        refit_fold = refit_fold,
        fold_status = fold_status_df,
        grid = grid,
        seed = seed,
        provenance = .bio_capture_provenance()
      )
    ),
    class = "LeakTune"
  )
}

# Delta Leakage Sensitivity Index (ΔLSI) ────────────────────────────────────
# Compares a naive (potentially leaky) CV pipeline against a guarded pipeline
# and quantifies leakage-induced performance inflation.

# ── Internal helpers ──────────────────────────────────────────────────────────

# Resolve which learner to use from a LeakFit's @metrics.
.dlsi_resolve_learner <- function(fit, learner) {
  if (!is.null(learner)) return(as.character(learner[1L]))
  if ("learner" %in% names(fit@metrics)) {
    lrns <- unique(as.character(fit@metrics$learner))
    lrns <- lrns[nzchar(lrns)]
    if (!length(lrns)) return(NULL)
    if (length(lrns) > 1L)
      warning(sprintf(
        "[delta_lsi] Multiple learners in fit@metrics (%s); using '%s'.",
        paste(lrns, collapse = ", "), lrns[1L]
      ))
    return(lrns[1L])
  }
  NULL
}

# Extract per-fold metric values from fit@metrics.
# Returns data.frame(fold, metric).
.dlsi_fold_metrics <- function(fit, metric, learner) {
  mdf <- fit@metrics

  if (!is.null(learner) && "learner" %in% names(mdf))
    mdf <- mdf[as.character(mdf$learner) == learner, , drop = FALSE]

  if (metric %in% names(mdf) && "fold" %in% names(mdf)) {
    out <- mdf[, c("fold", metric), drop = FALSE]
    names(out)[2L] <- "metric"
    return(out)
  }

  # Fallback: recompute fold metrics from @predictions
  .dlsi_fold_metrics_from_preds(fit, metric, learner)
}

# Recompute fold metrics from @predictions when not in @metrics.
.dlsi_fold_metrics_from_preds <- function(fit, metric, learner) {
  task    <- if (length(fit@task)) fit@task else "binomial"
  pred_df <- .dlsi_combine_preds(fit, learner)
  if (is.null(pred_df) || !"fold" %in% names(pred_df))
    stop(sprintf(
      "[delta_lsi] Metric '%s' not in fit@metrics and no usable predictions.",
      metric
    ))
  folds <- sort(unique(pred_df$fold))
  vals  <- vapply(folds, function(f) {
    sub <- pred_df[pred_df$fold == f, , drop = FALSE]
    tryCatch(
      .metric_value(metric, task, sub$truth, sub$pred, pred_df = sub),
      error = function(e) NA_real_
    )
  }, numeric(1L))
  data.frame(fold = folds, metric = vals, stringsAsFactors = FALSE)
}

# Combine all per-fold prediction data frames into one.
.dlsi_combine_preds <- function(fit, learner = NULL) {
  preds <- fit@predictions
  if (!length(preds)) return(NULL)
  pred_df <- tryCatch(
    do.call(rbind, Filter(is.data.frame, preds)),
    error = function(e) NULL
  )
  if (is.null(pred_df) || !nrow(pred_df)) return(NULL)
  if (!is.null(learner) && "learner" %in% names(pred_df)) {
    pred_df <- pred_df[as.character(pred_df$learner) == learner, , drop = FALSE]
  } else if ("learner" %in% names(pred_df)) {
    first_lrn <- unique(as.character(pred_df$learner))[1L]
    pred_df   <- pred_df[as.character(pred_df$learner) == first_lrn, , drop = FALSE]
  }
  pred_df
}

# Get repeat_id for each sequential fold position (1..N).
.dlsi_repeat_ids <- function(fit) {
  vapply(fit@splits@indices, function(z) {
    as.integer(z$repeat_id %||% 1L)
  }, integer(1L))
}

# Get test-set size for each sequential fold position.
# Falls back to predictions-based count when splits use compact form.
.dlsi_fold_sizes <- function(fit, pred_df = NULL) {
  sizes <- vapply(fit@splits@indices, function(z) {
    if (!is.null(z$test)) as.integer(length(z$test)) else NA_integer_
  }, integer(1L))

  if (any(is.na(sizes)) && !is.null(pred_df) && "fold" %in% names(pred_df)) {
    tab      <- table(pred_df$fold)
    fold_seq <- seq_along(fit@splits@indices)
    for (i in which(is.na(sizes))) {
      k <- as.character(fold_seq[i])
      if (k %in% names(tab)) sizes[i] <- as.integer(tab[[k]])
    }
  }
  sizes[is.na(sizes)] <- 1L
  sizes
}

# Check whether two fits share exactly the same fold-level test indices.
# Returns TRUE only when every fold's test set is identical (same observations,
# same order after sorting).  This guards against pairing fits that happen to
# have the same repeat count but were run on different data splits.
.dlsi_splits_match <- function(fit_naive, fit_guarded) {
  idx_n <- fit_naive@splits@indices
  idx_g <- fit_guarded@splits@indices
  if (length(idx_n) != length(idx_g)) return(FALSE)
  all(vapply(seq_along(idx_n), function(i) {
    te_n <- sort(as.integer(idx_n[[i]]$test %||% integer(0)))
    te_g <- sort(as.integer(idx_g[[i]]$test %||% integer(0)))
    identical(te_n, te_g)
  }, logical(1L)))
}

# Huber M-estimator of location (iterative; Huber k = 1.345).
.dlsi_huber <- function(x, k = 1.345, tol = 1e-8, max_iter = 100L) {
  x <- as.numeric(x[is.finite(x)])
  n <- length(x)
  if (!n)    return(NA_real_)
  if (n == 1L) return(x[1L])
  mu    <- stats::median(x)
  sigma <- stats::mad(x, constant = 1.4826)
  if (!is.finite(sigma) || sigma < tol) return(mu)
  for (iter in seq_len(max_iter)) {
    r     <- (x - mu) / sigma
    w     <- ifelse(abs(r) <= k, 1.0, k / abs(r))
    mu_new <- sum(w * x) / sum(w)
    if (abs(mu_new - mu) < tol * (1.0 + abs(mu))) break
    mu <- mu_new
  }
  mu
}

# BCa bootstrap confidence interval (Efron, 1987).
.dlsi_bca_ci <- function(x, statfn, M = 2000L, conf = 0.95, seed = 42L) {
  x <- as.numeric(x[is.finite(x)])
  n <- length(x)
  if (n < 2L) return(c(NA_real_, NA_real_))

  set.seed(seed)
  theta_hat <- statfn(x)
  if (!is.finite(theta_hat)) return(c(NA_real_, NA_real_))

  boot_stats <- replicate(M, statfn(sample(x, n, replace = TRUE)))
  boot_stats <- as.numeric(boot_stats[is.finite(boot_stats)])
  M_eff      <- length(boot_stats)
  if (M_eff < 10L) return(c(NA_real_, NA_real_))

  # Bias correction z0
  prop <- mean(boot_stats < theta_hat, na.rm = TRUE)
  prop <- max(1.0 / M_eff, min(1.0 - 1.0 / M_eff, prop))
  z0   <- stats::qnorm(prop)

  # Acceleration a_hat via jackknife
  jack   <- vapply(seq_len(n),
                   function(i) tryCatch(statfn(x[-i]), error = function(e) NA_real_),
                   numeric(1L))
  jack   <- as.numeric(jack[is.finite(jack)])
  a_hat  <- if (length(jack) >= 2L) {
    j_bar <- mean(jack)
    diff  <- j_bar - jack
    denom <- 6.0 * sum(diff^2)^1.5
    if (is.finite(denom) && denom > 1e-12) sum(diff^3) / denom else 0.0
  } else 0.0

  # BCa adjusted quantiles
  alpha    <- 1.0 - conf
  adj_p    <- function(z_a) {
    denom <- 1.0 - a_hat * (z0 + z_a)
    if (abs(denom) < 1e-10) return(alpha / 2.0)
    stats::pnorm(z0 + (z0 + z_a) / denom)
  }
  p_lo <- adj_p(stats::qnorm(alpha / 2.0))
  p_hi <- adj_p(stats::qnorm(1.0 - alpha / 2.0))
  eps  <- 1.0 / M_eff
  p_lo <- max(eps, min(1.0 - eps, p_lo))
  p_hi <- max(eps, min(1.0 - eps, p_hi))

  unname(stats::quantile(boot_stats, c(p_lo, p_hi), na.rm = TRUE))
}

# Sign-flip randomization test (Phipson & Smyth, 2010).
# H0: mean(delta_r) = 0 (no leakage inflation).
# Test statistic: arithmetic mean (well-defined under sign-flip null).
#
# Exact enumeration (R <= 15): all 2^R sign combinations are evaluated.
#   p = count(|null| >= |obs|) / 2^R  — no continuity correction.
#
# Monte Carlo (R > 15): Phipson & Smyth (2010) correction:
#   p = (count + 1) / (M + 1)  — guards against p = 0 from sampling.
.dlsi_sign_flip <- function(delta_r, M_flip = 10000L, seed = 42L) {
  delta_r  <- as.numeric(delta_r[is.finite(delta_r)])
  R        <- length(delta_r)
  if (R < 2L) return(NA_real_)

  obs_stat <- mean(delta_r)
  if (!is.finite(obs_stat)) return(NA_real_)

  set.seed(seed)

  if (R <= 15L) {
    # ── Exact enumeration ──────────────────────────────────────────────────
    n_combos   <- 2L^R
    null_stats <- vapply(seq_len(n_combos), function(j) {
      bits  <- as.integer(intToBits(j - 1L)[seq_len(R)])
      signs <- ifelse(bits == 1L, 1.0, -1.0)
      mean(signs * delta_r)
    }, numeric(1L))
    null_stats <- as.numeric(null_stats[is.finite(null_stats)])
    if (!length(null_stats)) return(NA_real_)
    # Exact p-value: proportion of null stats at least as extreme as observed.
    # No continuity correction — the null distribution is fully enumerated.
    sum(abs(null_stats) >= abs(obs_stat) - .Machine$double.eps * 100) /
      length(null_stats)
  } else {
    # ── Monte Carlo sign-flip ──────────────────────────────────────────────
    null_stats <- replicate(M_flip, {
      signs <- sample(c(-1.0, 1.0), R, replace = TRUE)
      mean(signs * delta_r)
    })
    null_stats <- as.numeric(null_stats[is.finite(null_stats)])
    if (!length(null_stats)) return(NA_real_)
    # Phipson & Smyth (2010) +1 correction prevents p = 0 from Monte Carlo.
    (sum(abs(null_stats) >= abs(obs_stat) - .Machine$double.eps * 100) + 1L) /
      (length(null_stats) + 1L)
  }
}

# Auto-estimate block size from AR(1) autocorrelation of Δ_r for blocked_time
# sign-flip.  With small R (5-20 values) the AR(1) estimate is noisy; the result
# is capped at floor(R/3) to guarantee at least three independent blocks.
.dlsi_resolve_block_size <- function(delta_r, block_size = NULL) {
  R <- length(delta_r)
  if (!is.null(block_size)) return(max(1L, as.integer(block_size)))
  if (R < 4L) return(1L)
  rho1 <- tryCatch(
    stats::cor(delta_r[-R], delta_r[-1L]),
    error = function(e) 0.0
  )
  rho1 <- if (is.finite(rho1)) max(0.0, rho1) else 0.0
  # Effective block size from AR(1) rate: b ≈ 1/(1-rho1)
  bs <- if (rho1 >= 0.99) R else max(1L, ceiling(1.0 / (1.0 - rho1)))
  # Cap so there are at least 3 blocks; floor(R/3) is the upper limit
  as.integer(min(bs, max(1L, floor(R / 3L))))
}

# Block sign-flip randomization test for temporally structured Δ_r.
# Flips contiguous blocks of repeats together to preserve serial autocorrelation
# under the null.  Returns a list(p_value, n_blocks, block_size_used).
#
# Exact enumeration (n_blocks <= 15): all 2^n_blocks block-sign vectors.
#   p = count(|null| >= |obs|) / 2^n_blocks  — no continuity correction.
#
# Monte Carlo (n_blocks > 15): Phipson & Smyth (2010) correction:
#   p = (count + 1) / (M + 1).
.dlsi_sign_flip_blocked <- function(delta_r, block_size, M_flip = 10000L,
                                    seed = 42L) {
  delta_r    <- as.numeric(delta_r[is.finite(delta_r)])
  R          <- length(delta_r)
  block_size <- max(1L, as.integer(block_size))

  if (R < 2L) return(list(p_value = NA_real_, n_blocks = 0L,
                           block_size_used = block_size))

  block_id <- ceiling(seq_len(R) / block_size)
  n_blocks <- max(block_id)

  # Single block: every sign vector flips all Δ_r simultaneously; p is always 1.
  if (n_blocks < 2L)
    return(list(p_value = 1.0, n_blocks = n_blocks,
                block_size_used = block_size))

  obs_stat <- mean(delta_r)
  if (!is.finite(obs_stat))
    return(list(p_value = NA_real_, n_blocks = n_blocks,
                block_size_used = block_size))

  set.seed(seed)

  if (n_blocks <= 15L) {
    # ── Exact enumeration over 2^n_blocks block-sign vectors ─────────────────
    combos     <- as.matrix(expand.grid(rep(list(c(-1.0, 1.0)), n_blocks)))
    null_stats <- apply(combos, 1L, function(bsigns) {
      mean(bsigns[block_id] * delta_r)
    })
    null_stats <- as.numeric(null_stats[is.finite(null_stats)])
    if (!length(null_stats))
      return(list(p_value = NA_real_, n_blocks = n_blocks,
                  block_size_used = block_size))
    p_val <- sum(abs(null_stats) >= abs(obs_stat) - .Machine$double.eps * 100) /
      length(null_stats)
  } else {
    # ── Monte Carlo with Phipson & Smyth (2010) correction ───────────────────
    null_stats <- replicate(M_flip, {
      bsigns <- sample(c(-1.0, 1.0), n_blocks, replace = TRUE)
      mean(bsigns[block_id] * delta_r)
    })
    null_stats <- as.numeric(null_stats[is.finite(null_stats)])
    if (!length(null_stats))
      return(list(p_value = NA_real_, n_blocks = n_blocks,
                  block_size_used = block_size))
    p_val <- (sum(abs(null_stats) >= abs(obs_stat) - .Machine$double.eps * 100) +
                1L) / (length(null_stats) + 1L)
  }

  list(p_value = p_val, n_blocks = n_blocks, block_size_used = block_size)
}

# ── Main function ─────────────────────────────────────────────────────────────

#' Delta Leakage Sensitivity Index (Delta LSI)
#'
#' Compares a naive (potentially leaky) cross-validation pipeline against a
#' guarded (leakage-protected) pipeline and quantifies leakage-induced
#' performance inflation using the Leakage Sensitivity Index (LSI).
#'
#' @details
#' \subsection{Method}{
#' For each fit, per-fold metric values are extracted from \code{fit@@metrics}
#' (or recomputed from \code{fit@@predictions} if necessary). Fold test-set
#' sizes are used as weights to aggregate fold metrics into per-repeat
#' estimates \eqn{\mu_r}. The repeat-level delta
#' \eqn{\Delta_r = s \cdot (\mu_r^{\text{naive}} - \mu_r^{\text{guarded}})}
#' captures leakage-induced performance inflation for each CV repeat, where
#' \eqn{s = +1} for higher-is-better metrics (e.g., AUC) and \eqn{s = -1}
#' for lower-is-better metrics (e.g., RMSE), so that \eqn{\Delta_r > 0}
#' always indicates the naive pipeline is more optimistic than the guarded one.
#'
#' The \strong{delta_lsi} point estimate is the Huber M-estimator (k = 1.345)
#' applied to \eqn{\{\Delta_r\}}, which is robust to occasional outlier
#' repeats. \strong{delta_metric} is the arithmetic mean of \eqn{\{\Delta_r\}}
#' for easy interpretation in the original metric's units.
#'
#' Pairing requires that \code{fit_leaky} and \code{fit_guarded} share
#' \emph{identical fold structures} (same test-set membership per fold) in
#' addition to the same number of repeats.  When repeat counts match but fold
#' structures differ, a warning is issued and the fits are treated as unpaired.
#'
#' When \eqn{R_{\text{eff}} \geq 5} (equal, paired repeats), a sign-flip
#' randomization test (Phipson & Smyth, 2010) is performed: under
#' \eqn{H_0} (no leakage) the sign of each \eqn{\Delta_r} is exchangeable.
#' All \eqn{2^R} sign combinations are enumerated exactly for
#' \eqn{R \leq 15} (no continuity correction); Monte Carlo sampling is used
#' for larger \eqn{R} with the Phipson & Smyth (2010) correction.
#'
#' BCa bootstrap confidence intervals (Efron, 1987) require
#' \eqn{R_{\text{eff}} \geq 10}.
#' }
#'
#' \subsection{Inference tiers}{
#' \describe{
#'   \item{\code{"A_full_inference"}}{R_eff >= 20: point + BCa CI + sign-flip p-value; \code{inference_ok = TRUE}}
#'   \item{\code{"B_signflip_ci"}}{10 <= R_eff < 20: point + sign-flip p-value + BCa CI}
#'   \item{\code{"C_signflip"}}{5 <= R_eff < 10: point + sign-flip p-value (no CI)}
#'   \item{\code{"D_insufficient"}}{R_eff < 5 or unpaired: point estimate only}
#' }}
#'
#' @param fit_leaky A \code{\linkS4class{LeakFit}} object from the leaky
#'   (unprotected) evaluation pipeline.
#' @param fit_guarded A \code{\linkS4class{LeakFit}} object from the guarded
#'   (leakage-protected) evaluation pipeline.
#' @param metric Character. Performance metric to compare. Must appear in
#'   \code{fit@@metrics} of both fits (e.g., \code{"auc"}, \code{"rmse"}).
#' @param exchangeability Character. Exchangeability assumption for the
#'   sign-flip test. One of \code{"iid"} (default), \code{"by_group"},
#'   \code{"within_batch"}, \code{"blocked_time"}.
#'   \code{"blocked_time"} activates a block sign-flip procedure that flips
#'   contiguous groups of repeats together, preserving serial autocorrelation
#'   under the null; see \code{block_size}.  \code{"by_group"} and
#'   \code{"within_batch"} are stored and reported but inference still uses the
#'   iid sign-flip procedure (a warning is issued; contributions welcome).
#'   \code{"iid"} (default) applies the standard independent sign-flip test.
#' @param block_size Integer or \code{NULL}. Block length for the block
#'   sign-flip test, used only when \code{exchangeability = "blocked_time"}.
#'   When \code{NULL} (default), the block size is auto-estimated from the
#'   first-order autocorrelation of \eqn{\{\Delta_r\}} and capped at
#'   \code{floor(R/3)} to ensure at least three independent blocks.  A warning
#'   is issued when the estimate is used with \code{R_eff < 20} because the
#'   AR(1) estimate is noisy at small sample sizes.  Provide an explicit integer
#'   to override auto-estimation.
#' @param learner Optional character. Learner name to select from multi-learner
#'   fits. If \code{NULL}, the first learner found in \code{fit@@metrics} is used.
#' @param higher_is_better Logical or \code{NULL}. Whether a higher value of
#'   \code{metric} indicates better performance. When \code{NULL} (default),
#'   auto-detected from the metric name: \code{"rmse"}, \code{"mse"},
#'   \code{"mae"}, \code{"log_loss"}, \code{"brier"}, \code{"error"},
#'   \code{"loss"}, and \code{"deviance"} are treated as lower-is-better;
#'   all others default to higher-is-better. Setting this correctly ensures
#'   that a positive \code{delta_lsi} always indicates leakage inflation
#'   (the naive pipeline is artificially more optimistic than the guarded one).
#' @param M_boot Integer. Number of bootstrap samples for BCa CI (default 2000).
#' @param M_flip Integer. Maximum Monte Carlo samples for sign-flip test when
#'   R_eff > 15 (default 10000).
#' @param strict Logical. If \code{TRUE}, error on insufficient R_eff instead
#'   of a warning.
#' @param return_details Logical. If \code{TRUE}, include the per-repeat
#'   \eqn{\Delta_r} vector and the original fit objects in the \code{info} slot.
#' @param seed Integer. Random seed for bootstrap and sign-flip test.
#' @param ... Unused. Reserved for deprecated aliases such as
#'   \code{fit_naive}.
#'
#' @return A \code{\linkS4class{LeakDeltaLSI}} object.
#'
#' @seealso \code{\link{audit_leakage}}, \code{\link{fit_resample}}
#' @export
delta_lsi <- function(
    fit_leaky,
    fit_guarded,
    metric          = "auc",
    exchangeability = c("iid", "by_group", "within_batch", "blocked_time"),
    learner         = NULL,
    higher_is_better = NULL,
    block_size      = NULL,
    M_boot          = 2000L,
    M_flip          = 10000L,
    strict          = FALSE,
    return_details  = FALSE,
    seed            = 42L,
    ...
) {
  extras <- list(...)
  if ("fit_naive" %in% names(extras)) {
    if (!missing(fit_leaky)) {
      stop("Use either fit_leaky or the deprecated fit_naive argument, not both.")
    }
    fit_leaky <- extras$fit_naive
    extras$fit_naive <- NULL
    warning("delta_lsi(): fit_naive is deprecated; use fit_leaky instead.")
  }
  if (length(extras)) {
    stop(
      sprintf(
        "Unused argument(s): %s",
        paste(names(extras), collapse = ", ")
      )
    )
  }

  exchangeability <- match.arg(exchangeability)

  if (!inherits(fit_leaky,   "LeakFit")) stop("fit_leaky must be a LeakFit object.")
  if (!inherits(fit_guarded, "LeakFit")) stop("fit_guarded must be a LeakFit object.")

  # ── Resolve metric directionality ────────────────────────────────────────────
  if (is.null(higher_is_better)) {
    lower_is_better <- grepl(
      "rmse|mse|mae|log_loss|logloss|brier|error|loss|deviance",
      tolower(metric)
    )
    higher_is_better <- !lower_is_better
  }
  # sign_factor > 0 means "naive - guarded > 0 is leakage"; flip for lower-is-better
  sign_factor <- if (isTRUE(higher_is_better)) 1.0 else -1.0

  # Resolve learners (one per fit; may differ)
  lrn_n <- .dlsi_resolve_learner(fit_leaky,   learner)
  lrn_g <- .dlsi_resolve_learner(fit_guarded, learner)

  # Per-fold metric data frames: columns fold, metric
  fm_n <- .dlsi_fold_metrics(fit_leaky,   metric, lrn_n)
  fm_g <- .dlsi_fold_metrics(fit_guarded, metric, lrn_g)

  # Sequential fold positions (1..N) for each fit
  fold_seq_n <- seq_along(fit_leaky@splits@indices)
  fold_seq_g <- seq_along(fit_guarded@splits@indices)

  # Repeat IDs and fold sizes
  rep_ids_n <- .dlsi_repeat_ids(fit_leaky)
  rep_ids_g <- .dlsi_repeat_ids(fit_guarded)

  pred_n <- .dlsi_combine_preds(fit_leaky,   lrn_n)
  pred_g <- .dlsi_combine_preds(fit_guarded, lrn_g)

  sizes_n <- .dlsi_fold_sizes(fit_leaky,   pred_n)
  sizes_g <- .dlsi_fold_sizes(fit_guarded, pred_g)

  # Build fold-level data frames
  .build_folds <- function(fm, rep_ids, sizes, fold_seq) {
    idx <- match(fm$fold, fold_seq)
    data.frame(
      fold      = fm$fold,
      metric    = fm$metric,
      repeat_id = rep_ids[idx],
      n         = sizes[idx],
      stringsAsFactors = FALSE
    )
  }

  folds_n <- .build_folds(fm_n, rep_ids_n, sizes_n, fold_seq_n)
  folds_g <- .build_folds(fm_g, rep_ids_g, sizes_g, fold_seq_g)
  folds_n$n[is.na(folds_n$n)] <- 1L
  folds_g$n[is.na(folds_g$n)] <- 1L

  # Aggregate folds → per-repeat size-weighted mean metric
  .agg_repeats <- function(folds_df) {
    reps <- sort(unique(folds_df$repeat_id[is.finite(folds_df$metric)]))
    rows <- lapply(reps, function(r) {
      sub <- folds_df[folds_df$repeat_id == r & is.finite(folds_df$metric), ]
      if (!nrow(sub)) return(NULL)
      w  <- pmax(as.numeric(sub$n), 1.0)
      mu <- sum(w * sub$metric) / sum(w)
      data.frame(
        repeat_id = r, metric = mu,
        n_folds = nrow(sub), total_n = sum(sub$n),
        stringsAsFactors = FALSE
      )
    })
    out <- do.call(rbind, Filter(Negate(is.null), rows))
    if (is.null(out)) {
      out <- data.frame(repeat_id = integer(0), metric = numeric(0),
                        n_folds = integer(0), total_n = integer(0),
                        stringsAsFactors = FALSE)
    }
    out
  }

  reps_n <- .agg_repeats(folds_n)
  reps_g <- .agg_repeats(folds_g)

  R_naive   <- nrow(reps_n)
  R_guarded <- nrow(reps_g)

  # ── Pairing: equal count AND matching fold structures ─────────────────────
  same_count <- (R_naive > 0L && R_guarded > 0L && R_naive == R_guarded)
  if (same_count) {
    splits_ok <- .dlsi_splits_match(fit_leaky, fit_guarded)
    if (!splits_ok) {
      warning(sprintf(
        paste0("[delta_lsi] fit_leaky and fit_guarded have R=%d repeats each ",
               "but different fold structures; treating as unpaired."),
        R_naive
      ))
    }
    paired <- splits_ok
    if (paired) {
      # Normalize guarded repeat_ids to match naive's positional mapping.
      # .dlsi_splits_match() verified that sequential position i has identical
      # test members in both fits, so naive's repeat_id labels are canonical.
      rid_lookup <- setNames(folds_n$repeat_id, folds_n$fold)
      folds_g$repeat_id <- unname(rid_lookup[as.character(folds_g$fold)])
      reps_g <- .agg_repeats(folds_g)
    }
  } else {
    paired <- FALSE
  }
  R_eff <- if (paired) as.integer(R_naive) else 0L

  # Per-repeat Δ values — sign_factor applied so positive = leakage inflation.
  # Align by repeat_id (both are sorted by .agg_repeats; guard against rare
  # case where different repeats yield all-NA metrics in each fit).
  delta_r <- if (paired) {
    if (identical(reps_n$repeat_id, reps_g$repeat_id)) {
      sign_factor * (reps_n$metric - reps_g$metric)
    } else {
      common <- intersect(reps_n$repeat_id, reps_g$repeat_id)
      rn <- reps_n[match(common, reps_n$repeat_id), ]
      rg <- reps_g[match(common, reps_g$repeat_id), ]
      sign_factor * (rn$metric - rg$metric)
    }
  } else NULL

  # Update R_eff if intersection reduced pairing
  if (!is.null(delta_r) && length(delta_r) < R_eff) {
    warning(sprintf(
      "[delta_lsi] %d of %d repeats dropped (all-NA metrics); R_eff reduced from %d to %d.",
      R_eff - length(delta_r), R_eff, R_eff, length(delta_r)
    ))
    R_eff <- length(delta_r)
  }

  # Inference tier (computed after R_eff adjustment)
  tier <- if (R_eff >= 20L) "A_full_inference" else
          if (R_eff >= 10L) "B_signflip_ci"    else
          if (R_eff >=  5L) "C_signflip"       else
                            "D_insufficient"

  if (tier == "D_insufficient") {
    msg <- sprintf(
      paste0("[delta_lsi] R_eff = %d (R_leaky=%d, R_guarded=%d, paired=%s). ",
             "Need paired R_eff >= 5 for sign-flip inference; reporting point estimate only."),
      R_eff, R_naive, R_guarded, paired
    )
    if (isTRUE(strict)) stop(msg) else warning(msg)
  }

  # Point estimates
  delta_lsi_est    <- if (!is.null(delta_r)) {
    .dlsi_huber(delta_r)
  } else {
    sign_factor * (.dlsi_huber(reps_n$metric) - .dlsi_huber(reps_g$metric))
  }
  delta_metric_est <- if (!is.null(delta_r)) {
    mean(delta_r, na.rm = TRUE)
  } else {
    sign_factor * (mean(reps_n$metric, na.rm = TRUE) - mean(reps_g$metric, na.rm = TRUE))
  }

  # BCa CI (requires R_eff >= 10 and paired)
  ci_lsi <- c(NA_real_, NA_real_)
  ci_dm  <- c(NA_real_, NA_real_)
  if (!is.null(delta_r) && R_eff >= 10L) {
    ci_lsi <- tryCatch(
      .dlsi_bca_ci(delta_r, .dlsi_huber,
                   M = M_boot, seed = seed),
      error = function(e) c(NA_real_, NA_real_)
    )
    ci_dm  <- tryCatch(
      .dlsi_bca_ci(delta_r, function(x) mean(x, na.rm = TRUE),
                   M = M_boot, seed = seed + 1L),
      error = function(e) c(NA_real_, NA_real_)
    )
  }

  # Sign-flip test (requires R_eff >= 5 and paired).
  # Routing depends on exchangeability:
  #   "blocked_time"       -> block sign-flip (contiguous blocks of Δ_r)
  #   "by_group" / "within_batch" -> iid sign-flip with an explanatory warning
  #   "iid"                -> standard independent sign-flip
  p_val            <- NA_real_
  block_size_used  <- NA_integer_
  n_blocks_used    <- NA_integer_

  if (!is.null(delta_r) && R_eff >= 5L) {
    if (exchangeability == "blocked_time") {
      bs     <- .dlsi_resolve_block_size(delta_r, block_size)
      if (is.null(block_size) && R_eff < 20L)
        warning(sprintf(
          paste0("[delta_lsi] exchangeability='blocked_time': block_size ",
                 "auto-estimated as %d from AR(1) of \u0394r (R_eff=%d; ",
                 "estimate is noisy at small R). Supply block_size explicitly ",
                 "to override."),
          bs, R_eff
        ))
      sf_out         <- .dlsi_sign_flip_blocked(delta_r, block_size = bs,
                                                M_flip = M_flip,
                                                seed = seed + 2L)
      p_val          <- sf_out$p_value
      block_size_used <- as.integer(sf_out$block_size_used)
      n_blocks_used   <- as.integer(sf_out$n_blocks)
      # If too few independent blocks for p < 0.05, set p to NA and warn.
      if (!is.na(n_blocks_used) && n_blocks_used < 5L) {
        min_p <- if (n_blocks_used >= 1L) 1.0 / 2.0^n_blocks_used else 1.0
        warning(sprintf(
          paste0("[delta_lsi] blocked_time sign-flip: n_blocks = %d < 5; ",
                 "minimum achievable p = %.4f > 0.05. Setting p_value = NA. ",
                 "Increase R_eff or reduce block_size."),
          n_blocks_used, min_p
        ))
        p_val <- NA_real_
      }
    } else {
      if (exchangeability %in% c("by_group", "within_batch"))
        warning(sprintf(
          paste0("[delta_lsi] exchangeability = '%s' is stored but the ",
                 "sign-flip test still assumes iid repeat structure. ",
                 "Use 'blocked_time' for temporal autocorrelation or 'iid' ",
                 "to suppress this warning."),
          exchangeability
        ))
      p_val <- tryCatch(
        .dlsi_sign_flip(delta_r, M_flip = M_flip, seed = seed + 2L),
        error = function(e) NA_real_
      )
    }
  }

  inference_ok <- (R_eff >= 20L && is.finite(p_val) && all(is.finite(ci_lsi)))

  # Metadata
  info_list <- list(
    paired           = paired,
    R_naive          = R_naive,
    R_guarded        = R_guarded,
    higher_is_better = higher_is_better,
    metric_naive     = .dlsi_huber(reps_n$metric),
    metric_guarded   = .dlsi_huber(reps_g$metric),
    block_size_used  = block_size_used,
    n_blocks         = n_blocks_used
  )
  if (return_details) {
    info_list$delta_r     <- delta_r
    info_list$fit_leaky   <- fit_leaky
    info_list$fit_naive   <- fit_leaky
    info_list$fit_guarded <- fit_guarded
  }

  methods::new("LeakDeltaLSI",
    metric          = metric,
    exchangeability = exchangeability,
    tier            = tier,
    strict          = strict,
    R_eff           = R_eff,
    delta_lsi       = delta_lsi_est,
    delta_lsi_ci    = as.numeric(ci_lsi),
    delta_metric    = delta_metric_est,
    delta_metric_ci = as.numeric(ci_dm),
    p_value         = p_val,
    inference_ok    = inference_ok,
    folds_naive     = folds_n,
    folds_guarded   = folds_g,
    repeats_naive   = reps_n,
    repeats_guarded = reps_g,
    info            = info_list
  )
}

# ── S4 show method ─────────────────────────────────────────────────────────────

#' @rdname LeakClasses
#' @param object A \code{\linkS4class{LeakDeltaLSI}} object.
setMethod("show", "LeakDeltaLSI", function(object) {
  hib <- object@info[["higher_is_better"]]
  hib_str <- if (is.null(hib)) "" else
             if (isTRUE(hib)) " [higher=better]" else " [lower=better]"
  cat("A LeakDeltaLSI object\n")
  cat(sprintf("  Metric:         %s%s\n", object@metric, hib_str))
  cat(sprintf("  Tier:           %s  (R_eff = %d)\n", object@tier, object@R_eff))
  fmt <- function(x) if (is.finite(x)) sprintf("%.4f", x) else "NA"
  cat(sprintf("  delta_metric:   %s", fmt(object@delta_metric)))
  if (all(is.finite(object@delta_metric_ci)))
    cat(sprintf("  [%.4f, %.4f] 95%% CI", object@delta_metric_ci[1L], object@delta_metric_ci[2L]))
  cat("\n")
  cat(sprintf("  delta_lsi:      %s", fmt(object@delta_lsi)))
  if (all(is.finite(object@delta_lsi_ci)))
    cat(sprintf("  [%.4f, %.4f] 95%% CI", object@delta_lsi_ci[1L], object@delta_lsi_ci[2L]))
  cat("\n")
  if (is.finite(object@p_value))
    cat(sprintf("  p-value:        %.4f  (sign-flip; tests mean \u0394r)\n",
                object@p_value))
  invisible(object)
})

# ── summary method ─────────────────────────────────────────────────────────────

#' Summarize a LeakDeltaLSI object
#'
#' Prints a human-readable summary of the Delta LSI analysis comparing leaky vs
#' guarded evaluation pipelines.
#'
#' @param object A \code{LeakDeltaLSI} object from \code{\link{delta_lsi}}.
#' @param digits Integer. Number of decimal places to show (default 3).
#' @param ... Unused.
#' @return Invisibly returns \code{object}.
#' @export
summary.LeakDeltaLSI <- function(object, digits = 3L, ...) {
  fmt <- function(x) if (is.finite(x)) formatC(x, digits = digits, format = "f") else "NA"
  hib <- object@info[["higher_is_better"]] %||% TRUE

  cat("\n=====================================\n")
  cat(" bioLeak Delta LSI Summary\n")
  cat("=====================================\n\n")
  cat(sprintf("Metric:            %s\n", object@metric))
  exch <- object@exchangeability
  cat(sprintf("Exchangeability:   %s\n", exch))
  if (identical(exch, "blocked_time")) {
    bs   <- object@info$block_size_used
    nblk <- object@info$n_blocks
    if (!is.null(bs) && !is.na(bs))
      cat(sprintf("  Block sign-flip: block_size = %d, n_blocks = %d\n", bs, nblk))
    else
      cat("  Block sign-flip: not computed (R_eff < 5)\n")
  } else if (exch %in% c("by_group", "within_batch")) {
    cat(sprintf(
      "  [Note: '%s' is stored; inference uses iid sign-flip.]\n", exch
    ))
  }
  cat(sprintf("Inference tier:    %s\n", object@tier))
  cat(sprintf("R_eff:             %d  (R_leaky=%s, R_guarded=%s, paired=%s)\n",
              object@R_eff,
              object@info[["R_naive"]]   %||% "?",
              object@info[["R_guarded"]] %||% "?",
              object@info[["paired"]]    %||% "?"))

  cat("\nLeaky pipeline:    ", fmt(object@info[["metric_naive"]]   %||% NA_real_), "\n", sep = "")
  cat("Guarded pipeline:  ", fmt(object@info[["metric_guarded"]] %||% NA_real_), "\n", sep = "")

  direction <- if (isTRUE(hib)) "leaky - guarded" else "-(leaky - guarded)"
  cat(sprintf("\nPoint estimates (%s; positive = leakage inflation):\n", direction))
  cat(sprintf("  delta_metric:  %s  (raw metric difference)\n", fmt(object@delta_metric)))
  ci_dm <- object@delta_metric_ci
  if (all(is.finite(ci_dm)))
    cat(sprintf("  95%% BCa CI:    [%s, %s]\n", fmt(ci_dm[1L]), fmt(ci_dm[2L])))
  cat(sprintf("  delta_lsi:     %s  (Huber-robust)\n", fmt(object@delta_lsi)))
  ci_lsi <- object@delta_lsi_ci
  if (all(is.finite(ci_lsi)))
    cat(sprintf("  95%% BCa CI:    [%s, %s]\n", fmt(ci_lsi[1L]), fmt(ci_lsi[2L])))

  cat("\nHypothesis test  [H0: no systematic inflation; paired repeat signs exchangeable]:\n")
  if (is.finite(object@p_value)) {
    p   <- object@p_value
    sig <- if (p < 0.001) "***" else if (p < 0.01) "**" else if (p < 0.05) "*" else ""
    cat(sprintf("  Sign-flip p:   %.4f %s\n", p, sig))
    # Diagnostic: warn when p-value and BCa CI lead to different conclusions.
    # This happens when the arithmetic mean (tested by p) and the Huber estimate
    # (covered by the CI) diverge — a sign of outlier repeats skewing the mean.
    if (all(is.finite(object@delta_lsi_ci))) {
      pval_sig    <- p < 0.05
      ci_exc_zero <- (object@delta_lsi_ci[1L] > 0 || object@delta_lsi_ci[2L] < 0)
      if (pval_sig != ci_exc_zero)
        cat(paste0("  [Note: p-value and BCa CI disagree. The p-value tests delta_metric\n",
                   "   (mean); the CI covers delta_lsi (Huber robust). Outlier repeats\n",
                   "   may be pulling the mean. Prefer delta_lsi for point estimates.]\n"))
    }
  } else {
    reason <- if (identical(object@exchangeability, "blocked_time") &&
                  !is.null(object@info$n_blocks) &&
                  !is.na(object@info$n_blocks) &&
                  object@info$n_blocks < 5L)
      sprintf("blocked_time: n_blocks = %d < 5", object@info$n_blocks)
    else
      "requires paired R_eff >= 5"
    cat(sprintf("  Sign-flip p:   NA  (%s)\n", reason))
  }

  cat(sprintf("\nInference valid:   %s\n",
              if (object@inference_ok) "YES (tier A)" else
                sprintf("NO (%s; need paired R_eff >= 20)", object@tier)))
  cat("\n")
  invisible(object)
}

# tests/testthat/test-delta-lsi.R
# Unit tests for delta_lsi() and its internal helpers.

# ── Helper: build minimal LeakFit objects ─────────────────────────────────────

# Build a LeakSplits with explicit test/train indices.
.make_dlsi_splits <- function(n_folds, n_obs, n_repeats = 1L) {
  fold_size <- floor(n_obs / n_folds)
  indices   <- lapply(seq_len(n_folds * n_repeats), function(i) {
    fold_num <- ((i - 1L) %% n_folds) + 1L
    rep_num  <- ceiling(i / n_folds)
    te       <- ((fold_num - 1L) * fold_size + 1L):(fold_num * fold_size)
    tr       <- setdiff(seq_len(n_obs), te)
    list(fold = fold_num, repeat_id = rep_num, test = te, train = tr)
  })
  methods::new("LeakSplits",
    mode    = "subject_grouped",
    indices = indices,
    info    = list()
  )
}

# Build a minimal LeakFit with controlled per-fold AUC values.
.make_dlsi_fit <- function(fold_aucs, n_obs = 30L, n_repeats = 1L,
                           seed = 1L) {
  n_folds <- length(fold_aucs) %/% n_repeats
  splits  <- .make_dlsi_splits(n_folds, n_obs, n_repeats)

  metrics_df <- data.frame(
    fold    = seq_along(fold_aucs),
    learner = "glm",
    auc     = as.numeric(fold_aucs),
    stringsAsFactors = FALSE
  )

  set.seed(seed)
  fold_size <- floor(n_obs / n_folds)
  preds <- lapply(seq_along(fold_aucs), function(i) {
    te <- splits@indices[[i]]$test
    n  <- length(te)
    # Predictions with spread proportional to the AUC value
    y  <- rbinom(n, 1L, 0.5)
    p  <- pmin(1, pmax(0, rnorm(n, mean = 0.5, sd = fold_aucs[i] / 2)))
    data.frame(
      id = te, fold = i, learner = "glm",
      truth = y, pred = p,
      stringsAsFactors = FALSE
    )
  })

  methods::new("LeakFit",
    splits        = splits,
    metrics       = metrics_df,
    metric_summary = data.frame(),
    audit         = data.frame(),
    predictions   = preds,
    preprocess    = list(),
    learners      = list(),
    outcome       = "y",
    task          = "binomial",
    feature_names = character(0L),
    info          = list()
  )
}

# ── Basic construction ─────────────────────────────────────────────────────────

test_that("delta_lsi returns a LeakDeltaLSI object", {
  set.seed(1)
  fit <- .make_dlsi_fit(rep(0.65, 6L), n_repeats = 2L)
  res <- suppressWarnings(delta_lsi(fit, fit, metric = "auc", seed = 1L))
  expect_s4_class(res, "LeakDeltaLSI")
})

test_that("delta_lsi with same fit: delta_metric and delta_lsi are zero", {
  set.seed(1)
  # 6 folds over 2 repeats → R_eff = 2 (D_insufficient tier but point estimate fine)
  fit <- .make_dlsi_fit(rep(0.7, 6L), n_repeats = 2L)
  res <- suppressWarnings(delta_lsi(fit, fit, metric = "auc", seed = 1L))
  expect_equal(res@delta_metric, 0.0, tolerance = 1e-10)
  expect_equal(res@delta_lsi,    0.0, tolerance = 1e-10)
})

test_that("delta_lsi detects inflated naive pipeline", {
  set.seed(2)
  # Naive: AUC ~ 0.85 per fold; guarded: AUC ~ 0.60
  fit_n <- .make_dlsi_fit(rep(0.85, 10L), n_repeats = 2L)
  fit_g <- .make_dlsi_fit(rep(0.60, 10L), n_repeats = 2L)
  res   <- suppressWarnings(delta_lsi(fit_n, fit_g, metric = "auc", seed = 2L))
  expect_gt(res@delta_lsi,    0.0)
  expect_gt(res@delta_metric, 0.0)
})

# ── Tier assignment ────────────────────────────────────────────────────────────

test_that("tier D_insufficient for R_eff < 5", {
  fit_n <- .make_dlsi_fit(rep(0.75, 3L), n_repeats = 1L)
  fit_g <- .make_dlsi_fit(rep(0.60, 3L), n_repeats = 1L)
  expect_warning(
    res <- delta_lsi(fit_n, fit_g, metric = "auc"),
    regexp = "R_eff"
  )
  expect_equal(res@tier, "D_insufficient")
  expect_equal(res@R_eff, 1L)  # paired R=1 → R_eff=1, but below the 5-repeat threshold
})

test_that("tier C_signflip for R_eff in 5..9", {
  set.seed(3)
  # 5 repeats × 3 folds = 15 folds each
  fit_n <- .make_dlsi_fit(rep(0.8, 15L), n_obs = 45L, n_repeats = 5L)
  fit_g <- .make_dlsi_fit(rep(0.6, 15L), n_obs = 45L, n_repeats = 5L)
  res   <- delta_lsi(fit_n, fit_g, metric = "auc", seed = 3L)
  expect_equal(res@tier, "C_signflip")
  expect_equal(res@R_eff, 5L)
  # CI not available at tier C
  expect_true(all(!is.finite(res@delta_lsi_ci)))
  # Sign-flip p-value is available at tier C
  expect_true(is.finite(res@p_value))
})

test_that("tier B_signflip_ci for R_eff in 10..19", {
  set.seed(4)
  fit_n <- .make_dlsi_fit(rep(0.8, 30L), n_obs = 90L, n_repeats = 10L)
  fit_g <- .make_dlsi_fit(rep(0.6, 30L), n_obs = 90L, n_repeats = 10L)
  res   <- delta_lsi(fit_n, fit_g, metric = "auc", seed = 4L,
                     M_boot = 200L, M_flip = 500L)
  expect_equal(res@tier, "B_signflip_ci")
  expect_equal(res@R_eff, 10L)
  # Both p-value and CI are available at tier B
  expect_true(is.finite(res@p_value))
  expect_true(all(is.finite(res@delta_lsi_ci)))
})

test_that("tier A_full_inference for R_eff >= 20", {
  set.seed(5)
  fit_n <- .make_dlsi_fit(rep(0.8, 60L), n_obs = 120L, n_repeats = 20L)
  fit_g <- .make_dlsi_fit(rep(0.6, 60L), n_obs = 120L, n_repeats = 20L)
  res   <- delta_lsi(fit_n, fit_g, metric = "auc", seed = 5L,
                     M_boot = 200L, M_flip = 1000L)
  expect_equal(res@tier, "A_full_inference")
  expect_true(res@inference_ok)
  expect_true(all(is.finite(res@delta_lsi_ci)))
  expect_true(is.finite(res@p_value))
})

# ── strict mode ───────────────────────────────────────────────────────────────

test_that("strict = TRUE errors on insufficient R_eff", {
  fit_n <- .make_dlsi_fit(rep(0.7, 2L), n_repeats = 1L)
  fit_g <- .make_dlsi_fit(rep(0.5, 2L), n_repeats = 1L)
  expect_error(
    delta_lsi(fit_n, fit_g, metric = "auc", strict = TRUE),
    regexp = "R_eff"
  )
})

# ── Input validation ───────────────────────────────────────────────────────────

test_that("non-LeakFit input errors informatively", {
  fit <- .make_dlsi_fit(rep(0.7, 4L))
  expect_error(delta_lsi("not_a_fit", fit), regexp = "LeakFit")
  expect_error(delta_lsi(fit, 42),          regexp = "LeakFit")
})

# ── Unpaired fits ─────────────────────────────────────────────────────────────

test_that("unpaired fits give D_insufficient with point estimate", {
  fit_n <- .make_dlsi_fit(rep(0.8, 6L), n_obs = 30L, n_repeats = 2L)  # R=2
  fit_g <- .make_dlsi_fit(rep(0.6, 9L), n_obs = 30L, n_repeats = 3L)  # R=3
  res   <- suppressWarnings(
    delta_lsi(fit_n, fit_g, metric = "auc", seed = 1L)
  )
  expect_equal(res@tier, "D_insufficient")
  expect_equal(res@R_eff, 0L)
  # Point estimate should still be computed
  expect_true(is.finite(res@delta_lsi))
  expect_true(is.finite(res@delta_metric))
  expect_gt(res@delta_lsi, 0.0)
})

# ── Return details ────────────────────────────────────────────────────────────

test_that("return_details = TRUE stores delta_r and fits in info", {
  set.seed(6)
  fit_n <- .make_dlsi_fit(rep(0.75, 15L), n_obs = 45L, n_repeats = 5L)
  fit_g <- .make_dlsi_fit(rep(0.55, 15L), n_obs = 45L, n_repeats = 5L)
  res   <- delta_lsi(fit_n, fit_g, metric = "auc",
                     return_details = TRUE, seed = 6L)
  expect_true("delta_r"     %in% names(res@info))
  expect_true("fit_leaky"   %in% names(res@info))
  expect_true("fit_naive"   %in% names(res@info))
  expect_true("fit_guarded" %in% names(res@info))
  expect_length(res@info$delta_r, 5L)
})

test_that("deprecated fit_naive argument maps to fit_leaky", {
  set.seed(6)
  fit_n <- .make_dlsi_fit(rep(0.75, 15L), n_obs = 45L, n_repeats = 5L)
  fit_g <- .make_dlsi_fit(rep(0.55, 15L), n_obs = 45L, n_repeats = 5L)

  expect_warning(
    res <- delta_lsi(fit_naive = fit_n, fit_guarded = fit_g, metric = "auc",
                     seed = 6L),
    regexp = "fit_naive is deprecated; use fit_leaky instead"
  )
  expect_s4_class(res, "LeakDeltaLSI")
})

# ── BCa CI covers zero when naive = guarded ───────────────────────────────────

test_that("BCa CI covers 0 when no leakage (same AUC)", {
  set.seed(7)
  # 10 repeats × 3 folds; same AUC → delta = 0
  aucs <- rep(0.65, 30L)
  fit_n <- .make_dlsi_fit(aucs, n_obs = 90L, n_repeats = 10L)
  fit_g <- .make_dlsi_fit(aucs, n_obs = 90L, n_repeats = 10L)
  res   <- delta_lsi(fit_n, fit_g, metric = "auc",
                     M_boot = 500L, seed = 7L)
  expect_equal(res@delta_lsi, 0.0, tolerance = 1e-10)
  ci <- res@delta_lsi_ci
  if (all(is.finite(ci))) {
    expect_lte(ci[1L], 0.0 + 1e-8)
    expect_gte(ci[2L], 0.0 - 1e-8)
  }
})

# ── Sign-flip p-value ──────────────────────────────────────────────────────────

test_that("sign-flip p-value is in [0, 1]", {
  set.seed(8)
  fit_n <- .make_dlsi_fit(rep(0.8, 15L), n_obs = 45L, n_repeats = 5L)
  fit_g <- .make_dlsi_fit(rep(0.6, 15L), n_obs = 45L, n_repeats = 5L)
  res   <- delta_lsi(fit_n, fit_g, metric = "auc",
                     M_flip = 500L, seed = 8L)
  expect_gte(res@p_value, 0.0)
  expect_lte(res@p_value, 1.0)
})

test_that("sign-flip p-value small when inflation is large and consistent", {
  set.seed(9)
  # 20 repeats × 3 folds; large consistent gap
  fit_n <- .make_dlsi_fit(rep(0.95, 60L), n_obs = 120L, n_repeats = 20L)
  fit_g <- .make_dlsi_fit(rep(0.50, 60L), n_obs = 120L, n_repeats = 20L)
  res   <- delta_lsi(fit_n, fit_g, metric = "auc",
                     M_boot = 500L, M_flip = 2000L, seed = 9L)
  expect_lt(res@p_value, 0.05)
})

# ── show / summary don't error ────────────────────────────────────────────────

test_that("show() and summary() run without error", {
  set.seed(10)
  fit_n <- .make_dlsi_fit(rep(0.75, 15L), n_obs = 45L, n_repeats = 5L)
  fit_g <- .make_dlsi_fit(rep(0.55, 15L), n_obs = 45L, n_repeats = 5L)
  res   <- delta_lsi(fit_n, fit_g, metric = "auc",
                     M_boot = 100L, M_flip = 200L, seed = 10L)
  expect_output(show(res),     "LeakDeltaLSI")
  expect_output(summary(res),  "Delta LSI")
})

# ── Fold / repeat data frames ─────────────────────────────────────────────────

test_that("folds_naive and folds_guarded have expected columns", {
  set.seed(11)
  fit_n <- .make_dlsi_fit(c(0.7, 0.8, 0.75), n_obs = 30L, n_repeats = 1L)
  fit_g <- .make_dlsi_fit(c(0.6, 0.65, 0.6), n_obs = 30L, n_repeats = 1L)
  res   <- suppressWarnings(delta_lsi(fit_n, fit_g, metric = "auc"))
  expected_cols <- c("fold", "metric", "repeat_id", "n")
  expect_true(all(expected_cols %in% names(res@folds_naive)))
  expect_true(all(expected_cols %in% names(res@folds_guarded)))
})

# ── Internal helpers ──────────────────────────────────────────────────────────

test_that(".dlsi_huber returns median for single value", {
  expect_equal(bioLeak:::.dlsi_huber(42), 42)
})

test_that(".dlsi_huber returns NA for empty input", {
  expect_true(is.na(bioLeak:::.dlsi_huber(numeric(0))))
})

test_that(".dlsi_huber is robust to one outlier", {
  x   <- c(0, 0, 0, 0, 0, 100)
  est <- bioLeak:::.dlsi_huber(x)
  # Huber M-estimator should be much closer to 0 than arithmetic mean (100/6 ≈ 16.7)
  expect_lt(abs(est), abs(mean(x)))
})

test_that(".dlsi_bca_ci covers known mean for symmetric distribution", {
  set.seed(42)
  x  <- rnorm(20, mean = 0.2, sd = 0.05)
  ci <- bioLeak:::.dlsi_bca_ci(x, mean, M = 2000L, seed = 42L)
  expect_true(all(is.finite(ci)))
  expect_lt(ci[1L], 0.2)
  expect_gt(ci[2L], 0.2)
})

test_that(".dlsi_sign_flip p-value is NA for length-1 input", {
  expect_true(is.na(bioLeak:::.dlsi_sign_flip(0.5)))
})

test_that(".dlsi_sign_flip exact enumeration: known result for R=2", {
  # delta_r = c(1, 1); obs_stat = mean(1, 1) = 1
  # All 4 sign combos (R=2 <= 15 → exact):
  #   (--) → mean(-1, -1) = -1;  (-+) → mean(-1,  1) =  0
  #   (+-) → mean( 1, -1) =  0;  (++) → mean( 1,  1) =  1
  # |null| >= |obs| = 1: combos (--) and (++) → count = 2
  # Exact p = 2 / 4 = 0.5  (no continuity correction for full enumeration)
  p <- bioLeak:::.dlsi_sign_flip(c(1, 1), seed = 1L)
  expect_equal(p, 0.5, tolerance = 1e-10)
})

# ── Adversarial tests ─────────────────────────────────────────────────────────

test_that("equal repeat count but mismatched fold structures warns and goes unpaired", {
  # fit_n uses n_obs=30, fit_g uses n_obs=24 → different test indices → unpaired
  fit_n <- .make_dlsi_fit(rep(0.8, 6L), n_obs = 30L, n_repeats = 2L)
  fit_g <- .make_dlsi_fit(rep(0.6, 6L), n_obs = 24L, n_repeats = 2L)
  res <- expect_warning_match(
    delta_lsi(fit_n, fit_g, metric = "auc"),
    pattern = "fold structures"
  )
  expect_equal(res@tier, "D_insufficient")
  expect_equal(res@R_eff, 0L)
  # Point estimate should still be computed unpaired
  expect_true(is.finite(res@delta_lsi))
})

test_that("relabeled repeat_ids do not corrupt pairing (regression)", {
  n_folds <- 3L; n_obs <- 45L; n_repeats <- 5L

  # Naive: per-repeat means decrease from 0.9 to 0.7
  naive_aucs <- rep(c(0.9, 0.85, 0.80, 0.75, 0.70), each = n_folds)
  fit_n <- .make_dlsi_fit(naive_aucs, n_obs = n_obs, n_repeats = n_repeats)

  # Guarded: per-repeat means decrease from 0.8 to 0.6
  guard_aucs <- rep(c(0.80, 0.75, 0.70, 0.65, 0.60), each = n_folds)
  fit_g <- .make_dlsi_fit(guard_aucs, n_obs = n_obs, n_repeats = n_repeats)

  # Reverse the repeat_id labels in guarded fit's splits.
  # Original: positions 1-3 → rid=1, 4-6 → rid=2, ..., 13-15 → rid=5
  # Reversed: positions 1-3 → rid=5, 4-6 → rid=4, ..., 13-15 → rid=1
  new_idx <- fit_g@splits@indices
  for (i in seq_along(new_idx)) {
    new_idx[[i]]$repeat_id <- (n_repeats + 1L) - new_idx[[i]]$repeat_id
  }
  fit_g@splits@indices <- new_idx

  res <- delta_lsi(fit_n, fit_g, metric = "auc",
                   return_details = TRUE, seed = 42L)

  # With correct positional pairing, every repeat delta should be exactly 0.1.
  # Bug (before fix): deltas would be 0.3, 0.2, 0.1, 0.0, -0.1 due to
  # repeat_id sort-order mismatch between the two fits.
  expect_equal(res@info$delta_r, rep(0.1, n_repeats), tolerance = 1e-10)
  expect_true(res@info$paired)
  expect_equal(res@R_eff, n_repeats)
})

test_that("lower-is-better metric: delta > 0 when naive has lower (better) values", {
  set.seed(1)
  # For lower-is-better metrics: naive = 0.3 (better), guarded = 0.5 (worse)
  # Without sign flip: naive - guarded = -0.2 < 0 (looks like naive is worse)
  # With higher_is_better = FALSE: sign_factor = -1 → delta = -1 * (-0.2) = +0.2 > 0
  fit_n <- .make_dlsi_fit(rep(0.3, 10L), n_obs = 30L, n_repeats = 2L)
  fit_g <- .make_dlsi_fit(rep(0.5, 10L), n_obs = 30L, n_repeats = 2L)
  res   <- suppressWarnings(delta_lsi(
    fit_n, fit_g, metric = "auc",
    higher_is_better = FALSE, seed = 1L
  ))
  expect_gt(res@delta_lsi,    0.0)
  expect_gt(res@delta_metric, 0.0)
  expect_false(isTRUE(res@info[["higher_is_better"]]))
})

test_that("metrics df without fold column falls back to predictions gracefully", {
  # Build a LeakFit whose @metrics has the requested metric but no 'fold' column.
  # delta_lsi should fall back to .dlsi_fold_metrics_from_preds() without crashing.
  splits <- methods::new("LeakSplits",
    mode    = "subject_grouped",
    indices = list(
      list(fold = 1L, repeat_id = 1L, test = 1:15, train = 16:30),
      list(fold = 2L, repeat_id = 1L, test = 16:30, train = 1:15)
    ),
    info = list()
  )
  # @metrics has auc but NO fold column
  mdf <- data.frame(learner = "glm", auc = c(0.7, 0.75), stringsAsFactors = FALSE)

  set.seed(1)
  make_preds <- function(te, auc_val) {
    n <- length(te)
    y <- rbinom(n, 1L, 0.5)
    p <- pmin(1, pmax(0, rnorm(n, 0.5, auc_val / 2)))
    data.frame(id = te, fold = if (min(te) == 1L) 1L else 2L,
               learner = "glm", truth = y, pred = p, stringsAsFactors = FALSE)
  }
  preds <- list(make_preds(1:15, 0.7), make_preds(16:30, 0.75))

  fit_bad <- methods::new("LeakFit",
    splits = splits, metrics = mdf, metric_summary = data.frame(),
    audit = data.frame(), predictions = preds, preprocess = list(),
    learners = list(), outcome = "y", task = "binomial",
    feature_names = character(0L), info = list()
  )
  fit_good <- .make_dlsi_fit(c(0.6, 0.65), n_obs = 30L, n_repeats = 1L)

  # Should not crash; falls back to prediction-based metric recomputation
  expect_no_error(suppressWarnings(
    delta_lsi(fit_bad, fit_good, metric = "auc")
  ))
})

test_that("auto-detect higher_is_better = FALSE for rmse via metric name", {
  set.seed(2)
  # Naive RMSE = 0.3 (lower = better), guarded RMSE = 0.5
  fit_n <- .make_dlsi_fit(rep(0.3, 10L), n_obs = 30L, n_repeats = 2L)
  fit_g <- .make_dlsi_fit(rep(0.5, 10L), n_obs = 30L, n_repeats = 2L)
  # Add rmse column so .dlsi_fold_metrics can find it
  fit_n@metrics$rmse <- fit_n@metrics$auc
  fit_g@metrics$rmse <- fit_g@metrics$auc
  # Auto-detect: metric="rmse" with higher_is_better=NULL (default)
  res <- suppressWarnings(delta_lsi(fit_n, fit_g, metric = "rmse", seed = 2L))
  expect_false(isTRUE(res@info[["higher_is_better"]]))
  # Naive RMSE is lower (better) → positive delta (leakage inflation)
  expect_gt(res@delta_lsi,    0.0)
  expect_gt(res@delta_metric, 0.0)
})

# ── Exchangeability: by_group / within_batch warnings ─────────────────────────

test_that("exchangeability 'by_group' emits iid-fallback warning", {
  set.seed(20)
  fit_n <- .make_dlsi_fit(rep(0.8, 15L), n_obs = 45L, n_repeats = 5L)
  fit_g <- .make_dlsi_fit(rep(0.6, 15L), n_obs = 45L, n_repeats = 5L)
  expect_warning(
    res <- delta_lsi(fit_n, fit_g, metric = "auc",
                     exchangeability = "by_group", seed = 20L),
    regexp = "iid repeat structure"
  )
  # Despite the warning, a p-value is still computed (using iid sign-flip)
  expect_true(is.finite(res@p_value))
  expect_equal(res@exchangeability, "by_group")
})

test_that("exchangeability 'within_batch' emits iid-fallback warning", {
  set.seed(21)
  fit_n <- .make_dlsi_fit(rep(0.8, 15L), n_obs = 45L, n_repeats = 5L)
  fit_g <- .make_dlsi_fit(rep(0.6, 15L), n_obs = 45L, n_repeats = 5L)
  expect_warning(
    delta_lsi(fit_n, fit_g, metric = "auc",
              exchangeability = "within_batch", seed = 21L),
    regexp = "iid repeat structure"
  )
})

# ── Exchangeability: blocked_time block sign-flip ──────────────────────────────

test_that(".dlsi_resolve_block_size returns 1 for uncorrelated Δ_r", {
  # iid Δ_r has rho1 ≈ 0, so block_size should be 1
  set.seed(42)
  delta_r <- rnorm(20)
  bs <- bioLeak:::.dlsi_resolve_block_size(delta_r)
  expect_gte(bs, 1L)
  # With low autocorrelation, expected block_size is 1 (or very small)
  expect_lte(bs, 3L)
})

test_that(".dlsi_resolve_block_size respects explicit block_size argument", {
  delta_r <- rnorm(20)
  expect_equal(bioLeak:::.dlsi_resolve_block_size(delta_r, block_size = 4L), 4L)
})

test_that(".dlsi_resolve_block_size caps at floor(R/3)", {
  # Highly correlated Δ_r: AR(1) rho ~ 0.99, would give block_size = R
  # but cap forces at most floor(R/3)
  delta_r <- cumsum(rnorm(12))   # near unit root, very high autocorrelation
  bs      <- bioLeak:::.dlsi_resolve_block_size(delta_r)
  expect_lte(bs, floor(length(delta_r) / 3L))
})

test_that(".dlsi_sign_flip_blocked exact: known result for R=4, block_size=2", {
  # delta_r = c(1, 1, 1, 1); block_size=2 → block_id = c(1,1,2,2); n_blocks=2
  # All 4 block-sign combos (exact enumeration, n_blocks=2 <= 15):
  #   (--): mean(-1,-1,-1,-1) = -1
  #   (-+): mean(-1,-1, 1, 1) =  0
  #   (+-): mean( 1, 1,-1,-1) =  0
  #   (++): mean( 1, 1, 1, 1) =  1
  # obs_stat = 1; |null| >= |obs|=1: combos (--) and (++) → count = 2
  # Exact p = 2/4 = 0.5
  res <- bioLeak:::.dlsi_sign_flip_blocked(c(1, 1, 1, 1), block_size = 2L, seed = 1L)
  expect_equal(res$p_value,        0.5, tolerance = 1e-10)
  expect_equal(res$n_blocks,       2L)
  expect_equal(res$block_size_used, 2L)
})

test_that("blocked_time sign-flip gives more conservative p than iid for AR Δ_r", {
  set.seed(30)
  # Generate strongly AR(1) Δ_r with a positive mean (leakage signal)
  R <- 15L
  delta_r <- 0.2 + as.numeric(arima.sim(list(ar = 0.85), n = R, sd = 0.05))

  # iid sign-flip
  p_iid <- bioLeak:::.dlsi_sign_flip(delta_r, M_flip = 5000L, seed = 30L)

  # Blocked sign-flip with block_size=3 (5 blocks from 15 repeats)
  sf_blk <- bioLeak:::.dlsi_sign_flip_blocked(delta_r, block_size = 3L,
                                               M_flip = 5000L, seed = 30L)
  p_blk <- sf_blk$p_value

  # Block p-value should be >= iid p-value (more conservative under autocorrelation)
  # Allow tolerance for MC noise; the key property is p_blk is generally larger
  expect_gte(sf_blk$n_blocks, 5L)  # at least 5 blocks for inference
  expect_true(is.finite(p_blk))
  expect_true(is.finite(p_iid))
})

test_that("blocked_time with n_blocks < 5 sets p_value = NA and warns", {
  set.seed(31)
  # 6 repeats, block_size=3 → 2 blocks → below threshold of 5
  fit_n <- .make_dlsi_fit(rep(0.8, 18L), n_obs = 54L, n_repeats = 6L)
  fit_g <- .make_dlsi_fit(rep(0.6, 18L), n_obs = 54L, n_repeats = 6L)
  expect_warning(
    res <- delta_lsi(fit_n, fit_g, metric = "auc",
                     exchangeability = "blocked_time",
                     block_size = 3L, seed = 31L),
    regexp = "n_blocks"
  )
  expect_true(is.na(res@p_value))
  expect_equal(res@info$n_blocks, 2L)
})

test_that("blocked_time with explicit block_size=1 behaves like iid sign-flip", {
  set.seed(32)
  fit_n <- .make_dlsi_fit(rep(0.8, 15L), n_obs = 45L, n_repeats = 5L)
  fit_g <- .make_dlsi_fit(rep(0.6, 15L), n_obs = 45L, n_repeats = 5L)
  # block_size=1 → each repeat is its own block → equivalent to iid
  res_blk <- delta_lsi(fit_n, fit_g, metric = "auc",
                        exchangeability = "blocked_time",
                        block_size = 1L, seed = 32L)
  res_iid <- delta_lsi(fit_n, fit_g, metric = "auc",
                        exchangeability = "iid", seed = 32L)
  # p-values should be identical (same enumeration, same seed offset)
  expect_equal(res_blk@p_value, res_iid@p_value, tolerance = 1e-10)
})

test_that("blocked_time stores block_size_used and n_blocks in info", {
  set.seed(33)
  fit_n <- .make_dlsi_fit(rep(0.8, 30L), n_obs = 90L, n_repeats = 10L)
  fit_g <- .make_dlsi_fit(rep(0.6, 30L), n_obs = 90L, n_repeats = 10L)
  res <- delta_lsi(fit_n, fit_g, metric = "auc",
                    exchangeability = "blocked_time",
                    block_size = 2L, seed = 33L,
                    M_boot = 100L, M_flip = 500L)
  expect_equal(res@info$block_size_used, 2L)
  expect_equal(res@info$n_blocks,        5L)   # 10 repeats / block_size=2 = 5 blocks
  expect_true(is.finite(res@p_value))          # 5 blocks >= 5 → p_value available
})

# Guard-to-recipe bridge -----------------------------------------------------

#' Convert guard preprocessing steps to a recipes recipe
#'
#' Maps bioLeak guard preprocessing steps (impute, normalize, filter, fs) to
#' their closest \pkg{recipes} equivalents. Requires the \pkg{recipes} package.
#' Steps that have no direct recipe equivalent are skipped with a warning.
#'
#' @param steps A named list of guard preprocessing steps, e.g.,
#'   \code{list(impute = list(method = "median"), normalize = list(method = "zscore"))}.
#' @param formula A model formula (e.g., \code{outcome ~ .}).
#' @param training_data A data.frame used to initialize the recipe.
#' @return A \code{recipes::recipe} object with the mapped steps added.
#' @details
#' Mapping:
#' \itemize{
#'   \item \code{impute$method = "median"}: \code{step_impute_median(all_numeric_predictors())}
#'   \item \code{impute$method = "knn"}: \code{step_impute_knn(all_predictors(), neighbors = k)}
#'   \item \code{impute$method = "missForest"} or \code{"mice"}: Warning + \code{step_impute_median()} fallback
#'   \item \code{normalize$method = "zscore"}: \code{step_normalize(all_numeric_predictors())}
#'   \item \code{normalize$method = "robust"}: Warning + \code{step_normalize()} fallback
#'   \item \code{normalize$method = "none"}: No step added
#'   \item \code{filter$var_thresh > 0}: \code{step_nzv(all_numeric_predictors())}
#'   \item \code{fs$method = "pca"}: \code{step_pca(all_numeric_predictors(), num_comp = ncomp)}
#'   \item \code{fs$method = "ttest"} or \code{"lasso"}: Warning, skipped (no recipe equivalent)
#' }
#' @export
guard_to_recipe <- function(steps, formula, training_data) {
  if (!requireNamespace("recipes", quietly = TRUE)) {
    .bio_stop("The 'recipes' package is required for guard_to_recipe(). Install it with install.packages('recipes').",
              "bioLeak_dependency_error")
  }

  stopifnot(is.list(steps))
  stopifnot(inherits(formula, "formula"))
  stopifnot(is.data.frame(training_data))

  rec <- recipes::recipe(formula, data = training_data)

  # Imputation
  imp <- steps$impute
  if (!is.null(imp) && !is.null(imp$method)) {
    if (identical(imp$method, "median")) {
      rec <- recipes::step_impute_median(rec, recipes::all_numeric_predictors())
    } else if (identical(imp$method, "knn")) {
      k <- imp$k %||% imp$neighbors %||% 5L
      rec <- recipes::step_impute_knn(rec, recipes::all_predictors(), neighbors = k)
    } else if (imp$method %in% c("missForest", "mice")) {
      .bio_warn(sprintf("Imputation method '%s' has no direct recipe equivalent; falling back to step_impute_median().",
                       imp$method),
                "bioLeak_fallback_warning")
      rec <- recipes::step_impute_median(rec, recipes::all_numeric_predictors())
    }
  }

  # Normalization
  norm <- steps$normalize
  if (!is.null(norm) && !is.null(norm$method)) {
    if (identical(norm$method, "zscore")) {
      rec <- recipes::step_normalize(rec, recipes::all_numeric_predictors())
    } else if (identical(norm$method, "robust")) {
      .bio_warn("Normalization method 'robust' has no direct recipe equivalent; falling back to step_normalize().",
                "bioLeak_fallback_warning")
      rec <- recipes::step_normalize(rec, recipes::all_numeric_predictors())
    } else if (!identical(norm$method, "none")) {
      # Unknown normalization method — skip with a note
      .bio_warn(sprintf("Normalization method '%s' not recognized; skipping.", norm$method),
                "bioLeak_fallback_warning")
    }
  }

  # Variance filter
  filt <- steps$filter
  if (!is.null(filt)) {
    var_thresh <- filt$var_thresh %||% 0
    if (var_thresh > 0) {
      rec <- recipes::step_nzv(rec, recipes::all_numeric_predictors())
    }
  }

  # Feature selection
  fs <- steps$fs
  if (!is.null(fs) && !is.null(fs$method)) {
    if (identical(fs$method, "pca")) {
      ncomp <- fs$ncomp %||% fs$num_comp %||% 5L
      rec <- recipes::step_pca(rec, recipes::all_numeric_predictors(), num_comp = ncomp)
    } else if (fs$method %in% c("ttest", "lasso")) {
      .bio_warn(sprintf("Feature selection method '%s' has no direct recipe equivalent; skipping.",
                       fs$method),
                "bioLeak_fallback_warning")
    }
  }

  rec
}

skip_if_not_installed("recipes")

test_that("guard_to_recipe returns a recipe object", {
  df <- data.frame(outcome = c(0, 1, 0, 1), x1 = rnorm(4), x2 = rnorm(4))
  steps <- list(impute = list(method = "median"),
                normalize = list(method = "zscore"))
  rec <- guard_to_recipe(steps, outcome ~ ., df)
  expect_s3_class(rec, "recipe")
})

test_that("median + zscore adds correct step classes", {
  df <- data.frame(outcome = c(0, 1, 0, 1), x1 = rnorm(4), x2 = rnorm(4))
  steps <- list(impute = list(method = "median"),
                normalize = list(method = "zscore"))
  rec <- guard_to_recipe(steps, outcome ~ ., df)
  step_classes <- vapply(rec$steps, function(s) class(s)[1], character(1))
  expect_true("step_impute_median" %in% step_classes)
  expect_true("step_normalize" %in% step_classes)
})

test_that("kNN imputation adds step_impute_knn", {
  df <- data.frame(outcome = c(0, 1, 0, 1), x1 = rnorm(4), x2 = rnorm(4))
  steps <- list(impute = list(method = "knn", k = 3))
  rec <- guard_to_recipe(steps, outcome ~ ., df)
  step_classes <- vapply(rec$steps, function(s) class(s)[1], character(1))
  expect_true("step_impute_knn" %in% step_classes)
})

test_that("warns for unsupported FS methods (ttest, lasso)", {
  df <- data.frame(outcome = c(0, 1, 0, 1), x1 = rnorm(4), x2 = rnorm(4))
  expect_warning(
    guard_to_recipe(list(fs = list(method = "ttest")), outcome ~ ., df),
    "ttest"
  )
  expect_warning(
    guard_to_recipe(list(fs = list(method = "lasso")), outcome ~ ., df),
    "lasso"
  )
})

test_that("warns for missForest imputation fallback", {
  df <- data.frame(outcome = c(0, 1, 0, 1), x1 = rnorm(4), x2 = rnorm(4))
  expect_warning(
    guard_to_recipe(list(impute = list(method = "missForest")), outcome ~ ., df),
    "missForest"
  )
})

test_that("output recipe can be prep'd and bake'd without error", {
  set.seed(1)
  df <- data.frame(outcome = c(0, 1, 0, 1, 0, 1),
                   x1 = c(1, NA, 3, 4, 5, 6),
                   x2 = c(10, 20, 30, 40, 50, 60))
  steps <- list(impute = list(method = "median"),
                normalize = list(method = "zscore"))
  rec <- guard_to_recipe(steps, outcome ~ ., df)
  prepped <- recipes::prep(rec, training = df)
  baked <- recipes::bake(prepped, new_data = df)
  expect_true(is.data.frame(baked))
  expect_false(any(is.na(baked$x1)))
})

#' Fit and evaluate with leakage guards over predefined splits
#'
#' Performs cross-validated model training and evaluation using
#' leakage-protected preprocessing (guard_fit) and user-specified learners.
#'
#' @param x SummarizedExperiment or matrix/data.frame
#' @param outcome outcome column name (if x is SE or data.frame), or a length-2
#'   character vector of time/event column names for survival outcomes.
#' @param splits LeakSplits object from make_split_plan(), or an `rsample` rset/rsplit.
#' @param split_cols Optional named list/character vector or `"auto"` (default)
#'   overriding group/batch/study/time column names when `splits` is an rsample
#'   object and its attributes are missing. `"auto"` falls back to common
#'   metadata column names (e.g., `group`, `subject`, `batch`, `study`, `time`).
#'   Supported names are `group`, `batch`, `study`, and `time`.
#' @param store_refit_data Logical; when TRUE (default), stores the original
#'   data and learner configuration inside the fit to enable refit-based
#'   permutation tests without manual `perm_refit_spec` setup.
#' @param preprocess list(impute, normalize, filter=list(...), fs) or a
#'   `recipes::recipe` object. When a recipe is supplied, the guarded preprocessing
#'   pipeline is bypassed and the recipe is prepped on training data only.
#'   Recipe/workflow leakage guardrails run before fitting; configure policy via
#'   \code{options(bioLeak.validation_mode = "warn" | "error" | "off")}.
#' @param learner parsnip model_spec (or list of model_spec objects) describing
#'   the model(s) to fit, or a `workflows::workflow`. For legacy use, a character
#'   vector of learner names (e.g., "glmnet", "ranger") or custom learner IDs is
#'   still supported.
#' @param learner_args list of additional arguments passed to legacy learners
#'   (ignored when `learner` is a parsnip model_spec).
#' @param custom_learners named list of custom learner definitions used only
#'   with legacy character learners. Each entry
#'   must contain \code{fit} and \code{predict} functions. The \code{fit} function
#'   should accept \code{x}, \code{y}, \code{task}, and \code{weights}, and return
#'   a model object. The \code{predict} function should accept \code{object},
#'   \code{newdata}, and \code{task}. For binomial/regression/survival tasks it
#'   should return a numeric vector; for multiclass tasks it should return either
#'   class labels or a matrix/data.frame of class probabilities.
#' @param metrics named list of metric functions, vector of metric names, or a
#'   `yardstick::metric_set`. When a yardstick metric set (or list of yardstick
#'   metric functions) is supplied, metrics are computed using yardstick with the
#'   positive class set to the second factor level.
#' @param class_weights optional named numeric vector of weights for binomial or
#'   multiclass outcomes
#' @param positive_class optional value indicating the positive class for binomial outcomes.
#'   When set, the outcome levels are reordered so that \code{positive_class} is treated
#'   as the positive class (level 2). If NULL, the second factor level is used.
#' @param classification_threshold Numeric threshold in \code{[0, 1]} used to
#'   convert binomial probabilities into class predictions for
#'   \code{pred_class} and accuracy metrics. Ignored for non-binomial tasks.
#' @param parallel logical, use future.apply for multicore execution
#' @param refit logical, if TRUE retrain final model on full data
#' @param seed integer, for reproducibility
#' @return A \code{\linkS4class{LeakFit}} S4 object containing:
#'   \describe{
#'     \item{\code{splits}}{The \code{LeakSplits} object used for resampling.}
#'     \item{\code{metrics}}{Data.frame of per-fold, per-learner performance
#'       metrics with columns \code{fold}, \code{learner}, and one column per
#'       requested metric.}
#'     \item{\code{metric_summary}}{Data.frame summarizing metrics across folds
#'       for each learner with columns \code{learner}, and \code{<metric>_mean}
#'       and \code{<metric>_sd} for each requested metric.}
#'     \item{\code{audit}}{Data.frame with per-fold audit information including
#'       \code{fold}, \code{n_train}, \code{n_test}, \code{learner}, and
#'       \code{features_final} (number of features after preprocessing).}
#'     \item{\code{predictions}}{List of data.frames containing out-of-fold
#'       predictions with columns \code{id} (sample identifier), \code{truth}
#'       (true outcome), \code{pred} (predicted value or probability), \code{fold},
#'       and \code{learner}. For classification tasks, includes \code{pred_class}.
#'       For multiclass, includes per-class probability columns.}
#'     \item{\code{preprocess}}{List of preprocessing state objects from each fold,
#'       storing imputation parameters, normalization statistics, and feature
#'       selection results.}
#'     \item{\code{learners}}{List of fitted model objects from each fold.}
#'     \item{\code{outcome}}{Character string naming the outcome variable.}
#'     \item{\code{task}}{Character string indicating the task type
#'       (\code{"binomial"}, \code{"multiclass"}, \code{"gaussian"}, or
#'       \code{"survival"}).}
#'     \item{\code{feature_names}}{Character vector of feature names after
#'       preprocessing.}
#'     \item{\code{info}}{List of additional metadata including \code{hash},
#'       \code{metrics_used}, \code{class_weights}, \code{positive_class},
#'       \code{sample_ids}, \code{fold_status}, \code{refit}, \code{final_model} (refitted model if
#'       \code{refit = TRUE}), \code{final_preprocess}, \code{learner_names},
#'       and \code{perm_refit_spec} (for permutation-based audits).}
#'   }
#'   Use \code{summary()} to print a formatted report, or access slots directly
#'   with \code{@}.
#' @details
#' Preprocessing is fit on the training fold and applied to the test fold,
#' preventing leakage from global imputation, scaling, or feature selection.
#' When a `recipes::recipe` or `workflows::workflow` is supplied, the recipe is
#' prepped on the training fold and baked on the test fold.
#' For data.frame or matrix inputs, columns used to define splits
#' (outcome, group, batch, study, time) are excluded from the predictor matrix.
#' Use \code{learner_args} to pass model-specific arguments, either as a named
#' list keyed by learner or a single list applied to all learners. For custom
#' learners, \code{learner_args[[name]]} may be a list with \code{fit} and
#' \code{predict} sublists to pass distinct arguments to each stage. For binomial
#' tasks, predictions and metrics assume the positive class is the second factor
#' level; use \code{positive_class} to control this. Use
#' \code{classification_threshold} to change the probability cutoff used for
#' class labels and accuracy. Parsnip learners must support
#' probability predictions for binomial metrics (AUC/PR-AUC/accuracy) and
#' multiclass log-loss when requested.
#' @examples
#' set.seed(1)
#' df <- data.frame(
#'   subject = rep(1:10, each = 2),
#'   outcome = rbinom(20, 1, 0.5),
#'   x1 = rnorm(20),
#'   x2 = rnorm(20)
#' )
#' splits <- make_split_plan(df, outcome = "outcome",
#'                       mode = "subject_grouped", group = "subject", v = 5)
#'
#' # glmnet learner (requires glmnet package)
#' fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                       learner = "glmnet", metrics = "auc")
#' summary(fit)
#'
#' # Custom learner (logistic regression) - no extra packages needed
#' custom <- list(
#'   glm = list(
#'     fit = function(x, y, task, weights, ...) {
#'       stats::glm(y ~ ., data = as.data.frame(x),
#'                  family = stats::binomial(), weights = weights)
#'     },
#'     predict = function(object, newdata, task, ...) {
#'       as.numeric(stats::predict(object, newdata = as.data.frame(newdata), type = "response"))
#'     }
#'   )
#' )
#' fit2 <- fit_resample(df, outcome = "outcome", splits = splits,
#'                      learner = "glm", custom_learners = custom,
#'                      metrics = "accuracy")
#'
#' summary(fit2)
#' @export
fit_resample <- function(x, outcome, splits,
                         preprocess = list(
                           impute = list(method = "median"),
                           normalize = list(method = "zscore"),
                           filter = list(var_thresh = 0, iqr_thresh = 0),
                           fs = list(method = "none")
                         ),
                         learner = c("glmnet", "ranger"),
                         learner_args = list(),
                         custom_learners = list(),
                         metrics = c("auc", "pr_auc", "accuracy"),
                         class_weights = NULL,
                         positive_class = NULL,
                         classification_threshold = 0.5,
                         parallel = FALSE,
                         refit = TRUE,
                         seed = 1,
                         split_cols = "auto",
                         store_refit_data = TRUE) {

  set.seed(seed)
  .bio_strict_checks(context = "fit_resample", seed = seed)
  classification_threshold_supplied <- !missing(classification_threshold)
  learner_input <- learner
  if (!is.numeric(classification_threshold) || length(classification_threshold) != 1L ||
      !is.finite(classification_threshold) || classification_threshold < 0 ||
      classification_threshold > 1) {
    .bio_stop("classification_threshold must be a single numeric value in [0, 1].",
              "bioLeak_input_error")
  }
  if (is.null(custom_learners)) custom_learners <- list()
  if (!is.list(custom_learners)) {
    .bio_stop("custom_learners must be a named list of learner definitions.",
              "bioLeak_input_error")
  }
  if (length(custom_learners) && is.null(names(custom_learners))) {
    .bio_stop("custom_learners must be a named list.",
              "bioLeak_input_error")
  }
  if (length(custom_learners)) {
    dup_names <- intersect(names(custom_learners), c("glmnet", "ranger"))
    if (length(dup_names)) {
      .bio_stop(sprintf("custom_learners cannot override built-in learners: %s",
                       paste(dup_names, collapse = ", ")),
                "bioLeak_input_error")
    }
    bad <- vapply(custom_learners, function(def) {
      !is.list(def) || !is.function(def$fit) || !is.function(def$predict)
    }, logical(1))
    if (any(bad)) {
      .bio_stop("Each custom learner must be a list with `fit` and `predict` functions.",
                "bioLeak_input_error")
    }
  }

  is_parsnip_spec <- function(obj) inherits(obj, "model_spec")
  is_workflow <- function(obj) inherits(obj, "workflow")
  use_parsnip <- FALSE
  use_workflow <- FALSE
  learner_specs <- NULL
  learner_names <- NULL

  if (is_workflow(learner)) {
    use_workflow <- TRUE
    learner_specs <- list(learner)
  } else if (is.list(learner) && length(learner) &&
             all(vapply(learner, is_workflow, logical(1)))) {
    use_workflow <- TRUE
    learner_specs <- learner
  } else if (is_parsnip_spec(learner)) {
    use_parsnip <- TRUE
    learner_specs <- list(learner)
  } else if (is.list(learner) && length(learner) &&
             all(vapply(learner, is_parsnip_spec, logical(1)))) {
    use_parsnip <- TRUE
    learner_specs <- learner
  }

  if (use_workflow) {
    if (!requireNamespace("workflows", quietly = TRUE)) {
      .bio_stop("Package 'workflows' is required when passing a workflow to learner.",
                "bioLeak_dependency_error")
    }
    if (length(custom_learners)) {
      warning("custom_learners ignored when learner is a workflow.")
    }
    if (length(learner_args)) {
      warning("learner_args ignored when learner is a workflow.")
    }
    learner_names <- names(learner_specs)
    if (is.null(learner_names)) learner_names <- rep("", length(learner_specs))
    missing_names <- !nzchar(learner_names)
    if (any(missing_names)) {
      fallback <- paste0("workflow_", seq_along(learner_specs))
      learner_names[missing_names] <- fallback[missing_names]
    }
  } else if (use_parsnip) {
    if (!requireNamespace("parsnip", quietly = TRUE)) {
      .bio_stop("Package 'parsnip' is required when passing a model_spec to learner.",
                "bioLeak_dependency_error")
    }
    if (length(custom_learners)) {
      warning("custom_learners ignored when learner is a parsnip model_spec.")
    }
    if (length(learner_args)) {
      warning("learner_args ignored when learner is a parsnip model_spec.")
    }
    parsnip_label <- function(spec, fallback) {
      model_class <- class(spec)
      model_class <- model_class[model_class != "model_spec"]
      model_class <- model_class[1]
      label <- if (!is.null(model_class) && nzchar(model_class)) model_class else fallback
      engine <- NULL
      if (!is.null(spec$engine) && nzchar(spec$engine)) {
        engine <- spec$engine
      } else if (!is.null(spec$method) && !is.null(spec$method$engine) &&
                 nzchar(spec$method$engine)) {
        engine <- spec$method$engine
      }
      if (!is.null(engine)) label <- paste0(label, "/", engine)
      label
    }
    learner_names <- names(learner_specs)
    if (is.null(learner_names)) learner_names <- rep("", length(learner_specs))
    missing_names <- !nzchar(learner_names)
    if (any(missing_names)) {
      fallback <- paste0("spec_", seq_along(learner_specs))
      spec_labels <- vapply(seq_along(learner_specs), function(i) {
        parsnip_label(learner_specs[[i]], fallback[[i]])
      }, character(1))
      learner_names[missing_names] <- spec_labels[missing_names]
    }
  } else {
    builtin_learners <- c("glmnet", "ranger")
    all_learners <- c(builtin_learners, names(custom_learners))
    if (!is.character(learner)) {
      .bio_stop("learner must be a character vector of legacy learners, a parsnip model_spec, or a workflow.",
                "bioLeak_input_error")
    }
    learner <- match.arg(learner, choices = all_learners, several.ok = TRUE)
    learner_names <- learner
  }
  use_recipe <- .bio_is_recipe(preprocess)
  if (use_recipe && !requireNamespace("recipes", quietly = TRUE)) {
    .bio_stop("Package 'recipes' is required when preprocess is a recipe.",
              "bioLeak_dependency_error")
  }
  if (use_workflow && isTRUE(use_recipe)) {
    warning("Recipe preprocess ignored when learner is a workflow.")
    use_recipe <- FALSE
  }
  preprocess_mode <- if (use_workflow) "workflow" else if (use_recipe) "recipe" else "guard"
  validation_mode <- .bio_validation_mode()
  if (use_recipe) {
    .bio_validate_recipe_graph(
      preprocess,
      context = "fit_resample",
      mode = validation_mode
    )
  }
  if (use_workflow && length(learner_specs)) {
    for (wf in learner_specs) {
      .bio_validate_workflow_graph(
        wf,
        context = "fit_resample",
        mode = validation_mode
      )
    }
  }

  Xall <- .bio_get_x(x)
  Xall_raw <- Xall
  yall <- .bio_get_y(x, outcome)

  if (!inherits(splits, "LeakSplits")) {
    if (.bio_is_rsample(splits)) {
      coldata <- if (.bio_is_se(x)) {
        .bio_coldata_df(x)
      } else if (is.data.frame(x)) {
        x
      } else if (is.matrix(x)) {
        data.frame(row_id = seq_len(nrow(Xall)))
      } else {
        NULL
      }
      splits <- .bio_as_leaksplits_from_rsample(splits, n = nrow(Xall), coldata = coldata,
                                                split_cols = split_cols)
    } else {
      .bio_stop("splits must be a LeakSplits or rsample rset/rsplit.",
                "bioLeak_input_error")
    }
  }
  drop_cols <- outcome
  if (inherits(splits, "LeakSplits")) {
    split_info <- splits@info
    drop_cols <- unique(c(drop_cols,
                          split_info$group,
                          split_info$batch,
                          split_info$study,
                          split_info$time))
  }
  drop_cols <- drop_cols[!is.na(drop_cols) & nzchar(drop_cols)]
  if (length(drop_cols) && !is.null(colnames(Xall))) {
    drop_cols <- intersect(colnames(Xall), drop_cols)
  }
  if (length(drop_cols)) {
    Xall <- Xall[, setdiff(colnames(Xall), drop_cols), drop = FALSE]
  }
  sample_ids <- NULL
  if (inherits(splits, "LeakSplits") && !is.null(splits@info$coldata)) {
    cd <- splits@info$coldata
    rn_cd <- rownames(cd)
    if (!is.null(rn_cd) && !anyNA(rn_cd) && all(nzchar(rn_cd)) && !anyDuplicated(rn_cd)) {
      sample_ids <- rn_cd
    } else if ("row_id" %in% names(cd)) {
      rid <- as.character(cd[["row_id"]])
      if (length(rid) == nrow(cd) && !anyNA(rid) && !anyDuplicated(rid) && all(nzchar(rid))) {
        sample_ids <- rid
      }
    }
  }
  if (is.null(sample_ids)) {
    rn <- rownames(Xall)
    if (!is.null(rn) && !anyNA(rn) && all(nzchar(rn)) && !anyDuplicated(rn)) {
      sample_ids <- rn
    }
  }
  if (is.null(sample_ids) || length(sample_ids) != nrow(Xall)) {
    sample_ids <- as.character(seq_len(nrow(Xall)))
  }
  ids <- sample_ids
  compact <- isTRUE(splits@info$compact)
  fold_assignments <- splits@info$fold_assignments
  split_mode <- splits@mode
  split_time <- splits@info$time
  split_horizon <- splits@info$horizon %||% 0
  split_purge <- splits@info$purge %||% 0
  split_embargo <- splits@info$embargo %||% 0
  split_coldata <- splits@info$coldata
  time_vec <- NULL
  if (compact && identical(split_mode, "time_series")) {
    if (is.null(split_coldata) || is.null(split_time) || !split_time %in% names(split_coldata)) {
      stop("time_series compact splits require time column in coldata.")
    }
    time_vec <- split_coldata[[split_time]]
  }
  task <- if (.bio_is_survival(yall)) "survival"
  else if (.bio_is_binomial(yall)) "binomial"
  else if (.bio_is_multiclass(yall)) "multiclass"
  else if (.bio_is_regression(yall)) "gaussian"
  else if (is.factor(yall) && nlevels(yall) == 2) "binomial"
  else if (is.factor(yall) && nlevels(yall) > 2) "multiclass"
  else .bio_stop("Unsupported outcome type: require binomial/multiclass factor, numeric regression, or survival outcome.",
                 "bioLeak_input_error")

  if (task == "binomial") {
    if (!is.factor(yall)) yall <- factor(yall)
    yall <- droplevels(yall)
    if (anyNA(yall)) {
      .bio_stop("Binomial outcome contains NA values. Remove or impute missing outcomes before fitting.",
                "bioLeak_input_error")
    }
    if (nlevels(yall) != 2) {
      .bio_stop("Binomial task requires exactly two outcome levels after preprocessing.",
                "bioLeak_input_error")
    }
    if (!is.null(positive_class)) {
      pos_chr <- as.character(positive_class)
      if (length(pos_chr) != 1L) {
        .bio_stop("positive_class must be a single value.",
                  "bioLeak_input_error")
      }
      levels_y <- levels(yall)
      if (!pos_chr %in% levels_y) {
        .bio_stop(sprintf("positive_class '%s' not found in outcome levels: %s",
                          pos_chr, paste(levels_y, collapse = ", ")),
                  "bioLeak_input_error")
      }
      if (!identical(pos_chr, levels_y[2])) {
        levels_y <- c(setdiff(levels_y, pos_chr), pos_chr)
        yall <- factor(yall, levels = levels_y)
      }
    }
    class_levels <- levels(yall)
    if (!is.null(class_weights)) {
      if (!is.numeric(class_weights)) stop("class_weights must be numeric.")
      if (is.null(names(class_weights))) {
        if (length(class_weights) != length(class_levels)) {
          stop("class_weights must align with outcome levels.")
        }
        names(class_weights) <- class_levels[seq_along(class_weights)]
      }
      missing_cw <- setdiff(class_levels, names(class_weights))
      if (length(missing_cw)) {
        stop(sprintf("class_weights missing levels: %s", paste(missing_cw, collapse = ", ")))
      }
      class_weights <- class_weights[class_levels]
    }
  } else if (task == "multiclass") {
    if (!is.factor(yall)) yall <- factor(yall)
    yall <- droplevels(yall)
    if (anyNA(yall)) {
      stop("Multiclass outcome contains NA values. Remove or impute missing outcomes before fitting.", call. = FALSE)
    }
    if (nlevels(yall) < 3) {
      stop("Multiclass task requires 3 or more outcome levels after preprocessing.")
    }
    class_levels <- levels(yall)
    if (!is.null(class_weights)) {
      if (!is.numeric(class_weights)) stop("class_weights must be numeric.")
      if (is.null(names(class_weights))) {
        if (length(class_weights) != length(class_levels)) {
          stop("class_weights must align with outcome levels.")
        }
        names(class_weights) <- class_levels[seq_along(class_weights)]
      }
      missing_cw <- setdiff(class_levels, names(class_weights))
      if (length(missing_cw)) {
        stop(sprintf("class_weights missing levels: %s", paste(missing_cw, collapse = ", ")))
      }
      class_weights <- class_weights[class_levels]
    }
    if (!is.null(positive_class)) {
      warning("positive_class is ignored for multiclass tasks.")
    }
    if (classification_threshold_supplied) {
      warning("classification_threshold is ignored for multiclass tasks.")
    }
  } else if (task == "gaussian") {
    if (!is.numeric(yall)) {
      yall <- as.numeric(yall)
      if (anyNA(yall)) stop("Gaussian task requires numeric outcome values.")
    }
    class_levels <- NULL
    if (!is.null(class_weights)) {
      warning("class_weights is ignored for gaussian tasks.")
    }
    if (!is.null(positive_class)) {
      warning("positive_class is ignored for gaussian tasks.")
    }
    if (classification_threshold_supplied) {
      warning("classification_threshold is ignored for gaussian tasks.")
    }
  } else {
    class_levels <- NULL
    if (!inherits(yall, "Surv")) {
      stop("Survival task requires a Surv outcome.")
    }
    if (anyNA(yall)) {
      stop("Survival outcome contains NA values. Remove or impute missing outcomes before fitting.", call. = FALSE)
    }
    if (!is.null(class_weights)) {
      warning("class_weights is ignored for survival tasks.")
    }
    if (!is.null(positive_class)) {
      warning("positive_class is ignored for survival tasks.")
    }
    if (classification_threshold_supplied) {
      warning("classification_threshold is ignored for survival tasks.")
    }
  }

  if (use_workflow && task == "binomial" && !is.null(class_weights)) {
    warning("class_weights are ignored for workflow learners unless explicitly handled in the workflow.")
  }
  if (use_workflow && task == "multiclass" && !is.null(class_weights)) {
    warning("class_weights are ignored for workflow learners unless explicitly handled in the workflow.")
  }

  metrics_input <- metrics
  metric_mode <- "legacy"
  yardstick_set <- NULL
  yardstick_metrics <- NULL

  if (!is.null(metrics) && inherits(metrics, "metric_set")) {
    metric_mode <- "yardstick"
    yardstick_set <- metrics
  } else if (!is.null(metrics) && .bio_is_yardstick_metric(metrics)) {
    metric_mode <- "yardstick"
    yardstick_metrics <- list(metrics)
  } else if (is.list(metrics) && length(metrics) &&
             all(vapply(metrics, .bio_is_yardstick_metric, logical(1)))) {
    metric_mode <- "yardstick"
    yardstick_metrics <- metrics
  }

  if (identical(metric_mode, "yardstick")) {
    if (!requireNamespace("yardstick", quietly = TRUE)) {
      stop("Package 'yardstick' is required when metrics are a yardstick set.", call. = FALSE)
    }
    if (is.null(yardstick_set)) {
      yardstick_set <- do.call(yardstick::metric_set, yardstick_metrics)
    }
    metric_labels <- character(0)
  } else {
    if (is.null(metrics)) {
      metrics <- if (task == "binomial") c("auc", "pr_auc", "accuracy")
      else if (task == "multiclass") c("accuracy", "macro_f1")
      else if (task == "survival") c("cindex")
      else c("rmse")
    }

    if (is.character(metrics)) {
      allowed <- if (task == "binomial") c("auc", "pr_auc", "accuracy")
      else if (task == "multiclass") c("accuracy", "macro_f1", "log_loss")
      else if (task == "survival") c("cindex")
      else c("rmse", "cindex")
      invalid <- setdiff(metrics, allowed)
      if (length(invalid)) {
        warning(sprintf("Dropping metrics not applicable to %s task: %s", task,
                        paste(invalid, collapse = ", ")))
        metrics <- setdiff(metrics, invalid)
      }
      if (!length(metrics)) {
        metrics <- if (task == "binomial") c("auc", "pr_auc", "accuracy")
        else if (task == "multiclass") c("accuracy", "macro_f1")
        else if (task == "survival") c("cindex")
        else c("rmse")
      }
    }

    metrics <- if (is.list(metrics)) metrics else as.list(metrics)
    metric_labels <- vapply(seq_along(metrics), function(i) {
      nm <- names(metrics)[i]
      if (!is.null(nm) && !is.na(nm) && nzchar(nm)) return(nm)
      mi <- metrics[[i]]
      if (is.character(mi) && length(mi) == 1) return(mi)
      paste0("metric_", i)
    }, character(1))
  }

  learner_objs <- if (use_parsnip || use_workflow) learner_specs else as.list(learner_names)

  # helper: safe metric computation -------------------------------------------
  compute_metric <- function(name, y, pred) {
    if (is.function(name)) return(name(y, pred))
    yb <- NULL
    if (task == "binomial") {
      yb <- if (is.factor(y)) as.numeric(y) - 1 else as.numeric(y)
    }
    if (name == "auc" && task == "binomial") {
      if (requireNamespace("pROC", quietly = TRUE))
        return(as.numeric(pROC::auc(pROC::roc(y, pred, quiet = TRUE))))
      pos <- pred[yb == 1]
      neg <- pred[yb == 0]
      comp <- outer(pos, neg, function(a, b) (a > b) + 0.5 * (a == b))
      return(mean(comp))
    }
    if (name == "pr_auc" && task == "binomial") {
      if (requireNamespace("PRROC", quietly = TRUE)) {
        pr <- PRROC::pr.curve(scores.class0 = pred[yb == 1],
                              scores.class1 = pred[yb == 0],
                              curve = FALSE)
        return(pr$auc.integral)
      }
      return(NA_real_)
    }
    if (name == "accuracy" && task == "binomial") {
      return(mean((pred >= classification_threshold) == as.logical(yb)))
    }
    if (name == "rmse" && task == "gaussian")
      return(sqrt(mean((as.numeric(y) - pred)^2)))
    if (name == "cindex" && task == "gaussian")
      return(.cindex_pairwise(pred, y))
    if (name == "cindex" && task == "survival")
      return(.cindex_survival(pred, y))
    NA_real_
  }

  compute_yardstick <- function(y, pred, pred_class, prob = NULL) {
    if (task == "survival") {
      stop("Yardstick metrics are not supported for survival tasks.", call. = FALSE)
    }
    if (task == "multiclass") {
      df <- data.frame(truth = y, pred_class = pred_class, stringsAsFactors = FALSE)
      if (!is.null(prob)) {
        prob <- as.data.frame(prob, check.names = FALSE)
        prob_cols <- paste0(".pred_", make.names(class_levels))
        if (ncol(prob) == length(class_levels)) {
          names(prob) <- prob_cols
        }
        df <- cbind(df, prob)
      }
      pred_cols <- grep("^\\.pred_", names(df), value = TRUE)
      args <- list(df, truth = quote(truth), estimate = quote(pred_class))
      for (col in pred_cols) {
        args[[col]] <- as.name(col)
      }
      res <- try(do.call(yardstick_set, args), silent = TRUE)
    } else if (task == "binomial") {
      df <- data.frame(truth = y, .pred = as.numeric(pred), stringsAsFactors = FALSE)
      if (!is.null(pred_class)) df$.pred_class <- pred_class
      res <- try(yardstick_set(df, truth = truth, estimate = .pred_class,
                               .pred, event_level = "second"),
                 silent = TRUE)
    } else {
      df <- data.frame(truth = y, .pred = as.numeric(pred), stringsAsFactors = FALSE)
      res <- try(yardstick_set(df, truth = truth, estimate = .pred),
                 silent = TRUE)
    }
    if (inherits(res, "try-error")) {
      err_msg <- attr(res, "condition")$message
      stop(sprintf("Yardstick metrics failed: %s", err_msg), call. = FALSE)
    }
    stats::setNames(res$.estimate, res$.metric)
  }

  align_probabilities <- function(prob, class_levels) {
    prob_mat <- if (is.data.frame(prob)) as.matrix(prob) else as.matrix(prob)
    if (is.null(colnames(prob_mat))) {
      if (ncol(prob_mat) != length(class_levels)) {
        stop("Probability predictions do not match class levels.", call. = FALSE)
      }
      colnames(prob_mat) <- class_levels
      return(prob_mat)
    }
    exp_cols <- paste0(".pred_", make.names(class_levels))
    if (all(exp_cols %in% colnames(prob_mat))) {
      prob_mat <- prob_mat[, exp_cols, drop = FALSE]
    } else if (all(class_levels %in% colnames(prob_mat))) {
      prob_mat <- prob_mat[, class_levels, drop = FALSE]
    } else if (all(make.names(class_levels) %in% colnames(prob_mat))) {
      prob_mat <- prob_mat[, make.names(class_levels), drop = FALSE]
    } else if (ncol(prob_mat) == length(class_levels)) {
      prob_mat <- prob_mat[, seq_len(ncol(prob_mat)), drop = FALSE]
    } else {
      stop("Probability predictions do not align with class levels.", call. = FALSE)
    }
    colnames(prob_mat) <- class_levels
    prob_mat
  }

  # --- Robust design matrix builder ------------------------------------------
  make_design_matrix <- function(X, ref_cols = NULL) {
    X <- as.data.frame(X)
    X <- X[, !names(X) %in% c("y", "outcome"), drop = FALSE]

    is_num <- vapply(X, is.numeric, logical(1))
    if (all(is_num)) {
      mm <- as.matrix(X)
    } else {
      mf <- stats::model.frame(~ ., data = X, na.action = stats::na.pass)
      mm <- stats::model.matrix(~ . - 1, data = mf)
      mm <- as.matrix(mm)
    }

    if (!is.null(ref_cols)) {
      missing_cols <- setdiff(ref_cols, colnames(mm))
      if (length(missing_cols)) {
        mm <- cbind(
          mm,
          matrix(0, nrow = nrow(mm), ncol = length(missing_cols),
                 dimnames = list(NULL, missing_cols))
        )
      }
      extra_cols <- setdiff(colnames(mm), ref_cols)
      if (length(extra_cols)) {
        mm <- mm[, setdiff(colnames(mm), extra_cols), drop = FALSE]
      }
      mm <- mm[, ref_cols, drop = FALSE]
      return(list(matrix = mm, columns = ref_cols))
    }

    if (!ncol(mm)) {
      stop("All predictors have zero variance after preprocessing.")
    }

    keep <- apply(mm, 2, sd, na.rm = TRUE) > 0
    if (!any(keep)) {
      stop("All predictors have zero variance after preprocessing.")
    }
    mm <- mm[, keep, drop = FALSE]

    list(matrix = mm, columns = colnames(mm))
  }

  resolve_weights <- function(y, weights_spec) {
    if (is.null(weights_spec) || !task %in% c("binomial", "multiclass")) return(NULL)
    if (!is.numeric(weights_spec)) stop("class_weights must be numeric.")
    cw <- weights_spec
    if (is.null(names(cw))) {
      if (length(cw) != length(class_levels)) {
        stop("Provide class_weights as a named vector matching outcome levels.")
      }
      names(cw) <- class_levels[seq_along(cw)]
    }
    missing_levels <- setdiff(class_levels, names(cw))
    if (length(missing_levels)) {
      stop(sprintf("class_weights missing levels: %s", paste(missing_levels, collapse = ", ")))
    }
    cw[as.character(y)]
  }

  resolve_args <- function(name, defaults = list()) {
    if (!length(learner_args)) return(modifyList(defaults, list()))
    if (!is.null(names(learner_args)) && all(names(learner_args) %in% learner)) {
      extras <- learner_args[[name]] %||% list()
    } else {
      extras <- learner_args
    }
    modifyList(defaults, extras)
  }

  resolve_custom_args <- function(name) {
    if (!length(learner_args)) return(list(fit = list(), predict = list()))
    if (!is.null(names(learner_args)) && all(names(learner_args) %in% learner)) {
      extras <- learner_args[[name]] %||% list()
    } else {
      extras <- learner_args
    }
    if (is.list(extras) && (("fit" %in% names(extras)) || ("predict" %in% names(extras)))) {
      return(list(fit = extras$fit %||% list(),
                  predict = extras$predict %||% list()))
    }
    list(fit = extras, predict = extras)
  }

  # single learner wrapper ----------------------------------------------------
  train_one_learner <- function(learner_obj, learner_label, Xtrg, ytr, Xteg, yte, weights = NULL) {
    if (inherits(learner_obj, "model_spec")) {
      if (!requireNamespace("parsnip", quietly = TRUE)) {
        stop("Package 'parsnip' is required when learner is a model_spec.")
      }
      if (!is.null(weights) && !inherits(weights, "hardhat_case_weights")) {
        if (!requireNamespace("hardhat", quietly = TRUE)) {
          stop("Package 'hardhat' is required for case weights with parsnip learners.")
        }
        weights <- hardhat::frequency_weights(weights)
      }
      y_for_fit <- if (task %in% c("binomial", "multiclass")) {
        factor(ytr, levels = class_levels)
      } else if (task == "survival") {
        ytr
      } else {
        as.numeric(ytr)
      }
      fit <- if (is.null(weights)) {
        parsnip::fit_xy(learner_obj, x = Xtrg, y = y_for_fit)
      } else {
        parsnip::fit_xy(learner_obj, x = Xtrg, y = y_for_fit, case_weights = weights)
      }
      if (task == "binomial") {
        prob <- try(stats::predict(fit, new_data = Xteg, type = "prob"), silent = TRUE)
        if (inherits(prob, "try-error")) {
          # Extract the actual error text from parsnip/xgboost
          err_msg <- attr(prob, "condition")$message
          stop(sprintf("Parsnip learner '%s' failed to predict: %s",
                       learner_label, err_msg))
        }
        prob_df <- as.data.frame(prob)
        pos_col <- paste0(".pred_", make.names(class_levels[2]))
        if (!pos_col %in% names(prob_df)) {
          if (ncol(prob_df) >= 2L) {
            pos_col <- names(prob_df)[2]
          } else {
            stop(sprintf("Parsnip learner '%s' did not return class probabilities.",
                         learner_label))
          }
        }
        pred <- as.numeric(prob_df[[pos_col]])
        pred_class <- factor(ifelse(pred >= classification_threshold, class_levels[2], class_levels[1]),
                             levels = class_levels)
        return(list(pred = pred, pred_class = pred_class, fit = fit))
      }
      if (task == "multiclass") {
        prob <- try(stats::predict(fit, new_data = Xteg, type = "prob"), silent = TRUE)
        if (inherits(prob, "try-error")) {
          err_msg <- attr(prob, "condition")$message
          stop(sprintf("Parsnip learner '%s' failed to predict: %s",
                       learner_label, err_msg))
        }
        prob_mat <- align_probabilities(prob, class_levels)
        class_pred <- try(stats::predict(fit, new_data = Xteg, type = "class"), silent = TRUE)
        if (inherits(class_pred, "try-error")) {
          err_msg <- attr(class_pred, "condition")$message
          stop(sprintf("Parsnip learner '%s' failed to predict classes: %s",
                       learner_label, err_msg))
        }
        class_df <- as.data.frame(class_pred)
        pred_class <- factor(as.character(class_df[[1]]), levels = class_levels)
        return(list(pred = pred_class, pred_class = pred_class, prob = prob_mat, fit = fit))
      }
      if (task == "survival") {
        pred_df <- try(stats::predict(fit, new_data = Xteg, type = "numeric"), silent = TRUE)
        if (inherits(pred_df, "try-error")) {
          pred_df <- try(stats::predict(fit, new_data = Xteg, type = "risk"), silent = TRUE)
        }
        if (inherits(pred_df, "try-error")) {
          err_msg <- attr(pred_df, "condition")$message
          stop(sprintf("Parsnip learner '%s' failed to predict: %s",
                       learner_label, err_msg))
        }
        pred <- as.numeric(as.data.frame(pred_df)[[1]])
        return(list(pred = pred, fit = fit))
      } else {
        pred_df <- stats::predict(fit, new_data = Xteg, type = "numeric")
        pred <- as.numeric(pred_df[[1]])
        return(list(pred = pred, fit = fit))
      }
    }
    if (learner_obj %in% names(custom_learners)) {
      def <- custom_learners[[learner_obj]]
      args <- resolve_custom_args(learner_obj)
      fit_args <- c(list(x = Xtrg, y = ytr, task = task, weights = weights), args$fit)
      model <- do.call(def$fit, fit_args)
      pred_args <- c(list(object = model, newdata = Xteg, task = task), args$predict)
      pred <- do.call(def$predict, pred_args)
      if (task == "multiclass") {
        if (is.data.frame(pred) || is.matrix(pred)) {
          prob_mat <- align_probabilities(pred, class_levels)
          pred_class <- factor(class_levels[max.col(prob_mat, ties.method = "first")],
                               levels = class_levels)
          return(list(pred = pred_class, pred_class = pred_class, prob = prob_mat, fit = model))
        }
        if (is.factor(pred) || is.character(pred)) {
          pred_class <- factor(as.character(pred), levels = class_levels)
          if (length(pred_class) != nrow(Xteg)) {
            stop(sprintf("Custom learner '%s' returned %d predictions for %d rows.",
                         learner_obj, length(pred_class), nrow(Xteg)))
          }
          return(list(pred = pred_class, pred_class = pred_class, fit = model))
        }
        stop(sprintf("Custom learner '%s' must return class labels or class probabilities for multiclass tasks.",
                     learner_obj))
      }
      pred <- as.numeric(pred)
      if (length(pred) != nrow(Xteg)) {
        stop(sprintf("Custom learner '%s' returned %d predictions for %d rows.",
                     learner_obj, length(pred), nrow(Xteg)))
      }
      return(list(pred = pred, fit = model))
    }
    if (learner_obj == "glmnet") {
      if (!requireNamespace("glmnet", quietly = TRUE)) stop("Install 'glmnet'.")
      fam <- if (task == "binomial") "binomial"
      else if (task == "multiclass") "multinomial"
      else if (task == "survival") "cox"
      else "gaussian"
      la  <- resolve_args("glmnet", list(alpha = 0.9, standardize = FALSE))
      Xtr_design <- make_design_matrix(Xtrg)
      Xte_design <- make_design_matrix(Xteg, ref_cols = Xtr_design$columns)
      y_for_fit <- if (task == "binomial") {
        as.numeric(factor(ytr, levels = class_levels)) - 1
      } else if (task == "multiclass") {
        factor(ytr, levels = class_levels)
      } else if (task == "survival") {
        ytr
      } else {
        as.numeric(ytr)
      }
      cv_args <- c(list(x = Xtr_design$matrix,
                        y = y_for_fit,
                        family = fam,
                        alpha = la$alpha,
                        standardize = la$standardize %||% FALSE,
                        weights = weights),
                   la[setdiff(names(la), c("alpha", "standardize"))])
      cvfit <- do.call(glmnet::cv.glmnet, cv_args)
      if (task == "multiclass") {
        pred_arr <- predict(cvfit, Xte_design$matrix, s = "lambda.min", type = "response")
        if (length(dim(pred_arr)) == 3L) {
          prob_mat <- pred_arr[, , 1, drop = FALSE][,,1]
        } else {
          prob_mat <- pred_arr
        }
        prob_mat <- align_probabilities(prob_mat, class_levels)
        pred_class <- factor(class_levels[max.col(prob_mat, ties.method = "first")],
                             levels = class_levels)
        return(list(pred = pred_class, pred_class = pred_class, prob = prob_mat, fit = cvfit))
      }
      pred_type <- if (task == "survival") "link" else "response"
      pred  <- as.numeric(predict(cvfit, Xte_design$matrix, s = "lambda.min",
                                  type = pred_type))

      return(list(pred = pred, fit = cvfit))
    }
    if (learner_obj == "ranger") {
      if (!requireNamespace("ranger", quietly = TRUE)) stop("Install 'ranger'.")
      if (task == "survival") {
        stop("Learner 'ranger' does not support survival tasks in bioLeak; use parsnip/workflow or a custom learner.")
      }
      y_for_fit <- if (task %in% c("binomial", "multiclass")) {
        factor(ytr, levels = class_levels)
      } else {
        as.numeric(ytr)
      }
      dftr <- data.frame(y = y_for_fit, Xtrg, check.names = FALSE)
      frm  <- stats::as.formula("y ~ .")
      rng_args <- resolve_args("ranger", list())
      if (task %in% c("binomial", "multiclass")) {
        rng_args$class.weights <- rng_args$class.weights %||% class_weights
      }
      rg   <- do.call(ranger::ranger, c(list(formula = frm, data = dftr,
                                             probability = (task %in% c("binomial", "multiclass"))),
                                        rng_args))
      dfte <- data.frame(Xteg, check.names = FALSE)
      pr   <- predict(rg, dfte)
      if (task == "binomial") {
        pred <- pr$predictions[, 2]
        pred_class <- factor(ifelse(pred >= classification_threshold, class_levels[2], class_levels[1]),
                             levels = class_levels)
        return(list(pred = pred, pred_class = pred_class, fit = rg))
      }
      if (task == "multiclass") {
        prob_mat <- pr$predictions
        prob_mat <- align_probabilities(prob_mat, class_levels)
        pred_class <- factor(class_levels[max.col(prob_mat, ties.method = "first")],
                             levels = class_levels)
        return(list(pred = pred_class, pred_class = pred_class, prob = prob_mat, fit = rg))
      }
      pred <- pr$predictions
      return(list(pred = pred, fit = rg))
    }
    stop("Unsupported learner.")
  }

  train_one_workflow <- function(learner_obj, learner_label, dftr, dfte, weights = NULL) {
    if (!requireNamespace("workflows", quietly = TRUE)) {
      stop("Package 'workflows' is required when learner is a workflow.")
    }
    fit <- try(generics::fit(learner_obj, data = dftr), silent = TRUE)
    if (inherits(fit, "try-error")) {
      err_msg <- attr(fit, "condition")$message
      stop(sprintf("Workflow learner '%s' failed to fit: %s", learner_label, err_msg))
    }
    if (task == "binomial") {
      prob <- try(stats::predict(fit, new_data = dfte, type = "prob"), silent = TRUE)
      if (inherits(prob, "try-error")) {
        err_msg <- attr(prob, "condition")$message
        stop(sprintf("Workflow learner '%s' failed to predict: %s", learner_label, err_msg))
      }
      prob_df <- as.data.frame(prob)
      pos_col <- paste0(".pred_", make.names(class_levels[2]))
      if (!pos_col %in% names(prob_df)) {
        if (ncol(prob_df) >= 2L) {
          pos_col <- names(prob_df)[2]
        } else {
          stop(sprintf("Workflow learner '%s' did not return class probabilities.",
                       learner_label))
        }
      }
      pred <- as.numeric(prob_df[[pos_col]])
      pred_class <- factor(ifelse(pred >= classification_threshold, class_levels[2], class_levels[1]),
                           levels = class_levels)
      return(list(pred = pred, pred_class = pred_class, fit = fit))
    }
    if (task == "multiclass") {
      prob <- try(stats::predict(fit, new_data = dfte, type = "prob"), silent = TRUE)
      if (inherits(prob, "try-error")) {
        err_msg <- attr(prob, "condition")$message
        stop(sprintf("Workflow learner '%s' failed to predict: %s", learner_label, err_msg))
      }
      prob_mat <- align_probabilities(prob, class_levels)
      class_pred <- try(stats::predict(fit, new_data = dfte, type = "class"), silent = TRUE)
      if (inherits(class_pred, "try-error")) {
        err_msg <- attr(class_pred, "condition")$message
        stop(sprintf("Workflow learner '%s' failed to predict classes: %s",
                     learner_label, err_msg))
      }
      class_df <- as.data.frame(class_pred)
      pred_class <- factor(as.character(class_df[[1]]), levels = class_levels)
      return(list(pred = pred_class, pred_class = pred_class, prob = prob_mat, fit = fit))
    }
    if (task == "survival") {
      pred_df <- try(stats::predict(fit, new_data = dfte, type = "numeric"), silent = TRUE)
      if (inherits(pred_df, "try-error")) {
        pred_df <- try(stats::predict(fit, new_data = dfte, type = "risk"), silent = TRUE)
      }
      if (inherits(pred_df, "try-error")) {
        err_msg <- attr(pred_df, "condition")$message
        stop(sprintf("Workflow learner '%s' failed to predict: %s", learner_label, err_msg))
      }
      pred <- as.numeric(as.data.frame(pred_df)[[1]])
      return(list(pred = pred, fit = fit))
    } else {
      pred_df <- stats::predict(fit, new_data = dfte, type = "numeric")
      pred <- as.numeric(pred_df[[1]])
      return(list(pred = pred, fit = fit))
    }
  }

  resolve_fold_indices <- function(fold) {
    if (!isTRUE(compact) || !is.null(fold$train)) return(fold)
    if (is.null(fold_assignments) || !length(fold_assignments)) {
      stop("Compact splits require fold assignments to compute indices.")
    }
    r <- fold$repeat_id
    if (is.null(r) || !is.finite(r)) r <- 1L
    r <- as.integer(r)
    if (r < 1L || r > length(fold_assignments)) {
      stop(sprintf("repeat_id %d out of range [1, %d].", r, length(fold_assignments)))
    }
    assign_vec <- fold_assignments[[r]]
    if (is.null(assign_vec)) {
      stop(sprintf("Missing fold assignments for repeat %s.", r))
    }
    test <- which(assign_vec == fold$fold)
    if (identical(split_mode, "time_series")) {
      if (is.null(time_vec) || !length(time_vec)) {
        stop("time_series compact splits require time column values.")
      }
      train <- .bio_time_series_train_indices(
        time_vec = time_vec,
        test_idx = test,
        candidate_idx = seq_len(nrow(Xall)),
        horizon = split_horizon,
        purge = split_purge,
        embargo = split_embargo
      )
    } else if (identical(split_mode, "combined")) {
      # Combined mode: exclude samples sharing constraint-axis levels with test
      remaining <- setdiff(seq_len(nrow(Xall)), test)
      cst <- splits@info$constraints
      if (!is.null(cst) && length(cst) > 1L && !is.null(split_coldata)) {
        for (ax in cst[-1]) {
          ax_vec <- split_coldata[[ax$col]]
          test_levels <- unique(ax_vec[test])
          remaining <- remaining[!ax_vec[remaining] %in% test_levels]
        }
      }
      train <- remaining
    } else {
      train <- setdiff(seq_len(nrow(Xall)), test)
    }
    fold_seq <- fold$fold_seq %||% fold$fold
    list(train = train, test = test, fold = fold$fold,
         repeat_id = fold$repeat_id, fold_seq = fold_seq)
  }

  make_fold_df <- function(X, y) {
    df <- as.data.frame(X, check.names = FALSE)
    if (length(outcome) == 2L) {
      if (!inherits(y, "Surv")) {
        stop("Survival tasks require a Surv outcome when building fold data.")
      }
      y_mat <- as.matrix(y)
      if (ncol(y_mat) < 2L) {
        stop("Survival outcome must include time and event columns.")
      }
      df[[outcome[[1]]]] <- y_mat[, 1]
      df[[outcome[[2]]]] <- y_mat[, ncol(y_mat)]
      df <- df[, c(outcome, setdiff(names(df), outcome)), drop = FALSE]
      return(df)
    }
    df[[outcome]] <- y
    df[, c(outcome, setdiff(names(df), outcome)), drop = FALSE]
  }

  # fold-level function -------------------------------------------------------
  do_fold <- function(fold) {
    fold_full <- resolve_fold_indices(fold)
    fold_id <- fold_full$fold_seq %||% fold_full$fold
    set.seed(seed + fold_id)
    tr <- fold_full$train
    te <- fold_full$test

    Xtr <- Xall[tr, , drop = FALSE]
    Xte <- Xall[te, , drop = FALSE]
    Xtr_raw <- Xall_raw[tr, , drop = FALSE]
    Xte_raw <- Xall_raw[te, , drop = FALSE]
    ytr <- yall[tr]
    yte <- yall[te]

    if (task %in% c("binomial", "multiclass")) {
      ytr <- factor(ytr, levels = class_levels)
      yte <- factor(yte, levels = class_levels)
      if (nlevels(droplevels(ytr)) < 2) {
        warning(sprintf("Fold %s skipped: only one class in training data", fold_id))
        empty_metrics <- if (identical(metric_mode, "yardstick")) {
          numeric(0)
        } else {
          setNames(rep(NA_real_, length(metrics)), metric_labels)
        }
        skipped <- lapply(learner_names, function(ln) {
          list(
            metrics = empty_metrics,
            pred = data.frame(
              id = integer(0),
              truth = factor(character(0), levels = class_levels),
              pred = numeric(0),
              fold = integer(0),
              learner = character(0),
              stringsAsFactors = FALSE
            ),
            guard = list(state = NULL),
            learner = NULL,
            feat_names = colnames(Xtr)
          )
        })
        names(skipped) <- learner_names
        return(skipped)
      }
      fold_weights <- resolve_weights(ytr, class_weights)
    } else if (task == "survival") {
      fold_weights <- NULL
    } else {
      ytr <- as.numeric(ytr)
      yte <- as.numeric(yte)
      fold_weights <- NULL
    }

    preprocess_state <- NULL
    Xtrg <- Xtr
    Xteg <- Xte
    dftr <- NULL
    dfte <- NULL

    if (identical(preprocess_mode, "guard")) {
      guard <- guard_fit(
        X = Xtr,
        y = ytr,
        steps = if (exists("preprocess") && is.list(preprocess)) preprocess else list(),
        task  = task
      )
      Xtrg <- guard$transform(Xtr)
      Xteg <- guard$transform(Xte)
      preprocess_state <- guard$state
      colnames(Xtrg) <- make.names(colnames(Xtrg))
      colnames(Xteg) <- make.names(colnames(Xteg))
    } else if (identical(preprocess_mode, "recipe")) {
      dftr <- make_fold_df(Xtr_raw, ytr)
      dfte <- make_fold_df(Xte_raw, yte)
      recipe_prep <- recipes::prep(preprocess, training = dftr, retain = TRUE)
      Xtrg <- as.data.frame(recipes::juice(recipe_prep, recipes::all_predictors()),
                            check.names = FALSE)
      Xteg <- as.data.frame(recipes::bake(recipe_prep, new_data = dfte,
                                          recipes::all_predictors()),
                            check.names = FALSE)
      if (length(drop_cols)) {
        keep_cols <- setdiff(names(Xtrg), drop_cols)
        Xtrg <- Xtrg[, keep_cols, drop = FALSE]
        Xteg <- Xteg[, intersect(keep_cols, names(Xteg)), drop = FALSE]
        missing_in_test <- setdiff(names(Xtrg), names(Xteg))
        if (length(missing_in_test)) {
          for (nm in missing_in_test) {
            Xteg[[nm]] <- if (is.numeric(Xtrg[[nm]])) mean(Xtrg[[nm]], na.rm = TRUE) else 0
          }
          warning("Test fold missing columns filled from training statistics: ",
                  paste(missing_in_test, collapse = ", "), call. = FALSE)
        }
        Xteg <- Xteg[, names(Xtrg), drop = FALSE]
      }
      if (!ncol(Xtrg)) stop("All predictors have zero variance after preprocessing.")
      preprocess_state <- list(type = "recipe", recipe = recipe_prep)
      colnames(Xtrg) <- make.names(colnames(Xtrg))
      colnames(Xteg) <- make.names(colnames(Xteg))
    } else {
      dftr <- make_fold_df(Xtr, ytr)
      dfte <- make_fold_df(Xte, yte)
      preprocess_state <- list(type = "workflow")
    }

    results <- list()
    for (i in seq_along(learner_names)) {
      ln <- learner_names[[i]]
      learner_obj <- learner_objs[[i]]
      # Train one learner; ensure consistent level matching
      if (identical(preprocess_mode, "workflow")) {
        dfte_pred <- dfte[, setdiff(names(dfte), outcome), drop = FALSE]
        model <- train_one_workflow(learner_obj, ln, dftr, dfte_pred, weights = fold_weights)
        feat_names <- setdiff(names(dftr), outcome)
      } else {
        model <- train_one_learner(learner_obj, ln, Xtrg, ytr, Xteg, yte, weights = fold_weights)
        feat_names <- colnames(Xtrg)
      }

      pred_class <- if (task == "binomial") {
        if (!is.null(model$pred_class)) {
          model$pred_class
        } else if (is.factor(model$pred) || is.character(model$pred)) {
          factor(as.character(model$pred), levels = class_levels)
        } else {
          factor(ifelse(model$pred >= classification_threshold, class_levels[2], class_levels[1]),
                 levels = class_levels)
        }
      } else if (task == "multiclass") {
        model$pred_class %||% {
          if (is.factor(model$pred) || is.character(model$pred)) {
            factor(as.character(model$pred), levels = class_levels)
          } else {
            NULL
          }
        }
      } else {
        NULL
      }

      prob_mat <- model$prob %||% NULL

      ms <- if (identical(metric_mode, "yardstick")) {
        compute_yardstick(yte, model$pred, pred_class, prob = prob_mat)
      } else if (task == "multiclass") {
        vals <- vapply(seq_along(metrics), function(idx) {
          mname <- metrics[[idx]]
          if (is.function(mname)) return(mname(yte, pred_class))
          if (identical(mname, "accuracy")) return(.multiclass_accuracy(yte, pred_class))
          if (identical(mname, "macro_f1")) return(.multiclass_macro_f1(yte, pred_class))
          if (identical(mname, "log_loss")) {
            if (is.null(prob_mat)) {
              warning("log_loss requires class probability predictions; returning NA.",
                      call. = FALSE)
              return(NA_real_)
            }
            return(.multiclass_log_loss(yte, prob_mat))
          }
          NA_real_
        }, numeric(1))
        names(vals) <- metric_labels
        vals
      } else {
        vals <- vapply(seq_along(metrics), function(idx) {
          compute_metric(metrics[[idx]], yte, model$pred)
        }, numeric(1))
        names(vals) <- metric_labels
        vals
      }
      pred_tbl <- if (task == "survival") {
        yte_mat <- as.matrix(yte)
        time_col <- if ("time" %in% colnames(yte_mat)) "time" else colnames(yte_mat)[1]
        status_col <- if ("status" %in% colnames(yte_mat)) "status" else colnames(yte_mat)[ncol(yte_mat)]
        data.frame(
          id = ids[te],
          truth_time = yte_mat[, time_col],
          truth_event = yte_mat[, status_col],
          pred = model$pred,
          fold = fold_id,
          learner = ln,
          stringsAsFactors = FALSE
        )
      } else {
        data.frame(
          id = ids[te],
          truth = yte,
          pred = model$pred,
          fold = fold_id,
          learner = ln,
          stringsAsFactors = FALSE
        )
      }
      if (!is.null(pred_class)) pred_tbl$pred_class <- pred_class
      if (!is.null(prob_mat) && task == "multiclass") {
        prob_df <- as.data.frame(prob_mat, check.names = FALSE)
        names(prob_df) <- paste0(".pred_", make.names(class_levels))
        pred_tbl <- cbind(pred_tbl, prob_df)
      }
      results[[ln]] <- list(
        metrics = ms,
        pred = pred_tbl,
        guard = preprocess_state,
        learner = model$fit,
        feat_names = feat_names
      )
    }

    results
  }


  folds <- splits@indices
  nfold <- length(folds)
  fold_errors <- rep(NA_character_, nfold)
  parallel_run <- parallel && requireNamespace("future.apply", quietly = TRUE)

  # progress bar --------------------------------------------------------------
  pb <- utils::txtProgressBar(min = 0, max = nfold, style = 3)
  pb_counter <- 0
  progress_wrap <- function(f) {
    local_fold_id <- f$fold_seq %||% f$fold
    start_time <- proc.time()
    err_msg <- NULL
    res <- tryCatch(do_fold(f), error = function(e) {
      err_msg <<- conditionMessage(e)
      fold_errors[[local_fold_id]] <<- conditionMessage(e)
      warning(sprintf("Fold %s failed: %s", local_fold_id, e$message)); NULL
    })
    elapsed <- (proc.time() - start_time)[["elapsed"]]
    if (is.null(res)) {
      res <- structure(list(), elapsed_sec = elapsed, fold_error = err_msg)
    } else {
      attr(res, "elapsed_sec") <- elapsed
    }
    if (!isTRUE(parallel_run)) {
      pb_counter <<- pb_counter + 1
      utils::setTxtProgressBar(pb, pb_counter)
    }
    res
  }

  # parallel or sequential execution -----------------------------------------
  if (isTRUE(parallel_run)) {
    out <- future.apply::future_lapply(seq_along(folds), function(i) {
      fold <- folds[[i]]
      fold$fold_seq <- i
      progress_wrap(fold)
    }, future.seed = TRUE)
  } else {
    out <- lapply(seq_along(folds), function(i) {
      fold <- folds[[i]]
      fold$fold_seq <- i
      progress_wrap(fold)
    })
  }

  close(pb)

  # collect results -----------------------------------------------------------
  met_rows <- list()
  preds <- list()
  guards <- list()
  lears <- list()
  featn <- NULL
  audit_rows <- list()
  fold_status_rows <- vector("list", length(out))

  for (fold_idx in seq_along(out)) {
    fold_res <- out[[fold_idx]]
    learner_total <- length(learner_names)
    learner_success <- 0L
    fold_elapsed <- attr(fold_res, "elapsed_sec") %||% NA_real_
    if (is.null(fold_res) || !length(names(fold_res))) next
    fold_info <- resolve_fold_indices(folds[[fold_idx]])
    fold_id <- fold_idx
    for (ln in names(fold_res)) {
      res <- fold_res[[ln]]
      if (is.null(res)) next
      m <- res$metrics
      if (all(is.na(m))) {
        next
      }
      learner_success <- learner_success + 1L
      metric_row <- c(list(fold = fold_id, learner = ln), as.list(m))
      met_rows[[length(met_rows) + 1]] <- as.data.frame(metric_row,
                                                        row.names = NULL,
                                                        check.names = FALSE)
      if (is.data.frame(res$pred) || is.matrix(res$pred)) {
        preds[[length(preds) + 1]] <- res$pred
      }
      if (!is.null(res$guard) && is.list(res$guard)) {
        guards[[length(guards) + 1]] <- res$guard
        if (is.null(featn) && !is.null(res$feat_names)) featn <- res$feat_names
      } else {
        guards[[length(guards) + 1]] <- NULL
      }
      lears[[length(lears) + 1]] <- res$learner
      audit_rows[[length(audit_rows) + 1]] <- data.frame(
        fold = fold_id,
        n_train = length(fold_info$train),
        n_test = length(fold_info$test),
        learner = ln,
        features_final = if (!is.null(res$guard) &&
                              is.list(res$guard) &&
                              !is.null(res$guard$filter) &&
                              !is.null(res$guard$filter$keep)) {
          sum(res$guard$filter$keep)
        } else if (!is.null(res$feat_names)) {
          length(res$feat_names)
        } else NA_integer_,
        row.names = NULL
      )
    }
    all_preds_empty <- length(fold_res) > 0 && all(vapply(fold_res, function(res) {
      is.null(res$pred) || nrow(res$pred) == 0L
    }, logical(1)))
    fold_status_rows[[fold_idx]] <- data.frame(
      fold = fold_idx,
      stage = if (learner_success > 0L) "fold_run" else "fold_precheck",
      status = if (learner_success > 0L) "success" else "skipped",
      reason = if (learner_success > 0L) NA_character_ else if (all_preds_empty) {
        "single_class_training"
      } else {
        "no_valid_metrics"
      },
      notes = sprintf("Successful learners: %d/%d", learner_success, learner_total),
      elapsed_sec = fold_elapsed,
      stringsAsFactors = FALSE
    )
  }

  for (fold_idx in seq_along(out)) {
    if (!is.null(fold_status_rows[[fold_idx]])) next
    err_note <- attr(out[[fold_idx]], "fold_error") %||% fold_errors[[fold_idx]]
    if (is.null(err_note) || is.na(err_note)) err_note <- "Fold failed before producing any learner results."
    fold_elapsed <- attr(out[[fold_idx]], "elapsed_sec") %||% NA_real_
    fold_status_rows[[fold_idx]] <- data.frame(
      fold = fold_idx,
      stage = "fold_run",
      status = "failed",
      reason = "fold_error",
      notes = err_note,
      elapsed_sec = fold_elapsed,
      stringsAsFactors = FALSE
    )
  }
  fold_status_df <- do.call(rbind, fold_status_rows)

  if (!length(met_rows)) {
    .bio_stop("No successful folds were completed. Check learner and preprocessing settings.",
              "bioLeak_fit_error")
  }

  met_df <- do.call(rbind, met_rows)
  audit_df <- do.call(rbind, audit_rows)
  metrics_used <- setdiff(colnames(met_df), c("fold", "learner"))

  # summarize metrics ---------------------------------------------------------
  metric_summary_raw <- aggregate(. ~ learner, data = met_df[, -1, drop = FALSE],
                                  FUN = function(x) c(mean = mean(x, na.rm = TRUE),
                                                      sd = sd(x, na.rm = TRUE)),
                                  na.action = stats::na.pass)
  # flatten embedded matrices from aggregate() into separate columns
  metric_summary <- data.frame(learner = metric_summary_raw$learner,
                               stringsAsFactors = FALSE)
  for (col in setdiff(colnames(metric_summary_raw), "learner")) {
    mat <- metric_summary_raw[[col]]
    if (is.matrix(mat)) {
      metric_summary[[paste0(col, "_mean")]] <- mat[, "mean"]
      metric_summary[[paste0(col, "_sd")]] <- mat[, "sd"]
    } else {
      metric_summary[[col]] <- mat
    }
  }

  # confidence intervals -------------------------------------------------------
  tryCatch({
    avg_n_train <- mean(audit_df$n_train, na.rm = TRUE)
    avg_n_test <- mean(audit_df$n_test, na.rm = TRUE)
    ci_df <- cv_ci(met_df, method = "nadeau_bengio",
                   n_train = avg_n_train, n_test = avg_n_test)
    # merge CI columns into metric_summary
    ci_cols <- grep("_ci_lo$|_ci_hi$", names(ci_df), value = TRUE)
    if (length(ci_cols) && nrow(ci_df) == nrow(metric_summary)) {
      for (cc in ci_cols) {
        metric_summary[[cc]] <- ci_df[[cc]]
      }
    }
  }, error = function(e) NULL)

  # optional refit ------------------------------------------------------------
  final_model <- NULL
  final_guard <- NULL
  if (refit) {
    ln <- learner_names[[1]]
    learner_obj <- learner_objs[[1]]
    full_weights <- if (task %in% c("binomial", "multiclass")) resolve_weights(yall, class_weights) else NULL
    if (identical(preprocess_mode, "guard")) {
      guard_full <- guard_fit(Xall, yall, preprocess, task)
      Xfullg <- guard_full$transform(Xall)
      colnames(Xfullg) <- make.names(colnames(Xfullg))
      final_model <- train_one_learner(learner_obj, ln, Xfullg, yall, Xfullg, yall,
                                       weights = full_weights)$fit
      final_guard <- guard_full$state
    } else if (identical(preprocess_mode, "recipe")) {
      df_full <- make_fold_df(Xall_raw, yall)
      recipe_prep <- recipes::prep(preprocess, training = df_full, retain = TRUE)
      Xfullg <- as.data.frame(recipes::juice(recipe_prep, recipes::all_predictors()),
                              check.names = FALSE)
      if (length(drop_cols)) {
        keep_cols <- setdiff(names(Xfullg), drop_cols)
        Xfullg <- Xfullg[, keep_cols, drop = FALSE]
      }
      colnames(Xfullg) <- make.names(colnames(Xfullg))
      final_model <- train_one_learner(learner_obj, ln, Xfullg, yall, Xfullg, yall,
                                       weights = full_weights)$fit
      final_guard <- list(type = "recipe", recipe = recipe_prep)
    } else {
      df_full <- make_fold_df(Xall, yall)
      final_model <- generics::fit(learner_obj, data = df_full)
      final_guard <- list(type = "workflow")
    }
  }

  # assemble LeakFit object ---------------------------------------------------
  perm_refit_spec <- NULL
  if (isTRUE(store_refit_data)) {
      perm_refit_spec <- list(
        x = x,
        outcome = outcome,
        preprocess = preprocess,
        learner = learner_input,
        learner_args = learner_args,
        custom_learners = custom_learners,
        class_weights = class_weights,
        positive_class = positive_class,
        classification_threshold = classification_threshold,
        parallel = parallel
      )
    }

  new("LeakFit",
      splits = splits,
      metrics = met_df,
      metric_summary = metric_summary,
      audit = audit_df,
      predictions = preds,
      preprocess = guards,
      learners = lears,
      outcome = outcome,
      task = task,
      feature_names = featn,
      info = list(hash = .bio_hash_indices(folds),
                  metrics_requested = metrics_input,
                  metrics_used = metrics_used,
                  class_weights = class_weights,
                  positive_class = if (task == "binomial") class_levels[2] else NULL,
                  classification_threshold = if (task == "binomial") classification_threshold else NULL,
                  sample_ids = ids,
                  fold_seeds = setNames(seed + seq_along(folds),
                                        paste0("fold", seq_along(folds))),
                  fold_status = fold_status_df,
                  refit = refit,
                  final_model = final_model,
                  final_preprocess = final_guard,
                  learner_names = learner_names,
                  perm_refit_spec = perm_refit_spec,
                  provenance = .bio_capture_provenance()))
}


# ---- LeakFit: show() method --------------------------------------------
# Brief auto-print representation. Detailed reporting is produced by
# summary(). Added in 0.3.7 (Comment 8 response): standard R idiom is
# that every public S4 class should have a show() method so that
# auto-printing at the console is informative.

#' @title Display summary for LeakFit objects
#' @description Prints a brief one-screen summary of a LeakFit, including
#'   task and outcome, fold count and status (successful, skipped, failed),
#'   and the headline cross-validated metric. Use \code{summary()} for the
#'   full per-fold diagnostic report.
#' @param object A \code{LeakFit} object.
#' @return No return value, called for side effects (prints a brief summary
#'   to the console). Returns \code{object} invisibly.
#' @examples
#' set.seed(1)
#' df <- data.frame(
#'   subject = rep(1:6, each = 2),
#'   outcome = rbinom(12, 1, 0.5),
#'   x1 = rnorm(12),
#'   x2 = rnorm(12)
#' )
#' splits <- make_split_plan(df, outcome = "outcome",
#'                       mode = "subject_grouped", group = "subject", v = 3,
#'                       progress = FALSE)
#' custom <- list(
#'   glm = list(
#'     fit = function(x, y, task, weights, ...) {
#'       stats::glm(y ~ ., data = as.data.frame(x),
#'                  family = stats::binomial(), weights = weights)
#'     },
#'     predict = function(object, newdata, task, ...) {
#'       as.numeric(stats::predict(object,
#'                                 newdata = as.data.frame(newdata),
#'                                 type = "response"))
#'     }
#'   )
#' )
#' fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                     learner = "glm", custom_learners = custom,
#'                     metrics = "auc", refit = FALSE, seed = 1)
#' show(fit)
#' @importMethodsFrom methods show
#' @export
setMethod("show", "LeakFit", function(object) {
  n_folds <- length(object@splits@indices)
  info <- object@info
  fold_status_str <- ""
  if (is.data.frame(info$fold_status) && nrow(info$fold_status) > 0L) {
    ok <- sum(info$fold_status$status == "success", na.rm = TRUE)
    sk <- sum(info$fold_status$status == "skipped", na.rm = TRUE)
    fl <- sum(info$fold_status$status == "failed",  na.rm = TRUE)
    fold_status_str <- sprintf("  Fold status:    %d success, %d skipped, %d failed\n",
                               ok, sk, fl)
  }
  learner_str <- if (length(object@learners) && !is.null(names(object@learners))) {
    paste(unique(names(object@learners)), collapse = ", ")
  } else if (nrow(object@metrics) > 0L && "learner" %in% names(object@metrics)) {
    paste(unique(object@metrics$learner), collapse = ", ")
  } else {
    "(unspecified)"
  }
  cat("A LeakFit object\n")
  cat(sprintf("  Task:           %s\n", object@task))
  cat(sprintf("  Outcome:        %s\n", object@outcome))
  cat(sprintf("  Learners:       %s\n", learner_str))
  cat(sprintf("  Folds:          %d  (mode = %s)\n",
              n_folds, object@splits@mode))
  cat(fold_status_str)
  cat("Use summary(<obj>) for the full diagnostic report.\n")
  invisible(object)
})

# Platform-safe symbols for console output.
.bio_symbol <- function(name) {
  use_unicode <- getOption("bioLeak.use_unicode", NA)
  if (is.na(use_unicode)) {
    utf8_ok <- FALSE
    if (requireNamespace("cli", quietly = TRUE)) {
      utf8_ok <- isTRUE(cli::is_utf8_output())
    } else {
      loc <- tryCatch(l10n_info(), error = function(e) NULL)
      if (!is.null(loc) && isTRUE(loc[["UTF-8"]])) utf8_ok <- TRUE
    }
    if (.Platform$OS.type == "windows") utf8_ok <- FALSE
    use_unicode <- utf8_ok
  } else {
    use_unicode <- isTRUE(use_unicode)
  }

  if (!use_unicode) {
    return(switch(name,
      pm = "+/-",
      chi_sq = "Chi^2",
      ge = ">=",
      check = "OK:",
      warn = "WARNING:",
      name
    ))
  }

  switch(name,
    pm = "\u00b1",
    chi_sq = "\u03c7\u00b2",
    ge = "\u2265",
    check = "\u2713",
    warn = "\u26a0",
    name
  )
}

#' S4 Classes for bioLeak Pipeline
#'
#' These classes capture splits, model fits, and audit diagnostics produced by
#' \code{make_split_plan()}, \code{fit_resample()}, and \code{audit_leakage()}.
#'
#' @slot mode Splitting mode. One of "subject_grouped", "batch_blocked",
#'   "study_loocv", "time_series", or "combined".
#' @slot indices List of resampling descriptors (train/test indices when available)
#' @slot info (LeakSplits) Metadata associated with the split plan (mode, coldata, hash, etc.)
#' @return An S4 object of the respective class.
#' @seealso [make_split_plan()], [fit_resample()], [audit_leakage()]
#' @rdname LeakClasses
#' @exportClass LeakSplits
setClass("LeakSplits",
         slots = c(mode = "character",
                   indices = "list",
                   info = "list"))

#' @rdname LeakClasses
#' @slot splits A [`LeakSplits`] object used for resampling
#' @slot metrics Model performance metrics per resample
#' @slot metric_summary Summary of metrics across resamples
#' @slot audit Audit information per resample
#' @slot predictions List of prediction objects
#' @slot preprocess Preprocessing steps used during fitting
#' @slot learners Learner definitions used in the pipeline
#' @slot outcome Outcome variable name
#' @slot task Modeling task name
#' @slot feature_names Feature names included in the model
#' @slot info (LeakFit) Metadata about the model fit (sample IDs, timings, provenance, etc.)
#' @seealso [fit_resample()]
#' @exportClass LeakFit
setClass("LeakFit",
         slots = c(
           splits = "LeakSplits",
           metrics = "data.frame",
           metric_summary = "data.frame",
           audit = "data.frame",
           predictions = "list",
           preprocess = "list",
           learners = "list",
           outcome = "character",
           task = "character",
           feature_names = "character",
           info = "list"
         ))

#' @rdname LeakClasses
#' @slot fit A [`LeakFit`] object used to generate the audit
#' @slot permutation_gap Data frame summarising permutation gaps
#' @slot perm_values Numeric vector of permutation-based scores
#' @slot batch_assoc Data frame of batch associations
#' @slot target_assoc Data frame of feature-wise outcome associations
#' @slot duplicates Data frame detailing duplicate records
#' @slot trail List capturing audit trail information
#' @slot info (LeakAudit) Metadata about the audit (mechanism summary, settings, provenance, etc.)
#' @seealso [audit_leakage()], [audit_report()]
#' @exportClass LeakAudit
setClass("LeakAudit",
         slots = c(
           fit = "LeakFit",
           permutation_gap = "data.frame",
           perm_values = "numeric",
           batch_assoc = "data.frame",
           target_assoc = "data.frame",
           duplicates = "data.frame",
           trail = "list",
           info = "list"))

setValidity("LeakSplits", function(object) {
  if (!length(object@indices)) return("indices list cannot be empty")
  TRUE
})

setValidity("LeakFit", function(object) {
  if (!inherits(object@splits, "LeakSplits")) return("splits must be a LeakSplits object")
  TRUE
})

setValidity("LeakAudit", function(object) {
  if (!inherits(object@fit, "LeakFit")) return("fit must be a LeakFit object")
  TRUE
})

LeakSplits <- function(mode = NA_character_, indices = list(), info = list()) {
  methods::new("LeakSplits", mode = mode, indices = indices, info = info)
}

LeakFit <- function(...) methods::new("LeakFit", ...)

LeakAudit <- function(...) methods::new("LeakAudit", ...)

#' @rdname LeakClasses
#' @slot metric Performance metric compared between pipelines
#' @slot exchangeability Exchangeability assumption used for the sign-flip test
#' @slot tier Inference tier label based on effective number of repeats
#' @slot strict Whether strict mode was requested
#' @slot R_eff Effective number of paired repeats available for inference
#' @slot delta_lsi Huber-robust point estimate of repeat-level metric difference
#' @slot delta_lsi_ci BCa 95\% CI for delta_lsi (NA when R_eff < 10)
#' @slot delta_metric Arithmetic mean of repeat-level metric differences
#' @slot delta_metric_ci BCa 95\% CI for delta_metric (NA when R_eff < 10)
#' @slot p_value Sign-flip randomization test p-value (NA when R_eff < 5 or unpaired)
#' @slot inference_ok TRUE when tier A (R_eff >= 20, paired, finite p and CI)
#' @slot folds_naive Per-fold data frame for the naive pipeline
#' @slot folds_guarded Per-fold data frame for the guarded pipeline
#' @slot repeats_naive Per-repeat aggregate data frame for the naive pipeline
#' @slot repeats_guarded Per-repeat aggregate data frame for the guarded pipeline
#' @slot info (LeakDeltaLSI) Metadata including R_naive, R_guarded, paired status, and block details
#' @seealso [delta_lsi()]
#' @exportClass LeakDeltaLSI
setClass("LeakDeltaLSI",
  slots = c(
    metric          = "character",
    exchangeability = "character",
    tier            = "character",
    strict          = "logical",
    R_eff           = "integer",
    delta_lsi       = "numeric",
    delta_lsi_ci    = "numeric",
    delta_metric    = "numeric",
    delta_metric_ci = "numeric",
    p_value         = "numeric",
    inference_ok    = "logical",
    folds_naive     = "data.frame",
    folds_guarded   = "data.frame",
    repeats_naive   = "data.frame",
    repeats_guarded = "data.frame",
    info            = "list"
  ),
  prototype = list(
    metric          = character(0),
    exchangeability = character(0),
    tier            = "D_insufficient",
    strict          = FALSE,
    R_eff           = 0L,
    delta_lsi       = NA_real_,
    delta_lsi_ci    = c(NA_real_, NA_real_),
    delta_metric    = NA_real_,
    delta_metric_ci = c(NA_real_, NA_real_),
    p_value         = NA_real_,
    inference_ok    = FALSE,
    folds_naive     = data.frame(),
    folds_guarded   = data.frame(),
    repeats_naive   = data.frame(),
    repeats_guarded = data.frame(),
    info            = list()
  )
)

LeakDeltaLSI <- function(...) methods::new("LeakDeltaLSI", ...)

# ---- LeakAudit: show() method ------------------------------------------
# Brief auto-print representation. Detailed reporting is produced by
# summary(). Added in 0.3.7 (Comment 8 response): standard R idiom is
# that every public S4 class should have a show() method so that
# auto-printing at the console is informative.

#' @title Display summary for LeakAudit objects
#' @description Prints a brief one-screen summary of a LeakAudit, including
#'   task and outcome, the permutation-gap statistic, and counts of
#'   batch-association rows, target-leakage features, and duplicate pairs.
#'   Use \code{summary()} for the full diagnostic report.
#' @param object A \code{LeakAudit} object.
#' @return No return value, called for side effects (prints a brief summary
#'   to the console). Returns \code{object} invisibly.
#' @examples
#' set.seed(1)
#' df <- data.frame(
#'   subject = rep(1:6, each = 2),
#'   outcome = rbinom(12, 1, 0.5),
#'   x1 = rnorm(12),
#'   x2 = rnorm(12)
#' )
#' splits <- make_split_plan(df, outcome = "outcome",
#'                       mode = "subject_grouped", group = "subject", v = 3,
#'                       progress = FALSE)
#' custom <- list(
#'   glm = list(
#'     fit = function(x, y, task, weights, ...) {
#'       stats::glm(y ~ ., data = as.data.frame(x),
#'                  family = stats::binomial(), weights = weights)
#'     },
#'     predict = function(object, newdata, task, ...) {
#'       as.numeric(stats::predict(object,
#'                                 newdata = as.data.frame(newdata),
#'                                 type = "response"))
#'     }
#'   )
#' )
#' fit <- fit_resample(df, outcome = "outcome", splits = splits,
#'                     learner = "glm", custom_learners = custom,
#'                     metrics = "auc", refit = FALSE, seed = 1)
#' aud <- audit_leakage(fit, metric = "auc", B = 10,
#'                      X_ref = df[, c("x1", "x2")])
#' show(aud)
#' @importMethodsFrom methods show
#' @export
setMethod("show", "LeakAudit", function(object) {
  perm_str <- if (is.data.frame(object@permutation_gap) &&
                  nrow(object@permutation_gap) > 0L &&
                  "metric_obs" %in% names(object@permutation_gap)) {
    obs <- as.numeric(object@permutation_gap$metric_obs[1L])
    gap <- if ("gap" %in% names(object@permutation_gap))
      as.numeric(object@permutation_gap$gap[1L]) else NA_real_
    pv <- if ("p_value" %in% names(object@permutation_gap))
      as.numeric(object@permutation_gap$p_value[1L]) else NA_real_
    sprintf("metric=%.3f, gap=%.3f, p=%.3f", obs, gap, pv)
  } else {
    "(not computed)"
  }
  task_str <- if (length(object@fit@task) && nzchar(object@fit@task))
    object@fit@task else "(unknown)"
  outcome_str <- if (length(object@fit@outcome) && nzchar(object@fit@outcome))
    object@fit@outcome else "(unknown)"

  cat("A LeakAudit object\n")
  cat(sprintf("  Task:               %s\n", task_str))
  cat(sprintf("  Outcome:            %s\n", outcome_str))
  cat(sprintf("  Permutation-gap:    %s\n", perm_str))
  cat(sprintf("  Batch association:  %d row(s)\n",
              if (is.data.frame(object@batch_assoc)) nrow(object@batch_assoc) else 0L))
  cat(sprintf("  Target leakage:     %d feature(s)\n",
              if (is.data.frame(object@target_assoc)) nrow(object@target_assoc) else 0L))
  cat(sprintf("  Duplicate pairs:    %d\n",
              if (is.data.frame(object@duplicates)) nrow(object@duplicates) else 0L))
  cat("Use summary(<obj>) for the full diagnostic report.\n")
  invisible(object)
})

test_that("LeakSplits validity and show method work", {
  expect_error(bioLeak:::LeakSplits(mode = "x", indices = list(), info = list()),
               "indices list cannot be empty")

  df <- make_class_df(8)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 2, seed = 1)
  out <- capture.output(show(splits))
  expect_true(any(grepl("LeakSplits object", out)))
})

test_that("LeakFit and LeakAudit validity checks enforce class constraints", {
  splits <- make_split_plan_quiet(make_class_df(8), outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 2, seed = 1)
  expect_error(bioLeak:::LeakFit(splits = list(), metrics = data.frame(),
                                 metric_summary = data.frame(), audit = data.frame(),
                                 predictions = list(), preprocess = list(), learners = list(),
                                 outcome = "outcome", task = "binomial",
                                 feature_names = character(), info = list()),
               "slot \"splits\"")

  fit <- bioLeak:::LeakFit(splits = splits, metrics = data.frame(),
                           metric_summary = data.frame(), audit = data.frame(),
                           predictions = list(), preprocess = list(), learners = list(),
                           outcome = "outcome", task = "binomial",
                           feature_names = character(), info = list())
  expect_error(bioLeak:::LeakAudit(fit = list(), permutation_gap = data.frame(),
                                   perm_values = numeric(), batch_assoc = data.frame(),
                                   target_assoc = data.frame(), duplicates = data.frame(),
                                   trail = list(), info = list()),
               "slot \"fit\"")
})

# test_that("summary methods emit expected headers", {
#   df <- make_class_df(10)
#   splits <- make_split_plan_quiet(df, outcome = "outcome",
#                               mode = "subject_grouped", group = "subject",
#                               v = 2, seed = 1)
#   custom <- make_custom_learners()
#   fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
#                             learner = "glm", custom_learners = custom,
#                             metrics = "auc", refit = FALSE, seed = 1)
#   audit <- audit_leakage(fit, metric = "auc", B = 3, perm_stratify = FALSE)
#
#   fit_out <- capture.output(summary(fit))
#   expect_true(any(grepl("Model Fit Summary", fit_out)))
#
#   aud_out <- capture.output(summary(audit))
#   expect_true(any(grepl("Leakage Audit Summary", aud_out)))
# })

test_that("GuardFit print and summary methods work", {
  X <- data.frame(a = c(1, 2, NA), b = c(3, 4, 5))
  fit <- guard_fit(X, y = c(1, 2, 3),
                              steps = list(impute = list(method = "median")),
                              task = "gaussian")
  out <- capture.output(print(fit))
  expect_true(any(grepl("Guarded preprocessing pipeline", out)))

  sm <- summary(fit)
  expect_true(inherits(sm, "summary.GuardFit"))
  out2 <- capture.output(print(sm))
  expect_true(any(grepl("GuardFit summary", out2)))
})

test_that("fit_resample validates custom learner specifications", {
  df <- make_class_df(10)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 2, seed = 1)

  expect_error(fit_resample_quiet(df, outcome = "outcome", splits = splits,
                                  learner = "glm", custom_learners = "bad"),
               "custom_learners must be a named list")

  expect_error(fit_resample_quiet(df, outcome = "outcome", splits = splits,
                                  learner = "glm", custom_learners = list(list())),
               "custom_learners must be a named list")

  bad_custom <- list(glm = list(fit = function(...) NULL))
  expect_error(fit_resample_quiet(df, outcome = "outcome", splits = splits,
                                  learner = "glm", custom_learners = bad_custom),
               "fit.*predict")
})

test_that("fit_resample validates class weights and positive class", {
  df <- make_class_df(12)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 3, seed = 1)
  custom <- make_custom_learners()

  expect_error(fit_resample_quiet(df, outcome = "outcome", splits = splits,
                                  learner = "glm", custom_learners = custom,
                                  class_weights = c(a = 1),
                                  metrics = "auc", refit = FALSE),
               "missing levels")

  expect_error(fit_resample_quiet(df, outcome = "outcome", splits = splits,
                                  learner = "glm", custom_learners = custom,
                                  positive_class = c(0, 1),
                                  metrics = "auc", refit = FALSE),
               "single value")
})

test_that("fit_resample supports class weights with parsnip learners", {
  skip_if_not_installed("parsnip")
  df <- make_class_df(12)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 3, seed = 1)
  spec <- parsnip::logistic_reg() |> parsnip::set_engine("glm")
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = spec,
                            class_weights = c("0" = 1, "1" = 1),
                            metrics = "auc", refit = FALSE)
  expect_true(nrow(fit@metrics) > 0)
})

test_that("fit_resample drops invalid metrics with warnings", {
  df <- make_class_df(10)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 2, seed = 1)
  custom <- make_custom_learners()

  fit <- expect_warning_match(
    fit_resample_quiet(df, outcome = "outcome", splits = splits,
                       learner = "glm", custom_learners = custom,
                       metrics = c("auc", "rmse"), refit = FALSE),
    "Dropping metrics"
  )
  expect_true(nrow(fit@metrics) > 0)
})

test_that("fit_resample summarizes metrics with all-NA columns", {
  df <- make_class_df(10)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 2, seed = 1)
  custom <- make_custom_learners()
  na_metric <- function(y, pred) NA_real_

  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "glm", custom_learners = custom,
                            metrics = list(auc = "auc", na_metric = na_metric),
                            refit = FALSE)
  expect_true(nrow(fit@metric_summary) > 0)
  expect_true("na_metric_mean" %in% colnames(fit@metric_summary))
})

test_that("fit_resample handles gaussian tasks and ignores classification options", {
  df <- make_regression_df(12)
  splits <- make_split_plan_quiet(df, outcome = "y",
                              mode = "subject_grouped", group = "subject",
                              v = 3, seed = 1)
  custom <- make_custom_learners()

  fit <- expect_warning_match(
    fit_resample_quiet(df, outcome = "y", splits = splits,
                       learner = "glm", custom_learners = custom,
                       class_weights = c(a = 1, b = 2),
                       positive_class = "a",
                       metrics = c("rmse", "cindex"), refit = FALSE),
    "ignored",
    all = TRUE
  )
  expect_equal(fit@task, "gaussian")
})

# test_that("fit_resample errors on unsupported outcomes", {
#   df <- make_class_df(10)
#   df$outcome <- factor(c("a", "b", "c", "a", "b", "c", "a", "b", "c", "a"))
#   splits <- make_split_plan_quiet(df, outcome = "outcome",
#                               mode = "subject_grouped", group = "subject",
#                               v = 2, seed = 1)
#   custom <- make_custom_learners()
#   expect_error(fit_resample_quiet(df, outcome = "outcome", splits = splits,
#                                   learner = "glm", custom_learners = custom),
#                "No successful folds were completed. Check learner and preprocessing settings.")
# })

test_that("fit_resample respects positive_class releveling", {
  df <- make_class_df(12)
  df$outcome <- factor(ifelse(df$outcome == 1, "yes", "no"), levels = c("no", "yes"))
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 3, seed = 1)
  custom <- make_custom_learners()
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "glm", custom_learners = custom,
                            positive_class = "yes", metrics = "auc",
                            refit = FALSE)
  expect_equal(fit@info$positive_class, "yes")
})

test_that("fit_resample applies classification_threshold for binomial tasks", {
  df <- make_class_df(12)
  df$x1 <- seq(0.05, 0.95, length.out = nrow(df))
  custom <- list(
    fixed_prob = list(
      fit = function(x, y, task, weights, ...) list(),
      predict = function(object, newdata, task, ...) as.numeric(newdata$x1)
    )
  )
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 3, seed = 2)

  fit_low <- fit_resample_quiet(
    df, outcome = "outcome", splits = splits,
    learner = "fixed_prob", custom_learners = custom,
    metrics = "accuracy", classification_threshold = 0.2,
    refit = FALSE
  )
  fit_high <- fit_resample_quiet(
    df, outcome = "outcome", splits = splits,
    learner = "fixed_prob", custom_learners = custom,
    metrics = "accuracy", classification_threshold = 0.8,
    refit = FALSE
  )

  pred_low <- do.call(rbind, fit_low@predictions)
  pred_high <- do.call(rbind, fit_high@predictions)
  key_low <- order(pred_low$fold, pred_low$id)
  key_high <- order(pred_high$fold, pred_high$id)
  expect_false(identical(as.character(pred_low$pred_class[key_low]),
                         as.character(pred_high$pred_class[key_high])))
  expect_false(isTRUE(all.equal(fit_low@metrics$accuracy, fit_high@metrics$accuracy)))
  expect_equal(fit_low@info$classification_threshold, 0.2)
  expect_equal(fit_high@info$classification_threshold, 0.8)

  expect_error(
    fit_resample_quiet(df, outcome = "outcome", splits = splits,
                       learner = "fixed_prob", custom_learners = custom,
                       classification_threshold = 1.2, refit = FALSE),
    "classification_threshold must be"
  )
})

test_that("fit_resample surfaces custom learner prediction length errors", {
  df <- make_class_df(10)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 2, seed = 1)
  custom <- list(
    bad = list(
      fit = function(x, y, task, weights, ...) suppressWarnings(stats::glm(
        y ~ ., data = data.frame(y = y, x), family = stats::binomial()
      )),
      predict = function(object, newdata, task, ...) rep(0.5, nrow(newdata) + 1)
    )
  )
  expect_warning_match(
    expect_error(fit_resample_quiet(df, outcome = "outcome", splits = splits,
                                    learner = "bad", custom_learners = custom,
                                    metrics = "auc", refit = FALSE),
                 "No successful folds were completed"),
    "Custom learner 'bad' returned"
  )
})

test_that("fit_resample warns when a fold lacks both classes", {
  df <- make_class_df(8)
  df$outcome <- c(rep(0, 4), rep(1, 4))
  indices <- list(
    list(train = 1:4, test = 5:6, fold = 1, repeat_id = 1),
    list(train = c(1:3, 5:8), test = 4, fold = 2, repeat_id = 1)
  )
  splits <- bioLeak:::LeakSplits(mode = "custom", indices = indices,
                                 info = list(outcome = "outcome", coldata = df))
  custom <- make_custom_learners()
  fit <- expect_warning_match(
    fit_resample_quiet(df, outcome = "outcome", splits = splits,
                       learner = "glm", custom_learners = custom,
                       metrics = "accuracy", refit = FALSE),
    "only one class"
  )
  expect_true(nrow(fit@metrics) > 0)
  expect_true(is.data.frame(fit@info$fold_status))
  expect_equal(nrow(fit@info$fold_status), length(splits@indices))
  expect_true(any(fit@info$fold_status$status == "success"))
  expect_true(any(fit@info$fold_status$status == "skipped"))
})

test_that("fit_resample errors for compact time_series without time metadata", {
  df <- make_class_df(10)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "time_series", time = "time",
                              v = 2, seed = 1, compact = TRUE)
  splits@info$time <- NULL
  custom <- make_custom_learners()
  expect_error(fit_resample_quiet(df, outcome = "outcome", splits = splits,
                                  learner = "glm", custom_learners = custom,
                                  metrics = "auc", refit = FALSE),
               "time_series compact splits")
})

test_that("fit_resample handles compact splits with repeats", {
  df <- make_class_df(30)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped",
                              group = "subject",
                              v = 3, repeats = 2,
                              compact = TRUE, seed = 1)
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "glm", custom_learners = make_custom_learners(),
                            metrics = "auc", refit = FALSE, seed = 1)
  pred_df <- do.call(rbind, fit@predictions)
  fold_counts <- table(pred_df$fold)
  expect_equal(length(fold_counts), length(splits@indices))
  expect_true(all(fold_counts > 0))
})

test_that("fit_resample resolves compact time_series splits with purge and embargo", {
  df <- make_class_df(12)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "time_series", time = "time",
                              v = 3, compact = TRUE, seed = 1,
                              purge = 2, embargo = 5)
  fit <- fit_resample_quiet(df, outcome = "outcome", splits = splits,
                            learner = "glm", custom_learners = make_custom_learners(),
                            metrics = "accuracy", refit = FALSE, seed = 1)

  expect_equal(fit@audit$n_train, c(3, 7))
  expect_equal(fit@splits@info$purge, 2)
  expect_equal(fit@splits@info$embargo, 5)
})

test_that("fit_resample supports parsnip learners when available", {
  skip_if_not_installed("parsnip")
  df <- make_class_df(12)
  splits <- make_split_plan_quiet(df, outcome = "outcome",
                              mode = "subject_grouped", group = "subject",
                              v = 3, seed = 1)
  spec <- parsnip::logistic_reg() |>
    parsnip::set_engine("glm")
  fit <- expect_warning_match(
    fit_resample_quiet(df, outcome = "outcome", splits = splits,
                       learner = spec, custom_learners = make_custom_learners(),
                       learner_args = list(alpha = 1),
                       metrics = "auc", refit = FALSE, seed = 1),
    "ignored"
  )
  expect_true(nrow(fit@metrics) > 0)
  expect_true(any(nzchar(fit@metrics$learner)))
})

test_that("audit_leakage warns when observed metric is non-finite", {
  indices <- list(list(train = 1:2, test = 3:4, fold = 1, repeat_id = 1))
  splits <- bioLeak:::LeakSplits(mode = "custom", indices = indices,
                                 info = list(outcome = "outcome"))
  preds <- list(data.frame(
    id = 1:2,
    truth = c(1, 2),
    pred = c(NA_real_, 1),
    fold = 1,
    learner = "glm",
    stringsAsFactors = FALSE
  ))
  fit <- bioLeak:::LeakFit(
    splits = splits,
    metrics = data.frame(fold = 1, learner = "glm", rmse = NA_real_),
    metric_summary = data.frame(),
    audit = data.frame(),
    predictions = preds,
    preprocess = list(),
    learners = list(),
    outcome = "outcome",
    task = "gaussian",
    feature_names = character(),
    info = list()
  )
  audit <- expect_warning_match(
    audit_leakage(fit, metric = "rmse", B = 3),
    "Observed metric is NA"
  )
  expect_true(is.na(audit@permutation_gap$metric_obs[1]))
})

## ---------------------------------------------------------------------
## test-show-methods.R
## ---------------------------------------------------------------------
## Tests for the 0.3.7 addition of show()/print() methods to the public
## result classes. These tests assert that:
##   (1) `methods(class = "LeakFit")`   includes "show",
##   (2) `methods(class = "LeakAudit")` includes "show",
##   (3) `methods(class = "LeakTune")`  includes "print",
##       all addressing the editor's Comment 8 complaint that these
##       classes were missing print/show methods.
## We also assert that auto-print is non-empty and contains the class
## name and at least one diagnostic field.
## ---------------------------------------------------------------------

# ---- Build minimal fixtures by direct slot construction --------------

make_leakfit_fixture <- function() {
  splits <- make_split_plan_quiet(
    make_class_df(8),
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 2,
    seed = 1
  )
  metrics_df <- data.frame(
    learner = "stub",
    fold = 1:2,
    repeat_id = 1L,
    auc = c(0.71, 0.65),
    stringsAsFactors = FALSE
  )
  bioLeak:::LeakFit(
    splits         = splits,
    metrics        = metrics_df,
    metric_summary = data.frame(),
    audit          = data.frame(),
    predictions    = list(),
    preprocess     = list(),
    learners       = list(),
    outcome        = "outcome",
    task           = "binomial",
    feature_names  = character(),
    info           = list()
  )
}

make_leakaudit_fixture <- function() {
  fit <- make_leakfit_fixture()
  perm_gap <- data.frame(
    mechanism_class = "non_random_signal",
    repeat_id = 1L,
    metric_obs = 0.68,
    perm_mean = 0.50,
    perm_sd = 0.05,
    gap = 0.18,
    z = 3.6,
    p_value = 0.02,
    n_perm = 50L,
    stringsAsFactors = FALSE
  )
  bioLeak:::LeakAudit(
    fit             = fit,
    permutation_gap = perm_gap,
    perm_values     = c(0.48, 0.51, 0.55),
    batch_assoc     = data.frame(),
    target_assoc    = data.frame(),
    duplicates      = data.frame(),
    trail           = list(),
    info            = list()
  )
}

make_leaktune_fixture <- function() {
  ## LeakTune is an S3-classed list with $info and $outer_fits;
  ## construct a minimal one that satisfies print.LeakTune's accesses.
  fit <- make_leakfit_fixture()
  tune_obj <- list(
    outer_fits = list(fit, fit, fit),
    outer_summary = data.frame(),
    info = list(
      grid = 5L,
      refit_metric = "roc_auc",
      selection_rule = "one_std_err",
      refit = TRUE
    )
  )
  class(tune_obj) <- "LeakTune"
  tune_obj
}


# ---- methods(class = ...) registration tests -------------------------

test_that("methods(class = 'LeakFit') includes show (Comment 8 check)", {
  cls_methods <- as.character(methods(class = "LeakFit"))
  expect_true("show,LeakFit-method" %in% cls_methods ||
              "show.LeakFit"        %in% cls_methods)
  expect_true("summary.LeakFit"     %in% cls_methods)  # pre-existing
})

test_that("methods(class = 'LeakAudit') includes show (Comment 8 check)", {
  cls_methods <- as.character(methods(class = "LeakAudit"))
  expect_true("show,LeakAudit-method" %in% cls_methods ||
              "show.LeakAudit"        %in% cls_methods)
  expect_true("summary.LeakAudit"     %in% cls_methods)  # pre-existing
})

test_that("methods(class = 'LeakTune') includes print (Comment 8 check)", {
  cls_methods <- as.character(methods(class = "LeakTune"))
  expect_true("print.LeakTune"   %in% cls_methods)
  expect_true("summary.LeakTune" %in% cls_methods)       # pre-existing
})


# ---- Auto-print correctness (S4 show methods) -----------------------

test_that("show(LeakFit) auto-prints the class name and key fields", {
  fit <- make_leakfit_fixture()
  txt <- paste(capture.output(show(fit)), collapse = "\n")
  expect_match(txt, "A LeakFit object")
  expect_match(txt, "Task:\\s+binomial")
  expect_match(txt, "Outcome:\\s+outcome")
  expect_match(txt, "Folds:\\s+2\\b")
  ## Hint about summary() should be present.
  expect_match(txt, "summary\\(<obj>\\)")
})

test_that("show(LeakAudit) auto-prints the class name and key fields", {
  audit <- make_leakaudit_fixture()
  txt <- paste(capture.output(show(audit)), collapse = "\n")
  expect_match(txt, "A LeakAudit object")
  expect_match(txt, "Permutation-gap:\\s+metric=0\\.680")
  expect_match(txt, "Batch association:\\s+0 row\\(s\\)")
  expect_match(txt, "Target leakage:\\s+0 feature\\(s\\)")
  expect_match(txt, "Duplicate pairs:\\s+0")
  expect_match(txt, "summary\\(<obj>\\)")
})

test_that("print(LeakTune) auto-prints the class name and key fields", {
  tn <- make_leaktune_fixture()
  txt <- paste(capture.output(print(tn)), collapse = "\n")
  expect_match(txt, "A LeakTune object")
  expect_match(txt, "Outer folds:\\s+3 successful / 3 total")
  expect_match(txt, "Tuning grid:\\s+5 candidates")
  expect_match(txt, "Selection rule:\\s+one_std_err")
  expect_match(txt, "Refit:\\s+yes")
  expect_match(txt, "summary\\(<obj>\\)")
})


# ---- Auto-print is the default at the console (no explicit show/print) ----

test_that("Auto-printing at the console invokes show()/print()", {
  fit  <- make_leakfit_fixture()
  audit <- make_leakaudit_fixture()
  tn   <- make_leaktune_fixture()
  ## When R auto-prints an object, it calls show() (S4) or print() (S3).
  expect_match(paste(capture.output(fit),  collapse = "\n"), "A LeakFit object")
  expect_match(paste(capture.output(audit),collapse = "\n"), "A LeakAudit object")
  expect_match(paste(capture.output(tn),   collapse = "\n"), "A LeakTune object")
})


# ---- Each method returns its argument invisibly --------------------

test_that("show()/print() methods return their argument invisibly", {
  fit  <- make_leakfit_fixture()
  audit <- make_leakaudit_fixture()
  tn   <- make_leaktune_fixture()
  capture.output({
    expect_identical(show(fit), fit)
    expect_identical(show(audit), audit)
    expect_identical(print(tn), tn)
  })
})

## ---------------------------------------------------------------------
## test-vignette-suggests.R
## ---------------------------------------------------------------------
## Tests for the 0.3.7 cleanup of Suggests-package handling in the
## vignette. The editor's Comment 9 noted that the vignette mixed
## `requireNamespace("recipes", quietly = TRUE)` (defensive) with
## `library(recipes)` (will hard-error if recipes is not installed),
## defeating the defensive check. After the fix, every code chunk
## that calls `library(<suggests_pkg>)` in eval = TRUE mode must be
## gated by a chunk option `eval = requireNamespace(...)` or by an
## inline `if (requireNamespace(...))` guard.
## ---------------------------------------------------------------------

vignette_lines <- function() {
  paths <- c(
    file.path("..", "..", "vignettes", "bioLeak-intro.Rmd"),
    file.path("vignettes", "bioLeak-intro.Rmd"),
    system.file("doc", "bioLeak-intro.Rmd", package = "bioLeak"),
    system.file("doc", "bioLeak-intro.R", package = "bioLeak")
  )
  paths <- paths[nzchar(paths) & file.exists(paths)]
  if (!length(paths)) skip("vignette source not found in this build")
  readLines(paths[1])
}

## Packages listed in DESCRIPTION's Suggests field. These are the
## packages whose `library()` calls must be guarded.
suggests_pkgs <- function() {
  desc_path <- c(
    file.path("..", "..", "DESCRIPTION"),
    "DESCRIPTION",
    system.file("DESCRIPTION", package = "bioLeak")
  )
  desc_path <- desc_path[nzchar(desc_path) & file.exists(desc_path)]
  if (!length(desc_path)) skip("DESCRIPTION not found in this build")
  d <- read.dcf(desc_path[1])
  if (!"Suggests" %in% colnames(d)) return(character(0))
  raw <- d[1, "Suggests"]
  pkgs <- trimws(strsplit(raw, ",")[[1]])
  ## Strip version constraints like " (>= 3.0.0)"
  pkgs <- sub(" *\\(.*\\)$", "", pkgs)
  pkgs[nzchar(pkgs)]
}

test_that("Vignette has no unguarded `library(<suggests_pkg>)` calls", {
  lines <- vignette_lines()
  pkgs  <- suggests_pkgs()
  if (!length(pkgs)) skip("no Suggests packages declared")

  ## Walk the vignette, tracking which chunk each line is in and
  ## whether that chunk is gated by an `eval = ...` option that
  ## references requireNamespace() or by an explicit eval = FALSE.
  in_chunk <- FALSE
  chunk_header <- ""
  offending <- character(0)

  for (i in seq_along(lines)) {
    L <- lines[i]
    if (grepl("^```\\{r", L)) {
      in_chunk <- TRUE
      chunk_header <- L
      next
    }
    if (grepl("^```\\s*$", L)) {
      in_chunk <- FALSE
      chunk_header <- ""
      next
    }
    if (!in_chunk) next

    ## Match `library(<pkg>)` (top-level call, not commented).
    m <- regmatches(L, regexec("^\\s*library\\(([A-Za-z0-9._]+)\\)", L))[[1]]
    if (length(m) == 2L) {
      pkg <- m[2]
      if (!pkg %in% pkgs) next   # Imports/Depends → fine
      ## Suggests pkg → check if the chunk header contains a guard
      header_safe <-
        grepl("eval\\s*=\\s*FALSE", chunk_header) ||
        grepl(sprintf("requireNamespace\\(\\s*[\"']%s[\"']", pkg),
              chunk_header)
      if (!header_safe) {
        offending <- c(offending,
                       sprintf("line %d: %s   (chunk: %s)", i, L, chunk_header))
      }
    }
  }

  expect_length(offending, 0L)
})

test_that("Suggests-using chunks at the documented locations are properly gated", {
  lines <- vignette_lines()

  ## tidymodels-interop chunk header should carry an
  ## `eval = requireNamespace(...)` gate.
  idx <- grep("^```\\{r tidymodels-interop", lines)
  expect_length(idx, 1L)
  hdr <- lines[idx]
  expect_match(hdr, "requireNamespace\\(\\s*['\"]recipes['\"]")
  expect_match(hdr, "requireNamespace\\(\\s*['\"]yardstick['\"]")

  ## parallel-setup chunk should remain eval = FALSE (documentation-only).
  idx <- grep("^```\\{r parallel-setup", lines)
  expect_length(idx, 1L)
  expect_match(lines[idx], "eval\\s*=\\s*FALSE")
})

test_that("Defensive requireNamespace guards remain in place for all gated chunks", {
  lines <- vignette_lines()
  ## Confirm no chunks regressed to bare library() inside an unguarded
  ## chunk (the editor's specific complaint pattern).
  ## Count: at least one inline `requireNamespace(<pkg>)` guard for every
  ## Suggests package that the vignette uses.
  used_with_guard <- regmatches(
    lines,
    gregexpr("requireNamespace\\(\\s*['\"]([^'\"]+)['\"]", lines)
  )
  flat <- unique(unlist(used_with_guard))
  pkg_names <- gsub("requireNamespace\\(\\s*['\"]([^'\"]+)['\"]", "\\1", flat)
  pkg_names <- pkg_names[nzchar(pkg_names)]

  ## At minimum, recipes/yardstick guards must exist somewhere in the
  ## vignette (either in chunk options or inline `if`).
  expect_true("recipes"   %in% pkg_names)
  expect_true("yardstick" %in% pkg_names)
})

## ---------------------------------------------------------------------
## test-guard-fit-rename.R
## ---------------------------------------------------------------------
## Tests for the 0.3.7 rename of `.guard_fit` -> `guard_fit` and
## `.guard_ensure_levels` -> `guard_ensure_levels`. These tests ensure
## the editor's Comment 6 fix stays in place: no public bioLeak function
## should have a leading-dot name.
## ---------------------------------------------------------------------

test_that("guard_fit() (no leading dot) is exported and works", {
  x <- data.frame(a = c(1, 2, NA), b = c(3, 4, 5))
  fit <- guard_fit(
    x, y = c(1, 2, 3),
    steps = list(impute = list(method = "median")),
    task = "gaussian"
  )
  expect_s3_class(fit, "GuardFit")
  expect_true(is.function(fit$transform))
})

test_that("guard_ensure_levels() (no leading dot) is exported and works", {
  df <- data.frame(
    site = c("A", "B", "B"),
    status = c("yes", "no", "yes"),
    stringsAsFactors = FALSE
  )
  out <- guard_ensure_levels(df)
  expect_named(out, c("data", "levels"))
  expect_true(is.factor(out$data$site))
  expect_true(is.factor(out$data$status))
})

test_that("Old dot-prefixed names `.guard_fit` and `.guard_ensure_levels` are no longer exported", {
  exported <- getNamespaceExports("bioLeak")

  ## The dot-prefixed names should NOT appear in the export list.
  expect_false(".guard_fit"           %in% exported)
  expect_false(".guard_ensure_levels" %in% exported)

  ## The renamed functions SHOULD appear in the export list.
  expect_true("guard_fit"           %in% exported)
  expect_true("guard_ensure_levels" %in% exported)
})

test_that("help(package = 'bioLeak') no longer indexes any user-facing '.<name>' export", {
  ## The package-help index lists each export. We assert that no
  ## user-facing entry in the export list begins with a leading dot.
  ## Auto-generated S4 mangled names (".__C__<Class>", ".__T__<gen>:<pkg>")
  ## are excluded; those are R's own internal class/method registry
  ## and never appear in `help(package = "bioLeak")` output. The check
  ## targets only names that begin with `.` followed by an alphabetic
  ## character (the form the editor's Comment 6 flagged).
  exported <- getNamespaceExports("bioLeak")
  dot_prefixed_user <- grep("^\\.[a-zA-Z]", exported, value = TRUE)
  expect_length(dot_prefixed_user, 0L)
})

---
title: "bioLeak Audit Report"
output:
  html_document:
    toc: true
    toc_depth: 2
params:
  audit: !r NULL
---

```{r setup, include=FALSE}
if (is.null(params$audit) || !inherits(params$audit, "LeakAudit")) {
  stop("Parameter 'audit' must be a LeakAudit object.", call. = FALSE)
}

audit <- params$audit

empty_df <- function(msg) data.frame(note = msg, stringsAsFactors = FALSE)
safe_df <- function(x, msg = "No results available.") {
  if (is.null(x) || !is.data.frame(x) || !nrow(x)) return(empty_df(msg))
  x
}

task <- tryCatch(audit@fit@task, error = function(e) NA_character_)
outcome <- tryCatch(audit@fit@outcome, error = function(e) NA_character_)
metric_summary <- tryCatch(audit@fit@metric_summary, error = function(e) NULL)
perm_gap <- tryCatch(audit@permutation_gap, error = function(e) NULL)
batch_assoc <- tryCatch(audit@batch_assoc, error = function(e) NULL)
target_assoc <- tryCatch(audit@target_assoc, error = function(e) NULL)
duplicates <- tryCatch(audit@duplicates, error = function(e) NULL)
perm_values <- tryCatch(audit@perm_values, error = function(e) numeric(0))
fold_status <- tryCatch(audit@fit@info$fold_status, error = function(e) NULL)
provenance <- tryCatch(audit@fit@info$provenance, error = function(e) NULL)
bio_version <- tryCatch(as.character(utils::packageVersion("bioLeak")),
                        error = function(e) "unknown")
has_ggplot2 <- requireNamespace("ggplot2", quietly = TRUE)
```

# Overview

- **bioLeak version**: `r bio_version`
- **Generated**: `r format(Sys.time(), "%Y-%m-%d %H:%M:%S %Z")`
- **Task**: `r ifelse(is.na(task), "unknown", task)`
- **Outcome**: `r ifelse(is.na(outcome), "unknown", outcome)`

## Cross-validated Metrics

```{r}
print(safe_df(metric_summary, "Metric summary is unavailable."))
```

## Permutation-Gap Test

The permutation-gap test compares the observed cross-validated metric against a
null distribution constructed by permuting labels. A large gap with a small
p-value indicates the model has learned signal beyond random label assignment.

```{r}
print(safe_df(perm_gap, "Permutation results are unavailable."))
```

```{r fig.height=4, fig.width=7}
if (length(perm_values) > 0L) {
  obs <- NA_real_
  if (is.data.frame(perm_gap) && nrow(perm_gap) > 0L && "metric_obs" %in% names(perm_gap)) {
    obs <- as.numeric(perm_gap$metric_obs[[1]])
  }
  hist(
    perm_values,
    breaks = max(10L, min(50L, floor(sqrt(length(perm_values))))),
    main = "Permutation Metric Distribution",
    xlab = "Metric",
    col = "grey85",
    border = "white"
  )
  if (is.finite(obs)) {
    abline(v = obs, col = "firebrick", lwd = 2)
  }
} else {
  plot.new()
  text(0.5, 0.5, "No permutation values available.")
}
```

## Batch / Study Association

```{r}
print(safe_df(batch_assoc, "No batch/study association results available."))
```

## Confounder Sensitivity

```{r}
tryCatch({
  cs <- confounder_sensitivity(audit@fit)
  if (is.data.frame(cs) && nrow(cs) > 0L) {
    print(cs)
  } else {
    print(empty_df("Confounder sensitivity analysis returned no results."))
  }
}, error = function(e) {
  print(empty_df(paste("Confounder sensitivity unavailable:", e$message)))
})
```

## Calibration

```{r}
tryCatch({
  if (identical(task, "binomial")) {
    cal <- calibration_summary(audit@fit)
    if (is.data.frame(cal) && nrow(cal) > 0L) {
      print(cal)
    } else {
      print(empty_df("Calibration summary returned no results."))
    }
  } else {
    print(empty_df("Calibration analysis is only available for binomial tasks."))
  }
}, error = function(e) {
  print(empty_df(paste("Calibration unavailable:", e$message)))
})
```

```{r fig.height=4, fig.width=5}
tryCatch({
  if (identical(task, "binomial") && has_ggplot2) {
    cal <- calibration_summary(audit@fit)
    if (is.data.frame(cal) && nrow(cal) > 0L &&
        all(c("bin_midpoint", "observed_rate") %in% names(cal))) {
      p <- ggplot2::ggplot(cal, ggplot2::aes(x = bin_midpoint, y = observed_rate)) +
        ggplot2::geom_point() +
        ggplot2::geom_abline(slope = 1, intercept = 0, linetype = "dashed", color = "grey50") +
        ggplot2::labs(x = "Predicted probability", y = "Observed rate",
                      title = "Calibration Curve") +
        ggplot2::theme_minimal()
      print(p)
    }
  }
}, error = function(e) NULL)
```

## Target Leakage Scan

```{r}
if (is.data.frame(target_assoc) && nrow(target_assoc) > 0L) {
  print(utils::head(target_assoc, 20L))
} else {
  print(empty_df("No target association results available."))
}
```

## Duplicate Detection

```{r}
if (is.data.frame(duplicates) && nrow(duplicates) > 0L) {
  print(utils::head(duplicates, 20L))
} else {
  print(empty_df("No near-duplicate samples detected."))
}
```

## Fold Balance

```{r fig.height=4, fig.width=6}
tryCatch({
  if (!is.null(fold_status) && is.data.frame(fold_status) && nrow(fold_status) > 0L) {
    status_tbl <- table(fold_status$status)
    barplot(status_tbl, col = c(success = "steelblue", skipped = "orange",
                                 failed = "firebrick")[names(status_tbl)],
            main = "Fold Status Distribution",
            ylab = "Count", las = 1)
  } else {
    plot.new()
    text(0.5, 0.5, "No fold status information available.")
  }
}, error = function(e) {
  plot.new()
  text(0.5, 0.5, "Fold balance plot unavailable.")
})
```

## Methods and Interpretation

This report summarizes a post-hoc leakage audit performed by bioLeak. Each
section targets a different aspect of data leakage and confounding:

- **Cross-validated Metrics**: Performance estimates from the resampling scheme.
  Confidence intervals (when present) use the Nadeau-Bengio (2003) corrected
  variance to account for training-set overlap.
- **Permutation-Gap Test**: Compares observed performance to a null built by
  shuffling labels. A significant gap indicates the model captures genuine
  signal; a non-significant gap warns of potential overfitting or leakage.
- **Batch / Study Association**: Chi-squared tests for confounding between fold
  assignment and batch or study labels. Significant associations suggest the
  experimental design may inflate performance.
- **Confounder Sensitivity**: Quantifies how much each confounder contributes to
  apparent model performance.
- **Calibration**: (Binomial tasks only.) Compares predicted probabilities to
  observed event rates. Good calibration means predicted probabilities are
  trustworthy.
- **Target Leakage Scan**: Flags features with unusually strong univariate
  associations with the outcome, which may indicate information leaking from the
  target into the feature set.
- **Duplicate Detection**: Identifies near-duplicate samples that could inflate
  performance when they span train and test sets.
- **Fold Balance**: Summarizes the success, skip, and failure rate across
  cross-validation folds.

## Session Information

```{r}
tryCatch({
  if (!is.null(provenance) && is.list(provenance)) {
    cat("R version:", provenance$r_version, "\n")
    cat("Platform:", provenance$platform, "\n")
    cat("Timestamp:", format(provenance$timestamp), "\n")
    if (!is.na(provenance$git_sha)) {
      cat("Git SHA:", provenance$git_sha, "\n")
    }
    if (!is.null(provenance$hardware)) {
      cat("System:", provenance$hardware$sysname, "/",
          provenance$hardware$machine, "\n")
    }
    cat("\nKey packages:\n")
    print(utils::head(provenance$packages, 20))
  } else {
    print(utils::sessionInfo())
  }
}, error = function(e) {
  print(utils::sessionInfo())
})
```

---
title: "bioLeak: Leakage-Aware Biomedical Modeling"
author: "Selçuk Korkmaz"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_document:
    toc: true
    toc_float: true
    number_sections: true
    theme: flatly
    highlight: tango
vignette: >
  %\VignetteIndexEntry{bioLeak: Leakage-Aware Biomedical Modeling}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE,
  eval = TRUE
)
```

# Why bioLeak

bioLeak is a leakage-aware modeling toolkit for biomedical and machine-learning
analyses. Its purpose is to prevent and diagnose information leakage across
resampling workflows where training and evaluation data are not truly
independent because samples share subjects, batches, studies, or time.

Standard workflows are often insufficient. Random, row-wise cross-validation
assumes samples are independent. Global preprocessing (imputation, scaling,
feature selection) done before resampling lets test-fold information shape the
training process. These choices inflate performance and can lead to incorrect
biomarker discovery, misleading clinical signals, or models that fail in
deployment.

Data leakage means any direct or indirect use of evaluation data in training or
feature engineering. In biomedical data, leakage commonly appears as:

- repeated measurements from the same patient split across folds
- batch or study effects that align with the outcome
- duplicates or near-duplicates across train and test
- time-series lookahead or random splits that use future information
- preprocessing that uses all samples instead of training-only statistics

bioLeak addresses these issues with leakage-aware splitting, guarded
preprocessing that is fit only on training data, and audit diagnostics that
surface overlaps, confounding, and duplicates.

# Guided workflow

The sections below walk through a leakage-aware workflow from data setup to
audits. Each step includes a leaky failure mode and a corrected alternative.

## Example data

```{r example-data}
library(bioLeak)

set.seed(123)
n <- 160
subject <- rep(seq_len(40), each = 4)
batch <- sample(paste0("B", 1:6), n, replace = TRUE)
study <- sample(paste0("S", 1:4), n, replace = TRUE)
time <- seq_len(n)

x1 <- rnorm(n)
x2 <- rnorm(n)
x3 <- rnorm(n)
linpred <- 0.7 * x1 - 0.4 * x2 + 0.2 * x3 + rnorm(n, sd = 0.5)
p <- stats::plogis(linpred)
outcome <- factor(ifelse(runif(n) < p, "case", "control"),
                  levels = c("control", "case"))

df <- data.frame(
  subject = subject,
  batch = batch,
  study = study,
  time = time,
  outcome = outcome,
  x1 = x1,
  x2 = x2,
  x3 = x3
)

df_leaky <- within(df, {
  leak_subject <- ave(as.numeric(outcome == "case"), subject, FUN = mean)
  leak_batch <- ave(as.numeric(outcome == "case"), batch, FUN = mean)
  leak_global <- mean(as.numeric(outcome == "case"))
})

df_time <- df
df_time$leak_future <- c(as.numeric(df_time$outcome == "case")[-1], 0)
predictors <- c("x1", "x2", "x3")


# Example data (first 6 rows)
head(df)

# Outcome class counts
as.data.frame(table(df$outcome))
```

The table preview displays the metadata columns (`subject`, `batch`, `study`, `time`), the binary outcome, and three numeric predictors (`x1`, `x2`, `x3`).

The class count table shows the baseline prevalence. Stratified splits try to preserve these proportions (for grouped modes, stratification is applied at the group level).


## Create leakage-aware splits with `make_split_plan()`

`make_split_plan()` is the foundation. It returns a `LeakSplits` object with
explicit train/test indices (or compact fold assignments) and metadata. For
data.frame inputs, a unique `row_id` column is added automatically, so
`group = "row_id"` is the explicit way to request sample-wise CV. It assumes
that the grouping columns you provide are complete and that samples sharing a
group must not cross folds. Grouped stratification uses the majority class per
group and is ignored for `study_loocv`, `time_series`, and survival outcomes.
Misuse to avoid:

- using `group = "row_id"` (sample-wise CV) when subjects repeat
- using the wrong grouping column (for example batch instead of subject)
- using time-series CV with unsorted or missing time values
- relying on stratification when the outcome is missing, single-class at the
  group level, or a time/event outcome
- assuming `time_series` will always return exactly `v` folds

**Leaky example: row-wise CV when subjects repeat**

```{r leaky-splits}
leaky_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "row_id",
  v = 5,
  repeats = 1,
  stratify = TRUE
)

cat("Leaky splits summary (sample-wise CV):\n")
leaky_splits
```

The printed `LeakSplits` summary reports the split mode, number of folds, and fold-level training and test set sizes.

Because `group = "row_id"`, each sample is treated as its own group. As a result, repeated samples from the same subject may appear in both the training and test sets, which introduces information leakage.


**Leakage-safe alternative: group by subject**

```{r safe-splits}
safe_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "subject",
  v = 5,
  repeats = 1,
  stratify = TRUE,
  seed = 10
)

cat("Leakage-safe splits summary (subject-grouped CV):\n")
safe_splits
```

Here, each subject is confined to a single fold. Test set sizes are multiples
of the number of samples per subject, confirming that subjects were not split
across folds.

The fold sizes remain comparable because `stratify = TRUE` balances outcome
proportions across folds while respecting subject-level boundaries.


**Other leakage-aware modes**

```{r splits-other-modes}
batch_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "batch_blocked",
  batch = "batch",
  v = 4,
  stratify = TRUE
)

cat("Batch-blocked splits summary:\n")
batch_splits

study_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "study_loocv",
  study = "study"
)

cat("Study leave-one-out splits summary:\n")
study_splits

time_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "time_series",
  time = "time",
  v = 4,
  horizon = 2
)

cat("Time-series splits summary:\n")
time_splits

nested_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "subject",
  v = 3,
  nested = TRUE,
  stratify = TRUE
)

cat("Nested CV splits summary:\n")
nested_splits
```

**RNG seed lineage**

bioLeak uses offset-based seeding so that every sub-operation gets a
deterministic, distinct seed without requiring isolated RNG streams:

| Function | Base seed | Per-unit offset |
|---|---|---|
| `make_split_plan()` | `set.seed(seed)` | repeats: `seed + 1000 * r`; nested: `seed + 1` |
| `fit_resample()` | `set.seed(seed)` | per-fold: `seed + fold_id` |
| `audit_leakage()` | `set.seed(seed)` | per-permutation: `seed + b` |

These are **not** isolated RNG streams but simple offsets. Collisions do not
occur for practical parameter values (e.g., `repeats < 1000`, `v < 1000`).
Always set an explicit `seed` argument for reproducibility; strict mode warns
when `seed` is `NULL` or `NA`.

Each summary reports the number of folds and the fold sizes for the corresponding split strategy.

- **Batch & Study:** The test set sizes vary because batches/studies are not evenly sized. `study_loocv` uses one study per fold, so `v` and `repeats` are ignored.

- **Time series:** The training set size increases across folds while the test set size follows the block size. Early blocks with no prior training data (or too-small test windows) are skipped, so you may see fewer than `v` folds.

**Combined (N-axis) mode**

When leakage can occur along multiple axes simultaneously (for example, both
subject and batch), `mode = "combined"` handles multi-axis constraint splitting.
The first element in `constraints` is the **primary fold driver** (determines
which groups go to test); subsequent elements are **exclusion constraints**
(training samples sharing constraint-axis levels with the test set are removed).

```{r combined-splits}
# For combined mode, constraint-axis levels should not span all folds.
# Here each subject belongs to exactly one site, so site exclusion only
# removes training samples from the same site as the test subjects.
df_comb <- df
df_comb$site <- paste0("site", rep(1:8, each = 5)[df_comb$subject])

combined_splits <- make_split_plan(
  df_comb,
  outcome = "outcome",
  mode = "combined",
  constraints = list(
    list(type = "subject", col = "subject"),
    list(type = "batch",   col = "site")
  ),
  v = 4,
  stratify = TRUE,
  seed = 42
)

cat("Combined (N-axis) splits summary:\n")
combined_splits
```

The fold sizes may be smaller than single-axis modes because training samples
are excluded whenever their site (or any constraint-axis level) also appears in
the test set. You can add more than two axes by appending additional elements to
`constraints`.

The legacy `primary_axis`/`secondary_axis` parameters are still accepted for
backward compatibility but `constraints` supersedes them and supports arbitrary
numbers of axes.

```{r combined-overlap-check}
# Verify zero overlap on all constraint axes
overlap_result <- check_split_overlap(combined_splits)
overlap_result
```

**Assumptions and intent by mode**

- `subject_grouped`: keeps all samples from a subject together
- `batch_blocked`: keeps batches/centers together (leave-one-group-out when `v`
  is at least the number of batches; otherwise multiple batches per fold)
- `study_loocv`: holds out each study in turn for external validation
- `time_series`: rolling-origin splits with an optional horizon to prevent lookahead
- `combined`: primary axis drives fold assignment; all additional constraint axes
  are enforced as exclusion constraints on the training set

Use `repeats` for repeated CV in grouped/batch modes (it is ignored for
`study_loocv` and `time_series`), `stratify = TRUE` to balance outcome
proportions at the group level when applicable, and `nested = TRUE` to attach
inner folds (one repeat). For large datasets, `progress = TRUE` reports
progress; storing explicit indices can be memory intensive.


## Validating splits with `check_split_overlap()`

`check_split_overlap()` verifies that no grouping-column levels appear in both
the training and test sets of any fold. It returns a data.frame with one row per
fold-column combination showing the overlap count and a pass/fail flag.

```{r check-split-overlap}
# Run on the subject-grouped splits from earlier
overlap_safe <- check_split_overlap(safe_splits)
overlap_safe
```

When `cols` is not supplied, the function auto-infers the relevant columns from
the split mode (e.g., `subject` for `subject_grouped`, all constraint axes for
`combined`). By default `stop_on_fail = TRUE`, so the function raises a
`bioLeak_overlap_error` when any overlap is detected. Set `stop_on_fail = FALSE`
to return the results without stopping.

When strict mode is active (`options(bioLeak.strict = TRUE)`),
`make_split_plan()` automatically runs `check_split_overlap()` after building
non-compact splits, so you do not need to call it manually.


## Scalability

**Handling Large Datasets (Compact Mode)**

For large datasets (e.g., $N > 50,000$) with many repeats, storing explicit
integer indices for every fold can consume gigabytes of memory. Use `compact = TRUE`
to store a lightweight vector of fold assignments instead. `fit_resample()` will
automatically reconstruct the indices on the fly.
Compact mode is not supported when `nested = TRUE` (it falls back to full
indices). For `time_series` compact splits, the time column must be present in
the stored split metadata so folds can be reconstructed.

```{r compact-splits}
# Efficient storage for large N
big_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "subject",
  v = 5,
  compact = TRUE  # <--- Saves memory
)

cat("Compact-mode splits summary:\n")
big_splits
cat(sprintf("Compact storage enabled: %s\n", big_splits@info$compact))
```

The summary is identical to a regular split, but the underlying storage is a
compact fold-assignment vector. Use the `compact` flag to confirm the memory-
saving mode is active.

## Strict leakage mode

Setting `options(bioLeak.strict = TRUE)` activates a global safety switch that
tightens validation across the entire workflow. In strict mode:

1. **Trained recipe/workflow detection escalates to an error.** Normally,
   passing a pre-trained recipe or workflow to `fit_resample()` or
   `tune_resample()` produces a warning. In strict mode, it raises an error so
   fold-global leakage cannot proceed silently.

2. **Seed warnings.** When `seed` is `NULL` or `NA` in `make_split_plan()`,
   `fit_resample()`, or `tune_resample()`, strict mode emits a
   `bioLeak_validation_warning` reminding you to set an explicit seed for
   reproducibility.

3. **Combined + nested warning.** Using `nested = TRUE` with
   `mode = "combined"` may produce empty inner folds; strict mode warns about
   this before proceeding.

4. **Automatic overlap check.** After building non-compact splits,
   `make_split_plan()` auto-runs `check_split_overlap()` to verify that no
   grouping-column levels cross fold boundaries.

```{r strict-mode-demo, error=TRUE}
# Enable strict mode temporarily
withr::with_options(list(bioLeak.strict = TRUE), {
  strict_splits <- make_split_plan(
    df,
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 5,
    stratify = TRUE,
    seed = 42
  )
  cat("Strict mode splits completed without error.\n")
})
```

```{r strict-mode-trained-recipe, error=TRUE}
# Strict mode catches a pre-trained recipe
if (requireNamespace("recipes", quietly = TRUE)) {
  rec <- recipes::recipe(outcome ~ ., data = df[, c("outcome", predictors)]) |>
    recipes::step_normalize(recipes::all_numeric_predictors()) |>
    recipes::prep(training = df[, c("outcome", predictors)])

  withr::with_options(list(bioLeak.strict = TRUE), {
    tryCatch(
      fit_resample(
        df, outcome = "outcome",
        splits = safe_splits,
        learner = "glmnet",
        preprocess = rec
      ),
      error = function(e) cat("Strict mode error:", conditionMessage(e), "\n")
    )
  })
}
```

Strict mode is off by default. Use `withr::with_options()` for temporary
activation or set `options(bioLeak.strict = TRUE)` in your `.Rprofile` for
project-wide enforcement.


## Guarded preprocessing and imputation

`bioLeak` uses guarded preprocessing to prevent leakage from global imputation,
scaling, and feature selection. The low-level API is:

- `guard_fit()` to fit preprocessing on training data only
- `predict()` (dispatching to `predict.GuardFit()`) to apply the trained guard
  to new data; the legacy alias `predict_guard()` is also kept for
  backward compatibility
- `guard_ensure_levels()` to align factor levels across train/test
- `impute_guarded()` as a convenience wrapper for leakage-safe imputation

`guard_fit()` fits a pipeline that can winsorize, impute, normalize, filter,
and select features. It one-hot encodes non-numeric columns and carries factor
levels forward to new data. Assumptions and misuse to avoid:

- `fs = "ttest"` requires a binary outcome with enough samples per class
- `fs = "lasso"` requires `glmnet`
- `impute = "mice"` is not supported for guarded prediction
- `impute = "none"` with missing values will trigger a median fallback and
  missingness indicators to avoid errors

Supported preprocessing options include imputation (`median`, `knn`,
`missForest`, `none`), normalization (`zscore`, `robust`, `none`), filtering by
variance or IQR (optionally `min_keep`), and feature selection (`ttest`,
`lasso`, `pca`). Winsorization is controlled via `impute$winsor` and
`impute$winsor_k` in guarded steps; in `impute_guarded()`, use `winsor` and
`winsor_thresh`. `impute_guarded()` returns a `LeakImpute` object with guarded
data, diagnostics, and the fitted guard state.

**Leaky example: global scaling before CV**

```{r leaky-scaling}
df_leaky_scaled <- df
df_leaky_scaled[predictors] <- scale(df_leaky_scaled[predictors])
scaled_summary <- data.frame(
  feature = predictors,
  mean = colMeans(df_leaky_scaled[predictors]),
  sd = apply(df_leaky_scaled[predictors], 2, stats::sd)
)
scaled_summary$mean <- round(scaled_summary$mean, 3)
scaled_summary$sd <- round(scaled_summary$sd, 3)

# Leaky global scaling: means ~0 and SDs ~1 computed on all samples
scaled_summary
```

The summary shows that scaling used the full dataset, so test-fold statistics
influenced the training transformation. This violates the train/test separation
and can bias performance estimates.

**Leakage-safe alternative: fit preprocessing on training only**

```{r guarded-preprocess}
fold1 <- safe_splits@indices[[1]]
train_x <- df[fold1$train, predictors]
test_x <- df[fold1$test, predictors]

guard <- guard_fit(
  X = train_x,
  y = df$outcome[fold1$train],
  steps = list(
    impute = list(method = "median", winsor = TRUE),
    normalize = list(method = "zscore"),
    filter = list(var_thresh = 0, iqr_thresh = 0),
    fs = list(method = "none")
  ),
  task = "binomial"
)

train_x_guarded <- predict(guard, train_x)
test_x_guarded <- predict(guard, test_x)

cat("GuardFit object:\n")
guard
cat("\nGuardFit summary:\n")
summary(guard)

# Guarded training data (first 6 rows)
head(train_x_guarded)

# Guarded test data (first 6 rows)
head(test_x_guarded)
```

The `GuardFit` object records the preprocessing steps and the number of
features retained after filtering. The summary reports how many features were
removed and the preprocessing audit trail. The guarded train/test previews show
that missing values were imputed and (if requested) scaled using training-only
statistics; the test data never influences these values.

**Factor level guard (advanced)**

```{r guard-ensure-levels}
raw_levels <- data.frame(
  site = c("A", "B", "B"),
  status = c("yes", "no", "yes"),
  stringsAsFactors = FALSE
)

level_state <- guard_ensure_levels(raw_levels)

# Aligned factor data with consistent levels
level_state$data

# Levels map
level_state$levels
```

The returned `data` keeps factor levels consistent across folds, while the
`levels` list records the training-time levels (including any dummy levels added
to preserve one-hot columns). Use these when transforming new data to avoid
misaligned model matrices.

**Leaky example: imputation using train and test together**

```{r leaky-impute}
train <- data.frame(a = c(1, 2, NA, 4), b = c(NA, 1, 1, 0))
test <- data.frame(a = c(NA, 5), b = c(1, NA))

all_median <- vapply(rbind(train, test),
                     function(col) median(col, na.rm = TRUE),
                     numeric(1))
train_leaky <- as.data.frame(Map(function(col, m) { col[is.na(col)] <- m; col },
                                 train, all_median))
test_leaky <- as.data.frame(Map(function(col, m) { col[is.na(col)] <- m; col },
                                test, all_median))

# Leaky medians computed on train + test
data.frame(feature = names(all_median), median = all_median)

# Leaky-imputed training data
train_leaky

# Leaky-imputed test data
test_leaky
```

The medians above were computed using both train and test data, so the test set
influences the imputation values. This is a classic leakage pathway because
test information directly alters the training features.

**Leakage-safe alternative: `impute_guarded()`**

```{r safe-impute}
imp <- impute_guarded(
  train = train,
  test = test,
  method = "median",
  winsor = FALSE
)

# Guarded-imputed training data
imp$train

# Guarded-imputed test data
imp$test
```

Here, the imputation statistics are computed from the training set only. The missing value in the test set (column `a`) is replaced by 2, which is the training median. In contrast, in the leaky example above, it was replaced by 3, the global median.

This confirms that the test set is transformed using fixed values learned from the training data, preserving a clean separation between training and evaluation.
The `LeakImpute` object also contains missingness diagnostics in
`imp$summary$diagnostics`, and guarded outputs use the same encoding as
`guard_fit()` (categorical variables are one-hot encoded). Use `vars` to
impute only a subset of columns when needed.

**Bridging guard steps to recipes: `guard_to_recipe()`**

`guard_to_recipe()` converts a guard preprocessing specification into a
`recipes::recipe`, enabling a smooth transition from guard-based to
tidymodels-based workflows. It maps supported guard steps to their closest
recipe equivalents:

| Guard step | Recipe equivalent |
|---|---|
| `impute$method = "median"` | `step_impute_median()` |
| `impute$method = "knn"` | `step_impute_knn()` |
| `normalize$method = "zscore"` | `step_normalize()` |
| `filter$var_thresh > 0` | `step_nzv()` |
| `fs$method = "pca"` | `step_pca()` |

Steps with no direct recipe equivalent (`missForest`, `mice`, `robust`
normalization, `ttest`, `lasso` feature selection) produce a
`bioLeak_fallback_warning` and either fall back to the closest available step or
are skipped.

```{r guard-to-recipe}
if (requireNamespace("recipes", quietly = TRUE)) {
  guard_steps <- list(
    impute    = list(method = "median"),
    normalize = list(method = "zscore")
  )

  rec <- guard_to_recipe(
    steps = guard_steps,
    formula = outcome ~ .,
    training_data = df[, c("outcome", predictors)]
  )

  cat("Converted recipe:\n")
  rec
}
```

The returned recipe is untrained and can be passed directly to `fit_resample()`
or `tune_resample()` as the `preprocess` argument.

## Fit and resample with `fit_resample()`

`fit_resample()` combines leakage-aware splits with guarded preprocessing and
model fitting. It supports:

- built-in learners (`glmnet`, `ranger`)
- `parsnip` model specifications or `workflows::workflow` objects (recommended)
- `recipes::recipe` preprocessing (prepped on training folds only)
- `rsample` resamples (`rset`/`rsplit`) via `splits`, with `split_cols` to map metadata
- custom learners via `custom_learners` (legacy/advanced)
- multiple metrics, class weights, positive class override, and yardstick metrics
- fold-level run diagnostics in `fit@info$fold_status` (`success`/`skipped`/`failed`)

Assumptions and misuse to avoid:

- outcome must be binary or multiclass (factor) for classification, numeric for
  regression, or a survival outcome (a `Surv` column or time/event columns)
- `positive_class` must be a single value that exists in the outcome levels
- `class_weights` only applies to binomial/multiclass tasks; workflows must
  handle weights internally
- if you supply a matrix without metadata, you must remove leakage columns
  yourself (group, batch, study, time)

Use `learner_args` to pass model-specific arguments. For custom learners
(advanced), you can supply separate `fit` and `predict` argument lists. Set
`parallel = TRUE` to use future.apply for fold-level parallelism when available.
When `learner` is a parsnip specification, `learner_args` and
`custom_learners` are ignored. When `learner` is a workflow, `preprocess` and
`learner_args` are ignored.
When a recipe or workflow is used, the built-in guarded preprocessing list is
bypassed, so ensure the recipe/workflow itself is leakage-safe.

Metrics used in this vignette:

- **AUC**: area under the ROC curve (0.5 = random, 1.0 = perfect; higher is better)
- **PR AUC**: area under the precision-recall curve (0 to 1; higher is better)
- **Accuracy**: fraction of correct predictions (0 to 1; higher is better)
- **RMSE**: root mean squared error (0 to infinity; lower is better)
- **Macro F1**: average F1 across classes (multiclass; higher is better)
- **Log loss**: cross-entropy for probabilities (lower is better)
- **C-index**: concordance for regression/survival (higher is better)

Always report the mean and variability across folds, not a single fold value.
When using yardstick metrics, the positive class is the second factor level;
set `positive_class` to control this.

**Parsnip model specification (recommended)**

```{r parsnip-spec}
spec <- parsnip::logistic_reg(mode = "classification") |>
  parsnip::set_engine("glm")
```

This spec uses base R `glm` under the hood and does not require extra model
packages. Use it in the examples below; custom learners are covered in the
Advanced section.

**Leaky example: leaky features and row-wise splits**

```{r fit-leaky}
fit_leaky <- fit_resample(
  df_leaky,
  outcome = "outcome",
  splits = leaky_splits,
  learner = spec,
  metrics = c("auc", "accuracy"),
  preprocess = list(
    impute = list(method = "median"),
    normalize = list(method = "zscore"),
    filter = list(var_thresh = 0),
    fs = list(method = "none")
  )
)

cat("Leaky fit summary:\n")
summary(fit_leaky)
metrics_leaky <- as.data.frame(fit_leaky@metric_summary)
num_cols <- vapply(metrics_leaky, is.numeric, logical(1))
metrics_leaky[num_cols] <- lapply(metrics_leaky[num_cols], round, digits = 3)

# Leaky fit: mean and SD of metrics across folds
metrics_leaky
```

The summary reports cross-validated performance. AUC ranges from 0.5 (random)
to 1.0 (perfect), while accuracy is the fraction of correct predictions.
Because the splits are leaky, these metrics can be artificially high and should
not be used for reporting.

**Leakage-safe alternative: grouped splits and clean predictors**

```{r fit-safe}
fit_safe <- fit_resample(
  df,
  outcome = "outcome",
  splits = safe_splits,
  learner = spec,
  metrics = c("auc", "accuracy"),
  preprocess = list(
    impute = list(method = "median"),
    normalize = list(method = "zscore"),
    filter = list(var_thresh = 0),
    fs = list(method = "none")
  ),
  positive_class = "case",
  class_weights = c(control = 1, case = 1),
  refit = TRUE
)

cat("Leakage-safe fit summary:\n")
summary(fit_safe)
metrics_safe <- as.data.frame(fit_safe@metric_summary)
num_cols <- vapply(metrics_safe, is.numeric, logical(1))
metrics_safe[num_cols] <- lapply(metrics_safe[num_cols], round, digits = 3)

# Leakage-safe fit: mean and SD of metrics across folds
metrics_safe

# Per-fold metrics (first 6 rows)
head(fit_metrics(fit_safe))
```

`fit_resample()` returns a `LeakFit` object. Use `summary(fit_safe)` for a
compact report, `fit_metrics(fit_safe)` for the per-fold metric data frame,
and inspect `fit_safe@predictions` for fold-level predictions. When
`refit = TRUE`, the final model and preprocessing state are stored
in `fit_safe@info$final_model` and `fit_safe@info$final_preprocess`.
Fold-level status diagnostics are available in `fit_safe@info$fold_status`.

The mean +/- SD table is the primary performance summary to report, while the
per-fold metrics reveal variability and potential instability across folds.

By default, `fit_resample()` stores refit inputs in `fit_safe@info$perm_refit_spec`
to enable `audit_leakage(perm_refit = "auto")`. Set `store_refit_data = FALSE`
to reduce memory and provide `perm_refit_spec` manually when you want refit-based
permutations. When multiple learners are passed, `refit = TRUE` refits only the
first learner; set `refit = FALSE` if you do not need a final model.

```{r fit-fold-status}
# Fold-level diagnostics for reproducible troubleshooting
head(fit_safe@info$fold_status, 5)
```

Interpretation of `fold_status`:

- `status = "success"` means at least one learner produced valid metrics on that fold.
- `status = "skipped"` typically means a data-structure issue in that fold (for example, single-class training data in classification).
- `status = "failed"` indicates a runtime failure (for example, model fitting or preprocessing error).

Use `reason` and `notes` to decide whether to:

- adjust split design (for example, stronger grouping, stratification, or fewer folds),
- adjust preprocessing/model complexity,
- or exclude unstable configurations before formal comparison.

If many folds are skipped/failed, do **not** trust aggregate metric means until the instability is resolved.

The examples above use a parsnip model specification. To swap in other models,
replace `spec` with another parsnip spec (see the gradient boosting example
below).

**Multiclass classification (optional)**

```{r multiclass-example}
if (requireNamespace("ranger", quietly = TRUE)) {
  set.seed(11)
  df_multi <- df
  df_multi$outcome3 <- factor(sample(c("A", "B", "C"),
                                     nrow(df_multi), replace = TRUE))

  multi_splits <- make_split_plan(
    df_multi,
    outcome = "outcome3",
    mode = "subject_grouped",
    group = "subject",
    v = 4,
    stratify = TRUE,
    seed = 11
  )

  fit_multi <- fit_resample(
    df_multi,
    outcome = "outcome3",
    splits = multi_splits,
    learner = "ranger",
    metrics = c("accuracy", "macro_f1", "log_loss"),
    refit = FALSE
  )

  cat("Multiclass fit summary:\n")
  summary(fit_multi)
} else {
  cat("ranger not installed; skipping multiclass example.\n")
}
```

Multiclass fits compute accuracy, macro-F1, and log loss when class
probabilities are available. `positive_class` is ignored for multiclass tasks.

**Survival outcomes (optional)**

```{r survival-example}
if (requireNamespace("survival", quietly = TRUE) &&
    requireNamespace("glmnet", quietly = TRUE)) {
  set.seed(12)
  df_surv <- df
  df_surv$time_to_event <- rexp(nrow(df_surv), rate = 0.1)
  df_surv$event <- rbinom(nrow(df_surv), 1, 0.7)

  surv_splits <- make_split_plan(
    df_surv,
    outcome = c("time_to_event", "event"),
    mode = "subject_grouped",
    group = "subject",
    v = 4,
    stratify = FALSE,
    seed = 12
  )

  fit_surv <- fit_resample(
    df_surv,
    outcome = c("time_to_event", "event"),
    splits = surv_splits,
    learner = "glmnet",
    metrics = "cindex",
    refit = FALSE
  )

  cat("Survival fit summary:\n")
  summary(fit_surv)
} else {
  cat("survival or glmnet not installed; skipping survival example.\n")
}
```

For survival tasks, supply `outcome = c(time_col, event_col)` or a `Surv`
column; `stratify` is ignored for time/event outcomes and `class_weights` are
not used.

**Passing learner-specific arguments (optional)**

```{r learner-args}
if (requireNamespace("glmnet", quietly = TRUE)) {
  fit_glmnet <- fit_resample(
    df,
    outcome = "outcome",
    splits = safe_splits,
    learner = "glmnet",
    metrics = "auc",
    learner_args = list(glmnet = list(alpha = 0.5)),
    preprocess = list(
      impute = list(method = "median"),
      normalize = list(method = "zscore"),
      filter = list(var_thresh = 0),
      fs = list(method = "none")
    )
  )
  cat("GLMNET summary with learner-specific arguments:\n")
  summary(fit_glmnet)
} else {
  cat("glmnet not installed; skipping learner_args example.\n")
}
```

This summary reflects the same guarded preprocessing but a different model
configuration (here, `alpha = 0.5`). Use the mean +/- SD metrics to compare
learner settings under identical splits.

**SummarizedExperiment inputs (optional)**

```{r se-example}
if (requireNamespace("SummarizedExperiment", quietly = TRUE)) {
  se <- SummarizedExperiment::SummarizedExperiment(
    assays = list(counts = t(as.matrix(df[, predictors]))),
    colData = df[, c("subject", "batch", "study", "time", "outcome"), drop = FALSE]
  )

  se_splits <- make_split_plan(
    se,
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 3
  )

  se_fit <- fit_resample(
    se,
    outcome = "outcome",
    splits = se_splits,
    learner = spec,
    metrics = "auc"
  )
  cat("SummarizedExperiment fit summary:\n")
  summary(se_fit)
} else {
  cat("SummarizedExperiment not installed; skipping SE example.\n")
}
```

The summary is identical in structure to the `data.frame` case because `fit_resample()` extracts predictors and metadata from the `SummarizedExperiment` object.

Note that `features_final` here reflects only the assay predictors (`x1`, `x2`,
`x3`), because the assay was constructed without metadata columns.


## Tidymodels interoperability

`bioLeak` integrates with `rsample`, `recipes`, `workflows`, and `yardstick`:

- `splits` can be an `rsample` rset/rsplit; use `split_cols` (default `"auto"`)
  or `bioLeak_mode` / `bioLeak_perm_mode` attributes to map
  group/batch/study/time metadata.
- `preprocess` can be a `recipes::recipe` (prepped on training folds only).
- `learner` can be a `workflows::workflow`.
- `metrics` can be a `yardstick::metric_set`.
- `as_rsample()` converts `LeakSplits` to an `rsample` object.

```{r tidymodels-interop, eval = requireNamespace("recipes", quietly = TRUE) && requireNamespace("yardstick", quietly = TRUE)}
## This chunk requires the recipes and yardstick packages, which are
## listed in DESCRIPTION's Suggests rather than Imports. The eval=
## option above gates the chunk on their availability, so the vignette
## still builds when those packages are absent. Users copying this
## code into their own R session must have both packages installed
## (e.g., install.packages(c("recipes", "yardstick"))).
library(bioLeak)
library(parsnip)
library(recipes)
library(yardstick)

set.seed(123)
N <- 60
df <- data.frame(
  subject = factor(rep(paste0("S", 1:20), length.out = N)), 
  outcome = factor(rep(c("ClassA", "ClassB"), length.out = N)),
  x1 = rnorm(N),
  x2 = rnorm(N),
  x3 = rnorm(N)
)

spec <- logistic_reg() |> set_engine("glm")

# Use bioLeak's native split planner to avoid conversion errors
set.seed(13)

# Use make_split_plan instead of rsample::group_vfold_cv
# This creates a subject-grouped CV directly compatible with fit_resample
splits <- make_split_plan(
  df, 
  outcome = "outcome", 
  mode = "subject_grouped", 
  group = "subject", 
  v = 3
)

rec <- recipes::recipe(outcome ~ x1 + x2 + x3, data = df) |>
  recipes::step_impute_median(recipes::all_numeric_predictors()) |>
  recipes::step_normalize(recipes::all_numeric_predictors())

metrics_set <- yardstick::metric_set(yardstick::roc_auc, yardstick::accuracy)

fit_rs <- fit_resample(
  df,
  outcome = "outcome",
  splits = splits,
  learner = spec,
  preprocess = rec,
  metrics = metrics_set,
  refit = FALSE
)

if (exists("as_rsample", where = asNamespace("bioLeak"), mode = "function")) {
    rs_export <- as_rsample(fit_rs@splits, data = df)
    print(rs_export)
}
```

Use `split_cols` to ensure split-defining metadata are dropped from predictors
when using rsample inputs.

```{r workflow-example}
if (requireNamespace("workflows", quietly = TRUE)) {
  wf <- workflows::workflow() |>
    workflows::add_model(spec) |>
    workflows::add_formula(outcome ~ x1 + x2 + x3)

  fit_wf <- fit_resample(
    df,
    outcome = "outcome",
    splits = splits,
    learner = wf,
    metrics = "auc",
    refit = FALSE
  )

  cat("Workflow fit summary:\n")
  summary(fit_wf)
} else {
  cat("workflows not installed; skipping workflow example.\n")
}
```

When auditing rsample-based fits, pass `perm_mode = "subject_grouped"` (or set
`attr(rs, "bioLeak_perm_mode") <- "subject_grouped"`) so restricted permutations
respect the intended split design.


## Nested tuning with `tune_resample()`

`tune_resample()` runs leakage-aware nested tuning with tidymodels. It accepts a
parsnip model specification or workflow, and supports two inner-CV modes:

- precomputed inner folds (for example from `make_split_plan(..., nested = TRUE)`)
- on-the-fly inner split generation for `LeakSplits` using `inner_v`,
  `inner_repeats`, and `inner_seed`

For rsample inputs, inner folds must already be present. Survival tasks are not
yet supported.

```{r tune-example}
if (requireNamespace("parsnip", quietly = TRUE) &&
    requireNamespace("recipes", quietly = TRUE) &&
    requireNamespace("tune", quietly = TRUE) &&
    requireNamespace("glmnet", quietly = TRUE)) {
  # --- 1. Create Data ---
  set.seed(123)
  N <- 60
  df <- data.frame(
    subject = factor(rep(paste0("S", 1:20), length.out = N)),
    outcome = factor(sample(c("ClassA", "ClassB"), N, replace = TRUE)),
    x1 = rnorm(N),
    x2 = rnorm(N),
    x3 = rnorm(N)
  )

  # --- 2. Generate Nested Splits ---
  set.seed(1)
  nested_splits <- make_split_plan(
    df,
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 3,
    nested = TRUE
  )

  # --- 3. Define Recipe & Model ---
  rec <- recipes::recipe(outcome ~ x1 + x2 + x3, data = df) |>
    recipes::step_impute_median(recipes::all_numeric_predictors()) |>
    recipes::step_normalize(recipes::all_numeric_predictors())

  spec_tune <- parsnip::logistic_reg(
    penalty = tune::tune(),
    mixture = 1,
    mode = "classification"
  ) |>
    parsnip::set_engine("glmnet")

  # --- 4. Run Tuning ---
  tuned <- tune_resample(
    df,
    outcome = "outcome",
    splits = nested_splits,
    learner = spec_tune,
    preprocess = rec,
    inner_v = 2,
    grid = 3,
    metrics = c("auc", "accuracy"),
    selection = "one_std_err",
    refit = TRUE,
    seed = 14
  )

  summary(tuned)
} else {
  cat("parsnip/recipes/tune/glmnet not installed; skipping nested tuning example.\n")
}
```

```{r tune-diagnostics}
if (exists("tuned")) {
  # Fold-level status from the outer loop
  head(tuned$fold_status, 5)

  # Final tuned model is available when refit = TRUE
  is.null(tuned$final_model)
} else {
  cat("Nested tuning object not available (dependencies missing).\n")
}
```

How to interpret this tuning output:

- `selection = "one_std_err"` chooses a model within one standard error of the best inner-CV score, then applies a deterministic simplicity tie-break. This usually favors more regularized/simpler models with similar expected performance.
- `tuned$metrics` and `tuned$metric_summary` remain the primary unbiased performance estimates (outer folds only).
- `tuned$final_model` is the deployable model fit on full data using the selected hyperparameters; it is for downstream prediction, not for estimating generalization.
- `tuned$fold_status` should be reviewed before reporting results. A large number of skipped/failed folds indicates unstable tuning and should trigger split/parameter redesign.


## Advanced: Using Gradient Boosting with Parsnip

`bioLeak` natively supports `tidymodels` specifications. You can pass a `parsnip`
model specification directly to `fit_resample()`. This allows you to use
state-of-the-art algorithms like XGBoost, LightGBM, or SVMs while ensuring all
preprocessing remains guarded.

```{r fit-xgboost}
if (requireNamespace("parsnip", quietly = TRUE) &&
    requireNamespace("xgboost", quietly = TRUE) &&
    requireNamespace("recipes", quietly = TRUE)) {
  
  # 1. Define the model spec
  xgb_spec <- parsnip::boost_tree(
    mode = "classification",
    trees = 100,
    tree_depth = 6,
    learn_rate = 0.01
  ) |>
    parsnip::set_engine("xgboost")
  
  # 2. Define a recipe on data without 'subject' because split metadata
  # columns are excluded from predictors by fit_resample().
  df_for_rec <- df[, !names(df) %in% "subject"]
  
  rec_xgb <- recipes::recipe(outcome ~ ., data = df_for_rec) |>
    recipes::step_dummy(recipes::all_nominal_predictors()) |>
    recipes::step_impute_median(recipes::all_numeric_predictors())
  # Note: No need for step_rm(subject) because it's already gone!
  
  # 3. Fit
  fit_xgb <- fit_resample(
    df,
    outcome = "outcome",
    splits = splits,
    learner = xgb_spec,
    metrics = "auc",
    preprocess = rec_xgb 
  )
  
  cat("XGBoost parsnip fit summary:\n")
  print(summary(fit_xgb))
} else {
  cat("parsnip/xgboost/recipes not installed.\n")
}
```

The summary reports cross-validated AUC for a non-linear gradient boosting
model. Use the mean $\pm$ SD table to compare against baseline models, and
confirm that any gains do not coincide with leakage signals in the audit
diagnostics.


## Advanced: Custom learners

Custom learners are used when a model is not available as a *parsnip* specification or when a lightweight wrapper around base R is needed.

Each custom learner must define `fit` and `predict` components. The `fit` function must accept `x`, `y`, `task`, and `weights` as inputs, while the `predict` function must accept `object` and `newdata`.

```{r custom-learners}
custom_learners <- list(
  glm = list(
    fit = function(x, y, task, weights, ...) {
      df_fit <- data.frame(y = y, x, check.names = FALSE)
      stats::glm(y ~ ., data = df_fit,
                 family = stats::binomial(), weights = weights)
    },
    predict = function(object, newdata, task, ...) {
      as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
                                type = "response"))
    }
  )
)

cat("Custom learner names:\n")
names(custom_learners)
cat("Custom learner components (fit/predict):\n")
lapply(custom_learners, names)
```

This output confirms that each custom learner provides both a `fit` and a `predict` function.

Custom learners can be used with `fit_resample()` as follows:

```r
fit_resample(
  ...,
  learner = "glm",
  custom_learners = custom_learners
)
```

## Visual diagnostics

**bioLeak** includes plotting helpers for quick diagnostic checks:

- `plot_fold_balance()` shows class balance per fold.  
- `plot_overlap_checks()` highlights train/test overlap for a metadata column.  
- `plot_time_acf()` checks autocorrelation in time-series predictions.
- `plot_calibration()` shows reliability of binomial probabilities.  
- `plot_confounder_sensitivity()` summarizes performance by batch/study strata.

Tabular helpers are also available:

- `calibration_summary()` returns calibration curve data and ECE/MCE/Brier metrics.
- `confounder_sensitivity()` returns per-stratum performance summaries.

**Misuse to avoid**

- Plotting without predictions (fits with no successful folds).  
- Using `plot_overlap_checks()` when the column is not present in the split metadata.  
- Using `plot_time_acf()` without a time column or for non-temporal data.
- Using `plot_calibration()` outside binomial tasks.
- Using `plot_confounder_sensitivity()` with a metric unsupported by the task.

For classification, `plot_fold_balance()` uses the positive class recorded in
`fit@info$positive_class`, or defaults to the second factor level if this is not
set. For multiclass tasks, it shows per-class counts without a proportion line.

```{r plot-fold-balance}
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_fold_balance(fit_safe)
} else {
  cat("ggplot2 not installed; skipping fold balance plot.\n")
}
```

The bar chart shows the counts of Positives (blue) and Negatives (tan) in each
fold. The dashed blue line represents the proportion of the positive class.

In a well-stratified fit, this line remains relatively stable across folds.
Large fluctuations indicate poor stratification, which can lead to unstable
fold-level performance estimates.

```{r plot-calibration}
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_calibration(fit_safe, bins = 10)
} else {
  cat("ggplot2 not installed; skipping calibration plot.\n")
}
```

Calibration curves compare predicted probabilities to observed event rates. Large deviations from the diagonal indicate miscalibration.
Use `min_bin_n` to suppress tiny bins and `learner =` to select a model when
multiple learners are stored in the fit.

```{r plot-confounder-sensitivity}
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_confounder_sensitivity(fit_safe, confounders = c("batch", "study"),
                              metric = "auc", min_n = 3)
} else {
  cat("ggplot2 not installed; skipping confounder sensitivity plot.\n")
}
```

Confounder sensitivity plots highlight whether performance varies substantially across batch or study strata.

```{r diagnostics-tables}
cal <- calibration_summary(fit_safe, bins = 10, min_bin_n = 5)
conf_tbl <- confounder_sensitivity(fit_safe,
                                   confounders = c("batch", "study"),
                                   metric = "auc",
                                   min_n = 3)

# Calibration metrics
cal$metrics

# Confounder sensitivity table (first 6 rows)
head(conf_tbl)
```


```{r plot-overlap}
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_overlap_checks(fit_leaky, column = "subject")
  plot_overlap_checks(fit_safe, column = "subject")
} else {
  cat("ggplot2 not installed; skipping overlap plots.\n")
}
```

The overlap plots compare the number of unique subjects appearing in the training and test sets for each fold.

**Top plot (`fit_leaky`)**  
The red line shows high overlap counts, accompanied by the explicit *“WARNING: Overlaps detected!”* annotation. This confirms that the same subjects appear in both the training and test sets, indicating information leakage.

**Bottom plot (`fit_safe`)**  
The overlap line remains flat at zero across all folds. This confirms that `subject_grouped` splitting successfully keeps subjects isolated, ensuring that no subject appears in both sets simultaneously.


## Audit leakage with `audit_leakage()`

`audit_leakage()` combines four diagnostics in one object:

- permutation gap: tests whether model signal is non-random
- batch or study association with folds (chi-square and Cramers V)
- target leakage scan: feature-wise outcome association to flag proxy columns
- duplicate and near-duplicate detection across train/test (by default) using feature similarity

Assumptions and misuse to avoid:

- `coldata` (or stored split metadata) must align with prediction IDs; rownames
  or a `row_id` column help alignment
- `X_ref` must align to prediction IDs (rownames, `row_id`, or matching order)
- `plot_perm_distribution()` requires `return_perm = TRUE`
- `target_threshold` should be set high enough to flag only strong proxies
- target leakage scan includes univariate associations plus an optional
  multivariate/interaction check (`target_scan_multivariate`, `*_components`,
  `*_interactions`, `*_B`); proxies outside `X_ref` can still pass undetected
- for rsample splits, set `perm_mode` (or `bioLeak_perm_mode`) so restricted
  permutations respect the intended design
- for time-series audits, ensure a valid time column and set `time_block` /
  `block_len` as needed

**Interpretation Note:** The label-permutation test refits models by default when
refit inputs are available (`perm_refit = "auto"` with `store_refit_data = TRUE`)
and `B <= perm_refit_auto_max`. Otherwise it keeps predictions fixed, so the
p-value quantifies prediction-label association rather than a full refit null.
A large gap indicates a non-random association. To determine if that signal is
*real* or *leaked*, check the **Batch Association** (confounding),
**Target Leakage Scan** (proxy features), and **Duplicate Detection**
(memorization) tables. Use `perm_refit = FALSE` to force fixed-prediction
shuffles or `perm_refit = TRUE` with `perm_refit_spec` to always refit.

Use `feature_space = "rank"` to compare samples by rank profiles, and
`sim_method` (`cosine` or `pearson`) to control similarity. For large `n`, `nn_k`
and `max_pairs` limit duplicate searches. Use `duplicate_scope = "all"` to
include within-fold duplicates (data-quality checks). Use `ci_method = "if"`
(influence-function) or `ci_method = "bootstrap"` with `boot_B` to obtain a
confidence interval for the permutation gap; `include_z` controls whether a
z-score is reported. Set `perm_stratify = "auto"` to stratify permutations only
when outcomes support it.

**Leakage-aware audit**

```{r audit-leakage}
# Use df_leaky (the original 160-row data) so X_ref aligns with fit_safe's predictions.
# df may have been redefined to a smaller dataset in the tidymodels section above.
X_ref <- df_leaky[, predictors]
X_ref[c(1, 5), ] <- X_ref[1, ]

audit <- audit_leakage(
  fit_safe,
  metric = "auc",
  B = 20,
  perm_stratify = TRUE,
  batch_cols = c("batch", "study"),
  X_ref = X_ref,
  sim_method = "cosine",
  sim_threshold = 0.995,
  return_perm = TRUE
)

cat("Leakage audit summary:\n")
summary(audit)
pg <- audit_perm_gap(audit)
if (!is.null(pg) && nrow(pg) > 0) {
  # Permutation significance results
  pg
}
ba <- audit_batch_assoc(audit)
if (!is.null(ba) && nrow(ba) > 0) {
  # Batch/study association with folds (Cramer's V)
  ba
} else {
  cat("No batch or study associations detected.\n")
}
ta <- audit_target_assoc(audit)
if (!is.null(ta) && nrow(ta) > 0) {
  # Top features by target association score
  head(ta)
} else {
  cat("No target leakage scan results available.\n")
}
du <- audit_duplicates(audit)
if (!is.null(du) && nrow(du) > 0) {
  # Top duplicate/near-duplicate pairs by similarity
  head(du)
} else {
  cat("No near-duplicates detected.\n")
}
```

The permutation table reports the observed metric, the mean under random label permutation, the gap (difference), and a permutation *p*-value. For metrics where higher values indicate better performance, larger gaps reflect stronger non-random signal.

The batch association table reports chi-square statistics and Cramer's V. Large
p-values and small V values indicate that folds are not aligned with batch or
study labels (which is the desired outcome when these should be independent).

The target scan table lists features with the strongest associations with the
outcome. For numeric features, the score is \(|\mathrm{AUC} - 0.5| \times 2\) for
classification or the absolute correlation for regression. For categorical
features, the score is Cramér’s V or eta-squared. Scores closer to 1 indicate
stronger outcome association.

The duplicate table lists pairs of samples with near-identical profiles that
cross train/test folds (by default). In this setup, the artificially duplicated
pair (`X_ref[c(1, 5), ] <- X_ref[1, ]`) should appear near the top of the list.
Use `duplicate_scope = "all"` to include within-fold duplicates.

**Mechanism risk assessment**

`summary(audit)` now includes a **Mechanism Risk Assessment** section that
classifies leakage evidence into four mechanism classes:

| Mechanism class | Evidence slot | Flagging rule |
|---|---|---|
| `non_random_signal` | permutation gap | p ≤ 0.05 and gap > 0 |
| `confounding_alignment` | batch association | p ≤ 0.05 and Cramér's V ≥ 0.1 |
| `proxy_target_leakage` | target scan | any feature flagged (FDR or raw) |
| `duplicate_overlap` | duplicates | any cross-fold duplicates present |
| `temporal_lookahead` | duplicates | cross-fold duplicates in time-series splits |

The summary is stored in `audit@info$mechanism_summary` and can be accessed
directly for programmatic triage:

```{r mechanism-summary}
mech <- audit@info$mechanism_summary
if (is.data.frame(mech) && nrow(mech) > 0) {
  mech
} else {
  cat("No mechanism summary available.\n")
}
```

Each row reports whether the mechanism was flagged, the evidence type, and the
associated test statistic and p-value. This provides a quick triage overview
before drilling into individual audit slots.

```{r plot-perm}
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_perm_distribution(audit)
} else {
  cat("ggplot2 not installed; skipping permutation plot.\n")
}
```

The histogram shows the null distribution (gray bars) of the performance metric under random label permutation.

The blue dashed line represents the average performance of a random model (the permuted mean). The red solid line represents the observed performance of the fitted model.

A genuine signal is indicated when the red line lies well to the right of the
gray distribution; overlap suggests weak or unstable signal.

Use the mechanism risk assessment (`audit@info$mechanism_summary`) as a quick
triage tool: flagged mechanisms point you to the specific audit slots that
warrant deeper investigation.


**Audit per learner**

```{r audit-by-learner}
if (requireNamespace("ranger", quietly = TRUE) && 
    requireNamespace("parsnip", quietly = TRUE)) {
    
    # 1. Define specs
    # Standard GLM (no tuning)
    spec_glm <- parsnip::logistic_reg() |> 
        parsnip::set_engine("glm")
    
    # Random Forest
    spec_rf <- parsnip::rand_forest(
        mode = "classification",
        trees = 100
    ) |>
        parsnip::set_engine("ranger")
    
    # 2. Fit using the current split object
    fit_multi <- fit_resample(
        df,
        outcome = "outcome",
        splits = splits,
        learner = list(glm = spec_glm, rf = spec_rf),
        metrics = "auc"
    )
    
    # 3. Run the audit
    audits <- audit_leakage_by_learner(fit_multi, metric = "auc", B = 20)
    cat("Per-learner audit summary:\n")
    print(audits)
    
} else {
    cat("ranger/parsnip not installed.\n")
}
```

This example uses **ranger** for the random forest specification; if it is not installed, the code chunk is skipped.

Use `parallel_learners = TRUE` to audit learners concurrently when **future.apply** is available.

The printed table summarizes each learner’s observed metric, permutation gap,
p-value, and key batch/duplicate summaries. Use it to compare signal strength
across models while checking for leakage risks. Pass `learners =` to audit a
subset, or `mc.cores =` to control parallel workers.


**HTML audit report**

`audit_report()` accepts either a `LeakAudit` or a `LeakFit` object. When a
`LeakFit` is provided, it first runs `audit_leakage()` and then forwards any
additional arguments to the audit step. If multiple learners were fit, pass
`learner =` via `...` to select one. Use `open = TRUE` to open the report in a
browser after rendering.

```{r audit-report, eval = FALSE}
if (requireNamespace("rmarkdown", quietly = TRUE) && rmarkdown::pandoc_available()) {
  report_path <- audit_report(audit, output_dir = ".")
  cat("HTML report written to:\n", report_path, "\n")
} else {
  cat("rmarkdown or pandoc not available; skipping audit report rendering.\n")
}
```

The report path points to a standalone HTML file containing the same audit tables and plots, suitable for sharing with collaborators or archiving as a quality control record.

## Time-series leakage checks

Time-series data require special handling. Random splits can leak information
from the future into the past. Use `mode = "time_series"` with a prediction
`horizon`, and audit with block permutations. Choose `time_block = "circular"`
or `"stationary"`; when `block_len` is NULL, the audit uses a default block
length (~10% of the test block size, minimum 5).

**Leaky example: lookahead feature**

```{r time-series-fit}
time_splits <- make_split_plan(
  df_time,
  outcome = "outcome",
  mode = "time_series",
  time = "time",
  v = 4,
  horizon = 1
)

cat("Time-series splits summary:\n")
time_splits

fit_time_leaky <- fit_resample(
  df_time,
  outcome = "outcome",
  splits = time_splits,
  learner = spec,
  metrics = "auc"
)

cat("Time-series leaky fit summary:\n")
summary(fit_time_leaky)
```

Time-series splitting trains on growing windows, so performance can differ from
standard CV simply because early folds have smaller training sets. Regardless of
the score, the presence of the `leak_future` feature makes this estimate
methodologically invalid.

**Leakage-safe alternative: remove lookahead and audit with blocks**

```{r time-series-safe}
df_time_safe <- df_time
df_time_safe$leak_future <- NULL

fit_time_safe <- fit_resample(
  df_time_safe,
  outcome = "outcome",
  splits = time_splits,
  learner = spec,
  metrics = "auc"
)

cat("Time-series safe fit summary:\n")
summary(fit_time_safe)

audit_time <- audit_leakage(
  fit_time_safe,
  metric = "auc",
  B = 20,
  time_block = "stationary",
  block_len = 5
)

cat("Time-series leakage audit summary:\n")
summary(audit_time)
if (!is.null(audit_time@permutation_gap) && nrow(audit_time@permutation_gap) > 0) {
  # Time-series permutation significance results
  audit_time@permutation_gap
}

if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_time_acf(fit_time_safe, lag.max = 20)
} else {
  cat("ggplot2 not installed; skipping ACF plot.\n")
}
```

The safe fit summary provides the leakage-resistant performance estimate. Compare
leaky vs. safe fits to gauge inflation risk; `features_final` should drop when
the lookahead feature is removed.

The ACF plot shows the autocorrelation of out-of-fold predictions by lag. The
dashed red lines represent the 95% confidence interval for white noise; large
bars outside the bands indicate residual temporal dependence. `plot_time_acf()`
requires numeric predictions and time metadata aligned to the fit.

## Cross-validation uncertainty with `cv_ci()`

Standard cross-validation yields one metric value per fold. Because folds share
training data, the per-fold values are positively correlated: the naive standard
error `sd / sqrt(K)` systematically underestimates variability, making the
resulting confidence intervals too narrow. `cv_ci()` addresses this with two
approaches:

- **`"normal"`** (default): uses the naive SEM `sd / sqrt(K)` with a
  *t*-distribution on `K − 1` degrees of freedom. Fast and appropriate when
  fold overlap is low (large datasets, many folds).

- **`"nadeau_bengio"`**: applies the Nadeau–Bengio (2003) variance correction
  `(1/K + n_test / n_train) × var(x)`, explicitly accounting for the fraction of
  training samples shared across folds. Recommended for small datasets or
  few-fold CV where fold correlation is non-negligible.

The function accepts any data frame with `fold` and `learner` columns — such as
`fit@metrics` from a `LeakFit` object — and returns one row per learner with
four columns per metric: `<metric>_mean`, `<metric>_sd`, `<metric>_ci_lo`, and
`<metric>_ci_hi`.

```{r cv-ci}
# Standard 95% CI from the leakage-safe fit
ci_std <- cv_ci(fit_safe@metrics, level = 0.95, method = "normal")
cat("Standard 95% CI:\n")
print(ci_std)

# Nadeau-Bengio corrected CI
# For K-fold CV: n_train ≈ (K-1)/K × n, n_test ≈ n/K
K_folds  <- length(unique(fit_safe@metrics$fold))
n_total  <- nrow(fit_safe@splits@info$coldata)

ci_nb <- cv_ci(
  fit_safe@metrics,
  level   = 0.95,
  method  = "nadeau_bengio",
  n_train = round(n_total * (K_folds - 1L) / K_folds),
  n_test  = round(n_total / K_folds)
)
cat("Nadeau-Bengio corrected 95% CI:\n")
print(ci_nb)
```

The Nadeau–Bengio interval is typically wider because it correctly inflates the
variance to account for fold overlap. Use it whenever the training-set fraction
shared between consecutive folds is large — for example, 5-fold CV reuses 80%
of training data across fold pairs, so correcting for this correlation matters.

When multiple learners were fit, `cv_ci()` returns one row per learner.
For repeated K-fold CV, average `n_train` and `n_test` across folds before
passing them in; `fit@metrics` already stores per-fold counts if `audit_leakage()`
annotated the object.

# Delta Leakage Sensitivity Index: Quantifying Performance Inflation

## Motivation

`audit_leakage()` detects whether leakage is present via permutation tests and
confounding checks. It cannot, however, directly answer: *by how much does
leakage inflate my reported performance?* The **Delta Leakage Sensitivity
Index** (ΔLSI) addresses this by rigorously comparing a naive (potentially
leaky) cross-validation pipeline against a leakage-guarded one run on the same
data.

## Mathematical framework

### Notation

Let $\mathcal{D} = \{(x_i, y_i)\}_{i=1}^n$ denote a dataset of $n$
observations. A repeated $K$-fold cross-validation design with $R$ independent
repeats partitions the observation indices into $RK$ folds. Let
$\mathcal{T}_{r,k} \subset \{1, \ldots, n\}$ denote the test index set for
fold $k$ in repeat $r$, with $n_{r,k} = |\mathcal{T}_{r,k}|$. Let
$m_{r,k}^{(p)}$ denote the performance metric evaluated on fold $(r,k)$ under
pipeline $p \in \{\text{naive},\, \text{guarded}\}$.

### Stage 1: Within-repeat aggregation

For each repeat $r$ and pipeline $p$, the $K$ fold-level metric values are
combined into a size-weighted repeat mean:

$$\mu_r^{(p)} = \frac{\displaystyle\sum_{k=1}^{K} n_{r,k}\, m_{r,k}^{(p)}}{\displaystyle\sum_{k=1}^{K} n_{r,k}}$$

The test-fold size $n_{r,k}$ is the weight, so folds with more held-out
observations contribute proportionally more to the repeat-level estimate. Folds
with missing metric values are excluded from both the numerator and denominator.
This produces the `@repeats_naive` and `@repeats_guarded` data frames (column
`metric`), one row per repeat.

### Stage 2: Per-repeat inflation score

For each repeat $r$, the signed performance gap between the two pipelines is:

$$\Delta_r = s \cdot \bigl(\mu_r^{\text{naive}} - \mu_r^{\text{guarded}}\bigr),
\qquad r = 1, \ldots, R$$

where

$$s = \begin{cases}
+1 & \text{if the metric is higher-is-better (e.g.\ AUC, accuracy)} \\
-1 & \text{if the metric is lower-is-better (e.g.\ RMSE, MAE, log-loss)}
\end{cases}$$

By construction $\Delta_r > 0$ if and only if the naive pipeline is more
optimistic than the guarded pipeline in repeat $r$, irrespective of metric
direction. Because both pipelines are evaluated on the same test fold
$\mathcal{T}_{r,k}$ (the paired design), any between-fold variability common to
both pipelines cancels in $\Delta_r$, concentrating the signal on genuine
pipeline differences rather than partition noise.

### Point estimates

Two summaries of the inflation vector
$\boldsymbol{\Delta} = (\Delta_1, \ldots, \Delta_R)^\top$ are reported.

**Arithmetic mean — `delta_metric`:**

$$\hat{\delta}_{\text{metric}} = \frac{1}{R} \sum_{r=1}^{R} \Delta_r$$

This is the most directly interpretable quantity. For AUC,
$\hat{\delta}_{\text{metric}} = 0.08$ means the naive pipeline's reported AUC
exceeds the guarded pipeline's by approximately $0.08$ units, averaged across
all repeats. This value is stored in `@delta_metric`.

**Huber M-estimate — `delta_lsi`:**

$$\hat{\delta}_{\text{LSI}} = \operatorname*{arg\,min}_{\mu} \sum_{r=1}^{R}
\rho_{k}\!\left(\frac{\Delta_r - \mu}{\hat{\sigma}}\right)$$

where $\rho_k$ is Huber's loss function with tuning constant $k = 1.345$:

$$\rho_k(u) = \begin{cases}
\tfrac{1}{2}u^2 & |u| \leq k \\
k|u| - \tfrac{1}{2}k^2 & |u| > k
\end{cases}$$

and $\hat{\sigma} = 1.4826 \times \operatorname{MAD}(\boldsymbol{\Delta})$ is a
fixed robust scale estimate. This minimisation is solved via iteratively
reweighted least squares (IRLS):

1. Initialise: $\hat{\mu}^{(0)} = \operatorname{median}(\boldsymbol{\Delta})$,
   $\hat{\sigma} = 1.4826 \times \operatorname{MAD}(\boldsymbol{\Delta})$
2. Compute standardised residuals:
   $u_r^{(t)} = (\Delta_r - \hat{\mu}^{(t)}) / \hat{\sigma}$
3. Assign Huber influence weights:
   $w_r^{(t)} = \min\!\left(1,\; k / |u_r^{(t)}|\right)$
4. Update location:
   $\hat{\mu}^{(t+1)} = \displaystyle\frac{\sum_{r=1}^{R} w_r^{(t)} \Delta_r}{\sum_{r=1}^{R} w_r^{(t)}}$
5. Repeat until $|\hat{\mu}^{(t+1)} - \hat{\mu}^{(t)}| < 10^{-8}(1 + |\hat{\mu}^{(t)}|)$

Note that $\hat{\sigma}$ is **fixed** at its initial value throughout iteration;
only the location $\hat{\mu}$ is updated. Outlier repeats with
$|u_r^{(t)}| > k$ receive weight $w_r < 1$, reducing their contribution to the
update. When $\hat{\sigma} = 0$ (all $\Delta_r$ identical), the estimator
returns the common value without iteration. This quantity is stored in
`@delta_lsi` and is the primary reported point estimate.

The two estimates coincide when $\{\Delta_r\}$ are symmetric and free of
outliers. When they differ substantially, `delta_lsi` is more trustworthy
because it is not pulled by a single anomalous repeat.

**Unpaired fallback.** When the two fits do not share the same fold structure
(`@info$paired = FALSE`), the paired vector $\boldsymbol{\Delta}$ cannot be
formed. The point estimates are instead computed from the separately summarised
repeat-mean vectors $\boldsymbol{\mu}^{\text{naive}}$ and
$\boldsymbol{\mu}^{\text{guarded}}$:

$$\hat{\delta}_{\text{metric}}^{\text{unpaired}}
  = s \cdot \bigl(\bar{\mu}^{\text{naive}} - \bar{\mu}^{\text{guarded}}\bigr),
\qquad
\hat{\delta}_{\text{LSI}}^{\text{unpaired}}
  = s \cdot \bigl(\hat{H}(\boldsymbol{\mu}^{\text{naive}})
               - \hat{H}(\boldsymbol{\mu}^{\text{guarded}})\bigr)$$

where $\bar{\mu}$ is the arithmetic mean and $\hat{H}$ is the Huber
M-estimator applied to each pipeline's repeat-means vector independently. No
inference (sign-flip test or BCa interval) is available in the unpaired case
($R_{\text{eff}} = 0$).

## Setting up a two-pipeline comparison

A ΔLSI analysis requires two `LeakFit` objects:

1. **Naive fit** — standard cross-validation without leakage safeguards (e.g.,
   random row splits, global preprocessing, or leaky features).
2. **Guarded fit** — leakage-aware cross-validation using subject-grouped splits
   and preprocessing fit only on training folds.

For the most powerful **paired** design, generate a single `LeakSplits` object
and pass it to both `fit_resample()` calls. Pairing is validated automatically
by comparing the sorted test-set indices of every fold; if the structures differ
despite equal repeat counts, a warning is issued and analysis falls back to
unpaired mode.

```{r delta-lsi-data}
set.seed(100)
n_d    <- 120L
subj_d <- rep(paste0("P", seq_len(30L)), each = 4L)   # 30 subjects, 4 obs each

df_dlsi_guarded <- data.frame(
  subject = subj_d,
  outcome = factor(sample(c("case", "control"), n_d, replace = TRUE)),
  x1      = rnorm(n_d),
  x2      = rnorm(n_d)
)

# Leaky feature: subject-level mean of outcome
# With subject-grouped splits, test subjects' leak value encodes their outcome
df_dlsi_naive <- df_dlsi_guarded
df_dlsi_naive$leak <- ave(
  as.numeric(df_dlsi_guarded$outcome == "case"),
  subj_d, FUN = mean
)

# Shared splits: 5-fold × 5 repeats → R_eff = 5 (tier C: point + p-value)
splits_dlsi <- make_split_plan(
  df_dlsi_guarded,
  outcome = "outcome",
  mode    = "subject_grouped",
  group   = "subject",
  v       = 5L,
  repeats = 5L
)
```

```{r delta-lsi-fits}
# Reuse the parsnip spec (base R glm) defined earlier
# Naive fit: uses leaky feature (leak encodes test-subject outcome label)
fit_dlsi_naive <- fit_resample(
  df_dlsi_naive,
  outcome = "outcome",
  splits  = splits_dlsi,
  learner = spec,
  metrics = "auc",
  seed    = 1L
)

# Guarded fit: same splits, clean features only
fit_dlsi_guarded <- fit_resample(
  df_dlsi_guarded,
  outcome = "outcome",
  splits  = splits_dlsi,
  learner = spec,
  metrics = "auc",
  seed    = 1L
)

cat("Naive   AUC:", round(mean(fit_dlsi_naive@metrics$auc,   na.rm = TRUE), 3), "\n")
cat("Guarded AUC:", round(mean(fit_dlsi_guarded@metrics$auc, na.rm = TRUE), 3), "\n")
```

## Running `delta_lsi()`

```{r delta-lsi-run}
result_dlsi <- delta_lsi(
  fit_leaky   = fit_dlsi_naive,
  fit_guarded = fit_dlsi_guarded,
  metric      = "auc",
  M_boot      = 300L,
  M_flip      = 300L,
  seed        = 1L
)

summary(result_dlsi)
```

`summary()` reports seven quantities; here is how to read each one:

- **Naive / Guarded pipeline values**: the Huber-robust mean AUC across repeats
  for each pipeline in isolation. Compare these directly to see the raw
  performance level of each pipeline before computing the gap.

- **`delta_metric`**: the arithmetic mean of the per-repeat inflation scores
  $\Delta_r$. Interpret this in the original metric's units: a value of 0.10
  means the naive pipeline reports AUC ~0.10 higher than the guarded pipeline,
  averaged across repeats. This is the most intuitive summary for reporting.

- **`delta_lsi`**: the Huber M-estimate of $\{\Delta_r\}$. Because it
  down-weights outlier repeats, it is the more reliable point estimate when
  repeat-to-repeat variation is high. Values very close to `delta_metric`
  indicate a consistent, non-skewed inflation pattern.

- **95% BCa CI** (shown when `R_eff ≥ 10`): a bootstrap confidence interval
  for `delta_lsi`. If the lower bound is above zero the inflation is unlikely
  to be a sampling artefact. If the interval spans zero the inflation estimate
  is uncertain given the number of repeats.

- **Sign-flip p-value** (shown when `R_eff ≥ 5`): tests H₀: no leakage
  inflation (mean $\Delta_r = 0$) via a randomization test. A small p-value
  (e.g., < 0.05) indicates the observed inflation is unlikely under the null
  of exchangeable pipeline labels.

- **Inference tier** and **`inference_ok`**: summarise how much statistical
  weight the design supports. Only tier A (`R_eff ≥ 20`, `inference_ok = TRUE`)
  provides a full combination of point estimate + CI + p-value. Lower tiers
  report what is reliable given the available repeats. The formal tier
  definitions and their rationale are given in the Statistical inference section
  below.

## Designing for a target inference tier

The formal tier thresholds have direct implications for how you set up your
repeated CV design. Plan the number of folds and repeats around the tier you
need:

- **Tier D** (`D_insufficient`, point estimates only): any $R_{\text{eff}} < 5$ — useful for
  exploratory comparisons where no formal test is required.
- **Tier C** (`C_signflip`, sign-flip p-value available): 5-fold $\times$ 5 repeats
  $\to R_{\text{eff}} = 5$. The smallest achievable p-value is $1/2^5 = 0.031$.
  For `exchangeability = "blocked_time"` the minimum p-value is $1/2^{n_{\text{blocks}}}$;
  at least 5 blocks are required for the test to be available.
- **Tier B** (`B_signflip_ci`, BCa CI additionally available): 5-fold $\times$ 10 repeats
  $\to R_{\text{eff}} = 10$. Both sign-flip p-value and BCa CI are reported.
- **Tier A** (`A_full_inference`, `inference_ok = TRUE`): 5-fold $\times$ 20
  repeats $\to R_{\text{eff}} = 20$. Recommended for any published or
  pre-registered ΔLSI analysis.

For paired designs `R_eff` equals the number of matched repeats. For unpaired
designs `R_eff = 0` and only the raw point estimates are available.

```{r delta-lsi-tier}
cat("Tier:          ", result_dlsi@tier, "\n")
cat("R_eff:         ", result_dlsi@R_eff, "\n")
cat("inference_ok:  ", result_dlsi@inference_ok, "\n")
```

## Statistical inference

### Sign-flip randomization test

The sign-flip test (Phipson & Smyth, 2010) assesses the null hypothesis that
there is no systematic performance inflation:

$$H_0: \mathbb{E}[\Delta_r] = 0 \quad \text{(no leakage-induced inflation)}$$

The test exploits sign-exchangeability: under $H_0$, each $\Delta_r$ is
symmetrically distributed around zero, so randomly flipping the sign of any
$\Delta_r$ leaves the distribution of the test statistic unchanged. The test
statistic is the arithmetic mean of the inflation scores:

$$T_{\text{obs}} = \frac{1}{R} \sum_{r=1}^{R} \Delta_r$$

For each sign vector $\boldsymbol{\epsilon} = (\epsilon_1, \ldots, \epsilon_R)$
with $\epsilon_r \in \{-1, +1\}$, a null-distribution realisation is:

$$T_{\boldsymbol{\epsilon}} = \frac{1}{R} \sum_{r=1}^{R} \epsilon_r \Delta_r$$

The two-sided p-value is the proportion of sign vectors yielding a statistic at
least as extreme as observed:

$$p = \frac{\#\bigl\{\boldsymbol{\epsilon} \in \mathcal{E} :
           |T_{\boldsymbol{\epsilon}}| \geq |T_{\text{obs}}|\bigr\}}{|\mathcal{E}|}$$

Two computation regimes are used:

- **Exact enumeration** ($R_{\text{eff}} \leq 15$): all $2^R$ sign vectors are
  evaluated — $|\mathcal{E}| = 2^R$. The p-value is $\text{count}/2^R$ with
  **no** continuity correction, because every sign combination including the
  observed one is explicitly enumerated. The minimum achievable p-value is
  $1/2^R$.

- **Monte Carlo** ($R_{\text{eff}} > 15$): $M_{\text{flip}}$ sign vectors are
  drawn uniformly from $\{-1,+1\}^R$. The Phipson & Smyth (2010) correction
  prevents $p = 0$ in the extreme tail:
$$p = \frac{B^{+} + 1}{M_{\text{flip}} + 1},
\qquad B^{+} = \#\{b : |T_{\boldsymbol{\epsilon}_b}| \geq |T_{\text{obs}}|\}$$

The arithmetic mean $\hat{\delta}_{\text{metric}}$ is used as the test
statistic — not the Huber estimate $\hat{\delta}_{\text{LSI}}$. Under a
sign-flip null the mean has a symmetric, well-defined exchangeable distribution;
the Huber M-estimator's nonlinear weighting changes with each sign assignment
and therefore does not preserve sign-exchangeability. As a result,
`@p_value` tests `@delta_metric`, not `@delta_lsi`.

This distinction matters in practice. When the per-repeat inflation vector
$\{\Delta_r\}$ contains outliers, `delta_lsi` (Huber) and `delta_metric` (mean)
can diverge in magnitude while agreeing in sign. A user can therefore see a
significant `@p_value` with a BCa CI for `delta_lsi` that spans zero — or vice
versa — if a handful of repeats pull the mean but not the robust estimate. The
`summary()` method flags this disagreement explicitly when it occurs.

The threshold $R_{\text{eff}} \geq 5$ for the sign-flip test follows from the
minimum achievable p-value: with $R = 4$, $1/2^4 = 0.0625 > 0.05$, so a
significant result at the conventional level is impossible regardless of the
data. With $R = 5$, $1/2^5 = 0.03125 < 0.05$, making it attainable. For
`exchangeability = "blocked_time"`, the relevant threshold is $n_{\text{blocks}} \geq 5$
(not $R_{\text{eff}} \geq 5$), since blocks are the independent units under the
test; `@p_value` is set to `NA` and a warning is issued when
$n_{\text{blocks}} < 5$. For other exchangeability values, `@p_value` is `NA`
when $R_{\text{eff}} < 5$ or when the design is unpaired.

### Huber M-estimator and robustness properties

The Huber M-estimator was described formally in the Mathematical framework
section above. Its inferential properties in this context are:

- **Breakdown point**: $\hat{\delta}_{\text{LSI}}$ remains a consistent
  estimate of the true location even when up to $\sim 29\%$ of repeats are
  outliers ($k = 1.345$ yields 95% asymptotic efficiency relative to the mean
  under a Gaussian distribution of $\{\Delta_r\}$).
- **Fixed scale**: $\hat{\sigma}$ is set to $1.4826 \times \operatorname{MAD}(\boldsymbol{\Delta})$
  before iteration and never updated. Fixing the scale prevents a single
  influential $\Delta_r$ from collapsing $\hat{\sigma}$ toward zero, which
  would force all weights to 1 and reduce the estimator to the ordinary mean.
- **Degenerate case**: when $\hat{\sigma} = 0$ (all repeat differences
  identical), the algorithm returns $\operatorname{median}(\boldsymbol{\Delta})$
  directly, which equals the common value.

The `delta_lsi` and `delta_metric` estimates always have the same sign because
the Huber estimator and the arithmetic mean agree on direction. They differ only
in magnitude, and then only when one or more $\Delta_r$ are extreme relative to
the bulk of the distribution.

### BCa bootstrap confidence interval

The BCa (bias-corrected and accelerated) interval (Efron, 1987) for
$\hat{\delta}_{\text{LSI}}$ corrects for both median bias and distributional
skewness in the bootstrap. Given $M_{\text{boot}}$ bootstrap resamples
$\boldsymbol{\Delta}_b^*$ drawn with replacement from $\boldsymbol{\Delta}$,
let $\hat{\theta}_b^* = \hat{H}(\boldsymbol{\Delta}_b^*)$ be the Huber estimate
on resample $b$ and $\hat{\theta} = \hat{H}(\boldsymbol{\Delta})$ the observed
estimate.

**Bias-correction constant** $\hat{z}_0$ accounts for the median bias of the
bootstrap distribution relative to the observed estimate:

$$\hat{z}_0 = \Phi^{-1}\!\left(
  \frac{\#\{b : \hat{\theta}_b^* < \hat{\theta}\}}{M_{\text{boot}}}
\right)$$

When $\hat{\theta}$ coincides with the exact median of the bootstrap
distribution, $\hat{z}_0 = 0$ and no bias correction is applied. A nonzero
$\hat{z}_0$ shifts both interval endpoints in the same direction.

**Acceleration constant** $\hat{a}$ corrects for skewness via the jackknife.
Let $\hat{\theta}_{(-r)} = \hat{H}(\boldsymbol{\Delta}_{(-r)})$ be the Huber
estimate with observation $r$ removed and
$\bar{\theta}_{(\cdot)} = R^{-1}\sum_{r=1}^R \hat{\theta}_{(-r)}$:

$$\hat{a} = \frac{\displaystyle\sum_{r=1}^{R}
                  \bigl(\bar{\theta}_{(\cdot)} - \hat{\theta}_{(-r)}\bigr)^3}
                 {6\,\left[\displaystyle\sum_{r=1}^{R}
                  \bigl(\bar{\theta}_{(\cdot)} - \hat{\theta}_{(-r)}\bigr)^2
                 \right]^{3/2}}$$

Here $\hat{a} > 0$ indicates a right-skewed sampling distribution;
$\hat{a} < 0$ indicates left skew.

**BCa-adjusted quantile levels**: for nominal coverage $1 - \alpha = 0.95$,
let $z_L = \Phi^{-1}(\alpha/2)$ and $z_U = \Phi^{-1}(1-\alpha/2)$. The
adjusted quantile levels are:

$$p_j = \Phi\!\left(\hat{z}_0 +
  \frac{\hat{z}_0 + z_j}{1 - \hat{a}\,(\hat{z}_0 + z_j)}
\right), \quad j \in \{L,\, U\}$$

The 95% BCa interval is the $p_L$ and $p_U$ empirical quantiles of the
bootstrap distribution $\{\hat{\theta}_b^*\}$:

$$\bigl[\hat{\theta}^*_{(p_L)},\; \hat{\theta}^*_{(p_U)}\bigr]$$

A separate BCa interval is computed for $\hat{\delta}_{\text{metric}}$
(using the arithmetic mean as the bootstrap statistic) and stored in
`@delta_metric_ci`. Both intervals are `NA` when $R_{\text{eff}} < 10$, because
fewer than 10 leave-one-out jackknife samples do not provide a reliable estimate
of $\hat{a}$.

### Pairing condition and the inference tier system

Two `LeakFit` objects are treated as **paired** if and only if they have the
same number of repeats and every fold shares an identical test-index
composition:

$$\operatorname{sort}\!\bigl(\mathcal{T}_{r,k}^{\text{naive}}\bigr)
= \operatorname{sort}\!\bigl(\mathcal{T}_{r,k}^{\text{guarded}}\bigr)
\quad \forall\, r,\, k$$

This is verified by comparing the sorted integer index vectors for each
sequential fold position. When paired, $R_{\text{eff}} = R$; when unpaired,
$R_{\text{eff}} = 0$ and all inference is disabled. Pairing is automatic when
both fits are produced from the same `LeakSplits` object.

All three inference procedures are gated by $R_{\text{eff}}$. The four-tier
system makes these dependencies explicit:

| Tier | Condition | `@delta_metric` | `@delta_lsi` | `@p_value` | `@delta_lsi_ci` | `@inference_ok` |
|---|---|:---:|:---:|:---:|:---:|:---:|
| `D_insufficient` | $R_{\text{eff}} < 5$ | ✓ | ✓ | `NA` | `NA` | `FALSE` |
| `C_signflip` | $5 \leq R_{\text{eff}} < 10$ | ✓ | ✓ | ✓ | `NA` | `FALSE` |
| `B_signflip_ci` | $10 \leq R_{\text{eff}} < 20$ | ✓ | ✓ | ✓ | ✓ | `FALSE` |
| `A_full_inference` | $R_{\text{eff}} \geq 20$ | ✓ | ✓ | ✓ | ✓ | `TRUE` |

The threshold $R_{\text{eff}} \geq 5$ for the sign-flip test under `"iid"`
exchangeability follows from the minimum achievable p-value under exact
enumeration ($1/2^5 = 0.03125 < 0.05$; $1/2^4 = 0.0625 > 0.05$). For
`exchangeability = "blocked_time"`, the independent units are blocks of
repeats, so the effective threshold is $n_{\text{blocks}} \geq 5$; when fewer
than five blocks are formed from the available repeats, `@p_value` is set to
`NA` even if $R_{\text{eff}} \geq 5$. The threshold $R_{\text{eff}} \geq 10$
for BCa intervals ensures at least 10 jackknife leave-one-out estimates for a
reliable $\hat{a}$. The `inference_ok = TRUE` flag at tier A requires all
three of $R_{\text{eff}} \geq 20$, a finite `@p_value`, and finite
`@delta_lsi_ci` to hold simultaneously.

The code below shows direct slot access to confirm that the numbers correspond
exactly to the quantities defined above:

```{r delta-lsi-components}
# delta_metric: arithmetic mean of {Δ_r}
cat("delta_metric:  ", round(result_dlsi@delta_metric, 4), "\n")
# delta_lsi: Huber M-estimate of {Δ_r} (k=1.345, fixed MAD scale)
cat("delta_lsi:     ", round(result_dlsi@delta_lsi,    4), "\n")
# delta_lsi_ci: 95% BCa interval for delta_lsi (NA below tier B)
cat("95% BCa CI:    [",
    paste(round(result_dlsi@delta_lsi_ci, 4), collapse = ", "), "]\n")
# p_value: sign-flip randomization p-value (NA below tier C or if unpaired)
cat("p-value:       ",
    if (is.na(result_dlsi@p_value)) "NA (tier D or unpaired)"
    else format(result_dlsi@p_value, digits = 3), "\n")
```

## Paired vs. unpaired designs

**Paired** (recommended): both fits are produced from the same `LeakSplits`
object. The per-repeat inflation scores $\Delta_r$ are computed as within-repeat
differences, eliminating between-fold variability. $R_{\text{eff}} = R$ and all
inference tiers are reachable. Equal repeat counts with differing index patterns
produce a warning and fall back to unpaired mode automatically.

**Unpaired**: the two fits have different fold structures or different repeat
counts. The point estimates use separately computed pipeline means rather than
paired differences; $R_{\text{eff}} = 0$ and all inference is disabled. This
mode is appropriate when the splitting scheme itself is part of what the
comparison is testing — for example, evaluating whether subject-grouped splits
alone (not a leaky feature) account for a performance difference.

```{r delta-lsi-unpaired, warning=TRUE}
# Unpaired: naive uses sample-wise CV, guarded uses subject-grouped splits
splits_rand <- make_split_plan(
  df_dlsi_naive,
  outcome = "outcome",
  mode    = "subject_grouped",
  group   = "row_id",
  v       = 5L,
  repeats = 5L
)

fit_naive_rand <- fit_resample(
  df_dlsi_naive,
  outcome = "outcome",
  splits  = splits_rand,
  learner = spec,
  metrics = "auc",
  seed    = 1L
)

r_unpaired <- suppressWarnings(
  delta_lsi(fit_naive_rand, fit_dlsi_guarded, metric = "auc", seed = 1L)
)

cat(sprintf("Paired:        %s\n", r_unpaired@info$paired))
cat(sprintf("R_eff:         %d  (0 = unpaired, inference disabled)\n", r_unpaired@R_eff))
cat(sprintf("delta_metric:  %.4f  (naive raw mean - guarded raw mean)\n",
            r_unpaired@delta_metric))
cat(sprintf("p_value:       %s\n",
            if (is.na(r_unpaired@p_value)) "NA" else r_unpaired@p_value))
```

## Metric direction with `higher_is_better`

For lower-is-better metrics (RMSE, MAE, log-loss, Brier score, deviance),
set `higher_is_better = FALSE` or let `delta_lsi()` auto-detect from the
metric name. Auto-detected lower-is-better metric names:
`rmse`, `mse`, `mae`, `log_loss`, `logloss`, `brier`, `error`, `loss`, `deviance`.

When `higher_is_better = FALSE`, the sign of `delta_metric` is flipped so that
a **positive** value still means the naive pipeline is inflated relative to
guarded (i.e., naive has a *lower* error, which is paradoxically *better*
performance — indicating leakage):

```{r delta-lsi-direction}
r_hib_true  <- suppressWarnings(delta_lsi(
  fit_dlsi_naive, fit_dlsi_guarded,
  metric           = "auc",
  higher_is_better = TRUE,
  seed             = 1L
))
r_hib_false <- suppressWarnings(delta_lsi(
  fit_dlsi_naive, fit_dlsi_guarded,
  metric           = "auc",
  higher_is_better = FALSE,
  seed             = 1L
))

cat(sprintf("hib = TRUE:   delta_metric = %+.4f  (positive = naive inflated)\n",
            r_hib_true@delta_metric))
cat(sprintf("hib = FALSE:  delta_metric = %+.4f  (sign flipped)\n",
            r_hib_false@delta_metric))
```

The `info$higher_is_better` slot records the value used (after auto-detection),
so analyses are fully reproducible:

```{r delta-lsi-hib-slot}
cat("higher_is_better used:", result_dlsi@info$higher_is_better, "\n")
```

## Accessing the `LeakDeltaLSI` object

`delta_lsi()` returns a `LeakDeltaLSI` S4 object. All slots are accessible
directly for programmatic downstream use:

| Slot | Type | Description |
|------|------|-------------|
| `@metric` | character | Metric name used (e.g. `"auc"`, `"rmse"`) |
| `@exchangeability` | character | Exchangeability assumption for the sign-flip test. `"blocked_time"` activates a block sign-flip procedure; `"by_group"` and `"within_batch"` are stored but inference still uses iid sign-flip (a warning is issued); `"iid"` (default) applies the standard independent sign-flip |
| `@tier` | character | Inference tier: `"A_full_inference"`, `"B_signflip_ci"`, `"C_signflip"`, or `"D_insufficient"` |
| `@R_eff` | integer | Effective paired repeats (0 when unpaired; all inference is based on this count) |
| `@delta_metric` | numeric | Arithmetic mean of per-repeat inflation scores $\bar{\Delta}$; interpret directly in metric units |
| `@delta_lsi` | numeric | Huber M-estimate of per-repeat inflation scores; more robust than `delta_metric` when repeat-level outliers are present |
| `@delta_metric_ci` | numeric[2] | 95% BCa bootstrap CI for `delta_metric` (both elements `NA` below tier B) |
| `@delta_lsi_ci` | numeric[2] | 95% BCa bootstrap CI for `delta_lsi` (both elements `NA` below tier B) |
| `@p_value` | numeric | Sign-flip randomization p-value testing H₀: mean inflation = 0 (`NA` below tier C or when unpaired) |
| `@inference_ok` | logical | `TRUE` only at tier A (`R_eff ≥ 20`), indicating full point + CI + p-value inference is available |
| `@folds_naive` | data.frame | Per-fold data for the naive pipeline: columns `fold`, `metric` (fold-level metric value), `repeat_id`, `n` (test-fold size) |
| `@folds_guarded` | data.frame | Per-fold data for the guarded pipeline: columns `fold`, `metric`, `repeat_id`, `n` — same structure as `@folds_naive` |
| `@repeats_naive` | data.frame | Per-repeat aggregates for the naive pipeline: columns `repeat_id`, `metric` (size-weighted mean across folds), `n_folds`, `total_n` |
| `@repeats_guarded` | data.frame | Per-repeat aggregates for the guarded pipeline: same structure as `@repeats_naive` |
| `@info` | list | Metadata: `paired` (logical), `R_naive`, `R_guarded` (number of repeats in each fit), `higher_is_better` (logical, after auto-detection), `metric_naive` and `metric_guarded` (Huber-robust overall mean for each pipeline), `block_size_used` and `n_blocks` (for `exchangeability = "blocked_time"`; `NA` otherwise), and optionally `delta_r` (per-repeat $\Delta_r$ vector) and the original fit objects when `return_details = TRUE` |

```{r delta-lsi-slots}
# Per-repeat summaries for custom plotting
head(result_dlsi@repeats_naive,   3)
head(result_dlsi@repeats_guarded, 3)

# Per-fold diagnostics
head(result_dlsi@folds_naive,   3)
head(result_dlsi@folds_guarded, 3)

# Metadata
str(result_dlsi@info)
```

The `@repeats_naive` and `@repeats_guarded` data frames are the primary
substrate for diagnostic visualisation. Plotting `metric` by `repeat_id` for
both pipelines side-by-side reveals whether the performance gap is consistent
across all repeats — a clean, reproducible offset indicates structural leakage
— or concentrated in a handful of repeats, which may indicate a data-quality
issue or an unstable fold partition.

The `@folds_naive` and `@folds_guarded` data frames go one level deeper,
providing the per-fold metric value along with the test-fold size (`n`) and the
repeat it belongs to (`repeat_id`). These are useful for identifying individual
folds that behave as outliers within a repeat, for example because a
small test partition happened to produce extreme metric values.

## Connection to `audit_leakage()`

`delta_lsi()` and `audit_leakage()` answer complementary questions:

| Question | Tool |
|----------|------|
| Does leakage produce non-random performance? | `audit_leakage()` — permutation gap + p-value |
| What mechanisms drive the leakage? | `audit_leakage()` — `batch_assoc`, `target_assoc`, `duplicates` |
| How large is the performance inflation? | `delta_lsi()` — `delta_metric`, `delta_lsi` |
| Is the inflation statistically significant? | `delta_lsi()` — sign-flip p-value |
| Do guards fully eliminate the inflation? | `delta_lsi()` — compare before/after guarding |

A recommended two-step workflow:

1. Run `audit_leakage(fit_leaky)` to detect leakage and diagnose which
   mechanism is responsible.
2. Build a guarded pipeline that addresses those mechanisms, then run
   `delta_lsi(fit_leaky, fit_guarded)` to quantify how much inflation the
   leakage introduced and whether guarding eliminated it.

```{r delta-lsi-audit-connect}
# Step 1: detect and characterise leakage
audit_naive_dlsi <- audit_leakage(fit_dlsi_naive, metric = "auc", B = 20L)
cat("Naive audit summary:\n")
summary(audit_naive_dlsi)

# Step 2: confirmed guards reduced inflation; check residual DLSI
cat("\ndelta_metric (naive vs guarded):", round(result_dlsi@delta_metric, 4), "\n")
if (!is.na(result_dlsi@p_value)) {
  cat("Sign-flip p-value:              ", format(result_dlsi@p_value, digits = 3), "\n")
}
```

A ΔLSI near zero after guarding and a non-significant p-value together provide
strong evidence that the leakage source has been eliminated. A persistent
positive ΔLSI after guarding indicates residual inflation — revisit the
preprocessing pipeline for remaining leakage pathways.

# Parallel Processing

`bioLeak` uses the `future` framework for parallelism (Windows, macOS, Linux).
`fit_resample()`, `audit_leakage()`, `audit_leakage_by_learner()`, and
`simulate_leakage_suite()` honor the active plan when `parallel = TRUE` (or
`parallel_learners = TRUE`).

```{r parallel-setup, eval=FALSE}
## future is a Suggests dependency. Install with install.packages("future")
## before running this code; otherwise library(future) errors.
library(future)

# Use multiple cores (works on all OS)
plan(multisession, workers = 4)

# Run a heavy simulation
sim <- simulate_leakage_suite(..., parallel = TRUE)

# Parallel folds or audits
# fit_resample(..., parallel = TRUE)
# audit_leakage(..., parallel = TRUE)
# audit_leakage_by_learner(..., parallel_learners = TRUE)

# Return to sequential processing
plan(sequential)
```

This chunk only configures the parallel plan, so it does not produce a result
object. Use it as a template before running compute-heavy functions.

## Simulation suite

`simulate_leakage_suite()` runs Monte Carlo simulations that inject specific
leakage mechanisms and then evaluates detection with the audit pipeline. It is
useful for validating your leakage checks before applying them to real data.
Available leakage types include `none`, `subject_overlap`, `batch_confounded`,
`peek_norm`, and `lookahead`. Use `signal_strength`, `prevalence`, `rho`, `K`,
`repeats`, `horizon`, and `preprocess` to control the simulation. The output is
a `LeakSimResults` data frame with one row per seed. Metrics are selected
automatically based on outcome type (AUC for binary classification).

```{r simulate-suite}
if (requireNamespace("glmnet", quietly = TRUE)) {
  sim <- simulate_leakage_suite(
    n = 80,
    p = 6,
    mode = "subject_grouped",
    learner = "glmnet",
    leakage = "subject_overlap",
    seeds = 1:2,
    B = 20,
    parallel = FALSE,
    signal_strength = 1
  )
  
  # Simulation results (first 6 rows)
  head(sim)
} else {
  cat("glmnet not installed; skipping simulation suite example.\n")
}
```

Each row corresponds to one simulation seed.

- `metric_obs` is the observed performance (AUC for this simulation). Higher
  values suggest stronger apparent signal, which can be inflated by leakage.

- `gap` and `p_value` indicate that the leaked signal is statistically distinguishable from random noise.

Use `metric_obs` to gauge the magnitude of the leakage effect, and `gap` to assess detection sensitivity.


## Objects and summaries

bioLeak uses S4 classes and list-like results to capture provenance and diagnostics:

- `LeakSplits`: returned by `make_split_plan()`, printed with `show()`.
  Use `check_split_overlap()` to verify no-overlap invariants on grouping
  columns.
- `LeakFit`: returned by `fit_resample()`, summarized with `summary()`
  includes fold diagnostics in `fit@info$fold_status`
- `LeakAudit`: returned by `audit_leakage()`, summarized with `summary()`.
  Now includes `audit@info$mechanism_summary` — a data.frame classifying
  leakage evidence into mechanism classes (`non_random_signal`,
  `confounding_alignment`, `proxy_target_leakage`, `duplicate_overlap`,
  `temporal_lookahead`).
- `LeakAuditList`: returned by `audit_leakage_by_learner()`, printed directly
- `GuardFit`: returned by `guard_fit()`, printed and summarized.
  Use `guard_to_recipe()` to convert guard preprocessing specs to a
  `recipes::recipe` for tidymodels workflows.
- `LeakImpute`: returned by `impute_guarded()` (guarded data + diagnostics)
- `LeakTune`: returned by `tune_resample()` (nested tuning summaries)
  includes `fold_status`, selected parameters, and optional `final_model` /
  `final_workflow` when `refit = TRUE`
- `LeakSimResults`: returned by `simulate_leakage_suite()` (per-seed results)
- `LeakDeltaLSI`: returned by `delta_lsi()` (paired pipeline comparison).
  Key slots: `@metric`, `@exchangeability`, `@tier`, `@R_eff`,
  `@delta_metric`, `@delta_metric_ci`, `@delta_lsi`, `@delta_lsi_ci`,
  `@p_value`, `@inference_ok`, `@folds_naive`, `@folds_guarded`,
  `@repeats_naive`, `@repeats_guarded`, `@info`.
  Summarise with `summary()`.

Set `options(bioLeak.strict = TRUE)` to activate strict leakage mode, which
escalates trained-recipe/workflow warnings to errors, warns on missing seeds,
and auto-runs overlap checks after splitting.

These objects store hashes and metadata that help reproduce and audit the
workflow. Inspect their slots (`@metrics`, `@predictions`, `@permutation_gap`,
`@duplicates`) to drill into specific leakage signals.

---
title: "bioLeak: Leakage-Aware Biomedical Modeling"
author: "Selçuk Korkmaz"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_document:
    toc: true
    toc_float: true
    number_sections: true
    theme: flatly
    highlight: tango
vignette: >
  %\VignetteIndexEntry{bioLeak: Leakage-Aware Biomedical Modeling}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE,
  eval = TRUE
)
```

# Why bioLeak

bioLeak is a leakage-aware modeling toolkit for biomedical and machine-learning
analyses. Its purpose is to prevent and diagnose information leakage across
resampling workflows where training and evaluation data are not truly
independent because samples share subjects, batches, studies, or time.

Standard workflows are often insufficient. Random, row-wise cross-validation
assumes samples are independent. Global preprocessing (imputation, scaling,
feature selection) done before resampling lets test-fold information shape the
training process. These choices inflate performance and can lead to incorrect
biomarker discovery, misleading clinical signals, or models that fail in
deployment.

Data leakage means any direct or indirect use of evaluation data in training or
feature engineering. In biomedical data, leakage commonly appears as:

- repeated measurements from the same patient split across folds
- batch or study effects that align with the outcome
- duplicates or near-duplicates across train and test
- time-series lookahead or random splits that use future information
- preprocessing that uses all samples instead of training-only statistics

bioLeak addresses these issues with leakage-aware splitting, guarded
preprocessing that is fit only on training data, and audit diagnostics that
surface overlaps, confounding, and duplicates.

# Guided workflow

The sections below walk through a leakage-aware workflow from data setup to
audits. Each step includes a leaky failure mode and a corrected alternative.

## Example data

```{r example-data}
library(bioLeak)

set.seed(123)
n <- 160
subject <- rep(seq_len(40), each = 4)
batch <- sample(paste0("B", 1:6), n, replace = TRUE)
study <- sample(paste0("S", 1:4), n, replace = TRUE)
time <- seq_len(n)

x1 <- rnorm(n)
x2 <- rnorm(n)
x3 <- rnorm(n)
linpred <- 0.7 * x1 - 0.4 * x2 + 0.2 * x3 + rnorm(n, sd = 0.5)
p <- stats::plogis(linpred)
outcome <- factor(ifelse(runif(n) < p, "case", "control"),
                  levels = c("control", "case"))

df <- data.frame(
  subject = subject,
  batch = batch,
  study = study,
  time = time,
  outcome = outcome,
  x1 = x1,
  x2 = x2,
  x3 = x3
)

df_leaky <- within(df, {
  leak_subject <- ave(as.numeric(outcome == "case"), subject, FUN = mean)
  leak_batch <- ave(as.numeric(outcome == "case"), batch, FUN = mean)
  leak_global <- mean(as.numeric(outcome == "case"))
})

df_time <- df
df_time$leak_future <- c(as.numeric(df_time$outcome == "case")[-1], 0)
predictors <- c("x1", "x2", "x3")


# Example data (first 6 rows)
head(df)

# Outcome class counts
as.data.frame(table(df$outcome))
```

The table preview displays the metadata columns (`subject`, `batch`, `study`, `time`), the binary outcome, and three numeric predictors (`x1`, `x2`, `x3`).

The class count table shows the baseline prevalence. Stratified splits try to preserve these proportions (for grouped modes, stratification is applied at the group level).


## Create leakage-aware splits with `make_split_plan()`

`make_split_plan()` is the foundation. It returns a `LeakSplits` object with
explicit train/test indices (or compact fold assignments) and metadata. For
data.frame inputs, a unique `row_id` column is added automatically, so
`group = "row_id"` is the explicit way to request sample-wise CV. It assumes
that the grouping columns you provide are complete and that samples sharing a
group must not cross folds. Grouped stratification uses the majority class per
group and is ignored for `study_loocv`, `time_series`, and survival outcomes.
Misuse to avoid:

- using `group = "row_id"` (sample-wise CV) when subjects repeat
- using the wrong grouping column (for example batch instead of subject)
- using time-series CV with unsorted or missing time values
- relying on stratification when the outcome is missing, single-class at the
  group level, or a time/event outcome
- assuming `time_series` will always return exactly `v` folds

**Leaky example: row-wise CV when subjects repeat**

```{r leaky-splits}
leaky_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "row_id",
  v = 5,
  repeats = 1,
  stratify = TRUE
)

cat("Leaky splits summary (sample-wise CV):\n")
leaky_splits
```

The printed `LeakSplits` summary reports the split mode, number of folds, and fold-level training and test set sizes.

Because `group = "row_id"`, each sample is treated as its own group. As a result, repeated samples from the same subject may appear in both the training and test sets, which introduces information leakage.


**Leakage-safe alternative: group by subject**

```{r safe-splits}
safe_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "subject",
  v = 5,
  repeats = 1,
  stratify = TRUE,
  seed = 10
)

cat("Leakage-safe splits summary (subject-grouped CV):\n")
safe_splits
```

Here, each subject is confined to a single fold. Test set sizes are multiples
of the number of samples per subject, confirming that subjects were not split
across folds.

The fold sizes remain comparable because `stratify = TRUE` balances outcome
proportions across folds while respecting subject-level boundaries.


**Other leakage-aware modes**

```{r splits-other-modes}
batch_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "batch_blocked",
  batch = "batch",
  v = 4,
  stratify = TRUE
)

cat("Batch-blocked splits summary:\n")
batch_splits

study_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "study_loocv",
  study = "study"
)

cat("Study leave-one-out splits summary:\n")
study_splits

time_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "time_series",
  time = "time",
  v = 4,
  horizon = 2
)

cat("Time-series splits summary:\n")
time_splits

nested_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "subject",
  v = 3,
  nested = TRUE,
  stratify = TRUE
)

cat("Nested CV splits summary:\n")
nested_splits
```

**RNG seed lineage**

bioLeak uses offset-based seeding so that every sub-operation gets a
deterministic, distinct seed without requiring isolated RNG streams:

| Function | Base seed | Per-unit offset |
|---|---|---|
| `make_split_plan()` | `set.seed(seed)` | repeats: `seed + 1000 * r`; nested: `seed + 1` |
| `fit_resample()` | `set.seed(seed)` | per-fold: `seed + fold_id` |
| `audit_leakage()` | `set.seed(seed)` | per-permutation: `seed + b` |

These are **not** isolated RNG streams but simple offsets. Collisions do not
occur for practical parameter values (e.g., `repeats < 1000`, `v < 1000`).
Always set an explicit `seed` argument for reproducibility; strict mode warns
when `seed` is `NULL` or `NA`.

Each summary reports the number of folds and the fold sizes for the corresponding split strategy.

- **Batch & Study:** The test set sizes vary because batches/studies are not evenly sized. `study_loocv` uses one study per fold, so `v` and `repeats` are ignored.

- **Time series:** The training set size increases across folds while the test set size follows the block size. Early blocks with no prior training data (or too-small test windows) are skipped, so you may see fewer than `v` folds.

**Combined (N-axis) mode**

When leakage can occur along multiple axes simultaneously (for example, both
subject and batch), `mode = "combined"` handles multi-axis constraint splitting.
The first element in `constraints` is the **primary fold driver** (determines
which groups go to test); subsequent elements are **exclusion constraints**
(training samples sharing constraint-axis levels with the test set are removed).

```{r combined-splits}
# For combined mode, constraint-axis levels should not span all folds.
# Here each subject belongs to exactly one site, so site exclusion only
# removes training samples from the same site as the test subjects.
df_comb <- df
df_comb$site <- paste0("site", rep(1:8, each = 5)[df_comb$subject])

combined_splits <- make_split_plan(
  df_comb,
  outcome = "outcome",
  mode = "combined",
  constraints = list(
    list(type = "subject", col = "subject"),
    list(type = "batch",   col = "site")
  ),
  v = 4,
  stratify = TRUE,
  seed = 42
)

cat("Combined (N-axis) splits summary:\n")
combined_splits
```

The fold sizes may be smaller than single-axis modes because training samples
are excluded whenever their site (or any constraint-axis level) also appears in
the test set. You can add more than two axes by appending additional elements to
`constraints`.

The legacy `primary_axis`/`secondary_axis` parameters are still accepted for
backward compatibility but `constraints` supersedes them and supports arbitrary
numbers of axes.

```{r combined-overlap-check}
# Verify zero overlap on all constraint axes
overlap_result <- check_split_overlap(combined_splits)
overlap_result
```

**Assumptions and intent by mode**

- `subject_grouped`: keeps all samples from a subject together
- `batch_blocked`: keeps batches/centers together (leave-one-group-out when `v`
  is at least the number of batches; otherwise multiple batches per fold)
- `study_loocv`: holds out each study in turn for external validation
- `time_series`: rolling-origin splits with an optional horizon to prevent lookahead
- `combined`: primary axis drives fold assignment; all additional constraint axes
  are enforced as exclusion constraints on the training set

Use `repeats` for repeated CV in grouped/batch modes (it is ignored for
`study_loocv` and `time_series`), `stratify = TRUE` to balance outcome
proportions at the group level when applicable, and `nested = TRUE` to attach
inner folds (one repeat). For large datasets, `progress = TRUE` reports
progress; storing explicit indices can be memory intensive.


## Validating splits with `check_split_overlap()`

`check_split_overlap()` verifies that no grouping-column levels appear in both
the training and test sets of any fold. It returns a data.frame with one row per
fold-column combination showing the overlap count and a pass/fail flag.

```{r check-split-overlap}
# Run on the subject-grouped splits from earlier
overlap_safe <- check_split_overlap(safe_splits)
overlap_safe
```

When `cols` is not supplied, the function auto-infers the relevant columns from
the split mode (e.g., `subject` for `subject_grouped`, all constraint axes for
`combined`). By default `stop_on_fail = TRUE`, so the function raises a
`bioLeak_overlap_error` when any overlap is detected. Set `stop_on_fail = FALSE`
to return the results without stopping.

When strict mode is active (`options(bioLeak.strict = TRUE)`),
`make_split_plan()` automatically runs `check_split_overlap()` after building
non-compact splits, so you do not need to call it manually.


## Scalability

**Handling Large Datasets (Compact Mode)**

For large datasets (e.g., $N > 50,000$) with many repeats, storing explicit
integer indices for every fold can consume gigabytes of memory. Use `compact = TRUE`
to store a lightweight vector of fold assignments instead. `fit_resample()` will
automatically reconstruct the indices on the fly.
Compact mode is not supported when `nested = TRUE` (it falls back to full
indices). For `time_series` compact splits, the time column must be present in
the stored split metadata so folds can be reconstructed.

```{r compact-splits}
# Efficient storage for large N
big_splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "subject",
  v = 5,
  compact = TRUE  # <--- Saves memory
)

cat("Compact-mode splits summary:\n")
big_splits
cat(sprintf("Compact storage enabled: %s\n", big_splits@info$compact))
```

The summary is identical to a regular split, but the underlying storage is a
compact fold-assignment vector. Use the `compact` flag to confirm the memory-
saving mode is active.

## Strict leakage mode

Setting `options(bioLeak.strict = TRUE)` activates a global safety switch that
tightens validation across the entire workflow. In strict mode:

1. **Trained recipe/workflow detection escalates to an error.** Normally,
   passing a pre-trained recipe or workflow to `fit_resample()` or
   `tune_resample()` produces a warning. In strict mode, it raises an error so
   fold-global leakage cannot proceed silently.

2. **Seed warnings.** When `seed` is `NULL` or `NA` in `make_split_plan()`,
   `fit_resample()`, or `tune_resample()`, strict mode emits a
   `bioLeak_validation_warning` reminding you to set an explicit seed for
   reproducibility.

3. **Combined + nested warning.** Using `nested = TRUE` with
   `mode = "combined"` may produce empty inner folds; strict mode warns about
   this before proceeding.

4. **Automatic overlap check.** After building non-compact splits,
   `make_split_plan()` auto-runs `check_split_overlap()` to verify that no
   grouping-column levels cross fold boundaries.

```{r strict-mode-demo, error=TRUE}
# Enable strict mode temporarily
withr::with_options(list(bioLeak.strict = TRUE), {
  strict_splits <- make_split_plan(
    df,
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 5,
    stratify = TRUE,
    seed = 42
  )
  cat("Strict mode splits completed without error.\n")
})
```

```{r strict-mode-trained-recipe, error=TRUE}
# Strict mode catches a pre-trained recipe
if (requireNamespace("recipes", quietly = TRUE)) {
  rec <- recipes::recipe(outcome ~ ., data = df[, c("outcome", predictors)]) |>
    recipes::step_normalize(recipes::all_numeric_predictors()) |>
    recipes::prep(training = df[, c("outcome", predictors)])

  withr::with_options(list(bioLeak.strict = TRUE), {
    tryCatch(
      fit_resample(
        df, outcome = "outcome",
        splits = safe_splits,
        learner = "glmnet",
        preprocess = rec
      ),
      error = function(e) cat("Strict mode error:", conditionMessage(e), "\n")
    )
  })
}
```

Strict mode is off by default. Use `withr::with_options()` for temporary
activation or set `options(bioLeak.strict = TRUE)` in your `.Rprofile` for
project-wide enforcement.


## Guarded preprocessing and imputation

`bioLeak` uses guarded preprocessing to prevent leakage from global imputation,
scaling, and feature selection. The low-level API is:

- `.guard_fit()` to fit preprocessing on training data only
- `predict_guard()` to apply the trained guard to new data
- `.guard_ensure_levels()` to align factor levels across train/test
- `impute_guarded()` as a convenience wrapper for leakage-safe imputation

`.guard_fit()` fits a pipeline that can winsorize, impute, normalize, filter,
and select features. It one-hot encodes non-numeric columns and carries factor
levels forward to new data. Assumptions and misuse to avoid:

- `fs = "ttest"` requires a binary outcome with enough samples per class
- `fs = "lasso"` requires `glmnet`
- `impute = "mice"` is not supported for guarded prediction
- `impute = "none"` with missing values will trigger a median fallback and
  missingness indicators to avoid errors

Supported preprocessing options include imputation (`median`, `knn`,
`missForest`, `none`), normalization (`zscore`, `robust`, `none`), filtering by
variance or IQR (optionally `min_keep`), and feature selection (`ttest`,
`lasso`, `pca`). Winsorization is controlled via `impute$winsor` and
`impute$winsor_k` in guarded steps; in `impute_guarded()`, use `winsor` and
`winsor_thresh`. `impute_guarded()` returns a `LeakImpute` object with guarded
data, diagnostics, and the fitted guard state.

**Leaky example: global scaling before CV**

```{r leaky-scaling}
df_leaky_scaled <- df
df_leaky_scaled[predictors] <- scale(df_leaky_scaled[predictors])
scaled_summary <- data.frame(
  feature = predictors,
  mean = colMeans(df_leaky_scaled[predictors]),
  sd = apply(df_leaky_scaled[predictors], 2, stats::sd)
)
scaled_summary$mean <- round(scaled_summary$mean, 3)
scaled_summary$sd <- round(scaled_summary$sd, 3)

# Leaky global scaling: means ~0 and SDs ~1 computed on all samples
scaled_summary
```

The summary shows that scaling used the full dataset, so test-fold statistics
influenced the training transformation. This violates the train/test separation
and can bias performance estimates.

**Leakage-safe alternative: fit preprocessing on training only**

```{r guarded-preprocess}
fold1 <- safe_splits@indices[[1]]
train_x <- df[fold1$train, predictors]
test_x <- df[fold1$test, predictors]

guard <- .guard_fit(
  X = train_x,
  y = df$outcome[fold1$train],
  steps = list(
    impute = list(method = "median", winsor = TRUE),
    normalize = list(method = "zscore"),
    filter = list(var_thresh = 0, iqr_thresh = 0),
    fs = list(method = "none")
  ),
  task = "binomial"
)

train_x_guarded <- predict_guard(guard, train_x)
test_x_guarded <- predict_guard(guard, test_x)

cat("GuardFit object:\n")
guard
cat("\nGuardFit summary:\n")
summary(guard)

# Guarded training data (first 6 rows)
head(train_x_guarded)

# Guarded test data (first 6 rows)
head(test_x_guarded)
```

The `GuardFit` object records the preprocessing steps and the number of
features retained after filtering. The summary reports how many features were
removed and the preprocessing audit trail. The guarded train/test previews show
that missing values were imputed and (if requested) scaled using training-only
statistics; the test data never influences these values.

**Factor level guard (advanced)**

```{r guard-ensure-levels}
raw_levels <- data.frame(
  site = c("A", "B", "B"),
  status = c("yes", "no", "yes"),
  stringsAsFactors = FALSE
)

level_state <- .guard_ensure_levels(raw_levels)

# Aligned factor data with consistent levels
level_state$data

# Levels map
level_state$levels
```

The returned `data` keeps factor levels consistent across folds, while the
`levels` list records the training-time levels (including any dummy levels added
to preserve one-hot columns). Use these when transforming new data to avoid
misaligned model matrices.

**Leaky example: imputation using train and test together**

```{r leaky-impute}
train <- data.frame(a = c(1, 2, NA, 4), b = c(NA, 1, 1, 0))
test <- data.frame(a = c(NA, 5), b = c(1, NA))

all_median <- vapply(rbind(train, test),
                     function(col) median(col, na.rm = TRUE),
                     numeric(1))
train_leaky <- as.data.frame(Map(function(col, m) { col[is.na(col)] <- m; col },
                                 train, all_median))
test_leaky <- as.data.frame(Map(function(col, m) { col[is.na(col)] <- m; col },
                                test, all_median))

# Leaky medians computed on train + test
data.frame(feature = names(all_median), median = all_median)

# Leaky-imputed training data
train_leaky

# Leaky-imputed test data
test_leaky
```

The medians above were computed using both train and test data, so the test set
influences the imputation values. This is a classic leakage pathway because
test information directly alters the training features.

**Leakage-safe alternative: `impute_guarded()`**

```{r safe-impute}
imp <- impute_guarded(
  train = train,
  test = test,
  method = "median",
  winsor = FALSE
)

# Guarded-imputed training data
imp$train

# Guarded-imputed test data
imp$test
```

Here, the imputation statistics are computed from the training set only. The missing value in the test set (column `a`) is replaced by 2, which is the training median. In contrast, in the leaky example above, it was replaced by 3, the global median.

This confirms that the test set is transformed using fixed values learned from the training data, preserving a clean separation between training and evaluation.
The `LeakImpute` object also contains missingness diagnostics in
`imp$summary$diagnostics`, and guarded outputs use the same encoding as
`.guard_fit()` (categorical variables are one-hot encoded). Use `vars` to
impute only a subset of columns when needed.

**Bridging guard steps to recipes: `guard_to_recipe()`**

`guard_to_recipe()` converts a guard preprocessing specification into a
`recipes::recipe`, enabling a smooth transition from guard-based to
tidymodels-based workflows. It maps supported guard steps to their closest
recipe equivalents:

| Guard step | Recipe equivalent |
|---|---|
| `impute$method = "median"` | `step_impute_median()` |
| `impute$method = "knn"` | `step_impute_knn()` |
| `normalize$method = "zscore"` | `step_normalize()` |
| `filter$var_thresh > 0` | `step_nzv()` |
| `fs$method = "pca"` | `step_pca()` |

Steps with no direct recipe equivalent (`missForest`, `mice`, `robust`
normalization, `ttest`, `lasso` feature selection) produce a
`bioLeak_fallback_warning` and either fall back to the closest available step or
are skipped.

```{r guard-to-recipe}
if (requireNamespace("recipes", quietly = TRUE)) {
  guard_steps <- list(
    impute    = list(method = "median"),
    normalize = list(method = "zscore")
  )

  rec <- guard_to_recipe(
    steps = guard_steps,
    formula = outcome ~ .,
    training_data = df[, c("outcome", predictors)]
  )

  cat("Converted recipe:\n")
  rec
}
```

The returned recipe is untrained and can be passed directly to `fit_resample()`
or `tune_resample()` as the `preprocess` argument.

## Fit and resample with `fit_resample()`

`fit_resample()` combines leakage-aware splits with guarded preprocessing and
model fitting. It supports:

- built-in learners (`glmnet`, `ranger`)
- `parsnip` model specifications or `workflows::workflow` objects (recommended)
- `recipes::recipe` preprocessing (prepped on training folds only)
- `rsample` resamples (`rset`/`rsplit`) via `splits`, with `split_cols` to map metadata
- custom learners via `custom_learners` (legacy/advanced)
- multiple metrics, class weights, positive class override, and yardstick metrics
- fold-level run diagnostics in `fit@info$fold_status` (`success`/`skipped`/`failed`)

Assumptions and misuse to avoid:

- outcome must be binary or multiclass (factor) for classification, numeric for
  regression, or a survival outcome (a `Surv` column or time/event columns)
- `positive_class` must be a single value that exists in the outcome levels
- `class_weights` only applies to binomial/multiclass tasks; workflows must
  handle weights internally
- if you supply a matrix without metadata, you must remove leakage columns
  yourself (group, batch, study, time)

Use `learner_args` to pass model-specific arguments. For custom learners
(advanced), you can supply separate `fit` and `predict` argument lists. Set
`parallel = TRUE` to use future.apply for fold-level parallelism when available.
When `learner` is a parsnip specification, `learner_args` and
`custom_learners` are ignored. When `learner` is a workflow, `preprocess` and
`learner_args` are ignored.
When a recipe or workflow is used, the built-in guarded preprocessing list is
bypassed, so ensure the recipe/workflow itself is leakage-safe.

Metrics used in this vignette:

- **AUC**: area under the ROC curve (0.5 = random, 1.0 = perfect; higher is better)
- **PR AUC**: area under the precision-recall curve (0 to 1; higher is better)
- **Accuracy**: fraction of correct predictions (0 to 1; higher is better)
- **RMSE**: root mean squared error (0 to infinity; lower is better)
- **Macro F1**: average F1 across classes (multiclass; higher is better)
- **Log loss**: cross-entropy for probabilities (lower is better)
- **C-index**: concordance for regression/survival (higher is better)

Always report the mean and variability across folds, not a single fold value.
When using yardstick metrics, the positive class is the second factor level;
set `positive_class` to control this.

**Parsnip model specification (recommended)**

```{r parsnip-spec}
spec <- parsnip::logistic_reg(mode = "classification") |>
  parsnip::set_engine("glm")
```

This spec uses base R `glm` under the hood and does not require extra model
packages. Use it in the examples below; custom learners are covered in the
Advanced section.

**Leaky example: leaky features and row-wise splits**

```{r fit-leaky}
fit_leaky <- fit_resample(
  df_leaky,
  outcome = "outcome",
  splits = leaky_splits,
  learner = spec,
  metrics = c("auc", "accuracy"),
  preprocess = list(
    impute = list(method = "median"),
    normalize = list(method = "zscore"),
    filter = list(var_thresh = 0),
    fs = list(method = "none")
  )
)

cat("Leaky fit summary:\n")
summary(fit_leaky)
metrics_leaky <- as.data.frame(fit_leaky@metric_summary)
num_cols <- vapply(metrics_leaky, is.numeric, logical(1))
metrics_leaky[num_cols] <- lapply(metrics_leaky[num_cols], round, digits = 3)

# Leaky fit: mean and SD of metrics across folds
metrics_leaky
```

The summary reports cross-validated performance. AUC ranges from 0.5 (random)
to 1.0 (perfect), while accuracy is the fraction of correct predictions.
Because the splits are leaky, these metrics can be artificially high and should
not be used for reporting.

**Leakage-safe alternative: grouped splits and clean predictors**

```{r fit-safe}
fit_safe <- fit_resample(
  df,
  outcome = "outcome",
  splits = safe_splits,
  learner = spec,
  metrics = c("auc", "accuracy"),
  preprocess = list(
    impute = list(method = "median"),
    normalize = list(method = "zscore"),
    filter = list(var_thresh = 0),
    fs = list(method = "none")
  ),
  positive_class = "case",
  class_weights = c(control = 1, case = 1),
  refit = TRUE
)

cat("Leakage-safe fit summary:\n")
summary(fit_safe)
metrics_safe <- as.data.frame(fit_safe@metric_summary)
num_cols <- vapply(metrics_safe, is.numeric, logical(1))
metrics_safe[num_cols] <- lapply(metrics_safe[num_cols], round, digits = 3)

# Leakage-safe fit: mean and SD of metrics across folds
metrics_safe

# Per-fold metrics (first 6 rows)
head(fit_safe@metrics)
```

`fit_resample()` returns a `LeakFit` object. Use `summary(fit_safe)` for a
compact report and inspect `fit_safe@metrics` and `fit_safe@predictions` for
details. When `refit = TRUE`, the final model and preprocessing state are stored
in `fit_safe@info$final_model` and `fit_safe@info$final_preprocess`.
Fold-level status diagnostics are available in `fit_safe@info$fold_status`.

The mean +/- SD table is the primary performance summary to report, while the
per-fold metrics reveal variability and potential instability across folds.

By default, `fit_resample()` stores refit inputs in `fit_safe@info$perm_refit_spec`
to enable `audit_leakage(perm_refit = "auto")`. Set `store_refit_data = FALSE`
to reduce memory and provide `perm_refit_spec` manually when you want refit-based
permutations. When multiple learners are passed, `refit = TRUE` refits only the
first learner; set `refit = FALSE` if you do not need a final model.

```{r fit-fold-status}
# Fold-level diagnostics for reproducible troubleshooting
head(fit_safe@info$fold_status, 5)
```

Interpretation of `fold_status`:

- `status = "success"` means at least one learner produced valid metrics on that fold.
- `status = "skipped"` typically means a data-structure issue in that fold (for example, single-class training data in classification).
- `status = "failed"` indicates a runtime failure (for example, model fitting or preprocessing error).

Use `reason` and `notes` to decide whether to:

- adjust split design (for example, stronger grouping, stratification, or fewer folds),
- adjust preprocessing/model complexity,
- or exclude unstable configurations before formal comparison.

If many folds are skipped/failed, do **not** trust aggregate metric means until the instability is resolved.

The examples above use a parsnip model specification. To swap in other models,
replace `spec` with another parsnip spec (see the gradient boosting example
below).

**Multiclass classification (optional)**

```{r multiclass-example}
if (requireNamespace("ranger", quietly = TRUE)) {
  set.seed(11)
  df_multi <- df
  df_multi$outcome3 <- factor(sample(c("A", "B", "C"),
                                     nrow(df_multi), replace = TRUE))

  multi_splits <- make_split_plan(
    df_multi,
    outcome = "outcome3",
    mode = "subject_grouped",
    group = "subject",
    v = 4,
    stratify = TRUE,
    seed = 11
  )

  fit_multi <- fit_resample(
    df_multi,
    outcome = "outcome3",
    splits = multi_splits,
    learner = "ranger",
    metrics = c("accuracy", "macro_f1", "log_loss"),
    refit = FALSE
  )

  cat("Multiclass fit summary:\n")
  summary(fit_multi)
} else {
  cat("ranger not installed; skipping multiclass example.\n")
}
```

Multiclass fits compute accuracy, macro-F1, and log loss when class
probabilities are available. `positive_class` is ignored for multiclass tasks.

**Survival outcomes (optional)**

```{r survival-example}
if (requireNamespace("survival", quietly = TRUE) &&
    requireNamespace("glmnet", quietly = TRUE)) {
  set.seed(12)
  df_surv <- df
  df_surv$time_to_event <- rexp(nrow(df_surv), rate = 0.1)
  df_surv$event <- rbinom(nrow(df_surv), 1, 0.7)

  surv_splits <- make_split_plan(
    df_surv,
    outcome = c("time_to_event", "event"),
    mode = "subject_grouped",
    group = "subject",
    v = 4,
    stratify = FALSE,
    seed = 12
  )

  fit_surv <- fit_resample(
    df_surv,
    outcome = c("time_to_event", "event"),
    splits = surv_splits,
    learner = "glmnet",
    metrics = "cindex",
    refit = FALSE
  )

  cat("Survival fit summary:\n")
  summary(fit_surv)
} else {
  cat("survival or glmnet not installed; skipping survival example.\n")
}
```

For survival tasks, supply `outcome = c(time_col, event_col)` or a `Surv`
column; `stratify` is ignored for time/event outcomes and `class_weights` are
not used.

**Passing learner-specific arguments (optional)**

```{r learner-args}
if (requireNamespace("glmnet", quietly = TRUE)) {
  fit_glmnet <- fit_resample(
    df,
    outcome = "outcome",
    splits = safe_splits,
    learner = "glmnet",
    metrics = "auc",
    learner_args = list(glmnet = list(alpha = 0.5)),
    preprocess = list(
      impute = list(method = "median"),
      normalize = list(method = "zscore"),
      filter = list(var_thresh = 0),
      fs = list(method = "none")
    )
  )
  cat("GLMNET summary with learner-specific arguments:\n")
  summary(fit_glmnet)
} else {
  cat("glmnet not installed; skipping learner_args example.\n")
}
```

This summary reflects the same guarded preprocessing but a different model
configuration (here, `alpha = 0.5`). Use the mean +/- SD metrics to compare
learner settings under identical splits.

**SummarizedExperiment inputs (optional)**

```{r se-example}
if (requireNamespace("SummarizedExperiment", quietly = TRUE)) {
  se <- SummarizedExperiment::SummarizedExperiment(
    assays = list(counts = t(as.matrix(df[, predictors]))),
    colData = df[, c("subject", "batch", "study", "time", "outcome"), drop = FALSE]
  )

  se_splits <- make_split_plan(
    se,
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 3
  )

  se_fit <- fit_resample(
    se,
    outcome = "outcome",
    splits = se_splits,
    learner = spec,
    metrics = "auc"
  )
  cat("SummarizedExperiment fit summary:\n")
  summary(se_fit)
} else {
  cat("SummarizedExperiment not installed; skipping SE example.\n")
}
```

The summary is identical in structure to the `data.frame` case because `fit_resample()` extracts predictors and metadata from the `SummarizedExperiment` object.

Note that `features_final` here reflects only the assay predictors (`x1`, `x2`,
`x3`), because the assay was constructed without metadata columns.


## Tidymodels interoperability

`bioLeak` integrates with `rsample`, `recipes`, `workflows`, and `yardstick`:

- `splits` can be an `rsample` rset/rsplit; use `split_cols` (default `"auto"`)
  or `bioLeak_mode` / `bioLeak_perm_mode` attributes to map
  group/batch/study/time metadata.
- `preprocess` can be a `recipes::recipe` (prepped on training folds only).
- `learner` can be a `workflows::workflow`.
- `metrics` can be a `yardstick::metric_set`.
- `as_rsample()` converts `LeakSplits` to an `rsample` object.

```{r tidymodels-interop}
library(bioLeak)
library(parsnip)
library(recipes)
library(yardstick)

set.seed(123)
N <- 60
df <- data.frame(
  subject = factor(rep(paste0("S", 1:20), length.out = N)), 
  outcome = factor(rep(c("ClassA", "ClassB"), length.out = N)),
  x1 = rnorm(N),
  x2 = rnorm(N),
  x3 = rnorm(N)
)

spec <- logistic_reg() |> set_engine("glm")

# Use bioLeak's native split planner to avoid conversion errors
set.seed(13)

# Use make_split_plan instead of rsample::group_vfold_cv
# This creates a subject-grouped CV directly compatible with fit_resample
splits <- make_split_plan(
  df, 
  outcome = "outcome", 
  mode = "subject_grouped", 
  group = "subject", 
  v = 3
)

rec <- recipes::recipe(outcome ~ x1 + x2 + x3, data = df) |>
  recipes::step_impute_median(recipes::all_numeric_predictors()) |>
  recipes::step_normalize(recipes::all_numeric_predictors())

metrics_set <- yardstick::metric_set(yardstick::roc_auc, yardstick::accuracy)

fit_rs <- fit_resample(
  df,
  outcome = "outcome",
  splits = splits,
  learner = spec,
  preprocess = rec,
  metrics = metrics_set,
  refit = FALSE
)

if (exists("as_rsample", where = asNamespace("bioLeak"), mode = "function")) {
    rs_export <- as_rsample(fit_rs@splits, data = df)
    print(rs_export)
}
```

Use `split_cols` to ensure split-defining metadata are dropped from predictors
when using rsample inputs.

```{r workflow-example}
if (requireNamespace("workflows", quietly = TRUE)) {
  wf <- workflows::workflow() |>
    workflows::add_model(spec) |>
    workflows::add_formula(outcome ~ x1 + x2 + x3)

  fit_wf <- fit_resample(
    df,
    outcome = "outcome",
    splits = splits,
    learner = wf,
    metrics = "auc",
    refit = FALSE
  )

  cat("Workflow fit summary:\n")
  summary(fit_wf)
} else {
  cat("workflows not installed; skipping workflow example.\n")
}
```

When auditing rsample-based fits, pass `perm_mode = "subject_grouped"` (or set
`attr(rs, "bioLeak_perm_mode") <- "subject_grouped"`) so restricted permutations
respect the intended split design.


## Nested tuning with `tune_resample()`

`tune_resample()` runs leakage-aware nested tuning with tidymodels. It accepts a
parsnip model specification or workflow, and supports two inner-CV modes:

- precomputed inner folds (for example from `make_split_plan(..., nested = TRUE)`)
- on-the-fly inner split generation for `LeakSplits` using `inner_v`,
  `inner_repeats`, and `inner_seed`

For rsample inputs, inner folds must already be present. Survival tasks are not
yet supported.

```{r tune-example}
if (requireNamespace("parsnip", quietly = TRUE) &&
    requireNamespace("recipes", quietly = TRUE) &&
    requireNamespace("tune", quietly = TRUE) &&
    requireNamespace("glmnet", quietly = TRUE)) {
  # --- 1. Create Data ---
  set.seed(123)
  N <- 60
  df <- data.frame(
    subject = factor(rep(paste0("S", 1:20), length.out = N)),
    outcome = factor(sample(c("ClassA", "ClassB"), N, replace = TRUE)),
    x1 = rnorm(N),
    x2 = rnorm(N),
    x3 = rnorm(N)
  )

  # --- 2. Generate Nested Splits ---
  set.seed(1)
  nested_splits <- make_split_plan(
    df,
    outcome = "outcome",
    mode = "subject_grouped",
    group = "subject",
    v = 3,
    nested = TRUE
  )

  # --- 3. Define Recipe & Model ---
  rec <- recipes::recipe(outcome ~ x1 + x2 + x3, data = df) |>
    recipes::step_impute_median(recipes::all_numeric_predictors()) |>
    recipes::step_normalize(recipes::all_numeric_predictors())

  spec_tune <- parsnip::logistic_reg(
    penalty = tune::tune(),
    mixture = 1,
    mode = "classification"
  ) |>
    parsnip::set_engine("glmnet")

  # --- 4. Run Tuning ---
  tuned <- tune_resample(
    df,
    outcome = "outcome",
    splits = nested_splits,
    learner = spec_tune,
    preprocess = rec,
    inner_v = 2,
    grid = 3,
    metrics = c("auc", "accuracy"),
    selection = "one_std_err",
    refit = TRUE,
    seed = 14
  )

  summary(tuned)
} else {
  cat("parsnip/recipes/tune/glmnet not installed; skipping nested tuning example.\n")
}
```

```{r tune-diagnostics}
if (exists("tuned")) {
  # Fold-level status from the outer loop
  head(tuned$fold_status, 5)

  # Final tuned model is available when refit = TRUE
  is.null(tuned$final_model)
} else {
  cat("Nested tuning object not available (dependencies missing).\n")
}
```

How to interpret this tuning output:

- `selection = "one_std_err"` chooses a model within one standard error of the best inner-CV score, then applies a deterministic simplicity tie-break. This usually favors more regularized/simpler models with similar expected performance.
- `tuned$metrics` and `tuned$metric_summary` remain the primary unbiased performance estimates (outer folds only).
- `tuned$final_model` is the deployable model fit on full data using the selected hyperparameters; it is for downstream prediction, not for estimating generalization.
- `tuned$fold_status` should be reviewed before reporting results. A large number of skipped/failed folds indicates unstable tuning and should trigger split/parameter redesign.


## Advanced: Using Gradient Boosting with Parsnip

`bioLeak` natively supports `tidymodels` specifications. You can pass a `parsnip`
model specification directly to `fit_resample()`. This allows you to use
state-of-the-art algorithms like XGBoost, LightGBM, or SVMs while ensuring all
preprocessing remains guarded.

```{r fit-xgboost}
if (requireNamespace("parsnip", quietly = TRUE) &&
    requireNamespace("xgboost", quietly = TRUE) &&
    requireNamespace("recipes", quietly = TRUE)) {
  
  # 1. Define the model spec
  xgb_spec <- parsnip::boost_tree(
    mode = "classification",
    trees = 100,
    tree_depth = 6,
    learn_rate = 0.01
  ) |>
    parsnip::set_engine("xgboost")
  
  # 2. Define a recipe on data without 'subject' because split metadata
  # columns are excluded from predictors by fit_resample().
  df_for_rec <- df[, !names(df) %in% "subject"]
  
  rec_xgb <- recipes::recipe(outcome ~ ., data = df_for_rec) |>
    recipes::step_dummy(recipes::all_nominal_predictors()) |>
    recipes::step_impute_median(recipes::all_numeric_predictors())
  # Note: No need for step_rm(subject) because it's already gone!
  
  # 3. Fit
  fit_xgb <- fit_resample(
    df,
    outcome = "outcome",
    splits = splits,
    learner = xgb_spec,
    metrics = "auc",
    preprocess = rec_xgb 
  )
  
  cat("XGBoost parsnip fit summary:\n")
  print(summary(fit_xgb))
} else {
  cat("parsnip/xgboost/recipes not installed.\n")
}
```

The summary reports cross-validated AUC for a non-linear gradient boosting
model. Use the mean $\pm$ SD table to compare against baseline models, and
confirm that any gains do not coincide with leakage signals in the audit
diagnostics.


## Advanced: Custom learners

Custom learners are used when a model is not available as a *parsnip* specification or when a lightweight wrapper around base R is needed.

Each custom learner must define `fit` and `predict` components. The `fit` function must accept `x`, `y`, `task`, and `weights` as inputs, while the `predict` function must accept `object` and `newdata`.

```{r custom-learners}
custom_learners <- list(
  glm = list(
    fit = function(x, y, task, weights, ...) {
      df_fit <- data.frame(y = y, x, check.names = FALSE)
      stats::glm(y ~ ., data = df_fit,
                 family = stats::binomial(), weights = weights)
    },
    predict = function(object, newdata, task, ...) {
      as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
                                type = "response"))
    }
  )
)

cat("Custom learner names:\n")
names(custom_learners)
cat("Custom learner components (fit/predict):\n")
lapply(custom_learners, names)
```

This output confirms that each custom learner provides both a `fit` and a `predict` function.

Custom learners can be used with `fit_resample()` as follows:

```r
fit_resample(
  ...,
  learner = "glm",
  custom_learners = custom_learners
)
```

## Visual diagnostics

**bioLeak** includes plotting helpers for quick diagnostic checks:

- `plot_fold_balance()` shows class balance per fold.  
- `plot_overlap_checks()` highlights train/test overlap for a metadata column.  
- `plot_time_acf()` checks autocorrelation in time-series predictions.
- `plot_calibration()` shows reliability of binomial probabilities.  
- `plot_confounder_sensitivity()` summarizes performance by batch/study strata.

Tabular helpers are also available:

- `calibration_summary()` returns calibration curve data and ECE/MCE/Brier metrics.
- `confounder_sensitivity()` returns per-stratum performance summaries.

**Misuse to avoid**

- Plotting without predictions (fits with no successful folds).  
- Using `plot_overlap_checks()` when the column is not present in the split metadata.  
- Using `plot_time_acf()` without a time column or for non-temporal data.
- Using `plot_calibration()` outside binomial tasks.
- Using `plot_confounder_sensitivity()` with a metric unsupported by the task.

For classification, `plot_fold_balance()` uses the positive class recorded in
`fit@info$positive_class`, or defaults to the second factor level if this is not
set. For multiclass tasks, it shows per-class counts without a proportion line.

```{r plot-fold-balance}
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_fold_balance(fit_safe)
} else {
  cat("ggplot2 not installed; skipping fold balance plot.\n")
}
```

The bar chart shows the counts of Positives (blue) and Negatives (tan) in each
fold. The dashed blue line represents the proportion of the positive class.

In a well-stratified fit, this line remains relatively stable across folds.
Large fluctuations indicate poor stratification, which can lead to unstable
fold-level performance estimates.

```{r plot-calibration}
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_calibration(fit_safe, bins = 10)
} else {
  cat("ggplot2 not installed; skipping calibration plot.\n")
}
```

Calibration curves compare predicted probabilities to observed event rates. Large deviations from the diagonal indicate miscalibration.
Use `min_bin_n` to suppress tiny bins and `learner =` to select a model when
multiple learners are stored in the fit.

```{r plot-confounder-sensitivity}
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_confounder_sensitivity(fit_safe, confounders = c("batch", "study"),
                              metric = "auc", min_n = 3)
} else {
  cat("ggplot2 not installed; skipping confounder sensitivity plot.\n")
}
```

Confounder sensitivity plots highlight whether performance varies substantially across batch or study strata.

```{r diagnostics-tables}
cal <- calibration_summary(fit_safe, bins = 10, min_bin_n = 5)
conf_tbl <- confounder_sensitivity(fit_safe,
                                   confounders = c("batch", "study"),
                                   metric = "auc",
                                   min_n = 3)

# Calibration metrics
cal$metrics

# Confounder sensitivity table (first 6 rows)
head(conf_tbl)
```


```{r plot-overlap}
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_overlap_checks(fit_leaky, column = "subject")
  plot_overlap_checks(fit_safe, column = "subject")
} else {
  cat("ggplot2 not installed; skipping overlap plots.\n")
}
```

The overlap plots compare the number of unique subjects appearing in the training and test sets for each fold.

**Top plot (`fit_leaky`)**  
The red line shows high overlap counts, accompanied by the explicit *“WARNING: Overlaps detected!”* annotation. This confirms that the same subjects appear in both the training and test sets, indicating information leakage.

**Bottom plot (`fit_safe`)**  
The overlap line remains flat at zero across all folds. This confirms that `subject_grouped` splitting successfully keeps subjects isolated, ensuring that no subject appears in both sets simultaneously.


## Audit leakage with `audit_leakage()`

`audit_leakage()` combines four diagnostics in one object:

- permutation gap: tests whether model signal is non-random
- batch or study association with folds (chi-square and Cramers V)
- target leakage scan: feature-wise outcome association to flag proxy columns
- duplicate and near-duplicate detection across train/test (by default) using feature similarity

Assumptions and misuse to avoid:

- `coldata` (or stored split metadata) must align with prediction IDs; rownames
  or a `row_id` column help alignment
- `X_ref` must align to prediction IDs (rownames, `row_id`, or matching order)
- `plot_perm_distribution()` requires `return_perm = TRUE`
- `target_threshold` should be set high enough to flag only strong proxies
- target leakage scan includes univariate associations plus an optional
  multivariate/interaction check (`target_scan_multivariate`, `*_components`,
  `*_interactions`, `*_B`); proxies outside `X_ref` can still pass undetected
- for rsample splits, set `perm_mode` (or `bioLeak_perm_mode`) so restricted
  permutations respect the intended design
- for time-series audits, ensure a valid time column and set `time_block` /
  `block_len` as needed

**Interpretation Note:** The label-permutation test refits models by default when
refit inputs are available (`perm_refit = "auto"` with `store_refit_data = TRUE`)
and `B <= perm_refit_auto_max`. Otherwise it keeps predictions fixed, so the
p-value quantifies prediction-label association rather than a full refit null.
A large gap indicates a non-random association. To determine if that signal is
*real* or *leaked*, check the **Batch Association** (confounding),
**Target Leakage Scan** (proxy features), and **Duplicate Detection**
(memorization) tables. Use `perm_refit = FALSE` to force fixed-prediction
shuffles or `perm_refit = TRUE` with `perm_refit_spec` to always refit.

Use `feature_space = "rank"` to compare samples by rank profiles, and
`sim_method` (`cosine` or `pearson`) to control similarity. For large `n`, `nn_k`
and `max_pairs` limit duplicate searches. Use `duplicate_scope = "all"` to
include within-fold duplicates (data-quality checks). Use `ci_method = "if"`
(influence-function) or `ci_method = "bootstrap"` with `boot_B` to obtain a
confidence interval for the permutation gap; `include_z` controls whether a
z-score is reported. Set `perm_stratify = "auto"` to stratify permutations only
when outcomes support it.

**Leakage-aware audit**

```{r audit-leakage}
# Use df_leaky (the original 160-row data) so X_ref aligns with fit_safe's predictions.
# df may have been redefined to a smaller dataset in the tidymodels section above.
X_ref <- df_leaky[, predictors]
X_ref[c(1, 5), ] <- X_ref[1, ]

audit <- audit_leakage(
  fit_safe,
  metric = "auc",
  B = 20,
  perm_stratify = TRUE,
  batch_cols = c("batch", "study"),
  X_ref = X_ref,
  sim_method = "cosine",
  sim_threshold = 0.995,
  return_perm = TRUE
)

cat("Leakage audit summary:\n")
summary(audit)
if (!is.null(audit@permutation_gap) && nrow(audit@permutation_gap) > 0) {
  # Permutation significance results
  audit@permutation_gap
}
if (!is.null(audit@batch_assoc) && nrow(audit@batch_assoc) > 0) {
  # Batch/study association with folds (Cramer's V)
  audit@batch_assoc
} else {
  cat("No batch or study associations detected.\n")
}
if (!is.null(audit@target_assoc) && nrow(audit@target_assoc) > 0) {
  # Top features by target association score
  head(audit@target_assoc)
} else {
  cat("No target leakage scan results available.\n")
}
if (!is.null(audit@duplicates) && nrow(audit@duplicates) > 0) {
  # Top duplicate/near-duplicate pairs by similarity
  head(audit@duplicates)
} else {
  cat("No near-duplicates detected.\n")
}
```

The permutation table reports the observed metric, the mean under random label permutation, the gap (difference), and a permutation *p*-value. For metrics where higher values indicate better performance, larger gaps reflect stronger non-random signal.

The batch association table reports chi-square statistics and Cramer's V. Large
p-values and small V values indicate that folds are not aligned with batch or
study labels (which is the desired outcome when these should be independent).

The target scan table lists features with the strongest associations with the
outcome. For numeric features, the score is \(|\mathrm{AUC} - 0.5| \times 2\) for
classification or the absolute correlation for regression. For categorical
features, the score is Cramér’s V or eta-squared. Scores closer to 1 indicate
stronger outcome association.

The duplicate table lists pairs of samples with near-identical profiles that
cross train/test folds (by default). In this setup, the artificially duplicated
pair (`X_ref[c(1, 5), ] <- X_ref[1, ]`) should appear near the top of the list.
Use `duplicate_scope = "all"` to include within-fold duplicates.

**Mechanism risk assessment**

`summary(audit)` now includes a **Mechanism Risk Assessment** section that
classifies leakage evidence into four mechanism classes:

| Mechanism class | Evidence slot | Flagging rule |
|---|---|---|
| `non_random_signal` | permutation gap | p ≤ 0.05 and gap > 0 |
| `confounding_alignment` | batch association | p ≤ 0.05 and Cramér's V ≥ 0.1 |
| `proxy_target_leakage` | target scan | any feature flagged (FDR or raw) |
| `duplicate_overlap` | duplicates | any cross-fold duplicates present |
| `temporal_lookahead` | duplicates | cross-fold duplicates in time-series splits |

The summary is stored in `audit@info$mechanism_summary` and can be accessed
directly for programmatic triage:

```{r mechanism-summary}
mech <- audit@info$mechanism_summary
if (is.data.frame(mech) && nrow(mech) > 0) {
  mech
} else {
  cat("No mechanism summary available.\n")
}
```

Each row reports whether the mechanism was flagged, the evidence type, and the
associated test statistic and p-value. This provides a quick triage overview
before drilling into individual audit slots.

```{r plot-perm}
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_perm_distribution(audit)
} else {
  cat("ggplot2 not installed; skipping permutation plot.\n")
}
```

The histogram shows the null distribution (gray bars) of the performance metric under random label permutation.

The blue dashed line represents the average performance of a random model (the permuted mean). The red solid line represents the observed performance of the fitted model.

A genuine signal is indicated when the red line lies well to the right of the
gray distribution; overlap suggests weak or unstable signal.

Use the mechanism risk assessment (`audit@info$mechanism_summary`) as a quick
triage tool: flagged mechanisms point you to the specific audit slots that
warrant deeper investigation.


**Audit per learner**

```{r audit-by-learner}
if (requireNamespace("ranger", quietly = TRUE) && 
    requireNamespace("parsnip", quietly = TRUE)) {
    
    # 1. Define specs
    # Standard GLM (no tuning)
    spec_glm <- parsnip::logistic_reg() |> 
        parsnip::set_engine("glm")
    
    # Random Forest
    spec_rf <- parsnip::rand_forest(
        mode = "classification",
        trees = 100
    ) |>
        parsnip::set_engine("ranger")
    
    # 2. Fit using the current split object
    fit_multi <- fit_resample(
        df,
        outcome = "outcome",
        splits = splits,
        learner = list(glm = spec_glm, rf = spec_rf),
        metrics = "auc"
    )
    
    # 3. Run the audit
    audits <- audit_leakage_by_learner(fit_multi, metric = "auc", B = 20)
    cat("Per-learner audit summary:\n")
    print(audits)
    
} else {
    cat("ranger/parsnip not installed.\n")
}
```

This example uses **ranger** for the random forest specification; if it is not installed, the code chunk is skipped.

Use `parallel_learners = TRUE` to audit learners concurrently when **future.apply** is available.

The printed table summarizes each learner’s observed metric, permutation gap,
p-value, and key batch/duplicate summaries. Use it to compare signal strength
across models while checking for leakage risks. Pass `learners =` to audit a
subset, or `mc.cores =` to control parallel workers.


**HTML audit report**

`audit_report()` accepts either a `LeakAudit` or a `LeakFit` object. When a
`LeakFit` is provided, it first runs `audit_leakage()` and then forwards any
additional arguments to the audit step. If multiple learners were fit, pass
`learner =` via `...` to select one. Use `open = TRUE` to open the report in a
browser after rendering.

```{r audit-report, eval = FALSE}
if (requireNamespace("rmarkdown", quietly = TRUE) && rmarkdown::pandoc_available()) {
  report_path <- audit_report(audit, output_dir = ".")
  cat("HTML report written to:\n", report_path, "\n")
} else {
  cat("rmarkdown or pandoc not available; skipping audit report rendering.\n")
}
```

The report path points to a standalone HTML file containing the same audit tables and plots, suitable for sharing with collaborators or archiving as a quality control record.

## Time-series leakage checks

Time-series data require special handling. Random splits can leak information
from the future into the past. Use `mode = "time_series"` with a prediction
`horizon`, and audit with block permutations. Choose `time_block = "circular"`
or `"stationary"`; when `block_len` is NULL, the audit uses a default block
length (~10% of the test block size, minimum 5).

**Leaky example: lookahead feature**

```{r time-series-fit}
time_splits <- make_split_plan(
  df_time,
  outcome = "outcome",
  mode = "time_series",
  time = "time",
  v = 4,
  horizon = 1
)

cat("Time-series splits summary:\n")
time_splits

fit_time_leaky <- fit_resample(
  df_time,
  outcome = "outcome",
  splits = time_splits,
  learner = spec,
  metrics = "auc"
)

cat("Time-series leaky fit summary:\n")
summary(fit_time_leaky)
```

Time-series splitting trains on growing windows, so performance can differ from
standard CV simply because early folds have smaller training sets. Regardless of
the score, the presence of the `leak_future` feature makes this estimate
methodologically invalid.

**Leakage-safe alternative: remove lookahead and audit with blocks**

```{r time-series-safe}
df_time_safe <- df_time
df_time_safe$leak_future <- NULL

fit_time_safe <- fit_resample(
  df_time_safe,
  outcome = "outcome",
  splits = time_splits,
  learner = spec,
  metrics = "auc"
)

cat("Time-series safe fit summary:\n")
summary(fit_time_safe)

audit_time <- audit_leakage(
  fit_time_safe,
  metric = "auc",
  B = 20,
  time_block = "stationary",
  block_len = 5
)

cat("Time-series leakage audit summary:\n")
summary(audit_time)
if (!is.null(audit_time@permutation_gap) && nrow(audit_time@permutation_gap) > 0) {
  # Time-series permutation significance results
  audit_time@permutation_gap
}

if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot_time_acf(fit_time_safe, lag.max = 20)
} else {
  cat("ggplot2 not installed; skipping ACF plot.\n")
}
```

The safe fit summary provides the leakage-resistant performance estimate. Compare
leaky vs. safe fits to gauge inflation risk; `features_final` should drop when
the lookahead feature is removed.

The ACF plot shows the autocorrelation of out-of-fold predictions by lag. The
dashed red lines represent the 95% confidence interval for white noise; large
bars outside the bands indicate residual temporal dependence. `plot_time_acf()`
requires numeric predictions and time metadata aligned to the fit.

## Cross-validation uncertainty with `cv_ci()`

Standard cross-validation yields one metric value per fold. Because folds share
training data, the per-fold values are positively correlated: the naive standard
error `sd / sqrt(K)` systematically underestimates variability, making the
resulting confidence intervals too narrow. `cv_ci()` addresses this with two
approaches:

- **`"normal"`** (default): uses the naive SEM `sd / sqrt(K)` with a
  *t*-distribution on `K − 1` degrees of freedom. Fast and appropriate when
  fold overlap is low (large datasets, many folds).

- **`"nadeau_bengio"`**: applies the Nadeau–Bengio (2003) variance correction
  `(1/K + n_test / n_train) × var(x)`, explicitly accounting for the fraction of
  training samples shared across folds. Recommended for small datasets or
  few-fold CV where fold correlation is non-negligible.

The function accepts any data frame with `fold` and `learner` columns — such as
`fit@metrics` from a `LeakFit` object — and returns one row per learner with
four columns per metric: `<metric>_mean`, `<metric>_sd`, `<metric>_ci_lo`, and
`<metric>_ci_hi`.

```{r cv-ci}
# Standard 95% CI from the leakage-safe fit
ci_std <- cv_ci(fit_safe@metrics, level = 0.95, method = "normal")
cat("Standard 95% CI:\n")
print(ci_std)

# Nadeau-Bengio corrected CI
# For K-fold CV: n_train ≈ (K-1)/K × n, n_test ≈ n/K
K_folds  <- length(unique(fit_safe@metrics$fold))
n_total  <- nrow(fit_safe@splits@info$coldata)

ci_nb <- cv_ci(
  fit_safe@metrics,
  level   = 0.95,
  method  = "nadeau_bengio",
  n_train = round(n_total * (K_folds - 1L) / K_folds),
  n_test  = round(n_total / K_folds)
)
cat("Nadeau-Bengio corrected 95% CI:\n")
print(ci_nb)
```

The Nadeau–Bengio interval is typically wider because it correctly inflates the
variance to account for fold overlap. Use it whenever the training-set fraction
shared between consecutive folds is large — for example, 5-fold CV reuses 80%
of training data across fold pairs, so correcting for this correlation matters.

When multiple learners were fit, `cv_ci()` returns one row per learner.
For repeated K-fold CV, average `n_train` and `n_test` across folds before
passing them in; `fit@metrics` already stores per-fold counts if `audit_leakage()`
annotated the object.

# Delta Leakage Sensitivity Index: Quantifying Performance Inflation

## Motivation

`audit_leakage()` detects whether leakage is present via permutation tests and
confounding checks. It cannot, however, directly answer: *by how much does
leakage inflate my reported performance?* The **Delta Leakage Sensitivity
Index** (ΔLSI) addresses this by rigorously comparing a naive (potentially
leaky) cross-validation pipeline against a leakage-guarded one run on the same
data.

## Mathematical framework

### Notation

Let $\mathcal{D} = \{(x_i, y_i)\}_{i=1}^n$ denote a dataset of $n$
observations. A repeated $K$-fold cross-validation design with $R$ independent
repeats partitions the observation indices into $RK$ folds. Let
$\mathcal{T}_{r,k} \subset \{1, \ldots, n\}$ denote the test index set for
fold $k$ in repeat $r$, with $n_{r,k} = |\mathcal{T}_{r,k}|$. Let
$m_{r,k}^{(p)}$ denote the performance metric evaluated on fold $(r,k)$ under
pipeline $p \in \{\text{naive},\, \text{guarded}\}$.

### Stage 1: Within-repeat aggregation

For each repeat $r$ and pipeline $p$, the $K$ fold-level metric values are
combined into a size-weighted repeat mean:

$$\mu_r^{(p)} = \frac{\displaystyle\sum_{k=1}^{K} n_{r,k}\, m_{r,k}^{(p)}}{\displaystyle\sum_{k=1}^{K} n_{r,k}}$$

The test-fold size $n_{r,k}$ is the weight, so folds with more held-out
observations contribute proportionally more to the repeat-level estimate. Folds
with missing metric values are excluded from both the numerator and denominator.
This produces the `@repeats_naive` and `@repeats_guarded` data frames (column
`metric`), one row per repeat.

### Stage 2: Per-repeat inflation score

For each repeat $r$, the signed performance gap between the two pipelines is:

$$\Delta_r = s \cdot \bigl(\mu_r^{\text{naive}} - \mu_r^{\text{guarded}}\bigr),
\qquad r = 1, \ldots, R$$

where

$$s = \begin{cases}
+1 & \text{if the metric is higher-is-better (e.g.\ AUC, accuracy)} \\
-1 & \text{if the metric is lower-is-better (e.g.\ RMSE, MAE, log-loss)}
\end{cases}$$

By construction $\Delta_r > 0$ if and only if the naive pipeline is more
optimistic than the guarded pipeline in repeat $r$, irrespective of metric
direction. Because both pipelines are evaluated on the same test fold
$\mathcal{T}_{r,k}$ (the paired design), any between-fold variability common to
both pipelines cancels in $\Delta_r$, concentrating the signal on genuine
pipeline differences rather than partition noise.

### Point estimates

Two summaries of the inflation vector
$\boldsymbol{\Delta} = (\Delta_1, \ldots, \Delta_R)^\top$ are reported.

**Arithmetic mean — `delta_metric`:**

$$\hat{\delta}_{\text{metric}} = \frac{1}{R} \sum_{r=1}^{R} \Delta_r$$

This is the most directly interpretable quantity. For AUC,
$\hat{\delta}_{\text{metric}} = 0.08$ means the naive pipeline's reported AUC
exceeds the guarded pipeline's by approximately $0.08$ units, averaged across
all repeats. This value is stored in `@delta_metric`.

**Huber M-estimate — `delta_lsi`:**

$$\hat{\delta}_{\text{LSI}} = \operatorname*{arg\,min}_{\mu} \sum_{r=1}^{R}
\rho_{k}\!\left(\frac{\Delta_r - \mu}{\hat{\sigma}}\right)$$

where $\rho_k$ is Huber's loss function with tuning constant $k = 1.345$:

$$\rho_k(u) = \begin{cases}
\tfrac{1}{2}u^2 & |u| \leq k \\
k|u| - \tfrac{1}{2}k^2 & |u| > k
\end{cases}$$

and $\hat{\sigma} = 1.4826 \times \operatorname{MAD}(\boldsymbol{\Delta})$ is a
fixed robust scale estimate. This minimisation is solved via iteratively
reweighted least squares (IRLS):

1. Initialise: $\hat{\mu}^{(0)} = \operatorname{median}(\boldsymbol{\Delta})$,
   $\hat{\sigma} = 1.4826 \times \operatorname{MAD}(\boldsymbol{\Delta})$
2. Compute standardised residuals:
   $u_r^{(t)} = (\Delta_r - \hat{\mu}^{(t)}) / \hat{\sigma}$
3. Assign Huber influence weights:
   $w_r^{(t)} = \min\!\left(1,\; k / |u_r^{(t)}|\right)$
4. Update location:
   $\hat{\mu}^{(t+1)} = \displaystyle\frac{\sum_{r=1}^{R} w_r^{(t)} \Delta_r}{\sum_{r=1}^{R} w_r^{(t)}}$
5. Repeat until $|\hat{\mu}^{(t+1)} - \hat{\mu}^{(t)}| < 10^{-8}(1 + |\hat{\mu}^{(t)}|)$

Note that $\hat{\sigma}$ is **fixed** at its initial value throughout iteration;
only the location $\hat{\mu}$ is updated. Outlier repeats with
$|u_r^{(t)}| > k$ receive weight $w_r < 1$, reducing their contribution to the
update. When $\hat{\sigma} = 0$ (all $\Delta_r$ identical), the estimator
returns the common value without iteration. This quantity is stored in
`@delta_lsi` and is the primary reported point estimate.

The two estimates coincide when $\{\Delta_r\}$ are symmetric and free of
outliers. When they differ substantially, `delta_lsi` is more trustworthy
because it is not pulled by a single anomalous repeat.

**Unpaired fallback.** When the two fits do not share the same fold structure
(`@info$paired = FALSE`), the paired vector $\boldsymbol{\Delta}$ cannot be
formed. The point estimates are instead computed from the separately summarised
repeat-mean vectors $\boldsymbol{\mu}^{\text{naive}}$ and
$\boldsymbol{\mu}^{\text{guarded}}$:

$$\hat{\delta}_{\text{metric}}^{\text{unpaired}}
  = s \cdot \bigl(\bar{\mu}^{\text{naive}} - \bar{\mu}^{\text{guarded}}\bigr),
\qquad
\hat{\delta}_{\text{LSI}}^{\text{unpaired}}
  = s \cdot \bigl(\hat{H}(\boldsymbol{\mu}^{\text{naive}})
               - \hat{H}(\boldsymbol{\mu}^{\text{guarded}})\bigr)$$

where $\bar{\mu}$ is the arithmetic mean and $\hat{H}$ is the Huber
M-estimator applied to each pipeline's repeat-means vector independently. No
inference (sign-flip test or BCa interval) is available in the unpaired case
($R_{\text{eff}} = 0$).

## Setting up a two-pipeline comparison

A ΔLSI analysis requires two `LeakFit` objects:

1. **Naive fit** — standard cross-validation without leakage safeguards (e.g.,
   random row splits, global preprocessing, or leaky features).
2. **Guarded fit** — leakage-aware cross-validation using subject-grouped splits
   and preprocessing fit only on training folds.

For the most powerful **paired** design, generate a single `LeakSplits` object
and pass it to both `fit_resample()` calls. Pairing is validated automatically
by comparing the sorted test-set indices of every fold; if the structures differ
despite equal repeat counts, a warning is issued and analysis falls back to
unpaired mode.

```{r delta-lsi-data}
set.seed(100)
n_d    <- 120L
subj_d <- rep(paste0("P", seq_len(30L)), each = 4L)   # 30 subjects, 4 obs each

df_dlsi_guarded <- data.frame(
  subject = subj_d,
  outcome = factor(sample(c("case", "control"), n_d, replace = TRUE)),
  x1      = rnorm(n_d),
  x2      = rnorm(n_d)
)

# Leaky feature: subject-level mean of outcome
# With subject-grouped splits, test subjects' leak value encodes their outcome
df_dlsi_naive <- df_dlsi_guarded
df_dlsi_naive$leak <- ave(
  as.numeric(df_dlsi_guarded$outcome == "case"),
  subj_d, FUN = mean
)

# Shared splits: 5-fold × 5 repeats → R_eff = 5 (tier C: point + p-value)
splits_dlsi <- make_split_plan(
  df_dlsi_guarded,
  outcome = "outcome",
  mode    = "subject_grouped",
  group   = "subject",
  v       = 5L,
  repeats = 5L
)
```

```{r delta-lsi-fits}
# Reuse the parsnip spec (base R glm) defined earlier
# Naive fit: uses leaky feature (leak encodes test-subject outcome label)
fit_dlsi_naive <- fit_resample(
  df_dlsi_naive,
  outcome = "outcome",
  splits  = splits_dlsi,
  learner = spec,
  metrics = "auc",
  seed    = 1L
)

# Guarded fit: same splits, clean features only
fit_dlsi_guarded <- fit_resample(
  df_dlsi_guarded,
  outcome = "outcome",
  splits  = splits_dlsi,
  learner = spec,
  metrics = "auc",
  seed    = 1L
)

cat("Naive   AUC:", round(mean(fit_dlsi_naive@metrics$auc,   na.rm = TRUE), 3), "\n")
cat("Guarded AUC:", round(mean(fit_dlsi_guarded@metrics$auc, na.rm = TRUE), 3), "\n")
```

## Running `delta_lsi()`

```{r delta-lsi-run}
result_dlsi <- delta_lsi(
  fit_leaky   = fit_dlsi_naive,
  fit_guarded = fit_dlsi_guarded,
  metric      = "auc",
  M_boot      = 300L,
  M_flip      = 300L,
  seed        = 1L
)

summary(result_dlsi)
```

`summary()` reports seven quantities; here is how to read each one:

- **Naive / Guarded pipeline values**: the Huber-robust mean AUC across repeats
  for each pipeline in isolation. Compare these directly to see the raw
  performance level of each pipeline before computing the gap.

- **`delta_metric`**: the arithmetic mean of the per-repeat inflation scores
  $\Delta_r$. Interpret this in the original metric's units: a value of 0.10
  means the naive pipeline reports AUC ~0.10 higher than the guarded pipeline,
  averaged across repeats. This is the most intuitive summary for reporting.

- **`delta_lsi`**: the Huber M-estimate of $\{\Delta_r\}$. Because it
  down-weights outlier repeats, it is the more reliable point estimate when
  repeat-to-repeat variation is high. Values very close to `delta_metric`
  indicate a consistent, non-skewed inflation pattern.

- **95% BCa CI** (shown when `R_eff ≥ 10`): a bootstrap confidence interval
  for `delta_lsi`. If the lower bound is above zero the inflation is unlikely
  to be a sampling artefact. If the interval spans zero the inflation estimate
  is uncertain given the number of repeats.

- **Sign-flip p-value** (shown when `R_eff ≥ 5`): tests H₀: no leakage
  inflation (mean $\Delta_r = 0$) via a randomization test. A small p-value
  (e.g., < 0.05) indicates the observed inflation is unlikely under the null
  of exchangeable pipeline labels.

- **Inference tier** and **`inference_ok`**: summarise how much statistical
  weight the design supports. Only tier A (`R_eff ≥ 20`, `inference_ok = TRUE`)
  provides a full combination of point estimate + CI + p-value. Lower tiers
  report what is reliable given the available repeats. The formal tier
  definitions and their rationale are given in the Statistical inference section
  below.

## Designing for a target inference tier

The formal tier thresholds have direct implications for how you set up your
repeated CV design. Plan the number of folds and repeats around the tier you
need:

- **Tier D** (`D_insufficient`, point estimates only): any $R_{\text{eff}} < 5$ — useful for
  exploratory comparisons where no formal test is required.
- **Tier C** (`C_signflip`, sign-flip p-value available): 5-fold $\times$ 5 repeats
  $\to R_{\text{eff}} = 5$. The smallest achievable p-value is $1/2^5 = 0.031$.
  For `exchangeability = "blocked_time"` the minimum p-value is $1/2^{n_{\text{blocks}}}$;
  at least 5 blocks are required for the test to be available.
- **Tier B** (`B_signflip_ci`, BCa CI additionally available): 5-fold $\times$ 10 repeats
  $\to R_{\text{eff}} = 10$. Both sign-flip p-value and BCa CI are reported.
- **Tier A** (`A_full_inference`, `inference_ok = TRUE`): 5-fold $\times$ 20
  repeats $\to R_{\text{eff}} = 20$. Recommended for any published or
  pre-registered ΔLSI analysis.

For paired designs `R_eff` equals the number of matched repeats. For unpaired
designs `R_eff = 0` and only the raw point estimates are available.

```{r delta-lsi-tier}
cat("Tier:          ", result_dlsi@tier, "\n")
cat("R_eff:         ", result_dlsi@R_eff, "\n")
cat("inference_ok:  ", result_dlsi@inference_ok, "\n")
```

## Statistical inference

### Sign-flip randomization test

The sign-flip test (Phipson & Smyth, 2010) assesses the null hypothesis that
there is no systematic performance inflation:

$$H_0: \mathbb{E}[\Delta_r] = 0 \quad \text{(no leakage-induced inflation)}$$

The test exploits sign-exchangeability: under $H_0$, each $\Delta_r$ is
symmetrically distributed around zero, so randomly flipping the sign of any
$\Delta_r$ leaves the distribution of the test statistic unchanged. The test
statistic is the arithmetic mean of the inflation scores:

$$T_{\text{obs}} = \frac{1}{R} \sum_{r=1}^{R} \Delta_r$$

For each sign vector $\boldsymbol{\epsilon} = (\epsilon_1, \ldots, \epsilon_R)$
with $\epsilon_r \in \{-1, +1\}$, a null-distribution realisation is:

$$T_{\boldsymbol{\epsilon}} = \frac{1}{R} \sum_{r=1}^{R} \epsilon_r \Delta_r$$

The two-sided p-value is the proportion of sign vectors yielding a statistic at
least as extreme as observed:

$$p = \frac{\#\bigl\{\boldsymbol{\epsilon} \in \mathcal{E} :
           |T_{\boldsymbol{\epsilon}}| \geq |T_{\text{obs}}|\bigr\}}{|\mathcal{E}|}$$

Two computation regimes are used:

- **Exact enumeration** ($R_{\text{eff}} \leq 15$): all $2^R$ sign vectors are
  evaluated — $|\mathcal{E}| = 2^R$. The p-value is $\text{count}/2^R$ with
  **no** continuity correction, because every sign combination including the
  observed one is explicitly enumerated. The minimum achievable p-value is
  $1/2^R$.

- **Monte Carlo** ($R_{\text{eff}} > 15$): $M_{\text{flip}}$ sign vectors are
  drawn uniformly from $\{-1,+1\}^R$. The Phipson & Smyth (2010) correction
  prevents $p = 0$ in the extreme tail:
$$p = \frac{B^{+} + 1}{M_{\text{flip}} + 1},
\qquad B^{+} = \#\{b : |T_{\boldsymbol{\epsilon}_b}| \geq |T_{\text{obs}}|\}$$

The arithmetic mean $\hat{\delta}_{\text{metric}}$ is used as the test
statistic — not the Huber estimate $\hat{\delta}_{\text{LSI}}$. Under a
sign-flip null the mean has a symmetric, well-defined exchangeable distribution;
the Huber M-estimator's nonlinear weighting changes with each sign assignment
and therefore does not preserve sign-exchangeability. As a result,
`@p_value` tests `@delta_metric`, not `@delta_lsi`.

This distinction matters in practice. When the per-repeat inflation vector
$\{\Delta_r\}$ contains outliers, `delta_lsi` (Huber) and `delta_metric` (mean)
can diverge in magnitude while agreeing in sign. A user can therefore see a
significant `@p_value` with a BCa CI for `delta_lsi` that spans zero — or vice
versa — if a handful of repeats pull the mean but not the robust estimate. The
`summary()` method flags this disagreement explicitly when it occurs.

The threshold $R_{\text{eff}} \geq 5$ for the sign-flip test follows from the
minimum achievable p-value: with $R = 4$, $1/2^4 = 0.0625 > 0.05$, so a
significant result at the conventional level is impossible regardless of the
data. With $R = 5$, $1/2^5 = 0.03125 < 0.05$, making it attainable. For
`exchangeability = "blocked_time"`, the relevant threshold is $n_{\text{blocks}} \geq 5$
(not $R_{\text{eff}} \geq 5$), since blocks are the independent units under the
test; `@p_value` is set to `NA` and a warning is issued when
$n_{\text{blocks}} < 5$. For other exchangeability values, `@p_value` is `NA`
when $R_{\text{eff}} < 5$ or when the design is unpaired.

### Huber M-estimator and robustness properties

The Huber M-estimator was described formally in the Mathematical framework
section above. Its inferential properties in this context are:

- **Breakdown point**: $\hat{\delta}_{\text{LSI}}$ remains a consistent
  estimate of the true location even when up to $\sim 29\%$ of repeats are
  outliers ($k = 1.345$ yields 95% asymptotic efficiency relative to the mean
  under a Gaussian distribution of $\{\Delta_r\}$).
- **Fixed scale**: $\hat{\sigma}$ is set to $1.4826 \times \operatorname{MAD}(\boldsymbol{\Delta})$
  before iteration and never updated. Fixing the scale prevents a single
  influential $\Delta_r$ from collapsing $\hat{\sigma}$ toward zero, which
  would force all weights to 1 and reduce the estimator to the ordinary mean.
- **Degenerate case**: when $\hat{\sigma} = 0$ (all repeat differences
  identical), the algorithm returns $\operatorname{median}(\boldsymbol{\Delta})$
  directly, which equals the common value.

The `delta_lsi` and `delta_metric` estimates always have the same sign because
the Huber estimator and the arithmetic mean agree on direction. They differ only
in magnitude, and then only when one or more $\Delta_r$ are extreme relative to
the bulk of the distribution.

### BCa bootstrap confidence interval

The BCa (bias-corrected and accelerated) interval (Efron, 1987) for
$\hat{\delta}_{\text{LSI}}$ corrects for both median bias and distributional
skewness in the bootstrap. Given $M_{\text{boot}}$ bootstrap resamples
$\boldsymbol{\Delta}_b^*$ drawn with replacement from $\boldsymbol{\Delta}$,
let $\hat{\theta}_b^* = \hat{H}(\boldsymbol{\Delta}_b^*)$ be the Huber estimate
on resample $b$ and $\hat{\theta} = \hat{H}(\boldsymbol{\Delta})$ the observed
estimate.

**Bias-correction constant** $\hat{z}_0$ accounts for the median bias of the
bootstrap distribution relative to the observed estimate:

$$\hat{z}_0 = \Phi^{-1}\!\left(
  \frac{\#\{b : \hat{\theta}_b^* < \hat{\theta}\}}{M_{\text{boot}}}
\right)$$

When $\hat{\theta}$ coincides with the exact median of the bootstrap
distribution, $\hat{z}_0 = 0$ and no bias correction is applied. A nonzero
$\hat{z}_0$ shifts both interval endpoints in the same direction.

**Acceleration constant** $\hat{a}$ corrects for skewness via the jackknife.
Let $\hat{\theta}_{(-r)} = \hat{H}(\boldsymbol{\Delta}_{(-r)})$ be the Huber
estimate with observation $r$ removed and
$\bar{\theta}_{(\cdot)} = R^{-1}\sum_{r=1}^R \hat{\theta}_{(-r)}$:

$$\hat{a} = \frac{\displaystyle\sum_{r=1}^{R}
                  \bigl(\bar{\theta}_{(\cdot)} - \hat{\theta}_{(-r)}\bigr)^3}
                 {6\,\left[\displaystyle\sum_{r=1}^{R}
                  \bigl(\bar{\theta}_{(\cdot)} - \hat{\theta}_{(-r)}\bigr)^2
                 \right]^{3/2}}$$

Here $\hat{a} > 0$ indicates a right-skewed sampling distribution;
$\hat{a} < 0$ indicates left skew.

**BCa-adjusted quantile levels**: for nominal coverage $1 - \alpha = 0.95$,
let $z_L = \Phi^{-1}(\alpha/2)$ and $z_U = \Phi^{-1}(1-\alpha/2)$. The
adjusted quantile levels are:

$$p_j = \Phi\!\left(\hat{z}_0 +
  \frac{\hat{z}_0 + z_j}{1 - \hat{a}\,(\hat{z}_0 + z_j)}
\right), \quad j \in \{L,\, U\}$$

The 95% BCa interval is the $p_L$ and $p_U$ empirical quantiles of the
bootstrap distribution $\{\hat{\theta}_b^*\}$:

$$\bigl[\hat{\theta}^*_{(p_L)},\; \hat{\theta}^*_{(p_U)}\bigr]$$

A separate BCa interval is computed for $\hat{\delta}_{\text{metric}}$
(using the arithmetic mean as the bootstrap statistic) and stored in
`@delta_metric_ci`. Both intervals are `NA` when $R_{\text{eff}} < 10$, because
fewer than 10 leave-one-out jackknife samples do not provide a reliable estimate
of $\hat{a}$.

### Pairing condition and the inference tier system

Two `LeakFit` objects are treated as **paired** if and only if they have the
same number of repeats and every fold shares an identical test-index
composition:

$$\operatorname{sort}\!\bigl(\mathcal{T}_{r,k}^{\text{naive}}\bigr)
= \operatorname{sort}\!\bigl(\mathcal{T}_{r,k}^{\text{guarded}}\bigr)
\quad \forall\, r,\, k$$

This is verified by comparing the sorted integer index vectors for each
sequential fold position. When paired, $R_{\text{eff}} = R$; when unpaired,
$R_{\text{eff}} = 0$ and all inference is disabled. Pairing is automatic when
both fits are produced from the same `LeakSplits` object.

All three inference procedures are gated by $R_{\text{eff}}$. The four-tier
system makes these dependencies explicit:

| Tier | Condition | `@delta_metric` | `@delta_lsi` | `@p_value` | `@delta_lsi_ci` | `@inference_ok` |
|---|---|:---:|:---:|:---:|:---:|:---:|
| `D_insufficient` | $R_{\text{eff}} < 5$ | ✓ | ✓ | `NA` | `NA` | `FALSE` |
| `C_signflip` | $5 \leq R_{\text{eff}} < 10$ | ✓ | ✓ | ✓ | `NA` | `FALSE` |
| `B_signflip_ci` | $10 \leq R_{\text{eff}} < 20$ | ✓ | ✓ | ✓ | ✓ | `FALSE` |
| `A_full_inference` | $R_{\text{eff}} \geq 20$ | ✓ | ✓ | ✓ | ✓ | `TRUE` |

The threshold $R_{\text{eff}} \geq 5$ for the sign-flip test under `"iid"`
exchangeability follows from the minimum achievable p-value under exact
enumeration ($1/2^5 = 0.03125 < 0.05$; $1/2^4 = 0.0625 > 0.05$). For
`exchangeability = "blocked_time"`, the independent units are blocks of
repeats, so the effective threshold is $n_{\text{blocks}} \geq 5$; when fewer
than five blocks are formed from the available repeats, `@p_value` is set to
`NA` even if $R_{\text{eff}} \geq 5$. The threshold $R_{\text{eff}} \geq 10$
for BCa intervals ensures at least 10 jackknife leave-one-out estimates for a
reliable $\hat{a}$. The `inference_ok = TRUE` flag at tier A requires all
three of $R_{\text{eff}} \geq 20$, a finite `@p_value`, and finite
`@delta_lsi_ci` to hold simultaneously.

The code below shows direct slot access to confirm that the numbers correspond
exactly to the quantities defined above:

```{r delta-lsi-components}
# delta_metric: arithmetic mean of {Δ_r}
cat("delta_metric:  ", round(result_dlsi@delta_metric, 4), "\n")
# delta_lsi: Huber M-estimate of {Δ_r} (k=1.345, fixed MAD scale)
cat("delta_lsi:     ", round(result_dlsi@delta_lsi,    4), "\n")
# delta_lsi_ci: 95% BCa interval for delta_lsi (NA below tier B)
cat("95% BCa CI:    [",
    paste(round(result_dlsi@delta_lsi_ci, 4), collapse = ", "), "]\n")
# p_value: sign-flip randomization p-value (NA below tier C or if unpaired)
cat("p-value:       ",
    if (is.na(result_dlsi@p_value)) "NA (tier D or unpaired)"
    else format(result_dlsi@p_value, digits = 3), "\n")
```

## Paired vs. unpaired designs

**Paired** (recommended): both fits are produced from the same `LeakSplits`
object. The per-repeat inflation scores $\Delta_r$ are computed as within-repeat
differences, eliminating between-fold variability. $R_{\text{eff}} = R$ and all
inference tiers are reachable. Equal repeat counts with differing index patterns
produce a warning and fall back to unpaired mode automatically.

**Unpaired**: the two fits have different fold structures or different repeat
counts. The point estimates use separately computed pipeline means rather than
paired differences; $R_{\text{eff}} = 0$ and all inference is disabled. This
mode is appropriate when the splitting scheme itself is part of what the
comparison is testing — for example, evaluating whether subject-grouped splits
alone (not a leaky feature) account for a performance difference.

```{r delta-lsi-unpaired, warning=TRUE}
# Unpaired: naive uses sample-wise CV, guarded uses subject-grouped splits
splits_rand <- make_split_plan(
  df_dlsi_naive,
  outcome = "outcome",
  mode    = "subject_grouped",
  group   = "row_id",
  v       = 5L,
  repeats = 5L
)

fit_naive_rand <- fit_resample(
  df_dlsi_naive,
  outcome = "outcome",
  splits  = splits_rand,
  learner = spec,
  metrics = "auc",
  seed    = 1L
)

r_unpaired <- suppressWarnings(
  delta_lsi(fit_naive_rand, fit_dlsi_guarded, metric = "auc", seed = 1L)
)

cat(sprintf("Paired:        %s\n", r_unpaired@info$paired))
cat(sprintf("R_eff:         %d  (0 = unpaired, inference disabled)\n", r_unpaired@R_eff))
cat(sprintf("delta_metric:  %.4f  (naive raw mean - guarded raw mean)\n",
            r_unpaired@delta_metric))
cat(sprintf("p_value:       %s\n",
            if (is.na(r_unpaired@p_value)) "NA" else r_unpaired@p_value))
```

## Metric direction with `higher_is_better`

For lower-is-better metrics (RMSE, MAE, log-loss, Brier score, deviance),
set `higher_is_better = FALSE` or let `delta_lsi()` auto-detect from the
metric name. Auto-detected lower-is-better metric names:
`rmse`, `mse`, `mae`, `log_loss`, `logloss`, `brier`, `error`, `loss`, `deviance`.

When `higher_is_better = FALSE`, the sign of `delta_metric` is flipped so that
a **positive** value still means the naive pipeline is inflated relative to
guarded (i.e., naive has a *lower* error, which is paradoxically *better*
performance — indicating leakage):

```{r delta-lsi-direction}
r_hib_true  <- suppressWarnings(delta_lsi(
  fit_dlsi_naive, fit_dlsi_guarded,
  metric           = "auc",
  higher_is_better = TRUE,
  seed             = 1L
))
r_hib_false <- suppressWarnings(delta_lsi(
  fit_dlsi_naive, fit_dlsi_guarded,
  metric           = "auc",
  higher_is_better = FALSE,
  seed             = 1L
))

cat(sprintf("hib = TRUE:   delta_metric = %+.4f  (positive = naive inflated)\n",
            r_hib_true@delta_metric))
cat(sprintf("hib = FALSE:  delta_metric = %+.4f  (sign flipped)\n",
            r_hib_false@delta_metric))
```

The `info$higher_is_better` slot records the value used (after auto-detection),
so analyses are fully reproducible:

```{r delta-lsi-hib-slot}
cat("higher_is_better used:", result_dlsi@info$higher_is_better, "\n")
```

## Accessing the `LeakDeltaLSI` object

`delta_lsi()` returns a `LeakDeltaLSI` S4 object. All slots are accessible
directly for programmatic downstream use:

| Slot | Type | Description |
|------|------|-------------|
| `@metric` | character | Metric name used (e.g. `"auc"`, `"rmse"`) |
| `@exchangeability` | character | Exchangeability assumption for the sign-flip test. `"blocked_time"` activates a block sign-flip procedure; `"by_group"` and `"within_batch"` are stored but inference still uses iid sign-flip (a warning is issued); `"iid"` (default) applies the standard independent sign-flip |
| `@tier` | character | Inference tier: `"A_full_inference"`, `"B_signflip_ci"`, `"C_signflip"`, or `"D_insufficient"` |
| `@R_eff` | integer | Effective paired repeats (0 when unpaired; all inference is based on this count) |
| `@delta_metric` | numeric | Arithmetic mean of per-repeat inflation scores $\bar{\Delta}$; interpret directly in metric units |
| `@delta_lsi` | numeric | Huber M-estimate of per-repeat inflation scores; more robust than `delta_metric` when repeat-level outliers are present |
| `@delta_metric_ci` | numeric[2] | 95% BCa bootstrap CI for `delta_metric` (both elements `NA` below tier B) |
| `@delta_lsi_ci` | numeric[2] | 95% BCa bootstrap CI for `delta_lsi` (both elements `NA` below tier B) |
| `@p_value` | numeric | Sign-flip randomization p-value testing H₀: mean inflation = 0 (`NA` below tier C or when unpaired) |
| `@inference_ok` | logical | `TRUE` only at tier A (`R_eff ≥ 20`), indicating full point + CI + p-value inference is available |
| `@folds_naive` | data.frame | Per-fold data for the naive pipeline: columns `fold`, `metric` (fold-level metric value), `repeat_id`, `n` (test-fold size) |
| `@folds_guarded` | data.frame | Per-fold data for the guarded pipeline: columns `fold`, `metric`, `repeat_id`, `n` — same structure as `@folds_naive` |
| `@repeats_naive` | data.frame | Per-repeat aggregates for the naive pipeline: columns `repeat_id`, `metric` (size-weighted mean across folds), `n_folds`, `total_n` |
| `@repeats_guarded` | data.frame | Per-repeat aggregates for the guarded pipeline: same structure as `@repeats_naive` |
| `@info` | list | Metadata: `paired` (logical), `R_naive`, `R_guarded` (number of repeats in each fit), `higher_is_better` (logical, after auto-detection), `metric_naive` and `metric_guarded` (Huber-robust overall mean for each pipeline), `block_size_used` and `n_blocks` (for `exchangeability = "blocked_time"`; `NA` otherwise), and optionally `delta_r` (per-repeat $\Delta_r$ vector) and the original fit objects when `return_details = TRUE` |

```{r delta-lsi-slots}
# Per-repeat summaries for custom plotting
head(result_dlsi@repeats_naive,   3)
head(result_dlsi@repeats_guarded, 3)

# Per-fold diagnostics
head(result_dlsi@folds_naive,   3)
head(result_dlsi@folds_guarded, 3)

# Metadata
str(result_dlsi@info)
```

The `@repeats_naive` and `@repeats_guarded` data frames are the primary
substrate for diagnostic visualisation. Plotting `metric` by `repeat_id` for
both pipelines side-by-side reveals whether the performance gap is consistent
across all repeats — a clean, reproducible offset indicates structural leakage
— or concentrated in a handful of repeats, which may indicate a data-quality
issue or an unstable fold partition.

The `@folds_naive` and `@folds_guarded` data frames go one level deeper,
providing the per-fold metric value along with the test-fold size (`n`) and the
repeat it belongs to (`repeat_id`). These are useful for identifying individual
folds that behave as outliers within a repeat, for example because a
small test partition happened to produce extreme metric values.

## Connection to `audit_leakage()`

`delta_lsi()` and `audit_leakage()` answer complementary questions:

| Question | Tool |
|----------|------|
| Does leakage produce non-random performance? | `audit_leakage()` — permutation gap + p-value |
| What mechanisms drive the leakage? | `audit_leakage()` — `batch_assoc`, `target_assoc`, `duplicates` |
| How large is the performance inflation? | `delta_lsi()` — `delta_metric`, `delta_lsi` |
| Is the inflation statistically significant? | `delta_lsi()` — sign-flip p-value |
| Do guards fully eliminate the inflation? | `delta_lsi()` — compare before/after guarding |

A recommended two-step workflow:

1. Run `audit_leakage(fit_leaky)` to detect leakage and diagnose which
   mechanism is responsible.
2. Build a guarded pipeline that addresses those mechanisms, then run
   `delta_lsi(fit_leaky, fit_guarded)` to quantify how much inflation the
   leakage introduced and whether guarding eliminated it.

```{r delta-lsi-audit-connect}
# Step 1: detect and characterise leakage
audit_naive_dlsi <- audit_leakage(fit_dlsi_naive, metric = "auc", B = 20L)
cat("Naive audit summary:\n")
summary(audit_naive_dlsi)

# Step 2: confirmed guards reduced inflation; check residual DLSI
cat("\ndelta_metric (naive vs guarded):", round(result_dlsi@delta_metric, 4), "\n")
if (!is.na(result_dlsi@p_value)) {
  cat("Sign-flip p-value:              ", format(result_dlsi@p_value, digits = 3), "\n")
}
```

A ΔLSI near zero after guarding and a non-significant p-value together provide
strong evidence that the leakage source has been eliminated. A persistent
positive ΔLSI after guarding indicates residual inflation — revisit the
preprocessing pipeline for remaining leakage pathways.

# Parallel Processing

`bioLeak` uses the `future` framework for parallelism (Windows, macOS, Linux).
`fit_resample()`, `audit_leakage()`, `audit_leakage_by_learner()`, and
`simulate_leakage_suite()` honor the active plan when `parallel = TRUE` (or
`parallel_learners = TRUE`).

```{r parallel-setup, eval=FALSE}
library(future)

# Use multiple cores (works on all OS)
plan(multisession, workers = 4)

# Run a heavy simulation
sim <- simulate_leakage_suite(..., parallel = TRUE)

# Parallel folds or audits
# fit_resample(..., parallel = TRUE)
# audit_leakage(..., parallel = TRUE)
# audit_leakage_by_learner(..., parallel_learners = TRUE)

# Return to sequential processing
plan(sequential)
```

This chunk only configures the parallel plan, so it does not produce a result
object. Use it as a template before running compute-heavy functions.

## Simulation suite

`simulate_leakage_suite()` runs Monte Carlo simulations that inject specific
leakage mechanisms and then evaluates detection with the audit pipeline. It is
useful for validating your leakage checks before applying them to real data.
Available leakage types include `none`, `subject_overlap`, `batch_confounded`,
`peek_norm`, and `lookahead`. Use `signal_strength`, `prevalence`, `rho`, `K`,
`repeats`, `horizon`, and `preprocess` to control the simulation. The output is
a `LeakSimResults` data frame with one row per seed. Metrics are selected
automatically based on outcome type (AUC for binary classification).

```{r simulate-suite}
if (requireNamespace("glmnet", quietly = TRUE)) {
  sim <- simulate_leakage_suite(
    n = 80,
    p = 6,
    mode = "subject_grouped",
    learner = "glmnet",
    leakage = "subject_overlap",
    seeds = 1:2,
    B = 20,
    parallel = FALSE,
    signal_strength = 1
  )
  
  # Simulation results (first 6 rows)
  head(sim)
} else {
  cat("glmnet not installed; skipping simulation suite example.\n")
}
```

Each row corresponds to one simulation seed.

- `metric_obs` is the observed performance (AUC for this simulation). Higher
  values suggest stronger apparent signal, which can be inflated by leakage.

- `gap` and `p_value` indicate that the leaked signal is statistically distinguishable from random noise.

Use `metric_obs` to gauge the magnitude of the leakage effect, and `gap` to assess detection sensitivity.


## Objects and summaries

bioLeak uses S4 classes and list-like results to capture provenance and diagnostics:

- `LeakSplits`: returned by `make_split_plan()`, printed with `show()`.
  Use `check_split_overlap()` to verify no-overlap invariants on grouping
  columns.
- `LeakFit`: returned by `fit_resample()`, summarized with `summary()`
  includes fold diagnostics in `fit@info$fold_status`
- `LeakAudit`: returned by `audit_leakage()`, summarized with `summary()`.
  Now includes `audit@info$mechanism_summary` — a data.frame classifying
  leakage evidence into mechanism classes (`non_random_signal`,
  `confounding_alignment`, `proxy_target_leakage`, `duplicate_overlap`,
  `temporal_lookahead`).
- `LeakAuditList`: returned by `audit_leakage_by_learner()`, printed directly
- `GuardFit`: returned by `.guard_fit()`, printed and summarized.
  Use `guard_to_recipe()` to convert guard preprocessing specs to a
  `recipes::recipe` for tidymodels workflows.
- `LeakImpute`: returned by `impute_guarded()` (guarded data + diagnostics)
- `LeakTune`: returned by `tune_resample()` (nested tuning summaries)
  includes `fold_status`, selected parameters, and optional `final_model` /
  `final_workflow` when `refit = TRUE`
- `LeakSimResults`: returned by `simulate_leakage_suite()` (per-seed results)
- `LeakDeltaLSI`: returned by `delta_lsi()` (paired pipeline comparison).
  Key slots: `@metric`, `@exchangeability`, `@tier`, `@R_eff`,
  `@delta_metric`, `@delta_metric_ci`, `@delta_lsi`, `@delta_lsi_ci`,
  `@p_value`, `@inference_ok`, `@folds_naive`, `@folds_guarded`,
  `@repeats_naive`, `@repeats_guarded`, `@info`.
  Summarise with `summary()`.

Set `options(bioLeak.strict = TRUE)` to activate strict leakage mode, which
escalates trained-recipe/workflow warnings to errors, warns on missing seeds,
and auto-runs overlap checks after splitting.

These objects store hashes and metadata that help reproduce and audit the
workflow. Inspect their slots (`@metrics`, `@predictions`, `@permutation_gap`,
`@duplicates`) to drill into specific leakage signals.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{audit_perm_gap}
\alias{audit_perm_gap}
\title{Permutation-gap test results from a LeakAudit}
\usage{
audit_perm_gap(audit)
}
\arguments{
\item{audit}{A `LeakAudit` object returned by [audit_leakage()].}
}
\value{
A `data.frame` with one row per (mechanism class, repeat),
  summarising the permutation-gap test.
}
\description{
Returns the permutation-gap test data frame stored in a `LeakAudit`
object. Columns include the observed metric, permuted-null mean and
SD, gap, z-score, and permutation p-value.
}
\seealso{
[audit_leakage()], [audit_target_assoc()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/guards.R
\name{predict_guard}
\alias{predict_guard}
\title{Apply a fitted GuardFit transformer to new data}
\usage{
predict_guard(fit, newdata)
}
\arguments{
\item{fit}{A \code{GuardFit} object created by [.guard_fit()]. This required
argument (no default) contains the training-time preprocessing settings and
statistics. Changing \code{fit} (for example, a different imputation method
or feature selection step) changes the output columns and values.}

\item{newdata}{A matrix or data.frame of predictors with one row per sample.
This required argument (no default) is transformed using the training-time
parameters in \code{fit} only. Missing columns are added and filled, extra
columns are dropped, and factor levels are aligned to the training levels; if
the training fit was numeric-only, non-numeric columns in \code{newdata}
trigger an error.}
}
\value{
A data.frame of transformed predictors with the same number of rows
as \code{newdata}. Column order and content match the training pipeline and
may include derived features (one-hot encodings, missingness indicators, or
PCA components). This output is not a prediction; it is intended as input to
a downstream model and assumes the training-time preprocessing is valid for
the new data.
}
\description{
Applies the preprocessing steps stored in a \code{GuardFit} object to new
data without refitting any statistics. This is designed to prevent
validation leakage that would occur if imputation, scaling, filtering, or
feature selection were recomputed on evaluation data. It enforces the
training schema by aligning columns and factor levels, and it errors when a
numeric-only training fit receives non-numeric predictors. It does not
detect label leakage, duplicate samples, or train/test contamination.
}
\examples{
x_train <- data.frame(a = c(1, 2, NA, 4), b = c(10, 11, 12, 13))
fit <- .guard_fit(
  x_train,
  y = c(0.1, 0.2, 0.3, 0.4),
  steps = list(impute = list(method = "median")),
  task = "gaussian"
)
x_new <- data.frame(a = c(NA, 5), b = c(9, 14))
out <- predict_guard(fit, x_new)
out
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{dlsi_tier}
\alias{dlsi_tier}
\title{Inference tier from a LeakDeltaLSI}
\usage{
dlsi_tier(dlsi)
}
\arguments{
\item{dlsi}{A `LeakDeltaLSI` object returned by [delta_lsi()].}
}
\value{
A length-one character string giving the tier label.
}
\description{
Returns the inference tier label stored in a `LeakDeltaLSI` object.
Possible values are `"A_full_inference"` (`R_eff >= 20`),
`"B_signflip_ci"` (`R_eff >= 10`), `"C_signflip"`
(`R_eff >= 5`), or `"D_insufficient"` (`R_eff < 5`).
}
\seealso{
[delta_lsi()], [dlsi_R_eff()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provenance.R
\name{.bio_capture_provenance}
\alias{.bio_capture_provenance}
\title{Capture session provenance information}
\usage{
.bio_capture_provenance(capture_git = TRUE, capture_hardware = TRUE)
}
\arguments{
\item{capture_git}{Logical; if TRUE, attempts to capture the current
git short SHA via \code{system2("git", ...)}. Default FALSE.}

\item{capture_hardware}{Logical; if TRUE, captures basic hardware info
from \code{Sys.info()}. Default TRUE.}
}
\value{
A list with elements \code{r_version}, \code{packages} (data.frame),
  \code{platform}, \code{timestamp}, \code{git_sha}, and \code{hardware}.
}
\description{
Internal helper that records R version, loaded packages, platform,
timestamp, and optionally git SHA and hardware details.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{fit_metrics}
\alias{fit_metrics}
\title{Per-fold metric data frame from a LeakFit}
\usage{
fit_metrics(fit)
}
\arguments{
\item{fit}{A `LeakFit` object returned by [fit_resample()].}
}
\value{
A `data.frame` with one row per (fold, repeat, learner)
  combination.
}
\description{
Returns the per-fold metric data frame stored in a `LeakFit` object.
Each row is one (fold, repeat, learner) combination; columns include
the requested metric values such as `auc` and any task-specific
performance scores.
}
\examples{
set.seed(1)
df <- data.frame(
  subject = rep(1:6, each = 2),
  outcome = factor(rep(c("a","b"), 6), levels = c("a","b")),
  x1 = rnorm(12), x2 = rnorm(12)
)
splits <- make_split_plan(df, outcome = "outcome",
                          mode = "subject_grouped", group = "subject", v = 3)
\dontrun{
  fit <- fit_resample(df, outcome = "outcome", splits = splits,
                      learner = parsnip::logistic_reg() |>
                                parsnip::set_engine("glm"),
                      metrics = "auc")
  fit_metrics(fit)
}
}
\seealso{
[fit_resample()], [audit_perm_gap()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_split_plan.R
\name{make_split_plan}
\alias{make_split_plan}
\title{Create leakage-resistant splits}
\usage{
make_split_plan(
  x,
  outcome = NULL,
  mode = c("subject_grouped", "batch_blocked", "study_loocv", "time_series", "combined"),
  group = NULL,
  batch = NULL,
  study = NULL,
  time = NULL,
  primary_axis = NULL,
  secondary_axis = NULL,
  constraints = NULL,
  v = 5,
  repeats = 1,
  stratify = FALSE,
  nested = FALSE,
  seed = 1,
  horizon = 0,
  purge = 0,
  embargo = 0,
  progress = TRUE,
  compact = FALSE,
  strict = TRUE
)
}
\arguments{
\item{x}{SummarizedExperiment or data.frame/matrix (samples x features).
If SummarizedExperiment, metadata are taken from colData(x). If data.frame,
metadata are taken from x (columns referenced by \code{group}, \code{batch}, \code{study}, \code{time}, \code{outcome}).}

\item{outcome}{character, outcome column name (used for stratification).}

\item{mode}{one of "subject_grouped","batch_blocked","study_loocv","time_series","combined".}

\item{group}{subject/group id column (for subject_grouped). Required when
mode is `subject_grouped`; use `group = "row_id"` to explicitly request
sample-wise CV.}

\item{batch}{batch/plate/center column (for batch_blocked).}

\item{study}{study id column (for study_loocv).}

\item{time}{time column (numeric or POSIXct) for time_series.}

\item{primary_axis}{List with elements \code{type} (one of \code{"subject"},
\code{"batch"}, \code{"study"}) and \code{col} (column name). Used only
when \code{mode = "combined"} to define the primary grouping axis.
Deprecated in favor of \code{constraints}; still supported for backward
compatibility.}

\item{secondary_axis}{List with elements \code{type} and \code{col}. Used
only when \code{mode = "combined"} to define the secondary constraint axis.
Training sets exclude samples whose secondary-axis levels appear in the
test set. Deprecated in favor of \code{constraints}; still supported for
backward compatibility.}

\item{constraints}{A list of constraint specifications for \code{mode = "combined"}.
Each element is a list with \code{type} (one of \code{"subject"}, \code{"batch"},
\code{"study"}) and \code{col} (column name). The first element defines the
primary grouping axis (fold driver); subsequent elements define exclusion
constraints (training samples sharing constraint-axis levels with the test
set are removed). Requires at least 2 elements. Cannot be used together with
\code{primary_axis}/\code{secondary_axis}.}

\item{v}{integer, number of folds (k) or rolling partitions.}

\item{repeats}{integer, number of repeats (>=1) for non-LOOCV modes.}

\item{stratify}{logical, keep outcome proportions similar across folds.
For grouped modes, stratification is applied at the group level (by
majority class per group) if \code{outcome} is provided; otherwise ignored.}

\item{nested}{logical, whether to attach inner CV splits (per outer fold)
using the same \code{mode} on the outer training set (with \code{v} folds, 1 repeat).}

\item{seed}{integer seed.}

\item{horizon}{numeric (>=0), minimal time gap for time_series so that the
training set only contains samples with time < min(test_time) when horizon = 0,
and time <= min(test_time) - horizon otherwise.}

\item{purge}{numeric (>=0), additional gap removed immediately before each
time-series test block.}

\item{embargo}{numeric (>=0), additional exclusion window anchored at the end
of each time-series test block. Training rows with
\code{time > max(test_time) - embargo} are removed.}

\item{progress}{logical, print progress for large jobs.}

\item{compact}{logical; store fold assignments instead of explicit train/test
indices to reduce memory usage for large datasets. Not supported when
\code{nested = TRUE}.}

\item{strict}{logical; deprecated and ignored. `subject_grouped` always
requires a non-NULL `group`.}
}
\value{
A \code{\linkS4class{LeakSplits}} S4 object containing:
  \describe{
    \item{\code{mode}}{Character string indicating the splitting mode
      (\code{"subject_grouped"}, \code{"batch_blocked"}, \code{"study_loocv"},
      or \code{"time_series"}).}
    \item{\code{indices}}{List of fold descriptors, each containing
      \code{train} (integer vector of training indices), \code{test}
      (integer vector of test indices), \code{fold} (fold number), and
      \code{repeat_id} (repeat identifier). When \code{compact = TRUE},
      indices are stored as fold assignments instead.}
    \item{\code{info}}{List of metadata including \code{outcome}, \code{v},
      \code{repeats}, \code{seed}, grouping columns (\code{group},
      \code{batch}, \code{study}, \code{time}), \code{stratify},
      \code{nested}, \code{horizon}, \code{purge}, \code{embargo},
      \code{summary} (data.frame of fold
      sizes), \code{hash} (reproducibility checksum), \code{inner}
      (nested inner splits if \code{nested = TRUE}), and \code{coldata}
      (sample metadata).}
  }
  Use the \code{show} method to print a summary, or access slots directly
  with \code{@}.
}
\description{
Generates leakage-safe cross-validation splits for common biomedical setups:
subject-grouped, batch-blocked, study leave-one-out, and time-series
rolling-origin. Supports repeats, optional stratification, nested inner CV,
and optional prediction horizon/purge/embargo gaps for time series. Note that splits store
explicit indices, which can be memory-intensive for large \code{n} and many
repeats.
}
\examples{
set.seed(1)
df <- data.frame(
  subject = rep(1:10, each = 2),
  outcome = rbinom(20, 1, 0.5),
  x1 = rnorm(20),
  x2 = rnorm(20)
)
splits <- make_split_plan(df, outcome = "outcome",
                      mode = "subject_grouped", group = "subject", v = 5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/permute_labels.R
\name{.permute_labels_factory}
\alias{.permute_labels_factory}
\title{Restricted permutation label factory}
\usage{
.permute_labels_factory(
  cd,
  outcome,
  mode,
  folds,
  perm_stratify,
  time_block,
  block_len,
  seed,
  group_col = NULL,
  batch_col = NULL,
  study_col = NULL,
  time_col = NULL,
  perm_refit = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{cd}{data.frame of sample metadata.}

\item{outcome}{outcome column name.}

\item{mode}{resampling mode (subject_grouped, batch_blocked, study_loocv, time_series).}

\item{folds}{list of fold descriptors from \code{LeakSplits}. When compact
splits are used, fold assignments are read from the
\code{fold_assignments} attribute.}

\item{perm_stratify}{logical or "auto"; if TRUE, permute within strata.}

\item{time_block}{time-series block permutation method.}

\item{block_len}{block length for time-series permutations.}

\item{seed}{integer seed.}

\item{group_col, batch_col, study_col}{optional metadata columns.}

\item{time_col}{optional metadata column name for time-series ordering.}

\item{perm_refit}{logical; if TRUE model is retrained on permuted labels
(block permutation preserves subject structure); if FALSE predictions are
fixed and simple label shuffle is used for \code{subject_grouped} mode.}

\item{verbose}{logical; print progress messages.}
}
\value{
A function that returns a list of permuted outcome vectors, one per fold.
}
\description{
Builds a closure that generates permuted outcome vectors per fold while
respecting grouping/batch/study/time constraints used in
\code{audit_leakage()}. Numeric outcomes can be stratified by quantiles to
preserve outcome structure under permutation.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cv_uncertainty.R
\name{.nb_corrected_var}
\alias{.nb_corrected_var}
\title{Nadeau-Bengio corrected variance}
\usage{
.nb_corrected_var(per_fold_vals, n_train = NULL, n_test = NULL)
}
\arguments{
\item{per_fold_vals}{Numeric vector of per-fold metric values.}

\item{n_train}{Average number of training samples per fold. NULL to skip
correction.}

\item{n_test}{Average number of test samples per fold. NULL to skip
correction.}
}
\value{
Scalar corrected variance, or \code{NA_real_} for single-fold input.
}
\description{
Computes the corrected variance for repeated K-fold CV as described in
Nadeau & Bengio (2003). The correction factor accounts for the overlap
between training sets across folds.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary.LeakTune}
\alias{summary.LeakTune}
\title{Summarize a nested tuning result}
\usage{
\method{summary}{LeakTune}(object, digits = 3, ...)
}
\arguments{
\item{object}{A [LeakTune] object returned by [tune_resample()].}

\item{digits}{Integer scalar. Number of decimal places to print in numeric
summary tables. Defaults to 3.}

\item{...}{Unused. Included for S3 method compatibility.}
}
\value{
Invisibly returns `object$metric_summary`, the data frame of per-learner
  metric means and standard deviations computed across outer folds.
}
\description{
Prints a concise report for a `LeakTune` object produced by [tune_resample()].
The report highlights the tuning strategy, selection metric, and
cross-validated performance across outer folds, plus a glimpse of the selected
hyperparameters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/guards.R
\name{.guard_ensure_levels}
\alias{.guard_ensure_levels}
\title{Ensure consistent categorical levels for guarded preprocessing}
\usage{
.guard_ensure_levels(df, levels_map = NULL, dummy_prefix = "__dummy__")
}
\arguments{
\item{df}{data.frame to normalize factor levels.}

\item{levels_map}{optional named list of factor levels learned from training data.}

\item{dummy_prefix}{prefix used when adding a dummy level to single-level factors.}
}
\value{
List with elements \code{data} (data.frame) and \code{levels} (named list of levels).
}
\description{
Converts character/logical columns to factors and aligns factor levels with
a training-time \code{levels_map}. Adds a dummy level when a column has only
one observed level so that downstream one-hot encoding retains a column.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/guard_to_recipe.R
\name{guard_to_recipe}
\alias{guard_to_recipe}
\title{Convert guard preprocessing steps to a recipes recipe}
\usage{
guard_to_recipe(steps, formula, training_data)
}
\arguments{
\item{steps}{A named list of guard preprocessing steps, e.g.,
\code{list(impute = list(method = "median"), normalize = list(method = "zscore"))}.}

\item{formula}{A model formula (e.g., \code{outcome ~ .}).}

\item{training_data}{A data.frame used to initialize the recipe.}
}
\value{
A \code{recipes::recipe} object with the mapped steps added.
}
\description{
Maps bioLeak guard preprocessing steps (impute, normalize, filter, fs) to
their closest \pkg{recipes} equivalents. Requires the \pkg{recipes} package.
Steps that have no direct recipe equivalent are skipped with a warning.
}
\details{
Mapping:
\itemize{
  \item \code{impute$method = "median"}: \code{step_impute_median(all_numeric_predictors())}
  \item \code{impute$method = "knn"}: \code{step_impute_knn(all_predictors(), neighbors = k)}
  \item \code{impute$method = "missForest"} or \code{"mice"}: Warning + \code{step_impute_median()} fallback
  \item \code{normalize$method = "zscore"}: \code{step_normalize(all_numeric_predictors())}
  \item \code{normalize$method = "robust"}: Warning + \code{step_normalize()} fallback
  \item \code{normalize$method = "none"}: No step added
  \item \code{filter$var_thresh > 0}: \code{step_nzv(all_numeric_predictors())}
  \item \code{fs$method = "pca"}: \code{step_pca(all_numeric_predictors(), num_comp = ncomp)}
  \item \code{fs$method = "ttest"} or \code{"lasso"}: Warning, skipped (no recipe equivalent)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time_series_blocks.R
\name{.stationary_bootstrap}
\alias{.stationary_bootstrap}
\title{Stationary bootstrap indices}
\usage{
.stationary_bootstrap(idx, mean_block)
}
\arguments{
\item{idx}{Integer vector of ordered indices.}

\item{mean_block}{Positive numeric, expected block length.}
}
\value{
Integer vector of permuted indices of the same length as `idx`.
}
\description{
Implements the stationary bootstrap of Politis & Romano (1994),
which resamples contiguous blocks of variable length to preserve
weak temporal dependence while maintaining ergodicity.
}
\references{
Politis, D. N., & Romano, J. P. (1994).
*The stationary bootstrap.* Journal of the American Statistical Association, 89(428), 1303-1313.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate_suite.R
\name{simulate_leakage_suite}
\alias{simulate_leakage_suite}
\title{Simulate leakage scenarios and audit results}
\usage{
simulate_leakage_suite(
  n = 500,
  p = 20,
  prevalence = 0.5,
  mode = c("subject_grouped", "batch_blocked", "study_loocv", "time_series"),
  learner = c("glmnet", "ranger"),
  leakage = c("none", "subject_overlap", "batch_confounded", "peek_norm", "lookahead"),
  preprocess = NULL,
  rho = 0,
  K = 5,
  repeats = 1,
  horizon = 0,
  B = 200,
  seeds = 1:10,
  parallel = FALSE,
  signal_strength = 1,
  verbose = FALSE
)
}
\arguments{
\item{n}{Integer scalar. Number of samples to simulate (default 500). Larger
values stabilize the Monte Carlo summary but increase runtime.}

\item{p}{Integer scalar. Number of baseline predictors before any leakage
feature is added (default 20). Increasing \code{p} changes the signal-to-noise
ratio and increases fitting time.}

\item{prevalence}{Numeric scalar in (0, 1). Target prevalence of class 1 in
the simulated outcome (default 0.5). Changing this alters class imbalance
and can affect AUC and the permutation gap.}

\item{mode}{Character scalar. Cross-validation scheme passed to
\code{make_split_plan()}; one of \code{"subject_grouped"},
\code{"batch_blocked"}, \code{"study_loocv"}, \code{"time_series"}.
Defaults to \code{"subject_grouped"}. This controls how samples are grouped
into folds (by subject, batch, study, or time) and therefore which leakage
mechanisms are realistically challenged.}

\item{learner}{Character scalar. Base learner, \code{"glmnet"} (default) or
\code{"ranger"}. Requires the corresponding package in \code{Suggests}.
Switching learners changes the fitted model, runtime, and performance.}

\item{leakage}{Character scalar. Leakage mechanism to inject; one of
\code{"none"}, \code{"subject_overlap"}, \code{"batch_confounded"},
\code{"peek_norm"}, \code{"lookahead"}. Leakage is added as an extra
predictor: \code{"subject_overlap"} adds per-subject mean outcome,
\code{"batch_confounded"} adds per-batch mean outcome, \code{"peek_norm"}
adds the globally normalized (z-scored) outcome, and \code{"lookahead"} adds the next-time
outcome. Changing this controls whether and how leakage is present.}

\item{preprocess}{Optional preprocessing list or recipe passed to
[fit_resample()]. When NULL (default), the simulator uses the
fit_resample defaults; for \code{"peek_norm"} leakage, normalization is
set to \code{"none"} to avoid attenuating the constant leakage feature.}

\item{rho}{Numeric scalar in [-1, 1]. AR(1)-style autocorrelation applied to
each predictor across row order (default 0). Higher absolute values increase
serial correlation and make time-ordered leakage more pronounced.}

\item{K}{Integer scalar. Number of folds/partitions (default 5). Used as the
fold count for \code{"subject_grouped"} and \code{"batch_blocked"}, and as
the number of rolling partitions for \code{"time_series"}. Ignored for
\code{"study_loocv"} (folds equal the number of studies).}

\item{repeats}{Integer scalar >= 1. Number of repeated CV runs for
\code{"subject_grouped"} and \code{"batch_blocked"} (default 1). Increasing
\code{repeats} increases the number of folds and runtime. Ignored for
\code{"study_loocv"} and \code{"time_series"}.}

\item{horizon}{Numeric scalar >= 0. Minimum time gap enforced between train
and test for \code{"time_series"} splits (default 0). Larger values make the
split more conservative and can reduce leakage from temporal proximity.}

\item{B}{Integer scalar >= 1. Number of permutations used by
\code{audit_leakage()} to compute the permutation gap and p-value (default
200). Larger values yield more stable p-values but increase runtime.}

\item{seeds}{Integer vector. Monte Carlo seeds (default \code{1:10}). One row
of output is produced per seed; changing \code{seeds} changes the simulated
datasets and splits.}

\item{parallel}{Logical scalar. If \code{TRUE}, evaluates seeds in parallel
using \code{future.apply} (if installed). Results are identical to sequential
execution; only runtime changes.}

\item{signal_strength}{Numeric scalar. Scales the linear predictor before
sampling outcomes (default 1). Larger values increase class separation and
tend to increase AUC; smaller values make the task harder.}

\item{verbose}{Logical scalar. If \code{TRUE}, prints progress messages for
each seed. Does not affect results.}
}
\value{
A \code{LeakSimResults} data frame with one row per seed and columns:
\itemize{
  \item \code{seed}: seed used for data generation, splitting, and auditing.
  \item \code{metric_obs}: observed CV performance (AUC for this simulation).
  \item \code{gap}: permutation-gap statistic (observed minus permutation mean).
  \item \code{p_value}: permutation p-value for the gap.
  \item \code{leakage}: leakage scenario used.
  \item \code{mode}: CV mode used.
}
Only the permutation-gap summary is returned; fitted models, predictions, and
other audit components are not included.
}
\description{
Simulates synthetic binary classification datasets with optional leakage
mechanisms, fits a model using a leakage-aware cross-validation scheme, and
summarizes the permutation-gap audit for each Monte Carlo seed. The suite is
designed to surface validation failures such as subject overlap across folds,
batch-confounded outcomes, global normalization/summary leakage, and
time-series look-ahead. The output is a per-seed summary of observed CV
performance and its gap versus a label-permutation null; it does not return
fitted models or the full audit object. Results are limited to the built-in
data generator and leakage types implemented here, and should be interpreted
as a simulation-based sanity check rather than a comprehensive leakage
detector for real data.
}
\details{
The generator draws \code{p} standard normal predictors, builds a linear
predictor from the first \code{min(5, p)} features, scales it by
\code{signal_strength}, and samples a binary outcome to achieve the requested
\code{prevalence}. Outcomes are returned as a two-level factor, so the audited
metric is AUC. Simulated metadata include subject, batch, study, and time
fields used by \code{mode} to create leakage-aware splits. Leakage mechanisms
are injected by adding a single extra predictor as described in
\code{leakage}. Parallel execution uses \code{future.apply} when installed and
does not change results.
}
\note{
This function is a general-purpose utility and its data-generation logic
intentionally differs from the custom simulation used in the bioLeak
manuscript (\file{paper/run_simulation.R}). Specific differences:
\itemize{
  \item \strong{peek_norm leakage}: this function uses a z-scored binary
    outcome as the leak feature; the manuscript uses a noisy continuous
    version (\code{as.numeric(y) + rnorm(n, 0, 0.3)}).
  \item \strong{lookahead leakage}: this function shifts the binary outcome
    (\code{c(y[-1], y[n])}); the manuscript shifts a continuous biomarker
    (\code{linpred + noise}).
  \item \strong{signal generation}: this function applies AR correlation to
    predictors via \code{rho}; the manuscript adds AR(1) noise directly to
    the linear predictor.
  \item \strong{audit settings}: the manuscript uses
    \code{perm_refit = FALSE} and \code{perm_stratify = TRUE};
    this function uses \code{perm_refit = "auto"} and the
    \code{perm_stratify} default (\code{FALSE}).
}
Users wishing to reproduce manuscript figures should run
\file{paper/run_simulation.R} directly rather than calling this function.
}
\examples{
\donttest{
  if (requireNamespace("glmnet", quietly = TRUE)) {
    set.seed(1)
    res <- simulate_leakage_suite(
      n = 120, p = 6, prevalence = 0.4,
      mode = "subject_grouped",
      learner = "glmnet",
      leakage = "subject_overlap",
      K = 3, repeats = 1,
      B = 50, seeds = 1,
      parallel = FALSE
    )
    # One row per seed with observed AUC, permutation gap, and p-value
    res
  }
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit_report.R
\name{audit_report}
\alias{audit_report}
\title{Render an HTML audit report}
\usage{
audit_report(
  audit,
  output_file = "bioLeak_audit_report.html",
  output_dir = tempdir(),
  quiet = TRUE,
  open = FALSE,
  ...
)
}
\arguments{
\item{audit}{A [LeakAudit] object from [audit_leakage()] or a [LeakFit] object
from [fit_resample()]. If a [LeakAudit] is supplied, the report uses its
stored results verbatim. If a [LeakFit] is supplied, `audit_report()` first
computes a new audit via [audit_leakage(...)]; the fit must contain
predictions and split metadata. When multiple learners were fit, pass a
`learner` argument via `...` to select a single model.}

\item{output_file}{Character scalar. File name for the HTML report. Defaults
to `"bioLeak_audit_report.html"`. If a relative name is provided, it is
created inside `output_dir`. Changing this value only changes the file name,
not the audit content.}

\item{output_dir}{Character scalar. Directory path where the report is
written. Defaults to [tempdir()]. The directory must exist or be creatable
by `rmarkdown::render()`. Changing this value only changes the output
location.}

\item{quiet}{Logical scalar passed to `rmarkdown::render()`. Defaults to
`TRUE`. When `FALSE`, knitting output and warnings are printed to the
console. This does not change audit results.}

\item{open}{Logical scalar. Defaults to `FALSE`. When `TRUE`, opens the
generated report in a browser via [utils::browseURL()]. This does not change
the report contents.}

\item{...}{Additional named arguments forwarded to [audit_leakage()] only
when `audit` is a [LeakFit]. These control how the audit is computed and
therefore change the report. Typical examples include `metric` (character),
`B` (integer), `perm_stratify` (logical), `batch_cols` (character vector),
`X_ref` (matrix/data.frame), `sim_method` (character), and
`duplicate_scope` (character). When omitted, [audit_leakage()] defaults are
used. Ignored when `audit` is already a [LeakAudit].}
}
\value{
Character string containing the absolute file path to the generated
  HTML report. The report is a self-contained HTML file that can be opened
  in any web browser. It includes sections for: cross-validated metric
  summaries, label-permutation test results (gap, p-value), batch/study
  association tests, confounder sensitivity analysis, calibration diagnostics
  (for binomial tasks), target leakage scan results, and duplicate detection
  findings. The path can be used with \code{\link[utils]{browseURL}} to open
  the report programmatically.
}
\description{
Creates an HTML report that summarizes a leakage audit for a resampled model.
The report is built from a [LeakAudit] (or created from a [LeakFit]) and
presents: cross-validated metric summaries, a label-permutation association
test of the chosen performance metric (auto-refit when refit data are
available; otherwise fixed predictions), batch or study association tests
between metadata and predictions, confounder sensitivity plots, calibration
checks for binomial tasks, a target leakage scan based on feature-outcome
similarity (with multivariate scan enabled by default for supported tasks),
and duplicate
detection across training and test folds. The
output is a self-contained HTML file with tables and plots for these checks
plus the audit parameters used.
}
\details{
The report does not refit models or reprocess data unless `perm_refit` triggers
refitting (`TRUE` or `"auto"` with a valid `perm_refit_spec`); it otherwise
inspects the predictions and metadata stored in the input.
Results are conditional on the
provided splits, selected metric, and any feature matrix supplied to
[audit_leakage()]. The univariate target leakage scan can miss multivariate
proxies, interaction leakage, or features not included in `X_ref`; the
multivariate scan (enabled by default for supported tasks) adds a model-based
check but still only uses features in `X_ref`. A
non-significant result does not prove the absence of leakage, especially with
small `B` or incomplete metadata. Rendering requires the
`rmarkdown` package and `ggplot2` for plots.
}
\examples{
set.seed(1)
df <- data.frame(
  subject = rep(1:6, each = 2),
  outcome = factor(rep(c(0, 1), 6)),
  x1 = rnorm(12),
  x2 = rnorm(12)
)

splits <- make_split_plan(df, outcome = "outcome",
                      mode = "subject_grouped", group = "subject",
                      v = 3, progress = FALSE)

custom <- list(
  glm = list(
    fit = function(x, y, task, weights, ...) {
      stats::glm(y ~ ., data = data.frame(y = y, x),
                 family = stats::binomial(), weights = weights)
    },
    predict = function(object, newdata, task, ...) {
      as.numeric(stats::predict(object,
                                newdata = as.data.frame(newdata),
                                type = "response"))
    }
  )
)

fit <- fit_resample(df, outcome = "outcome", splits = splits,
                    learner = "glm", custom_learners = custom,
                    metrics = "auc", refit = FALSE, seed = 1)

audit <- audit_leakage(fit, metric = "auc", B = 5, perm_stratify = FALSE)

if (requireNamespace("rmarkdown", quietly = TRUE) &&
    requireNamespace("ggplot2", quietly = TRUE) &&
    isTRUE(rmarkdown::pandoc_available("1.12.3"))) {
  out_file <- audit_report(audit, output_dir = tempdir(), quiet = TRUE)
  out_file
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary.LeakFit}
\alias{summary.LeakFit}
\title{Summarize a LeakFit object}
\usage{
\method{summary}{LeakFit}(object, digits = 3, ...)
}
\arguments{
\item{object}{A [LeakFit] object returned by [fit_resample()]. It should
contain `metric_summary` and `audit` slots; missing entries result in empty
sections in the printed report.}

\item{digits}{Integer scalar. Number of decimal places to print in numeric
summary tables. Defaults to 3; affects printed output only, not the
returned data.}

\item{...}{Unused. Included for S3 method compatibility; changing these
values has no effect.}
}
\value{
Invisibly returns `object@metric_summary`, a data frame of per-learner
  metric means and standard deviations computed across folds. This function
  does not recompute metrics.
}
\description{
Prints a compact console report for a [LeakFit] object created by
[fit_resample()]. The report lists task/outcome metadata, learners,
total folds, and cross-validated metrics summarized as mean and standard
deviation across completed folds, plus a small audit table with per-fold
train/test sizes and retained feature counts.
}
\details{
This summary is meant for quick sanity checks of the resampling setup and
performance. It does not run leakage diagnostics and will not detect target
leakage, duplicate samples, or batch/study confounding; use [audit_leakage()]
or `summary()` on a [LeakAudit] object for those checks.
}
\examples{
set.seed(1)
df <- data.frame(
  subject = rep(1:6, each = 2),
  outcome = factor(rep(c(0, 1), each = 6)),
  x1 = rnorm(12),
  x2 = rnorm(12)
)
splits <- make_split_plan(
  df,
  outcome = "outcome",
  mode = "subject_grouped",
  group = "subject",
  v = 3,
  stratify = TRUE,
  progress = FALSE
)
custom <- list(
  glm = list(
    fit = function(x, y, task, weights, ...) {
      stats::glm(y ~ ., data = data.frame(y = y, x),
                 family = stats::binomial(), weights = weights)
    },
    predict = function(object, newdata, task, ...) {
      as.numeric(stats::predict(object,
                                newdata = as.data.frame(newdata),
                                type = "response"))
    }
  )
)
fit <- fit_resample(df, outcome = "outcome", splits = splits,
                    learner = "glm", custom_learners = custom,
                    metrics = "auc", seed = 1)
summary_df <- summary(fit)
summary_df

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cv_uncertainty.R
\name{cv_ci}
\alias{cv_ci}
\title{Confidence intervals for cross-validated metrics}
\usage{
cv_ci(
  metrics_df,
  level = 0.95,
  method = c("normal", "nadeau_bengio"),
  n_train = NULL,
  n_test = NULL
)
}
\arguments{
\item{metrics_df}{Data.frame with columns \code{fold}, \code{learner}, and
one or more numeric metric columns.}

\item{level}{Confidence level (default 0.95).}

\item{method}{One of \code{"normal"} or \code{"nadeau_bengio"}.}

\item{n_train}{Average number of training samples per fold. Used only when
\code{method = "nadeau_bengio"}. NULL to use fallback variance.}

\item{n_test}{Average number of test samples per fold. Used only when
\code{method = "nadeau_bengio"}. NULL to use fallback variance.}
}
\value{
A data.frame with \code{learner} and, for each metric, columns
  \code{<metric>_mean}, \code{<metric>_sd}, \code{<metric>_ci_lo}, and
  \code{<metric>_ci_hi}.
}
\description{
Computes per-learner confidence intervals for each metric column in a
per-fold metrics data.frame. Supports the standard normal/t approach and the
Nadeau-Bengio (2003) corrected variance for repeated K-fold CV.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics_extra.R
\name{confounder_sensitivity}
\alias{confounder_sensitivity}
\title{Confounder sensitivity summaries}
\usage{
confounder_sensitivity(
  fit,
  confounders = NULL,
  metric = NULL,
  min_n = 10,
  coldata = NULL,
  numeric_bins = 4,
  learner = NULL,
  strict_align = FALSE
)
}
\arguments{
\item{fit}{A [LeakFit] object from [fit_resample()].}

\item{confounders}{Character vector of columns in `coldata` to evaluate.
Defaults to common batch/study identifiers when available.}

\item{metric}{Metric name to compute within each stratum. Defaults to the
first metric used in the fit (or task defaults if unavailable).}

\item{min_n}{Minimum samples per stratum; smaller strata return NA metrics.}

\item{coldata}{Optional data.frame of sample metadata. Defaults to
`fit@splits@info$coldata` when available.}

\item{numeric_bins}{Integer number of quantile bins for numeric confounders
with many unique values.}

\item{learner}{Optional character scalar. When predictions include multiple
learners, selects the learner to summarize.}

\item{strict_align}{Logical scalar. If TRUE, errors when coldata cannot be
aligned by row names or IDs and would fall back to row-order matching.
Default is FALSE.}
}
\value{
A data.frame with per-confounder, per-level metrics and counts.
}
\description{
Computes performance metrics within confounder strata to surface potential
confounding. Requires aligned metadata in `coldata`.
}
\examples{
set.seed(42)
df <- data.frame(
  subject = rep(1:15, each = 2),
  outcome = factor(rep(c(0, 1), 15)),
  batch = factor(rep(c("A", "B", "C"), 10)),
  x1 = rnorm(30),
  x2 = rnorm(30)
)
splits <- make_split_plan(df, outcome = "outcome",
                          mode = "subject_grouped", group = "subject",
                          v = 3, progress = FALSE)
custom <- list(
  glm = list(
    fit = function(x, y, task, weights, ...) {
      stats::glm(y ~ ., data = as.data.frame(x),
                 family = stats::binomial(), weights = weights)
    },
    predict = function(object, newdata, task, ...) {
      as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
                                type = "response"))
    }
  )
)
fit <- fit_resample(df, outcome = "outcome", splits = splits,
                    learner = "glm", custom_learners = custom,
                    metrics = "auc", refit = FALSE, seed = 1)
confounder_sensitivity(fit, confounders = "batch", coldata = df)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/guards.R
\name{guard_fit}
\alias{guard_fit}
\title{Fit leakage-safe preprocessing pipeline}
\usage{
guard_fit(
  X,
  y = NULL,
  steps = list(),
  task = c("binomial", "multiclass", "gaussian", "survival")
)
}
\arguments{
\item{X}{matrix/data.frame of predictors (training).}

\item{y}{Optional outcome for supervised feature selection.}

\item{steps}{List of configuration options (see Details).}

\item{task}{"binomial", "multiclass", "gaussian", or "survival".}
}
\value{
An object of class "GuardFit" with elements `transform`, `state`, `p_out`, and `steps`.
}
\description{
Builds and fits a guarded preprocessing pipeline on training data,
then constructs a transformer for consistent application to new data.
}
\details{
The pipeline applies, in order:
\itemize{
  \item Winsorization (optional) to limit outliers.
  \item Imputation learned on training data only.
  \item Normalization (z-score or robust).
  \item Variance/IQR filtering.
  \item Feature selection (optional; t-test, lasso, PCA).
}
All statistics are estimated on the training data and re-used for new data.
}
\examples{
x <- data.frame(a = c(1, 2, NA), b = c(3, 4, 5))
fit <- guard_fit(x, y = c(1, 2, 3),
                 steps = list(impute = list(method = "median")),
                 task = "gaussian")
fit$transform(x)
}
\seealso{
[predict_guard()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_split_plan.R
\name{check_split_overlap}
\alias{check_split_overlap}
\title{Check split overlap invariants}
\usage{
check_split_overlap(splits, coldata = NULL, cols = NULL, stop_on_fail = TRUE)
}
\arguments{
\item{splits}{A \code{LeakSplits} object from \code{\link{make_split_plan}}.}

\item{coldata}{A data.frame of sample metadata. When \code{NULL} (default),
the function uses \code{splits@info$coldata} if available.}

\item{cols}{Character vector of column names to check for overlap. When
\code{NULL} (default), the function infers columns from the split mode
(e.g., \code{group} for \code{subject_grouped}, \code{batch} for
\code{batch_blocked}, both axes for \code{combined}).}

\item{stop_on_fail}{Logical; if \code{TRUE} (default), raises an error when
any overlap is detected.}
}
\value{
A data.frame with one row per fold-by-column combination and
  columns \code{fold}, \code{repeat_id}, \code{col}, \code{n_overlap}
  (number of overlapping group values), and \code{pass} (logical).
  Invisible. Raises an error if any fold fails and \code{stop_on_fail = TRUE}.
}
\description{
Verifies that a \code{\linkS4class{LeakSplits}} object satisfies the
expected no-overlap constraints for one or more grouping columns. For each
fold, the function checks that no group-level value appearing in the test
set is also present in the training set.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidymodels.R
\name{as_rsample}
\alias{as_rsample}
\title{Convert LeakSplits to an rsample resample set}
\usage{
as_rsample(x, data = NULL, ...)
}
\arguments{
\item{x}{LeakSplits object created by [make_split_plan()].}

\item{data}{Optional data.frame used to populate rsample splits. When NULL,
the stored `coldata` from `x` is used (if available).}

\item{...}{Additional arguments passed to methods (unused).}
}
\value{
An rsample \code{rset} object compatible with tidymodels workflows.
  The returned object is a tibble with class \code{rset} containing:
  \describe{
    \item{\code{splits}}{List-column of \code{rsplit} objects, each with
      \code{analysis} (training indices) and \code{assessment} (test indices).}
    \item{\code{id}}{Character column with fold identifiers (e.g., "Fold1").}
    \item{\code{id2}}{Character column with repeat identifiers (e.g., "Repeat1")
      when multiple repeats are present; otherwise absent.}
  }
  The object also carries attributes for \code{group}, \code{batch},
  \code{study}, \code{time} (when available from the original \code{LeakSplits}),
  and \code{bioLeak_mode} indicating the original splitting mode. This allows
  the splits to be used with \code{tune::tune_grid()}, \code{rsample::fit_resamples()},
  and other tidymodels functions.
}
\description{
Convert LeakSplits to an rsample resample set
}
\examples{
if (requireNamespace("rsample", quietly = TRUE)) {
  df <- data.frame(
    subject = rep(1:10, each = 2),
    outcome = rbinom(20, 1, 0.5),
    x1 = rnorm(20),
    x2 = rnorm(20)
  )
  splits <- make_split_plan(df, outcome = "outcome",
                        mode = "subject_grouped", group = "subject", v = 5)
  rset <- as_rsample(splits, data = df)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_audit.R
\name{plot_time_acf}
\alias{plot_time_acf}
\title{Plot ACF of test predictions for time-series leakage checks}
\usage{
plot_time_acf(fit, lag.max = 20)
}
\arguments{
\item{fit}{LeakFit.}

\item{lag.max}{maximum lag to show.}
}
\value{
A list with the autocorrelation results, \code{lag.max}, and a ggplot object.
}
\description{
Uses the autocorrelation function of out-of-fold predictions to detect
temporal dependence that may indicate leakage. Predictions are ordered by
the split time column before computing the ACF. Requires numeric predictions
(regression or survival). Requires ggplot2.
}
\examples{
if (requireNamespace("ggplot2", quietly = TRUE)) {
  set.seed(42)
  df <- data.frame(
    id = 1:30,
    time = seq.Date(as.Date("2020-01-01"), by = "day", length.out = 30),
    y = rnorm(30),
    x1 = rnorm(30),
    x2 = rnorm(30)
  )
  splits <- make_split_plan(df, outcome = "y", mode = "time_series",
                            time = "time", v = 3, progress = FALSE)
  custom <- list(
    lm = list(
      fit = function(x, y, task, weights, ...) {
        stats::lm(y ~ ., data = data.frame(y = y, x))
      },
      predict = function(object, newdata, task, ...) {
        as.numeric(stats::predict(object, newdata = as.data.frame(newdata)))
      }
    )
  )
  fit <- fit_resample(df, outcome = "y", splits = splits,
                      learner = "lm", custom_learners = custom,
                      metrics = "rmse", refit = FALSE, seed = 1)
  plot_time_acf(fit, lag.max = 10)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{audit_info}
\alias{audit_info}
\title{Auxiliary information list from a LeakAudit}
\usage{
audit_info(audit)
}
\arguments{
\item{audit}{A `LeakAudit` object returned by [audit_leakage()].}
}
\value{
A named list of supplementary audit components.
}
\description{
Returns the auxiliary `info` list stored in a `LeakAudit` object.
This list carries non-tabular audit components such as
multivariate-target-scan results, configuration flags, and
permutation-test diagnostics.
}
\seealso{
[audit_leakage()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_resample.R
\name{fit_resample}
\alias{fit_resample}
\title{Fit and evaluate with leakage guards over predefined splits}
\usage{
fit_resample(
  x,
  outcome,
  splits,
  preprocess = list(impute = list(method = "median"), normalize = list(method =
    "zscore"), filter = list(var_thresh = 0, iqr_thresh = 0), fs = list(method = "none")),
  learner = c("glmnet", "ranger"),
  learner_args = list(),
  custom_learners = list(),
  metrics = c("auc", "pr_auc", "accuracy"),
  class_weights = NULL,
  positive_class = NULL,
  classification_threshold = 0.5,
  parallel = FALSE,
  refit = TRUE,
  seed = 1,
  split_cols = "auto",
  store_refit_data = TRUE
)
}
\arguments{
\item{x}{SummarizedExperiment or matrix/data.frame}

\item{outcome}{outcome column name (if x is SE or data.frame), or a length-2
character vector of time/event column names for survival outcomes.}

\item{splits}{LeakSplits object from make_split_plan(), or an `rsample` rset/rsplit.}

\item{preprocess}{list(impute, normalize, filter=list(...), fs) or a
`recipes::recipe` object. When a recipe is supplied, the guarded preprocessing
pipeline is bypassed and the recipe is prepped on training data only.
Recipe/workflow leakage guardrails run before fitting; configure policy via
\code{options(bioLeak.validation_mode = "warn" | "error" | "off")}.}

\item{learner}{parsnip model_spec (or list of model_spec objects) describing
the model(s) to fit, or a `workflows::workflow`. For legacy use, a character
vector of learner names (e.g., "glmnet", "ranger") or custom learner IDs is
still supported.}

\item{learner_args}{list of additional arguments passed to legacy learners
(ignored when `learner` is a parsnip model_spec).}

\item{custom_learners}{named list of custom learner definitions used only
with legacy character learners. Each entry
must contain \code{fit} and \code{predict} functions. The \code{fit} function
should accept \code{x}, \code{y}, \code{task}, and \code{weights}, and return
a model object. The \code{predict} function should accept \code{object},
\code{newdata}, and \code{task}. For binomial/regression/survival tasks it
should return a numeric vector; for multiclass tasks it should return either
class labels or a matrix/data.frame of class probabilities.}

\item{metrics}{named list of metric functions, vector of metric names, or a
`yardstick::metric_set`. When a yardstick metric set (or list of yardstick
metric functions) is supplied, metrics are computed using yardstick with the
positive class set to the second factor level.}

\item{class_weights}{optional named numeric vector of weights for binomial or
multiclass outcomes}

\item{positive_class}{optional value indicating the positive class for binomial outcomes.
When set, the outcome levels are reordered so that \code{positive_class} is treated
as the positive class (level 2). If NULL, the second factor level is used.}

\item{classification_threshold}{Numeric threshold in \code{[0, 1]} used to
convert binomial probabilities into class predictions for
\code{pred_class} and accuracy metrics. Ignored for non-binomial tasks.}

\item{parallel}{logical, use future.apply for multicore execution}

\item{refit}{logical, if TRUE retrain final model on full data}

\item{seed}{integer, for reproducibility}

\item{split_cols}{Optional named list/character vector or `"auto"` (default)
overriding group/batch/study/time column names when `splits` is an rsample
object and its attributes are missing. `"auto"` falls back to common
metadata column names (e.g., `group`, `subject`, `batch`, `study`, `time`).
Supported names are `group`, `batch`, `study`, and `time`.}

\item{store_refit_data}{Logical; when TRUE (default), stores the original
data and learner configuration inside the fit to enable refit-based
permutation tests without manual `perm_refit_spec` setup.}
}
\value{
A \code{\linkS4class{LeakFit}} S4 object containing:
  \describe{
    \item{\code{splits}}{The \code{LeakSplits} object used for resampling.}
    \item{\code{metrics}}{Data.frame of per-fold, per-learner performance
      metrics with columns \code{fold}, \code{learner}, and one column per
      requested metric.}
    \item{\code{metric_summary}}{Data.frame summarizing metrics across folds
      for each learner with columns \code{learner}, and \code{<metric>_mean}
      and \code{<metric>_sd} for each requested metric.}
    \item{\code{audit}}{Data.frame with per-fold audit information including
      \code{fold}, \code{n_train}, \code{n_test}, \code{learner}, and
      \code{features_final} (number of features after preprocessing).}
    \item{\code{predictions}}{List of data.frames containing out-of-fold
      predictions with columns \code{id} (sample identifier), \code{truth}
      (true outcome), \code{pred} (predicted value or probability), \code{fold},
      and \code{learner}. For classification tasks, includes \code{pred_class}.
      For multiclass, includes per-class probability columns.}
    \item{\code{preprocess}}{List of preprocessing state objects from each fold,
      storing imputation parameters, normalization statistics, and feature
      selection results.}
    \item{\code{learners}}{List of fitted model objects from each fold.}
    \item{\code{outcome}}{Character string naming the outcome variable.}
    \item{\code{task}}{Character string indicating the task type
      (\code{"binomial"}, \code{"multiclass"}, \code{"gaussian"}, or
      \code{"survival"}).}
    \item{\code{feature_names}}{Character vector of feature names after
      preprocessing.}
    \item{\code{info}}{List of additional metadata including \code{hash},
      \code{metrics_used}, \code{class_weights}, \code{positive_class},
      \code{sample_ids}, \code{fold_status}, \code{refit}, \code{final_model} (refitted model if
      \code{refit = TRUE}), \code{final_preprocess}, \code{learner_names},
      and \code{perm_refit_spec} (for permutation-based audits).}
  }
  Use \code{summary()} to print a formatted report, or access slots directly
  with \code{@}.
}
\description{
Performs cross-validated model training and evaluation using
leakage-protected preprocessing (.guard_fit) and user-specified learners.
}
\details{
Preprocessing is fit on the training fold and applied to the test fold,
preventing leakage from global imputation, scaling, or feature selection.
When a `recipes::recipe` or `workflows::workflow` is supplied, the recipe is
prepped on the training fold and baked on the test fold.
For data.frame or matrix inputs, columns used to define splits
(outcome, group, batch, study, time) are excluded from the predictor matrix.
Use \code{learner_args} to pass model-specific arguments, either as a named
list keyed by learner or a single list applied to all learners. For custom
learners, \code{learner_args[[name]]} may be a list with \code{fit} and
\code{predict} sublists to pass distinct arguments to each stage. For binomial
tasks, predictions and metrics assume the positive class is the second factor
level; use \code{positive_class} to control this. Use
\code{classification_threshold} to change the probability cutoff used for
class labels and accuracy. Parsnip learners must support
probability predictions for binomial metrics (AUC/PR-AUC/accuracy) and
multiclass log-loss when requested.
}
\examples{
set.seed(1)
df <- data.frame(
  subject = rep(1:10, each = 2),
  outcome = rbinom(20, 1, 0.5),
  x1 = rnorm(20),
  x2 = rnorm(20)
)
splits <- make_split_plan(df, outcome = "outcome",
                      mode = "subject_grouped", group = "subject", v = 5)

# glmnet learner (requires glmnet package)
fit <- fit_resample(df, outcome = "outcome", splits = splits,
                      learner = "glmnet", metrics = "auc")
summary(fit)

# Custom learner (logistic regression) - no extra packages needed
custom <- list(
  glm = list(
    fit = function(x, y, task, weights, ...) {
      stats::glm(y ~ ., data = as.data.frame(x),
                 family = stats::binomial(), weights = weights)
    },
    predict = function(object, newdata, task, ...) {
      as.numeric(stats::predict(object, newdata = as.data.frame(newdata), type = "response"))
    }
  )
)
fit2 <- fit_resample(df, outcome = "outcome", splits = splits,
                     learner = "glm", custom_learners = custom,
                     metrics = "accuracy")

summary(fit2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/impute_guarded.R
\name{impute_guarded}
\alias{impute_guarded}
\title{Leakage-safe data imputation via guarded preprocessing}
\usage{
impute_guarded(
  train,
  test,
  method = c("median", "knn", "missForest", "none"),
  constant_value = 0,
  k = 5,
  seed = 123,
  winsor = TRUE,
  winsor_thresh = 3,
  parallel = FALSE,
  return_outliers = FALSE,
  vars = NULL
)
}
\arguments{
\item{train}{data frame (training set)}

\item{test}{data frame (test set)}

\item{method}{one of "median", "knn", "missForest", or "none"}

\item{constant_value}{unused; retained for backward compatibility}

\item{k}{number of neighbors for kNN imputation (if method = "knn")}

\item{seed}{unused; retained for backward compatibility. Set seed before
calling this function if reproducibility is needed.}

\item{winsor}{logical; apply MAD-based winsorization before imputation}

\item{winsor_thresh}{numeric; MAD cutoff (default = 3)}

\item{parallel}{logical; unused (kept for compatibility)}

\item{return_outliers}{logical; unused (outlier flags not returned)}

\item{vars}{optional character vector; impute only selected variables}
}
\value{
A list (S3 class \code{"LeakImpute"}) with elements \code{train},
  \code{test}, \code{model}, \code{method}, \code{summary}, and \code{outliers}.
}
\description{
Fits imputation parameters on the training data only, then applies the same
guarded transformation to the test data. This function is a thin wrapper
around the guarded preprocessing used by \code{fit_resample()}.
Output is the transformed feature matrix used by the guarded pipeline
(categorical variables are one-hot encoded).
}
\examples{
train <- data.frame(x = c(1, 2, NA, 4), y = c(NA, 1, 1, 0))
test <- data.frame(x = c(NA, 5), y = c(1, NA))
imp <- impute_guarded(train, test, method = "median", winsor = FALSE)
imp$train
imp$test
}
\seealso{
[fit_resample()], [predict_guard()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time_series_blocks.R
\name{.circular_block_permute}
\alias{.circular_block_permute}
\title{Circular block permutation indices}
\usage{
.circular_block_permute(idx, block_len)
}
\arguments{
\item{idx}{Integer vector of ordered indices.}

\item{block_len}{Positive integer block length (>= 1).}
}
\value{
Integer vector of permuted indices of the same length as `idx`.
}
\description{
Generates a permutation of time indices by concatenating random-length blocks
sampled circularly from the ordered sequence. Used for creating
block-permuted surrogates that preserve short-range temporal structure.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/delta_lsi.R
\name{summary.LeakDeltaLSI}
\alias{summary.LeakDeltaLSI}
\title{Summarize a LeakDeltaLSI object}
\usage{
\method{summary}{LeakDeltaLSI}(object, digits = 3L, ...)
}
\arguments{
\item{object}{A \code{LeakDeltaLSI} object from \code{\link{delta_lsi}}.}

\item{digits}{Integer. Number of decimal places to show (default 3).}

\item{...}{Unused.}
}
\value{
Invisibly returns \code{object}.
}
\description{
Prints a human-readable summary of the Delta LSI analysis comparing leaky vs
guarded evaluation pipelines.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{audit_duplicates}
\alias{audit_duplicates}
\title{Near-duplicate sample pairs from a LeakAudit}
\usage{
audit_duplicates(audit)
}
\arguments{
\item{audit}{A `LeakAudit` object returned by [audit_leakage()].}
}
\value{
A `data.frame` (possibly empty) with one row per detected
  near-duplicate pair.
}
\description{
Returns the near-duplicate detection data frame stored in a
`LeakAudit` object. Each row is one (i, j) sample pair whose
similarity exceeds the configured threshold.
}
\seealso{
[audit_leakage()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary.LeakAudit}
\alias{summary.LeakAudit}
\title{Summarize a leakage audit}
\usage{
\method{summary}{LeakAudit}(object, digits = 3, ...)
}
\arguments{
\item{object}{A `LeakAudit` object from [audit_leakage()]. The summary reads
stored results from `object` and prints them to the console.}

\item{digits}{Integer number of digits to show when formatting numeric
statistics in the console output. Defaults to `3`. Increasing `digits`
shows more precision; decreasing it shortens the printout without changing
the underlying values.}

\item{...}{Unused. Included for S3 method compatibility; additional
arguments are ignored.}
}
\value{
Invisibly returns `object` after printing the summary.
}
\description{
Prints a concise, human-readable report for a `LeakAudit` object produced by
[audit_leakage()]. The summary surfaces four diagnostics when available:
label-permutation gap (prediction-label association by default), batch/study
association tests (metadata aligned with fold splits), target leakage scan
(features strongly associated with the outcome), and near-duplicate detection
(high similarity in `X_ref`). The output reflects the stored audit results
only; it does not recompute any tests.
}
\details{
The permutation test quantifies prediction-label association when using fixed
predictions; refit-based permutations require `perm_refit = TRUE` (or `"auto"`
with refit data). It does not by itself prove or rule out leakage.
Batch association flags metadata that align with fold assignment; this may
reflect study design rather than leakage.
Target leakage scan uses univariate feature-outcome associations and can miss
multivariate proxies, interaction leakage, or features not included in `X_ref`.
The multivariate scan (enabled by default for supported tasks) reports an
additional model-based score.
Duplicate detection only considers the provided `X_ref` features and the
similarity threshold used during [audit_leakage()]. By default,
`duplicate_scope = "train_test"` filters to pairs that cross train/test;
set `duplicate_scope = "all"` to include within-fold duplicates.
Sections are reported as "not available" when the corresponding audit
component was not computed.
}
\examples{
set.seed(1)
df <- data.frame(
  subject = rep(1:6, each = 2),
  outcome = rbinom(12, 1, 0.5),
  x1 = rnorm(12),
  x2 = rnorm(12)
)
splits <- make_split_plan(df, outcome = "outcome",
                      mode = "subject_grouped", group = "subject", v = 3)
custom <- list(
  glm = list(
    fit = function(x, y, task, weights, ...) {
      stats::glm(y ~ ., data = as.data.frame(x),
                 family = stats::binomial(), weights = weights)
    },
    predict = function(object, newdata, task, ...) {
      as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
                                type = "response"))
    }
  )
)
fit <- fit_resample(df, outcome = "outcome", splits = splits,
                    learner = "glm", custom_learners = custom,
                    metrics = "auc", refit = FALSE, seed = 1)
audit <- audit_leakage(fit, metric = "auc", B = 5,
                       X_ref = df[, c("x1", "x2")], seed = 1)
summary(audit) # prints the audit report and returns `audit` invisibly

}
\seealso{
[plot_perm_distribution()], [plot_fold_balance()], [plot_overlap_checks()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_split_plan.R
\name{show,LeakSplits-method}
\alias{show,LeakSplits-method}
\title{Display summary for LeakSplits objects}
\usage{
\S4method{show}{LeakSplits}(object)
}
\arguments{
\item{object}{LeakSplits object.}
}
\value{
No return value, called for side effects (prints a summary to the
  console showing mode, fold count, repeats, outcome, stratification status,
  nested status, per-fold train/test sizes, and the reproducibility hash).
}
\description{
Prints fold counts, sizes, and hash metadata for quick inspection.
}
\examples{
df <- data.frame(
  subject = rep(1:10, each = 2),
  outcome = rbinom(20, 1, 0.5),
  x1 = rnorm(20),
  x2 = rnorm(20)
)
splits <- make_split_plan(df, outcome = "outcome",
                      mode = "subject_grouped", group = "subject", v = 5)
show(splits)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit.R
\name{audit_leakage}
\alias{audit_leakage}
\title{Audit leakage and confounding}
\usage{
audit_leakage(
  fit,
  metric = c("auc", "pr_auc", "accuracy", "macro_f1", "log_loss", "rmse", "cindex"),
  B = 200,
  perm_stratify = FALSE,
  perm_refit = "auto",
  perm_refit_auto_max = 200,
  perm_refit_spec = NULL,
  perm_mode = NULL,
  time_block = c("circular", "stationary"),
  block_len = NULL,
  include_z = TRUE,
  ci_method = c("if", "bootstrap"),
  boot_B = 400,
  parallel = FALSE,
  seed = 1,
  return_perm = TRUE,
  batch_cols = NULL,
  coldata = NULL,
  X_ref = NULL,
  target_scan = TRUE,
  target_scan_multivariate = TRUE,
  target_scan_multivariate_B = 100,
  target_scan_multivariate_components = 10,
  target_scan_multivariate_interactions = TRUE,
  target_threshold = 0.9,
  target_p_adjust = c("none", "BH", "BY", "holm", "bonferroni"),
  target_alpha = 0.05,
  feature_space = c("raw", "rank"),
  sim_method = c("cosine", "pearson"),
  sim_threshold = 0.995,
  nn_k = 50,
  max_pairs = 5000,
  duplicate_scope = c("train_test", "all"),
  learner = NULL,
  strict_align = FALSE
)
}
\arguments{
\item{fit}{A [LeakFit] object from [fit_resample()] containing cross-validated
predictions and split metadata. If predictions include learner IDs for
multiple models, you must supply `learner` to select one; if learner IDs are
absent, the audit uses all predictions and may mix learners.}

\item{metric}{Character scalar. One of `"auc"`, `"pr_auc"`, `"accuracy"`,
`"macro_f1"`, `"log_loss"`, `"rmse"`, or `"cindex"`. Defaults to `"auc"`.
This controls the observed performance
statistic, the permutation null, and the sign of the reported gap.}

\item{B}{Integer scalar. Number of permutations used to build the null
distribution (default 200). Larger values reduce Monte Carlo error but
increase runtime.}

\item{perm_stratify}{Logical scalar or `"auto"`. If TRUE, permutations
are stratified within each fold (factor levels; numeric outcomes are binned
into quantiles when enough non-missing values are available). If FALSE, no
stratification is used. Defaults to FALSE. Stratification only applies when
`coldata` supplies the outcome; otherwise labels are shuffled within each fold.}

\item{perm_refit}{Logical scalar or `"auto"`. If FALSE, permutations keep
predictions fixed and shuffle labels (association test). If TRUE, each
permutation refits the model on permuted outcomes using `perm_refit_spec`.
Refit-based permutations are slower but better approximate a full null
distribution. The default is `"auto"`, which refits only when
`perm_refit_spec` is provided and `B` is less than or equal to
`perm_refit_auto_max`; otherwise it falls back to fixed-prediction
permutations.}

\item{perm_refit_auto_max}{Integer scalar. Maximum `B` allowed for
`perm_refit = "auto"` to trigger refitting. Defaults to 200.}

\item{perm_refit_spec}{List of inputs used when `perm_refit = TRUE`.
Required elements: `x` (data used for fitting) and `learner` (parsnip
model_spec, workflow, or legacy learner). Optional elements: `outcome`
(defaults to `fit@outcome`), `preprocess`, `learner_args`,
`custom_learners`, `class_weights`, `positive_class`, and `parallel`.
Survival outcomes are not supported for refit-based permutations.}

\item{perm_mode}{Optional character scalar to override the permutation mode
used for restricted shuffles. One of `"subject_grouped"`, `"batch_blocked"`,
`"study_loocv"`, or `"time_series"`. Defaults to the split metadata when
available (including rsample-derived modes).}

\item{time_block}{Character scalar, `"circular"` or `"stationary"`. Controls
block permutation for `time_series` splits; ignored for other split modes.
Default is `"circular"`.}

\item{block_len}{Integer scalar or NULL. Block length for time-series
permutations. NULL selects `max(5, floor(0.1 * fold_size))`. Larger values
preserve more temporal structure and yield a more conservative null.}

\item{include_z}{Logical scalar. If TRUE (default), include the z-score for the
permutation gap when a standard error is available; if FALSE, `z` is NA.}

\item{ci_method}{Character scalar, `"if"` or `"bootstrap"`. Controls how the
standard error and confidence interval for the permutation gap are estimated.
Default is `"if"`. `"if"` uses an influence-function estimate when available; `"bootstrap"`
resamples permutation values `boot_B` times. Failed estimates yield NA.}

\item{boot_B}{Integer scalar. Number of bootstrap resamples when
`ci_method = "bootstrap"` (default 400). Larger values are more stable but slower.}

\item{parallel}{Logical scalar. If TRUE and `future.apply` is available,
permutations run in parallel. Results should match sequential execution.
Default is FALSE.}

\item{seed}{Integer scalar. Random seed used for permutations and bootstrap
resampling; changing it changes the randomization but not the observed metric.
Default is 1.}

\item{return_perm}{Logical scalar. If TRUE (default), stores the permutation
distribution in `audit@perm_values`. Set FALSE to reduce memory use.}

\item{batch_cols}{Character vector. Names of `coldata` columns to test for
association with fold assignment. If NULL, defaults to any of
`"batch"`, `"plate"`, `"center"`, `"site"`, `"study"` found in `coldata`.
Changing this controls which batch tests appear in `batch_assoc`.}

\item{coldata}{Optional data.frame of sample-level metadata. Rows must align
to prediction ids via row names, a `row_id` column, or row order. Used to
build restricted permutations (when the outcome column is present), compute
batch associations, and supply outcomes for target scans. If NULL, uses
`fit@splits@info$coldata` when available. If alignment fails, restricted
permutations are disabled with a warning.}

\item{X_ref}{Optional numeric matrix/data.frame (samples x features). Used for
duplicate detection and the target leakage scan. If NULL, uses
`fit@info$X_ref` when available. Rows must align to sample ids (split order)
via row names, a `row_id` column, or row order; misalignment disables these
checks.}

\item{target_scan}{Logical scalar. If TRUE (default), computes per-feature
outcome associations on `X_ref` and flags proxy features; if FALSE, or if
`X_ref`/outcomes are unavailable, `target_assoc` is empty. Not available
for survival outcomes.}

\item{target_scan_multivariate}{Logical scalar. If TRUE (default), fits a simple
multivariate/interaction model on `X_ref` using the stored splits and
reports a permutation-based score/p-value. This is slower and only
implemented for binomial and gaussian tasks.}

\item{target_scan_multivariate_B}{Integer scalar. Number of permutations for
the multivariate scan (default 100). Larger values stabilize the p-value.}

\item{target_scan_multivariate_components}{Integer scalar. Maximum number of
principal components used in the multivariate scan (default 10).}

\item{target_scan_multivariate_interactions}{Logical scalar. If TRUE (default),
adds pairwise interactions among the top components in the multivariate scan.}

\item{target_threshold}{Numeric scalar in (0,1). Threshold applied to the
association score used to flag proxy features. Higher values are stricter.
Default is 0.9.}

\item{target_p_adjust}{Character scalar. Multiple-testing correction method
applied to finite `target_assoc$p_value` values. One of `"none"` (default),
`"BH"`, `"BY"`, `"holm"`, or `"bonferroni"`. Adds columns
`p_value_adj` and `flag_fdr` to `target_assoc`.}

\item{target_alpha}{Numeric scalar in (0,1). Significance level used for
`flag_fdr` when `target_p_adjust != "none"`. Default is 0.05.}

\item{feature_space}{Character scalar, `"raw"` or `"rank"`. If `"rank"`,
each row of `X_ref` is rank-transformed before similarity calculations.
This affects duplicate detection only. Default is `"raw"`.}

\item{sim_method}{Character scalar, `"cosine"` or `"pearson"`. Similarity
metric for duplicate detection. `"pearson"` row-centers before cosine.
Default is `"cosine"`.}

\item{sim_threshold}{Numeric scalar in (0,1). Similarity cutoff for reporting
duplicate pairs (default 0.995). Higher values yield fewer pairs.}

\item{nn_k}{Integer scalar. For large datasets (`n > 3000`) with `RANN`
installed, checks only the nearest `nn_k` neighbors per row. Larger values
increase sensitivity but slow the search. Ignored when full comparisons are used.
Default is 50.}

\item{max_pairs}{Integer scalar. Maximum number of duplicate pairs returned.
If more pairs are found, only the most similar are kept. This does not
affect permutation results. Default is 5000.}

\item{duplicate_scope}{Character scalar. One of `"train_test"` (default) or
`"all"`. `"train_test"` retains only near-duplicate pairs that appear in
train vs test in at least one repeat; `"all"` reports all near-duplicate
pairs in `X_ref` regardless of fold assignment.}

\item{learner}{Optional character scalar. When predictions include multiple
learner IDs, selects the learner to audit. If NULL and multiple learners
are present, the function errors; if predictions lack learner IDs, this
argument is ignored with a warning. Default is NULL.}

\item{strict_align}{Logical scalar. If TRUE, errors instead of warning when
\code{X_ref} or \code{coldata} cannot be aligned to predictions by row
names or IDs and would otherwise fall back to row-order matching. Default
is FALSE for backward compatibility. Set to TRUE in production pipelines to
catch silent misalignment.}
}
\value{
A \code{\linkS4class{LeakAudit}} S4 object containing:
  \describe{
    \item{\code{fit}}{The \code{LeakFit} object that was audited.}
    \item{\code{permutation_gap}}{One-row data.frame from the permutation-gap
      test with columns: \code{metric_obs} (observed cross-validated metric),
      \code{perm_mean} (mean of permuted metrics), \code{perm_sd} (standard
      deviation), \code{gap} (observed minus permuted mean, or vice versa for
      loss metrics), \code{z} (standardized gap), \code{p_value}
      (permutation p-value), and \code{n_perm} (number of permutations). A
      large positive gap and small p-value suggest the model captures signal
      beyond random label assignment.}
    \item{\code{perm_values}}{Numeric vector of length \code{B} containing
      the metric value from each permutation. Useful for plotting the null
      distribution. Empty if \code{return_perm = FALSE}.}
    \item{\code{batch_assoc}}{Data.frame of chi-square association tests
      between fold assignment and batch/study metadata, with columns:
      \code{variable}, \code{stat} (chi-square statistic), \code{df}
      (degrees of freedom), \code{pval}, and \code{cramer_v} (effect size).
      Small p-values indicate potential confounding by design.}
    \item{\code{target_assoc}}{Data.frame of per-feature outcome associations
      with columns: \code{feature}, \code{type} (\code{"numeric"} or
      \code{"categorical"}), \code{metric} (AUC, correlation, eta_sq, or
      Cramer's V depending on task), \code{value}, \code{score} (scaled
      effect size), \code{p_value}, \code{n}, and \code{flag} (TRUE if
      \code{score >= target_threshold}). Flagged features may indicate
      target leakage.}
    \item{\code{duplicates}}{Data.frame of near-duplicate sample pairs with
      columns: \code{i}, \code{j} (row indices in \code{X_ref}), \code{sim}
      (similarity value), and \code{cross_fold} (whether the pair spans
      train vs test). Duplicates across folds can inflate performance.}
    \item{\code{trail}}{List capturing audit parameters and intermediate
      results for reproducibility, including \code{metric}, \code{B},
      \code{seed}, \code{perm_stratify}, \code{perm_refit}, and timing info.}
    \item{\code{info}}{List with additional metadata including multivariate
      scan results when \code{target_scan_multivariate = TRUE}.}
  }
  Use \code{summary()} to print a human-readable report, or access slots
  directly with \code{@}.
}
\description{
Computes a post-hoc leakage audit for a resampled model fit. The audit
(1) runs a permutation-gap test comparing observed cross-validated
performance to a label-permutation null (by default refitting when data
are available; otherwise using fixed predictions),
(2) tests whether fold assignments are associated with batch or study
metadata (confounding by design),
(3) scans features for unusually strong outcome proxies, and (4) flags
duplicate or near-duplicate samples in a reference feature matrix.

The returned [LeakAudit] summarizes these diagnostics. It relies on the
stored predictions, splits, and optional metadata; it does not refit models
unless `perm_refit = TRUE` (or `perm_refit = "auto"` with a valid
`perm_refit_spec`). Results are conditional on the chosen metric
and supplied metadata/features and should be interpreted as diagnostics,
not proof of leakage or its absence.
}
\details{
The `permutation_gap` slot reports `metric_obs`, `perm_mean`, `perm_sd`,
`gap`, `z`, `p_value`, and `n_perm`. The gap is defined as
`metric_obs - perm_mean` for metrics where higher is better (AUC, PR-AUC,
accuracy, macro-F1, C-index) and `perm_mean - metric_obs` for RMSE/log-loss.
By default, `perm_refit = "auto"` refits models when refit data are available
and `B` is not too large; otherwise it keeps predictions fixed and shuffles
labels. Fixed-prediction permutations quantify prediction-label association
rather than a full refit null. Set `perm_refit = FALSE` to force fixed
predictions, or `perm_refit = TRUE` (with `perm_refit_spec`) to always refit.

`batch_assoc` contains chi-square tests between fold assignment and each
`batch_cols` variable (`stat`, `df`, `pval`, `cramer_v`). `target_assoc`
reports feature-wise outcome associations on `X_ref`; numeric features use
AUC (binomial), `eta_sq` (multiclass), or correlation (gaussian), while
categorical features use Cramer's V (binomial/multiclass) or `eta_sq` from a
one-way ANOVA (gaussian). The
`score` column is the scaled effect size used for heuristic flagging
(`flag = score >= target_threshold`).
When `target_p_adjust != "none"`, finite `p_value` entries also receive
multiplicity-adjusted `p_value_adj` and `flag_fdr = (p_value_adj <= target_alpha)`.
The univariate target leakage scan can miss multivariate proxies, interaction
leakage, or features not included in `X_ref`. The multivariate scan (enabled
by default for supported tasks) adds a model-based proxy check but still only
covers features present in `X_ref`.

Duplicate detection compares rows of `X_ref` using the chosen `sim_method`
(cosine on L2-normalized rows, or Pearson via row-centering), optionally after
rank transformation (`feature_space = "rank"`). By default,
`duplicate_scope = "train_test"` filters to pairs that appear in train vs test
in at least one repeat; set `duplicate_scope = "all"` to include within-fold
duplicates. The `duplicates` slot returns index pairs and similarity values
for near-duplicate samples. Only duplicates present in `X_ref` can be
detected, and checks are skipped if inputs cannot be aligned to splits.
}
\examples{
set.seed(1)
df <- data.frame(
  subject = rep(1:6, each = 2),
  outcome = rbinom(12, 1, 0.5),
  x1 = rnorm(12),
  x2 = rnorm(12)
)

splits <- make_split_plan(df, outcome = "outcome",
                      mode = "subject_grouped", group = "subject", v = 3,
                      progress = FALSE)

custom <- list(
  glm = list(
    fit = function(x, y, task, weights, ...) {
      stats::glm(y ~ ., data = as.data.frame(x),
                 family = stats::binomial(), weights = weights)
    },
    predict = function(object, newdata, task, ...) {
      as.numeric(stats::predict(object,
                                newdata = as.data.frame(newdata),
                                type = "response"))
    }
  )
)

fit <- fit_resample(df, outcome = "outcome", splits = splits,
                    learner = "glm", custom_learners = custom,
                    metrics = "auc", refit = FALSE, seed = 1)

audit <- audit_leakage(fit, metric = "auc", B = 10,
                       X_ref = df[, c("x1", "x2")])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{dlsi_ci}
\alias{dlsi_ci}
\title{BCa confidence interval from a LeakDeltaLSI}
\usage{
dlsi_ci(dlsi, which = c("robust", "metric"))
}
\arguments{
\item{dlsi}{A `LeakDeltaLSI` object returned by [delta_lsi()].}

\item{which}{Either `"robust"` (default) for the Huber estimate's
confidence interval, or `"metric"` for the raw arithmetic mean's
confidence interval.}
}
\value{
A length-two numeric vector `c(lower, upper)`. Returns
  `c(NA_real_, NA_real_)` when the interval is not computed (e.g.,
  the inference tier did not include CIs).
}
\description{
Returns the bias-corrected and accelerated (BCa) bootstrap confidence
interval stored in a `LeakDeltaLSI` object. By default returns the
interval for the Huber-robust delta estimate; set
`which = "metric"` to return the interval for the raw metric
difference instead.
}
\seealso{
[delta_lsi()], [dlsi_robust()], [dlsi_metric()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{dlsi_p_value}
\alias{dlsi_p_value}
\title{Sign-flip p-value from a LeakDeltaLSI}
\usage{
dlsi_p_value(dlsi)
}
\arguments{
\item{dlsi}{A `LeakDeltaLSI` object returned by [delta_lsi()].}
}
\value{
A length-one numeric scalar in `[0, 1]`, or `NA_real_` when
  the inference tier did not include hypothesis testing.
}
\description{
Returns the paired sign-flip randomization-test p-value stored in a
`LeakDeltaLSI` object.
}
\seealso{
[delta_lsi()], [dlsi_tier()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R, R/delta_lsi.R
\docType{class}
\name{LeakSplits-class}
\alias{LeakSplits-class}
\alias{LeakFit-class}
\alias{LeakAudit-class}
\alias{LeakDeltaLSI-class}
\alias{show,LeakDeltaLSI-method}
\title{S4 Classes for bioLeak Pipeline}
\usage{
\S4method{show}{LeakDeltaLSI}(object)
}
\arguments{
\item{object}{A \code{\linkS4class{LeakDeltaLSI}} object.}
}
\value{
An S4 object of the respective class.
}
\description{
These classes capture splits, model fits, and audit diagnostics produced by
\code{make_split_plan()}, \code{fit_resample()}, and \code{audit_leakage()}.
}
\section{Slots}{

\describe{
\item{\code{mode}}{Splitting mode. One of "subject_grouped", "batch_blocked",
"study_loocv", "time_series", or "combined".}

\item{\code{indices}}{List of resampling descriptors (train/test indices when available)}

\item{\code{info}}{(LeakSplits) Metadata associated with the split plan (mode, coldata, hash, etc.)}

\item{\code{splits}}{A [`LeakSplits`] object used for resampling}

\item{\code{metrics}}{Model performance metrics per resample}

\item{\code{metric_summary}}{Summary of metrics across resamples}

\item{\code{audit}}{Audit information per resample}

\item{\code{predictions}}{List of prediction objects}

\item{\code{preprocess}}{Preprocessing steps used during fitting}

\item{\code{learners}}{Learner definitions used in the pipeline}

\item{\code{outcome}}{Outcome variable name}

\item{\code{task}}{Modeling task name}

\item{\code{feature_names}}{Feature names included in the model}

\item{\code{info}}{(LeakFit) Metadata about the model fit (sample IDs, timings, provenance, etc.)}

\item{\code{fit}}{A [`LeakFit`] object used to generate the audit}

\item{\code{permutation_gap}}{Data frame summarising permutation gaps}

\item{\code{perm_values}}{Numeric vector of permutation-based scores}

\item{\code{batch_assoc}}{Data frame of batch associations}

\item{\code{target_assoc}}{Data frame of feature-wise outcome associations}

\item{\code{duplicates}}{Data frame detailing duplicate records}

\item{\code{trail}}{List capturing audit trail information}

\item{\code{info}}{(LeakAudit) Metadata about the audit (mechanism summary, settings, provenance, etc.)}

\item{\code{metric}}{Performance metric compared between pipelines}

\item{\code{exchangeability}}{Exchangeability assumption used for the sign-flip test}

\item{\code{tier}}{Inference tier label based on effective number of repeats}

\item{\code{strict}}{Whether strict mode was requested}

\item{\code{R_eff}}{Effective number of paired repeats available for inference}

\item{\code{delta_lsi}}{Huber-robust point estimate of repeat-level metric difference}

\item{\code{delta_lsi_ci}}{BCa 95\% CI for delta_lsi (NA when R_eff < 10)}

\item{\code{delta_metric}}{Arithmetic mean of repeat-level metric differences}

\item{\code{delta_metric_ci}}{BCa 95\% CI for delta_metric (NA when R_eff < 10)}

\item{\code{p_value}}{Sign-flip randomization test p-value (NA when R_eff < 5 or unpaired)}

\item{\code{inference_ok}}{TRUE when tier A (R_eff >= 20, paired, finite p and CI)}

\item{\code{folds_naive}}{Per-fold data frame for the naive pipeline}

\item{\code{folds_guarded}}{Per-fold data frame for the guarded pipeline}

\item{\code{repeats_naive}}{Per-repeat aggregate data frame for the naive pipeline}

\item{\code{repeats_guarded}}{Per-repeat aggregate data frame for the guarded pipeline}

\item{\code{info}}{(LeakDeltaLSI) Metadata including R_naive, R_guarded, paired status, and block details}
}}

\seealso{
[make_split_plan()], [fit_resample()], [audit_leakage()]

[fit_resample()]

[audit_leakage()], [audit_report()]

[delta_lsi()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/guards.R
\name{predict.GuardFit}
\alias{predict.GuardFit}
\alias{predict_guard}
\title{Apply a fitted GuardFit transformer to new data}
\usage{
\method{predict}{GuardFit}(object, newdata, ...)

predict_guard(fit, newdata)
}
\arguments{
\item{object, fit}{A \code{GuardFit} object created by [guard_fit()].
Contains the training-time preprocessing settings and statistics. Changing
the object (for example, a different imputation method or feature
selection step) changes the output columns and values. \code{object} is
the canonical name (matching the S3 \code{predict()} generic);
\code{fit} is the legacy name accepted only by \code{predict_guard()}.}

\item{newdata}{A matrix or data.frame of predictors with one row per sample.
This required argument (no default) is transformed using the training-time
parameters in the fit only. Missing columns are added and filled, extra
columns are dropped, and factor levels are aligned to the training levels;
if the training fit was numeric-only, non-numeric columns in \code{newdata}
trigger an error.}

\item{...}{Ignored. Present so that the S3 method signature matches the
[stats::predict()] generic; additional arguments are silently dropped.}
}
\value{
A data.frame of transformed predictors with the same number of rows
as \code{newdata}. Column order and content match the training pipeline and
may include derived features (one-hot encodings, missingness indicators, or
PCA components). This output is not a prediction; it is intended as input
to a downstream model and assumes the training-time preprocessing is valid
for the new data.
}
\description{
Applies the preprocessing steps stored in a \code{GuardFit} object to new
data without refitting any statistics. This is designed to prevent
validation leakage that would occur if imputation, scaling, filtering, or
feature selection were recomputed on evaluation data. It enforces the
training schema by aligning columns and factor levels, and it errors when a
numeric-only training fit receives non-numeric predictors. It does not
detect label leakage, duplicate samples, or train/test contamination.

\code{predict.GuardFit()} is the canonical S3 method --- callers can use
\code{predict(fit, newdata)} on a \code{GuardFit} object and the right
method is dispatched. \code{predict_guard()} is retained as a
backward-compatible thin alias that simply forwards to the S3 method, so
existing code that calls \code{predict_guard(fit, x)} continues to work.
}
\examples{
x_train <- data.frame(a = c(1, 2, NA, 4), b = c(10, 11, 12, 13))
fit <- guard_fit(
  x_train,
  y = c(0.1, 0.2, 0.3, 0.4),
  steps = list(impute = list(method = "median")),
  task = "gaussian"
)
x_new <- data.frame(a = c(NA, 5), b = c(9, 14))
## Canonical: dispatch through the predict() generic.
out <- predict(fit, x_new)
out
## Equivalent legacy form (kept for backward compatibility).
identical(out, predict_guard(fit, x_new))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/guards.R
\name{.guard_fit}
\alias{.guard_fit}
\title{Fit leakage-safe preprocessing pipeline}
\usage{
.guard_fit(
  X,
  y = NULL,
  steps = list(),
  task = c("binomial", "multiclass", "gaussian", "survival")
)
}
\arguments{
\item{X}{matrix/data.frame of predictors (training).}

\item{y}{Optional outcome for supervised feature selection.}

\item{steps}{List of configuration options (see Details).}

\item{task}{"binomial", "multiclass", "gaussian", or "survival".}
}
\value{
An object of class "GuardFit" with elements `transform`, `state`, `p_out`, and `steps`.
}
\description{
Builds and fits a guarded preprocessing pipeline on training data,
then constructs a transformer for consistent application to new data.
}
\details{
The pipeline applies, in order:
\itemize{
  \item Winsorization (optional) to limit outliers.
  \item Imputation learned on training data only.
  \item Normalization (z-score or robust).
  \item Variance/IQR filtering.
  \item Feature selection (optional; t-test, lasso, PCA).
}
All statistics are estimated on the training data and re-used for new data.
}
\examples{
x <- data.frame(a = c(1, 2, NA), b = c(3, 4, 5))
fit <- .guard_fit(x, y = c(1, 2, 3),
                  steps = list(impute = list(method = "median")),
                  task = "gaussian")
fit$transform(x)
}
\seealso{
[predict_guard()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{dlsi_R_eff}
\alias{dlsi_R_eff}
\title{Effective number of paired repeats from a LeakDeltaLSI}
\usage{
dlsi_R_eff(dlsi)
}
\arguments{
\item{dlsi}{A `LeakDeltaLSI` object returned by [delta_lsi()].}
}
\value{
A length-one integer.
}
\description{
Returns the effective number of paired repeats `R_eff` stored in a
`LeakDeltaLSI` object. This is the count of repeats that contribute
to the inference; it equals the smaller of the leaky and guarded
fits' repeat counts when the comparison is paired.
}
\seealso{
[delta_lsi()], [dlsi_tier()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/delta_lsi.R
\name{delta_lsi}
\alias{delta_lsi}
\title{Delta Leakage Sensitivity Index (Delta LSI)}
\usage{
delta_lsi(
  fit_leaky,
  fit_guarded,
  metric = "auc",
  exchangeability = c("iid", "by_group", "within_batch", "blocked_time"),
  learner = NULL,
  higher_is_better = NULL,
  block_size = NULL,
  M_boot = 2000L,
  M_flip = 10000L,
  strict = FALSE,
  return_details = FALSE,
  seed = 42L,
  ...
)
}
\arguments{
\item{fit_leaky}{A \code{\linkS4class{LeakFit}} object from the leaky
(unprotected) evaluation pipeline.}

\item{fit_guarded}{A \code{\linkS4class{LeakFit}} object from the guarded
(leakage-protected) evaluation pipeline.}

\item{metric}{Character. Performance metric to compare. Must appear in
\code{fit@metrics} of both fits (e.g., \code{"auc"}, \code{"rmse"}).}

\item{exchangeability}{Character. Exchangeability assumption for the
sign-flip test. One of \code{"iid"} (default), \code{"by_group"},
\code{"within_batch"}, \code{"blocked_time"}.
\code{"blocked_time"} activates a block sign-flip procedure that flips
contiguous groups of repeats together, preserving serial autocorrelation
under the null; see \code{block_size}.  \code{"by_group"} and
\code{"within_batch"} are stored and reported but inference still uses the
iid sign-flip procedure (a warning is issued; contributions welcome).
\code{"iid"} (default) applies the standard independent sign-flip test.}

\item{learner}{Optional character. Learner name to select from multi-learner
fits. If \code{NULL}, the first learner found in \code{fit@metrics} is used.}

\item{higher_is_better}{Logical or \code{NULL}. Whether a higher value of
\code{metric} indicates better performance. When \code{NULL} (default),
auto-detected from the metric name: \code{"rmse"}, \code{"mse"},
\code{"mae"}, \code{"log_loss"}, \code{"brier"}, \code{"error"},
\code{"loss"}, and \code{"deviance"} are treated as lower-is-better;
all others default to higher-is-better. Setting this correctly ensures
that a positive \code{delta_lsi} always indicates leakage inflation
(the naive pipeline is artificially more optimistic than the guarded one).}

\item{block_size}{Integer or \code{NULL}. Block length for the block
sign-flip test, used only when \code{exchangeability = "blocked_time"}.
When \code{NULL} (default), the block size is auto-estimated from the
first-order autocorrelation of \eqn{\{\Delta_r\}} and capped at
\code{floor(R/3)} to ensure at least three independent blocks.  A warning
is issued when the estimate is used with \code{R_eff < 20} because the
AR(1) estimate is noisy at small sample sizes.  Provide an explicit integer
to override auto-estimation.}

\item{M_boot}{Integer. Number of bootstrap samples for BCa CI (default 2000).}

\item{M_flip}{Integer. Maximum Monte Carlo samples for sign-flip test when
R_eff > 15 (default 10000).}

\item{strict}{Logical. If \code{TRUE}, error on insufficient R_eff instead
of a warning.}

\item{return_details}{Logical. If \code{TRUE}, include the per-repeat
\eqn{\Delta_r} vector and the original fit objects in the \code{info} slot.}

\item{seed}{Integer. Random seed for bootstrap and sign-flip test.}

\item{...}{Unused. Reserved for deprecated aliases such as
\code{fit_naive}.}
}
\value{
A \code{\linkS4class{LeakDeltaLSI}} object.
}
\description{
Compares a naive (potentially leaky) cross-validation pipeline against a
guarded (leakage-protected) pipeline and quantifies leakage-induced
performance inflation using the Leakage Sensitivity Index (LSI).
}
\details{
\subsection{Method}{
For each fit, per-fold metric values are extracted from \code{fit@metrics}
(or recomputed from \code{fit@predictions} if necessary). Fold test-set
sizes are used as weights to aggregate fold metrics into per-repeat
estimates \eqn{\mu_r}. The repeat-level delta
\eqn{\Delta_r = s \cdot (\mu_r^{\text{naive}} - \mu_r^{\text{guarded}})}
captures leakage-induced performance inflation for each CV repeat, where
\eqn{s = +1} for higher-is-better metrics (e.g., AUC) and \eqn{s = -1}
for lower-is-better metrics (e.g., RMSE), so that \eqn{\Delta_r > 0}
always indicates the naive pipeline is more optimistic than the guarded one.

The \strong{delta_lsi} point estimate is the Huber M-estimator (k = 1.345)
applied to \eqn{\{\Delta_r\}}, which is robust to occasional outlier
repeats. \strong{delta_metric} is the arithmetic mean of \eqn{\{\Delta_r\}}
for easy interpretation in the original metric's units.

Pairing requires that \code{fit_leaky} and \code{fit_guarded} share
\emph{identical fold structures} (same test-set membership per fold) in
addition to the same number of repeats.  When repeat counts match but fold
structures differ, a warning is issued and the fits are treated as unpaired.

When \eqn{R_{\text{eff}} \geq 5} (equal, paired repeats), a sign-flip
randomization test (Phipson & Smyth, 2010) is performed: under
\eqn{H_0} (no leakage) the sign of each \eqn{\Delta_r} is exchangeable.
All \eqn{2^R} sign combinations are enumerated exactly for
\eqn{R \leq 15} (no continuity correction); Monte Carlo sampling is used
for larger \eqn{R} with the Phipson & Smyth (2010) correction.

BCa bootstrap confidence intervals (Efron, 1987) require
\eqn{R_{\text{eff}} \geq 10}.
}

\subsection{Inference tiers}{
\describe{
  \item{\code{"A_full_inference"}}{R_eff >= 20: point + BCa CI + sign-flip p-value; \code{inference_ok = TRUE}}
  \item{\code{"B_signflip_ci"}}{10 <= R_eff < 20: point + sign-flip p-value + BCa CI}
  \item{\code{"C_signflip"}}{5 <= R_eff < 10: point + sign-flip p-value (no CI)}
  \item{\code{"D_insufficient"}}{R_eff < 5 or unpaired: point estimate only}
}}
}
\seealso{
\code{\link{audit_leakage}}, \code{\link{fit_resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/permute_labels.R
\docType{data}
\name{.quantile_break_cache}
\alias{.quantile_break_cache}
\title{Quantile break cache for permutation stratification}
\format{
An environment used to cache quantile breakpoints.
}
\usage{
.quantile_break_cache
}
\value{
An environment (internal data object, not a function).
}
\description{
Internal environment used to cache quantile breakpoints for numeric
outcomes during restricted permutation testing. This avoids recomputing
quantiles across repeated calls in \code{audit_leakage()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics_extra.R
\name{calibration_summary}
\alias{calibration_summary}
\title{Calibration diagnostics for binomial predictions}
\usage{
calibration_summary(fit, bins = 10, min_bin_n = 5, learner = NULL)
}
\arguments{
\item{fit}{A [LeakFit] object from [fit_resample()].}

\item{bins}{Integer number of probability bins for the calibration curve.}

\item{min_bin_n}{Minimum samples per bin used in plotting; bins smaller than
this are retained in the output but can be filtered by the caller.}

\item{learner}{Optional character scalar. When predictions include multiple
learners, selects the learner to summarize.}
}
\value{
A list with a `curve` data.frame and a one-row `metrics` data.frame
  containing ECE, MCE, and Brier score.
}
\description{
Computes reliability curve summaries and calibration metrics for a
binomial [LeakFit] using out-of-fold predictions.
}
\examples{
set.seed(42)
df <- data.frame(
  subject = rep(1:15, each = 2),
  outcome = factor(rep(c(0, 1), 15)),
  x1 = rnorm(30),
  x2 = rnorm(30)
)
splits <- make_split_plan(df, outcome = "outcome",
                          mode = "subject_grouped", group = "subject",
                          v = 3, progress = FALSE)
custom <- list(
  glm = list(
    fit = function(x, y, task, weights, ...) {
      stats::glm(y ~ ., data = as.data.frame(x),
                 family = stats::binomial(), weights = weights)
    },
    predict = function(object, newdata, task, ...) {
      as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
                                type = "response"))
    }
  )
)
fit <- fit_resample(df, outcome = "outcome", splits = splits,
                    learner = "glm", custom_learners = custom,
                    metrics = "auc", refit = FALSE, seed = 1)
cal <- calibration_summary(fit, bins = 5)
cal$metrics

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{dlsi_robust}
\alias{dlsi_robust}
\title{Huber-robust delta_lsi point estimate from a LeakDeltaLSI}
\usage{
dlsi_robust(dlsi)
}
\arguments{
\item{dlsi}{A `LeakDeltaLSI` object returned by [delta_lsi()].}
}
\value{
A length-one numeric scalar.
}
\description{
Returns the Huber-robust point estimate of the per-repeat delta
values stored in a `LeakDeltaLSI` object.
}
\seealso{
[delta_lsi()], [dlsi_metric()], [dlsi_ci()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_audit.R
\name{plot_overlap_checks}
\alias{plot_overlap_checks}
\title{Plot overlap diagnostics between train/test groups}
\usage{
plot_overlap_checks(fit, column = NULL)
}
\arguments{
\item{fit}{A `LeakFit` object produced by [fit_resample()]. It must contain
the split indices and the associated metadata in `fit@splits@info$coldata`.
The metadata rows must align with the data used to create the splits.}

\item{column}{Character scalar naming the metadata column to check (for
example `"subject"` or `"batch"`). The function compares unique values of
this column between train and test within each resample. There is no
default: `NULL` or an unknown column triggers an error. Changing `column`
changes which kind of leakage (subject-level, batch-level, etc.) is tested
and therefore the overlap counts.}
}
\value{
A list returned invisibly with:
  \itemize{
    \item `overlap_counts`: data.frame with one row per resample and columns
      `fold` (resample index in `fit@splits@indices`), `overlap` (unique IDs
      shared by train and test), `train` (unique IDs in train), and `test`
      (unique IDs in test).
    \item `column`: the metadata column name used for the check.
    \item `plot`: the ggplot object showing the three count series across folds.
  }
  The plot is also printed. When any overlap is detected, the plot adds a
  warning annotation.
}
\description{
Checks whether the same group identifiers appear in both the training and
test partitions within each resample. This is designed to detect leakage
from grouped or repeated-measures data (for example, the same subject,
batch, plate, or study appearing on both sides of a fold) when group-wise
splitting is expected.
}
\details{
For each resample in `fit@splits@indices`, the function counts the number of
unique values of `column` in the train and test sets and the size of their
intersection. Any non-zero overlap indicates that at least one group appears
in both train and test for that resample. The check is metadata-based only:
it relies on exact matches of the supplied column and does not inspect
features or outcomes. It only checks train vs test within each resample, so
it will not detect overlaps across different resamples or other leakage
mechanisms. Inconsistent IDs or missing values in the metadata can hide or
inflate overlaps. `NA` values are treated as regular identifiers and will
count toward overlap if they appear in both partitions. Requires ggplot2.
}
\examples{
set.seed(1)
df <- data.frame(
  subject = rep(1:6, each = 2),
  outcome = rbinom(12, 1, 0.5),
  x1 = rnorm(12),
  x2 = rnorm(12)
)
splits <- make_split_plan(df, outcome = "outcome",
                      mode = "subject_grouped", group = "subject", v = 3)
custom <- list(
  glm = list(
    fit = function(x, y, task, weights, ...) {
      stats::glm(y ~ ., data = as.data.frame(x),
                 family = stats::binomial(), weights = weights)
    },
    predict = function(object, newdata, task, ...) {
      as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
                                type = "response"))
    }
  )
)
fit <- fit_resample(df, outcome = "outcome", splits = splits,
                    learner = "glm", custom_learners = custom,
                    metrics = "accuracy", refit = FALSE)
if (requireNamespace("ggplot2", quietly = TRUE)) {
  out <- plot_overlap_checks(fit, column = "subject")
  out$overlap_counts
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_resample.R
\name{tune_resample}
\alias{tune_resample}
\title{Leakage-aware nested tuning with tidymodels}
\usage{
tune_resample(
  x,
  outcome,
  splits,
  learner,
  preprocess = NULL,
  grid = 10,
  metrics = NULL,
  positive_class = NULL,
  selection = c("best", "one_std_err"),
  selection_metric = NULL,
  inner_v = NULL,
  inner_repeats = 1,
  inner_seed = NULL,
  control = NULL,
  parallel = FALSE,
  refit = FALSE,
  seed = 1,
  split_cols = "auto",
  tune_threshold = FALSE,
  threshold_grid = seq(0.1, 0.9, by = 0.05),
  threshold_metric = "accuracy"
)
}
\arguments{
\item{x}{SummarizedExperiment or matrix/data.frame.}

\item{outcome}{Outcome column name (if x is SE or data.frame).}

\item{splits}{LeakSplits object defining the outer resamples. If the splits
do not already include inner folds, they are created from each outer
training fold using the same split metadata. rsample splits must already
include inner folds.}

\item{learner}{A parsnip model_spec with tunable parameters, or a workflows
workflow. When a model_spec is provided, a workflow is built using
`preprocess` or a formula.}

\item{preprocess}{Optional `recipes::recipe`. Required when you need
preprocessing for tuning. Ignored when `learner` is already a workflow.
Recipe/workflow leakage guardrails run before tuning; configure policy via
\code{options(bioLeak.validation_mode = "warn" | "error" | "off")}.}

\item{grid}{Tuning grid passed to `tune::tune_grid()`. Can be a data.frame or
an integer size.}

\item{metrics}{Character vector of metric names (`auc`, `pr_auc`, `accuracy`,
`macro_f1`, `log_loss`, `rmse`) or a yardstick metric set/list. Metrics are
computed with yardstick; unsupported metrics are dropped with a warning.
For binomial tasks, if any inner assessment fold contains a single class,
probability metrics (`auc`, `roc_auc`, `pr_auc`) are dropped for tuning with
a warning.}

\item{positive_class}{Optional value indicating the positive class for
binomial outcomes. When set, the outcome levels are reordered so the
positive class is second.}

\item{selection}{Selection rule for tuning, either `"best"` or `"one_std_err"`.}

\item{selection_metric}{Metric name used for selecting hyperparameters.
Defaults to the first metric in `metrics`. If the chosen metric yields
no valid results, the first available metric is used with a warning.}

\item{inner_v}{Optional number of folds for inner CV when inner splits are
not precomputed. Defaults to the outer `v`.}

\item{inner_repeats}{Optional number of repeats for inner CV when inner
splits are not precomputed. Defaults to 1.}

\item{inner_seed}{Optional seed for inner split generation when inner splits
are not precomputed. Defaults to the outer split seed.}

\item{control}{Optional `tune::control_grid()` settings for tuning.}

\item{parallel}{Logical; passed to [fit_resample()] when evaluating outer
folds (single-fold, no refit).}

\item{refit}{Logical; if TRUE, refits a final tuned workflow on the full
dataset using aggregated hyperparameters across all outer folds (median
for numeric parameters, majority vote for categorical). This avoids
nested-CV leakage that would occur from selecting a single fold's params.}

\item{seed}{Integer seed for reproducibility.}

\item{split_cols}{Optional named list/character vector or `"auto"` (default)
overriding group/batch/study/time column names when `splits` is an rsample
object and its attributes are missing. `"auto"` falls back to common
metadata column names (e.g., `group`, `subject`, `batch`, `study`, `time`).
Supported names are `group`, `batch`, `study`, and `time`.}

\item{tune_threshold}{Logical; when `TRUE` for binomial tasks, selects a
probability threshold from inner-fold predictions and applies it only to the
corresponding outer-fold evaluation.}

\item{threshold_grid}{Numeric vector of thresholds in `[0, 1]` considered when
`tune_threshold = TRUE`.}

\item{threshold_metric}{Metric used to pick thresholds when
`tune_threshold = TRUE`. Supported values are `"accuracy"`,
`"balanced_accuracy"`, and `"f1"`, or a custom function with signature
`function(truth, pred_class, prob, threshold)`.}
}
\value{
A list of class `"LeakTune"` with components:
  \item{metrics}{Outer-fold metrics.}
  \item{metric_summary}{Mean/SD metrics across outer folds with columns
    \code{learner}, and \code{<metric>_mean} and \code{<metric>_sd} for
    each metric.}
  \item{best_params}{Best hyperparameters per outer fold.}
  \item{inner_results}{List of inner tuning results.}
  \item{outer_fits}{List of outer LeakFit objects.}
  \item{thresholds}{Per-fold threshold choices when threshold tuning is enabled.}
  \item{fold_status}{Outer-fold status log with stage, status, reason, and notes.}
  \item{final_model}{Optional final workflow fit when `refit = TRUE`.}
  \item{info}{Metadata about the tuning run.}
}
\description{
Runs nested cross-validation for hyperparameter tuning using leakage-aware
splits. Inner resamples are constructed from each outer training fold to
avoid information leakage during tuning. Requires tidymodels tuning
packages and a workflow or recipe-based preprocessing. Survival tasks are
not yet supported.
}
\examples{
\donttest{
  if (requireNamespace("tune", quietly = TRUE) &&
      requireNamespace("recipes", quietly = TRUE) &&
      requireNamespace("glmnet", quietly = TRUE) &&
      requireNamespace("rsample", quietly = TRUE) &&
      requireNamespace("workflows", quietly = TRUE) &&
      requireNamespace("yardstick", quietly = TRUE) &&
     requireNamespace("dials", quietly = TRUE)) {
    df <- data.frame(
      subject = rep(1:10, each = 2),
      outcome = factor(rep(c(0, 1), each = 10)),
      x1 = rnorm(20),
      x2 = rnorm(20)
    )
    splits <- make_split_plan(df, outcome = "outcome",
                         mode = "subject_grouped", group = "subject",
                         v = 3, nested = TRUE, stratify = TRUE)
    spec <- parsnip::logistic_reg(penalty = tune::tune(), mixture = 1) |>
      parsnip::set_engine("glmnet")
    rec <- recipes::recipe(outcome ~ x1 + x2, data = df)
    tuned <- tune_resample(df, outcome = "outcome", splits = splits,
                          learner = spec, preprocess = rec, grid = 5)
    tuned$metric_summary
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/guards.R
\name{guard_ensure_levels}
\alias{guard_ensure_levels}
\title{Ensure consistent categorical levels for guarded preprocessing}
\usage{
guard_ensure_levels(df, levels_map = NULL, dummy_prefix = "__dummy__")
}
\arguments{
\item{df}{data.frame to normalize factor levels.}

\item{levels_map}{optional named list of factor levels learned from training data.}

\item{dummy_prefix}{prefix used when adding a dummy level to single-level factors.}
}
\value{
List with elements \code{data} (data.frame) and \code{levels} (named list of levels).
}
\description{
Converts character/logical columns to factors and aligns factor levels with
a training-time \code{levels_map}. Adds a dummy level when a column has only
one observed level so that downstream one-hot encoding retains a column.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/benchmark_suite.R
\name{benchmark_leakage_suite}
\alias{benchmark_leakage_suite}
\title{Simulation benchmark matrix for leakage diagnostics}
\usage{
benchmark_leakage_suite(
  modalities = c("omics", "imaging_tabular", "ehr_tabular"),
  leakages = c("none", "subject_overlap", "batch_confounded", "peek_norm", "lookahead"),
  modes = c("subject_grouped", "batch_blocked", "time_series"),
  learner = c("glmnet", "ranger"),
  seeds = 1:5,
  B = 200,
  alpha = 0.05,
  parallel = FALSE
)
}
\arguments{
\item{modalities}{Character vector selecting predefined modality profiles.
Supported values: `"omics"`, `"imaging_tabular"`, `"ehr_tabular"`.}

\item{leakages}{Character vector of leakage mechanisms passed to
[simulate_leakage_suite()].}

\item{modes}{Character vector of split modes passed to [simulate_leakage_suite()].}

\item{learner}{Character scalar. `"glmnet"` (default) or `"ranger"`.}

\item{seeds}{Integer vector of Monte Carlo seeds.}

\item{B}{Integer scalar. Number of permutations per scenario.}

\item{alpha}{Numeric scalar in (0, 1). Detection threshold applied to
permutation p-values.}

\item{parallel}{Logical scalar. If TRUE, evaluates scenarios in parallel
when `future.apply` is available.}
}
\value{
A data.frame with one row per simulation seed/scenario and columns:
  `modality`, `leakage`, `mode`, `seed`, observed metric, gap, p-value, and
  a logical `detected` flag. A scenario-level summary is attached as
  `attr(x, "summary")`.
}
\description{
Runs a reproducible grid of simulation scenarios across modalities, leakage
mechanisms, and split modes using [simulate_leakage_suite()]. This function
is designed as a benchmarking harness to quantify detection rates and
performance inflation under controlled settings.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_audit.R
\name{plot_calibration}
\alias{plot_calibration}
\title{Plot calibration curve for binomial predictions}
\usage{
plot_calibration(fit, bins = 10, min_bin_n = 5, learner = NULL)
}
\arguments{
\item{fit}{LeakFit.}

\item{bins}{Number of probability bins to use.}

\item{min_bin_n}{Minimum samples per bin shown in the plot.}

\item{learner}{Optional character scalar. When predictions include multiple
learners, selects the learner to summarize.}
}
\value{
A list containing the calibration curve, metrics, and a ggplot object.
}
\description{
Visualizes observed outcome rates versus predicted probabilities across
bins to diagnose calibration (binomial tasks only). Requires ggplot2.
}
\examples{
if (requireNamespace("ggplot2", quietly = TRUE)) {
  set.seed(42)
  df <- data.frame(
    subject = rep(1:15, each = 2),
    outcome = factor(rep(c(0, 1), 15)),
    x1 = rnorm(30),
    x2 = rnorm(30)
  )
  splits <- make_split_plan(df, outcome = "outcome",
                            mode = "subject_grouped", group = "subject",
                            v = 3, progress = FALSE)
  custom <- list(
    glm = list(
      fit = function(x, y, task, weights, ...) {
        stats::glm(y ~ ., data = as.data.frame(x),
                   family = stats::binomial(), weights = weights)
      },
      predict = function(object, newdata, task, ...) {
        as.numeric(stats::predict(object, newdata = as.data.frame(newdata),
                                  type = "response"))
      }
    )
  )
  fit <- fit_resample(df, outcome = "outcome", splits = splits,
                      learner = "glm", custom_learners = custom,
                      metrics = "auc", refit = FALSE, seed = 1)
  plot_calibration(fit, bins = 5)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit.R
\name{audit_leakage_by_learner}
\alias{audit_leakage_by_learner}
\title{Audit leakage per learner}
\usage{
audit_leakage_by_learner(
  fit,
  metric = c("auc", "pr_auc", "accuracy", "macro_f1", "log_loss", "rmse", "cindex"),
  learners = NULL,
  parallel_learners = FALSE,
  mc.cores = NULL,
  ...
)
}
\arguments{
\item{fit}{A [LeakFit] object produced by [fit_resample()]. It must contain
predictions and split metadata. Learner IDs must be present in predictions
to audit multiple models.}

\item{metric}{Character scalar. One of `"auc"`, `"pr_auc"`, `"accuracy"`,
`"macro_f1"`, `"log_loss"`, `"rmse"`, or `"cindex"`. Controls which metric
is audited for each learner.}

\item{learners}{Character vector or NULL. If NULL (default), audits all
learners found in predictions. If provided, must match learner IDs stored in
the predictions. Supplying more than one learner requires learner IDs.}

\item{parallel_learners}{Logical scalar. If TRUE, runs per-learner audits in
parallel using `future.apply` (if installed). This changes runtime but not
the audit results.}

\item{mc.cores}{Integer scalar or NULL. Number of workers used when
`parallel_learners = TRUE`. Defaults to the minimum of available cores and
the number of learners.}

\item{...}{Additional named arguments forwarded to [audit_leakage()] for each
learner. These control the audit itself. Common options include:
`B` (integer permutations), `perm_stratify` (logical or `"auto"`),
`perm_refit` (logical), `perm_refit_spec` (list),
`time_block` (character), `block_len` (integer or NULL), `include_z`
(logical), `ci_method` (character), `boot_B` (integer), `parallel`
(logical), `seed` (integer), `return_perm` (logical), `batch_cols`
(character vector), `coldata` (data.frame), `X_ref` (matrix/data.frame),
`target_scan` (logical), `target_threshold` (numeric),
`target_p_adjust` (character), `target_alpha` (numeric),
`feature_space` (character), `sim_method` (character), `sim_threshold` (numeric),
`nn_k` (integer), `max_pairs` (integer), and `duplicate_scope` (character).
See [audit_leakage()] for full definitions; changing these values changes
each learner's audit.}
}
\value{
A named list of \code{\linkS4class{LeakAudit}} objects, where each
  element is keyed by the learner ID (character string). Each
  \code{LeakAudit} object contains the same slots as described in
  \code{\link{audit_leakage}}: \code{fit}, \code{permutation_gap},
  \code{perm_values}, \code{batch_assoc}, \code{target_assoc},
  \code{duplicates}, \code{trail}, and \code{info}. Use \code{names()} to
  retrieve learner IDs, and access individual audits with \code{[[learner_id]]}
  or \code{$learner_id}. Each audit reflects the performance and diagnostics
  for that specific learner's predictions.
}
\description{
Runs [audit_leakage()] separately for each learner recorded in a [LeakFit]
and returns a named list of [LeakAudit] objects. Use this when a single fit
contains predictions for multiple models and you want model-specific audits.
If predictions do not include learner IDs, only a single audit can be run and
requesting multiple learners is an error.
}
\examples{
set.seed(1)
df <- data.frame(
  subject = rep(1:6, each = 2),
  outcome = factor(rep(c(0, 1), 6)),
  x1 = rnorm(12),
  x2 = rnorm(12)
)
splits <- make_split_plan(df, outcome = "outcome",
                      mode = "subject_grouped", group = "subject",
                      v = 3, progress = FALSE)
custom <- list(
  glm = list(
    fit = function(x, y, task, weights, ...) {
      stats::glm(y ~ ., data = data.frame(y = y, x),
                 family = stats::binomial(), weights = weights)
    },
    predict = function(object, newdata, task, ...) {
      as.numeric(stats::predict(object,
                                newdata = as.data.frame(newdata),
                                type = "response"))
    }
  )
)
custom$glm2 <- custom$glm
fit <- fit_resample(df, outcome = "outcome", splits = splits,
                    learner = c("glm", "glm2"), custom_learners = custom,
                    metrics = "auc", refit = FALSE, seed = 1)
audits <- audit_leakage_by_learner(fit, metric = "auc", B = 10,
                                   perm_stratify = FALSE)
names(audits)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{dlsi_metric}
\alias{dlsi_metric}
\title{Raw delta-metric estimate from a LeakDeltaLSI}
\usage{
dlsi_metric(dlsi)
}
\arguments{
\item{dlsi}{A `LeakDeltaLSI` object returned by [delta_lsi()].}
}
\value{
A length-one numeric scalar.
}
\description{
Returns the arithmetic mean of the per-repeat raw metric differences
(leaky minus guarded) stored in a `LeakDeltaLSI` object.
}
\seealso{
[delta_lsi()], [dlsi_robust()], [dlsi_ci()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_resample.R
\name{print.LeakTune}
\alias{print.LeakTune}
\title{Print a LeakTune object}
\usage{
\method{print}{LeakTune}(x, ...)
}
\arguments{
\item{x}{A `LeakTune` object returned by [tune_resample()].}

\item{...}{Ignored; present so that the S3 signature matches
[base::print()].}
}
\value{
Invisibly returns `x`.
}
\description{
Brief one-screen auto-print representation of a `LeakTune` result
returned by [tune_resample()]. Use [summary()] for the full
diagnostic report (outer-loop metrics, selected hyperparameters,
fold-by-fold detail, and refit summary).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{audit_batch_assoc}
\alias{audit_batch_assoc}
\title{Batch / study association results from a LeakAudit}
\usage{
audit_batch_assoc(audit)
}
\arguments{
\item{audit}{A `LeakAudit` object returned by [audit_leakage()].}
}
\value{
A `data.frame` (possibly empty) with one row per
  (metadata column, repeat).
}
\description{
Returns the batch/study chi-squared association data frame stored in
a `LeakAudit` object. Columns include the metadata column, repeat,
chi-squared statistic, degrees of freedom, p-value, and Cramer's V.
}
\seealso{
[audit_leakage()], [audit_perm_gap()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{dlsi_repeats}
\alias{dlsi_repeats}
\title{Per-repeat metric data frames from a LeakDeltaLSI}
\usage{
dlsi_repeats(dlsi, which = c("naive", "guarded"))
}
\arguments{
\item{dlsi}{A `LeakDeltaLSI` object returned by [delta_lsi()].}

\item{which}{Either `"naive"` (default) for the naive/leaky
pipeline's per-repeat data frame, or `"guarded"` for the guarded
pipeline's per-repeat data frame.}
}
\value{
A `data.frame` with one row per repeat.
}
\description{
Returns the per-repeat metric data frame for one of the two
pipelines stored in a `LeakDeltaLSI` object. The naive (or leaky)
pipeline's repeats are returned by default; set
`which = "guarded"` to return the guarded pipeline's repeats.
}
\seealso{
[delta_lsi()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{show,LeakAudit-method}
\alias{show,LeakAudit-method}
\title{Display summary for LeakAudit objects}
\usage{
\S4method{show}{LeakAudit}(object)
}
\arguments{
\item{object}{A \code{LeakAudit} object.}
}
\value{
No return value, called for side effects (prints a brief summary
  to the console). Returns \code{object} invisibly.
}
\description{
Prints a brief one-screen summary of a LeakAudit, including
  task and outcome, the permutation-gap statistic, and counts of
  batch-association rows, target-leakage features, and duplicate pairs.
  Use \code{summary()} for the full diagnostic report.
}
\examples{
set.seed(1)
df <- data.frame(
  subject = rep(1:6, each = 2),
  outcome = rbinom(12, 1, 0.5),
  x1 = rnorm(12),
  x2 = rnorm(12)
)
splits <- make_split_plan(df, outcome = "outcome",
                      mode = "subject_grouped", group = "subject", v = 3,
                      progress = FALSE)
custom <- list(
  glm = list(
    fit = function(x, y, task, weights, ...) {
      stats::glm(y ~ ., data = as.data.frame(x),
                 family = stats::binomial(), weights = weights)
    },
    predict = function(object, newdata, task, ...) {
      as.numeric(stats::predict(object,
                                newdata = as.data.frame(newdata),
                                type = "response"))
    }
  )
)
fit <- fit_resample(df, outcome = "outcome", splits = splits,
                    learner = "glm", custom_learners = custom,
                    metrics = "auc", refit = FALSE, seed = 1)
aud <- audit_leakage(fit, metric = "auc", B = 10,
                     X_ref = df[, c("x1", "x2")])
show(aud)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/splitgraph_adapter.R
\name{as_leaksplits}
\alias{as_leaksplits}
\title{Convert a splitGraph split_spec into bioLeak splits}
\usage{
as_leaksplits(spec, data, outcome, sample_id_col = "sample_id", v = 5, ...)
}
\arguments{
\item{spec}{A \code{split_spec} object from \pkg{splitGraph}.}

\item{data}{A data.frame (or SummarizedExperiment) containing at least one
identifier column matching \code{sample_id_col} and an \code{outcome}
column.}

\item{outcome}{Name of the outcome column in \code{data}.}

\item{sample_id_col}{Name of the sample-id column in \code{data}
(default \code{"sample_id"}).}

\item{v}{Number of CV folds to request from \code{make_split_plan()}.}

\item{...}{Additional arguments forwarded to \code{\link{make_split_plan}}
(e.g. \code{stratify}, \code{seed}, \code{horizon}, \code{purge}).}
}
\value{
A \code{LeakSplits} object.
}
\description{
Consume a `split_spec` produced by \pkg{splitGraph} and build a
corresponding \code{LeakSplits} object via \code{\link{make_split_plan}}.
The spec supplies the grouping/blocking/ordering assignments; the caller
supplies the observation frame (features + outcome), joined on sample id.
}
\details{
The mapping from \code{spec$constraint_mode} to
\code{make_split_plan(mode=)} is:
\itemize{
  \item \code{"subject"}   -> \code{"subject_grouped"}
  \item \code{"batch"}     -> \code{"batch_blocked"}
  \item \code{"study"}     -> \code{"study_loocv"}
  \item \code{"time"}      -> \code{"time_series"}
  \item \code{"composite"} -> \code{"combined"}
}

Blocking variables declared on the spec (\code{batch_group},
\code{study_group}) and ordering (\code{order_rank}) are forwarded
automatically when relevant.
}
\seealso{
\code{\link{make_split_plan}}, \code{\link{as_rsample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{audit_target_assoc}
\alias{audit_target_assoc}
\title{Target leakage scan results from a LeakAudit}
\usage{
audit_target_assoc(audit)
}
\arguments{
\item{audit}{A `LeakAudit` object returned by [audit_leakage()].}
}
\value{
A `data.frame` (possibly empty) with one row per scanned
  feature.
}
\description{
Returns the per-feature target-association data frame stored in a
`LeakAudit` object. Columns include the feature name, association
value, score, and a logical flag indicating whether the feature
exceeded the configured `target_threshold`.
}
\seealso{
[audit_leakage()], [audit_perm_gap()]
}
